"""
Character Definition

Reference: `char.def` file.
"""
import regex
from collections import defaultdict


def get_emoji(text):
    import emoji

    _emojis = {}
    for lang in ["pt", "it", "es", "en"]:
        _emojis.update(emoji.UNICODE_EMOJI[lang])
    emoji_list = []
    flags = regex.findall("[\U0001F1E6-\U0001F1FF]", text)

    for grapheme in regex.findall(r"\X", text):
        if any(char in _emojis for char in grapheme):
            emoji_list.append(grapheme)

    return emoji_list + flags


categories = defaultdict(
    lambda: None,
    {
        0x20: "SPACE",
        0xD: "SPACE",
        0x9: "SPACE",
        0xB: "SPACE",
        0xA: "SPACE",
        0x21: "SYMBOL",
        0x22: "SYMBOL",
        0x23: "SYMBOL",
        0x24: "SYMBOL",
        0x25: "SYMBOL",
        0x26: "SYMBOL",
        0x27: "SYMBOL",
        0x28: "SYMBOL",
        0x29: "SYMBOL",
        0x2A: "SYMBOL",
        0x2B: "SYMBOL",
        0x2C: "SYMBOL",
        0x2D: "SYMBOL",
        0x2E: "SYMBOL",
        0x2F: "SYMBOL",
        0x30: "NUMERIC",
        0x31: "NUMERIC",
        0x32: "NUMERIC",
        0x33: "NUMERIC",
        0x34: "NUMERIC",
        0x35: "NUMERIC",
        0x36: "NUMERIC",
        0x37: "NUMERIC",
        0x38: "NUMERIC",
        0x39: "NUMERIC",
        0x3A: "SYMBOL",
        0x3B: "SYMBOL",
        0x3C: "SYMBOL",
        0x3D: "SYMBOL",
        0x3E: "SYMBOL",
        0x3F: "SYMBOL",
        0x40: "SYMBOL",
        0x41: "ALPHA",
        0x42: "ALPHA",
        0x43: "ALPHA",
        0x44: "ALPHA",
        0x45: "ALPHA",
        0x46: "ALPHA",
        0x47: "ALPHA",
        0x48: "ALPHA",
        0x49: "ALPHA",
        0x4A: "ALPHA",
        0x4B: "ALPHA",
        0x4C: "ALPHA",
        0x4D: "ALPHA",
        0x4E: "ALPHA",
        0x4F: "ALPHA",
        0x50: "ALPHA",
        0x51: "ALPHA",
        0x52: "ALPHA",
        0x53: "ALPHA",
        0x54: "ALPHA",
        0x55: "ALPHA",
        0x56: "ALPHA",
        0x57: "ALPHA",
        0x58: "ALPHA",
        0x59: "ALPHA",
        0x5A: "ALPHA",
        0x5B: "SYMBOL",
        0x5C: "SYMBOL",
        0x5D: "SYMBOL",
        0x5E: "SYMBOL",
        0x5F: "SYMBOL",
        0x60: "SYMBOL",
        0x61: "ALPHA",
        0x62: "ALPHA",
        0x63: "ALPHA",
        0x64: "ALPHA",
        0x65: "ALPHA",
        0x66: "ALPHA",
        0x67: "ALPHA",
        0x68: "ALPHA",
        0x69: "ALPHA",
        0x6A: "ALPHA",
        0x6B: "ALPHA",
        0x6C: "ALPHA",
        0x6D: "ALPHA",
        0x6E: "ALPHA",
        0x6F: "ALPHA",
        0x70: "ALPHA",
        0x71: "ALPHA",
        0x72: "ALPHA",
        0x73: "ALPHA",
        0x74: "ALPHA",
        0x75: "ALPHA",
        0x76: "ALPHA",
        0x77: "ALPHA",
        0x78: "ALPHA",
        0x79: "ALPHA",
        0x7A: "ALPHA",
        0x7B: "SYMBOL",
        0x7C: "SYMBOL",
        0x7D: "SYMBOL",
        0x7E: "SYMBOL",
        0xA1: "SYMBOL",
        0xA2: "SYMBOL",
        0xA3: "SYMBOL",
        0xA4: "SYMBOL",
        0xA5: "SYMBOL",
        0xA6: "SYMBOL",
        0xA7: "SYMBOL",
        0xA8: "SYMBOL",
        0xA9: "SYMBOL",
        0xAA: "SYMBOL",
        0xAB: "SYMBOL",
        0xAC: "SYMBOL",
        0xAD: "SYMBOL",
        0xAE: "SYMBOL",
        0xAF: "SYMBOL",
        0xB0: "SYMBOL",
        0xB1: "SYMBOL",
        0xB2: "SYMBOL",
        0xB3: "SYMBOL",
        0xB4: "SYMBOL",
        0xB5: "SYMBOL",
        0xB6: "SYMBOL",
        0xB7: "SYMBOL",
        0xB8: "SYMBOL",
        0xB9: "SYMBOL",
        0xBA: "SYMBOL",
        0xBB: "SYMBOL",
        0xBC: "SYMBOL",
        0xBD: "SYMBOL",
        0xBE: "SYMBOL",
        0xBF: "SYMBOL",
        0xC0: "ALPHA",
        0xC1: "ALPHA",
        0xC2: "ALPHA",
        0xC3: "ALPHA",
        0xC4: "ALPHA",
        0xC5: "ALPHA",
        0xC6: "ALPHA",
        0xC7: "ALPHA",
        0xC8: "ALPHA",
        0xC9: "ALPHA",
        0xCA: "ALPHA",
        0xCB: "ALPHA",
        0xCC: "ALPHA",
        0xCD: "ALPHA",
        0xCE: "ALPHA",
        0xCF: "ALPHA",
        0xD0: "ALPHA",
        0xD1: "ALPHA",
        0xD2: "ALPHA",
        0xD3: "ALPHA",
        0xD4: "ALPHA",
        0xD5: "ALPHA",
        0xD6: "ALPHA",
        0xD7: "ALPHA",
        0xD8: "ALPHA",
        0xD9: "ALPHA",
        0xDA: "ALPHA",
        0xDB: "ALPHA",
        0xDC: "ALPHA",
        0xDD: "ALPHA",
        0xDE: "ALPHA",
        0xDF: "ALPHA",
        0xE0: "ALPHA",
        0xE1: "ALPHA",
        0xE2: "ALPHA",
        0xE3: "ALPHA",
        0xE4: "ALPHA",
        0xE5: "ALPHA",
        0xE6: "ALPHA",
        0xE7: "ALPHA",
        0xE8: "ALPHA",
        0xE9: "ALPHA",
        0xEA: "ALPHA",
        0xEB: "ALPHA",
        0xEC: "ALPHA",
        0xED: "ALPHA",
        0xEE: "ALPHA",
        0xEF: "ALPHA",
        0xF0: "ALPHA",
        0xF1: "ALPHA",
        0xF2: "ALPHA",
        0xF3: "ALPHA",
        0xF4: "ALPHA",
        0xF5: "ALPHA",
        0xF6: "ALPHA",
        0xF7: "ALPHA",
        0xF8: "ALPHA",
        0xF9: "ALPHA",
        0xFA: "ALPHA",
        0xFB: "ALPHA",
        0xFC: "ALPHA",
        0xFD: "ALPHA",
        0xFE: "ALPHA",
        0xFF: "ALPHA",
        0x100: "ALPHA",
        0x101: "ALPHA",
        0x102: "ALPHA",
        0x103: "ALPHA",
        0x104: "ALPHA",
        0x105: "ALPHA",
        0x106: "ALPHA",
        0x107: "ALPHA",
        0x108: "ALPHA",
        0x109: "ALPHA",
        0x10A: "ALPHA",
        0x10B: "ALPHA",
        0x10C: "ALPHA",
        0x10D: "ALPHA",
        0x10E: "ALPHA",
        0x10F: "ALPHA",
        0x110: "ALPHA",
        0x111: "ALPHA",
        0x112: "ALPHA",
        0x113: "ALPHA",
        0x114: "ALPHA",
        0x115: "ALPHA",
        0x116: "ALPHA",
        0x117: "ALPHA",
        0x118: "ALPHA",
        0x119: "ALPHA",
        0x11A: "ALPHA",
        0x11B: "ALPHA",
        0x11C: "ALPHA",
        0x11D: "ALPHA",
        0x11E: "ALPHA",
        0x11F: "ALPHA",
        0x120: "ALPHA",
        0x121: "ALPHA",
        0x122: "ALPHA",
        0x123: "ALPHA",
        0x124: "ALPHA",
        0x125: "ALPHA",
        0x126: "ALPHA",
        0x127: "ALPHA",
        0x128: "ALPHA",
        0x129: "ALPHA",
        0x12A: "ALPHA",
        0x12B: "ALPHA",
        0x12C: "ALPHA",
        0x12D: "ALPHA",
        0x12E: "ALPHA",
        0x12F: "ALPHA",
        0x130: "ALPHA",
        0x131: "ALPHA",
        0x132: "ALPHA",
        0x133: "ALPHA",
        0x134: "ALPHA",
        0x135: "ALPHA",
        0x136: "ALPHA",
        0x137: "ALPHA",
        0x138: "ALPHA",
        0x139: "ALPHA",
        0x13A: "ALPHA",
        0x13B: "ALPHA",
        0x13C: "ALPHA",
        0x13D: "ALPHA",
        0x13E: "ALPHA",
        0x13F: "ALPHA",
        0x140: "ALPHA",
        0x141: "ALPHA",
        0x142: "ALPHA",
        0x143: "ALPHA",
        0x144: "ALPHA",
        0x145: "ALPHA",
        0x146: "ALPHA",
        0x147: "ALPHA",
        0x148: "ALPHA",
        0x149: "ALPHA",
        0x14A: "ALPHA",
        0x14B: "ALPHA",
        0x14C: "ALPHA",
        0x14D: "ALPHA",
        0x14E: "ALPHA",
        0x14F: "ALPHA",
        0x150: "ALPHA",
        0x151: "ALPHA",
        0x152: "ALPHA",
        0x153: "ALPHA",
        0x154: "ALPHA",
        0x155: "ALPHA",
        0x156: "ALPHA",
        0x157: "ALPHA",
        0x158: "ALPHA",
        0x159: "ALPHA",
        0x15A: "ALPHA",
        0x15B: "ALPHA",
        0x15C: "ALPHA",
        0x15D: "ALPHA",
        0x15E: "ALPHA",
        0x15F: "ALPHA",
        0x160: "ALPHA",
        0x161: "ALPHA",
        0x162: "ALPHA",
        0x163: "ALPHA",
        0x164: "ALPHA",
        0x165: "ALPHA",
        0x166: "ALPHA",
        0x167: "ALPHA",
        0x168: "ALPHA",
        0x169: "ALPHA",
        0x16A: "ALPHA",
        0x16B: "ALPHA",
        0x16C: "ALPHA",
        0x16D: "ALPHA",
        0x16E: "ALPHA",
        0x16F: "ALPHA",
        0x170: "ALPHA",
        0x171: "ALPHA",
        0x172: "ALPHA",
        0x173: "ALPHA",
        0x174: "ALPHA",
        0x175: "ALPHA",
        0x176: "ALPHA",
        0x177: "ALPHA",
        0x178: "ALPHA",
        0x179: "ALPHA",
        0x17A: "ALPHA",
        0x17B: "ALPHA",
        0x17C: "ALPHA",
        0x17D: "ALPHA",
        0x17E: "ALPHA",
        0x17F: "ALPHA",
        0x180: "ALPHA",
        0x181: "ALPHA",
        0x182: "ALPHA",
        0x183: "ALPHA",
        0x184: "ALPHA",
        0x185: "ALPHA",
        0x186: "ALPHA",
        0x187: "ALPHA",
        0x188: "ALPHA",
        0x189: "ALPHA",
        0x18A: "ALPHA",
        0x18B: "ALPHA",
        0x18C: "ALPHA",
        0x18D: "ALPHA",
        0x18E: "ALPHA",
        0x18F: "ALPHA",
        0x190: "ALPHA",
        0x191: "ALPHA",
        0x192: "ALPHA",
        0x193: "ALPHA",
        0x194: "ALPHA",
        0x195: "ALPHA",
        0x196: "ALPHA",
        0x197: "ALPHA",
        0x198: "ALPHA",
        0x199: "ALPHA",
        0x19A: "ALPHA",
        0x19B: "ALPHA",
        0x19C: "ALPHA",
        0x19D: "ALPHA",
        0x19E: "ALPHA",
        0x19F: "ALPHA",
        0x1A0: "ALPHA",
        0x1A1: "ALPHA",
        0x1A2: "ALPHA",
        0x1A3: "ALPHA",
        0x1A4: "ALPHA",
        0x1A5: "ALPHA",
        0x1A6: "ALPHA",
        0x1A7: "ALPHA",
        0x1A8: "ALPHA",
        0x1A9: "ALPHA",
        0x1AA: "ALPHA",
        0x1AB: "ALPHA",
        0x1AC: "ALPHA",
        0x1AD: "ALPHA",
        0x1AE: "ALPHA",
        0x1AF: "ALPHA",
        0x1B0: "ALPHA",
        0x1B1: "ALPHA",
        0x1B2: "ALPHA",
        0x1B3: "ALPHA",
        0x1B4: "ALPHA",
        0x1B5: "ALPHA",
        0x1B6: "ALPHA",
        0x1B7: "ALPHA",
        0x1B8: "ALPHA",
        0x1B9: "ALPHA",
        0x1BA: "ALPHA",
        0x1BB: "ALPHA",
        0x1BC: "ALPHA",
        0x1BD: "ALPHA",
        0x1BE: "ALPHA",
        0x1BF: "ALPHA",
        0x1C0: "ALPHA",
        0x1C1: "ALPHA",
        0x1C2: "ALPHA",
        0x1C3: "ALPHA",
        0x1C4: "ALPHA",
        0x1C5: "ALPHA",
        0x1C6: "ALPHA",
        0x1C7: "ALPHA",
        0x1C8: "ALPHA",
        0x1C9: "ALPHA",
        0x1CA: "ALPHA",
        0x1CB: "ALPHA",
        0x1CC: "ALPHA",
        0x1CD: "ALPHA",
        0x1CE: "ALPHA",
        0x1CF: "ALPHA",
        0x1D0: "ALPHA",
        0x1D1: "ALPHA",
        0x1D2: "ALPHA",
        0x1D3: "ALPHA",
        0x1D4: "ALPHA",
        0x1D5: "ALPHA",
        0x1D6: "ALPHA",
        0x1D7: "ALPHA",
        0x1D8: "ALPHA",
        0x1D9: "ALPHA",
        0x1DA: "ALPHA",
        0x1DB: "ALPHA",
        0x1DC: "ALPHA",
        0x1DD: "ALPHA",
        0x1DE: "ALPHA",
        0x1DF: "ALPHA",
        0x1E0: "ALPHA",
        0x1E1: "ALPHA",
        0x1E2: "ALPHA",
        0x1E3: "ALPHA",
        0x1E4: "ALPHA",
        0x1E5: "ALPHA",
        0x1E6: "ALPHA",
        0x1E7: "ALPHA",
        0x1E8: "ALPHA",
        0x1E9: "ALPHA",
        0x1EA: "ALPHA",
        0x1EB: "ALPHA",
        0x1EC: "ALPHA",
        0x1ED: "ALPHA",
        0x1EE: "ALPHA",
        0x1EF: "ALPHA",
        0x1F0: "ALPHA",
        0x1F1: "ALPHA",
        0x1F2: "ALPHA",
        0x1F3: "ALPHA",
        0x1F4: "ALPHA",
        0x1F5: "ALPHA",
        0x1F6: "ALPHA",
        0x1F7: "ALPHA",
        0x1F8: "ALPHA",
        0x1F9: "ALPHA",
        0x1FA: "ALPHA",
        0x1FB: "ALPHA",
        0x1FC: "ALPHA",
        0x1FD: "ALPHA",
        0x1FE: "ALPHA",
        0x1FF: "ALPHA",
        0x200: "ALPHA",
        0x201: "ALPHA",
        0x202: "ALPHA",
        0x203: "ALPHA",
        0x204: "ALPHA",
        0x205: "ALPHA",
        0x206: "ALPHA",
        0x207: "ALPHA",
        0x208: "ALPHA",
        0x209: "ALPHA",
        0x20A: "ALPHA",
        0x20B: "ALPHA",
        0x20C: "ALPHA",
        0x20D: "ALPHA",
        0x20E: "ALPHA",
        0x20F: "ALPHA",
        0x210: "ALPHA",
        0x211: "ALPHA",
        0x212: "ALPHA",
        0x213: "ALPHA",
        0x214: "ALPHA",
        0x215: "ALPHA",
        0x216: "ALPHA",
        0x217: "ALPHA",
        0x218: "ALPHA",
        0x219: "ALPHA",
        0x21A: "ALPHA",
        0x21B: "ALPHA",
        0x21C: "ALPHA",
        0x21D: "ALPHA",
        0x21E: "ALPHA",
        0x21F: "ALPHA",
        0x220: "ALPHA",
        0x221: "ALPHA",
        0x222: "ALPHA",
        0x223: "ALPHA",
        0x224: "ALPHA",
        0x225: "ALPHA",
        0x226: "ALPHA",
        0x227: "ALPHA",
        0x228: "ALPHA",
        0x229: "ALPHA",
        0x22A: "ALPHA",
        0x22B: "ALPHA",
        0x22C: "ALPHA",
        0x22D: "ALPHA",
        0x22E: "ALPHA",
        0x22F: "ALPHA",
        0x230: "ALPHA",
        0x231: "ALPHA",
        0x232: "ALPHA",
        0x233: "ALPHA",
        0x234: "ALPHA",
        0x235: "ALPHA",
        0x236: "ALPHA",
        0x1E00: "ALPHA",
        0x1E01: "ALPHA",
        0x1E02: "ALPHA",
        0x1E03: "ALPHA",
        0x1E04: "ALPHA",
        0x1E05: "ALPHA",
        0x1E06: "ALPHA",
        0x1E07: "ALPHA",
        0x1E08: "ALPHA",
        0x1E09: "ALPHA",
        0x1E0A: "ALPHA",
        0x1E0B: "ALPHA",
        0x1E0C: "ALPHA",
        0x1E0D: "ALPHA",
        0x1E0E: "ALPHA",
        0x1E0F: "ALPHA",
        0x1E10: "ALPHA",
        0x1E11: "ALPHA",
        0x1E12: "ALPHA",
        0x1E13: "ALPHA",
        0x1E14: "ALPHA",
        0x1E15: "ALPHA",
        0x1E16: "ALPHA",
        0x1E17: "ALPHA",
        0x1E18: "ALPHA",
        0x1E19: "ALPHA",
        0x1E1A: "ALPHA",
        0x1E1B: "ALPHA",
        0x1E1C: "ALPHA",
        0x1E1D: "ALPHA",
        0x1E1E: "ALPHA",
        0x1E1F: "ALPHA",
        0x1E20: "ALPHA",
        0x1E21: "ALPHA",
        0x1E22: "ALPHA",
        0x1E23: "ALPHA",
        0x1E24: "ALPHA",
        0x1E25: "ALPHA",
        0x1E26: "ALPHA",
        0x1E27: "ALPHA",
        0x1E28: "ALPHA",
        0x1E29: "ALPHA",
        0x1E2A: "ALPHA",
        0x1E2B: "ALPHA",
        0x1E2C: "ALPHA",
        0x1E2D: "ALPHA",
        0x1E2E: "ALPHA",
        0x1E2F: "ALPHA",
        0x1E30: "ALPHA",
        0x1E31: "ALPHA",
        0x1E32: "ALPHA",
        0x1E33: "ALPHA",
        0x1E34: "ALPHA",
        0x1E35: "ALPHA",
        0x1E36: "ALPHA",
        0x1E37: "ALPHA",
        0x1E38: "ALPHA",
        0x1E39: "ALPHA",
        0x1E3A: "ALPHA",
        0x1E3B: "ALPHA",
        0x1E3C: "ALPHA",
        0x1E3D: "ALPHA",
        0x1E3E: "ALPHA",
        0x1E3F: "ALPHA",
        0x1E40: "ALPHA",
        0x1E41: "ALPHA",
        0x1E42: "ALPHA",
        0x1E43: "ALPHA",
        0x1E44: "ALPHA",
        0x1E45: "ALPHA",
        0x1E46: "ALPHA",
        0x1E47: "ALPHA",
        0x1E48: "ALPHA",
        0x1E49: "ALPHA",
        0x1E4A: "ALPHA",
        0x1E4B: "ALPHA",
        0x1E4C: "ALPHA",
        0x1E4D: "ALPHA",
        0x1E4E: "ALPHA",
        0x1E4F: "ALPHA",
        0x1E50: "ALPHA",
        0x1E51: "ALPHA",
        0x1E52: "ALPHA",
        0x1E53: "ALPHA",
        0x1E54: "ALPHA",
        0x1E55: "ALPHA",
        0x1E56: "ALPHA",
        0x1E57: "ALPHA",
        0x1E58: "ALPHA",
        0x1E59: "ALPHA",
        0x1E5A: "ALPHA",
        0x1E5B: "ALPHA",
        0x1E5C: "ALPHA",
        0x1E5D: "ALPHA",
        0x1E5E: "ALPHA",
        0x1E5F: "ALPHA",
        0x1E60: "ALPHA",
        0x1E61: "ALPHA",
        0x1E62: "ALPHA",
        0x1E63: "ALPHA",
        0x1E64: "ALPHA",
        0x1E65: "ALPHA",
        0x1E66: "ALPHA",
        0x1E67: "ALPHA",
        0x1E68: "ALPHA",
        0x1E69: "ALPHA",
        0x1E6A: "ALPHA",
        0x1E6B: "ALPHA",
        0x1E6C: "ALPHA",
        0x1E6D: "ALPHA",
        0x1E6E: "ALPHA",
        0x1E6F: "ALPHA",
        0x1E70: "ALPHA",
        0x1E71: "ALPHA",
        0x1E72: "ALPHA",
        0x1E73: "ALPHA",
        0x1E74: "ALPHA",
        0x1E75: "ALPHA",
        0x1E76: "ALPHA",
        0x1E77: "ALPHA",
        0x1E78: "ALPHA",
        0x1E79: "ALPHA",
        0x1E7A: "ALPHA",
        0x1E7B: "ALPHA",
        0x1E7C: "ALPHA",
        0x1E7D: "ALPHA",
        0x1E7E: "ALPHA",
        0x1E7F: "ALPHA",
        0x1E80: "ALPHA",
        0x1E81: "ALPHA",
        0x1E82: "ALPHA",
        0x1E83: "ALPHA",
        0x1E84: "ALPHA",
        0x1E85: "ALPHA",
        0x1E86: "ALPHA",
        0x1E87: "ALPHA",
        0x1E88: "ALPHA",
        0x1E89: "ALPHA",
        0x1E8A: "ALPHA",
        0x1E8B: "ALPHA",
        0x1E8C: "ALPHA",
        0x1E8D: "ALPHA",
        0x1E8E: "ALPHA",
        0x1E8F: "ALPHA",
        0x1E90: "ALPHA",
        0x1E91: "ALPHA",
        0x1E92: "ALPHA",
        0x1E93: "ALPHA",
        0x1E94: "ALPHA",
        0x1E95: "ALPHA",
        0x1E96: "ALPHA",
        0x1E97: "ALPHA",
        0x1E98: "ALPHA",
        0x1E99: "ALPHA",
        0x1E9A: "ALPHA",
        0x1E9B: "ALPHA",
        0x1E9C: "ALPHA",
        0x1E9D: "ALPHA",
        0x1E9E: "ALPHA",
        0x1E9F: "ALPHA",
        0x1EA0: "ALPHA",
        0x1EA1: "ALPHA",
        0x1EA2: "ALPHA",
        0x1EA3: "ALPHA",
        0x1EA4: "ALPHA",
        0x1EA5: "ALPHA",
        0x1EA6: "ALPHA",
        0x1EA7: "ALPHA",
        0x1EA8: "ALPHA",
        0x1EA9: "ALPHA",
        0x1EAA: "ALPHA",
        0x1EAB: "ALPHA",
        0x1EAC: "ALPHA",
        0x1EAD: "ALPHA",
        0x1EAE: "ALPHA",
        0x1EAF: "ALPHA",
        0x1EB0: "ALPHA",
        0x1EB1: "ALPHA",
        0x1EB2: "ALPHA",
        0x1EB3: "ALPHA",
        0x1EB4: "ALPHA",
        0x1EB5: "ALPHA",
        0x1EB6: "ALPHA",
        0x1EB7: "ALPHA",
        0x1EB8: "ALPHA",
        0x1EB9: "ALPHA",
        0x1EBA: "ALPHA",
        0x1EBB: "ALPHA",
        0x1EBC: "ALPHA",
        0x1EBD: "ALPHA",
        0x1EBE: "ALPHA",
        0x1EBF: "ALPHA",
        0x1EC0: "ALPHA",
        0x1EC1: "ALPHA",
        0x1EC2: "ALPHA",
        0x1EC3: "ALPHA",
        0x1EC4: "ALPHA",
        0x1EC5: "ALPHA",
        0x1EC6: "ALPHA",
        0x1EC7: "ALPHA",
        0x1EC8: "ALPHA",
        0x1EC9: "ALPHA",
        0x1ECA: "ALPHA",
        0x1ECB: "ALPHA",
        0x1ECC: "ALPHA",
        0x1ECD: "ALPHA",
        0x1ECE: "ALPHA",
        0x1ECF: "ALPHA",
        0x1ED0: "ALPHA",
        0x1ED1: "ALPHA",
        0x1ED2: "ALPHA",
        0x1ED3: "ALPHA",
        0x1ED4: "ALPHA",
        0x1ED5: "ALPHA",
        0x1ED6: "ALPHA",
        0x1ED7: "ALPHA",
        0x1ED8: "ALPHA",
        0x1ED9: "ALPHA",
        0x1EDA: "ALPHA",
        0x1EDB: "ALPHA",
        0x1EDC: "ALPHA",
        0x1EDD: "ALPHA",
        0x1EDE: "ALPHA",
        0x1EDF: "ALPHA",
        0x1EE0: "ALPHA",
        0x1EE1: "ALPHA",
        0x1EE2: "ALPHA",
        0x1EE3: "ALPHA",
        0x1EE4: "ALPHA",
        0x1EE5: "ALPHA",
        0x1EE6: "ALPHA",
        0x1EE7: "ALPHA",
        0x1EE8: "ALPHA",
        0x1EE9: "ALPHA",
        0x1EEA: "ALPHA",
        0x1EEB: "ALPHA",
        0x1EEC: "ALPHA",
        0x1EED: "ALPHA",
        0x1EEE: "ALPHA",
        0x1EEF: "ALPHA",
        0x1EF0: "ALPHA",
        0x1EF1: "ALPHA",
        0x1EF2: "ALPHA",
        0x1EF3: "ALPHA",
        0x1EF4: "ALPHA",
        0x1EF5: "ALPHA",
        0x1EF6: "ALPHA",
        0x1EF7: "ALPHA",
        0x1EF8: "ALPHA",
        0x1EF9: "ALPHA",
        0x400: "CYRILLIC",
        0x401: "CYRILLIC",
        0x402: "CYRILLIC",
        0x403: "CYRILLIC",
        0x404: "CYRILLIC",
        0x405: "CYRILLIC",
        0x406: "CYRILLIC",
        0x407: "CYRILLIC",
        0x408: "CYRILLIC",
        0x409: "CYRILLIC",
        0x40A: "CYRILLIC",
        0x40B: "CYRILLIC",
        0x40C: "CYRILLIC",
        0x40D: "CYRILLIC",
        0x40E: "CYRILLIC",
        0x40F: "CYRILLIC",
        0x410: "CYRILLIC",
        0x411: "CYRILLIC",
        0x412: "CYRILLIC",
        0x413: "CYRILLIC",
        0x414: "CYRILLIC",
        0x415: "CYRILLIC",
        0x416: "CYRILLIC",
        0x417: "CYRILLIC",
        0x418: "CYRILLIC",
        0x419: "CYRILLIC",
        0x41A: "CYRILLIC",
        0x41B: "CYRILLIC",
        0x41C: "CYRILLIC",
        0x41D: "CYRILLIC",
        0x41E: "CYRILLIC",
        0x41F: "CYRILLIC",
        0x420: "CYRILLIC",
        0x421: "CYRILLIC",
        0x422: "CYRILLIC",
        0x423: "CYRILLIC",
        0x424: "CYRILLIC",
        0x425: "CYRILLIC",
        0x426: "CYRILLIC",
        0x427: "CYRILLIC",
        0x428: "CYRILLIC",
        0x429: "CYRILLIC",
        0x42A: "CYRILLIC",
        0x42B: "CYRILLIC",
        0x42C: "CYRILLIC",
        0x42D: "CYRILLIC",
        0x42E: "CYRILLIC",
        0x42F: "CYRILLIC",
        0x430: "CYRILLIC",
        0x431: "CYRILLIC",
        0x432: "CYRILLIC",
        0x433: "CYRILLIC",
        0x434: "CYRILLIC",
        0x435: "CYRILLIC",
        0x436: "CYRILLIC",
        0x437: "CYRILLIC",
        0x438: "CYRILLIC",
        0x439: "CYRILLIC",
        0x43A: "CYRILLIC",
        0x43B: "CYRILLIC",
        0x43C: "CYRILLIC",
        0x43D: "CYRILLIC",
        0x43E: "CYRILLIC",
        0x43F: "CYRILLIC",
        0x440: "CYRILLIC",
        0x441: "CYRILLIC",
        0x442: "CYRILLIC",
        0x443: "CYRILLIC",
        0x444: "CYRILLIC",
        0x445: "CYRILLIC",
        0x446: "CYRILLIC",
        0x447: "CYRILLIC",
        0x448: "CYRILLIC",
        0x449: "CYRILLIC",
        0x44A: "CYRILLIC",
        0x44B: "CYRILLIC",
        0x44C: "CYRILLIC",
        0x44D: "CYRILLIC",
        0x44E: "CYRILLIC",
        0x44F: "CYRILLIC",
        0x450: "CYRILLIC",
        0x451: "CYRILLIC",
        0x452: "CYRILLIC",
        0x453: "CYRILLIC",
        0x454: "CYRILLIC",
        0x455: "CYRILLIC",
        0x456: "CYRILLIC",
        0x457: "CYRILLIC",
        0x458: "CYRILLIC",
        0x459: "CYRILLIC",
        0x45A: "CYRILLIC",
        0x45B: "CYRILLIC",
        0x45C: "CYRILLIC",
        0x45D: "CYRILLIC",
        0x45E: "CYRILLIC",
        0x45F: "CYRILLIC",
        0x460: "CYRILLIC",
        0x461: "CYRILLIC",
        0x462: "CYRILLIC",
        0x463: "CYRILLIC",
        0x464: "CYRILLIC",
        0x465: "CYRILLIC",
        0x466: "CYRILLIC",
        0x467: "CYRILLIC",
        0x468: "CYRILLIC",
        0x469: "CYRILLIC",
        0x46A: "CYRILLIC",
        0x46B: "CYRILLIC",
        0x46C: "CYRILLIC",
        0x46D: "CYRILLIC",
        0x46E: "CYRILLIC",
        0x46F: "CYRILLIC",
        0x470: "CYRILLIC",
        0x471: "CYRILLIC",
        0x472: "CYRILLIC",
        0x473: "CYRILLIC",
        0x474: "CYRILLIC",
        0x475: "CYRILLIC",
        0x476: "CYRILLIC",
        0x477: "CYRILLIC",
        0x478: "CYRILLIC",
        0x479: "CYRILLIC",
        0x47A: "CYRILLIC",
        0x47B: "CYRILLIC",
        0x47C: "CYRILLIC",
        0x47D: "CYRILLIC",
        0x47E: "CYRILLIC",
        0x47F: "CYRILLIC",
        0x480: "CYRILLIC",
        0x481: "CYRILLIC",
        0x482: "CYRILLIC",
        0x483: "CYRILLIC",
        0x484: "CYRILLIC",
        0x485: "CYRILLIC",
        0x486: "CYRILLIC",
        0x487: "CYRILLIC",
        0x488: "CYRILLIC",
        0x489: "CYRILLIC",
        0x48A: "CYRILLIC",
        0x48B: "CYRILLIC",
        0x48C: "CYRILLIC",
        0x48D: "CYRILLIC",
        0x48E: "CYRILLIC",
        0x48F: "CYRILLIC",
        0x490: "CYRILLIC",
        0x491: "CYRILLIC",
        0x492: "CYRILLIC",
        0x493: "CYRILLIC",
        0x494: "CYRILLIC",
        0x495: "CYRILLIC",
        0x496: "CYRILLIC",
        0x497: "CYRILLIC",
        0x498: "CYRILLIC",
        0x499: "CYRILLIC",
        0x49A: "CYRILLIC",
        0x49B: "CYRILLIC",
        0x49C: "CYRILLIC",
        0x49D: "CYRILLIC",
        0x49E: "CYRILLIC",
        0x49F: "CYRILLIC",
        0x4A0: "CYRILLIC",
        0x4A1: "CYRILLIC",
        0x4A2: "CYRILLIC",
        0x4A3: "CYRILLIC",
        0x4A4: "CYRILLIC",
        0x4A5: "CYRILLIC",
        0x4A6: "CYRILLIC",
        0x4A7: "CYRILLIC",
        0x4A8: "CYRILLIC",
        0x4A9: "CYRILLIC",
        0x4AA: "CYRILLIC",
        0x4AB: "CYRILLIC",
        0x4AC: "CYRILLIC",
        0x4AD: "CYRILLIC",
        0x4AE: "CYRILLIC",
        0x4AF: "CYRILLIC",
        0x4B0: "CYRILLIC",
        0x4B1: "CYRILLIC",
        0x4B2: "CYRILLIC",
        0x4B3: "CYRILLIC",
        0x4B4: "CYRILLIC",
        0x4B5: "CYRILLIC",
        0x4B6: "CYRILLIC",
        0x4B7: "CYRILLIC",
        0x4B8: "CYRILLIC",
        0x4B9: "CYRILLIC",
        0x4BA: "CYRILLIC",
        0x4BB: "CYRILLIC",
        0x4BC: "CYRILLIC",
        0x4BD: "CYRILLIC",
        0x4BE: "CYRILLIC",
        0x4BF: "CYRILLIC",
        0x4C0: "CYRILLIC",
        0x4C1: "CYRILLIC",
        0x4C2: "CYRILLIC",
        0x4C3: "CYRILLIC",
        0x4C4: "CYRILLIC",
        0x4C5: "CYRILLIC",
        0x4C6: "CYRILLIC",
        0x4C7: "CYRILLIC",
        0x4C8: "CYRILLIC",
        0x4C9: "CYRILLIC",
        0x4CA: "CYRILLIC",
        0x4CB: "CYRILLIC",
        0x4CC: "CYRILLIC",
        0x4CD: "CYRILLIC",
        0x4CE: "CYRILLIC",
        0x4CF: "CYRILLIC",
        0x4D0: "CYRILLIC",
        0x4D1: "CYRILLIC",
        0x4D2: "CYRILLIC",
        0x4D3: "CYRILLIC",
        0x4D4: "CYRILLIC",
        0x4D5: "CYRILLIC",
        0x4D6: "CYRILLIC",
        0x4D7: "CYRILLIC",
        0x4D8: "CYRILLIC",
        0x4D9: "CYRILLIC",
        0x4DA: "CYRILLIC",
        0x4DB: "CYRILLIC",
        0x4DC: "CYRILLIC",
        0x4DD: "CYRILLIC",
        0x4DE: "CYRILLIC",
        0x4DF: "CYRILLIC",
        0x4E0: "CYRILLIC",
        0x4E1: "CYRILLIC",
        0x4E2: "CYRILLIC",
        0x4E3: "CYRILLIC",
        0x4E4: "CYRILLIC",
        0x4E5: "CYRILLIC",
        0x4E6: "CYRILLIC",
        0x4E7: "CYRILLIC",
        0x4E8: "CYRILLIC",
        0x4E9: "CYRILLIC",
        0x4EA: "CYRILLIC",
        0x4EB: "CYRILLIC",
        0x4EC: "CYRILLIC",
        0x4ED: "CYRILLIC",
        0x4EE: "CYRILLIC",
        0x4EF: "CYRILLIC",
        0x4F0: "CYRILLIC",
        0x4F1: "CYRILLIC",
        0x4F2: "CYRILLIC",
        0x4F3: "CYRILLIC",
        0x4F4: "CYRILLIC",
        0x4F5: "CYRILLIC",
        0x4F6: "CYRILLIC",
        0x4F7: "CYRILLIC",
        0x4F8: "CYRILLIC",
        0x4F9: "CYRILLIC",
        0x500: "CYRILLIC",
        0x501: "CYRILLIC",
        0x502: "CYRILLIC",
        0x503: "CYRILLIC",
        0x504: "CYRILLIC",
        0x505: "CYRILLIC",
        0x506: "CYRILLIC",
        0x507: "CYRILLIC",
        0x508: "CYRILLIC",
        0x509: "CYRILLIC",
        0x50A: "CYRILLIC",
        0x50B: "CYRILLIC",
        0x50C: "CYRILLIC",
        0x50D: "CYRILLIC",
        0x50E: "CYRILLIC",
        0x50F: "CYRILLIC",
        0x374: "GREEK",
        0x375: "GREEK",
        0x376: "GREEK",
        0x377: "GREEK",
        0x378: "GREEK",
        0x379: "GREEK",
        0x37A: "GREEK",
        0x37B: "GREEK",
        0x37C: "GREEK",
        0x37D: "GREEK",
        0x37E: "GREEK",
        0x37F: "GREEK",
        0x380: "GREEK",
        0x381: "GREEK",
        0x382: "GREEK",
        0x383: "GREEK",
        0x384: "GREEK",
        0x385: "GREEK",
        0x386: "GREEK",
        0x387: "GREEK",
        0x388: "GREEK",
        0x389: "GREEK",
        0x38A: "GREEK",
        0x38B: "GREEK",
        0x38C: "GREEK",
        0x38D: "GREEK",
        0x38E: "GREEK",
        0x38F: "GREEK",
        0x390: "GREEK",
        0x391: "GREEK",
        0x392: "GREEK",
        0x393: "GREEK",
        0x394: "GREEK",
        0x395: "GREEK",
        0x396: "GREEK",
        0x397: "GREEK",
        0x398: "GREEK",
        0x399: "GREEK",
        0x39A: "GREEK",
        0x39B: "GREEK",
        0x39C: "GREEK",
        0x39D: "GREEK",
        0x39E: "GREEK",
        0x39F: "GREEK",
        0x3A0: "GREEK",
        0x3A1: "GREEK",
        0x3A2: "GREEK",
        0x3A3: "GREEK",
        0x3A4: "GREEK",
        0x3A5: "GREEK",
        0x3A6: "GREEK",
        0x3A7: "GREEK",
        0x3A8: "GREEK",
        0x3A9: "GREEK",
        0x3AA: "GREEK",
        0x3AB: "GREEK",
        0x3AC: "GREEK",
        0x3AD: "GREEK",
        0x3AE: "GREEK",
        0x3AF: "GREEK",
        0x3B0: "GREEK",
        0x3B1: "GREEK",
        0x3B2: "GREEK",
        0x3B3: "GREEK",
        0x3B4: "GREEK",
        0x3B5: "GREEK",
        0x3B6: "GREEK",
        0x3B7: "GREEK",
        0x3B8: "GREEK",
        0x3B9: "GREEK",
        0x3BA: "GREEK",
        0x3BB: "GREEK",
        0x3BC: "GREEK",
        0x3BD: "GREEK",
        0x3BE: "GREEK",
        0x3BF: "GREEK",
        0x3C0: "GREEK",
        0x3C1: "GREEK",
        0x3C2: "GREEK",
        0x3C3: "GREEK",
        0x3C4: "GREEK",
        0x3C5: "GREEK",
        0x3C6: "GREEK",
        0x3C7: "GREEK",
        0x3C8: "GREEK",
        0x3C9: "GREEK",
        0x3CA: "GREEK",
        0x3CB: "GREEK",
        0x3CC: "GREEK",
        0x3CD: "GREEK",
        0x3CE: "GREEK",
        0x3CF: "GREEK",
        0x3D0: "GREEK",
        0x3D1: "GREEK",
        0x3D2: "GREEK",
        0x3D3: "GREEK",
        0x3D4: "GREEK",
        0x3D5: "GREEK",
        0x3D6: "GREEK",
        0x3D7: "GREEK",
        0x3D8: "GREEK",
        0x3D9: "GREEK",
        0x3DA: "GREEK",
        0x3DB: "GREEK",
        0x3DC: "GREEK",
        0x3DD: "GREEK",
        0x3DE: "GREEK",
        0x3DF: "GREEK",
        0x3E0: "GREEK",
        0x3E1: "GREEK",
        0x3E2: "GREEK",
        0x3E3: "GREEK",
        0x3E4: "GREEK",
        0x3E5: "GREEK",
        0x3E6: "GREEK",
        0x3E7: "GREEK",
        0x3E8: "GREEK",
        0x3E9: "GREEK",
        0x3EA: "GREEK",
        0x3EB: "GREEK",
        0x3EC: "GREEK",
        0x3ED: "GREEK",
        0x3EE: "GREEK",
        0x3EF: "GREEK",
        0x3F0: "GREEK",
        0x3F1: "GREEK",
        0x3F2: "GREEK",
        0x3F3: "GREEK",
        0x3F4: "GREEK",
        0x3F5: "GREEK",
        0x3F6: "GREEK",
        0x3F7: "GREEK",
        0x3F8: "GREEK",
        0x3F9: "GREEK",
        0x3FA: "GREEK",
        0x3FB: "GREEK",
        0xAC00: "HANGUL",
        0xAC01: "HANGUL",
        0xAC02: "HANGUL",
        0xAC03: "HANGUL",
        0xAC04: "HANGUL",
        0xAC05: "HANGUL",
        0xAC06: "HANGUL",
        0xAC07: "HANGUL",
        0xAC08: "HANGUL",
        0xAC09: "HANGUL",
        0xAC0A: "HANGUL",
        0xAC0B: "HANGUL",
        0xAC0C: "HANGUL",
        0xAC0D: "HANGUL",
        0xAC0E: "HANGUL",
        0xAC0F: "HANGUL",
        0xAC10: "HANGUL",
        0xAC11: "HANGUL",
        0xAC12: "HANGUL",
        0xAC13: "HANGUL",
        0xAC14: "HANGUL",
        0xAC15: "HANGUL",
        0xAC16: "HANGUL",
        0xAC17: "HANGUL",
        0xAC18: "HANGUL",
        0xAC19: "HANGUL",
        0xAC1A: "HANGUL",
        0xAC1B: "HANGUL",
        0xAC1C: "HANGUL",
        0xAC1D: "HANGUL",
        0xAC1E: "HANGUL",
        0xAC1F: "HANGUL",
        0xAC20: "HANGUL",
        0xAC21: "HANGUL",
        0xAC22: "HANGUL",
        0xAC23: "HANGUL",
        0xAC24: "HANGUL",
        0xAC25: "HANGUL",
        0xAC26: "HANGUL",
        0xAC27: "HANGUL",
        0xAC28: "HANGUL",
        0xAC29: "HANGUL",
        0xAC2A: "HANGUL",
        0xAC2B: "HANGUL",
        0xAC2C: "HANGUL",
        0xAC2D: "HANGUL",
        0xAC2E: "HANGUL",
        0xAC2F: "HANGUL",
        0xAC30: "HANGUL",
        0xAC31: "HANGUL",
        0xAC32: "HANGUL",
        0xAC33: "HANGUL",
        0xAC34: "HANGUL",
        0xAC35: "HANGUL",
        0xAC36: "HANGUL",
        0xAC37: "HANGUL",
        0xAC38: "HANGUL",
        0xAC39: "HANGUL",
        0xAC3A: "HANGUL",
        0xAC3B: "HANGUL",
        0xAC3C: "HANGUL",
        0xAC3D: "HANGUL",
        0xAC3E: "HANGUL",
        0xAC3F: "HANGUL",
        0xAC40: "HANGUL",
        0xAC41: "HANGUL",
        0xAC42: "HANGUL",
        0xAC43: "HANGUL",
        0xAC44: "HANGUL",
        0xAC45: "HANGUL",
        0xAC46: "HANGUL",
        0xAC47: "HANGUL",
        0xAC48: "HANGUL",
        0xAC49: "HANGUL",
        0xAC4A: "HANGUL",
        0xAC4B: "HANGUL",
        0xAC4C: "HANGUL",
        0xAC4D: "HANGUL",
        0xAC4E: "HANGUL",
        0xAC4F: "HANGUL",
        0xAC50: "HANGUL",
        0xAC51: "HANGUL",
        0xAC52: "HANGUL",
        0xAC53: "HANGUL",
        0xAC54: "HANGUL",
        0xAC55: "HANGUL",
        0xAC56: "HANGUL",
        0xAC57: "HANGUL",
        0xAC58: "HANGUL",
        0xAC59: "HANGUL",
        0xAC5A: "HANGUL",
        0xAC5B: "HANGUL",
        0xAC5C: "HANGUL",
        0xAC5D: "HANGUL",
        0xAC5E: "HANGUL",
        0xAC5F: "HANGUL",
        0xAC60: "HANGUL",
        0xAC61: "HANGUL",
        0xAC62: "HANGUL",
        0xAC63: "HANGUL",
        0xAC64: "HANGUL",
        0xAC65: "HANGUL",
        0xAC66: "HANGUL",
        0xAC67: "HANGUL",
        0xAC68: "HANGUL",
        0xAC69: "HANGUL",
        0xAC6A: "HANGUL",
        0xAC6B: "HANGUL",
        0xAC6C: "HANGUL",
        0xAC6D: "HANGUL",
        0xAC6E: "HANGUL",
        0xAC6F: "HANGUL",
        0xAC70: "HANGUL",
        0xAC71: "HANGUL",
        0xAC72: "HANGUL",
        0xAC73: "HANGUL",
        0xAC74: "HANGUL",
        0xAC75: "HANGUL",
        0xAC76: "HANGUL",
        0xAC77: "HANGUL",
        0xAC78: "HANGUL",
        0xAC79: "HANGUL",
        0xAC7A: "HANGUL",
        0xAC7B: "HANGUL",
        0xAC7C: "HANGUL",
        0xAC7D: "HANGUL",
        0xAC7E: "HANGUL",
        0xAC7F: "HANGUL",
        0xAC80: "HANGUL",
        0xAC81: "HANGUL",
        0xAC82: "HANGUL",
        0xAC83: "HANGUL",
        0xAC84: "HANGUL",
        0xAC85: "HANGUL",
        0xAC86: "HANGUL",
        0xAC87: "HANGUL",
        0xAC88: "HANGUL",
        0xAC89: "HANGUL",
        0xAC8A: "HANGUL",
        0xAC8B: "HANGUL",
        0xAC8C: "HANGUL",
        0xAC8D: "HANGUL",
        0xAC8E: "HANGUL",
        0xAC8F: "HANGUL",
        0xAC90: "HANGUL",
        0xAC91: "HANGUL",
        0xAC92: "HANGUL",
        0xAC93: "HANGUL",
        0xAC94: "HANGUL",
        0xAC95: "HANGUL",
        0xAC96: "HANGUL",
        0xAC97: "HANGUL",
        0xAC98: "HANGUL",
        0xAC99: "HANGUL",
        0xAC9A: "HANGUL",
        0xAC9B: "HANGUL",
        0xAC9C: "HANGUL",
        0xAC9D: "HANGUL",
        0xAC9E: "HANGUL",
        0xAC9F: "HANGUL",
        0xACA0: "HANGUL",
        0xACA1: "HANGUL",
        0xACA2: "HANGUL",
        0xACA3: "HANGUL",
        0xACA4: "HANGUL",
        0xACA5: "HANGUL",
        0xACA6: "HANGUL",
        0xACA7: "HANGUL",
        0xACA8: "HANGUL",
        0xACA9: "HANGUL",
        0xACAA: "HANGUL",
        0xACAB: "HANGUL",
        0xACAC: "HANGUL",
        0xACAD: "HANGUL",
        0xACAE: "HANGUL",
        0xACAF: "HANGUL",
        0xACB0: "HANGUL",
        0xACB1: "HANGUL",
        0xACB2: "HANGUL",
        0xACB3: "HANGUL",
        0xACB4: "HANGUL",
        0xACB5: "HANGUL",
        0xACB6: "HANGUL",
        0xACB7: "HANGUL",
        0xACB8: "HANGUL",
        0xACB9: "HANGUL",
        0xACBA: "HANGUL",
        0xACBB: "HANGUL",
        0xACBC: "HANGUL",
        0xACBD: "HANGUL",
        0xACBE: "HANGUL",
        0xACBF: "HANGUL",
        0xACC0: "HANGUL",
        0xACC1: "HANGUL",
        0xACC2: "HANGUL",
        0xACC3: "HANGUL",
        0xACC4: "HANGUL",
        0xACC5: "HANGUL",
        0xACC6: "HANGUL",
        0xACC7: "HANGUL",
        0xACC8: "HANGUL",
        0xACC9: "HANGUL",
        0xACCA: "HANGUL",
        0xACCB: "HANGUL",
        0xACCC: "HANGUL",
        0xACCD: "HANGUL",
        0xACCE: "HANGUL",
        0xACCF: "HANGUL",
        0xACD0: "HANGUL",
        0xACD1: "HANGUL",
        0xACD2: "HANGUL",
        0xACD3: "HANGUL",
        0xACD4: "HANGUL",
        0xACD5: "HANGUL",
        0xACD6: "HANGUL",
        0xACD7: "HANGUL",
        0xACD8: "HANGUL",
        0xACD9: "HANGUL",
        0xACDA: "HANGUL",
        0xACDB: "HANGUL",
        0xACDC: "HANGUL",
        0xACDD: "HANGUL",
        0xACDE: "HANGUL",
        0xACDF: "HANGUL",
        0xACE0: "HANGUL",
        0xACE1: "HANGUL",
        0xACE2: "HANGUL",
        0xACE3: "HANGUL",
        0xACE4: "HANGUL",
        0xACE5: "HANGUL",
        0xACE6: "HANGUL",
        0xACE7: "HANGUL",
        0xACE8: "HANGUL",
        0xACE9: "HANGUL",
        0xACEA: "HANGUL",
        0xACEB: "HANGUL",
        0xACEC: "HANGUL",
        0xACED: "HANGUL",
        0xACEE: "HANGUL",
        0xACEF: "HANGUL",
        0xACF0: "HANGUL",
        0xACF1: "HANGUL",
        0xACF2: "HANGUL",
        0xACF3: "HANGUL",
        0xACF4: "HANGUL",
        0xACF5: "HANGUL",
        0xACF6: "HANGUL",
        0xACF7: "HANGUL",
        0xACF8: "HANGUL",
        0xACF9: "HANGUL",
        0xACFA: "HANGUL",
        0xACFB: "HANGUL",
        0xACFC: "HANGUL",
        0xACFD: "HANGUL",
        0xACFE: "HANGUL",
        0xACFF: "HANGUL",
        0xAD00: "HANGUL",
        0xAD01: "HANGUL",
        0xAD02: "HANGUL",
        0xAD03: "HANGUL",
        0xAD04: "HANGUL",
        0xAD05: "HANGUL",
        0xAD06: "HANGUL",
        0xAD07: "HANGUL",
        0xAD08: "HANGUL",
        0xAD09: "HANGUL",
        0xAD0A: "HANGUL",
        0xAD0B: "HANGUL",
        0xAD0C: "HANGUL",
        0xAD0D: "HANGUL",
        0xAD0E: "HANGUL",
        0xAD0F: "HANGUL",
        0xAD10: "HANGUL",
        0xAD11: "HANGUL",
        0xAD12: "HANGUL",
        0xAD13: "HANGUL",
        0xAD14: "HANGUL",
        0xAD15: "HANGUL",
        0xAD16: "HANGUL",
        0xAD17: "HANGUL",
        0xAD18: "HANGUL",
        0xAD19: "HANGUL",
        0xAD1A: "HANGUL",
        0xAD1B: "HANGUL",
        0xAD1C: "HANGUL",
        0xAD1D: "HANGUL",
        0xAD1E: "HANGUL",
        0xAD1F: "HANGUL",
        0xAD20: "HANGUL",
        0xAD21: "HANGUL",
        0xAD22: "HANGUL",
        0xAD23: "HANGUL",
        0xAD24: "HANGUL",
        0xAD25: "HANGUL",
        0xAD26: "HANGUL",
        0xAD27: "HANGUL",
        0xAD28: "HANGUL",
        0xAD29: "HANGUL",
        0xAD2A: "HANGUL",
        0xAD2B: "HANGUL",
        0xAD2C: "HANGUL",
        0xAD2D: "HANGUL",
        0xAD2E: "HANGUL",
        0xAD2F: "HANGUL",
        0xAD30: "HANGUL",
        0xAD31: "HANGUL",
        0xAD32: "HANGUL",
        0xAD33: "HANGUL",
        0xAD34: "HANGUL",
        0xAD35: "HANGUL",
        0xAD36: "HANGUL",
        0xAD37: "HANGUL",
        0xAD38: "HANGUL",
        0xAD39: "HANGUL",
        0xAD3A: "HANGUL",
        0xAD3B: "HANGUL",
        0xAD3C: "HANGUL",
        0xAD3D: "HANGUL",
        0xAD3E: "HANGUL",
        0xAD3F: "HANGUL",
        0xAD40: "HANGUL",
        0xAD41: "HANGUL",
        0xAD42: "HANGUL",
        0xAD43: "HANGUL",
        0xAD44: "HANGUL",
        0xAD45: "HANGUL",
        0xAD46: "HANGUL",
        0xAD47: "HANGUL",
        0xAD48: "HANGUL",
        0xAD49: "HANGUL",
        0xAD4A: "HANGUL",
        0xAD4B: "HANGUL",
        0xAD4C: "HANGUL",
        0xAD4D: "HANGUL",
        0xAD4E: "HANGUL",
        0xAD4F: "HANGUL",
        0xAD50: "HANGUL",
        0xAD51: "HANGUL",
        0xAD52: "HANGUL",
        0xAD53: "HANGUL",
        0xAD54: "HANGUL",
        0xAD55: "HANGUL",
        0xAD56: "HANGUL",
        0xAD57: "HANGUL",
        0xAD58: "HANGUL",
        0xAD59: "HANGUL",
        0xAD5A: "HANGUL",
        0xAD5B: "HANGUL",
        0xAD5C: "HANGUL",
        0xAD5D: "HANGUL",
        0xAD5E: "HANGUL",
        0xAD5F: "HANGUL",
        0xAD60: "HANGUL",
        0xAD61: "HANGUL",
        0xAD62: "HANGUL",
        0xAD63: "HANGUL",
        0xAD64: "HANGUL",
        0xAD65: "HANGUL",
        0xAD66: "HANGUL",
        0xAD67: "HANGUL",
        0xAD68: "HANGUL",
        0xAD69: "HANGUL",
        0xAD6A: "HANGUL",
        0xAD6B: "HANGUL",
        0xAD6C: "HANGUL",
        0xAD6D: "HANGUL",
        0xAD6E: "HANGUL",
        0xAD6F: "HANGUL",
        0xAD70: "HANGUL",
        0xAD71: "HANGUL",
        0xAD72: "HANGUL",
        0xAD73: "HANGUL",
        0xAD74: "HANGUL",
        0xAD75: "HANGUL",
        0xAD76: "HANGUL",
        0xAD77: "HANGUL",
        0xAD78: "HANGUL",
        0xAD79: "HANGUL",
        0xAD7A: "HANGUL",
        0xAD7B: "HANGUL",
        0xAD7C: "HANGUL",
        0xAD7D: "HANGUL",
        0xAD7E: "HANGUL",
        0xAD7F: "HANGUL",
        0xAD80: "HANGUL",
        0xAD81: "HANGUL",
        0xAD82: "HANGUL",
        0xAD83: "HANGUL",
        0xAD84: "HANGUL",
        0xAD85: "HANGUL",
        0xAD86: "HANGUL",
        0xAD87: "HANGUL",
        0xAD88: "HANGUL",
        0xAD89: "HANGUL",
        0xAD8A: "HANGUL",
        0xAD8B: "HANGUL",
        0xAD8C: "HANGUL",
        0xAD8D: "HANGUL",
        0xAD8E: "HANGUL",
        0xAD8F: "HANGUL",
        0xAD90: "HANGUL",
        0xAD91: "HANGUL",
        0xAD92: "HANGUL",
        0xAD93: "HANGUL",
        0xAD94: "HANGUL",
        0xAD95: "HANGUL",
        0xAD96: "HANGUL",
        0xAD97: "HANGUL",
        0xAD98: "HANGUL",
        0xAD99: "HANGUL",
        0xAD9A: "HANGUL",
        0xAD9B: "HANGUL",
        0xAD9C: "HANGUL",
        0xAD9D: "HANGUL",
        0xAD9E: "HANGUL",
        0xAD9F: "HANGUL",
        0xADA0: "HANGUL",
        0xADA1: "HANGUL",
        0xADA2: "HANGUL",
        0xADA3: "HANGUL",
        0xADA4: "HANGUL",
        0xADA5: "HANGUL",
        0xADA6: "HANGUL",
        0xADA7: "HANGUL",
        0xADA8: "HANGUL",
        0xADA9: "HANGUL",
        0xADAA: "HANGUL",
        0xADAB: "HANGUL",
        0xADAC: "HANGUL",
        0xADAD: "HANGUL",
        0xADAE: "HANGUL",
        0xADAF: "HANGUL",
        0xADB0: "HANGUL",
        0xADB1: "HANGUL",
        0xADB2: "HANGUL",
        0xADB3: "HANGUL",
        0xADB4: "HANGUL",
        0xADB5: "HANGUL",
        0xADB6: "HANGUL",
        0xADB7: "HANGUL",
        0xADB8: "HANGUL",
        0xADB9: "HANGUL",
        0xADBA: "HANGUL",
        0xADBB: "HANGUL",
        0xADBC: "HANGUL",
        0xADBD: "HANGUL",
        0xADBE: "HANGUL",
        0xADBF: "HANGUL",
        0xADC0: "HANGUL",
        0xADC1: "HANGUL",
        0xADC2: "HANGUL",
        0xADC3: "HANGUL",
        0xADC4: "HANGUL",
        0xADC5: "HANGUL",
        0xADC6: "HANGUL",
        0xADC7: "HANGUL",
        0xADC8: "HANGUL",
        0xADC9: "HANGUL",
        0xADCA: "HANGUL",
        0xADCB: "HANGUL",
        0xADCC: "HANGUL",
        0xADCD: "HANGUL",
        0xADCE: "HANGUL",
        0xADCF: "HANGUL",
        0xADD0: "HANGUL",
        0xADD1: "HANGUL",
        0xADD2: "HANGUL",
        0xADD3: "HANGUL",
        0xADD4: "HANGUL",
        0xADD5: "HANGUL",
        0xADD6: "HANGUL",
        0xADD7: "HANGUL",
        0xADD8: "HANGUL",
        0xADD9: "HANGUL",
        0xADDA: "HANGUL",
        0xADDB: "HANGUL",
        0xADDC: "HANGUL",
        0xADDD: "HANGUL",
        0xADDE: "HANGUL",
        0xADDF: "HANGUL",
        0xADE0: "HANGUL",
        0xADE1: "HANGUL",
        0xADE2: "HANGUL",
        0xADE3: "HANGUL",
        0xADE4: "HANGUL",
        0xADE5: "HANGUL",
        0xADE6: "HANGUL",
        0xADE7: "HANGUL",
        0xADE8: "HANGUL",
        0xADE9: "HANGUL",
        0xADEA: "HANGUL",
        0xADEB: "HANGUL",
        0xADEC: "HANGUL",
        0xADED: "HANGUL",
        0xADEE: "HANGUL",
        0xADEF: "HANGUL",
        0xADF0: "HANGUL",
        0xADF1: "HANGUL",
        0xADF2: "HANGUL",
        0xADF3: "HANGUL",
        0xADF4: "HANGUL",
        0xADF5: "HANGUL",
        0xADF6: "HANGUL",
        0xADF7: "HANGUL",
        0xADF8: "HANGUL",
        0xADF9: "HANGUL",
        0xADFA: "HANGUL",
        0xADFB: "HANGUL",
        0xADFC: "HANGUL",
        0xADFD: "HANGUL",
        0xADFE: "HANGUL",
        0xADFF: "HANGUL",
        0xAE00: "HANGUL",
        0xAE01: "HANGUL",
        0xAE02: "HANGUL",
        0xAE03: "HANGUL",
        0xAE04: "HANGUL",
        0xAE05: "HANGUL",
        0xAE06: "HANGUL",
        0xAE07: "HANGUL",
        0xAE08: "HANGUL",
        0xAE09: "HANGUL",
        0xAE0A: "HANGUL",
        0xAE0B: "HANGUL",
        0xAE0C: "HANGUL",
        0xAE0D: "HANGUL",
        0xAE0E: "HANGUL",
        0xAE0F: "HANGUL",
        0xAE10: "HANGUL",
        0xAE11: "HANGUL",
        0xAE12: "HANGUL",
        0xAE13: "HANGUL",
        0xAE14: "HANGUL",
        0xAE15: "HANGUL",
        0xAE16: "HANGUL",
        0xAE17: "HANGUL",
        0xAE18: "HANGUL",
        0xAE19: "HANGUL",
        0xAE1A: "HANGUL",
        0xAE1B: "HANGUL",
        0xAE1C: "HANGUL",
        0xAE1D: "HANGUL",
        0xAE1E: "HANGUL",
        0xAE1F: "HANGUL",
        0xAE20: "HANGUL",
        0xAE21: "HANGUL",
        0xAE22: "HANGUL",
        0xAE23: "HANGUL",
        0xAE24: "HANGUL",
        0xAE25: "HANGUL",
        0xAE26: "HANGUL",
        0xAE27: "HANGUL",
        0xAE28: "HANGUL",
        0xAE29: "HANGUL",
        0xAE2A: "HANGUL",
        0xAE2B: "HANGUL",
        0xAE2C: "HANGUL",
        0xAE2D: "HANGUL",
        0xAE2E: "HANGUL",
        0xAE2F: "HANGUL",
        0xAE30: "HANGUL",
        0xAE31: "HANGUL",
        0xAE32: "HANGUL",
        0xAE33: "HANGUL",
        0xAE34: "HANGUL",
        0xAE35: "HANGUL",
        0xAE36: "HANGUL",
        0xAE37: "HANGUL",
        0xAE38: "HANGUL",
        0xAE39: "HANGUL",
        0xAE3A: "HANGUL",
        0xAE3B: "HANGUL",
        0xAE3C: "HANGUL",
        0xAE3D: "HANGUL",
        0xAE3E: "HANGUL",
        0xAE3F: "HANGUL",
        0xAE40: "HANGUL",
        0xAE41: "HANGUL",
        0xAE42: "HANGUL",
        0xAE43: "HANGUL",
        0xAE44: "HANGUL",
        0xAE45: "HANGUL",
        0xAE46: "HANGUL",
        0xAE47: "HANGUL",
        0xAE48: "HANGUL",
        0xAE49: "HANGUL",
        0xAE4A: "HANGUL",
        0xAE4B: "HANGUL",
        0xAE4C: "HANGUL",
        0xAE4D: "HANGUL",
        0xAE4E: "HANGUL",
        0xAE4F: "HANGUL",
        0xAE50: "HANGUL",
        0xAE51: "HANGUL",
        0xAE52: "HANGUL",
        0xAE53: "HANGUL",
        0xAE54: "HANGUL",
        0xAE55: "HANGUL",
        0xAE56: "HANGUL",
        0xAE57: "HANGUL",
        0xAE58: "HANGUL",
        0xAE59: "HANGUL",
        0xAE5A: "HANGUL",
        0xAE5B: "HANGUL",
        0xAE5C: "HANGUL",
        0xAE5D: "HANGUL",
        0xAE5E: "HANGUL",
        0xAE5F: "HANGUL",
        0xAE60: "HANGUL",
        0xAE61: "HANGUL",
        0xAE62: "HANGUL",
        0xAE63: "HANGUL",
        0xAE64: "HANGUL",
        0xAE65: "HANGUL",
        0xAE66: "HANGUL",
        0xAE67: "HANGUL",
        0xAE68: "HANGUL",
        0xAE69: "HANGUL",
        0xAE6A: "HANGUL",
        0xAE6B: "HANGUL",
        0xAE6C: "HANGUL",
        0xAE6D: "HANGUL",
        0xAE6E: "HANGUL",
        0xAE6F: "HANGUL",
        0xAE70: "HANGUL",
        0xAE71: "HANGUL",
        0xAE72: "HANGUL",
        0xAE73: "HANGUL",
        0xAE74: "HANGUL",
        0xAE75: "HANGUL",
        0xAE76: "HANGUL",
        0xAE77: "HANGUL",
        0xAE78: "HANGUL",
        0xAE79: "HANGUL",
        0xAE7A: "HANGUL",
        0xAE7B: "HANGUL",
        0xAE7C: "HANGUL",
        0xAE7D: "HANGUL",
        0xAE7E: "HANGUL",
        0xAE7F: "HANGUL",
        0xAE80: "HANGUL",
        0xAE81: "HANGUL",
        0xAE82: "HANGUL",
        0xAE83: "HANGUL",
        0xAE84: "HANGUL",
        0xAE85: "HANGUL",
        0xAE86: "HANGUL",
        0xAE87: "HANGUL",
        0xAE88: "HANGUL",
        0xAE89: "HANGUL",
        0xAE8A: "HANGUL",
        0xAE8B: "HANGUL",
        0xAE8C: "HANGUL",
        0xAE8D: "HANGUL",
        0xAE8E: "HANGUL",
        0xAE8F: "HANGUL",
        0xAE90: "HANGUL",
        0xAE91: "HANGUL",
        0xAE92: "HANGUL",
        0xAE93: "HANGUL",
        0xAE94: "HANGUL",
        0xAE95: "HANGUL",
        0xAE96: "HANGUL",
        0xAE97: "HANGUL",
        0xAE98: "HANGUL",
        0xAE99: "HANGUL",
        0xAE9A: "HANGUL",
        0xAE9B: "HANGUL",
        0xAE9C: "HANGUL",
        0xAE9D: "HANGUL",
        0xAE9E: "HANGUL",
        0xAE9F: "HANGUL",
        0xAEA0: "HANGUL",
        0xAEA1: "HANGUL",
        0xAEA2: "HANGUL",
        0xAEA3: "HANGUL",
        0xAEA4: "HANGUL",
        0xAEA5: "HANGUL",
        0xAEA6: "HANGUL",
        0xAEA7: "HANGUL",
        0xAEA8: "HANGUL",
        0xAEA9: "HANGUL",
        0xAEAA: "HANGUL",
        0xAEAB: "HANGUL",
        0xAEAC: "HANGUL",
        0xAEAD: "HANGUL",
        0xAEAE: "HANGUL",
        0xAEAF: "HANGUL",
        0xAEB0: "HANGUL",
        0xAEB1: "HANGUL",
        0xAEB2: "HANGUL",
        0xAEB3: "HANGUL",
        0xAEB4: "HANGUL",
        0xAEB5: "HANGUL",
        0xAEB6: "HANGUL",
        0xAEB7: "HANGUL",
        0xAEB8: "HANGUL",
        0xAEB9: "HANGUL",
        0xAEBA: "HANGUL",
        0xAEBB: "HANGUL",
        0xAEBC: "HANGUL",
        0xAEBD: "HANGUL",
        0xAEBE: "HANGUL",
        0xAEBF: "HANGUL",
        0xAEC0: "HANGUL",
        0xAEC1: "HANGUL",
        0xAEC2: "HANGUL",
        0xAEC3: "HANGUL",
        0xAEC4: "HANGUL",
        0xAEC5: "HANGUL",
        0xAEC6: "HANGUL",
        0xAEC7: "HANGUL",
        0xAEC8: "HANGUL",
        0xAEC9: "HANGUL",
        0xAECA: "HANGUL",
        0xAECB: "HANGUL",
        0xAECC: "HANGUL",
        0xAECD: "HANGUL",
        0xAECE: "HANGUL",
        0xAECF: "HANGUL",
        0xAED0: "HANGUL",
        0xAED1: "HANGUL",
        0xAED2: "HANGUL",
        0xAED3: "HANGUL",
        0xAED4: "HANGUL",
        0xAED5: "HANGUL",
        0xAED6: "HANGUL",
        0xAED7: "HANGUL",
        0xAED8: "HANGUL",
        0xAED9: "HANGUL",
        0xAEDA: "HANGUL",
        0xAEDB: "HANGUL",
        0xAEDC: "HANGUL",
        0xAEDD: "HANGUL",
        0xAEDE: "HANGUL",
        0xAEDF: "HANGUL",
        0xAEE0: "HANGUL",
        0xAEE1: "HANGUL",
        0xAEE2: "HANGUL",
        0xAEE3: "HANGUL",
        0xAEE4: "HANGUL",
        0xAEE5: "HANGUL",
        0xAEE6: "HANGUL",
        0xAEE7: "HANGUL",
        0xAEE8: "HANGUL",
        0xAEE9: "HANGUL",
        0xAEEA: "HANGUL",
        0xAEEB: "HANGUL",
        0xAEEC: "HANGUL",
        0xAEED: "HANGUL",
        0xAEEE: "HANGUL",
        0xAEEF: "HANGUL",
        0xAEF0: "HANGUL",
        0xAEF1: "HANGUL",
        0xAEF2: "HANGUL",
        0xAEF3: "HANGUL",
        0xAEF4: "HANGUL",
        0xAEF5: "HANGUL",
        0xAEF6: "HANGUL",
        0xAEF7: "HANGUL",
        0xAEF8: "HANGUL",
        0xAEF9: "HANGUL",
        0xAEFA: "HANGUL",
        0xAEFB: "HANGUL",
        0xAEFC: "HANGUL",
        0xAEFD: "HANGUL",
        0xAEFE: "HANGUL",
        0xAEFF: "HANGUL",
        0xAF00: "HANGUL",
        0xAF01: "HANGUL",
        0xAF02: "HANGUL",
        0xAF03: "HANGUL",
        0xAF04: "HANGUL",
        0xAF05: "HANGUL",
        0xAF06: "HANGUL",
        0xAF07: "HANGUL",
        0xAF08: "HANGUL",
        0xAF09: "HANGUL",
        0xAF0A: "HANGUL",
        0xAF0B: "HANGUL",
        0xAF0C: "HANGUL",
        0xAF0D: "HANGUL",
        0xAF0E: "HANGUL",
        0xAF0F: "HANGUL",
        0xAF10: "HANGUL",
        0xAF11: "HANGUL",
        0xAF12: "HANGUL",
        0xAF13: "HANGUL",
        0xAF14: "HANGUL",
        0xAF15: "HANGUL",
        0xAF16: "HANGUL",
        0xAF17: "HANGUL",
        0xAF18: "HANGUL",
        0xAF19: "HANGUL",
        0xAF1A: "HANGUL",
        0xAF1B: "HANGUL",
        0xAF1C: "HANGUL",
        0xAF1D: "HANGUL",
        0xAF1E: "HANGUL",
        0xAF1F: "HANGUL",
        0xAF20: "HANGUL",
        0xAF21: "HANGUL",
        0xAF22: "HANGUL",
        0xAF23: "HANGUL",
        0xAF24: "HANGUL",
        0xAF25: "HANGUL",
        0xAF26: "HANGUL",
        0xAF27: "HANGUL",
        0xAF28: "HANGUL",
        0xAF29: "HANGUL",
        0xAF2A: "HANGUL",
        0xAF2B: "HANGUL",
        0xAF2C: "HANGUL",
        0xAF2D: "HANGUL",
        0xAF2E: "HANGUL",
        0xAF2F: "HANGUL",
        0xAF30: "HANGUL",
        0xAF31: "HANGUL",
        0xAF32: "HANGUL",
        0xAF33: "HANGUL",
        0xAF34: "HANGUL",
        0xAF35: "HANGUL",
        0xAF36: "HANGUL",
        0xAF37: "HANGUL",
        0xAF38: "HANGUL",
        0xAF39: "HANGUL",
        0xAF3A: "HANGUL",
        0xAF3B: "HANGUL",
        0xAF3C: "HANGUL",
        0xAF3D: "HANGUL",
        0xAF3E: "HANGUL",
        0xAF3F: "HANGUL",
        0xAF40: "HANGUL",
        0xAF41: "HANGUL",
        0xAF42: "HANGUL",
        0xAF43: "HANGUL",
        0xAF44: "HANGUL",
        0xAF45: "HANGUL",
        0xAF46: "HANGUL",
        0xAF47: "HANGUL",
        0xAF48: "HANGUL",
        0xAF49: "HANGUL",
        0xAF4A: "HANGUL",
        0xAF4B: "HANGUL",
        0xAF4C: "HANGUL",
        0xAF4D: "HANGUL",
        0xAF4E: "HANGUL",
        0xAF4F: "HANGUL",
        0xAF50: "HANGUL",
        0xAF51: "HANGUL",
        0xAF52: "HANGUL",
        0xAF53: "HANGUL",
        0xAF54: "HANGUL",
        0xAF55: "HANGUL",
        0xAF56: "HANGUL",
        0xAF57: "HANGUL",
        0xAF58: "HANGUL",
        0xAF59: "HANGUL",
        0xAF5A: "HANGUL",
        0xAF5B: "HANGUL",
        0xAF5C: "HANGUL",
        0xAF5D: "HANGUL",
        0xAF5E: "HANGUL",
        0xAF5F: "HANGUL",
        0xAF60: "HANGUL",
        0xAF61: "HANGUL",
        0xAF62: "HANGUL",
        0xAF63: "HANGUL",
        0xAF64: "HANGUL",
        0xAF65: "HANGUL",
        0xAF66: "HANGUL",
        0xAF67: "HANGUL",
        0xAF68: "HANGUL",
        0xAF69: "HANGUL",
        0xAF6A: "HANGUL",
        0xAF6B: "HANGUL",
        0xAF6C: "HANGUL",
        0xAF6D: "HANGUL",
        0xAF6E: "HANGUL",
        0xAF6F: "HANGUL",
        0xAF70: "HANGUL",
        0xAF71: "HANGUL",
        0xAF72: "HANGUL",
        0xAF73: "HANGUL",
        0xAF74: "HANGUL",
        0xAF75: "HANGUL",
        0xAF76: "HANGUL",
        0xAF77: "HANGUL",
        0xAF78: "HANGUL",
        0xAF79: "HANGUL",
        0xAF7A: "HANGUL",
        0xAF7B: "HANGUL",
        0xAF7C: "HANGUL",
        0xAF7D: "HANGUL",
        0xAF7E: "HANGUL",
        0xAF7F: "HANGUL",
        0xAF80: "HANGUL",
        0xAF81: "HANGUL",
        0xAF82: "HANGUL",
        0xAF83: "HANGUL",
        0xAF84: "HANGUL",
        0xAF85: "HANGUL",
        0xAF86: "HANGUL",
        0xAF87: "HANGUL",
        0xAF88: "HANGUL",
        0xAF89: "HANGUL",
        0xAF8A: "HANGUL",
        0xAF8B: "HANGUL",
        0xAF8C: "HANGUL",
        0xAF8D: "HANGUL",
        0xAF8E: "HANGUL",
        0xAF8F: "HANGUL",
        0xAF90: "HANGUL",
        0xAF91: "HANGUL",
        0xAF92: "HANGUL",
        0xAF93: "HANGUL",
        0xAF94: "HANGUL",
        0xAF95: "HANGUL",
        0xAF96: "HANGUL",
        0xAF97: "HANGUL",
        0xAF98: "HANGUL",
        0xAF99: "HANGUL",
        0xAF9A: "HANGUL",
        0xAF9B: "HANGUL",
        0xAF9C: "HANGUL",
        0xAF9D: "HANGUL",
        0xAF9E: "HANGUL",
        0xAF9F: "HANGUL",
        0xAFA0: "HANGUL",
        0xAFA1: "HANGUL",
        0xAFA2: "HANGUL",
        0xAFA3: "HANGUL",
        0xAFA4: "HANGUL",
        0xAFA5: "HANGUL",
        0xAFA6: "HANGUL",
        0xAFA7: "HANGUL",
        0xAFA8: "HANGUL",
        0xAFA9: "HANGUL",
        0xAFAA: "HANGUL",
        0xAFAB: "HANGUL",
        0xAFAC: "HANGUL",
        0xAFAD: "HANGUL",
        0xAFAE: "HANGUL",
        0xAFAF: "HANGUL",
        0xAFB0: "HANGUL",
        0xAFB1: "HANGUL",
        0xAFB2: "HANGUL",
        0xAFB3: "HANGUL",
        0xAFB4: "HANGUL",
        0xAFB5: "HANGUL",
        0xAFB6: "HANGUL",
        0xAFB7: "HANGUL",
        0xAFB8: "HANGUL",
        0xAFB9: "HANGUL",
        0xAFBA: "HANGUL",
        0xAFBB: "HANGUL",
        0xAFBC: "HANGUL",
        0xAFBD: "HANGUL",
        0xAFBE: "HANGUL",
        0xAFBF: "HANGUL",
        0xAFC0: "HANGUL",
        0xAFC1: "HANGUL",
        0xAFC2: "HANGUL",
        0xAFC3: "HANGUL",
        0xAFC4: "HANGUL",
        0xAFC5: "HANGUL",
        0xAFC6: "HANGUL",
        0xAFC7: "HANGUL",
        0xAFC8: "HANGUL",
        0xAFC9: "HANGUL",
        0xAFCA: "HANGUL",
        0xAFCB: "HANGUL",
        0xAFCC: "HANGUL",
        0xAFCD: "HANGUL",
        0xAFCE: "HANGUL",
        0xAFCF: "HANGUL",
        0xAFD0: "HANGUL",
        0xAFD1: "HANGUL",
        0xAFD2: "HANGUL",
        0xAFD3: "HANGUL",
        0xAFD4: "HANGUL",
        0xAFD5: "HANGUL",
        0xAFD6: "HANGUL",
        0xAFD7: "HANGUL",
        0xAFD8: "HANGUL",
        0xAFD9: "HANGUL",
        0xAFDA: "HANGUL",
        0xAFDB: "HANGUL",
        0xAFDC: "HANGUL",
        0xAFDD: "HANGUL",
        0xAFDE: "HANGUL",
        0xAFDF: "HANGUL",
        0xAFE0: "HANGUL",
        0xAFE1: "HANGUL",
        0xAFE2: "HANGUL",
        0xAFE3: "HANGUL",
        0xAFE4: "HANGUL",
        0xAFE5: "HANGUL",
        0xAFE6: "HANGUL",
        0xAFE7: "HANGUL",
        0xAFE8: "HANGUL",
        0xAFE9: "HANGUL",
        0xAFEA: "HANGUL",
        0xAFEB: "HANGUL",
        0xAFEC: "HANGUL",
        0xAFED: "HANGUL",
        0xAFEE: "HANGUL",
        0xAFEF: "HANGUL",
        0xAFF0: "HANGUL",
        0xAFF1: "HANGUL",
        0xAFF2: "HANGUL",
        0xAFF3: "HANGUL",
        0xAFF4: "HANGUL",
        0xAFF5: "HANGUL",
        0xAFF6: "HANGUL",
        0xAFF7: "HANGUL",
        0xAFF8: "HANGUL",
        0xAFF9: "HANGUL",
        0xAFFA: "HANGUL",
        0xAFFB: "HANGUL",
        0xAFFC: "HANGUL",
        0xAFFD: "HANGUL",
        0xAFFE: "HANGUL",
        0xAFFF: "HANGUL",
        0xB000: "HANGUL",
        0xB001: "HANGUL",
        0xB002: "HANGUL",
        0xB003: "HANGUL",
        0xB004: "HANGUL",
        0xB005: "HANGUL",
        0xB006: "HANGUL",
        0xB007: "HANGUL",
        0xB008: "HANGUL",
        0xB009: "HANGUL",
        0xB00A: "HANGUL",
        0xB00B: "HANGUL",
        0xB00C: "HANGUL",
        0xB00D: "HANGUL",
        0xB00E: "HANGUL",
        0xB00F: "HANGUL",
        0xB010: "HANGUL",
        0xB011: "HANGUL",
        0xB012: "HANGUL",
        0xB013: "HANGUL",
        0xB014: "HANGUL",
        0xB015: "HANGUL",
        0xB016: "HANGUL",
        0xB017: "HANGUL",
        0xB018: "HANGUL",
        0xB019: "HANGUL",
        0xB01A: "HANGUL",
        0xB01B: "HANGUL",
        0xB01C: "HANGUL",
        0xB01D: "HANGUL",
        0xB01E: "HANGUL",
        0xB01F: "HANGUL",
        0xB020: "HANGUL",
        0xB021: "HANGUL",
        0xB022: "HANGUL",
        0xB023: "HANGUL",
        0xB024: "HANGUL",
        0xB025: "HANGUL",
        0xB026: "HANGUL",
        0xB027: "HANGUL",
        0xB028: "HANGUL",
        0xB029: "HANGUL",
        0xB02A: "HANGUL",
        0xB02B: "HANGUL",
        0xB02C: "HANGUL",
        0xB02D: "HANGUL",
        0xB02E: "HANGUL",
        0xB02F: "HANGUL",
        0xB030: "HANGUL",
        0xB031: "HANGUL",
        0xB032: "HANGUL",
        0xB033: "HANGUL",
        0xB034: "HANGUL",
        0xB035: "HANGUL",
        0xB036: "HANGUL",
        0xB037: "HANGUL",
        0xB038: "HANGUL",
        0xB039: "HANGUL",
        0xB03A: "HANGUL",
        0xB03B: "HANGUL",
        0xB03C: "HANGUL",
        0xB03D: "HANGUL",
        0xB03E: "HANGUL",
        0xB03F: "HANGUL",
        0xB040: "HANGUL",
        0xB041: "HANGUL",
        0xB042: "HANGUL",
        0xB043: "HANGUL",
        0xB044: "HANGUL",
        0xB045: "HANGUL",
        0xB046: "HANGUL",
        0xB047: "HANGUL",
        0xB048: "HANGUL",
        0xB049: "HANGUL",
        0xB04A: "HANGUL",
        0xB04B: "HANGUL",
        0xB04C: "HANGUL",
        0xB04D: "HANGUL",
        0xB04E: "HANGUL",
        0xB04F: "HANGUL",
        0xB050: "HANGUL",
        0xB051: "HANGUL",
        0xB052: "HANGUL",
        0xB053: "HANGUL",
        0xB054: "HANGUL",
        0xB055: "HANGUL",
        0xB056: "HANGUL",
        0xB057: "HANGUL",
        0xB058: "HANGUL",
        0xB059: "HANGUL",
        0xB05A: "HANGUL",
        0xB05B: "HANGUL",
        0xB05C: "HANGUL",
        0xB05D: "HANGUL",
        0xB05E: "HANGUL",
        0xB05F: "HANGUL",
        0xB060: "HANGUL",
        0xB061: "HANGUL",
        0xB062: "HANGUL",
        0xB063: "HANGUL",
        0xB064: "HANGUL",
        0xB065: "HANGUL",
        0xB066: "HANGUL",
        0xB067: "HANGUL",
        0xB068: "HANGUL",
        0xB069: "HANGUL",
        0xB06A: "HANGUL",
        0xB06B: "HANGUL",
        0xB06C: "HANGUL",
        0xB06D: "HANGUL",
        0xB06E: "HANGUL",
        0xB06F: "HANGUL",
        0xB070: "HANGUL",
        0xB071: "HANGUL",
        0xB072: "HANGUL",
        0xB073: "HANGUL",
        0xB074: "HANGUL",
        0xB075: "HANGUL",
        0xB076: "HANGUL",
        0xB077: "HANGUL",
        0xB078: "HANGUL",
        0xB079: "HANGUL",
        0xB07A: "HANGUL",
        0xB07B: "HANGUL",
        0xB07C: "HANGUL",
        0xB07D: "HANGUL",
        0xB07E: "HANGUL",
        0xB07F: "HANGUL",
        0xB080: "HANGUL",
        0xB081: "HANGUL",
        0xB082: "HANGUL",
        0xB083: "HANGUL",
        0xB084: "HANGUL",
        0xB085: "HANGUL",
        0xB086: "HANGUL",
        0xB087: "HANGUL",
        0xB088: "HANGUL",
        0xB089: "HANGUL",
        0xB08A: "HANGUL",
        0xB08B: "HANGUL",
        0xB08C: "HANGUL",
        0xB08D: "HANGUL",
        0xB08E: "HANGUL",
        0xB08F: "HANGUL",
        0xB090: "HANGUL",
        0xB091: "HANGUL",
        0xB092: "HANGUL",
        0xB093: "HANGUL",
        0xB094: "HANGUL",
        0xB095: "HANGUL",
        0xB096: "HANGUL",
        0xB097: "HANGUL",
        0xB098: "HANGUL",
        0xB099: "HANGUL",
        0xB09A: "HANGUL",
        0xB09B: "HANGUL",
        0xB09C: "HANGUL",
        0xB09D: "HANGUL",
        0xB09E: "HANGUL",
        0xB09F: "HANGUL",
        0xB0A0: "HANGUL",
        0xB0A1: "HANGUL",
        0xB0A2: "HANGUL",
        0xB0A3: "HANGUL",
        0xB0A4: "HANGUL",
        0xB0A5: "HANGUL",
        0xB0A6: "HANGUL",
        0xB0A7: "HANGUL",
        0xB0A8: "HANGUL",
        0xB0A9: "HANGUL",
        0xB0AA: "HANGUL",
        0xB0AB: "HANGUL",
        0xB0AC: "HANGUL",
        0xB0AD: "HANGUL",
        0xB0AE: "HANGUL",
        0xB0AF: "HANGUL",
        0xB0B0: "HANGUL",
        0xB0B1: "HANGUL",
        0xB0B2: "HANGUL",
        0xB0B3: "HANGUL",
        0xB0B4: "HANGUL",
        0xB0B5: "HANGUL",
        0xB0B6: "HANGUL",
        0xB0B7: "HANGUL",
        0xB0B8: "HANGUL",
        0xB0B9: "HANGUL",
        0xB0BA: "HANGUL",
        0xB0BB: "HANGUL",
        0xB0BC: "HANGUL",
        0xB0BD: "HANGUL",
        0xB0BE: "HANGUL",
        0xB0BF: "HANGUL",
        0xB0C0: "HANGUL",
        0xB0C1: "HANGUL",
        0xB0C2: "HANGUL",
        0xB0C3: "HANGUL",
        0xB0C4: "HANGUL",
        0xB0C5: "HANGUL",
        0xB0C6: "HANGUL",
        0xB0C7: "HANGUL",
        0xB0C8: "HANGUL",
        0xB0C9: "HANGUL",
        0xB0CA: "HANGUL",
        0xB0CB: "HANGUL",
        0xB0CC: "HANGUL",
        0xB0CD: "HANGUL",
        0xB0CE: "HANGUL",
        0xB0CF: "HANGUL",
        0xB0D0: "HANGUL",
        0xB0D1: "HANGUL",
        0xB0D2: "HANGUL",
        0xB0D3: "HANGUL",
        0xB0D4: "HANGUL",
        0xB0D5: "HANGUL",
        0xB0D6: "HANGUL",
        0xB0D7: "HANGUL",
        0xB0D8: "HANGUL",
        0xB0D9: "HANGUL",
        0xB0DA: "HANGUL",
        0xB0DB: "HANGUL",
        0xB0DC: "HANGUL",
        0xB0DD: "HANGUL",
        0xB0DE: "HANGUL",
        0xB0DF: "HANGUL",
        0xB0E0: "HANGUL",
        0xB0E1: "HANGUL",
        0xB0E2: "HANGUL",
        0xB0E3: "HANGUL",
        0xB0E4: "HANGUL",
        0xB0E5: "HANGUL",
        0xB0E6: "HANGUL",
        0xB0E7: "HANGUL",
        0xB0E8: "HANGUL",
        0xB0E9: "HANGUL",
        0xB0EA: "HANGUL",
        0xB0EB: "HANGUL",
        0xB0EC: "HANGUL",
        0xB0ED: "HANGUL",
        0xB0EE: "HANGUL",
        0xB0EF: "HANGUL",
        0xB0F0: "HANGUL",
        0xB0F1: "HANGUL",
        0xB0F2: "HANGUL",
        0xB0F3: "HANGUL",
        0xB0F4: "HANGUL",
        0xB0F5: "HANGUL",
        0xB0F6: "HANGUL",
        0xB0F7: "HANGUL",
        0xB0F8: "HANGUL",
        0xB0F9: "HANGUL",
        0xB0FA: "HANGUL",
        0xB0FB: "HANGUL",
        0xB0FC: "HANGUL",
        0xB0FD: "HANGUL",
        0xB0FE: "HANGUL",
        0xB0FF: "HANGUL",
        0xB100: "HANGUL",
        0xB101: "HANGUL",
        0xB102: "HANGUL",
        0xB103: "HANGUL",
        0xB104: "HANGUL",
        0xB105: "HANGUL",
        0xB106: "HANGUL",
        0xB107: "HANGUL",
        0xB108: "HANGUL",
        0xB109: "HANGUL",
        0xB10A: "HANGUL",
        0xB10B: "HANGUL",
        0xB10C: "HANGUL",
        0xB10D: "HANGUL",
        0xB10E: "HANGUL",
        0xB10F: "HANGUL",
        0xB110: "HANGUL",
        0xB111: "HANGUL",
        0xB112: "HANGUL",
        0xB113: "HANGUL",
        0xB114: "HANGUL",
        0xB115: "HANGUL",
        0xB116: "HANGUL",
        0xB117: "HANGUL",
        0xB118: "HANGUL",
        0xB119: "HANGUL",
        0xB11A: "HANGUL",
        0xB11B: "HANGUL",
        0xB11C: "HANGUL",
        0xB11D: "HANGUL",
        0xB11E: "HANGUL",
        0xB11F: "HANGUL",
        0xB120: "HANGUL",
        0xB121: "HANGUL",
        0xB122: "HANGUL",
        0xB123: "HANGUL",
        0xB124: "HANGUL",
        0xB125: "HANGUL",
        0xB126: "HANGUL",
        0xB127: "HANGUL",
        0xB128: "HANGUL",
        0xB129: "HANGUL",
        0xB12A: "HANGUL",
        0xB12B: "HANGUL",
        0xB12C: "HANGUL",
        0xB12D: "HANGUL",
        0xB12E: "HANGUL",
        0xB12F: "HANGUL",
        0xB130: "HANGUL",
        0xB131: "HANGUL",
        0xB132: "HANGUL",
        0xB133: "HANGUL",
        0xB134: "HANGUL",
        0xB135: "HANGUL",
        0xB136: "HANGUL",
        0xB137: "HANGUL",
        0xB138: "HANGUL",
        0xB139: "HANGUL",
        0xB13A: "HANGUL",
        0xB13B: "HANGUL",
        0xB13C: "HANGUL",
        0xB13D: "HANGUL",
        0xB13E: "HANGUL",
        0xB13F: "HANGUL",
        0xB140: "HANGUL",
        0xB141: "HANGUL",
        0xB142: "HANGUL",
        0xB143: "HANGUL",
        0xB144: "HANGUL",
        0xB145: "HANGUL",
        0xB146: "HANGUL",
        0xB147: "HANGUL",
        0xB148: "HANGUL",
        0xB149: "HANGUL",
        0xB14A: "HANGUL",
        0xB14B: "HANGUL",
        0xB14C: "HANGUL",
        0xB14D: "HANGUL",
        0xB14E: "HANGUL",
        0xB14F: "HANGUL",
        0xB150: "HANGUL",
        0xB151: "HANGUL",
        0xB152: "HANGUL",
        0xB153: "HANGUL",
        0xB154: "HANGUL",
        0xB155: "HANGUL",
        0xB156: "HANGUL",
        0xB157: "HANGUL",
        0xB158: "HANGUL",
        0xB159: "HANGUL",
        0xB15A: "HANGUL",
        0xB15B: "HANGUL",
        0xB15C: "HANGUL",
        0xB15D: "HANGUL",
        0xB15E: "HANGUL",
        0xB15F: "HANGUL",
        0xB160: "HANGUL",
        0xB161: "HANGUL",
        0xB162: "HANGUL",
        0xB163: "HANGUL",
        0xB164: "HANGUL",
        0xB165: "HANGUL",
        0xB166: "HANGUL",
        0xB167: "HANGUL",
        0xB168: "HANGUL",
        0xB169: "HANGUL",
        0xB16A: "HANGUL",
        0xB16B: "HANGUL",
        0xB16C: "HANGUL",
        0xB16D: "HANGUL",
        0xB16E: "HANGUL",
        0xB16F: "HANGUL",
        0xB170: "HANGUL",
        0xB171: "HANGUL",
        0xB172: "HANGUL",
        0xB173: "HANGUL",
        0xB174: "HANGUL",
        0xB175: "HANGUL",
        0xB176: "HANGUL",
        0xB177: "HANGUL",
        0xB178: "HANGUL",
        0xB179: "HANGUL",
        0xB17A: "HANGUL",
        0xB17B: "HANGUL",
        0xB17C: "HANGUL",
        0xB17D: "HANGUL",
        0xB17E: "HANGUL",
        0xB17F: "HANGUL",
        0xB180: "HANGUL",
        0xB181: "HANGUL",
        0xB182: "HANGUL",
        0xB183: "HANGUL",
        0xB184: "HANGUL",
        0xB185: "HANGUL",
        0xB186: "HANGUL",
        0xB187: "HANGUL",
        0xB188: "HANGUL",
        0xB189: "HANGUL",
        0xB18A: "HANGUL",
        0xB18B: "HANGUL",
        0xB18C: "HANGUL",
        0xB18D: "HANGUL",
        0xB18E: "HANGUL",
        0xB18F: "HANGUL",
        0xB190: "HANGUL",
        0xB191: "HANGUL",
        0xB192: "HANGUL",
        0xB193: "HANGUL",
        0xB194: "HANGUL",
        0xB195: "HANGUL",
        0xB196: "HANGUL",
        0xB197: "HANGUL",
        0xB198: "HANGUL",
        0xB199: "HANGUL",
        0xB19A: "HANGUL",
        0xB19B: "HANGUL",
        0xB19C: "HANGUL",
        0xB19D: "HANGUL",
        0xB19E: "HANGUL",
        0xB19F: "HANGUL",
        0xB1A0: "HANGUL",
        0xB1A1: "HANGUL",
        0xB1A2: "HANGUL",
        0xB1A3: "HANGUL",
        0xB1A4: "HANGUL",
        0xB1A5: "HANGUL",
        0xB1A6: "HANGUL",
        0xB1A7: "HANGUL",
        0xB1A8: "HANGUL",
        0xB1A9: "HANGUL",
        0xB1AA: "HANGUL",
        0xB1AB: "HANGUL",
        0xB1AC: "HANGUL",
        0xB1AD: "HANGUL",
        0xB1AE: "HANGUL",
        0xB1AF: "HANGUL",
        0xB1B0: "HANGUL",
        0xB1B1: "HANGUL",
        0xB1B2: "HANGUL",
        0xB1B3: "HANGUL",
        0xB1B4: "HANGUL",
        0xB1B5: "HANGUL",
        0xB1B6: "HANGUL",
        0xB1B7: "HANGUL",
        0xB1B8: "HANGUL",
        0xB1B9: "HANGUL",
        0xB1BA: "HANGUL",
        0xB1BB: "HANGUL",
        0xB1BC: "HANGUL",
        0xB1BD: "HANGUL",
        0xB1BE: "HANGUL",
        0xB1BF: "HANGUL",
        0xB1C0: "HANGUL",
        0xB1C1: "HANGUL",
        0xB1C2: "HANGUL",
        0xB1C3: "HANGUL",
        0xB1C4: "HANGUL",
        0xB1C5: "HANGUL",
        0xB1C6: "HANGUL",
        0xB1C7: "HANGUL",
        0xB1C8: "HANGUL",
        0xB1C9: "HANGUL",
        0xB1CA: "HANGUL",
        0xB1CB: "HANGUL",
        0xB1CC: "HANGUL",
        0xB1CD: "HANGUL",
        0xB1CE: "HANGUL",
        0xB1CF: "HANGUL",
        0xB1D0: "HANGUL",
        0xB1D1: "HANGUL",
        0xB1D2: "HANGUL",
        0xB1D3: "HANGUL",
        0xB1D4: "HANGUL",
        0xB1D5: "HANGUL",
        0xB1D6: "HANGUL",
        0xB1D7: "HANGUL",
        0xB1D8: "HANGUL",
        0xB1D9: "HANGUL",
        0xB1DA: "HANGUL",
        0xB1DB: "HANGUL",
        0xB1DC: "HANGUL",
        0xB1DD: "HANGUL",
        0xB1DE: "HANGUL",
        0xB1DF: "HANGUL",
        0xB1E0: "HANGUL",
        0xB1E1: "HANGUL",
        0xB1E2: "HANGUL",
        0xB1E3: "HANGUL",
        0xB1E4: "HANGUL",
        0xB1E5: "HANGUL",
        0xB1E6: "HANGUL",
        0xB1E7: "HANGUL",
        0xB1E8: "HANGUL",
        0xB1E9: "HANGUL",
        0xB1EA: "HANGUL",
        0xB1EB: "HANGUL",
        0xB1EC: "HANGUL",
        0xB1ED: "HANGUL",
        0xB1EE: "HANGUL",
        0xB1EF: "HANGUL",
        0xB1F0: "HANGUL",
        0xB1F1: "HANGUL",
        0xB1F2: "HANGUL",
        0xB1F3: "HANGUL",
        0xB1F4: "HANGUL",
        0xB1F5: "HANGUL",
        0xB1F6: "HANGUL",
        0xB1F7: "HANGUL",
        0xB1F8: "HANGUL",
        0xB1F9: "HANGUL",
        0xB1FA: "HANGUL",
        0xB1FB: "HANGUL",
        0xB1FC: "HANGUL",
        0xB1FD: "HANGUL",
        0xB1FE: "HANGUL",
        0xB1FF: "HANGUL",
        0xB200: "HANGUL",
        0xB201: "HANGUL",
        0xB202: "HANGUL",
        0xB203: "HANGUL",
        0xB204: "HANGUL",
        0xB205: "HANGUL",
        0xB206: "HANGUL",
        0xB207: "HANGUL",
        0xB208: "HANGUL",
        0xB209: "HANGUL",
        0xB20A: "HANGUL",
        0xB20B: "HANGUL",
        0xB20C: "HANGUL",
        0xB20D: "HANGUL",
        0xB20E: "HANGUL",
        0xB20F: "HANGUL",
        0xB210: "HANGUL",
        0xB211: "HANGUL",
        0xB212: "HANGUL",
        0xB213: "HANGUL",
        0xB214: "HANGUL",
        0xB215: "HANGUL",
        0xB216: "HANGUL",
        0xB217: "HANGUL",
        0xB218: "HANGUL",
        0xB219: "HANGUL",
        0xB21A: "HANGUL",
        0xB21B: "HANGUL",
        0xB21C: "HANGUL",
        0xB21D: "HANGUL",
        0xB21E: "HANGUL",
        0xB21F: "HANGUL",
        0xB220: "HANGUL",
        0xB221: "HANGUL",
        0xB222: "HANGUL",
        0xB223: "HANGUL",
        0xB224: "HANGUL",
        0xB225: "HANGUL",
        0xB226: "HANGUL",
        0xB227: "HANGUL",
        0xB228: "HANGUL",
        0xB229: "HANGUL",
        0xB22A: "HANGUL",
        0xB22B: "HANGUL",
        0xB22C: "HANGUL",
        0xB22D: "HANGUL",
        0xB22E: "HANGUL",
        0xB22F: "HANGUL",
        0xB230: "HANGUL",
        0xB231: "HANGUL",
        0xB232: "HANGUL",
        0xB233: "HANGUL",
        0xB234: "HANGUL",
        0xB235: "HANGUL",
        0xB236: "HANGUL",
        0xB237: "HANGUL",
        0xB238: "HANGUL",
        0xB239: "HANGUL",
        0xB23A: "HANGUL",
        0xB23B: "HANGUL",
        0xB23C: "HANGUL",
        0xB23D: "HANGUL",
        0xB23E: "HANGUL",
        0xB23F: "HANGUL",
        0xB240: "HANGUL",
        0xB241: "HANGUL",
        0xB242: "HANGUL",
        0xB243: "HANGUL",
        0xB244: "HANGUL",
        0xB245: "HANGUL",
        0xB246: "HANGUL",
        0xB247: "HANGUL",
        0xB248: "HANGUL",
        0xB249: "HANGUL",
        0xB24A: "HANGUL",
        0xB24B: "HANGUL",
        0xB24C: "HANGUL",
        0xB24D: "HANGUL",
        0xB24E: "HANGUL",
        0xB24F: "HANGUL",
        0xB250: "HANGUL",
        0xB251: "HANGUL",
        0xB252: "HANGUL",
        0xB253: "HANGUL",
        0xB254: "HANGUL",
        0xB255: "HANGUL",
        0xB256: "HANGUL",
        0xB257: "HANGUL",
        0xB258: "HANGUL",
        0xB259: "HANGUL",
        0xB25A: "HANGUL",
        0xB25B: "HANGUL",
        0xB25C: "HANGUL",
        0xB25D: "HANGUL",
        0xB25E: "HANGUL",
        0xB25F: "HANGUL",
        0xB260: "HANGUL",
        0xB261: "HANGUL",
        0xB262: "HANGUL",
        0xB263: "HANGUL",
        0xB264: "HANGUL",
        0xB265: "HANGUL",
        0xB266: "HANGUL",
        0xB267: "HANGUL",
        0xB268: "HANGUL",
        0xB269: "HANGUL",
        0xB26A: "HANGUL",
        0xB26B: "HANGUL",
        0xB26C: "HANGUL",
        0xB26D: "HANGUL",
        0xB26E: "HANGUL",
        0xB26F: "HANGUL",
        0xB270: "HANGUL",
        0xB271: "HANGUL",
        0xB272: "HANGUL",
        0xB273: "HANGUL",
        0xB274: "HANGUL",
        0xB275: "HANGUL",
        0xB276: "HANGUL",
        0xB277: "HANGUL",
        0xB278: "HANGUL",
        0xB279: "HANGUL",
        0xB27A: "HANGUL",
        0xB27B: "HANGUL",
        0xB27C: "HANGUL",
        0xB27D: "HANGUL",
        0xB27E: "HANGUL",
        0xB27F: "HANGUL",
        0xB280: "HANGUL",
        0xB281: "HANGUL",
        0xB282: "HANGUL",
        0xB283: "HANGUL",
        0xB284: "HANGUL",
        0xB285: "HANGUL",
        0xB286: "HANGUL",
        0xB287: "HANGUL",
        0xB288: "HANGUL",
        0xB289: "HANGUL",
        0xB28A: "HANGUL",
        0xB28B: "HANGUL",
        0xB28C: "HANGUL",
        0xB28D: "HANGUL",
        0xB28E: "HANGUL",
        0xB28F: "HANGUL",
        0xB290: "HANGUL",
        0xB291: "HANGUL",
        0xB292: "HANGUL",
        0xB293: "HANGUL",
        0xB294: "HANGUL",
        0xB295: "HANGUL",
        0xB296: "HANGUL",
        0xB297: "HANGUL",
        0xB298: "HANGUL",
        0xB299: "HANGUL",
        0xB29A: "HANGUL",
        0xB29B: "HANGUL",
        0xB29C: "HANGUL",
        0xB29D: "HANGUL",
        0xB29E: "HANGUL",
        0xB29F: "HANGUL",
        0xB2A0: "HANGUL",
        0xB2A1: "HANGUL",
        0xB2A2: "HANGUL",
        0xB2A3: "HANGUL",
        0xB2A4: "HANGUL",
        0xB2A5: "HANGUL",
        0xB2A6: "HANGUL",
        0xB2A7: "HANGUL",
        0xB2A8: "HANGUL",
        0xB2A9: "HANGUL",
        0xB2AA: "HANGUL",
        0xB2AB: "HANGUL",
        0xB2AC: "HANGUL",
        0xB2AD: "HANGUL",
        0xB2AE: "HANGUL",
        0xB2AF: "HANGUL",
        0xB2B0: "HANGUL",
        0xB2B1: "HANGUL",
        0xB2B2: "HANGUL",
        0xB2B3: "HANGUL",
        0xB2B4: "HANGUL",
        0xB2B5: "HANGUL",
        0xB2B6: "HANGUL",
        0xB2B7: "HANGUL",
        0xB2B8: "HANGUL",
        0xB2B9: "HANGUL",
        0xB2BA: "HANGUL",
        0xB2BB: "HANGUL",
        0xB2BC: "HANGUL",
        0xB2BD: "HANGUL",
        0xB2BE: "HANGUL",
        0xB2BF: "HANGUL",
        0xB2C0: "HANGUL",
        0xB2C1: "HANGUL",
        0xB2C2: "HANGUL",
        0xB2C3: "HANGUL",
        0xB2C4: "HANGUL",
        0xB2C5: "HANGUL",
        0xB2C6: "HANGUL",
        0xB2C7: "HANGUL",
        0xB2C8: "HANGUL",
        0xB2C9: "HANGUL",
        0xB2CA: "HANGUL",
        0xB2CB: "HANGUL",
        0xB2CC: "HANGUL",
        0xB2CD: "HANGUL",
        0xB2CE: "HANGUL",
        0xB2CF: "HANGUL",
        0xB2D0: "HANGUL",
        0xB2D1: "HANGUL",
        0xB2D2: "HANGUL",
        0xB2D3: "HANGUL",
        0xB2D4: "HANGUL",
        0xB2D5: "HANGUL",
        0xB2D6: "HANGUL",
        0xB2D7: "HANGUL",
        0xB2D8: "HANGUL",
        0xB2D9: "HANGUL",
        0xB2DA: "HANGUL",
        0xB2DB: "HANGUL",
        0xB2DC: "HANGUL",
        0xB2DD: "HANGUL",
        0xB2DE: "HANGUL",
        0xB2DF: "HANGUL",
        0xB2E0: "HANGUL",
        0xB2E1: "HANGUL",
        0xB2E2: "HANGUL",
        0xB2E3: "HANGUL",
        0xB2E4: "HANGUL",
        0xB2E5: "HANGUL",
        0xB2E6: "HANGUL",
        0xB2E7: "HANGUL",
        0xB2E8: "HANGUL",
        0xB2E9: "HANGUL",
        0xB2EA: "HANGUL",
        0xB2EB: "HANGUL",
        0xB2EC: "HANGUL",
        0xB2ED: "HANGUL",
        0xB2EE: "HANGUL",
        0xB2EF: "HANGUL",
        0xB2F0: "HANGUL",
        0xB2F1: "HANGUL",
        0xB2F2: "HANGUL",
        0xB2F3: "HANGUL",
        0xB2F4: "HANGUL",
        0xB2F5: "HANGUL",
        0xB2F6: "HANGUL",
        0xB2F7: "HANGUL",
        0xB2F8: "HANGUL",
        0xB2F9: "HANGUL",
        0xB2FA: "HANGUL",
        0xB2FB: "HANGUL",
        0xB2FC: "HANGUL",
        0xB2FD: "HANGUL",
        0xB2FE: "HANGUL",
        0xB2FF: "HANGUL",
        0xB300: "HANGUL",
        0xB301: "HANGUL",
        0xB302: "HANGUL",
        0xB303: "HANGUL",
        0xB304: "HANGUL",
        0xB305: "HANGUL",
        0xB306: "HANGUL",
        0xB307: "HANGUL",
        0xB308: "HANGUL",
        0xB309: "HANGUL",
        0xB30A: "HANGUL",
        0xB30B: "HANGUL",
        0xB30C: "HANGUL",
        0xB30D: "HANGUL",
        0xB30E: "HANGUL",
        0xB30F: "HANGUL",
        0xB310: "HANGUL",
        0xB311: "HANGUL",
        0xB312: "HANGUL",
        0xB313: "HANGUL",
        0xB314: "HANGUL",
        0xB315: "HANGUL",
        0xB316: "HANGUL",
        0xB317: "HANGUL",
        0xB318: "HANGUL",
        0xB319: "HANGUL",
        0xB31A: "HANGUL",
        0xB31B: "HANGUL",
        0xB31C: "HANGUL",
        0xB31D: "HANGUL",
        0xB31E: "HANGUL",
        0xB31F: "HANGUL",
        0xB320: "HANGUL",
        0xB321: "HANGUL",
        0xB322: "HANGUL",
        0xB323: "HANGUL",
        0xB324: "HANGUL",
        0xB325: "HANGUL",
        0xB326: "HANGUL",
        0xB327: "HANGUL",
        0xB328: "HANGUL",
        0xB329: "HANGUL",
        0xB32A: "HANGUL",
        0xB32B: "HANGUL",
        0xB32C: "HANGUL",
        0xB32D: "HANGUL",
        0xB32E: "HANGUL",
        0xB32F: "HANGUL",
        0xB330: "HANGUL",
        0xB331: "HANGUL",
        0xB332: "HANGUL",
        0xB333: "HANGUL",
        0xB334: "HANGUL",
        0xB335: "HANGUL",
        0xB336: "HANGUL",
        0xB337: "HANGUL",
        0xB338: "HANGUL",
        0xB339: "HANGUL",
        0xB33A: "HANGUL",
        0xB33B: "HANGUL",
        0xB33C: "HANGUL",
        0xB33D: "HANGUL",
        0xB33E: "HANGUL",
        0xB33F: "HANGUL",
        0xB340: "HANGUL",
        0xB341: "HANGUL",
        0xB342: "HANGUL",
        0xB343: "HANGUL",
        0xB344: "HANGUL",
        0xB345: "HANGUL",
        0xB346: "HANGUL",
        0xB347: "HANGUL",
        0xB348: "HANGUL",
        0xB349: "HANGUL",
        0xB34A: "HANGUL",
        0xB34B: "HANGUL",
        0xB34C: "HANGUL",
        0xB34D: "HANGUL",
        0xB34E: "HANGUL",
        0xB34F: "HANGUL",
        0xB350: "HANGUL",
        0xB351: "HANGUL",
        0xB352: "HANGUL",
        0xB353: "HANGUL",
        0xB354: "HANGUL",
        0xB355: "HANGUL",
        0xB356: "HANGUL",
        0xB357: "HANGUL",
        0xB358: "HANGUL",
        0xB359: "HANGUL",
        0xB35A: "HANGUL",
        0xB35B: "HANGUL",
        0xB35C: "HANGUL",
        0xB35D: "HANGUL",
        0xB35E: "HANGUL",
        0xB35F: "HANGUL",
        0xB360: "HANGUL",
        0xB361: "HANGUL",
        0xB362: "HANGUL",
        0xB363: "HANGUL",
        0xB364: "HANGUL",
        0xB365: "HANGUL",
        0xB366: "HANGUL",
        0xB367: "HANGUL",
        0xB368: "HANGUL",
        0xB369: "HANGUL",
        0xB36A: "HANGUL",
        0xB36B: "HANGUL",
        0xB36C: "HANGUL",
        0xB36D: "HANGUL",
        0xB36E: "HANGUL",
        0xB36F: "HANGUL",
        0xB370: "HANGUL",
        0xB371: "HANGUL",
        0xB372: "HANGUL",
        0xB373: "HANGUL",
        0xB374: "HANGUL",
        0xB375: "HANGUL",
        0xB376: "HANGUL",
        0xB377: "HANGUL",
        0xB378: "HANGUL",
        0xB379: "HANGUL",
        0xB37A: "HANGUL",
        0xB37B: "HANGUL",
        0xB37C: "HANGUL",
        0xB37D: "HANGUL",
        0xB37E: "HANGUL",
        0xB37F: "HANGUL",
        0xB380: "HANGUL",
        0xB381: "HANGUL",
        0xB382: "HANGUL",
        0xB383: "HANGUL",
        0xB384: "HANGUL",
        0xB385: "HANGUL",
        0xB386: "HANGUL",
        0xB387: "HANGUL",
        0xB388: "HANGUL",
        0xB389: "HANGUL",
        0xB38A: "HANGUL",
        0xB38B: "HANGUL",
        0xB38C: "HANGUL",
        0xB38D: "HANGUL",
        0xB38E: "HANGUL",
        0xB38F: "HANGUL",
        0xB390: "HANGUL",
        0xB391: "HANGUL",
        0xB392: "HANGUL",
        0xB393: "HANGUL",
        0xB394: "HANGUL",
        0xB395: "HANGUL",
        0xB396: "HANGUL",
        0xB397: "HANGUL",
        0xB398: "HANGUL",
        0xB399: "HANGUL",
        0xB39A: "HANGUL",
        0xB39B: "HANGUL",
        0xB39C: "HANGUL",
        0xB39D: "HANGUL",
        0xB39E: "HANGUL",
        0xB39F: "HANGUL",
        0xB3A0: "HANGUL",
        0xB3A1: "HANGUL",
        0xB3A2: "HANGUL",
        0xB3A3: "HANGUL",
        0xB3A4: "HANGUL",
        0xB3A5: "HANGUL",
        0xB3A6: "HANGUL",
        0xB3A7: "HANGUL",
        0xB3A8: "HANGUL",
        0xB3A9: "HANGUL",
        0xB3AA: "HANGUL",
        0xB3AB: "HANGUL",
        0xB3AC: "HANGUL",
        0xB3AD: "HANGUL",
        0xB3AE: "HANGUL",
        0xB3AF: "HANGUL",
        0xB3B0: "HANGUL",
        0xB3B1: "HANGUL",
        0xB3B2: "HANGUL",
        0xB3B3: "HANGUL",
        0xB3B4: "HANGUL",
        0xB3B5: "HANGUL",
        0xB3B6: "HANGUL",
        0xB3B7: "HANGUL",
        0xB3B8: "HANGUL",
        0xB3B9: "HANGUL",
        0xB3BA: "HANGUL",
        0xB3BB: "HANGUL",
        0xB3BC: "HANGUL",
        0xB3BD: "HANGUL",
        0xB3BE: "HANGUL",
        0xB3BF: "HANGUL",
        0xB3C0: "HANGUL",
        0xB3C1: "HANGUL",
        0xB3C2: "HANGUL",
        0xB3C3: "HANGUL",
        0xB3C4: "HANGUL",
        0xB3C5: "HANGUL",
        0xB3C6: "HANGUL",
        0xB3C7: "HANGUL",
        0xB3C8: "HANGUL",
        0xB3C9: "HANGUL",
        0xB3CA: "HANGUL",
        0xB3CB: "HANGUL",
        0xB3CC: "HANGUL",
        0xB3CD: "HANGUL",
        0xB3CE: "HANGUL",
        0xB3CF: "HANGUL",
        0xB3D0: "HANGUL",
        0xB3D1: "HANGUL",
        0xB3D2: "HANGUL",
        0xB3D3: "HANGUL",
        0xB3D4: "HANGUL",
        0xB3D5: "HANGUL",
        0xB3D6: "HANGUL",
        0xB3D7: "HANGUL",
        0xB3D8: "HANGUL",
        0xB3D9: "HANGUL",
        0xB3DA: "HANGUL",
        0xB3DB: "HANGUL",
        0xB3DC: "HANGUL",
        0xB3DD: "HANGUL",
        0xB3DE: "HANGUL",
        0xB3DF: "HANGUL",
        0xB3E0: "HANGUL",
        0xB3E1: "HANGUL",
        0xB3E2: "HANGUL",
        0xB3E3: "HANGUL",
        0xB3E4: "HANGUL",
        0xB3E5: "HANGUL",
        0xB3E6: "HANGUL",
        0xB3E7: "HANGUL",
        0xB3E8: "HANGUL",
        0xB3E9: "HANGUL",
        0xB3EA: "HANGUL",
        0xB3EB: "HANGUL",
        0xB3EC: "HANGUL",
        0xB3ED: "HANGUL",
        0xB3EE: "HANGUL",
        0xB3EF: "HANGUL",
        0xB3F0: "HANGUL",
        0xB3F1: "HANGUL",
        0xB3F2: "HANGUL",
        0xB3F3: "HANGUL",
        0xB3F4: "HANGUL",
        0xB3F5: "HANGUL",
        0xB3F6: "HANGUL",
        0xB3F7: "HANGUL",
        0xB3F8: "HANGUL",
        0xB3F9: "HANGUL",
        0xB3FA: "HANGUL",
        0xB3FB: "HANGUL",
        0xB3FC: "HANGUL",
        0xB3FD: "HANGUL",
        0xB3FE: "HANGUL",
        0xB3FF: "HANGUL",
        0xB400: "HANGUL",
        0xB401: "HANGUL",
        0xB402: "HANGUL",
        0xB403: "HANGUL",
        0xB404: "HANGUL",
        0xB405: "HANGUL",
        0xB406: "HANGUL",
        0xB407: "HANGUL",
        0xB408: "HANGUL",
        0xB409: "HANGUL",
        0xB40A: "HANGUL",
        0xB40B: "HANGUL",
        0xB40C: "HANGUL",
        0xB40D: "HANGUL",
        0xB40E: "HANGUL",
        0xB40F: "HANGUL",
        0xB410: "HANGUL",
        0xB411: "HANGUL",
        0xB412: "HANGUL",
        0xB413: "HANGUL",
        0xB414: "HANGUL",
        0xB415: "HANGUL",
        0xB416: "HANGUL",
        0xB417: "HANGUL",
        0xB418: "HANGUL",
        0xB419: "HANGUL",
        0xB41A: "HANGUL",
        0xB41B: "HANGUL",
        0xB41C: "HANGUL",
        0xB41D: "HANGUL",
        0xB41E: "HANGUL",
        0xB41F: "HANGUL",
        0xB420: "HANGUL",
        0xB421: "HANGUL",
        0xB422: "HANGUL",
        0xB423: "HANGUL",
        0xB424: "HANGUL",
        0xB425: "HANGUL",
        0xB426: "HANGUL",
        0xB427: "HANGUL",
        0xB428: "HANGUL",
        0xB429: "HANGUL",
        0xB42A: "HANGUL",
        0xB42B: "HANGUL",
        0xB42C: "HANGUL",
        0xB42D: "HANGUL",
        0xB42E: "HANGUL",
        0xB42F: "HANGUL",
        0xB430: "HANGUL",
        0xB431: "HANGUL",
        0xB432: "HANGUL",
        0xB433: "HANGUL",
        0xB434: "HANGUL",
        0xB435: "HANGUL",
        0xB436: "HANGUL",
        0xB437: "HANGUL",
        0xB438: "HANGUL",
        0xB439: "HANGUL",
        0xB43A: "HANGUL",
        0xB43B: "HANGUL",
        0xB43C: "HANGUL",
        0xB43D: "HANGUL",
        0xB43E: "HANGUL",
        0xB43F: "HANGUL",
        0xB440: "HANGUL",
        0xB441: "HANGUL",
        0xB442: "HANGUL",
        0xB443: "HANGUL",
        0xB444: "HANGUL",
        0xB445: "HANGUL",
        0xB446: "HANGUL",
        0xB447: "HANGUL",
        0xB448: "HANGUL",
        0xB449: "HANGUL",
        0xB44A: "HANGUL",
        0xB44B: "HANGUL",
        0xB44C: "HANGUL",
        0xB44D: "HANGUL",
        0xB44E: "HANGUL",
        0xB44F: "HANGUL",
        0xB450: "HANGUL",
        0xB451: "HANGUL",
        0xB452: "HANGUL",
        0xB453: "HANGUL",
        0xB454: "HANGUL",
        0xB455: "HANGUL",
        0xB456: "HANGUL",
        0xB457: "HANGUL",
        0xB458: "HANGUL",
        0xB459: "HANGUL",
        0xB45A: "HANGUL",
        0xB45B: "HANGUL",
        0xB45C: "HANGUL",
        0xB45D: "HANGUL",
        0xB45E: "HANGUL",
        0xB45F: "HANGUL",
        0xB460: "HANGUL",
        0xB461: "HANGUL",
        0xB462: "HANGUL",
        0xB463: "HANGUL",
        0xB464: "HANGUL",
        0xB465: "HANGUL",
        0xB466: "HANGUL",
        0xB467: "HANGUL",
        0xB468: "HANGUL",
        0xB469: "HANGUL",
        0xB46A: "HANGUL",
        0xB46B: "HANGUL",
        0xB46C: "HANGUL",
        0xB46D: "HANGUL",
        0xB46E: "HANGUL",
        0xB46F: "HANGUL",
        0xB470: "HANGUL",
        0xB471: "HANGUL",
        0xB472: "HANGUL",
        0xB473: "HANGUL",
        0xB474: "HANGUL",
        0xB475: "HANGUL",
        0xB476: "HANGUL",
        0xB477: "HANGUL",
        0xB478: "HANGUL",
        0xB479: "HANGUL",
        0xB47A: "HANGUL",
        0xB47B: "HANGUL",
        0xB47C: "HANGUL",
        0xB47D: "HANGUL",
        0xB47E: "HANGUL",
        0xB47F: "HANGUL",
        0xB480: "HANGUL",
        0xB481: "HANGUL",
        0xB482: "HANGUL",
        0xB483: "HANGUL",
        0xB484: "HANGUL",
        0xB485: "HANGUL",
        0xB486: "HANGUL",
        0xB487: "HANGUL",
        0xB488: "HANGUL",
        0xB489: "HANGUL",
        0xB48A: "HANGUL",
        0xB48B: "HANGUL",
        0xB48C: "HANGUL",
        0xB48D: "HANGUL",
        0xB48E: "HANGUL",
        0xB48F: "HANGUL",
        0xB490: "HANGUL",
        0xB491: "HANGUL",
        0xB492: "HANGUL",
        0xB493: "HANGUL",
        0xB494: "HANGUL",
        0xB495: "HANGUL",
        0xB496: "HANGUL",
        0xB497: "HANGUL",
        0xB498: "HANGUL",
        0xB499: "HANGUL",
        0xB49A: "HANGUL",
        0xB49B: "HANGUL",
        0xB49C: "HANGUL",
        0xB49D: "HANGUL",
        0xB49E: "HANGUL",
        0xB49F: "HANGUL",
        0xB4A0: "HANGUL",
        0xB4A1: "HANGUL",
        0xB4A2: "HANGUL",
        0xB4A3: "HANGUL",
        0xB4A4: "HANGUL",
        0xB4A5: "HANGUL",
        0xB4A6: "HANGUL",
        0xB4A7: "HANGUL",
        0xB4A8: "HANGUL",
        0xB4A9: "HANGUL",
        0xB4AA: "HANGUL",
        0xB4AB: "HANGUL",
        0xB4AC: "HANGUL",
        0xB4AD: "HANGUL",
        0xB4AE: "HANGUL",
        0xB4AF: "HANGUL",
        0xB4B0: "HANGUL",
        0xB4B1: "HANGUL",
        0xB4B2: "HANGUL",
        0xB4B3: "HANGUL",
        0xB4B4: "HANGUL",
        0xB4B5: "HANGUL",
        0xB4B6: "HANGUL",
        0xB4B7: "HANGUL",
        0xB4B8: "HANGUL",
        0xB4B9: "HANGUL",
        0xB4BA: "HANGUL",
        0xB4BB: "HANGUL",
        0xB4BC: "HANGUL",
        0xB4BD: "HANGUL",
        0xB4BE: "HANGUL",
        0xB4BF: "HANGUL",
        0xB4C0: "HANGUL",
        0xB4C1: "HANGUL",
        0xB4C2: "HANGUL",
        0xB4C3: "HANGUL",
        0xB4C4: "HANGUL",
        0xB4C5: "HANGUL",
        0xB4C6: "HANGUL",
        0xB4C7: "HANGUL",
        0xB4C8: "HANGUL",
        0xB4C9: "HANGUL",
        0xB4CA: "HANGUL",
        0xB4CB: "HANGUL",
        0xB4CC: "HANGUL",
        0xB4CD: "HANGUL",
        0xB4CE: "HANGUL",
        0xB4CF: "HANGUL",
        0xB4D0: "HANGUL",
        0xB4D1: "HANGUL",
        0xB4D2: "HANGUL",
        0xB4D3: "HANGUL",
        0xB4D4: "HANGUL",
        0xB4D5: "HANGUL",
        0xB4D6: "HANGUL",
        0xB4D7: "HANGUL",
        0xB4D8: "HANGUL",
        0xB4D9: "HANGUL",
        0xB4DA: "HANGUL",
        0xB4DB: "HANGUL",
        0xB4DC: "HANGUL",
        0xB4DD: "HANGUL",
        0xB4DE: "HANGUL",
        0xB4DF: "HANGUL",
        0xB4E0: "HANGUL",
        0xB4E1: "HANGUL",
        0xB4E2: "HANGUL",
        0xB4E3: "HANGUL",
        0xB4E4: "HANGUL",
        0xB4E5: "HANGUL",
        0xB4E6: "HANGUL",
        0xB4E7: "HANGUL",
        0xB4E8: "HANGUL",
        0xB4E9: "HANGUL",
        0xB4EA: "HANGUL",
        0xB4EB: "HANGUL",
        0xB4EC: "HANGUL",
        0xB4ED: "HANGUL",
        0xB4EE: "HANGUL",
        0xB4EF: "HANGUL",
        0xB4F0: "HANGUL",
        0xB4F1: "HANGUL",
        0xB4F2: "HANGUL",
        0xB4F3: "HANGUL",
        0xB4F4: "HANGUL",
        0xB4F5: "HANGUL",
        0xB4F6: "HANGUL",
        0xB4F7: "HANGUL",
        0xB4F8: "HANGUL",
        0xB4F9: "HANGUL",
        0xB4FA: "HANGUL",
        0xB4FB: "HANGUL",
        0xB4FC: "HANGUL",
        0xB4FD: "HANGUL",
        0xB4FE: "HANGUL",
        0xB4FF: "HANGUL",
        0xB500: "HANGUL",
        0xB501: "HANGUL",
        0xB502: "HANGUL",
        0xB503: "HANGUL",
        0xB504: "HANGUL",
        0xB505: "HANGUL",
        0xB506: "HANGUL",
        0xB507: "HANGUL",
        0xB508: "HANGUL",
        0xB509: "HANGUL",
        0xB50A: "HANGUL",
        0xB50B: "HANGUL",
        0xB50C: "HANGUL",
        0xB50D: "HANGUL",
        0xB50E: "HANGUL",
        0xB50F: "HANGUL",
        0xB510: "HANGUL",
        0xB511: "HANGUL",
        0xB512: "HANGUL",
        0xB513: "HANGUL",
        0xB514: "HANGUL",
        0xB515: "HANGUL",
        0xB516: "HANGUL",
        0xB517: "HANGUL",
        0xB518: "HANGUL",
        0xB519: "HANGUL",
        0xB51A: "HANGUL",
        0xB51B: "HANGUL",
        0xB51C: "HANGUL",
        0xB51D: "HANGUL",
        0xB51E: "HANGUL",
        0xB51F: "HANGUL",
        0xB520: "HANGUL",
        0xB521: "HANGUL",
        0xB522: "HANGUL",
        0xB523: "HANGUL",
        0xB524: "HANGUL",
        0xB525: "HANGUL",
        0xB526: "HANGUL",
        0xB527: "HANGUL",
        0xB528: "HANGUL",
        0xB529: "HANGUL",
        0xB52A: "HANGUL",
        0xB52B: "HANGUL",
        0xB52C: "HANGUL",
        0xB52D: "HANGUL",
        0xB52E: "HANGUL",
        0xB52F: "HANGUL",
        0xB530: "HANGUL",
        0xB531: "HANGUL",
        0xB532: "HANGUL",
        0xB533: "HANGUL",
        0xB534: "HANGUL",
        0xB535: "HANGUL",
        0xB536: "HANGUL",
        0xB537: "HANGUL",
        0xB538: "HANGUL",
        0xB539: "HANGUL",
        0xB53A: "HANGUL",
        0xB53B: "HANGUL",
        0xB53C: "HANGUL",
        0xB53D: "HANGUL",
        0xB53E: "HANGUL",
        0xB53F: "HANGUL",
        0xB540: "HANGUL",
        0xB541: "HANGUL",
        0xB542: "HANGUL",
        0xB543: "HANGUL",
        0xB544: "HANGUL",
        0xB545: "HANGUL",
        0xB546: "HANGUL",
        0xB547: "HANGUL",
        0xB548: "HANGUL",
        0xB549: "HANGUL",
        0xB54A: "HANGUL",
        0xB54B: "HANGUL",
        0xB54C: "HANGUL",
        0xB54D: "HANGUL",
        0xB54E: "HANGUL",
        0xB54F: "HANGUL",
        0xB550: "HANGUL",
        0xB551: "HANGUL",
        0xB552: "HANGUL",
        0xB553: "HANGUL",
        0xB554: "HANGUL",
        0xB555: "HANGUL",
        0xB556: "HANGUL",
        0xB557: "HANGUL",
        0xB558: "HANGUL",
        0xB559: "HANGUL",
        0xB55A: "HANGUL",
        0xB55B: "HANGUL",
        0xB55C: "HANGUL",
        0xB55D: "HANGUL",
        0xB55E: "HANGUL",
        0xB55F: "HANGUL",
        0xB560: "HANGUL",
        0xB561: "HANGUL",
        0xB562: "HANGUL",
        0xB563: "HANGUL",
        0xB564: "HANGUL",
        0xB565: "HANGUL",
        0xB566: "HANGUL",
        0xB567: "HANGUL",
        0xB568: "HANGUL",
        0xB569: "HANGUL",
        0xB56A: "HANGUL",
        0xB56B: "HANGUL",
        0xB56C: "HANGUL",
        0xB56D: "HANGUL",
        0xB56E: "HANGUL",
        0xB56F: "HANGUL",
        0xB570: "HANGUL",
        0xB571: "HANGUL",
        0xB572: "HANGUL",
        0xB573: "HANGUL",
        0xB574: "HANGUL",
        0xB575: "HANGUL",
        0xB576: "HANGUL",
        0xB577: "HANGUL",
        0xB578: "HANGUL",
        0xB579: "HANGUL",
        0xB57A: "HANGUL",
        0xB57B: "HANGUL",
        0xB57C: "HANGUL",
        0xB57D: "HANGUL",
        0xB57E: "HANGUL",
        0xB57F: "HANGUL",
        0xB580: "HANGUL",
        0xB581: "HANGUL",
        0xB582: "HANGUL",
        0xB583: "HANGUL",
        0xB584: "HANGUL",
        0xB585: "HANGUL",
        0xB586: "HANGUL",
        0xB587: "HANGUL",
        0xB588: "HANGUL",
        0xB589: "HANGUL",
        0xB58A: "HANGUL",
        0xB58B: "HANGUL",
        0xB58C: "HANGUL",
        0xB58D: "HANGUL",
        0xB58E: "HANGUL",
        0xB58F: "HANGUL",
        0xB590: "HANGUL",
        0xB591: "HANGUL",
        0xB592: "HANGUL",
        0xB593: "HANGUL",
        0xB594: "HANGUL",
        0xB595: "HANGUL",
        0xB596: "HANGUL",
        0xB597: "HANGUL",
        0xB598: "HANGUL",
        0xB599: "HANGUL",
        0xB59A: "HANGUL",
        0xB59B: "HANGUL",
        0xB59C: "HANGUL",
        0xB59D: "HANGUL",
        0xB59E: "HANGUL",
        0xB59F: "HANGUL",
        0xB5A0: "HANGUL",
        0xB5A1: "HANGUL",
        0xB5A2: "HANGUL",
        0xB5A3: "HANGUL",
        0xB5A4: "HANGUL",
        0xB5A5: "HANGUL",
        0xB5A6: "HANGUL",
        0xB5A7: "HANGUL",
        0xB5A8: "HANGUL",
        0xB5A9: "HANGUL",
        0xB5AA: "HANGUL",
        0xB5AB: "HANGUL",
        0xB5AC: "HANGUL",
        0xB5AD: "HANGUL",
        0xB5AE: "HANGUL",
        0xB5AF: "HANGUL",
        0xB5B0: "HANGUL",
        0xB5B1: "HANGUL",
        0xB5B2: "HANGUL",
        0xB5B3: "HANGUL",
        0xB5B4: "HANGUL",
        0xB5B5: "HANGUL",
        0xB5B6: "HANGUL",
        0xB5B7: "HANGUL",
        0xB5B8: "HANGUL",
        0xB5B9: "HANGUL",
        0xB5BA: "HANGUL",
        0xB5BB: "HANGUL",
        0xB5BC: "HANGUL",
        0xB5BD: "HANGUL",
        0xB5BE: "HANGUL",
        0xB5BF: "HANGUL",
        0xB5C0: "HANGUL",
        0xB5C1: "HANGUL",
        0xB5C2: "HANGUL",
        0xB5C3: "HANGUL",
        0xB5C4: "HANGUL",
        0xB5C5: "HANGUL",
        0xB5C6: "HANGUL",
        0xB5C7: "HANGUL",
        0xB5C8: "HANGUL",
        0xB5C9: "HANGUL",
        0xB5CA: "HANGUL",
        0xB5CB: "HANGUL",
        0xB5CC: "HANGUL",
        0xB5CD: "HANGUL",
        0xB5CE: "HANGUL",
        0xB5CF: "HANGUL",
        0xB5D0: "HANGUL",
        0xB5D1: "HANGUL",
        0xB5D2: "HANGUL",
        0xB5D3: "HANGUL",
        0xB5D4: "HANGUL",
        0xB5D5: "HANGUL",
        0xB5D6: "HANGUL",
        0xB5D7: "HANGUL",
        0xB5D8: "HANGUL",
        0xB5D9: "HANGUL",
        0xB5DA: "HANGUL",
        0xB5DB: "HANGUL",
        0xB5DC: "HANGUL",
        0xB5DD: "HANGUL",
        0xB5DE: "HANGUL",
        0xB5DF: "HANGUL",
        0xB5E0: "HANGUL",
        0xB5E1: "HANGUL",
        0xB5E2: "HANGUL",
        0xB5E3: "HANGUL",
        0xB5E4: "HANGUL",
        0xB5E5: "HANGUL",
        0xB5E6: "HANGUL",
        0xB5E7: "HANGUL",
        0xB5E8: "HANGUL",
        0xB5E9: "HANGUL",
        0xB5EA: "HANGUL",
        0xB5EB: "HANGUL",
        0xB5EC: "HANGUL",
        0xB5ED: "HANGUL",
        0xB5EE: "HANGUL",
        0xB5EF: "HANGUL",
        0xB5F0: "HANGUL",
        0xB5F1: "HANGUL",
        0xB5F2: "HANGUL",
        0xB5F3: "HANGUL",
        0xB5F4: "HANGUL",
        0xB5F5: "HANGUL",
        0xB5F6: "HANGUL",
        0xB5F7: "HANGUL",
        0xB5F8: "HANGUL",
        0xB5F9: "HANGUL",
        0xB5FA: "HANGUL",
        0xB5FB: "HANGUL",
        0xB5FC: "HANGUL",
        0xB5FD: "HANGUL",
        0xB5FE: "HANGUL",
        0xB5FF: "HANGUL",
        0xB600: "HANGUL",
        0xB601: "HANGUL",
        0xB602: "HANGUL",
        0xB603: "HANGUL",
        0xB604: "HANGUL",
        0xB605: "HANGUL",
        0xB606: "HANGUL",
        0xB607: "HANGUL",
        0xB608: "HANGUL",
        0xB609: "HANGUL",
        0xB60A: "HANGUL",
        0xB60B: "HANGUL",
        0xB60C: "HANGUL",
        0xB60D: "HANGUL",
        0xB60E: "HANGUL",
        0xB60F: "HANGUL",
        0xB610: "HANGUL",
        0xB611: "HANGUL",
        0xB612: "HANGUL",
        0xB613: "HANGUL",
        0xB614: "HANGUL",
        0xB615: "HANGUL",
        0xB616: "HANGUL",
        0xB617: "HANGUL",
        0xB618: "HANGUL",
        0xB619: "HANGUL",
        0xB61A: "HANGUL",
        0xB61B: "HANGUL",
        0xB61C: "HANGUL",
        0xB61D: "HANGUL",
        0xB61E: "HANGUL",
        0xB61F: "HANGUL",
        0xB620: "HANGUL",
        0xB621: "HANGUL",
        0xB622: "HANGUL",
        0xB623: "HANGUL",
        0xB624: "HANGUL",
        0xB625: "HANGUL",
        0xB626: "HANGUL",
        0xB627: "HANGUL",
        0xB628: "HANGUL",
        0xB629: "HANGUL",
        0xB62A: "HANGUL",
        0xB62B: "HANGUL",
        0xB62C: "HANGUL",
        0xB62D: "HANGUL",
        0xB62E: "HANGUL",
        0xB62F: "HANGUL",
        0xB630: "HANGUL",
        0xB631: "HANGUL",
        0xB632: "HANGUL",
        0xB633: "HANGUL",
        0xB634: "HANGUL",
        0xB635: "HANGUL",
        0xB636: "HANGUL",
        0xB637: "HANGUL",
        0xB638: "HANGUL",
        0xB639: "HANGUL",
        0xB63A: "HANGUL",
        0xB63B: "HANGUL",
        0xB63C: "HANGUL",
        0xB63D: "HANGUL",
        0xB63E: "HANGUL",
        0xB63F: "HANGUL",
        0xB640: "HANGUL",
        0xB641: "HANGUL",
        0xB642: "HANGUL",
        0xB643: "HANGUL",
        0xB644: "HANGUL",
        0xB645: "HANGUL",
        0xB646: "HANGUL",
        0xB647: "HANGUL",
        0xB648: "HANGUL",
        0xB649: "HANGUL",
        0xB64A: "HANGUL",
        0xB64B: "HANGUL",
        0xB64C: "HANGUL",
        0xB64D: "HANGUL",
        0xB64E: "HANGUL",
        0xB64F: "HANGUL",
        0xB650: "HANGUL",
        0xB651: "HANGUL",
        0xB652: "HANGUL",
        0xB653: "HANGUL",
        0xB654: "HANGUL",
        0xB655: "HANGUL",
        0xB656: "HANGUL",
        0xB657: "HANGUL",
        0xB658: "HANGUL",
        0xB659: "HANGUL",
        0xB65A: "HANGUL",
        0xB65B: "HANGUL",
        0xB65C: "HANGUL",
        0xB65D: "HANGUL",
        0xB65E: "HANGUL",
        0xB65F: "HANGUL",
        0xB660: "HANGUL",
        0xB661: "HANGUL",
        0xB662: "HANGUL",
        0xB663: "HANGUL",
        0xB664: "HANGUL",
        0xB665: "HANGUL",
        0xB666: "HANGUL",
        0xB667: "HANGUL",
        0xB668: "HANGUL",
        0xB669: "HANGUL",
        0xB66A: "HANGUL",
        0xB66B: "HANGUL",
        0xB66C: "HANGUL",
        0xB66D: "HANGUL",
        0xB66E: "HANGUL",
        0xB66F: "HANGUL",
        0xB670: "HANGUL",
        0xB671: "HANGUL",
        0xB672: "HANGUL",
        0xB673: "HANGUL",
        0xB674: "HANGUL",
        0xB675: "HANGUL",
        0xB676: "HANGUL",
        0xB677: "HANGUL",
        0xB678: "HANGUL",
        0xB679: "HANGUL",
        0xB67A: "HANGUL",
        0xB67B: "HANGUL",
        0xB67C: "HANGUL",
        0xB67D: "HANGUL",
        0xB67E: "HANGUL",
        0xB67F: "HANGUL",
        0xB680: "HANGUL",
        0xB681: "HANGUL",
        0xB682: "HANGUL",
        0xB683: "HANGUL",
        0xB684: "HANGUL",
        0xB685: "HANGUL",
        0xB686: "HANGUL",
        0xB687: "HANGUL",
        0xB688: "HANGUL",
        0xB689: "HANGUL",
        0xB68A: "HANGUL",
        0xB68B: "HANGUL",
        0xB68C: "HANGUL",
        0xB68D: "HANGUL",
        0xB68E: "HANGUL",
        0xB68F: "HANGUL",
        0xB690: "HANGUL",
        0xB691: "HANGUL",
        0xB692: "HANGUL",
        0xB693: "HANGUL",
        0xB694: "HANGUL",
        0xB695: "HANGUL",
        0xB696: "HANGUL",
        0xB697: "HANGUL",
        0xB698: "HANGUL",
        0xB699: "HANGUL",
        0xB69A: "HANGUL",
        0xB69B: "HANGUL",
        0xB69C: "HANGUL",
        0xB69D: "HANGUL",
        0xB69E: "HANGUL",
        0xB69F: "HANGUL",
        0xB6A0: "HANGUL",
        0xB6A1: "HANGUL",
        0xB6A2: "HANGUL",
        0xB6A3: "HANGUL",
        0xB6A4: "HANGUL",
        0xB6A5: "HANGUL",
        0xB6A6: "HANGUL",
        0xB6A7: "HANGUL",
        0xB6A8: "HANGUL",
        0xB6A9: "HANGUL",
        0xB6AA: "HANGUL",
        0xB6AB: "HANGUL",
        0xB6AC: "HANGUL",
        0xB6AD: "HANGUL",
        0xB6AE: "HANGUL",
        0xB6AF: "HANGUL",
        0xB6B0: "HANGUL",
        0xB6B1: "HANGUL",
        0xB6B2: "HANGUL",
        0xB6B3: "HANGUL",
        0xB6B4: "HANGUL",
        0xB6B5: "HANGUL",
        0xB6B6: "HANGUL",
        0xB6B7: "HANGUL",
        0xB6B8: "HANGUL",
        0xB6B9: "HANGUL",
        0xB6BA: "HANGUL",
        0xB6BB: "HANGUL",
        0xB6BC: "HANGUL",
        0xB6BD: "HANGUL",
        0xB6BE: "HANGUL",
        0xB6BF: "HANGUL",
        0xB6C0: "HANGUL",
        0xB6C1: "HANGUL",
        0xB6C2: "HANGUL",
        0xB6C3: "HANGUL",
        0xB6C4: "HANGUL",
        0xB6C5: "HANGUL",
        0xB6C6: "HANGUL",
        0xB6C7: "HANGUL",
        0xB6C8: "HANGUL",
        0xB6C9: "HANGUL",
        0xB6CA: "HANGUL",
        0xB6CB: "HANGUL",
        0xB6CC: "HANGUL",
        0xB6CD: "HANGUL",
        0xB6CE: "HANGUL",
        0xB6CF: "HANGUL",
        0xB6D0: "HANGUL",
        0xB6D1: "HANGUL",
        0xB6D2: "HANGUL",
        0xB6D3: "HANGUL",
        0xB6D4: "HANGUL",
        0xB6D5: "HANGUL",
        0xB6D6: "HANGUL",
        0xB6D7: "HANGUL",
        0xB6D8: "HANGUL",
        0xB6D9: "HANGUL",
        0xB6DA: "HANGUL",
        0xB6DB: "HANGUL",
        0xB6DC: "HANGUL",
        0xB6DD: "HANGUL",
        0xB6DE: "HANGUL",
        0xB6DF: "HANGUL",
        0xB6E0: "HANGUL",
        0xB6E1: "HANGUL",
        0xB6E2: "HANGUL",
        0xB6E3: "HANGUL",
        0xB6E4: "HANGUL",
        0xB6E5: "HANGUL",
        0xB6E6: "HANGUL",
        0xB6E7: "HANGUL",
        0xB6E8: "HANGUL",
        0xB6E9: "HANGUL",
        0xB6EA: "HANGUL",
        0xB6EB: "HANGUL",
        0xB6EC: "HANGUL",
        0xB6ED: "HANGUL",
        0xB6EE: "HANGUL",
        0xB6EF: "HANGUL",
        0xB6F0: "HANGUL",
        0xB6F1: "HANGUL",
        0xB6F2: "HANGUL",
        0xB6F3: "HANGUL",
        0xB6F4: "HANGUL",
        0xB6F5: "HANGUL",
        0xB6F6: "HANGUL",
        0xB6F7: "HANGUL",
        0xB6F8: "HANGUL",
        0xB6F9: "HANGUL",
        0xB6FA: "HANGUL",
        0xB6FB: "HANGUL",
        0xB6FC: "HANGUL",
        0xB6FD: "HANGUL",
        0xB6FE: "HANGUL",
        0xB6FF: "HANGUL",
        0xB700: "HANGUL",
        0xB701: "HANGUL",
        0xB702: "HANGUL",
        0xB703: "HANGUL",
        0xB704: "HANGUL",
        0xB705: "HANGUL",
        0xB706: "HANGUL",
        0xB707: "HANGUL",
        0xB708: "HANGUL",
        0xB709: "HANGUL",
        0xB70A: "HANGUL",
        0xB70B: "HANGUL",
        0xB70C: "HANGUL",
        0xB70D: "HANGUL",
        0xB70E: "HANGUL",
        0xB70F: "HANGUL",
        0xB710: "HANGUL",
        0xB711: "HANGUL",
        0xB712: "HANGUL",
        0xB713: "HANGUL",
        0xB714: "HANGUL",
        0xB715: "HANGUL",
        0xB716: "HANGUL",
        0xB717: "HANGUL",
        0xB718: "HANGUL",
        0xB719: "HANGUL",
        0xB71A: "HANGUL",
        0xB71B: "HANGUL",
        0xB71C: "HANGUL",
        0xB71D: "HANGUL",
        0xB71E: "HANGUL",
        0xB71F: "HANGUL",
        0xB720: "HANGUL",
        0xB721: "HANGUL",
        0xB722: "HANGUL",
        0xB723: "HANGUL",
        0xB724: "HANGUL",
        0xB725: "HANGUL",
        0xB726: "HANGUL",
        0xB727: "HANGUL",
        0xB728: "HANGUL",
        0xB729: "HANGUL",
        0xB72A: "HANGUL",
        0xB72B: "HANGUL",
        0xB72C: "HANGUL",
        0xB72D: "HANGUL",
        0xB72E: "HANGUL",
        0xB72F: "HANGUL",
        0xB730: "HANGUL",
        0xB731: "HANGUL",
        0xB732: "HANGUL",
        0xB733: "HANGUL",
        0xB734: "HANGUL",
        0xB735: "HANGUL",
        0xB736: "HANGUL",
        0xB737: "HANGUL",
        0xB738: "HANGUL",
        0xB739: "HANGUL",
        0xB73A: "HANGUL",
        0xB73B: "HANGUL",
        0xB73C: "HANGUL",
        0xB73D: "HANGUL",
        0xB73E: "HANGUL",
        0xB73F: "HANGUL",
        0xB740: "HANGUL",
        0xB741: "HANGUL",
        0xB742: "HANGUL",
        0xB743: "HANGUL",
        0xB744: "HANGUL",
        0xB745: "HANGUL",
        0xB746: "HANGUL",
        0xB747: "HANGUL",
        0xB748: "HANGUL",
        0xB749: "HANGUL",
        0xB74A: "HANGUL",
        0xB74B: "HANGUL",
        0xB74C: "HANGUL",
        0xB74D: "HANGUL",
        0xB74E: "HANGUL",
        0xB74F: "HANGUL",
        0xB750: "HANGUL",
        0xB751: "HANGUL",
        0xB752: "HANGUL",
        0xB753: "HANGUL",
        0xB754: "HANGUL",
        0xB755: "HANGUL",
        0xB756: "HANGUL",
        0xB757: "HANGUL",
        0xB758: "HANGUL",
        0xB759: "HANGUL",
        0xB75A: "HANGUL",
        0xB75B: "HANGUL",
        0xB75C: "HANGUL",
        0xB75D: "HANGUL",
        0xB75E: "HANGUL",
        0xB75F: "HANGUL",
        0xB760: "HANGUL",
        0xB761: "HANGUL",
        0xB762: "HANGUL",
        0xB763: "HANGUL",
        0xB764: "HANGUL",
        0xB765: "HANGUL",
        0xB766: "HANGUL",
        0xB767: "HANGUL",
        0xB768: "HANGUL",
        0xB769: "HANGUL",
        0xB76A: "HANGUL",
        0xB76B: "HANGUL",
        0xB76C: "HANGUL",
        0xB76D: "HANGUL",
        0xB76E: "HANGUL",
        0xB76F: "HANGUL",
        0xB770: "HANGUL",
        0xB771: "HANGUL",
        0xB772: "HANGUL",
        0xB773: "HANGUL",
        0xB774: "HANGUL",
        0xB775: "HANGUL",
        0xB776: "HANGUL",
        0xB777: "HANGUL",
        0xB778: "HANGUL",
        0xB779: "HANGUL",
        0xB77A: "HANGUL",
        0xB77B: "HANGUL",
        0xB77C: "HANGUL",
        0xB77D: "HANGUL",
        0xB77E: "HANGUL",
        0xB77F: "HANGUL",
        0xB780: "HANGUL",
        0xB781: "HANGUL",
        0xB782: "HANGUL",
        0xB783: "HANGUL",
        0xB784: "HANGUL",
        0xB785: "HANGUL",
        0xB786: "HANGUL",
        0xB787: "HANGUL",
        0xB788: "HANGUL",
        0xB789: "HANGUL",
        0xB78A: "HANGUL",
        0xB78B: "HANGUL",
        0xB78C: "HANGUL",
        0xB78D: "HANGUL",
        0xB78E: "HANGUL",
        0xB78F: "HANGUL",
        0xB790: "HANGUL",
        0xB791: "HANGUL",
        0xB792: "HANGUL",
        0xB793: "HANGUL",
        0xB794: "HANGUL",
        0xB795: "HANGUL",
        0xB796: "HANGUL",
        0xB797: "HANGUL",
        0xB798: "HANGUL",
        0xB799: "HANGUL",
        0xB79A: "HANGUL",
        0xB79B: "HANGUL",
        0xB79C: "HANGUL",
        0xB79D: "HANGUL",
        0xB79E: "HANGUL",
        0xB79F: "HANGUL",
        0xB7A0: "HANGUL",
        0xB7A1: "HANGUL",
        0xB7A2: "HANGUL",
        0xB7A3: "HANGUL",
        0xB7A4: "HANGUL",
        0xB7A5: "HANGUL",
        0xB7A6: "HANGUL",
        0xB7A7: "HANGUL",
        0xB7A8: "HANGUL",
        0xB7A9: "HANGUL",
        0xB7AA: "HANGUL",
        0xB7AB: "HANGUL",
        0xB7AC: "HANGUL",
        0xB7AD: "HANGUL",
        0xB7AE: "HANGUL",
        0xB7AF: "HANGUL",
        0xB7B0: "HANGUL",
        0xB7B1: "HANGUL",
        0xB7B2: "HANGUL",
        0xB7B3: "HANGUL",
        0xB7B4: "HANGUL",
        0xB7B5: "HANGUL",
        0xB7B6: "HANGUL",
        0xB7B7: "HANGUL",
        0xB7B8: "HANGUL",
        0xB7B9: "HANGUL",
        0xB7BA: "HANGUL",
        0xB7BB: "HANGUL",
        0xB7BC: "HANGUL",
        0xB7BD: "HANGUL",
        0xB7BE: "HANGUL",
        0xB7BF: "HANGUL",
        0xB7C0: "HANGUL",
        0xB7C1: "HANGUL",
        0xB7C2: "HANGUL",
        0xB7C3: "HANGUL",
        0xB7C4: "HANGUL",
        0xB7C5: "HANGUL",
        0xB7C6: "HANGUL",
        0xB7C7: "HANGUL",
        0xB7C8: "HANGUL",
        0xB7C9: "HANGUL",
        0xB7CA: "HANGUL",
        0xB7CB: "HANGUL",
        0xB7CC: "HANGUL",
        0xB7CD: "HANGUL",
        0xB7CE: "HANGUL",
        0xB7CF: "HANGUL",
        0xB7D0: "HANGUL",
        0xB7D1: "HANGUL",
        0xB7D2: "HANGUL",
        0xB7D3: "HANGUL",
        0xB7D4: "HANGUL",
        0xB7D5: "HANGUL",
        0xB7D6: "HANGUL",
        0xB7D7: "HANGUL",
        0xB7D8: "HANGUL",
        0xB7D9: "HANGUL",
        0xB7DA: "HANGUL",
        0xB7DB: "HANGUL",
        0xB7DC: "HANGUL",
        0xB7DD: "HANGUL",
        0xB7DE: "HANGUL",
        0xB7DF: "HANGUL",
        0xB7E0: "HANGUL",
        0xB7E1: "HANGUL",
        0xB7E2: "HANGUL",
        0xB7E3: "HANGUL",
        0xB7E4: "HANGUL",
        0xB7E5: "HANGUL",
        0xB7E6: "HANGUL",
        0xB7E7: "HANGUL",
        0xB7E8: "HANGUL",
        0xB7E9: "HANGUL",
        0xB7EA: "HANGUL",
        0xB7EB: "HANGUL",
        0xB7EC: "HANGUL",
        0xB7ED: "HANGUL",
        0xB7EE: "HANGUL",
        0xB7EF: "HANGUL",
        0xB7F0: "HANGUL",
        0xB7F1: "HANGUL",
        0xB7F2: "HANGUL",
        0xB7F3: "HANGUL",
        0xB7F4: "HANGUL",
        0xB7F5: "HANGUL",
        0xB7F6: "HANGUL",
        0xB7F7: "HANGUL",
        0xB7F8: "HANGUL",
        0xB7F9: "HANGUL",
        0xB7FA: "HANGUL",
        0xB7FB: "HANGUL",
        0xB7FC: "HANGUL",
        0xB7FD: "HANGUL",
        0xB7FE: "HANGUL",
        0xB7FF: "HANGUL",
        0xB800: "HANGUL",
        0xB801: "HANGUL",
        0xB802: "HANGUL",
        0xB803: "HANGUL",
        0xB804: "HANGUL",
        0xB805: "HANGUL",
        0xB806: "HANGUL",
        0xB807: "HANGUL",
        0xB808: "HANGUL",
        0xB809: "HANGUL",
        0xB80A: "HANGUL",
        0xB80B: "HANGUL",
        0xB80C: "HANGUL",
        0xB80D: "HANGUL",
        0xB80E: "HANGUL",
        0xB80F: "HANGUL",
        0xB810: "HANGUL",
        0xB811: "HANGUL",
        0xB812: "HANGUL",
        0xB813: "HANGUL",
        0xB814: "HANGUL",
        0xB815: "HANGUL",
        0xB816: "HANGUL",
        0xB817: "HANGUL",
        0xB818: "HANGUL",
        0xB819: "HANGUL",
        0xB81A: "HANGUL",
        0xB81B: "HANGUL",
        0xB81C: "HANGUL",
        0xB81D: "HANGUL",
        0xB81E: "HANGUL",
        0xB81F: "HANGUL",
        0xB820: "HANGUL",
        0xB821: "HANGUL",
        0xB822: "HANGUL",
        0xB823: "HANGUL",
        0xB824: "HANGUL",
        0xB825: "HANGUL",
        0xB826: "HANGUL",
        0xB827: "HANGUL",
        0xB828: "HANGUL",
        0xB829: "HANGUL",
        0xB82A: "HANGUL",
        0xB82B: "HANGUL",
        0xB82C: "HANGUL",
        0xB82D: "HANGUL",
        0xB82E: "HANGUL",
        0xB82F: "HANGUL",
        0xB830: "HANGUL",
        0xB831: "HANGUL",
        0xB832: "HANGUL",
        0xB833: "HANGUL",
        0xB834: "HANGUL",
        0xB835: "HANGUL",
        0xB836: "HANGUL",
        0xB837: "HANGUL",
        0xB838: "HANGUL",
        0xB839: "HANGUL",
        0xB83A: "HANGUL",
        0xB83B: "HANGUL",
        0xB83C: "HANGUL",
        0xB83D: "HANGUL",
        0xB83E: "HANGUL",
        0xB83F: "HANGUL",
        0xB840: "HANGUL",
        0xB841: "HANGUL",
        0xB842: "HANGUL",
        0xB843: "HANGUL",
        0xB844: "HANGUL",
        0xB845: "HANGUL",
        0xB846: "HANGUL",
        0xB847: "HANGUL",
        0xB848: "HANGUL",
        0xB849: "HANGUL",
        0xB84A: "HANGUL",
        0xB84B: "HANGUL",
        0xB84C: "HANGUL",
        0xB84D: "HANGUL",
        0xB84E: "HANGUL",
        0xB84F: "HANGUL",
        0xB850: "HANGUL",
        0xB851: "HANGUL",
        0xB852: "HANGUL",
        0xB853: "HANGUL",
        0xB854: "HANGUL",
        0xB855: "HANGUL",
        0xB856: "HANGUL",
        0xB857: "HANGUL",
        0xB858: "HANGUL",
        0xB859: "HANGUL",
        0xB85A: "HANGUL",
        0xB85B: "HANGUL",
        0xB85C: "HANGUL",
        0xB85D: "HANGUL",
        0xB85E: "HANGUL",
        0xB85F: "HANGUL",
        0xB860: "HANGUL",
        0xB861: "HANGUL",
        0xB862: "HANGUL",
        0xB863: "HANGUL",
        0xB864: "HANGUL",
        0xB865: "HANGUL",
        0xB866: "HANGUL",
        0xB867: "HANGUL",
        0xB868: "HANGUL",
        0xB869: "HANGUL",
        0xB86A: "HANGUL",
        0xB86B: "HANGUL",
        0xB86C: "HANGUL",
        0xB86D: "HANGUL",
        0xB86E: "HANGUL",
        0xB86F: "HANGUL",
        0xB870: "HANGUL",
        0xB871: "HANGUL",
        0xB872: "HANGUL",
        0xB873: "HANGUL",
        0xB874: "HANGUL",
        0xB875: "HANGUL",
        0xB876: "HANGUL",
        0xB877: "HANGUL",
        0xB878: "HANGUL",
        0xB879: "HANGUL",
        0xB87A: "HANGUL",
        0xB87B: "HANGUL",
        0xB87C: "HANGUL",
        0xB87D: "HANGUL",
        0xB87E: "HANGUL",
        0xB87F: "HANGUL",
        0xB880: "HANGUL",
        0xB881: "HANGUL",
        0xB882: "HANGUL",
        0xB883: "HANGUL",
        0xB884: "HANGUL",
        0xB885: "HANGUL",
        0xB886: "HANGUL",
        0xB887: "HANGUL",
        0xB888: "HANGUL",
        0xB889: "HANGUL",
        0xB88A: "HANGUL",
        0xB88B: "HANGUL",
        0xB88C: "HANGUL",
        0xB88D: "HANGUL",
        0xB88E: "HANGUL",
        0xB88F: "HANGUL",
        0xB890: "HANGUL",
        0xB891: "HANGUL",
        0xB892: "HANGUL",
        0xB893: "HANGUL",
        0xB894: "HANGUL",
        0xB895: "HANGUL",
        0xB896: "HANGUL",
        0xB897: "HANGUL",
        0xB898: "HANGUL",
        0xB899: "HANGUL",
        0xB89A: "HANGUL",
        0xB89B: "HANGUL",
        0xB89C: "HANGUL",
        0xB89D: "HANGUL",
        0xB89E: "HANGUL",
        0xB89F: "HANGUL",
        0xB8A0: "HANGUL",
        0xB8A1: "HANGUL",
        0xB8A2: "HANGUL",
        0xB8A3: "HANGUL",
        0xB8A4: "HANGUL",
        0xB8A5: "HANGUL",
        0xB8A6: "HANGUL",
        0xB8A7: "HANGUL",
        0xB8A8: "HANGUL",
        0xB8A9: "HANGUL",
        0xB8AA: "HANGUL",
        0xB8AB: "HANGUL",
        0xB8AC: "HANGUL",
        0xB8AD: "HANGUL",
        0xB8AE: "HANGUL",
        0xB8AF: "HANGUL",
        0xB8B0: "HANGUL",
        0xB8B1: "HANGUL",
        0xB8B2: "HANGUL",
        0xB8B3: "HANGUL",
        0xB8B4: "HANGUL",
        0xB8B5: "HANGUL",
        0xB8B6: "HANGUL",
        0xB8B7: "HANGUL",
        0xB8B8: "HANGUL",
        0xB8B9: "HANGUL",
        0xB8BA: "HANGUL",
        0xB8BB: "HANGUL",
        0xB8BC: "HANGUL",
        0xB8BD: "HANGUL",
        0xB8BE: "HANGUL",
        0xB8BF: "HANGUL",
        0xB8C0: "HANGUL",
        0xB8C1: "HANGUL",
        0xB8C2: "HANGUL",
        0xB8C3: "HANGUL",
        0xB8C4: "HANGUL",
        0xB8C5: "HANGUL",
        0xB8C6: "HANGUL",
        0xB8C7: "HANGUL",
        0xB8C8: "HANGUL",
        0xB8C9: "HANGUL",
        0xB8CA: "HANGUL",
        0xB8CB: "HANGUL",
        0xB8CC: "HANGUL",
        0xB8CD: "HANGUL",
        0xB8CE: "HANGUL",
        0xB8CF: "HANGUL",
        0xB8D0: "HANGUL",
        0xB8D1: "HANGUL",
        0xB8D2: "HANGUL",
        0xB8D3: "HANGUL",
        0xB8D4: "HANGUL",
        0xB8D5: "HANGUL",
        0xB8D6: "HANGUL",
        0xB8D7: "HANGUL",
        0xB8D8: "HANGUL",
        0xB8D9: "HANGUL",
        0xB8DA: "HANGUL",
        0xB8DB: "HANGUL",
        0xB8DC: "HANGUL",
        0xB8DD: "HANGUL",
        0xB8DE: "HANGUL",
        0xB8DF: "HANGUL",
        0xB8E0: "HANGUL",
        0xB8E1: "HANGUL",
        0xB8E2: "HANGUL",
        0xB8E3: "HANGUL",
        0xB8E4: "HANGUL",
        0xB8E5: "HANGUL",
        0xB8E6: "HANGUL",
        0xB8E7: "HANGUL",
        0xB8E8: "HANGUL",
        0xB8E9: "HANGUL",
        0xB8EA: "HANGUL",
        0xB8EB: "HANGUL",
        0xB8EC: "HANGUL",
        0xB8ED: "HANGUL",
        0xB8EE: "HANGUL",
        0xB8EF: "HANGUL",
        0xB8F0: "HANGUL",
        0xB8F1: "HANGUL",
        0xB8F2: "HANGUL",
        0xB8F3: "HANGUL",
        0xB8F4: "HANGUL",
        0xB8F5: "HANGUL",
        0xB8F6: "HANGUL",
        0xB8F7: "HANGUL",
        0xB8F8: "HANGUL",
        0xB8F9: "HANGUL",
        0xB8FA: "HANGUL",
        0xB8FB: "HANGUL",
        0xB8FC: "HANGUL",
        0xB8FD: "HANGUL",
        0xB8FE: "HANGUL",
        0xB8FF: "HANGUL",
        0xB900: "HANGUL",
        0xB901: "HANGUL",
        0xB902: "HANGUL",
        0xB903: "HANGUL",
        0xB904: "HANGUL",
        0xB905: "HANGUL",
        0xB906: "HANGUL",
        0xB907: "HANGUL",
        0xB908: "HANGUL",
        0xB909: "HANGUL",
        0xB90A: "HANGUL",
        0xB90B: "HANGUL",
        0xB90C: "HANGUL",
        0xB90D: "HANGUL",
        0xB90E: "HANGUL",
        0xB90F: "HANGUL",
        0xB910: "HANGUL",
        0xB911: "HANGUL",
        0xB912: "HANGUL",
        0xB913: "HANGUL",
        0xB914: "HANGUL",
        0xB915: "HANGUL",
        0xB916: "HANGUL",
        0xB917: "HANGUL",
        0xB918: "HANGUL",
        0xB919: "HANGUL",
        0xB91A: "HANGUL",
        0xB91B: "HANGUL",
        0xB91C: "HANGUL",
        0xB91D: "HANGUL",
        0xB91E: "HANGUL",
        0xB91F: "HANGUL",
        0xB920: "HANGUL",
        0xB921: "HANGUL",
        0xB922: "HANGUL",
        0xB923: "HANGUL",
        0xB924: "HANGUL",
        0xB925: "HANGUL",
        0xB926: "HANGUL",
        0xB927: "HANGUL",
        0xB928: "HANGUL",
        0xB929: "HANGUL",
        0xB92A: "HANGUL",
        0xB92B: "HANGUL",
        0xB92C: "HANGUL",
        0xB92D: "HANGUL",
        0xB92E: "HANGUL",
        0xB92F: "HANGUL",
        0xB930: "HANGUL",
        0xB931: "HANGUL",
        0xB932: "HANGUL",
        0xB933: "HANGUL",
        0xB934: "HANGUL",
        0xB935: "HANGUL",
        0xB936: "HANGUL",
        0xB937: "HANGUL",
        0xB938: "HANGUL",
        0xB939: "HANGUL",
        0xB93A: "HANGUL",
        0xB93B: "HANGUL",
        0xB93C: "HANGUL",
        0xB93D: "HANGUL",
        0xB93E: "HANGUL",
        0xB93F: "HANGUL",
        0xB940: "HANGUL",
        0xB941: "HANGUL",
        0xB942: "HANGUL",
        0xB943: "HANGUL",
        0xB944: "HANGUL",
        0xB945: "HANGUL",
        0xB946: "HANGUL",
        0xB947: "HANGUL",
        0xB948: "HANGUL",
        0xB949: "HANGUL",
        0xB94A: "HANGUL",
        0xB94B: "HANGUL",
        0xB94C: "HANGUL",
        0xB94D: "HANGUL",
        0xB94E: "HANGUL",
        0xB94F: "HANGUL",
        0xB950: "HANGUL",
        0xB951: "HANGUL",
        0xB952: "HANGUL",
        0xB953: "HANGUL",
        0xB954: "HANGUL",
        0xB955: "HANGUL",
        0xB956: "HANGUL",
        0xB957: "HANGUL",
        0xB958: "HANGUL",
        0xB959: "HANGUL",
        0xB95A: "HANGUL",
        0xB95B: "HANGUL",
        0xB95C: "HANGUL",
        0xB95D: "HANGUL",
        0xB95E: "HANGUL",
        0xB95F: "HANGUL",
        0xB960: "HANGUL",
        0xB961: "HANGUL",
        0xB962: "HANGUL",
        0xB963: "HANGUL",
        0xB964: "HANGUL",
        0xB965: "HANGUL",
        0xB966: "HANGUL",
        0xB967: "HANGUL",
        0xB968: "HANGUL",
        0xB969: "HANGUL",
        0xB96A: "HANGUL",
        0xB96B: "HANGUL",
        0xB96C: "HANGUL",
        0xB96D: "HANGUL",
        0xB96E: "HANGUL",
        0xB96F: "HANGUL",
        0xB970: "HANGUL",
        0xB971: "HANGUL",
        0xB972: "HANGUL",
        0xB973: "HANGUL",
        0xB974: "HANGUL",
        0xB975: "HANGUL",
        0xB976: "HANGUL",
        0xB977: "HANGUL",
        0xB978: "HANGUL",
        0xB979: "HANGUL",
        0xB97A: "HANGUL",
        0xB97B: "HANGUL",
        0xB97C: "HANGUL",
        0xB97D: "HANGUL",
        0xB97E: "HANGUL",
        0xB97F: "HANGUL",
        0xB980: "HANGUL",
        0xB981: "HANGUL",
        0xB982: "HANGUL",
        0xB983: "HANGUL",
        0xB984: "HANGUL",
        0xB985: "HANGUL",
        0xB986: "HANGUL",
        0xB987: "HANGUL",
        0xB988: "HANGUL",
        0xB989: "HANGUL",
        0xB98A: "HANGUL",
        0xB98B: "HANGUL",
        0xB98C: "HANGUL",
        0xB98D: "HANGUL",
        0xB98E: "HANGUL",
        0xB98F: "HANGUL",
        0xB990: "HANGUL",
        0xB991: "HANGUL",
        0xB992: "HANGUL",
        0xB993: "HANGUL",
        0xB994: "HANGUL",
        0xB995: "HANGUL",
        0xB996: "HANGUL",
        0xB997: "HANGUL",
        0xB998: "HANGUL",
        0xB999: "HANGUL",
        0xB99A: "HANGUL",
        0xB99B: "HANGUL",
        0xB99C: "HANGUL",
        0xB99D: "HANGUL",
        0xB99E: "HANGUL",
        0xB99F: "HANGUL",
        0xB9A0: "HANGUL",
        0xB9A1: "HANGUL",
        0xB9A2: "HANGUL",
        0xB9A3: "HANGUL",
        0xB9A4: "HANGUL",
        0xB9A5: "HANGUL",
        0xB9A6: "HANGUL",
        0xB9A7: "HANGUL",
        0xB9A8: "HANGUL",
        0xB9A9: "HANGUL",
        0xB9AA: "HANGUL",
        0xB9AB: "HANGUL",
        0xB9AC: "HANGUL",
        0xB9AD: "HANGUL",
        0xB9AE: "HANGUL",
        0xB9AF: "HANGUL",
        0xB9B0: "HANGUL",
        0xB9B1: "HANGUL",
        0xB9B2: "HANGUL",
        0xB9B3: "HANGUL",
        0xB9B4: "HANGUL",
        0xB9B5: "HANGUL",
        0xB9B6: "HANGUL",
        0xB9B7: "HANGUL",
        0xB9B8: "HANGUL",
        0xB9B9: "HANGUL",
        0xB9BA: "HANGUL",
        0xB9BB: "HANGUL",
        0xB9BC: "HANGUL",
        0xB9BD: "HANGUL",
        0xB9BE: "HANGUL",
        0xB9BF: "HANGUL",
        0xB9C0: "HANGUL",
        0xB9C1: "HANGUL",
        0xB9C2: "HANGUL",
        0xB9C3: "HANGUL",
        0xB9C4: "HANGUL",
        0xB9C5: "HANGUL",
        0xB9C6: "HANGUL",
        0xB9C7: "HANGUL",
        0xB9C8: "HANGUL",
        0xB9C9: "HANGUL",
        0xB9CA: "HANGUL",
        0xB9CB: "HANGUL",
        0xB9CC: "HANGUL",
        0xB9CD: "HANGUL",
        0xB9CE: "HANGUL",
        0xB9CF: "HANGUL",
        0xB9D0: "HANGUL",
        0xB9D1: "HANGUL",
        0xB9D2: "HANGUL",
        0xB9D3: "HANGUL",
        0xB9D4: "HANGUL",
        0xB9D5: "HANGUL",
        0xB9D6: "HANGUL",
        0xB9D7: "HANGUL",
        0xB9D8: "HANGUL",
        0xB9D9: "HANGUL",
        0xB9DA: "HANGUL",
        0xB9DB: "HANGUL",
        0xB9DC: "HANGUL",
        0xB9DD: "HANGUL",
        0xB9DE: "HANGUL",
        0xB9DF: "HANGUL",
        0xB9E0: "HANGUL",
        0xB9E1: "HANGUL",
        0xB9E2: "HANGUL",
        0xB9E3: "HANGUL",
        0xB9E4: "HANGUL",
        0xB9E5: "HANGUL",
        0xB9E6: "HANGUL",
        0xB9E7: "HANGUL",
        0xB9E8: "HANGUL",
        0xB9E9: "HANGUL",
        0xB9EA: "HANGUL",
        0xB9EB: "HANGUL",
        0xB9EC: "HANGUL",
        0xB9ED: "HANGUL",
        0xB9EE: "HANGUL",
        0xB9EF: "HANGUL",
        0xB9F0: "HANGUL",
        0xB9F1: "HANGUL",
        0xB9F2: "HANGUL",
        0xB9F3: "HANGUL",
        0xB9F4: "HANGUL",
        0xB9F5: "HANGUL",
        0xB9F6: "HANGUL",
        0xB9F7: "HANGUL",
        0xB9F8: "HANGUL",
        0xB9F9: "HANGUL",
        0xB9FA: "HANGUL",
        0xB9FB: "HANGUL",
        0xB9FC: "HANGUL",
        0xB9FD: "HANGUL",
        0xB9FE: "HANGUL",
        0xB9FF: "HANGUL",
        0xBA00: "HANGUL",
        0xBA01: "HANGUL",
        0xBA02: "HANGUL",
        0xBA03: "HANGUL",
        0xBA04: "HANGUL",
        0xBA05: "HANGUL",
        0xBA06: "HANGUL",
        0xBA07: "HANGUL",
        0xBA08: "HANGUL",
        0xBA09: "HANGUL",
        0xBA0A: "HANGUL",
        0xBA0B: "HANGUL",
        0xBA0C: "HANGUL",
        0xBA0D: "HANGUL",
        0xBA0E: "HANGUL",
        0xBA0F: "HANGUL",
        0xBA10: "HANGUL",
        0xBA11: "HANGUL",
        0xBA12: "HANGUL",
        0xBA13: "HANGUL",
        0xBA14: "HANGUL",
        0xBA15: "HANGUL",
        0xBA16: "HANGUL",
        0xBA17: "HANGUL",
        0xBA18: "HANGUL",
        0xBA19: "HANGUL",
        0xBA1A: "HANGUL",
        0xBA1B: "HANGUL",
        0xBA1C: "HANGUL",
        0xBA1D: "HANGUL",
        0xBA1E: "HANGUL",
        0xBA1F: "HANGUL",
        0xBA20: "HANGUL",
        0xBA21: "HANGUL",
        0xBA22: "HANGUL",
        0xBA23: "HANGUL",
        0xBA24: "HANGUL",
        0xBA25: "HANGUL",
        0xBA26: "HANGUL",
        0xBA27: "HANGUL",
        0xBA28: "HANGUL",
        0xBA29: "HANGUL",
        0xBA2A: "HANGUL",
        0xBA2B: "HANGUL",
        0xBA2C: "HANGUL",
        0xBA2D: "HANGUL",
        0xBA2E: "HANGUL",
        0xBA2F: "HANGUL",
        0xBA30: "HANGUL",
        0xBA31: "HANGUL",
        0xBA32: "HANGUL",
        0xBA33: "HANGUL",
        0xBA34: "HANGUL",
        0xBA35: "HANGUL",
        0xBA36: "HANGUL",
        0xBA37: "HANGUL",
        0xBA38: "HANGUL",
        0xBA39: "HANGUL",
        0xBA3A: "HANGUL",
        0xBA3B: "HANGUL",
        0xBA3C: "HANGUL",
        0xBA3D: "HANGUL",
        0xBA3E: "HANGUL",
        0xBA3F: "HANGUL",
        0xBA40: "HANGUL",
        0xBA41: "HANGUL",
        0xBA42: "HANGUL",
        0xBA43: "HANGUL",
        0xBA44: "HANGUL",
        0xBA45: "HANGUL",
        0xBA46: "HANGUL",
        0xBA47: "HANGUL",
        0xBA48: "HANGUL",
        0xBA49: "HANGUL",
        0xBA4A: "HANGUL",
        0xBA4B: "HANGUL",
        0xBA4C: "HANGUL",
        0xBA4D: "HANGUL",
        0xBA4E: "HANGUL",
        0xBA4F: "HANGUL",
        0xBA50: "HANGUL",
        0xBA51: "HANGUL",
        0xBA52: "HANGUL",
        0xBA53: "HANGUL",
        0xBA54: "HANGUL",
        0xBA55: "HANGUL",
        0xBA56: "HANGUL",
        0xBA57: "HANGUL",
        0xBA58: "HANGUL",
        0xBA59: "HANGUL",
        0xBA5A: "HANGUL",
        0xBA5B: "HANGUL",
        0xBA5C: "HANGUL",
        0xBA5D: "HANGUL",
        0xBA5E: "HANGUL",
        0xBA5F: "HANGUL",
        0xBA60: "HANGUL",
        0xBA61: "HANGUL",
        0xBA62: "HANGUL",
        0xBA63: "HANGUL",
        0xBA64: "HANGUL",
        0xBA65: "HANGUL",
        0xBA66: "HANGUL",
        0xBA67: "HANGUL",
        0xBA68: "HANGUL",
        0xBA69: "HANGUL",
        0xBA6A: "HANGUL",
        0xBA6B: "HANGUL",
        0xBA6C: "HANGUL",
        0xBA6D: "HANGUL",
        0xBA6E: "HANGUL",
        0xBA6F: "HANGUL",
        0xBA70: "HANGUL",
        0xBA71: "HANGUL",
        0xBA72: "HANGUL",
        0xBA73: "HANGUL",
        0xBA74: "HANGUL",
        0xBA75: "HANGUL",
        0xBA76: "HANGUL",
        0xBA77: "HANGUL",
        0xBA78: "HANGUL",
        0xBA79: "HANGUL",
        0xBA7A: "HANGUL",
        0xBA7B: "HANGUL",
        0xBA7C: "HANGUL",
        0xBA7D: "HANGUL",
        0xBA7E: "HANGUL",
        0xBA7F: "HANGUL",
        0xBA80: "HANGUL",
        0xBA81: "HANGUL",
        0xBA82: "HANGUL",
        0xBA83: "HANGUL",
        0xBA84: "HANGUL",
        0xBA85: "HANGUL",
        0xBA86: "HANGUL",
        0xBA87: "HANGUL",
        0xBA88: "HANGUL",
        0xBA89: "HANGUL",
        0xBA8A: "HANGUL",
        0xBA8B: "HANGUL",
        0xBA8C: "HANGUL",
        0xBA8D: "HANGUL",
        0xBA8E: "HANGUL",
        0xBA8F: "HANGUL",
        0xBA90: "HANGUL",
        0xBA91: "HANGUL",
        0xBA92: "HANGUL",
        0xBA93: "HANGUL",
        0xBA94: "HANGUL",
        0xBA95: "HANGUL",
        0xBA96: "HANGUL",
        0xBA97: "HANGUL",
        0xBA98: "HANGUL",
        0xBA99: "HANGUL",
        0xBA9A: "HANGUL",
        0xBA9B: "HANGUL",
        0xBA9C: "HANGUL",
        0xBA9D: "HANGUL",
        0xBA9E: "HANGUL",
        0xBA9F: "HANGUL",
        0xBAA0: "HANGUL",
        0xBAA1: "HANGUL",
        0xBAA2: "HANGUL",
        0xBAA3: "HANGUL",
        0xBAA4: "HANGUL",
        0xBAA5: "HANGUL",
        0xBAA6: "HANGUL",
        0xBAA7: "HANGUL",
        0xBAA8: "HANGUL",
        0xBAA9: "HANGUL",
        0xBAAA: "HANGUL",
        0xBAAB: "HANGUL",
        0xBAAC: "HANGUL",
        0xBAAD: "HANGUL",
        0xBAAE: "HANGUL",
        0xBAAF: "HANGUL",
        0xBAB0: "HANGUL",
        0xBAB1: "HANGUL",
        0xBAB2: "HANGUL",
        0xBAB3: "HANGUL",
        0xBAB4: "HANGUL",
        0xBAB5: "HANGUL",
        0xBAB6: "HANGUL",
        0xBAB7: "HANGUL",
        0xBAB8: "HANGUL",
        0xBAB9: "HANGUL",
        0xBABA: "HANGUL",
        0xBABB: "HANGUL",
        0xBABC: "HANGUL",
        0xBABD: "HANGUL",
        0xBABE: "HANGUL",
        0xBABF: "HANGUL",
        0xBAC0: "HANGUL",
        0xBAC1: "HANGUL",
        0xBAC2: "HANGUL",
        0xBAC3: "HANGUL",
        0xBAC4: "HANGUL",
        0xBAC5: "HANGUL",
        0xBAC6: "HANGUL",
        0xBAC7: "HANGUL",
        0xBAC8: "HANGUL",
        0xBAC9: "HANGUL",
        0xBACA: "HANGUL",
        0xBACB: "HANGUL",
        0xBACC: "HANGUL",
        0xBACD: "HANGUL",
        0xBACE: "HANGUL",
        0xBACF: "HANGUL",
        0xBAD0: "HANGUL",
        0xBAD1: "HANGUL",
        0xBAD2: "HANGUL",
        0xBAD3: "HANGUL",
        0xBAD4: "HANGUL",
        0xBAD5: "HANGUL",
        0xBAD6: "HANGUL",
        0xBAD7: "HANGUL",
        0xBAD8: "HANGUL",
        0xBAD9: "HANGUL",
        0xBADA: "HANGUL",
        0xBADB: "HANGUL",
        0xBADC: "HANGUL",
        0xBADD: "HANGUL",
        0xBADE: "HANGUL",
        0xBADF: "HANGUL",
        0xBAE0: "HANGUL",
        0xBAE1: "HANGUL",
        0xBAE2: "HANGUL",
        0xBAE3: "HANGUL",
        0xBAE4: "HANGUL",
        0xBAE5: "HANGUL",
        0xBAE6: "HANGUL",
        0xBAE7: "HANGUL",
        0xBAE8: "HANGUL",
        0xBAE9: "HANGUL",
        0xBAEA: "HANGUL",
        0xBAEB: "HANGUL",
        0xBAEC: "HANGUL",
        0xBAED: "HANGUL",
        0xBAEE: "HANGUL",
        0xBAEF: "HANGUL",
        0xBAF0: "HANGUL",
        0xBAF1: "HANGUL",
        0xBAF2: "HANGUL",
        0xBAF3: "HANGUL",
        0xBAF4: "HANGUL",
        0xBAF5: "HANGUL",
        0xBAF6: "HANGUL",
        0xBAF7: "HANGUL",
        0xBAF8: "HANGUL",
        0xBAF9: "HANGUL",
        0xBAFA: "HANGUL",
        0xBAFB: "HANGUL",
        0xBAFC: "HANGUL",
        0xBAFD: "HANGUL",
        0xBAFE: "HANGUL",
        0xBAFF: "HANGUL",
        0xBB00: "HANGUL",
        0xBB01: "HANGUL",
        0xBB02: "HANGUL",
        0xBB03: "HANGUL",
        0xBB04: "HANGUL",
        0xBB05: "HANGUL",
        0xBB06: "HANGUL",
        0xBB07: "HANGUL",
        0xBB08: "HANGUL",
        0xBB09: "HANGUL",
        0xBB0A: "HANGUL",
        0xBB0B: "HANGUL",
        0xBB0C: "HANGUL",
        0xBB0D: "HANGUL",
        0xBB0E: "HANGUL",
        0xBB0F: "HANGUL",
        0xBB10: "HANGUL",
        0xBB11: "HANGUL",
        0xBB12: "HANGUL",
        0xBB13: "HANGUL",
        0xBB14: "HANGUL",
        0xBB15: "HANGUL",
        0xBB16: "HANGUL",
        0xBB17: "HANGUL",
        0xBB18: "HANGUL",
        0xBB19: "HANGUL",
        0xBB1A: "HANGUL",
        0xBB1B: "HANGUL",
        0xBB1C: "HANGUL",
        0xBB1D: "HANGUL",
        0xBB1E: "HANGUL",
        0xBB1F: "HANGUL",
        0xBB20: "HANGUL",
        0xBB21: "HANGUL",
        0xBB22: "HANGUL",
        0xBB23: "HANGUL",
        0xBB24: "HANGUL",
        0xBB25: "HANGUL",
        0xBB26: "HANGUL",
        0xBB27: "HANGUL",
        0xBB28: "HANGUL",
        0xBB29: "HANGUL",
        0xBB2A: "HANGUL",
        0xBB2B: "HANGUL",
        0xBB2C: "HANGUL",
        0xBB2D: "HANGUL",
        0xBB2E: "HANGUL",
        0xBB2F: "HANGUL",
        0xBB30: "HANGUL",
        0xBB31: "HANGUL",
        0xBB32: "HANGUL",
        0xBB33: "HANGUL",
        0xBB34: "HANGUL",
        0xBB35: "HANGUL",
        0xBB36: "HANGUL",
        0xBB37: "HANGUL",
        0xBB38: "HANGUL",
        0xBB39: "HANGUL",
        0xBB3A: "HANGUL",
        0xBB3B: "HANGUL",
        0xBB3C: "HANGUL",
        0xBB3D: "HANGUL",
        0xBB3E: "HANGUL",
        0xBB3F: "HANGUL",
        0xBB40: "HANGUL",
        0xBB41: "HANGUL",
        0xBB42: "HANGUL",
        0xBB43: "HANGUL",
        0xBB44: "HANGUL",
        0xBB45: "HANGUL",
        0xBB46: "HANGUL",
        0xBB47: "HANGUL",
        0xBB48: "HANGUL",
        0xBB49: "HANGUL",
        0xBB4A: "HANGUL",
        0xBB4B: "HANGUL",
        0xBB4C: "HANGUL",
        0xBB4D: "HANGUL",
        0xBB4E: "HANGUL",
        0xBB4F: "HANGUL",
        0xBB50: "HANGUL",
        0xBB51: "HANGUL",
        0xBB52: "HANGUL",
        0xBB53: "HANGUL",
        0xBB54: "HANGUL",
        0xBB55: "HANGUL",
        0xBB56: "HANGUL",
        0xBB57: "HANGUL",
        0xBB58: "HANGUL",
        0xBB59: "HANGUL",
        0xBB5A: "HANGUL",
        0xBB5B: "HANGUL",
        0xBB5C: "HANGUL",
        0xBB5D: "HANGUL",
        0xBB5E: "HANGUL",
        0xBB5F: "HANGUL",
        0xBB60: "HANGUL",
        0xBB61: "HANGUL",
        0xBB62: "HANGUL",
        0xBB63: "HANGUL",
        0xBB64: "HANGUL",
        0xBB65: "HANGUL",
        0xBB66: "HANGUL",
        0xBB67: "HANGUL",
        0xBB68: "HANGUL",
        0xBB69: "HANGUL",
        0xBB6A: "HANGUL",
        0xBB6B: "HANGUL",
        0xBB6C: "HANGUL",
        0xBB6D: "HANGUL",
        0xBB6E: "HANGUL",
        0xBB6F: "HANGUL",
        0xBB70: "HANGUL",
        0xBB71: "HANGUL",
        0xBB72: "HANGUL",
        0xBB73: "HANGUL",
        0xBB74: "HANGUL",
        0xBB75: "HANGUL",
        0xBB76: "HANGUL",
        0xBB77: "HANGUL",
        0xBB78: "HANGUL",
        0xBB79: "HANGUL",
        0xBB7A: "HANGUL",
        0xBB7B: "HANGUL",
        0xBB7C: "HANGUL",
        0xBB7D: "HANGUL",
        0xBB7E: "HANGUL",
        0xBB7F: "HANGUL",
        0xBB80: "HANGUL",
        0xBB81: "HANGUL",
        0xBB82: "HANGUL",
        0xBB83: "HANGUL",
        0xBB84: "HANGUL",
        0xBB85: "HANGUL",
        0xBB86: "HANGUL",
        0xBB87: "HANGUL",
        0xBB88: "HANGUL",
        0xBB89: "HANGUL",
        0xBB8A: "HANGUL",
        0xBB8B: "HANGUL",
        0xBB8C: "HANGUL",
        0xBB8D: "HANGUL",
        0xBB8E: "HANGUL",
        0xBB8F: "HANGUL",
        0xBB90: "HANGUL",
        0xBB91: "HANGUL",
        0xBB92: "HANGUL",
        0xBB93: "HANGUL",
        0xBB94: "HANGUL",
        0xBB95: "HANGUL",
        0xBB96: "HANGUL",
        0xBB97: "HANGUL",
        0xBB98: "HANGUL",
        0xBB99: "HANGUL",
        0xBB9A: "HANGUL",
        0xBB9B: "HANGUL",
        0xBB9C: "HANGUL",
        0xBB9D: "HANGUL",
        0xBB9E: "HANGUL",
        0xBB9F: "HANGUL",
        0xBBA0: "HANGUL",
        0xBBA1: "HANGUL",
        0xBBA2: "HANGUL",
        0xBBA3: "HANGUL",
        0xBBA4: "HANGUL",
        0xBBA5: "HANGUL",
        0xBBA6: "HANGUL",
        0xBBA7: "HANGUL",
        0xBBA8: "HANGUL",
        0xBBA9: "HANGUL",
        0xBBAA: "HANGUL",
        0xBBAB: "HANGUL",
        0xBBAC: "HANGUL",
        0xBBAD: "HANGUL",
        0xBBAE: "HANGUL",
        0xBBAF: "HANGUL",
        0xBBB0: "HANGUL",
        0xBBB1: "HANGUL",
        0xBBB2: "HANGUL",
        0xBBB3: "HANGUL",
        0xBBB4: "HANGUL",
        0xBBB5: "HANGUL",
        0xBBB6: "HANGUL",
        0xBBB7: "HANGUL",
        0xBBB8: "HANGUL",
        0xBBB9: "HANGUL",
        0xBBBA: "HANGUL",
        0xBBBB: "HANGUL",
        0xBBBC: "HANGUL",
        0xBBBD: "HANGUL",
        0xBBBE: "HANGUL",
        0xBBBF: "HANGUL",
        0xBBC0: "HANGUL",
        0xBBC1: "HANGUL",
        0xBBC2: "HANGUL",
        0xBBC3: "HANGUL",
        0xBBC4: "HANGUL",
        0xBBC5: "HANGUL",
        0xBBC6: "HANGUL",
        0xBBC7: "HANGUL",
        0xBBC8: "HANGUL",
        0xBBC9: "HANGUL",
        0xBBCA: "HANGUL",
        0xBBCB: "HANGUL",
        0xBBCC: "HANGUL",
        0xBBCD: "HANGUL",
        0xBBCE: "HANGUL",
        0xBBCF: "HANGUL",
        0xBBD0: "HANGUL",
        0xBBD1: "HANGUL",
        0xBBD2: "HANGUL",
        0xBBD3: "HANGUL",
        0xBBD4: "HANGUL",
        0xBBD5: "HANGUL",
        0xBBD6: "HANGUL",
        0xBBD7: "HANGUL",
        0xBBD8: "HANGUL",
        0xBBD9: "HANGUL",
        0xBBDA: "HANGUL",
        0xBBDB: "HANGUL",
        0xBBDC: "HANGUL",
        0xBBDD: "HANGUL",
        0xBBDE: "HANGUL",
        0xBBDF: "HANGUL",
        0xBBE0: "HANGUL",
        0xBBE1: "HANGUL",
        0xBBE2: "HANGUL",
        0xBBE3: "HANGUL",
        0xBBE4: "HANGUL",
        0xBBE5: "HANGUL",
        0xBBE6: "HANGUL",
        0xBBE7: "HANGUL",
        0xBBE8: "HANGUL",
        0xBBE9: "HANGUL",
        0xBBEA: "HANGUL",
        0xBBEB: "HANGUL",
        0xBBEC: "HANGUL",
        0xBBED: "HANGUL",
        0xBBEE: "HANGUL",
        0xBBEF: "HANGUL",
        0xBBF0: "HANGUL",
        0xBBF1: "HANGUL",
        0xBBF2: "HANGUL",
        0xBBF3: "HANGUL",
        0xBBF4: "HANGUL",
        0xBBF5: "HANGUL",
        0xBBF6: "HANGUL",
        0xBBF7: "HANGUL",
        0xBBF8: "HANGUL",
        0xBBF9: "HANGUL",
        0xBBFA: "HANGUL",
        0xBBFB: "HANGUL",
        0xBBFC: "HANGUL",
        0xBBFD: "HANGUL",
        0xBBFE: "HANGUL",
        0xBBFF: "HANGUL",
        0xBC00: "HANGUL",
        0xBC01: "HANGUL",
        0xBC02: "HANGUL",
        0xBC03: "HANGUL",
        0xBC04: "HANGUL",
        0xBC05: "HANGUL",
        0xBC06: "HANGUL",
        0xBC07: "HANGUL",
        0xBC08: "HANGUL",
        0xBC09: "HANGUL",
        0xBC0A: "HANGUL",
        0xBC0B: "HANGUL",
        0xBC0C: "HANGUL",
        0xBC0D: "HANGUL",
        0xBC0E: "HANGUL",
        0xBC0F: "HANGUL",
        0xBC10: "HANGUL",
        0xBC11: "HANGUL",
        0xBC12: "HANGUL",
        0xBC13: "HANGUL",
        0xBC14: "HANGUL",
        0xBC15: "HANGUL",
        0xBC16: "HANGUL",
        0xBC17: "HANGUL",
        0xBC18: "HANGUL",
        0xBC19: "HANGUL",
        0xBC1A: "HANGUL",
        0xBC1B: "HANGUL",
        0xBC1C: "HANGUL",
        0xBC1D: "HANGUL",
        0xBC1E: "HANGUL",
        0xBC1F: "HANGUL",
        0xBC20: "HANGUL",
        0xBC21: "HANGUL",
        0xBC22: "HANGUL",
        0xBC23: "HANGUL",
        0xBC24: "HANGUL",
        0xBC25: "HANGUL",
        0xBC26: "HANGUL",
        0xBC27: "HANGUL",
        0xBC28: "HANGUL",
        0xBC29: "HANGUL",
        0xBC2A: "HANGUL",
        0xBC2B: "HANGUL",
        0xBC2C: "HANGUL",
        0xBC2D: "HANGUL",
        0xBC2E: "HANGUL",
        0xBC2F: "HANGUL",
        0xBC30: "HANGUL",
        0xBC31: "HANGUL",
        0xBC32: "HANGUL",
        0xBC33: "HANGUL",
        0xBC34: "HANGUL",
        0xBC35: "HANGUL",
        0xBC36: "HANGUL",
        0xBC37: "HANGUL",
        0xBC38: "HANGUL",
        0xBC39: "HANGUL",
        0xBC3A: "HANGUL",
        0xBC3B: "HANGUL",
        0xBC3C: "HANGUL",
        0xBC3D: "HANGUL",
        0xBC3E: "HANGUL",
        0xBC3F: "HANGUL",
        0xBC40: "HANGUL",
        0xBC41: "HANGUL",
        0xBC42: "HANGUL",
        0xBC43: "HANGUL",
        0xBC44: "HANGUL",
        0xBC45: "HANGUL",
        0xBC46: "HANGUL",
        0xBC47: "HANGUL",
        0xBC48: "HANGUL",
        0xBC49: "HANGUL",
        0xBC4A: "HANGUL",
        0xBC4B: "HANGUL",
        0xBC4C: "HANGUL",
        0xBC4D: "HANGUL",
        0xBC4E: "HANGUL",
        0xBC4F: "HANGUL",
        0xBC50: "HANGUL",
        0xBC51: "HANGUL",
        0xBC52: "HANGUL",
        0xBC53: "HANGUL",
        0xBC54: "HANGUL",
        0xBC55: "HANGUL",
        0xBC56: "HANGUL",
        0xBC57: "HANGUL",
        0xBC58: "HANGUL",
        0xBC59: "HANGUL",
        0xBC5A: "HANGUL",
        0xBC5B: "HANGUL",
        0xBC5C: "HANGUL",
        0xBC5D: "HANGUL",
        0xBC5E: "HANGUL",
        0xBC5F: "HANGUL",
        0xBC60: "HANGUL",
        0xBC61: "HANGUL",
        0xBC62: "HANGUL",
        0xBC63: "HANGUL",
        0xBC64: "HANGUL",
        0xBC65: "HANGUL",
        0xBC66: "HANGUL",
        0xBC67: "HANGUL",
        0xBC68: "HANGUL",
        0xBC69: "HANGUL",
        0xBC6A: "HANGUL",
        0xBC6B: "HANGUL",
        0xBC6C: "HANGUL",
        0xBC6D: "HANGUL",
        0xBC6E: "HANGUL",
        0xBC6F: "HANGUL",
        0xBC70: "HANGUL",
        0xBC71: "HANGUL",
        0xBC72: "HANGUL",
        0xBC73: "HANGUL",
        0xBC74: "HANGUL",
        0xBC75: "HANGUL",
        0xBC76: "HANGUL",
        0xBC77: "HANGUL",
        0xBC78: "HANGUL",
        0xBC79: "HANGUL",
        0xBC7A: "HANGUL",
        0xBC7B: "HANGUL",
        0xBC7C: "HANGUL",
        0xBC7D: "HANGUL",
        0xBC7E: "HANGUL",
        0xBC7F: "HANGUL",
        0xBC80: "HANGUL",
        0xBC81: "HANGUL",
        0xBC82: "HANGUL",
        0xBC83: "HANGUL",
        0xBC84: "HANGUL",
        0xBC85: "HANGUL",
        0xBC86: "HANGUL",
        0xBC87: "HANGUL",
        0xBC88: "HANGUL",
        0xBC89: "HANGUL",
        0xBC8A: "HANGUL",
        0xBC8B: "HANGUL",
        0xBC8C: "HANGUL",
        0xBC8D: "HANGUL",
        0xBC8E: "HANGUL",
        0xBC8F: "HANGUL",
        0xBC90: "HANGUL",
        0xBC91: "HANGUL",
        0xBC92: "HANGUL",
        0xBC93: "HANGUL",
        0xBC94: "HANGUL",
        0xBC95: "HANGUL",
        0xBC96: "HANGUL",
        0xBC97: "HANGUL",
        0xBC98: "HANGUL",
        0xBC99: "HANGUL",
        0xBC9A: "HANGUL",
        0xBC9B: "HANGUL",
        0xBC9C: "HANGUL",
        0xBC9D: "HANGUL",
        0xBC9E: "HANGUL",
        0xBC9F: "HANGUL",
        0xBCA0: "HANGUL",
        0xBCA1: "HANGUL",
        0xBCA2: "HANGUL",
        0xBCA3: "HANGUL",
        0xBCA4: "HANGUL",
        0xBCA5: "HANGUL",
        0xBCA6: "HANGUL",
        0xBCA7: "HANGUL",
        0xBCA8: "HANGUL",
        0xBCA9: "HANGUL",
        0xBCAA: "HANGUL",
        0xBCAB: "HANGUL",
        0xBCAC: "HANGUL",
        0xBCAD: "HANGUL",
        0xBCAE: "HANGUL",
        0xBCAF: "HANGUL",
        0xBCB0: "HANGUL",
        0xBCB1: "HANGUL",
        0xBCB2: "HANGUL",
        0xBCB3: "HANGUL",
        0xBCB4: "HANGUL",
        0xBCB5: "HANGUL",
        0xBCB6: "HANGUL",
        0xBCB7: "HANGUL",
        0xBCB8: "HANGUL",
        0xBCB9: "HANGUL",
        0xBCBA: "HANGUL",
        0xBCBB: "HANGUL",
        0xBCBC: "HANGUL",
        0xBCBD: "HANGUL",
        0xBCBE: "HANGUL",
        0xBCBF: "HANGUL",
        0xBCC0: "HANGUL",
        0xBCC1: "HANGUL",
        0xBCC2: "HANGUL",
        0xBCC3: "HANGUL",
        0xBCC4: "HANGUL",
        0xBCC5: "HANGUL",
        0xBCC6: "HANGUL",
        0xBCC7: "HANGUL",
        0xBCC8: "HANGUL",
        0xBCC9: "HANGUL",
        0xBCCA: "HANGUL",
        0xBCCB: "HANGUL",
        0xBCCC: "HANGUL",
        0xBCCD: "HANGUL",
        0xBCCE: "HANGUL",
        0xBCCF: "HANGUL",
        0xBCD0: "HANGUL",
        0xBCD1: "HANGUL",
        0xBCD2: "HANGUL",
        0xBCD3: "HANGUL",
        0xBCD4: "HANGUL",
        0xBCD5: "HANGUL",
        0xBCD6: "HANGUL",
        0xBCD7: "HANGUL",
        0xBCD8: "HANGUL",
        0xBCD9: "HANGUL",
        0xBCDA: "HANGUL",
        0xBCDB: "HANGUL",
        0xBCDC: "HANGUL",
        0xBCDD: "HANGUL",
        0xBCDE: "HANGUL",
        0xBCDF: "HANGUL",
        0xBCE0: "HANGUL",
        0xBCE1: "HANGUL",
        0xBCE2: "HANGUL",
        0xBCE3: "HANGUL",
        0xBCE4: "HANGUL",
        0xBCE5: "HANGUL",
        0xBCE6: "HANGUL",
        0xBCE7: "HANGUL",
        0xBCE8: "HANGUL",
        0xBCE9: "HANGUL",
        0xBCEA: "HANGUL",
        0xBCEB: "HANGUL",
        0xBCEC: "HANGUL",
        0xBCED: "HANGUL",
        0xBCEE: "HANGUL",
        0xBCEF: "HANGUL",
        0xBCF0: "HANGUL",
        0xBCF1: "HANGUL",
        0xBCF2: "HANGUL",
        0xBCF3: "HANGUL",
        0xBCF4: "HANGUL",
        0xBCF5: "HANGUL",
        0xBCF6: "HANGUL",
        0xBCF7: "HANGUL",
        0xBCF8: "HANGUL",
        0xBCF9: "HANGUL",
        0xBCFA: "HANGUL",
        0xBCFB: "HANGUL",
        0xBCFC: "HANGUL",
        0xBCFD: "HANGUL",
        0xBCFE: "HANGUL",
        0xBCFF: "HANGUL",
        0xBD00: "HANGUL",
        0xBD01: "HANGUL",
        0xBD02: "HANGUL",
        0xBD03: "HANGUL",
        0xBD04: "HANGUL",
        0xBD05: "HANGUL",
        0xBD06: "HANGUL",
        0xBD07: "HANGUL",
        0xBD08: "HANGUL",
        0xBD09: "HANGUL",
        0xBD0A: "HANGUL",
        0xBD0B: "HANGUL",
        0xBD0C: "HANGUL",
        0xBD0D: "HANGUL",
        0xBD0E: "HANGUL",
        0xBD0F: "HANGUL",
        0xBD10: "HANGUL",
        0xBD11: "HANGUL",
        0xBD12: "HANGUL",
        0xBD13: "HANGUL",
        0xBD14: "HANGUL",
        0xBD15: "HANGUL",
        0xBD16: "HANGUL",
        0xBD17: "HANGUL",
        0xBD18: "HANGUL",
        0xBD19: "HANGUL",
        0xBD1A: "HANGUL",
        0xBD1B: "HANGUL",
        0xBD1C: "HANGUL",
        0xBD1D: "HANGUL",
        0xBD1E: "HANGUL",
        0xBD1F: "HANGUL",
        0xBD20: "HANGUL",
        0xBD21: "HANGUL",
        0xBD22: "HANGUL",
        0xBD23: "HANGUL",
        0xBD24: "HANGUL",
        0xBD25: "HANGUL",
        0xBD26: "HANGUL",
        0xBD27: "HANGUL",
        0xBD28: "HANGUL",
        0xBD29: "HANGUL",
        0xBD2A: "HANGUL",
        0xBD2B: "HANGUL",
        0xBD2C: "HANGUL",
        0xBD2D: "HANGUL",
        0xBD2E: "HANGUL",
        0xBD2F: "HANGUL",
        0xBD30: "HANGUL",
        0xBD31: "HANGUL",
        0xBD32: "HANGUL",
        0xBD33: "HANGUL",
        0xBD34: "HANGUL",
        0xBD35: "HANGUL",
        0xBD36: "HANGUL",
        0xBD37: "HANGUL",
        0xBD38: "HANGUL",
        0xBD39: "HANGUL",
        0xBD3A: "HANGUL",
        0xBD3B: "HANGUL",
        0xBD3C: "HANGUL",
        0xBD3D: "HANGUL",
        0xBD3E: "HANGUL",
        0xBD3F: "HANGUL",
        0xBD40: "HANGUL",
        0xBD41: "HANGUL",
        0xBD42: "HANGUL",
        0xBD43: "HANGUL",
        0xBD44: "HANGUL",
        0xBD45: "HANGUL",
        0xBD46: "HANGUL",
        0xBD47: "HANGUL",
        0xBD48: "HANGUL",
        0xBD49: "HANGUL",
        0xBD4A: "HANGUL",
        0xBD4B: "HANGUL",
        0xBD4C: "HANGUL",
        0xBD4D: "HANGUL",
        0xBD4E: "HANGUL",
        0xBD4F: "HANGUL",
        0xBD50: "HANGUL",
        0xBD51: "HANGUL",
        0xBD52: "HANGUL",
        0xBD53: "HANGUL",
        0xBD54: "HANGUL",
        0xBD55: "HANGUL",
        0xBD56: "HANGUL",
        0xBD57: "HANGUL",
        0xBD58: "HANGUL",
        0xBD59: "HANGUL",
        0xBD5A: "HANGUL",
        0xBD5B: "HANGUL",
        0xBD5C: "HANGUL",
        0xBD5D: "HANGUL",
        0xBD5E: "HANGUL",
        0xBD5F: "HANGUL",
        0xBD60: "HANGUL",
        0xBD61: "HANGUL",
        0xBD62: "HANGUL",
        0xBD63: "HANGUL",
        0xBD64: "HANGUL",
        0xBD65: "HANGUL",
        0xBD66: "HANGUL",
        0xBD67: "HANGUL",
        0xBD68: "HANGUL",
        0xBD69: "HANGUL",
        0xBD6A: "HANGUL",
        0xBD6B: "HANGUL",
        0xBD6C: "HANGUL",
        0xBD6D: "HANGUL",
        0xBD6E: "HANGUL",
        0xBD6F: "HANGUL",
        0xBD70: "HANGUL",
        0xBD71: "HANGUL",
        0xBD72: "HANGUL",
        0xBD73: "HANGUL",
        0xBD74: "HANGUL",
        0xBD75: "HANGUL",
        0xBD76: "HANGUL",
        0xBD77: "HANGUL",
        0xBD78: "HANGUL",
        0xBD79: "HANGUL",
        0xBD7A: "HANGUL",
        0xBD7B: "HANGUL",
        0xBD7C: "HANGUL",
        0xBD7D: "HANGUL",
        0xBD7E: "HANGUL",
        0xBD7F: "HANGUL",
        0xBD80: "HANGUL",
        0xBD81: "HANGUL",
        0xBD82: "HANGUL",
        0xBD83: "HANGUL",
        0xBD84: "HANGUL",
        0xBD85: "HANGUL",
        0xBD86: "HANGUL",
        0xBD87: "HANGUL",
        0xBD88: "HANGUL",
        0xBD89: "HANGUL",
        0xBD8A: "HANGUL",
        0xBD8B: "HANGUL",
        0xBD8C: "HANGUL",
        0xBD8D: "HANGUL",
        0xBD8E: "HANGUL",
        0xBD8F: "HANGUL",
        0xBD90: "HANGUL",
        0xBD91: "HANGUL",
        0xBD92: "HANGUL",
        0xBD93: "HANGUL",
        0xBD94: "HANGUL",
        0xBD95: "HANGUL",
        0xBD96: "HANGUL",
        0xBD97: "HANGUL",
        0xBD98: "HANGUL",
        0xBD99: "HANGUL",
        0xBD9A: "HANGUL",
        0xBD9B: "HANGUL",
        0xBD9C: "HANGUL",
        0xBD9D: "HANGUL",
        0xBD9E: "HANGUL",
        0xBD9F: "HANGUL",
        0xBDA0: "HANGUL",
        0xBDA1: "HANGUL",
        0xBDA2: "HANGUL",
        0xBDA3: "HANGUL",
        0xBDA4: "HANGUL",
        0xBDA5: "HANGUL",
        0xBDA6: "HANGUL",
        0xBDA7: "HANGUL",
        0xBDA8: "HANGUL",
        0xBDA9: "HANGUL",
        0xBDAA: "HANGUL",
        0xBDAB: "HANGUL",
        0xBDAC: "HANGUL",
        0xBDAD: "HANGUL",
        0xBDAE: "HANGUL",
        0xBDAF: "HANGUL",
        0xBDB0: "HANGUL",
        0xBDB1: "HANGUL",
        0xBDB2: "HANGUL",
        0xBDB3: "HANGUL",
        0xBDB4: "HANGUL",
        0xBDB5: "HANGUL",
        0xBDB6: "HANGUL",
        0xBDB7: "HANGUL",
        0xBDB8: "HANGUL",
        0xBDB9: "HANGUL",
        0xBDBA: "HANGUL",
        0xBDBB: "HANGUL",
        0xBDBC: "HANGUL",
        0xBDBD: "HANGUL",
        0xBDBE: "HANGUL",
        0xBDBF: "HANGUL",
        0xBDC0: "HANGUL",
        0xBDC1: "HANGUL",
        0xBDC2: "HANGUL",
        0xBDC3: "HANGUL",
        0xBDC4: "HANGUL",
        0xBDC5: "HANGUL",
        0xBDC6: "HANGUL",
        0xBDC7: "HANGUL",
        0xBDC8: "HANGUL",
        0xBDC9: "HANGUL",
        0xBDCA: "HANGUL",
        0xBDCB: "HANGUL",
        0xBDCC: "HANGUL",
        0xBDCD: "HANGUL",
        0xBDCE: "HANGUL",
        0xBDCF: "HANGUL",
        0xBDD0: "HANGUL",
        0xBDD1: "HANGUL",
        0xBDD2: "HANGUL",
        0xBDD3: "HANGUL",
        0xBDD4: "HANGUL",
        0xBDD5: "HANGUL",
        0xBDD6: "HANGUL",
        0xBDD7: "HANGUL",
        0xBDD8: "HANGUL",
        0xBDD9: "HANGUL",
        0xBDDA: "HANGUL",
        0xBDDB: "HANGUL",
        0xBDDC: "HANGUL",
        0xBDDD: "HANGUL",
        0xBDDE: "HANGUL",
        0xBDDF: "HANGUL",
        0xBDE0: "HANGUL",
        0xBDE1: "HANGUL",
        0xBDE2: "HANGUL",
        0xBDE3: "HANGUL",
        0xBDE4: "HANGUL",
        0xBDE5: "HANGUL",
        0xBDE6: "HANGUL",
        0xBDE7: "HANGUL",
        0xBDE8: "HANGUL",
        0xBDE9: "HANGUL",
        0xBDEA: "HANGUL",
        0xBDEB: "HANGUL",
        0xBDEC: "HANGUL",
        0xBDED: "HANGUL",
        0xBDEE: "HANGUL",
        0xBDEF: "HANGUL",
        0xBDF0: "HANGUL",
        0xBDF1: "HANGUL",
        0xBDF2: "HANGUL",
        0xBDF3: "HANGUL",
        0xBDF4: "HANGUL",
        0xBDF5: "HANGUL",
        0xBDF6: "HANGUL",
        0xBDF7: "HANGUL",
        0xBDF8: "HANGUL",
        0xBDF9: "HANGUL",
        0xBDFA: "HANGUL",
        0xBDFB: "HANGUL",
        0xBDFC: "HANGUL",
        0xBDFD: "HANGUL",
        0xBDFE: "HANGUL",
        0xBDFF: "HANGUL",
        0xBE00: "HANGUL",
        0xBE01: "HANGUL",
        0xBE02: "HANGUL",
        0xBE03: "HANGUL",
        0xBE04: "HANGUL",
        0xBE05: "HANGUL",
        0xBE06: "HANGUL",
        0xBE07: "HANGUL",
        0xBE08: "HANGUL",
        0xBE09: "HANGUL",
        0xBE0A: "HANGUL",
        0xBE0B: "HANGUL",
        0xBE0C: "HANGUL",
        0xBE0D: "HANGUL",
        0xBE0E: "HANGUL",
        0xBE0F: "HANGUL",
        0xBE10: "HANGUL",
        0xBE11: "HANGUL",
        0xBE12: "HANGUL",
        0xBE13: "HANGUL",
        0xBE14: "HANGUL",
        0xBE15: "HANGUL",
        0xBE16: "HANGUL",
        0xBE17: "HANGUL",
        0xBE18: "HANGUL",
        0xBE19: "HANGUL",
        0xBE1A: "HANGUL",
        0xBE1B: "HANGUL",
        0xBE1C: "HANGUL",
        0xBE1D: "HANGUL",
        0xBE1E: "HANGUL",
        0xBE1F: "HANGUL",
        0xBE20: "HANGUL",
        0xBE21: "HANGUL",
        0xBE22: "HANGUL",
        0xBE23: "HANGUL",
        0xBE24: "HANGUL",
        0xBE25: "HANGUL",
        0xBE26: "HANGUL",
        0xBE27: "HANGUL",
        0xBE28: "HANGUL",
        0xBE29: "HANGUL",
        0xBE2A: "HANGUL",
        0xBE2B: "HANGUL",
        0xBE2C: "HANGUL",
        0xBE2D: "HANGUL",
        0xBE2E: "HANGUL",
        0xBE2F: "HANGUL",
        0xBE30: "HANGUL",
        0xBE31: "HANGUL",
        0xBE32: "HANGUL",
        0xBE33: "HANGUL",
        0xBE34: "HANGUL",
        0xBE35: "HANGUL",
        0xBE36: "HANGUL",
        0xBE37: "HANGUL",
        0xBE38: "HANGUL",
        0xBE39: "HANGUL",
        0xBE3A: "HANGUL",
        0xBE3B: "HANGUL",
        0xBE3C: "HANGUL",
        0xBE3D: "HANGUL",
        0xBE3E: "HANGUL",
        0xBE3F: "HANGUL",
        0xBE40: "HANGUL",
        0xBE41: "HANGUL",
        0xBE42: "HANGUL",
        0xBE43: "HANGUL",
        0xBE44: "HANGUL",
        0xBE45: "HANGUL",
        0xBE46: "HANGUL",
        0xBE47: "HANGUL",
        0xBE48: "HANGUL",
        0xBE49: "HANGUL",
        0xBE4A: "HANGUL",
        0xBE4B: "HANGUL",
        0xBE4C: "HANGUL",
        0xBE4D: "HANGUL",
        0xBE4E: "HANGUL",
        0xBE4F: "HANGUL",
        0xBE50: "HANGUL",
        0xBE51: "HANGUL",
        0xBE52: "HANGUL",
        0xBE53: "HANGUL",
        0xBE54: "HANGUL",
        0xBE55: "HANGUL",
        0xBE56: "HANGUL",
        0xBE57: "HANGUL",
        0xBE58: "HANGUL",
        0xBE59: "HANGUL",
        0xBE5A: "HANGUL",
        0xBE5B: "HANGUL",
        0xBE5C: "HANGUL",
        0xBE5D: "HANGUL",
        0xBE5E: "HANGUL",
        0xBE5F: "HANGUL",
        0xBE60: "HANGUL",
        0xBE61: "HANGUL",
        0xBE62: "HANGUL",
        0xBE63: "HANGUL",
        0xBE64: "HANGUL",
        0xBE65: "HANGUL",
        0xBE66: "HANGUL",
        0xBE67: "HANGUL",
        0xBE68: "HANGUL",
        0xBE69: "HANGUL",
        0xBE6A: "HANGUL",
        0xBE6B: "HANGUL",
        0xBE6C: "HANGUL",
        0xBE6D: "HANGUL",
        0xBE6E: "HANGUL",
        0xBE6F: "HANGUL",
        0xBE70: "HANGUL",
        0xBE71: "HANGUL",
        0xBE72: "HANGUL",
        0xBE73: "HANGUL",
        0xBE74: "HANGUL",
        0xBE75: "HANGUL",
        0xBE76: "HANGUL",
        0xBE77: "HANGUL",
        0xBE78: "HANGUL",
        0xBE79: "HANGUL",
        0xBE7A: "HANGUL",
        0xBE7B: "HANGUL",
        0xBE7C: "HANGUL",
        0xBE7D: "HANGUL",
        0xBE7E: "HANGUL",
        0xBE7F: "HANGUL",
        0xBE80: "HANGUL",
        0xBE81: "HANGUL",
        0xBE82: "HANGUL",
        0xBE83: "HANGUL",
        0xBE84: "HANGUL",
        0xBE85: "HANGUL",
        0xBE86: "HANGUL",
        0xBE87: "HANGUL",
        0xBE88: "HANGUL",
        0xBE89: "HANGUL",
        0xBE8A: "HANGUL",
        0xBE8B: "HANGUL",
        0xBE8C: "HANGUL",
        0xBE8D: "HANGUL",
        0xBE8E: "HANGUL",
        0xBE8F: "HANGUL",
        0xBE90: "HANGUL",
        0xBE91: "HANGUL",
        0xBE92: "HANGUL",
        0xBE93: "HANGUL",
        0xBE94: "HANGUL",
        0xBE95: "HANGUL",
        0xBE96: "HANGUL",
        0xBE97: "HANGUL",
        0xBE98: "HANGUL",
        0xBE99: "HANGUL",
        0xBE9A: "HANGUL",
        0xBE9B: "HANGUL",
        0xBE9C: "HANGUL",
        0xBE9D: "HANGUL",
        0xBE9E: "HANGUL",
        0xBE9F: "HANGUL",
        0xBEA0: "HANGUL",
        0xBEA1: "HANGUL",
        0xBEA2: "HANGUL",
        0xBEA3: "HANGUL",
        0xBEA4: "HANGUL",
        0xBEA5: "HANGUL",
        0xBEA6: "HANGUL",
        0xBEA7: "HANGUL",
        0xBEA8: "HANGUL",
        0xBEA9: "HANGUL",
        0xBEAA: "HANGUL",
        0xBEAB: "HANGUL",
        0xBEAC: "HANGUL",
        0xBEAD: "HANGUL",
        0xBEAE: "HANGUL",
        0xBEAF: "HANGUL",
        0xBEB0: "HANGUL",
        0xBEB1: "HANGUL",
        0xBEB2: "HANGUL",
        0xBEB3: "HANGUL",
        0xBEB4: "HANGUL",
        0xBEB5: "HANGUL",
        0xBEB6: "HANGUL",
        0xBEB7: "HANGUL",
        0xBEB8: "HANGUL",
        0xBEB9: "HANGUL",
        0xBEBA: "HANGUL",
        0xBEBB: "HANGUL",
        0xBEBC: "HANGUL",
        0xBEBD: "HANGUL",
        0xBEBE: "HANGUL",
        0xBEBF: "HANGUL",
        0xBEC0: "HANGUL",
        0xBEC1: "HANGUL",
        0xBEC2: "HANGUL",
        0xBEC3: "HANGUL",
        0xBEC4: "HANGUL",
        0xBEC5: "HANGUL",
        0xBEC6: "HANGUL",
        0xBEC7: "HANGUL",
        0xBEC8: "HANGUL",
        0xBEC9: "HANGUL",
        0xBECA: "HANGUL",
        0xBECB: "HANGUL",
        0xBECC: "HANGUL",
        0xBECD: "HANGUL",
        0xBECE: "HANGUL",
        0xBECF: "HANGUL",
        0xBED0: "HANGUL",
        0xBED1: "HANGUL",
        0xBED2: "HANGUL",
        0xBED3: "HANGUL",
        0xBED4: "HANGUL",
        0xBED5: "HANGUL",
        0xBED6: "HANGUL",
        0xBED7: "HANGUL",
        0xBED8: "HANGUL",
        0xBED9: "HANGUL",
        0xBEDA: "HANGUL",
        0xBEDB: "HANGUL",
        0xBEDC: "HANGUL",
        0xBEDD: "HANGUL",
        0xBEDE: "HANGUL",
        0xBEDF: "HANGUL",
        0xBEE0: "HANGUL",
        0xBEE1: "HANGUL",
        0xBEE2: "HANGUL",
        0xBEE3: "HANGUL",
        0xBEE4: "HANGUL",
        0xBEE5: "HANGUL",
        0xBEE6: "HANGUL",
        0xBEE7: "HANGUL",
        0xBEE8: "HANGUL",
        0xBEE9: "HANGUL",
        0xBEEA: "HANGUL",
        0xBEEB: "HANGUL",
        0xBEEC: "HANGUL",
        0xBEED: "HANGUL",
        0xBEEE: "HANGUL",
        0xBEEF: "HANGUL",
        0xBEF0: "HANGUL",
        0xBEF1: "HANGUL",
        0xBEF2: "HANGUL",
        0xBEF3: "HANGUL",
        0xBEF4: "HANGUL",
        0xBEF5: "HANGUL",
        0xBEF6: "HANGUL",
        0xBEF7: "HANGUL",
        0xBEF8: "HANGUL",
        0xBEF9: "HANGUL",
        0xBEFA: "HANGUL",
        0xBEFB: "HANGUL",
        0xBEFC: "HANGUL",
        0xBEFD: "HANGUL",
        0xBEFE: "HANGUL",
        0xBEFF: "HANGUL",
        0xBF00: "HANGUL",
        0xBF01: "HANGUL",
        0xBF02: "HANGUL",
        0xBF03: "HANGUL",
        0xBF04: "HANGUL",
        0xBF05: "HANGUL",
        0xBF06: "HANGUL",
        0xBF07: "HANGUL",
        0xBF08: "HANGUL",
        0xBF09: "HANGUL",
        0xBF0A: "HANGUL",
        0xBF0B: "HANGUL",
        0xBF0C: "HANGUL",
        0xBF0D: "HANGUL",
        0xBF0E: "HANGUL",
        0xBF0F: "HANGUL",
        0xBF10: "HANGUL",
        0xBF11: "HANGUL",
        0xBF12: "HANGUL",
        0xBF13: "HANGUL",
        0xBF14: "HANGUL",
        0xBF15: "HANGUL",
        0xBF16: "HANGUL",
        0xBF17: "HANGUL",
        0xBF18: "HANGUL",
        0xBF19: "HANGUL",
        0xBF1A: "HANGUL",
        0xBF1B: "HANGUL",
        0xBF1C: "HANGUL",
        0xBF1D: "HANGUL",
        0xBF1E: "HANGUL",
        0xBF1F: "HANGUL",
        0xBF20: "HANGUL",
        0xBF21: "HANGUL",
        0xBF22: "HANGUL",
        0xBF23: "HANGUL",
        0xBF24: "HANGUL",
        0xBF25: "HANGUL",
        0xBF26: "HANGUL",
        0xBF27: "HANGUL",
        0xBF28: "HANGUL",
        0xBF29: "HANGUL",
        0xBF2A: "HANGUL",
        0xBF2B: "HANGUL",
        0xBF2C: "HANGUL",
        0xBF2D: "HANGUL",
        0xBF2E: "HANGUL",
        0xBF2F: "HANGUL",
        0xBF30: "HANGUL",
        0xBF31: "HANGUL",
        0xBF32: "HANGUL",
        0xBF33: "HANGUL",
        0xBF34: "HANGUL",
        0xBF35: "HANGUL",
        0xBF36: "HANGUL",
        0xBF37: "HANGUL",
        0xBF38: "HANGUL",
        0xBF39: "HANGUL",
        0xBF3A: "HANGUL",
        0xBF3B: "HANGUL",
        0xBF3C: "HANGUL",
        0xBF3D: "HANGUL",
        0xBF3E: "HANGUL",
        0xBF3F: "HANGUL",
        0xBF40: "HANGUL",
        0xBF41: "HANGUL",
        0xBF42: "HANGUL",
        0xBF43: "HANGUL",
        0xBF44: "HANGUL",
        0xBF45: "HANGUL",
        0xBF46: "HANGUL",
        0xBF47: "HANGUL",
        0xBF48: "HANGUL",
        0xBF49: "HANGUL",
        0xBF4A: "HANGUL",
        0xBF4B: "HANGUL",
        0xBF4C: "HANGUL",
        0xBF4D: "HANGUL",
        0xBF4E: "HANGUL",
        0xBF4F: "HANGUL",
        0xBF50: "HANGUL",
        0xBF51: "HANGUL",
        0xBF52: "HANGUL",
        0xBF53: "HANGUL",
        0xBF54: "HANGUL",
        0xBF55: "HANGUL",
        0xBF56: "HANGUL",
        0xBF57: "HANGUL",
        0xBF58: "HANGUL",
        0xBF59: "HANGUL",
        0xBF5A: "HANGUL",
        0xBF5B: "HANGUL",
        0xBF5C: "HANGUL",
        0xBF5D: "HANGUL",
        0xBF5E: "HANGUL",
        0xBF5F: "HANGUL",
        0xBF60: "HANGUL",
        0xBF61: "HANGUL",
        0xBF62: "HANGUL",
        0xBF63: "HANGUL",
        0xBF64: "HANGUL",
        0xBF65: "HANGUL",
        0xBF66: "HANGUL",
        0xBF67: "HANGUL",
        0xBF68: "HANGUL",
        0xBF69: "HANGUL",
        0xBF6A: "HANGUL",
        0xBF6B: "HANGUL",
        0xBF6C: "HANGUL",
        0xBF6D: "HANGUL",
        0xBF6E: "HANGUL",
        0xBF6F: "HANGUL",
        0xBF70: "HANGUL",
        0xBF71: "HANGUL",
        0xBF72: "HANGUL",
        0xBF73: "HANGUL",
        0xBF74: "HANGUL",
        0xBF75: "HANGUL",
        0xBF76: "HANGUL",
        0xBF77: "HANGUL",
        0xBF78: "HANGUL",
        0xBF79: "HANGUL",
        0xBF7A: "HANGUL",
        0xBF7B: "HANGUL",
        0xBF7C: "HANGUL",
        0xBF7D: "HANGUL",
        0xBF7E: "HANGUL",
        0xBF7F: "HANGUL",
        0xBF80: "HANGUL",
        0xBF81: "HANGUL",
        0xBF82: "HANGUL",
        0xBF83: "HANGUL",
        0xBF84: "HANGUL",
        0xBF85: "HANGUL",
        0xBF86: "HANGUL",
        0xBF87: "HANGUL",
        0xBF88: "HANGUL",
        0xBF89: "HANGUL",
        0xBF8A: "HANGUL",
        0xBF8B: "HANGUL",
        0xBF8C: "HANGUL",
        0xBF8D: "HANGUL",
        0xBF8E: "HANGUL",
        0xBF8F: "HANGUL",
        0xBF90: "HANGUL",
        0xBF91: "HANGUL",
        0xBF92: "HANGUL",
        0xBF93: "HANGUL",
        0xBF94: "HANGUL",
        0xBF95: "HANGUL",
        0xBF96: "HANGUL",
        0xBF97: "HANGUL",
        0xBF98: "HANGUL",
        0xBF99: "HANGUL",
        0xBF9A: "HANGUL",
        0xBF9B: "HANGUL",
        0xBF9C: "HANGUL",
        0xBF9D: "HANGUL",
        0xBF9E: "HANGUL",
        0xBF9F: "HANGUL",
        0xBFA0: "HANGUL",
        0xBFA1: "HANGUL",
        0xBFA2: "HANGUL",
        0xBFA3: "HANGUL",
        0xBFA4: "HANGUL",
        0xBFA5: "HANGUL",
        0xBFA6: "HANGUL",
        0xBFA7: "HANGUL",
        0xBFA8: "HANGUL",
        0xBFA9: "HANGUL",
        0xBFAA: "HANGUL",
        0xBFAB: "HANGUL",
        0xBFAC: "HANGUL",
        0xBFAD: "HANGUL",
        0xBFAE: "HANGUL",
        0xBFAF: "HANGUL",
        0xBFB0: "HANGUL",
        0xBFB1: "HANGUL",
        0xBFB2: "HANGUL",
        0xBFB3: "HANGUL",
        0xBFB4: "HANGUL",
        0xBFB5: "HANGUL",
        0xBFB6: "HANGUL",
        0xBFB7: "HANGUL",
        0xBFB8: "HANGUL",
        0xBFB9: "HANGUL",
        0xBFBA: "HANGUL",
        0xBFBB: "HANGUL",
        0xBFBC: "HANGUL",
        0xBFBD: "HANGUL",
        0xBFBE: "HANGUL",
        0xBFBF: "HANGUL",
        0xBFC0: "HANGUL",
        0xBFC1: "HANGUL",
        0xBFC2: "HANGUL",
        0xBFC3: "HANGUL",
        0xBFC4: "HANGUL",
        0xBFC5: "HANGUL",
        0xBFC6: "HANGUL",
        0xBFC7: "HANGUL",
        0xBFC8: "HANGUL",
        0xBFC9: "HANGUL",
        0xBFCA: "HANGUL",
        0xBFCB: "HANGUL",
        0xBFCC: "HANGUL",
        0xBFCD: "HANGUL",
        0xBFCE: "HANGUL",
        0xBFCF: "HANGUL",
        0xBFD0: "HANGUL",
        0xBFD1: "HANGUL",
        0xBFD2: "HANGUL",
        0xBFD3: "HANGUL",
        0xBFD4: "HANGUL",
        0xBFD5: "HANGUL",
        0xBFD6: "HANGUL",
        0xBFD7: "HANGUL",
        0xBFD8: "HANGUL",
        0xBFD9: "HANGUL",
        0xBFDA: "HANGUL",
        0xBFDB: "HANGUL",
        0xBFDC: "HANGUL",
        0xBFDD: "HANGUL",
        0xBFDE: "HANGUL",
        0xBFDF: "HANGUL",
        0xBFE0: "HANGUL",
        0xBFE1: "HANGUL",
        0xBFE2: "HANGUL",
        0xBFE3: "HANGUL",
        0xBFE4: "HANGUL",
        0xBFE5: "HANGUL",
        0xBFE6: "HANGUL",
        0xBFE7: "HANGUL",
        0xBFE8: "HANGUL",
        0xBFE9: "HANGUL",
        0xBFEA: "HANGUL",
        0xBFEB: "HANGUL",
        0xBFEC: "HANGUL",
        0xBFED: "HANGUL",
        0xBFEE: "HANGUL",
        0xBFEF: "HANGUL",
        0xBFF0: "HANGUL",
        0xBFF1: "HANGUL",
        0xBFF2: "HANGUL",
        0xBFF3: "HANGUL",
        0xBFF4: "HANGUL",
        0xBFF5: "HANGUL",
        0xBFF6: "HANGUL",
        0xBFF7: "HANGUL",
        0xBFF8: "HANGUL",
        0xBFF9: "HANGUL",
        0xBFFA: "HANGUL",
        0xBFFB: "HANGUL",
        0xBFFC: "HANGUL",
        0xBFFD: "HANGUL",
        0xBFFE: "HANGUL",
        0xBFFF: "HANGUL",
        0xC000: "HANGUL",
        0xC001: "HANGUL",
        0xC002: "HANGUL",
        0xC003: "HANGUL",
        0xC004: "HANGUL",
        0xC005: "HANGUL",
        0xC006: "HANGUL",
        0xC007: "HANGUL",
        0xC008: "HANGUL",
        0xC009: "HANGUL",
        0xC00A: "HANGUL",
        0xC00B: "HANGUL",
        0xC00C: "HANGUL",
        0xC00D: "HANGUL",
        0xC00E: "HANGUL",
        0xC00F: "HANGUL",
        0xC010: "HANGUL",
        0xC011: "HANGUL",
        0xC012: "HANGUL",
        0xC013: "HANGUL",
        0xC014: "HANGUL",
        0xC015: "HANGUL",
        0xC016: "HANGUL",
        0xC017: "HANGUL",
        0xC018: "HANGUL",
        0xC019: "HANGUL",
        0xC01A: "HANGUL",
        0xC01B: "HANGUL",
        0xC01C: "HANGUL",
        0xC01D: "HANGUL",
        0xC01E: "HANGUL",
        0xC01F: "HANGUL",
        0xC020: "HANGUL",
        0xC021: "HANGUL",
        0xC022: "HANGUL",
        0xC023: "HANGUL",
        0xC024: "HANGUL",
        0xC025: "HANGUL",
        0xC026: "HANGUL",
        0xC027: "HANGUL",
        0xC028: "HANGUL",
        0xC029: "HANGUL",
        0xC02A: "HANGUL",
        0xC02B: "HANGUL",
        0xC02C: "HANGUL",
        0xC02D: "HANGUL",
        0xC02E: "HANGUL",
        0xC02F: "HANGUL",
        0xC030: "HANGUL",
        0xC031: "HANGUL",
        0xC032: "HANGUL",
        0xC033: "HANGUL",
        0xC034: "HANGUL",
        0xC035: "HANGUL",
        0xC036: "HANGUL",
        0xC037: "HANGUL",
        0xC038: "HANGUL",
        0xC039: "HANGUL",
        0xC03A: "HANGUL",
        0xC03B: "HANGUL",
        0xC03C: "HANGUL",
        0xC03D: "HANGUL",
        0xC03E: "HANGUL",
        0xC03F: "HANGUL",
        0xC040: "HANGUL",
        0xC041: "HANGUL",
        0xC042: "HANGUL",
        0xC043: "HANGUL",
        0xC044: "HANGUL",
        0xC045: "HANGUL",
        0xC046: "HANGUL",
        0xC047: "HANGUL",
        0xC048: "HANGUL",
        0xC049: "HANGUL",
        0xC04A: "HANGUL",
        0xC04B: "HANGUL",
        0xC04C: "HANGUL",
        0xC04D: "HANGUL",
        0xC04E: "HANGUL",
        0xC04F: "HANGUL",
        0xC050: "HANGUL",
        0xC051: "HANGUL",
        0xC052: "HANGUL",
        0xC053: "HANGUL",
        0xC054: "HANGUL",
        0xC055: "HANGUL",
        0xC056: "HANGUL",
        0xC057: "HANGUL",
        0xC058: "HANGUL",
        0xC059: "HANGUL",
        0xC05A: "HANGUL",
        0xC05B: "HANGUL",
        0xC05C: "HANGUL",
        0xC05D: "HANGUL",
        0xC05E: "HANGUL",
        0xC05F: "HANGUL",
        0xC060: "HANGUL",
        0xC061: "HANGUL",
        0xC062: "HANGUL",
        0xC063: "HANGUL",
        0xC064: "HANGUL",
        0xC065: "HANGUL",
        0xC066: "HANGUL",
        0xC067: "HANGUL",
        0xC068: "HANGUL",
        0xC069: "HANGUL",
        0xC06A: "HANGUL",
        0xC06B: "HANGUL",
        0xC06C: "HANGUL",
        0xC06D: "HANGUL",
        0xC06E: "HANGUL",
        0xC06F: "HANGUL",
        0xC070: "HANGUL",
        0xC071: "HANGUL",
        0xC072: "HANGUL",
        0xC073: "HANGUL",
        0xC074: "HANGUL",
        0xC075: "HANGUL",
        0xC076: "HANGUL",
        0xC077: "HANGUL",
        0xC078: "HANGUL",
        0xC079: "HANGUL",
        0xC07A: "HANGUL",
        0xC07B: "HANGUL",
        0xC07C: "HANGUL",
        0xC07D: "HANGUL",
        0xC07E: "HANGUL",
        0xC07F: "HANGUL",
        0xC080: "HANGUL",
        0xC081: "HANGUL",
        0xC082: "HANGUL",
        0xC083: "HANGUL",
        0xC084: "HANGUL",
        0xC085: "HANGUL",
        0xC086: "HANGUL",
        0xC087: "HANGUL",
        0xC088: "HANGUL",
        0xC089: "HANGUL",
        0xC08A: "HANGUL",
        0xC08B: "HANGUL",
        0xC08C: "HANGUL",
        0xC08D: "HANGUL",
        0xC08E: "HANGUL",
        0xC08F: "HANGUL",
        0xC090: "HANGUL",
        0xC091: "HANGUL",
        0xC092: "HANGUL",
        0xC093: "HANGUL",
        0xC094: "HANGUL",
        0xC095: "HANGUL",
        0xC096: "HANGUL",
        0xC097: "HANGUL",
        0xC098: "HANGUL",
        0xC099: "HANGUL",
        0xC09A: "HANGUL",
        0xC09B: "HANGUL",
        0xC09C: "HANGUL",
        0xC09D: "HANGUL",
        0xC09E: "HANGUL",
        0xC09F: "HANGUL",
        0xC0A0: "HANGUL",
        0xC0A1: "HANGUL",
        0xC0A2: "HANGUL",
        0xC0A3: "HANGUL",
        0xC0A4: "HANGUL",
        0xC0A5: "HANGUL",
        0xC0A6: "HANGUL",
        0xC0A7: "HANGUL",
        0xC0A8: "HANGUL",
        0xC0A9: "HANGUL",
        0xC0AA: "HANGUL",
        0xC0AB: "HANGUL",
        0xC0AC: "HANGUL",
        0xC0AD: "HANGUL",
        0xC0AE: "HANGUL",
        0xC0AF: "HANGUL",
        0xC0B0: "HANGUL",
        0xC0B1: "HANGUL",
        0xC0B2: "HANGUL",
        0xC0B3: "HANGUL",
        0xC0B4: "HANGUL",
        0xC0B5: "HANGUL",
        0xC0B6: "HANGUL",
        0xC0B7: "HANGUL",
        0xC0B8: "HANGUL",
        0xC0B9: "HANGUL",
        0xC0BA: "HANGUL",
        0xC0BB: "HANGUL",
        0xC0BC: "HANGUL",
        0xC0BD: "HANGUL",
        0xC0BE: "HANGUL",
        0xC0BF: "HANGUL",
        0xC0C0: "HANGUL",
        0xC0C1: "HANGUL",
        0xC0C2: "HANGUL",
        0xC0C3: "HANGUL",
        0xC0C4: "HANGUL",
        0xC0C5: "HANGUL",
        0xC0C6: "HANGUL",
        0xC0C7: "HANGUL",
        0xC0C8: "HANGUL",
        0xC0C9: "HANGUL",
        0xC0CA: "HANGUL",
        0xC0CB: "HANGUL",
        0xC0CC: "HANGUL",
        0xC0CD: "HANGUL",
        0xC0CE: "HANGUL",
        0xC0CF: "HANGUL",
        0xC0D0: "HANGUL",
        0xC0D1: "HANGUL",
        0xC0D2: "HANGUL",
        0xC0D3: "HANGUL",
        0xC0D4: "HANGUL",
        0xC0D5: "HANGUL",
        0xC0D6: "HANGUL",
        0xC0D7: "HANGUL",
        0xC0D8: "HANGUL",
        0xC0D9: "HANGUL",
        0xC0DA: "HANGUL",
        0xC0DB: "HANGUL",
        0xC0DC: "HANGUL",
        0xC0DD: "HANGUL",
        0xC0DE: "HANGUL",
        0xC0DF: "HANGUL",
        0xC0E0: "HANGUL",
        0xC0E1: "HANGUL",
        0xC0E2: "HANGUL",
        0xC0E3: "HANGUL",
        0xC0E4: "HANGUL",
        0xC0E5: "HANGUL",
        0xC0E6: "HANGUL",
        0xC0E7: "HANGUL",
        0xC0E8: "HANGUL",
        0xC0E9: "HANGUL",
        0xC0EA: "HANGUL",
        0xC0EB: "HANGUL",
        0xC0EC: "HANGUL",
        0xC0ED: "HANGUL",
        0xC0EE: "HANGUL",
        0xC0EF: "HANGUL",
        0xC0F0: "HANGUL",
        0xC0F1: "HANGUL",
        0xC0F2: "HANGUL",
        0xC0F3: "HANGUL",
        0xC0F4: "HANGUL",
        0xC0F5: "HANGUL",
        0xC0F6: "HANGUL",
        0xC0F7: "HANGUL",
        0xC0F8: "HANGUL",
        0xC0F9: "HANGUL",
        0xC0FA: "HANGUL",
        0xC0FB: "HANGUL",
        0xC0FC: "HANGUL",
        0xC0FD: "HANGUL",
        0xC0FE: "HANGUL",
        0xC0FF: "HANGUL",
        0xC100: "HANGUL",
        0xC101: "HANGUL",
        0xC102: "HANGUL",
        0xC103: "HANGUL",
        0xC104: "HANGUL",
        0xC105: "HANGUL",
        0xC106: "HANGUL",
        0xC107: "HANGUL",
        0xC108: "HANGUL",
        0xC109: "HANGUL",
        0xC10A: "HANGUL",
        0xC10B: "HANGUL",
        0xC10C: "HANGUL",
        0xC10D: "HANGUL",
        0xC10E: "HANGUL",
        0xC10F: "HANGUL",
        0xC110: "HANGUL",
        0xC111: "HANGUL",
        0xC112: "HANGUL",
        0xC113: "HANGUL",
        0xC114: "HANGUL",
        0xC115: "HANGUL",
        0xC116: "HANGUL",
        0xC117: "HANGUL",
        0xC118: "HANGUL",
        0xC119: "HANGUL",
        0xC11A: "HANGUL",
        0xC11B: "HANGUL",
        0xC11C: "HANGUL",
        0xC11D: "HANGUL",
        0xC11E: "HANGUL",
        0xC11F: "HANGUL",
        0xC120: "HANGUL",
        0xC121: "HANGUL",
        0xC122: "HANGUL",
        0xC123: "HANGUL",
        0xC124: "HANGUL",
        0xC125: "HANGUL",
        0xC126: "HANGUL",
        0xC127: "HANGUL",
        0xC128: "HANGUL",
        0xC129: "HANGUL",
        0xC12A: "HANGUL",
        0xC12B: "HANGUL",
        0xC12C: "HANGUL",
        0xC12D: "HANGUL",
        0xC12E: "HANGUL",
        0xC12F: "HANGUL",
        0xC130: "HANGUL",
        0xC131: "HANGUL",
        0xC132: "HANGUL",
        0xC133: "HANGUL",
        0xC134: "HANGUL",
        0xC135: "HANGUL",
        0xC136: "HANGUL",
        0xC137: "HANGUL",
        0xC138: "HANGUL",
        0xC139: "HANGUL",
        0xC13A: "HANGUL",
        0xC13B: "HANGUL",
        0xC13C: "HANGUL",
        0xC13D: "HANGUL",
        0xC13E: "HANGUL",
        0xC13F: "HANGUL",
        0xC140: "HANGUL",
        0xC141: "HANGUL",
        0xC142: "HANGUL",
        0xC143: "HANGUL",
        0xC144: "HANGUL",
        0xC145: "HANGUL",
        0xC146: "HANGUL",
        0xC147: "HANGUL",
        0xC148: "HANGUL",
        0xC149: "HANGUL",
        0xC14A: "HANGUL",
        0xC14B: "HANGUL",
        0xC14C: "HANGUL",
        0xC14D: "HANGUL",
        0xC14E: "HANGUL",
        0xC14F: "HANGUL",
        0xC150: "HANGUL",
        0xC151: "HANGUL",
        0xC152: "HANGUL",
        0xC153: "HANGUL",
        0xC154: "HANGUL",
        0xC155: "HANGUL",
        0xC156: "HANGUL",
        0xC157: "HANGUL",
        0xC158: "HANGUL",
        0xC159: "HANGUL",
        0xC15A: "HANGUL",
        0xC15B: "HANGUL",
        0xC15C: "HANGUL",
        0xC15D: "HANGUL",
        0xC15E: "HANGUL",
        0xC15F: "HANGUL",
        0xC160: "HANGUL",
        0xC161: "HANGUL",
        0xC162: "HANGUL",
        0xC163: "HANGUL",
        0xC164: "HANGUL",
        0xC165: "HANGUL",
        0xC166: "HANGUL",
        0xC167: "HANGUL",
        0xC168: "HANGUL",
        0xC169: "HANGUL",
        0xC16A: "HANGUL",
        0xC16B: "HANGUL",
        0xC16C: "HANGUL",
        0xC16D: "HANGUL",
        0xC16E: "HANGUL",
        0xC16F: "HANGUL",
        0xC170: "HANGUL",
        0xC171: "HANGUL",
        0xC172: "HANGUL",
        0xC173: "HANGUL",
        0xC174: "HANGUL",
        0xC175: "HANGUL",
        0xC176: "HANGUL",
        0xC177: "HANGUL",
        0xC178: "HANGUL",
        0xC179: "HANGUL",
        0xC17A: "HANGUL",
        0xC17B: "HANGUL",
        0xC17C: "HANGUL",
        0xC17D: "HANGUL",
        0xC17E: "HANGUL",
        0xC17F: "HANGUL",
        0xC180: "HANGUL",
        0xC181: "HANGUL",
        0xC182: "HANGUL",
        0xC183: "HANGUL",
        0xC184: "HANGUL",
        0xC185: "HANGUL",
        0xC186: "HANGUL",
        0xC187: "HANGUL",
        0xC188: "HANGUL",
        0xC189: "HANGUL",
        0xC18A: "HANGUL",
        0xC18B: "HANGUL",
        0xC18C: "HANGUL",
        0xC18D: "HANGUL",
        0xC18E: "HANGUL",
        0xC18F: "HANGUL",
        0xC190: "HANGUL",
        0xC191: "HANGUL",
        0xC192: "HANGUL",
        0xC193: "HANGUL",
        0xC194: "HANGUL",
        0xC195: "HANGUL",
        0xC196: "HANGUL",
        0xC197: "HANGUL",
        0xC198: "HANGUL",
        0xC199: "HANGUL",
        0xC19A: "HANGUL",
        0xC19B: "HANGUL",
        0xC19C: "HANGUL",
        0xC19D: "HANGUL",
        0xC19E: "HANGUL",
        0xC19F: "HANGUL",
        0xC1A0: "HANGUL",
        0xC1A1: "HANGUL",
        0xC1A2: "HANGUL",
        0xC1A3: "HANGUL",
        0xC1A4: "HANGUL",
        0xC1A5: "HANGUL",
        0xC1A6: "HANGUL",
        0xC1A7: "HANGUL",
        0xC1A8: "HANGUL",
        0xC1A9: "HANGUL",
        0xC1AA: "HANGUL",
        0xC1AB: "HANGUL",
        0xC1AC: "HANGUL",
        0xC1AD: "HANGUL",
        0xC1AE: "HANGUL",
        0xC1AF: "HANGUL",
        0xC1B0: "HANGUL",
        0xC1B1: "HANGUL",
        0xC1B2: "HANGUL",
        0xC1B3: "HANGUL",
        0xC1B4: "HANGUL",
        0xC1B5: "HANGUL",
        0xC1B6: "HANGUL",
        0xC1B7: "HANGUL",
        0xC1B8: "HANGUL",
        0xC1B9: "HANGUL",
        0xC1BA: "HANGUL",
        0xC1BB: "HANGUL",
        0xC1BC: "HANGUL",
        0xC1BD: "HANGUL",
        0xC1BE: "HANGUL",
        0xC1BF: "HANGUL",
        0xC1C0: "HANGUL",
        0xC1C1: "HANGUL",
        0xC1C2: "HANGUL",
        0xC1C3: "HANGUL",
        0xC1C4: "HANGUL",
        0xC1C5: "HANGUL",
        0xC1C6: "HANGUL",
        0xC1C7: "HANGUL",
        0xC1C8: "HANGUL",
        0xC1C9: "HANGUL",
        0xC1CA: "HANGUL",
        0xC1CB: "HANGUL",
        0xC1CC: "HANGUL",
        0xC1CD: "HANGUL",
        0xC1CE: "HANGUL",
        0xC1CF: "HANGUL",
        0xC1D0: "HANGUL",
        0xC1D1: "HANGUL",
        0xC1D2: "HANGUL",
        0xC1D3: "HANGUL",
        0xC1D4: "HANGUL",
        0xC1D5: "HANGUL",
        0xC1D6: "HANGUL",
        0xC1D7: "HANGUL",
        0xC1D8: "HANGUL",
        0xC1D9: "HANGUL",
        0xC1DA: "HANGUL",
        0xC1DB: "HANGUL",
        0xC1DC: "HANGUL",
        0xC1DD: "HANGUL",
        0xC1DE: "HANGUL",
        0xC1DF: "HANGUL",
        0xC1E0: "HANGUL",
        0xC1E1: "HANGUL",
        0xC1E2: "HANGUL",
        0xC1E3: "HANGUL",
        0xC1E4: "HANGUL",
        0xC1E5: "HANGUL",
        0xC1E6: "HANGUL",
        0xC1E7: "HANGUL",
        0xC1E8: "HANGUL",
        0xC1E9: "HANGUL",
        0xC1EA: "HANGUL",
        0xC1EB: "HANGUL",
        0xC1EC: "HANGUL",
        0xC1ED: "HANGUL",
        0xC1EE: "HANGUL",
        0xC1EF: "HANGUL",
        0xC1F0: "HANGUL",
        0xC1F1: "HANGUL",
        0xC1F2: "HANGUL",
        0xC1F3: "HANGUL",
        0xC1F4: "HANGUL",
        0xC1F5: "HANGUL",
        0xC1F6: "HANGUL",
        0xC1F7: "HANGUL",
        0xC1F8: "HANGUL",
        0xC1F9: "HANGUL",
        0xC1FA: "HANGUL",
        0xC1FB: "HANGUL",
        0xC1FC: "HANGUL",
        0xC1FD: "HANGUL",
        0xC1FE: "HANGUL",
        0xC1FF: "HANGUL",
        0xC200: "HANGUL",
        0xC201: "HANGUL",
        0xC202: "HANGUL",
        0xC203: "HANGUL",
        0xC204: "HANGUL",
        0xC205: "HANGUL",
        0xC206: "HANGUL",
        0xC207: "HANGUL",
        0xC208: "HANGUL",
        0xC209: "HANGUL",
        0xC20A: "HANGUL",
        0xC20B: "HANGUL",
        0xC20C: "HANGUL",
        0xC20D: "HANGUL",
        0xC20E: "HANGUL",
        0xC20F: "HANGUL",
        0xC210: "HANGUL",
        0xC211: "HANGUL",
        0xC212: "HANGUL",
        0xC213: "HANGUL",
        0xC214: "HANGUL",
        0xC215: "HANGUL",
        0xC216: "HANGUL",
        0xC217: "HANGUL",
        0xC218: "HANGUL",
        0xC219: "HANGUL",
        0xC21A: "HANGUL",
        0xC21B: "HANGUL",
        0xC21C: "HANGUL",
        0xC21D: "HANGUL",
        0xC21E: "HANGUL",
        0xC21F: "HANGUL",
        0xC220: "HANGUL",
        0xC221: "HANGUL",
        0xC222: "HANGUL",
        0xC223: "HANGUL",
        0xC224: "HANGUL",
        0xC225: "HANGUL",
        0xC226: "HANGUL",
        0xC227: "HANGUL",
        0xC228: "HANGUL",
        0xC229: "HANGUL",
        0xC22A: "HANGUL",
        0xC22B: "HANGUL",
        0xC22C: "HANGUL",
        0xC22D: "HANGUL",
        0xC22E: "HANGUL",
        0xC22F: "HANGUL",
        0xC230: "HANGUL",
        0xC231: "HANGUL",
        0xC232: "HANGUL",
        0xC233: "HANGUL",
        0xC234: "HANGUL",
        0xC235: "HANGUL",
        0xC236: "HANGUL",
        0xC237: "HANGUL",
        0xC238: "HANGUL",
        0xC239: "HANGUL",
        0xC23A: "HANGUL",
        0xC23B: "HANGUL",
        0xC23C: "HANGUL",
        0xC23D: "HANGUL",
        0xC23E: "HANGUL",
        0xC23F: "HANGUL",
        0xC240: "HANGUL",
        0xC241: "HANGUL",
        0xC242: "HANGUL",
        0xC243: "HANGUL",
        0xC244: "HANGUL",
        0xC245: "HANGUL",
        0xC246: "HANGUL",
        0xC247: "HANGUL",
        0xC248: "HANGUL",
        0xC249: "HANGUL",
        0xC24A: "HANGUL",
        0xC24B: "HANGUL",
        0xC24C: "HANGUL",
        0xC24D: "HANGUL",
        0xC24E: "HANGUL",
        0xC24F: "HANGUL",
        0xC250: "HANGUL",
        0xC251: "HANGUL",
        0xC252: "HANGUL",
        0xC253: "HANGUL",
        0xC254: "HANGUL",
        0xC255: "HANGUL",
        0xC256: "HANGUL",
        0xC257: "HANGUL",
        0xC258: "HANGUL",
        0xC259: "HANGUL",
        0xC25A: "HANGUL",
        0xC25B: "HANGUL",
        0xC25C: "HANGUL",
        0xC25D: "HANGUL",
        0xC25E: "HANGUL",
        0xC25F: "HANGUL",
        0xC260: "HANGUL",
        0xC261: "HANGUL",
        0xC262: "HANGUL",
        0xC263: "HANGUL",
        0xC264: "HANGUL",
        0xC265: "HANGUL",
        0xC266: "HANGUL",
        0xC267: "HANGUL",
        0xC268: "HANGUL",
        0xC269: "HANGUL",
        0xC26A: "HANGUL",
        0xC26B: "HANGUL",
        0xC26C: "HANGUL",
        0xC26D: "HANGUL",
        0xC26E: "HANGUL",
        0xC26F: "HANGUL",
        0xC270: "HANGUL",
        0xC271: "HANGUL",
        0xC272: "HANGUL",
        0xC273: "HANGUL",
        0xC274: "HANGUL",
        0xC275: "HANGUL",
        0xC276: "HANGUL",
        0xC277: "HANGUL",
        0xC278: "HANGUL",
        0xC279: "HANGUL",
        0xC27A: "HANGUL",
        0xC27B: "HANGUL",
        0xC27C: "HANGUL",
        0xC27D: "HANGUL",
        0xC27E: "HANGUL",
        0xC27F: "HANGUL",
        0xC280: "HANGUL",
        0xC281: "HANGUL",
        0xC282: "HANGUL",
        0xC283: "HANGUL",
        0xC284: "HANGUL",
        0xC285: "HANGUL",
        0xC286: "HANGUL",
        0xC287: "HANGUL",
        0xC288: "HANGUL",
        0xC289: "HANGUL",
        0xC28A: "HANGUL",
        0xC28B: "HANGUL",
        0xC28C: "HANGUL",
        0xC28D: "HANGUL",
        0xC28E: "HANGUL",
        0xC28F: "HANGUL",
        0xC290: "HANGUL",
        0xC291: "HANGUL",
        0xC292: "HANGUL",
        0xC293: "HANGUL",
        0xC294: "HANGUL",
        0xC295: "HANGUL",
        0xC296: "HANGUL",
        0xC297: "HANGUL",
        0xC298: "HANGUL",
        0xC299: "HANGUL",
        0xC29A: "HANGUL",
        0xC29B: "HANGUL",
        0xC29C: "HANGUL",
        0xC29D: "HANGUL",
        0xC29E: "HANGUL",
        0xC29F: "HANGUL",
        0xC2A0: "HANGUL",
        0xC2A1: "HANGUL",
        0xC2A2: "HANGUL",
        0xC2A3: "HANGUL",
        0xC2A4: "HANGUL",
        0xC2A5: "HANGUL",
        0xC2A6: "HANGUL",
        0xC2A7: "HANGUL",
        0xC2A8: "HANGUL",
        0xC2A9: "HANGUL",
        0xC2AA: "HANGUL",
        0xC2AB: "HANGUL",
        0xC2AC: "HANGUL",
        0xC2AD: "HANGUL",
        0xC2AE: "HANGUL",
        0xC2AF: "HANGUL",
        0xC2B0: "HANGUL",
        0xC2B1: "HANGUL",
        0xC2B2: "HANGUL",
        0xC2B3: "HANGUL",
        0xC2B4: "HANGUL",
        0xC2B5: "HANGUL",
        0xC2B6: "HANGUL",
        0xC2B7: "HANGUL",
        0xC2B8: "HANGUL",
        0xC2B9: "HANGUL",
        0xC2BA: "HANGUL",
        0xC2BB: "HANGUL",
        0xC2BC: "HANGUL",
        0xC2BD: "HANGUL",
        0xC2BE: "HANGUL",
        0xC2BF: "HANGUL",
        0xC2C0: "HANGUL",
        0xC2C1: "HANGUL",
        0xC2C2: "HANGUL",
        0xC2C3: "HANGUL",
        0xC2C4: "HANGUL",
        0xC2C5: "HANGUL",
        0xC2C6: "HANGUL",
        0xC2C7: "HANGUL",
        0xC2C8: "HANGUL",
        0xC2C9: "HANGUL",
        0xC2CA: "HANGUL",
        0xC2CB: "HANGUL",
        0xC2CC: "HANGUL",
        0xC2CD: "HANGUL",
        0xC2CE: "HANGUL",
        0xC2CF: "HANGUL",
        0xC2D0: "HANGUL",
        0xC2D1: "HANGUL",
        0xC2D2: "HANGUL",
        0xC2D3: "HANGUL",
        0xC2D4: "HANGUL",
        0xC2D5: "HANGUL",
        0xC2D6: "HANGUL",
        0xC2D7: "HANGUL",
        0xC2D8: "HANGUL",
        0xC2D9: "HANGUL",
        0xC2DA: "HANGUL",
        0xC2DB: "HANGUL",
        0xC2DC: "HANGUL",
        0xC2DD: "HANGUL",
        0xC2DE: "HANGUL",
        0xC2DF: "HANGUL",
        0xC2E0: "HANGUL",
        0xC2E1: "HANGUL",
        0xC2E2: "HANGUL",
        0xC2E3: "HANGUL",
        0xC2E4: "HANGUL",
        0xC2E5: "HANGUL",
        0xC2E6: "HANGUL",
        0xC2E7: "HANGUL",
        0xC2E8: "HANGUL",
        0xC2E9: "HANGUL",
        0xC2EA: "HANGUL",
        0xC2EB: "HANGUL",
        0xC2EC: "HANGUL",
        0xC2ED: "HANGUL",
        0xC2EE: "HANGUL",
        0xC2EF: "HANGUL",
        0xC2F0: "HANGUL",
        0xC2F1: "HANGUL",
        0xC2F2: "HANGUL",
        0xC2F3: "HANGUL",
        0xC2F4: "HANGUL",
        0xC2F5: "HANGUL",
        0xC2F6: "HANGUL",
        0xC2F7: "HANGUL",
        0xC2F8: "HANGUL",
        0xC2F9: "HANGUL",
        0xC2FA: "HANGUL",
        0xC2FB: "HANGUL",
        0xC2FC: "HANGUL",
        0xC2FD: "HANGUL",
        0xC2FE: "HANGUL",
        0xC2FF: "HANGUL",
        0xC300: "HANGUL",
        0xC301: "HANGUL",
        0xC302: "HANGUL",
        0xC303: "HANGUL",
        0xC304: "HANGUL",
        0xC305: "HANGUL",
        0xC306: "HANGUL",
        0xC307: "HANGUL",
        0xC308: "HANGUL",
        0xC309: "HANGUL",
        0xC30A: "HANGUL",
        0xC30B: "HANGUL",
        0xC30C: "HANGUL",
        0xC30D: "HANGUL",
        0xC30E: "HANGUL",
        0xC30F: "HANGUL",
        0xC310: "HANGUL",
        0xC311: "HANGUL",
        0xC312: "HANGUL",
        0xC313: "HANGUL",
        0xC314: "HANGUL",
        0xC315: "HANGUL",
        0xC316: "HANGUL",
        0xC317: "HANGUL",
        0xC318: "HANGUL",
        0xC319: "HANGUL",
        0xC31A: "HANGUL",
        0xC31B: "HANGUL",
        0xC31C: "HANGUL",
        0xC31D: "HANGUL",
        0xC31E: "HANGUL",
        0xC31F: "HANGUL",
        0xC320: "HANGUL",
        0xC321: "HANGUL",
        0xC322: "HANGUL",
        0xC323: "HANGUL",
        0xC324: "HANGUL",
        0xC325: "HANGUL",
        0xC326: "HANGUL",
        0xC327: "HANGUL",
        0xC328: "HANGUL",
        0xC329: "HANGUL",
        0xC32A: "HANGUL",
        0xC32B: "HANGUL",
        0xC32C: "HANGUL",
        0xC32D: "HANGUL",
        0xC32E: "HANGUL",
        0xC32F: "HANGUL",
        0xC330: "HANGUL",
        0xC331: "HANGUL",
        0xC332: "HANGUL",
        0xC333: "HANGUL",
        0xC334: "HANGUL",
        0xC335: "HANGUL",
        0xC336: "HANGUL",
        0xC337: "HANGUL",
        0xC338: "HANGUL",
        0xC339: "HANGUL",
        0xC33A: "HANGUL",
        0xC33B: "HANGUL",
        0xC33C: "HANGUL",
        0xC33D: "HANGUL",
        0xC33E: "HANGUL",
        0xC33F: "HANGUL",
        0xC340: "HANGUL",
        0xC341: "HANGUL",
        0xC342: "HANGUL",
        0xC343: "HANGUL",
        0xC344: "HANGUL",
        0xC345: "HANGUL",
        0xC346: "HANGUL",
        0xC347: "HANGUL",
        0xC348: "HANGUL",
        0xC349: "HANGUL",
        0xC34A: "HANGUL",
        0xC34B: "HANGUL",
        0xC34C: "HANGUL",
        0xC34D: "HANGUL",
        0xC34E: "HANGUL",
        0xC34F: "HANGUL",
        0xC350: "HANGUL",
        0xC351: "HANGUL",
        0xC352: "HANGUL",
        0xC353: "HANGUL",
        0xC354: "HANGUL",
        0xC355: "HANGUL",
        0xC356: "HANGUL",
        0xC357: "HANGUL",
        0xC358: "HANGUL",
        0xC359: "HANGUL",
        0xC35A: "HANGUL",
        0xC35B: "HANGUL",
        0xC35C: "HANGUL",
        0xC35D: "HANGUL",
        0xC35E: "HANGUL",
        0xC35F: "HANGUL",
        0xC360: "HANGUL",
        0xC361: "HANGUL",
        0xC362: "HANGUL",
        0xC363: "HANGUL",
        0xC364: "HANGUL",
        0xC365: "HANGUL",
        0xC366: "HANGUL",
        0xC367: "HANGUL",
        0xC368: "HANGUL",
        0xC369: "HANGUL",
        0xC36A: "HANGUL",
        0xC36B: "HANGUL",
        0xC36C: "HANGUL",
        0xC36D: "HANGUL",
        0xC36E: "HANGUL",
        0xC36F: "HANGUL",
        0xC370: "HANGUL",
        0xC371: "HANGUL",
        0xC372: "HANGUL",
        0xC373: "HANGUL",
        0xC374: "HANGUL",
        0xC375: "HANGUL",
        0xC376: "HANGUL",
        0xC377: "HANGUL",
        0xC378: "HANGUL",
        0xC379: "HANGUL",
        0xC37A: "HANGUL",
        0xC37B: "HANGUL",
        0xC37C: "HANGUL",
        0xC37D: "HANGUL",
        0xC37E: "HANGUL",
        0xC37F: "HANGUL",
        0xC380: "HANGUL",
        0xC381: "HANGUL",
        0xC382: "HANGUL",
        0xC383: "HANGUL",
        0xC384: "HANGUL",
        0xC385: "HANGUL",
        0xC386: "HANGUL",
        0xC387: "HANGUL",
        0xC388: "HANGUL",
        0xC389: "HANGUL",
        0xC38A: "HANGUL",
        0xC38B: "HANGUL",
        0xC38C: "HANGUL",
        0xC38D: "HANGUL",
        0xC38E: "HANGUL",
        0xC38F: "HANGUL",
        0xC390: "HANGUL",
        0xC391: "HANGUL",
        0xC392: "HANGUL",
        0xC393: "HANGUL",
        0xC394: "HANGUL",
        0xC395: "HANGUL",
        0xC396: "HANGUL",
        0xC397: "HANGUL",
        0xC398: "HANGUL",
        0xC399: "HANGUL",
        0xC39A: "HANGUL",
        0xC39B: "HANGUL",
        0xC39C: "HANGUL",
        0xC39D: "HANGUL",
        0xC39E: "HANGUL",
        0xC39F: "HANGUL",
        0xC3A0: "HANGUL",
        0xC3A1: "HANGUL",
        0xC3A2: "HANGUL",
        0xC3A3: "HANGUL",
        0xC3A4: "HANGUL",
        0xC3A5: "HANGUL",
        0xC3A6: "HANGUL",
        0xC3A7: "HANGUL",
        0xC3A8: "HANGUL",
        0xC3A9: "HANGUL",
        0xC3AA: "HANGUL",
        0xC3AB: "HANGUL",
        0xC3AC: "HANGUL",
        0xC3AD: "HANGUL",
        0xC3AE: "HANGUL",
        0xC3AF: "HANGUL",
        0xC3B0: "HANGUL",
        0xC3B1: "HANGUL",
        0xC3B2: "HANGUL",
        0xC3B3: "HANGUL",
        0xC3B4: "HANGUL",
        0xC3B5: "HANGUL",
        0xC3B6: "HANGUL",
        0xC3B7: "HANGUL",
        0xC3B8: "HANGUL",
        0xC3B9: "HANGUL",
        0xC3BA: "HANGUL",
        0xC3BB: "HANGUL",
        0xC3BC: "HANGUL",
        0xC3BD: "HANGUL",
        0xC3BE: "HANGUL",
        0xC3BF: "HANGUL",
        0xC3C0: "HANGUL",
        0xC3C1: "HANGUL",
        0xC3C2: "HANGUL",
        0xC3C3: "HANGUL",
        0xC3C4: "HANGUL",
        0xC3C5: "HANGUL",
        0xC3C6: "HANGUL",
        0xC3C7: "HANGUL",
        0xC3C8: "HANGUL",
        0xC3C9: "HANGUL",
        0xC3CA: "HANGUL",
        0xC3CB: "HANGUL",
        0xC3CC: "HANGUL",
        0xC3CD: "HANGUL",
        0xC3CE: "HANGUL",
        0xC3CF: "HANGUL",
        0xC3D0: "HANGUL",
        0xC3D1: "HANGUL",
        0xC3D2: "HANGUL",
        0xC3D3: "HANGUL",
        0xC3D4: "HANGUL",
        0xC3D5: "HANGUL",
        0xC3D6: "HANGUL",
        0xC3D7: "HANGUL",
        0xC3D8: "HANGUL",
        0xC3D9: "HANGUL",
        0xC3DA: "HANGUL",
        0xC3DB: "HANGUL",
        0xC3DC: "HANGUL",
        0xC3DD: "HANGUL",
        0xC3DE: "HANGUL",
        0xC3DF: "HANGUL",
        0xC3E0: "HANGUL",
        0xC3E1: "HANGUL",
        0xC3E2: "HANGUL",
        0xC3E3: "HANGUL",
        0xC3E4: "HANGUL",
        0xC3E5: "HANGUL",
        0xC3E6: "HANGUL",
        0xC3E7: "HANGUL",
        0xC3E8: "HANGUL",
        0xC3E9: "HANGUL",
        0xC3EA: "HANGUL",
        0xC3EB: "HANGUL",
        0xC3EC: "HANGUL",
        0xC3ED: "HANGUL",
        0xC3EE: "HANGUL",
        0xC3EF: "HANGUL",
        0xC3F0: "HANGUL",
        0xC3F1: "HANGUL",
        0xC3F2: "HANGUL",
        0xC3F3: "HANGUL",
        0xC3F4: "HANGUL",
        0xC3F5: "HANGUL",
        0xC3F6: "HANGUL",
        0xC3F7: "HANGUL",
        0xC3F8: "HANGUL",
        0xC3F9: "HANGUL",
        0xC3FA: "HANGUL",
        0xC3FB: "HANGUL",
        0xC3FC: "HANGUL",
        0xC3FD: "HANGUL",
        0xC3FE: "HANGUL",
        0xC3FF: "HANGUL",
        0xC400: "HANGUL",
        0xC401: "HANGUL",
        0xC402: "HANGUL",
        0xC403: "HANGUL",
        0xC404: "HANGUL",
        0xC405: "HANGUL",
        0xC406: "HANGUL",
        0xC407: "HANGUL",
        0xC408: "HANGUL",
        0xC409: "HANGUL",
        0xC40A: "HANGUL",
        0xC40B: "HANGUL",
        0xC40C: "HANGUL",
        0xC40D: "HANGUL",
        0xC40E: "HANGUL",
        0xC40F: "HANGUL",
        0xC410: "HANGUL",
        0xC411: "HANGUL",
        0xC412: "HANGUL",
        0xC413: "HANGUL",
        0xC414: "HANGUL",
        0xC415: "HANGUL",
        0xC416: "HANGUL",
        0xC417: "HANGUL",
        0xC418: "HANGUL",
        0xC419: "HANGUL",
        0xC41A: "HANGUL",
        0xC41B: "HANGUL",
        0xC41C: "HANGUL",
        0xC41D: "HANGUL",
        0xC41E: "HANGUL",
        0xC41F: "HANGUL",
        0xC420: "HANGUL",
        0xC421: "HANGUL",
        0xC422: "HANGUL",
        0xC423: "HANGUL",
        0xC424: "HANGUL",
        0xC425: "HANGUL",
        0xC426: "HANGUL",
        0xC427: "HANGUL",
        0xC428: "HANGUL",
        0xC429: "HANGUL",
        0xC42A: "HANGUL",
        0xC42B: "HANGUL",
        0xC42C: "HANGUL",
        0xC42D: "HANGUL",
        0xC42E: "HANGUL",
        0xC42F: "HANGUL",
        0xC430: "HANGUL",
        0xC431: "HANGUL",
        0xC432: "HANGUL",
        0xC433: "HANGUL",
        0xC434: "HANGUL",
        0xC435: "HANGUL",
        0xC436: "HANGUL",
        0xC437: "HANGUL",
        0xC438: "HANGUL",
        0xC439: "HANGUL",
        0xC43A: "HANGUL",
        0xC43B: "HANGUL",
        0xC43C: "HANGUL",
        0xC43D: "HANGUL",
        0xC43E: "HANGUL",
        0xC43F: "HANGUL",
        0xC440: "HANGUL",
        0xC441: "HANGUL",
        0xC442: "HANGUL",
        0xC443: "HANGUL",
        0xC444: "HANGUL",
        0xC445: "HANGUL",
        0xC446: "HANGUL",
        0xC447: "HANGUL",
        0xC448: "HANGUL",
        0xC449: "HANGUL",
        0xC44A: "HANGUL",
        0xC44B: "HANGUL",
        0xC44C: "HANGUL",
        0xC44D: "HANGUL",
        0xC44E: "HANGUL",
        0xC44F: "HANGUL",
        0xC450: "HANGUL",
        0xC451: "HANGUL",
        0xC452: "HANGUL",
        0xC453: "HANGUL",
        0xC454: "HANGUL",
        0xC455: "HANGUL",
        0xC456: "HANGUL",
        0xC457: "HANGUL",
        0xC458: "HANGUL",
        0xC459: "HANGUL",
        0xC45A: "HANGUL",
        0xC45B: "HANGUL",
        0xC45C: "HANGUL",
        0xC45D: "HANGUL",
        0xC45E: "HANGUL",
        0xC45F: "HANGUL",
        0xC460: "HANGUL",
        0xC461: "HANGUL",
        0xC462: "HANGUL",
        0xC463: "HANGUL",
        0xC464: "HANGUL",
        0xC465: "HANGUL",
        0xC466: "HANGUL",
        0xC467: "HANGUL",
        0xC468: "HANGUL",
        0xC469: "HANGUL",
        0xC46A: "HANGUL",
        0xC46B: "HANGUL",
        0xC46C: "HANGUL",
        0xC46D: "HANGUL",
        0xC46E: "HANGUL",
        0xC46F: "HANGUL",
        0xC470: "HANGUL",
        0xC471: "HANGUL",
        0xC472: "HANGUL",
        0xC473: "HANGUL",
        0xC474: "HANGUL",
        0xC475: "HANGUL",
        0xC476: "HANGUL",
        0xC477: "HANGUL",
        0xC478: "HANGUL",
        0xC479: "HANGUL",
        0xC47A: "HANGUL",
        0xC47B: "HANGUL",
        0xC47C: "HANGUL",
        0xC47D: "HANGUL",
        0xC47E: "HANGUL",
        0xC47F: "HANGUL",
        0xC480: "HANGUL",
        0xC481: "HANGUL",
        0xC482: "HANGUL",
        0xC483: "HANGUL",
        0xC484: "HANGUL",
        0xC485: "HANGUL",
        0xC486: "HANGUL",
        0xC487: "HANGUL",
        0xC488: "HANGUL",
        0xC489: "HANGUL",
        0xC48A: "HANGUL",
        0xC48B: "HANGUL",
        0xC48C: "HANGUL",
        0xC48D: "HANGUL",
        0xC48E: "HANGUL",
        0xC48F: "HANGUL",
        0xC490: "HANGUL",
        0xC491: "HANGUL",
        0xC492: "HANGUL",
        0xC493: "HANGUL",
        0xC494: "HANGUL",
        0xC495: "HANGUL",
        0xC496: "HANGUL",
        0xC497: "HANGUL",
        0xC498: "HANGUL",
        0xC499: "HANGUL",
        0xC49A: "HANGUL",
        0xC49B: "HANGUL",
        0xC49C: "HANGUL",
        0xC49D: "HANGUL",
        0xC49E: "HANGUL",
        0xC49F: "HANGUL",
        0xC4A0: "HANGUL",
        0xC4A1: "HANGUL",
        0xC4A2: "HANGUL",
        0xC4A3: "HANGUL",
        0xC4A4: "HANGUL",
        0xC4A5: "HANGUL",
        0xC4A6: "HANGUL",
        0xC4A7: "HANGUL",
        0xC4A8: "HANGUL",
        0xC4A9: "HANGUL",
        0xC4AA: "HANGUL",
        0xC4AB: "HANGUL",
        0xC4AC: "HANGUL",
        0xC4AD: "HANGUL",
        0xC4AE: "HANGUL",
        0xC4AF: "HANGUL",
        0xC4B0: "HANGUL",
        0xC4B1: "HANGUL",
        0xC4B2: "HANGUL",
        0xC4B3: "HANGUL",
        0xC4B4: "HANGUL",
        0xC4B5: "HANGUL",
        0xC4B6: "HANGUL",
        0xC4B7: "HANGUL",
        0xC4B8: "HANGUL",
        0xC4B9: "HANGUL",
        0xC4BA: "HANGUL",
        0xC4BB: "HANGUL",
        0xC4BC: "HANGUL",
        0xC4BD: "HANGUL",
        0xC4BE: "HANGUL",
        0xC4BF: "HANGUL",
        0xC4C0: "HANGUL",
        0xC4C1: "HANGUL",
        0xC4C2: "HANGUL",
        0xC4C3: "HANGUL",
        0xC4C4: "HANGUL",
        0xC4C5: "HANGUL",
        0xC4C6: "HANGUL",
        0xC4C7: "HANGUL",
        0xC4C8: "HANGUL",
        0xC4C9: "HANGUL",
        0xC4CA: "HANGUL",
        0xC4CB: "HANGUL",
        0xC4CC: "HANGUL",
        0xC4CD: "HANGUL",
        0xC4CE: "HANGUL",
        0xC4CF: "HANGUL",
        0xC4D0: "HANGUL",
        0xC4D1: "HANGUL",
        0xC4D2: "HANGUL",
        0xC4D3: "HANGUL",
        0xC4D4: "HANGUL",
        0xC4D5: "HANGUL",
        0xC4D6: "HANGUL",
        0xC4D7: "HANGUL",
        0xC4D8: "HANGUL",
        0xC4D9: "HANGUL",
        0xC4DA: "HANGUL",
        0xC4DB: "HANGUL",
        0xC4DC: "HANGUL",
        0xC4DD: "HANGUL",
        0xC4DE: "HANGUL",
        0xC4DF: "HANGUL",
        0xC4E0: "HANGUL",
        0xC4E1: "HANGUL",
        0xC4E2: "HANGUL",
        0xC4E3: "HANGUL",
        0xC4E4: "HANGUL",
        0xC4E5: "HANGUL",
        0xC4E6: "HANGUL",
        0xC4E7: "HANGUL",
        0xC4E8: "HANGUL",
        0xC4E9: "HANGUL",
        0xC4EA: "HANGUL",
        0xC4EB: "HANGUL",
        0xC4EC: "HANGUL",
        0xC4ED: "HANGUL",
        0xC4EE: "HANGUL",
        0xC4EF: "HANGUL",
        0xC4F0: "HANGUL",
        0xC4F1: "HANGUL",
        0xC4F2: "HANGUL",
        0xC4F3: "HANGUL",
        0xC4F4: "HANGUL",
        0xC4F5: "HANGUL",
        0xC4F6: "HANGUL",
        0xC4F7: "HANGUL",
        0xC4F8: "HANGUL",
        0xC4F9: "HANGUL",
        0xC4FA: "HANGUL",
        0xC4FB: "HANGUL",
        0xC4FC: "HANGUL",
        0xC4FD: "HANGUL",
        0xC4FE: "HANGUL",
        0xC4FF: "HANGUL",
        0xC500: "HANGUL",
        0xC501: "HANGUL",
        0xC502: "HANGUL",
        0xC503: "HANGUL",
        0xC504: "HANGUL",
        0xC505: "HANGUL",
        0xC506: "HANGUL",
        0xC507: "HANGUL",
        0xC508: "HANGUL",
        0xC509: "HANGUL",
        0xC50A: "HANGUL",
        0xC50B: "HANGUL",
        0xC50C: "HANGUL",
        0xC50D: "HANGUL",
        0xC50E: "HANGUL",
        0xC50F: "HANGUL",
        0xC510: "HANGUL",
        0xC511: "HANGUL",
        0xC512: "HANGUL",
        0xC513: "HANGUL",
        0xC514: "HANGUL",
        0xC515: "HANGUL",
        0xC516: "HANGUL",
        0xC517: "HANGUL",
        0xC518: "HANGUL",
        0xC519: "HANGUL",
        0xC51A: "HANGUL",
        0xC51B: "HANGUL",
        0xC51C: "HANGUL",
        0xC51D: "HANGUL",
        0xC51E: "HANGUL",
        0xC51F: "HANGUL",
        0xC520: "HANGUL",
        0xC521: "HANGUL",
        0xC522: "HANGUL",
        0xC523: "HANGUL",
        0xC524: "HANGUL",
        0xC525: "HANGUL",
        0xC526: "HANGUL",
        0xC527: "HANGUL",
        0xC528: "HANGUL",
        0xC529: "HANGUL",
        0xC52A: "HANGUL",
        0xC52B: "HANGUL",
        0xC52C: "HANGUL",
        0xC52D: "HANGUL",
        0xC52E: "HANGUL",
        0xC52F: "HANGUL",
        0xC530: "HANGUL",
        0xC531: "HANGUL",
        0xC532: "HANGUL",
        0xC533: "HANGUL",
        0xC534: "HANGUL",
        0xC535: "HANGUL",
        0xC536: "HANGUL",
        0xC537: "HANGUL",
        0xC538: "HANGUL",
        0xC539: "HANGUL",
        0xC53A: "HANGUL",
        0xC53B: "HANGUL",
        0xC53C: "HANGUL",
        0xC53D: "HANGUL",
        0xC53E: "HANGUL",
        0xC53F: "HANGUL",
        0xC540: "HANGUL",
        0xC541: "HANGUL",
        0xC542: "HANGUL",
        0xC543: "HANGUL",
        0xC544: "HANGUL",
        0xC545: "HANGUL",
        0xC546: "HANGUL",
        0xC547: "HANGUL",
        0xC548: "HANGUL",
        0xC549: "HANGUL",
        0xC54A: "HANGUL",
        0xC54B: "HANGUL",
        0xC54C: "HANGUL",
        0xC54D: "HANGUL",
        0xC54E: "HANGUL",
        0xC54F: "HANGUL",
        0xC550: "HANGUL",
        0xC551: "HANGUL",
        0xC552: "HANGUL",
        0xC553: "HANGUL",
        0xC554: "HANGUL",
        0xC555: "HANGUL",
        0xC556: "HANGUL",
        0xC557: "HANGUL",
        0xC558: "HANGUL",
        0xC559: "HANGUL",
        0xC55A: "HANGUL",
        0xC55B: "HANGUL",
        0xC55C: "HANGUL",
        0xC55D: "HANGUL",
        0xC55E: "HANGUL",
        0xC55F: "HANGUL",
        0xC560: "HANGUL",
        0xC561: "HANGUL",
        0xC562: "HANGUL",
        0xC563: "HANGUL",
        0xC564: "HANGUL",
        0xC565: "HANGUL",
        0xC566: "HANGUL",
        0xC567: "HANGUL",
        0xC568: "HANGUL",
        0xC569: "HANGUL",
        0xC56A: "HANGUL",
        0xC56B: "HANGUL",
        0xC56C: "HANGUL",
        0xC56D: "HANGUL",
        0xC56E: "HANGUL",
        0xC56F: "HANGUL",
        0xC570: "HANGUL",
        0xC571: "HANGUL",
        0xC572: "HANGUL",
        0xC573: "HANGUL",
        0xC574: "HANGUL",
        0xC575: "HANGUL",
        0xC576: "HANGUL",
        0xC577: "HANGUL",
        0xC578: "HANGUL",
        0xC579: "HANGUL",
        0xC57A: "HANGUL",
        0xC57B: "HANGUL",
        0xC57C: "HANGUL",
        0xC57D: "HANGUL",
        0xC57E: "HANGUL",
        0xC57F: "HANGUL",
        0xC580: "HANGUL",
        0xC581: "HANGUL",
        0xC582: "HANGUL",
        0xC583: "HANGUL",
        0xC584: "HANGUL",
        0xC585: "HANGUL",
        0xC586: "HANGUL",
        0xC587: "HANGUL",
        0xC588: "HANGUL",
        0xC589: "HANGUL",
        0xC58A: "HANGUL",
        0xC58B: "HANGUL",
        0xC58C: "HANGUL",
        0xC58D: "HANGUL",
        0xC58E: "HANGUL",
        0xC58F: "HANGUL",
        0xC590: "HANGUL",
        0xC591: "HANGUL",
        0xC592: "HANGUL",
        0xC593: "HANGUL",
        0xC594: "HANGUL",
        0xC595: "HANGUL",
        0xC596: "HANGUL",
        0xC597: "HANGUL",
        0xC598: "HANGUL",
        0xC599: "HANGUL",
        0xC59A: "HANGUL",
        0xC59B: "HANGUL",
        0xC59C: "HANGUL",
        0xC59D: "HANGUL",
        0xC59E: "HANGUL",
        0xC59F: "HANGUL",
        0xC5A0: "HANGUL",
        0xC5A1: "HANGUL",
        0xC5A2: "HANGUL",
        0xC5A3: "HANGUL",
        0xC5A4: "HANGUL",
        0xC5A5: "HANGUL",
        0xC5A6: "HANGUL",
        0xC5A7: "HANGUL",
        0xC5A8: "HANGUL",
        0xC5A9: "HANGUL",
        0xC5AA: "HANGUL",
        0xC5AB: "HANGUL",
        0xC5AC: "HANGUL",
        0xC5AD: "HANGUL",
        0xC5AE: "HANGUL",
        0xC5AF: "HANGUL",
        0xC5B0: "HANGUL",
        0xC5B1: "HANGUL",
        0xC5B2: "HANGUL",
        0xC5B3: "HANGUL",
        0xC5B4: "HANGUL",
        0xC5B5: "HANGUL",
        0xC5B6: "HANGUL",
        0xC5B7: "HANGUL",
        0xC5B8: "HANGUL",
        0xC5B9: "HANGUL",
        0xC5BA: "HANGUL",
        0xC5BB: "HANGUL",
        0xC5BC: "HANGUL",
        0xC5BD: "HANGUL",
        0xC5BE: "HANGUL",
        0xC5BF: "HANGUL",
        0xC5C0: "HANGUL",
        0xC5C1: "HANGUL",
        0xC5C2: "HANGUL",
        0xC5C3: "HANGUL",
        0xC5C4: "HANGUL",
        0xC5C5: "HANGUL",
        0xC5C6: "HANGUL",
        0xC5C7: "HANGUL",
        0xC5C8: "HANGUL",
        0xC5C9: "HANGUL",
        0xC5CA: "HANGUL",
        0xC5CB: "HANGUL",
        0xC5CC: "HANGUL",
        0xC5CD: "HANGUL",
        0xC5CE: "HANGUL",
        0xC5CF: "HANGUL",
        0xC5D0: "HANGUL",
        0xC5D1: "HANGUL",
        0xC5D2: "HANGUL",
        0xC5D3: "HANGUL",
        0xC5D4: "HANGUL",
        0xC5D5: "HANGUL",
        0xC5D6: "HANGUL",
        0xC5D7: "HANGUL",
        0xC5D8: "HANGUL",
        0xC5D9: "HANGUL",
        0xC5DA: "HANGUL",
        0xC5DB: "HANGUL",
        0xC5DC: "HANGUL",
        0xC5DD: "HANGUL",
        0xC5DE: "HANGUL",
        0xC5DF: "HANGUL",
        0xC5E0: "HANGUL",
        0xC5E1: "HANGUL",
        0xC5E2: "HANGUL",
        0xC5E3: "HANGUL",
        0xC5E4: "HANGUL",
        0xC5E5: "HANGUL",
        0xC5E6: "HANGUL",
        0xC5E7: "HANGUL",
        0xC5E8: "HANGUL",
        0xC5E9: "HANGUL",
        0xC5EA: "HANGUL",
        0xC5EB: "HANGUL",
        0xC5EC: "HANGUL",
        0xC5ED: "HANGUL",
        0xC5EE: "HANGUL",
        0xC5EF: "HANGUL",
        0xC5F0: "HANGUL",
        0xC5F1: "HANGUL",
        0xC5F2: "HANGUL",
        0xC5F3: "HANGUL",
        0xC5F4: "HANGUL",
        0xC5F5: "HANGUL",
        0xC5F6: "HANGUL",
        0xC5F7: "HANGUL",
        0xC5F8: "HANGUL",
        0xC5F9: "HANGUL",
        0xC5FA: "HANGUL",
        0xC5FB: "HANGUL",
        0xC5FC: "HANGUL",
        0xC5FD: "HANGUL",
        0xC5FE: "HANGUL",
        0xC5FF: "HANGUL",
        0xC600: "HANGUL",
        0xC601: "HANGUL",
        0xC602: "HANGUL",
        0xC603: "HANGUL",
        0xC604: "HANGUL",
        0xC605: "HANGUL",
        0xC606: "HANGUL",
        0xC607: "HANGUL",
        0xC608: "HANGUL",
        0xC609: "HANGUL",
        0xC60A: "HANGUL",
        0xC60B: "HANGUL",
        0xC60C: "HANGUL",
        0xC60D: "HANGUL",
        0xC60E: "HANGUL",
        0xC60F: "HANGUL",
        0xC610: "HANGUL",
        0xC611: "HANGUL",
        0xC612: "HANGUL",
        0xC613: "HANGUL",
        0xC614: "HANGUL",
        0xC615: "HANGUL",
        0xC616: "HANGUL",
        0xC617: "HANGUL",
        0xC618: "HANGUL",
        0xC619: "HANGUL",
        0xC61A: "HANGUL",
        0xC61B: "HANGUL",
        0xC61C: "HANGUL",
        0xC61D: "HANGUL",
        0xC61E: "HANGUL",
        0xC61F: "HANGUL",
        0xC620: "HANGUL",
        0xC621: "HANGUL",
        0xC622: "HANGUL",
        0xC623: "HANGUL",
        0xC624: "HANGUL",
        0xC625: "HANGUL",
        0xC626: "HANGUL",
        0xC627: "HANGUL",
        0xC628: "HANGUL",
        0xC629: "HANGUL",
        0xC62A: "HANGUL",
        0xC62B: "HANGUL",
        0xC62C: "HANGUL",
        0xC62D: "HANGUL",
        0xC62E: "HANGUL",
        0xC62F: "HANGUL",
        0xC630: "HANGUL",
        0xC631: "HANGUL",
        0xC632: "HANGUL",
        0xC633: "HANGUL",
        0xC634: "HANGUL",
        0xC635: "HANGUL",
        0xC636: "HANGUL",
        0xC637: "HANGUL",
        0xC638: "HANGUL",
        0xC639: "HANGUL",
        0xC63A: "HANGUL",
        0xC63B: "HANGUL",
        0xC63C: "HANGUL",
        0xC63D: "HANGUL",
        0xC63E: "HANGUL",
        0xC63F: "HANGUL",
        0xC640: "HANGUL",
        0xC641: "HANGUL",
        0xC642: "HANGUL",
        0xC643: "HANGUL",
        0xC644: "HANGUL",
        0xC645: "HANGUL",
        0xC646: "HANGUL",
        0xC647: "HANGUL",
        0xC648: "HANGUL",
        0xC649: "HANGUL",
        0xC64A: "HANGUL",
        0xC64B: "HANGUL",
        0xC64C: "HANGUL",
        0xC64D: "HANGUL",
        0xC64E: "HANGUL",
        0xC64F: "HANGUL",
        0xC650: "HANGUL",
        0xC651: "HANGUL",
        0xC652: "HANGUL",
        0xC653: "HANGUL",
        0xC654: "HANGUL",
        0xC655: "HANGUL",
        0xC656: "HANGUL",
        0xC657: "HANGUL",
        0xC658: "HANGUL",
        0xC659: "HANGUL",
        0xC65A: "HANGUL",
        0xC65B: "HANGUL",
        0xC65C: "HANGUL",
        0xC65D: "HANGUL",
        0xC65E: "HANGUL",
        0xC65F: "HANGUL",
        0xC660: "HANGUL",
        0xC661: "HANGUL",
        0xC662: "HANGUL",
        0xC663: "HANGUL",
        0xC664: "HANGUL",
        0xC665: "HANGUL",
        0xC666: "HANGUL",
        0xC667: "HANGUL",
        0xC668: "HANGUL",
        0xC669: "HANGUL",
        0xC66A: "HANGUL",
        0xC66B: "HANGUL",
        0xC66C: "HANGUL",
        0xC66D: "HANGUL",
        0xC66E: "HANGUL",
        0xC66F: "HANGUL",
        0xC670: "HANGUL",
        0xC671: "HANGUL",
        0xC672: "HANGUL",
        0xC673: "HANGUL",
        0xC674: "HANGUL",
        0xC675: "HANGUL",
        0xC676: "HANGUL",
        0xC677: "HANGUL",
        0xC678: "HANGUL",
        0xC679: "HANGUL",
        0xC67A: "HANGUL",
        0xC67B: "HANGUL",
        0xC67C: "HANGUL",
        0xC67D: "HANGUL",
        0xC67E: "HANGUL",
        0xC67F: "HANGUL",
        0xC680: "HANGUL",
        0xC681: "HANGUL",
        0xC682: "HANGUL",
        0xC683: "HANGUL",
        0xC684: "HANGUL",
        0xC685: "HANGUL",
        0xC686: "HANGUL",
        0xC687: "HANGUL",
        0xC688: "HANGUL",
        0xC689: "HANGUL",
        0xC68A: "HANGUL",
        0xC68B: "HANGUL",
        0xC68C: "HANGUL",
        0xC68D: "HANGUL",
        0xC68E: "HANGUL",
        0xC68F: "HANGUL",
        0xC690: "HANGUL",
        0xC691: "HANGUL",
        0xC692: "HANGUL",
        0xC693: "HANGUL",
        0xC694: "HANGUL",
        0xC695: "HANGUL",
        0xC696: "HANGUL",
        0xC697: "HANGUL",
        0xC698: "HANGUL",
        0xC699: "HANGUL",
        0xC69A: "HANGUL",
        0xC69B: "HANGUL",
        0xC69C: "HANGUL",
        0xC69D: "HANGUL",
        0xC69E: "HANGUL",
        0xC69F: "HANGUL",
        0xC6A0: "HANGUL",
        0xC6A1: "HANGUL",
        0xC6A2: "HANGUL",
        0xC6A3: "HANGUL",
        0xC6A4: "HANGUL",
        0xC6A5: "HANGUL",
        0xC6A6: "HANGUL",
        0xC6A7: "HANGUL",
        0xC6A8: "HANGUL",
        0xC6A9: "HANGUL",
        0xC6AA: "HANGUL",
        0xC6AB: "HANGUL",
        0xC6AC: "HANGUL",
        0xC6AD: "HANGUL",
        0xC6AE: "HANGUL",
        0xC6AF: "HANGUL",
        0xC6B0: "HANGUL",
        0xC6B1: "HANGUL",
        0xC6B2: "HANGUL",
        0xC6B3: "HANGUL",
        0xC6B4: "HANGUL",
        0xC6B5: "HANGUL",
        0xC6B6: "HANGUL",
        0xC6B7: "HANGUL",
        0xC6B8: "HANGUL",
        0xC6B9: "HANGUL",
        0xC6BA: "HANGUL",
        0xC6BB: "HANGUL",
        0xC6BC: "HANGUL",
        0xC6BD: "HANGUL",
        0xC6BE: "HANGUL",
        0xC6BF: "HANGUL",
        0xC6C0: "HANGUL",
        0xC6C1: "HANGUL",
        0xC6C2: "HANGUL",
        0xC6C3: "HANGUL",
        0xC6C4: "HANGUL",
        0xC6C5: "HANGUL",
        0xC6C6: "HANGUL",
        0xC6C7: "HANGUL",
        0xC6C8: "HANGUL",
        0xC6C9: "HANGUL",
        0xC6CA: "HANGUL",
        0xC6CB: "HANGUL",
        0xC6CC: "HANGUL",
        0xC6CD: "HANGUL",
        0xC6CE: "HANGUL",
        0xC6CF: "HANGUL",
        0xC6D0: "HANGUL",
        0xC6D1: "HANGUL",
        0xC6D2: "HANGUL",
        0xC6D3: "HANGUL",
        0xC6D4: "HANGUL",
        0xC6D5: "HANGUL",
        0xC6D6: "HANGUL",
        0xC6D7: "HANGUL",
        0xC6D8: "HANGUL",
        0xC6D9: "HANGUL",
        0xC6DA: "HANGUL",
        0xC6DB: "HANGUL",
        0xC6DC: "HANGUL",
        0xC6DD: "HANGUL",
        0xC6DE: "HANGUL",
        0xC6DF: "HANGUL",
        0xC6E0: "HANGUL",
        0xC6E1: "HANGUL",
        0xC6E2: "HANGUL",
        0xC6E3: "HANGUL",
        0xC6E4: "HANGUL",
        0xC6E5: "HANGUL",
        0xC6E6: "HANGUL",
        0xC6E7: "HANGUL",
        0xC6E8: "HANGUL",
        0xC6E9: "HANGUL",
        0xC6EA: "HANGUL",
        0xC6EB: "HANGUL",
        0xC6EC: "HANGUL",
        0xC6ED: "HANGUL",
        0xC6EE: "HANGUL",
        0xC6EF: "HANGUL",
        0xC6F0: "HANGUL",
        0xC6F1: "HANGUL",
        0xC6F2: "HANGUL",
        0xC6F3: "HANGUL",
        0xC6F4: "HANGUL",
        0xC6F5: "HANGUL",
        0xC6F6: "HANGUL",
        0xC6F7: "HANGUL",
        0xC6F8: "HANGUL",
        0xC6F9: "HANGUL",
        0xC6FA: "HANGUL",
        0xC6FB: "HANGUL",
        0xC6FC: "HANGUL",
        0xC6FD: "HANGUL",
        0xC6FE: "HANGUL",
        0xC6FF: "HANGUL",
        0xC700: "HANGUL",
        0xC701: "HANGUL",
        0xC702: "HANGUL",
        0xC703: "HANGUL",
        0xC704: "HANGUL",
        0xC705: "HANGUL",
        0xC706: "HANGUL",
        0xC707: "HANGUL",
        0xC708: "HANGUL",
        0xC709: "HANGUL",
        0xC70A: "HANGUL",
        0xC70B: "HANGUL",
        0xC70C: "HANGUL",
        0xC70D: "HANGUL",
        0xC70E: "HANGUL",
        0xC70F: "HANGUL",
        0xC710: "HANGUL",
        0xC711: "HANGUL",
        0xC712: "HANGUL",
        0xC713: "HANGUL",
        0xC714: "HANGUL",
        0xC715: "HANGUL",
        0xC716: "HANGUL",
        0xC717: "HANGUL",
        0xC718: "HANGUL",
        0xC719: "HANGUL",
        0xC71A: "HANGUL",
        0xC71B: "HANGUL",
        0xC71C: "HANGUL",
        0xC71D: "HANGUL",
        0xC71E: "HANGUL",
        0xC71F: "HANGUL",
        0xC720: "HANGUL",
        0xC721: "HANGUL",
        0xC722: "HANGUL",
        0xC723: "HANGUL",
        0xC724: "HANGUL",
        0xC725: "HANGUL",
        0xC726: "HANGUL",
        0xC727: "HANGUL",
        0xC728: "HANGUL",
        0xC729: "HANGUL",
        0xC72A: "HANGUL",
        0xC72B: "HANGUL",
        0xC72C: "HANGUL",
        0xC72D: "HANGUL",
        0xC72E: "HANGUL",
        0xC72F: "HANGUL",
        0xC730: "HANGUL",
        0xC731: "HANGUL",
        0xC732: "HANGUL",
        0xC733: "HANGUL",
        0xC734: "HANGUL",
        0xC735: "HANGUL",
        0xC736: "HANGUL",
        0xC737: "HANGUL",
        0xC738: "HANGUL",
        0xC739: "HANGUL",
        0xC73A: "HANGUL",
        0xC73B: "HANGUL",
        0xC73C: "HANGUL",
        0xC73D: "HANGUL",
        0xC73E: "HANGUL",
        0xC73F: "HANGUL",
        0xC740: "HANGUL",
        0xC741: "HANGUL",
        0xC742: "HANGUL",
        0xC743: "HANGUL",
        0xC744: "HANGUL",
        0xC745: "HANGUL",
        0xC746: "HANGUL",
        0xC747: "HANGUL",
        0xC748: "HANGUL",
        0xC749: "HANGUL",
        0xC74A: "HANGUL",
        0xC74B: "HANGUL",
        0xC74C: "HANGUL",
        0xC74D: "HANGUL",
        0xC74E: "HANGUL",
        0xC74F: "HANGUL",
        0xC750: "HANGUL",
        0xC751: "HANGUL",
        0xC752: "HANGUL",
        0xC753: "HANGUL",
        0xC754: "HANGUL",
        0xC755: "HANGUL",
        0xC756: "HANGUL",
        0xC757: "HANGUL",
        0xC758: "HANGUL",
        0xC759: "HANGUL",
        0xC75A: "HANGUL",
        0xC75B: "HANGUL",
        0xC75C: "HANGUL",
        0xC75D: "HANGUL",
        0xC75E: "HANGUL",
        0xC75F: "HANGUL",
        0xC760: "HANGUL",
        0xC761: "HANGUL",
        0xC762: "HANGUL",
        0xC763: "HANGUL",
        0xC764: "HANGUL",
        0xC765: "HANGUL",
        0xC766: "HANGUL",
        0xC767: "HANGUL",
        0xC768: "HANGUL",
        0xC769: "HANGUL",
        0xC76A: "HANGUL",
        0xC76B: "HANGUL",
        0xC76C: "HANGUL",
        0xC76D: "HANGUL",
        0xC76E: "HANGUL",
        0xC76F: "HANGUL",
        0xC770: "HANGUL",
        0xC771: "HANGUL",
        0xC772: "HANGUL",
        0xC773: "HANGUL",
        0xC774: "HANGUL",
        0xC775: "HANGUL",
        0xC776: "HANGUL",
        0xC777: "HANGUL",
        0xC778: "HANGUL",
        0xC779: "HANGUL",
        0xC77A: "HANGUL",
        0xC77B: "HANGUL",
        0xC77C: "HANGUL",
        0xC77D: "HANGUL",
        0xC77E: "HANGUL",
        0xC77F: "HANGUL",
        0xC780: "HANGUL",
        0xC781: "HANGUL",
        0xC782: "HANGUL",
        0xC783: "HANGUL",
        0xC784: "HANGUL",
        0xC785: "HANGUL",
        0xC786: "HANGUL",
        0xC787: "HANGUL",
        0xC788: "HANGUL",
        0xC789: "HANGUL",
        0xC78A: "HANGUL",
        0xC78B: "HANGUL",
        0xC78C: "HANGUL",
        0xC78D: "HANGUL",
        0xC78E: "HANGUL",
        0xC78F: "HANGUL",
        0xC790: "HANGUL",
        0xC791: "HANGUL",
        0xC792: "HANGUL",
        0xC793: "HANGUL",
        0xC794: "HANGUL",
        0xC795: "HANGUL",
        0xC796: "HANGUL",
        0xC797: "HANGUL",
        0xC798: "HANGUL",
        0xC799: "HANGUL",
        0xC79A: "HANGUL",
        0xC79B: "HANGUL",
        0xC79C: "HANGUL",
        0xC79D: "HANGUL",
        0xC79E: "HANGUL",
        0xC79F: "HANGUL",
        0xC7A0: "HANGUL",
        0xC7A1: "HANGUL",
        0xC7A2: "HANGUL",
        0xC7A3: "HANGUL",
        0xC7A4: "HANGUL",
        0xC7A5: "HANGUL",
        0xC7A6: "HANGUL",
        0xC7A7: "HANGUL",
        0xC7A8: "HANGUL",
        0xC7A9: "HANGUL",
        0xC7AA: "HANGUL",
        0xC7AB: "HANGUL",
        0xC7AC: "HANGUL",
        0xC7AD: "HANGUL",
        0xC7AE: "HANGUL",
        0xC7AF: "HANGUL",
        0xC7B0: "HANGUL",
        0xC7B1: "HANGUL",
        0xC7B2: "HANGUL",
        0xC7B3: "HANGUL",
        0xC7B4: "HANGUL",
        0xC7B5: "HANGUL",
        0xC7B6: "HANGUL",
        0xC7B7: "HANGUL",
        0xC7B8: "HANGUL",
        0xC7B9: "HANGUL",
        0xC7BA: "HANGUL",
        0xC7BB: "HANGUL",
        0xC7BC: "HANGUL",
        0xC7BD: "HANGUL",
        0xC7BE: "HANGUL",
        0xC7BF: "HANGUL",
        0xC7C0: "HANGUL",
        0xC7C1: "HANGUL",
        0xC7C2: "HANGUL",
        0xC7C3: "HANGUL",
        0xC7C4: "HANGUL",
        0xC7C5: "HANGUL",
        0xC7C6: "HANGUL",
        0xC7C7: "HANGUL",
        0xC7C8: "HANGUL",
        0xC7C9: "HANGUL",
        0xC7CA: "HANGUL",
        0xC7CB: "HANGUL",
        0xC7CC: "HANGUL",
        0xC7CD: "HANGUL",
        0xC7CE: "HANGUL",
        0xC7CF: "HANGUL",
        0xC7D0: "HANGUL",
        0xC7D1: "HANGUL",
        0xC7D2: "HANGUL",
        0xC7D3: "HANGUL",
        0xC7D4: "HANGUL",
        0xC7D5: "HANGUL",
        0xC7D6: "HANGUL",
        0xC7D7: "HANGUL",
        0xC7D8: "HANGUL",
        0xC7D9: "HANGUL",
        0xC7DA: "HANGUL",
        0xC7DB: "HANGUL",
        0xC7DC: "HANGUL",
        0xC7DD: "HANGUL",
        0xC7DE: "HANGUL",
        0xC7DF: "HANGUL",
        0xC7E0: "HANGUL",
        0xC7E1: "HANGUL",
        0xC7E2: "HANGUL",
        0xC7E3: "HANGUL",
        0xC7E4: "HANGUL",
        0xC7E5: "HANGUL",
        0xC7E6: "HANGUL",
        0xC7E7: "HANGUL",
        0xC7E8: "HANGUL",
        0xC7E9: "HANGUL",
        0xC7EA: "HANGUL",
        0xC7EB: "HANGUL",
        0xC7EC: "HANGUL",
        0xC7ED: "HANGUL",
        0xC7EE: "HANGUL",
        0xC7EF: "HANGUL",
        0xC7F0: "HANGUL",
        0xC7F1: "HANGUL",
        0xC7F2: "HANGUL",
        0xC7F3: "HANGUL",
        0xC7F4: "HANGUL",
        0xC7F5: "HANGUL",
        0xC7F6: "HANGUL",
        0xC7F7: "HANGUL",
        0xC7F8: "HANGUL",
        0xC7F9: "HANGUL",
        0xC7FA: "HANGUL",
        0xC7FB: "HANGUL",
        0xC7FC: "HANGUL",
        0xC7FD: "HANGUL",
        0xC7FE: "HANGUL",
        0xC7FF: "HANGUL",
        0xC800: "HANGUL",
        0xC801: "HANGUL",
        0xC802: "HANGUL",
        0xC803: "HANGUL",
        0xC804: "HANGUL",
        0xC805: "HANGUL",
        0xC806: "HANGUL",
        0xC807: "HANGUL",
        0xC808: "HANGUL",
        0xC809: "HANGUL",
        0xC80A: "HANGUL",
        0xC80B: "HANGUL",
        0xC80C: "HANGUL",
        0xC80D: "HANGUL",
        0xC80E: "HANGUL",
        0xC80F: "HANGUL",
        0xC810: "HANGUL",
        0xC811: "HANGUL",
        0xC812: "HANGUL",
        0xC813: "HANGUL",
        0xC814: "HANGUL",
        0xC815: "HANGUL",
        0xC816: "HANGUL",
        0xC817: "HANGUL",
        0xC818: "HANGUL",
        0xC819: "HANGUL",
        0xC81A: "HANGUL",
        0xC81B: "HANGUL",
        0xC81C: "HANGUL",
        0xC81D: "HANGUL",
        0xC81E: "HANGUL",
        0xC81F: "HANGUL",
        0xC820: "HANGUL",
        0xC821: "HANGUL",
        0xC822: "HANGUL",
        0xC823: "HANGUL",
        0xC824: "HANGUL",
        0xC825: "HANGUL",
        0xC826: "HANGUL",
        0xC827: "HANGUL",
        0xC828: "HANGUL",
        0xC829: "HANGUL",
        0xC82A: "HANGUL",
        0xC82B: "HANGUL",
        0xC82C: "HANGUL",
        0xC82D: "HANGUL",
        0xC82E: "HANGUL",
        0xC82F: "HANGUL",
        0xC830: "HANGUL",
        0xC831: "HANGUL",
        0xC832: "HANGUL",
        0xC833: "HANGUL",
        0xC834: "HANGUL",
        0xC835: "HANGUL",
        0xC836: "HANGUL",
        0xC837: "HANGUL",
        0xC838: "HANGUL",
        0xC839: "HANGUL",
        0xC83A: "HANGUL",
        0xC83B: "HANGUL",
        0xC83C: "HANGUL",
        0xC83D: "HANGUL",
        0xC83E: "HANGUL",
        0xC83F: "HANGUL",
        0xC840: "HANGUL",
        0xC841: "HANGUL",
        0xC842: "HANGUL",
        0xC843: "HANGUL",
        0xC844: "HANGUL",
        0xC845: "HANGUL",
        0xC846: "HANGUL",
        0xC847: "HANGUL",
        0xC848: "HANGUL",
        0xC849: "HANGUL",
        0xC84A: "HANGUL",
        0xC84B: "HANGUL",
        0xC84C: "HANGUL",
        0xC84D: "HANGUL",
        0xC84E: "HANGUL",
        0xC84F: "HANGUL",
        0xC850: "HANGUL",
        0xC851: "HANGUL",
        0xC852: "HANGUL",
        0xC853: "HANGUL",
        0xC854: "HANGUL",
        0xC855: "HANGUL",
        0xC856: "HANGUL",
        0xC857: "HANGUL",
        0xC858: "HANGUL",
        0xC859: "HANGUL",
        0xC85A: "HANGUL",
        0xC85B: "HANGUL",
        0xC85C: "HANGUL",
        0xC85D: "HANGUL",
        0xC85E: "HANGUL",
        0xC85F: "HANGUL",
        0xC860: "HANGUL",
        0xC861: "HANGUL",
        0xC862: "HANGUL",
        0xC863: "HANGUL",
        0xC864: "HANGUL",
        0xC865: "HANGUL",
        0xC866: "HANGUL",
        0xC867: "HANGUL",
        0xC868: "HANGUL",
        0xC869: "HANGUL",
        0xC86A: "HANGUL",
        0xC86B: "HANGUL",
        0xC86C: "HANGUL",
        0xC86D: "HANGUL",
        0xC86E: "HANGUL",
        0xC86F: "HANGUL",
        0xC870: "HANGUL",
        0xC871: "HANGUL",
        0xC872: "HANGUL",
        0xC873: "HANGUL",
        0xC874: "HANGUL",
        0xC875: "HANGUL",
        0xC876: "HANGUL",
        0xC877: "HANGUL",
        0xC878: "HANGUL",
        0xC879: "HANGUL",
        0xC87A: "HANGUL",
        0xC87B: "HANGUL",
        0xC87C: "HANGUL",
        0xC87D: "HANGUL",
        0xC87E: "HANGUL",
        0xC87F: "HANGUL",
        0xC880: "HANGUL",
        0xC881: "HANGUL",
        0xC882: "HANGUL",
        0xC883: "HANGUL",
        0xC884: "HANGUL",
        0xC885: "HANGUL",
        0xC886: "HANGUL",
        0xC887: "HANGUL",
        0xC888: "HANGUL",
        0xC889: "HANGUL",
        0xC88A: "HANGUL",
        0xC88B: "HANGUL",
        0xC88C: "HANGUL",
        0xC88D: "HANGUL",
        0xC88E: "HANGUL",
        0xC88F: "HANGUL",
        0xC890: "HANGUL",
        0xC891: "HANGUL",
        0xC892: "HANGUL",
        0xC893: "HANGUL",
        0xC894: "HANGUL",
        0xC895: "HANGUL",
        0xC896: "HANGUL",
        0xC897: "HANGUL",
        0xC898: "HANGUL",
        0xC899: "HANGUL",
        0xC89A: "HANGUL",
        0xC89B: "HANGUL",
        0xC89C: "HANGUL",
        0xC89D: "HANGUL",
        0xC89E: "HANGUL",
        0xC89F: "HANGUL",
        0xC8A0: "HANGUL",
        0xC8A1: "HANGUL",
        0xC8A2: "HANGUL",
        0xC8A3: "HANGUL",
        0xC8A4: "HANGUL",
        0xC8A5: "HANGUL",
        0xC8A6: "HANGUL",
        0xC8A7: "HANGUL",
        0xC8A8: "HANGUL",
        0xC8A9: "HANGUL",
        0xC8AA: "HANGUL",
        0xC8AB: "HANGUL",
        0xC8AC: "HANGUL",
        0xC8AD: "HANGUL",
        0xC8AE: "HANGUL",
        0xC8AF: "HANGUL",
        0xC8B0: "HANGUL",
        0xC8B1: "HANGUL",
        0xC8B2: "HANGUL",
        0xC8B3: "HANGUL",
        0xC8B4: "HANGUL",
        0xC8B5: "HANGUL",
        0xC8B6: "HANGUL",
        0xC8B7: "HANGUL",
        0xC8B8: "HANGUL",
        0xC8B9: "HANGUL",
        0xC8BA: "HANGUL",
        0xC8BB: "HANGUL",
        0xC8BC: "HANGUL",
        0xC8BD: "HANGUL",
        0xC8BE: "HANGUL",
        0xC8BF: "HANGUL",
        0xC8C0: "HANGUL",
        0xC8C1: "HANGUL",
        0xC8C2: "HANGUL",
        0xC8C3: "HANGUL",
        0xC8C4: "HANGUL",
        0xC8C5: "HANGUL",
        0xC8C6: "HANGUL",
        0xC8C7: "HANGUL",
        0xC8C8: "HANGUL",
        0xC8C9: "HANGUL",
        0xC8CA: "HANGUL",
        0xC8CB: "HANGUL",
        0xC8CC: "HANGUL",
        0xC8CD: "HANGUL",
        0xC8CE: "HANGUL",
        0xC8CF: "HANGUL",
        0xC8D0: "HANGUL",
        0xC8D1: "HANGUL",
        0xC8D2: "HANGUL",
        0xC8D3: "HANGUL",
        0xC8D4: "HANGUL",
        0xC8D5: "HANGUL",
        0xC8D6: "HANGUL",
        0xC8D7: "HANGUL",
        0xC8D8: "HANGUL",
        0xC8D9: "HANGUL",
        0xC8DA: "HANGUL",
        0xC8DB: "HANGUL",
        0xC8DC: "HANGUL",
        0xC8DD: "HANGUL",
        0xC8DE: "HANGUL",
        0xC8DF: "HANGUL",
        0xC8E0: "HANGUL",
        0xC8E1: "HANGUL",
        0xC8E2: "HANGUL",
        0xC8E3: "HANGUL",
        0xC8E4: "HANGUL",
        0xC8E5: "HANGUL",
        0xC8E6: "HANGUL",
        0xC8E7: "HANGUL",
        0xC8E8: "HANGUL",
        0xC8E9: "HANGUL",
        0xC8EA: "HANGUL",
        0xC8EB: "HANGUL",
        0xC8EC: "HANGUL",
        0xC8ED: "HANGUL",
        0xC8EE: "HANGUL",
        0xC8EF: "HANGUL",
        0xC8F0: "HANGUL",
        0xC8F1: "HANGUL",
        0xC8F2: "HANGUL",
        0xC8F3: "HANGUL",
        0xC8F4: "HANGUL",
        0xC8F5: "HANGUL",
        0xC8F6: "HANGUL",
        0xC8F7: "HANGUL",
        0xC8F8: "HANGUL",
        0xC8F9: "HANGUL",
        0xC8FA: "HANGUL",
        0xC8FB: "HANGUL",
        0xC8FC: "HANGUL",
        0xC8FD: "HANGUL",
        0xC8FE: "HANGUL",
        0xC8FF: "HANGUL",
        0xC900: "HANGUL",
        0xC901: "HANGUL",
        0xC902: "HANGUL",
        0xC903: "HANGUL",
        0xC904: "HANGUL",
        0xC905: "HANGUL",
        0xC906: "HANGUL",
        0xC907: "HANGUL",
        0xC908: "HANGUL",
        0xC909: "HANGUL",
        0xC90A: "HANGUL",
        0xC90B: "HANGUL",
        0xC90C: "HANGUL",
        0xC90D: "HANGUL",
        0xC90E: "HANGUL",
        0xC90F: "HANGUL",
        0xC910: "HANGUL",
        0xC911: "HANGUL",
        0xC912: "HANGUL",
        0xC913: "HANGUL",
        0xC914: "HANGUL",
        0xC915: "HANGUL",
        0xC916: "HANGUL",
        0xC917: "HANGUL",
        0xC918: "HANGUL",
        0xC919: "HANGUL",
        0xC91A: "HANGUL",
        0xC91B: "HANGUL",
        0xC91C: "HANGUL",
        0xC91D: "HANGUL",
        0xC91E: "HANGUL",
        0xC91F: "HANGUL",
        0xC920: "HANGUL",
        0xC921: "HANGUL",
        0xC922: "HANGUL",
        0xC923: "HANGUL",
        0xC924: "HANGUL",
        0xC925: "HANGUL",
        0xC926: "HANGUL",
        0xC927: "HANGUL",
        0xC928: "HANGUL",
        0xC929: "HANGUL",
        0xC92A: "HANGUL",
        0xC92B: "HANGUL",
        0xC92C: "HANGUL",
        0xC92D: "HANGUL",
        0xC92E: "HANGUL",
        0xC92F: "HANGUL",
        0xC930: "HANGUL",
        0xC931: "HANGUL",
        0xC932: "HANGUL",
        0xC933: "HANGUL",
        0xC934: "HANGUL",
        0xC935: "HANGUL",
        0xC936: "HANGUL",
        0xC937: "HANGUL",
        0xC938: "HANGUL",
        0xC939: "HANGUL",
        0xC93A: "HANGUL",
        0xC93B: "HANGUL",
        0xC93C: "HANGUL",
        0xC93D: "HANGUL",
        0xC93E: "HANGUL",
        0xC93F: "HANGUL",
        0xC940: "HANGUL",
        0xC941: "HANGUL",
        0xC942: "HANGUL",
        0xC943: "HANGUL",
        0xC944: "HANGUL",
        0xC945: "HANGUL",
        0xC946: "HANGUL",
        0xC947: "HANGUL",
        0xC948: "HANGUL",
        0xC949: "HANGUL",
        0xC94A: "HANGUL",
        0xC94B: "HANGUL",
        0xC94C: "HANGUL",
        0xC94D: "HANGUL",
        0xC94E: "HANGUL",
        0xC94F: "HANGUL",
        0xC950: "HANGUL",
        0xC951: "HANGUL",
        0xC952: "HANGUL",
        0xC953: "HANGUL",
        0xC954: "HANGUL",
        0xC955: "HANGUL",
        0xC956: "HANGUL",
        0xC957: "HANGUL",
        0xC958: "HANGUL",
        0xC959: "HANGUL",
        0xC95A: "HANGUL",
        0xC95B: "HANGUL",
        0xC95C: "HANGUL",
        0xC95D: "HANGUL",
        0xC95E: "HANGUL",
        0xC95F: "HANGUL",
        0xC960: "HANGUL",
        0xC961: "HANGUL",
        0xC962: "HANGUL",
        0xC963: "HANGUL",
        0xC964: "HANGUL",
        0xC965: "HANGUL",
        0xC966: "HANGUL",
        0xC967: "HANGUL",
        0xC968: "HANGUL",
        0xC969: "HANGUL",
        0xC96A: "HANGUL",
        0xC96B: "HANGUL",
        0xC96C: "HANGUL",
        0xC96D: "HANGUL",
        0xC96E: "HANGUL",
        0xC96F: "HANGUL",
        0xC970: "HANGUL",
        0xC971: "HANGUL",
        0xC972: "HANGUL",
        0xC973: "HANGUL",
        0xC974: "HANGUL",
        0xC975: "HANGUL",
        0xC976: "HANGUL",
        0xC977: "HANGUL",
        0xC978: "HANGUL",
        0xC979: "HANGUL",
        0xC97A: "HANGUL",
        0xC97B: "HANGUL",
        0xC97C: "HANGUL",
        0xC97D: "HANGUL",
        0xC97E: "HANGUL",
        0xC97F: "HANGUL",
        0xC980: "HANGUL",
        0xC981: "HANGUL",
        0xC982: "HANGUL",
        0xC983: "HANGUL",
        0xC984: "HANGUL",
        0xC985: "HANGUL",
        0xC986: "HANGUL",
        0xC987: "HANGUL",
        0xC988: "HANGUL",
        0xC989: "HANGUL",
        0xC98A: "HANGUL",
        0xC98B: "HANGUL",
        0xC98C: "HANGUL",
        0xC98D: "HANGUL",
        0xC98E: "HANGUL",
        0xC98F: "HANGUL",
        0xC990: "HANGUL",
        0xC991: "HANGUL",
        0xC992: "HANGUL",
        0xC993: "HANGUL",
        0xC994: "HANGUL",
        0xC995: "HANGUL",
        0xC996: "HANGUL",
        0xC997: "HANGUL",
        0xC998: "HANGUL",
        0xC999: "HANGUL",
        0xC99A: "HANGUL",
        0xC99B: "HANGUL",
        0xC99C: "HANGUL",
        0xC99D: "HANGUL",
        0xC99E: "HANGUL",
        0xC99F: "HANGUL",
        0xC9A0: "HANGUL",
        0xC9A1: "HANGUL",
        0xC9A2: "HANGUL",
        0xC9A3: "HANGUL",
        0xC9A4: "HANGUL",
        0xC9A5: "HANGUL",
        0xC9A6: "HANGUL",
        0xC9A7: "HANGUL",
        0xC9A8: "HANGUL",
        0xC9A9: "HANGUL",
        0xC9AA: "HANGUL",
        0xC9AB: "HANGUL",
        0xC9AC: "HANGUL",
        0xC9AD: "HANGUL",
        0xC9AE: "HANGUL",
        0xC9AF: "HANGUL",
        0xC9B0: "HANGUL",
        0xC9B1: "HANGUL",
        0xC9B2: "HANGUL",
        0xC9B3: "HANGUL",
        0xC9B4: "HANGUL",
        0xC9B5: "HANGUL",
        0xC9B6: "HANGUL",
        0xC9B7: "HANGUL",
        0xC9B8: "HANGUL",
        0xC9B9: "HANGUL",
        0xC9BA: "HANGUL",
        0xC9BB: "HANGUL",
        0xC9BC: "HANGUL",
        0xC9BD: "HANGUL",
        0xC9BE: "HANGUL",
        0xC9BF: "HANGUL",
        0xC9C0: "HANGUL",
        0xC9C1: "HANGUL",
        0xC9C2: "HANGUL",
        0xC9C3: "HANGUL",
        0xC9C4: "HANGUL",
        0xC9C5: "HANGUL",
        0xC9C6: "HANGUL",
        0xC9C7: "HANGUL",
        0xC9C8: "HANGUL",
        0xC9C9: "HANGUL",
        0xC9CA: "HANGUL",
        0xC9CB: "HANGUL",
        0xC9CC: "HANGUL",
        0xC9CD: "HANGUL",
        0xC9CE: "HANGUL",
        0xC9CF: "HANGUL",
        0xC9D0: "HANGUL",
        0xC9D1: "HANGUL",
        0xC9D2: "HANGUL",
        0xC9D3: "HANGUL",
        0xC9D4: "HANGUL",
        0xC9D5: "HANGUL",
        0xC9D6: "HANGUL",
        0xC9D7: "HANGUL",
        0xC9D8: "HANGUL",
        0xC9D9: "HANGUL",
        0xC9DA: "HANGUL",
        0xC9DB: "HANGUL",
        0xC9DC: "HANGUL",
        0xC9DD: "HANGUL",
        0xC9DE: "HANGUL",
        0xC9DF: "HANGUL",
        0xC9E0: "HANGUL",
        0xC9E1: "HANGUL",
        0xC9E2: "HANGUL",
        0xC9E3: "HANGUL",
        0xC9E4: "HANGUL",
        0xC9E5: "HANGUL",
        0xC9E6: "HANGUL",
        0xC9E7: "HANGUL",
        0xC9E8: "HANGUL",
        0xC9E9: "HANGUL",
        0xC9EA: "HANGUL",
        0xC9EB: "HANGUL",
        0xC9EC: "HANGUL",
        0xC9ED: "HANGUL",
        0xC9EE: "HANGUL",
        0xC9EF: "HANGUL",
        0xC9F0: "HANGUL",
        0xC9F1: "HANGUL",
        0xC9F2: "HANGUL",
        0xC9F3: "HANGUL",
        0xC9F4: "HANGUL",
        0xC9F5: "HANGUL",
        0xC9F6: "HANGUL",
        0xC9F7: "HANGUL",
        0xC9F8: "HANGUL",
        0xC9F9: "HANGUL",
        0xC9FA: "HANGUL",
        0xC9FB: "HANGUL",
        0xC9FC: "HANGUL",
        0xC9FD: "HANGUL",
        0xC9FE: "HANGUL",
        0xC9FF: "HANGUL",
        0xCA00: "HANGUL",
        0xCA01: "HANGUL",
        0xCA02: "HANGUL",
        0xCA03: "HANGUL",
        0xCA04: "HANGUL",
        0xCA05: "HANGUL",
        0xCA06: "HANGUL",
        0xCA07: "HANGUL",
        0xCA08: "HANGUL",
        0xCA09: "HANGUL",
        0xCA0A: "HANGUL",
        0xCA0B: "HANGUL",
        0xCA0C: "HANGUL",
        0xCA0D: "HANGUL",
        0xCA0E: "HANGUL",
        0xCA0F: "HANGUL",
        0xCA10: "HANGUL",
        0xCA11: "HANGUL",
        0xCA12: "HANGUL",
        0xCA13: "HANGUL",
        0xCA14: "HANGUL",
        0xCA15: "HANGUL",
        0xCA16: "HANGUL",
        0xCA17: "HANGUL",
        0xCA18: "HANGUL",
        0xCA19: "HANGUL",
        0xCA1A: "HANGUL",
        0xCA1B: "HANGUL",
        0xCA1C: "HANGUL",
        0xCA1D: "HANGUL",
        0xCA1E: "HANGUL",
        0xCA1F: "HANGUL",
        0xCA20: "HANGUL",
        0xCA21: "HANGUL",
        0xCA22: "HANGUL",
        0xCA23: "HANGUL",
        0xCA24: "HANGUL",
        0xCA25: "HANGUL",
        0xCA26: "HANGUL",
        0xCA27: "HANGUL",
        0xCA28: "HANGUL",
        0xCA29: "HANGUL",
        0xCA2A: "HANGUL",
        0xCA2B: "HANGUL",
        0xCA2C: "HANGUL",
        0xCA2D: "HANGUL",
        0xCA2E: "HANGUL",
        0xCA2F: "HANGUL",
        0xCA30: "HANGUL",
        0xCA31: "HANGUL",
        0xCA32: "HANGUL",
        0xCA33: "HANGUL",
        0xCA34: "HANGUL",
        0xCA35: "HANGUL",
        0xCA36: "HANGUL",
        0xCA37: "HANGUL",
        0xCA38: "HANGUL",
        0xCA39: "HANGUL",
        0xCA3A: "HANGUL",
        0xCA3B: "HANGUL",
        0xCA3C: "HANGUL",
        0xCA3D: "HANGUL",
        0xCA3E: "HANGUL",
        0xCA3F: "HANGUL",
        0xCA40: "HANGUL",
        0xCA41: "HANGUL",
        0xCA42: "HANGUL",
        0xCA43: "HANGUL",
        0xCA44: "HANGUL",
        0xCA45: "HANGUL",
        0xCA46: "HANGUL",
        0xCA47: "HANGUL",
        0xCA48: "HANGUL",
        0xCA49: "HANGUL",
        0xCA4A: "HANGUL",
        0xCA4B: "HANGUL",
        0xCA4C: "HANGUL",
        0xCA4D: "HANGUL",
        0xCA4E: "HANGUL",
        0xCA4F: "HANGUL",
        0xCA50: "HANGUL",
        0xCA51: "HANGUL",
        0xCA52: "HANGUL",
        0xCA53: "HANGUL",
        0xCA54: "HANGUL",
        0xCA55: "HANGUL",
        0xCA56: "HANGUL",
        0xCA57: "HANGUL",
        0xCA58: "HANGUL",
        0xCA59: "HANGUL",
        0xCA5A: "HANGUL",
        0xCA5B: "HANGUL",
        0xCA5C: "HANGUL",
        0xCA5D: "HANGUL",
        0xCA5E: "HANGUL",
        0xCA5F: "HANGUL",
        0xCA60: "HANGUL",
        0xCA61: "HANGUL",
        0xCA62: "HANGUL",
        0xCA63: "HANGUL",
        0xCA64: "HANGUL",
        0xCA65: "HANGUL",
        0xCA66: "HANGUL",
        0xCA67: "HANGUL",
        0xCA68: "HANGUL",
        0xCA69: "HANGUL",
        0xCA6A: "HANGUL",
        0xCA6B: "HANGUL",
        0xCA6C: "HANGUL",
        0xCA6D: "HANGUL",
        0xCA6E: "HANGUL",
        0xCA6F: "HANGUL",
        0xCA70: "HANGUL",
        0xCA71: "HANGUL",
        0xCA72: "HANGUL",
        0xCA73: "HANGUL",
        0xCA74: "HANGUL",
        0xCA75: "HANGUL",
        0xCA76: "HANGUL",
        0xCA77: "HANGUL",
        0xCA78: "HANGUL",
        0xCA79: "HANGUL",
        0xCA7A: "HANGUL",
        0xCA7B: "HANGUL",
        0xCA7C: "HANGUL",
        0xCA7D: "HANGUL",
        0xCA7E: "HANGUL",
        0xCA7F: "HANGUL",
        0xCA80: "HANGUL",
        0xCA81: "HANGUL",
        0xCA82: "HANGUL",
        0xCA83: "HANGUL",
        0xCA84: "HANGUL",
        0xCA85: "HANGUL",
        0xCA86: "HANGUL",
        0xCA87: "HANGUL",
        0xCA88: "HANGUL",
        0xCA89: "HANGUL",
        0xCA8A: "HANGUL",
        0xCA8B: "HANGUL",
        0xCA8C: "HANGUL",
        0xCA8D: "HANGUL",
        0xCA8E: "HANGUL",
        0xCA8F: "HANGUL",
        0xCA90: "HANGUL",
        0xCA91: "HANGUL",
        0xCA92: "HANGUL",
        0xCA93: "HANGUL",
        0xCA94: "HANGUL",
        0xCA95: "HANGUL",
        0xCA96: "HANGUL",
        0xCA97: "HANGUL",
        0xCA98: "HANGUL",
        0xCA99: "HANGUL",
        0xCA9A: "HANGUL",
        0xCA9B: "HANGUL",
        0xCA9C: "HANGUL",
        0xCA9D: "HANGUL",
        0xCA9E: "HANGUL",
        0xCA9F: "HANGUL",
        0xCAA0: "HANGUL",
        0xCAA1: "HANGUL",
        0xCAA2: "HANGUL",
        0xCAA3: "HANGUL",
        0xCAA4: "HANGUL",
        0xCAA5: "HANGUL",
        0xCAA6: "HANGUL",
        0xCAA7: "HANGUL",
        0xCAA8: "HANGUL",
        0xCAA9: "HANGUL",
        0xCAAA: "HANGUL",
        0xCAAB: "HANGUL",
        0xCAAC: "HANGUL",
        0xCAAD: "HANGUL",
        0xCAAE: "HANGUL",
        0xCAAF: "HANGUL",
        0xCAB0: "HANGUL",
        0xCAB1: "HANGUL",
        0xCAB2: "HANGUL",
        0xCAB3: "HANGUL",
        0xCAB4: "HANGUL",
        0xCAB5: "HANGUL",
        0xCAB6: "HANGUL",
        0xCAB7: "HANGUL",
        0xCAB8: "HANGUL",
        0xCAB9: "HANGUL",
        0xCABA: "HANGUL",
        0xCABB: "HANGUL",
        0xCABC: "HANGUL",
        0xCABD: "HANGUL",
        0xCABE: "HANGUL",
        0xCABF: "HANGUL",
        0xCAC0: "HANGUL",
        0xCAC1: "HANGUL",
        0xCAC2: "HANGUL",
        0xCAC3: "HANGUL",
        0xCAC4: "HANGUL",
        0xCAC5: "HANGUL",
        0xCAC6: "HANGUL",
        0xCAC7: "HANGUL",
        0xCAC8: "HANGUL",
        0xCAC9: "HANGUL",
        0xCACA: "HANGUL",
        0xCACB: "HANGUL",
        0xCACC: "HANGUL",
        0xCACD: "HANGUL",
        0xCACE: "HANGUL",
        0xCACF: "HANGUL",
        0xCAD0: "HANGUL",
        0xCAD1: "HANGUL",
        0xCAD2: "HANGUL",
        0xCAD3: "HANGUL",
        0xCAD4: "HANGUL",
        0xCAD5: "HANGUL",
        0xCAD6: "HANGUL",
        0xCAD7: "HANGUL",
        0xCAD8: "HANGUL",
        0xCAD9: "HANGUL",
        0xCADA: "HANGUL",
        0xCADB: "HANGUL",
        0xCADC: "HANGUL",
        0xCADD: "HANGUL",
        0xCADE: "HANGUL",
        0xCADF: "HANGUL",
        0xCAE0: "HANGUL",
        0xCAE1: "HANGUL",
        0xCAE2: "HANGUL",
        0xCAE3: "HANGUL",
        0xCAE4: "HANGUL",
        0xCAE5: "HANGUL",
        0xCAE6: "HANGUL",
        0xCAE7: "HANGUL",
        0xCAE8: "HANGUL",
        0xCAE9: "HANGUL",
        0xCAEA: "HANGUL",
        0xCAEB: "HANGUL",
        0xCAEC: "HANGUL",
        0xCAED: "HANGUL",
        0xCAEE: "HANGUL",
        0xCAEF: "HANGUL",
        0xCAF0: "HANGUL",
        0xCAF1: "HANGUL",
        0xCAF2: "HANGUL",
        0xCAF3: "HANGUL",
        0xCAF4: "HANGUL",
        0xCAF5: "HANGUL",
        0xCAF6: "HANGUL",
        0xCAF7: "HANGUL",
        0xCAF8: "HANGUL",
        0xCAF9: "HANGUL",
        0xCAFA: "HANGUL",
        0xCAFB: "HANGUL",
        0xCAFC: "HANGUL",
        0xCAFD: "HANGUL",
        0xCAFE: "HANGUL",
        0xCAFF: "HANGUL",
        0xCB00: "HANGUL",
        0xCB01: "HANGUL",
        0xCB02: "HANGUL",
        0xCB03: "HANGUL",
        0xCB04: "HANGUL",
        0xCB05: "HANGUL",
        0xCB06: "HANGUL",
        0xCB07: "HANGUL",
        0xCB08: "HANGUL",
        0xCB09: "HANGUL",
        0xCB0A: "HANGUL",
        0xCB0B: "HANGUL",
        0xCB0C: "HANGUL",
        0xCB0D: "HANGUL",
        0xCB0E: "HANGUL",
        0xCB0F: "HANGUL",
        0xCB10: "HANGUL",
        0xCB11: "HANGUL",
        0xCB12: "HANGUL",
        0xCB13: "HANGUL",
        0xCB14: "HANGUL",
        0xCB15: "HANGUL",
        0xCB16: "HANGUL",
        0xCB17: "HANGUL",
        0xCB18: "HANGUL",
        0xCB19: "HANGUL",
        0xCB1A: "HANGUL",
        0xCB1B: "HANGUL",
        0xCB1C: "HANGUL",
        0xCB1D: "HANGUL",
        0xCB1E: "HANGUL",
        0xCB1F: "HANGUL",
        0xCB20: "HANGUL",
        0xCB21: "HANGUL",
        0xCB22: "HANGUL",
        0xCB23: "HANGUL",
        0xCB24: "HANGUL",
        0xCB25: "HANGUL",
        0xCB26: "HANGUL",
        0xCB27: "HANGUL",
        0xCB28: "HANGUL",
        0xCB29: "HANGUL",
        0xCB2A: "HANGUL",
        0xCB2B: "HANGUL",
        0xCB2C: "HANGUL",
        0xCB2D: "HANGUL",
        0xCB2E: "HANGUL",
        0xCB2F: "HANGUL",
        0xCB30: "HANGUL",
        0xCB31: "HANGUL",
        0xCB32: "HANGUL",
        0xCB33: "HANGUL",
        0xCB34: "HANGUL",
        0xCB35: "HANGUL",
        0xCB36: "HANGUL",
        0xCB37: "HANGUL",
        0xCB38: "HANGUL",
        0xCB39: "HANGUL",
        0xCB3A: "HANGUL",
        0xCB3B: "HANGUL",
        0xCB3C: "HANGUL",
        0xCB3D: "HANGUL",
        0xCB3E: "HANGUL",
        0xCB3F: "HANGUL",
        0xCB40: "HANGUL",
        0xCB41: "HANGUL",
        0xCB42: "HANGUL",
        0xCB43: "HANGUL",
        0xCB44: "HANGUL",
        0xCB45: "HANGUL",
        0xCB46: "HANGUL",
        0xCB47: "HANGUL",
        0xCB48: "HANGUL",
        0xCB49: "HANGUL",
        0xCB4A: "HANGUL",
        0xCB4B: "HANGUL",
        0xCB4C: "HANGUL",
        0xCB4D: "HANGUL",
        0xCB4E: "HANGUL",
        0xCB4F: "HANGUL",
        0xCB50: "HANGUL",
        0xCB51: "HANGUL",
        0xCB52: "HANGUL",
        0xCB53: "HANGUL",
        0xCB54: "HANGUL",
        0xCB55: "HANGUL",
        0xCB56: "HANGUL",
        0xCB57: "HANGUL",
        0xCB58: "HANGUL",
        0xCB59: "HANGUL",
        0xCB5A: "HANGUL",
        0xCB5B: "HANGUL",
        0xCB5C: "HANGUL",
        0xCB5D: "HANGUL",
        0xCB5E: "HANGUL",
        0xCB5F: "HANGUL",
        0xCB60: "HANGUL",
        0xCB61: "HANGUL",
        0xCB62: "HANGUL",
        0xCB63: "HANGUL",
        0xCB64: "HANGUL",
        0xCB65: "HANGUL",
        0xCB66: "HANGUL",
        0xCB67: "HANGUL",
        0xCB68: "HANGUL",
        0xCB69: "HANGUL",
        0xCB6A: "HANGUL",
        0xCB6B: "HANGUL",
        0xCB6C: "HANGUL",
        0xCB6D: "HANGUL",
        0xCB6E: "HANGUL",
        0xCB6F: "HANGUL",
        0xCB70: "HANGUL",
        0xCB71: "HANGUL",
        0xCB72: "HANGUL",
        0xCB73: "HANGUL",
        0xCB74: "HANGUL",
        0xCB75: "HANGUL",
        0xCB76: "HANGUL",
        0xCB77: "HANGUL",
        0xCB78: "HANGUL",
        0xCB79: "HANGUL",
        0xCB7A: "HANGUL",
        0xCB7B: "HANGUL",
        0xCB7C: "HANGUL",
        0xCB7D: "HANGUL",
        0xCB7E: "HANGUL",
        0xCB7F: "HANGUL",
        0xCB80: "HANGUL",
        0xCB81: "HANGUL",
        0xCB82: "HANGUL",
        0xCB83: "HANGUL",
        0xCB84: "HANGUL",
        0xCB85: "HANGUL",
        0xCB86: "HANGUL",
        0xCB87: "HANGUL",
        0xCB88: "HANGUL",
        0xCB89: "HANGUL",
        0xCB8A: "HANGUL",
        0xCB8B: "HANGUL",
        0xCB8C: "HANGUL",
        0xCB8D: "HANGUL",
        0xCB8E: "HANGUL",
        0xCB8F: "HANGUL",
        0xCB90: "HANGUL",
        0xCB91: "HANGUL",
        0xCB92: "HANGUL",
        0xCB93: "HANGUL",
        0xCB94: "HANGUL",
        0xCB95: "HANGUL",
        0xCB96: "HANGUL",
        0xCB97: "HANGUL",
        0xCB98: "HANGUL",
        0xCB99: "HANGUL",
        0xCB9A: "HANGUL",
        0xCB9B: "HANGUL",
        0xCB9C: "HANGUL",
        0xCB9D: "HANGUL",
        0xCB9E: "HANGUL",
        0xCB9F: "HANGUL",
        0xCBA0: "HANGUL",
        0xCBA1: "HANGUL",
        0xCBA2: "HANGUL",
        0xCBA3: "HANGUL",
        0xCBA4: "HANGUL",
        0xCBA5: "HANGUL",
        0xCBA6: "HANGUL",
        0xCBA7: "HANGUL",
        0xCBA8: "HANGUL",
        0xCBA9: "HANGUL",
        0xCBAA: "HANGUL",
        0xCBAB: "HANGUL",
        0xCBAC: "HANGUL",
        0xCBAD: "HANGUL",
        0xCBAE: "HANGUL",
        0xCBAF: "HANGUL",
        0xCBB0: "HANGUL",
        0xCBB1: "HANGUL",
        0xCBB2: "HANGUL",
        0xCBB3: "HANGUL",
        0xCBB4: "HANGUL",
        0xCBB5: "HANGUL",
        0xCBB6: "HANGUL",
        0xCBB7: "HANGUL",
        0xCBB8: "HANGUL",
        0xCBB9: "HANGUL",
        0xCBBA: "HANGUL",
        0xCBBB: "HANGUL",
        0xCBBC: "HANGUL",
        0xCBBD: "HANGUL",
        0xCBBE: "HANGUL",
        0xCBBF: "HANGUL",
        0xCBC0: "HANGUL",
        0xCBC1: "HANGUL",
        0xCBC2: "HANGUL",
        0xCBC3: "HANGUL",
        0xCBC4: "HANGUL",
        0xCBC5: "HANGUL",
        0xCBC6: "HANGUL",
        0xCBC7: "HANGUL",
        0xCBC8: "HANGUL",
        0xCBC9: "HANGUL",
        0xCBCA: "HANGUL",
        0xCBCB: "HANGUL",
        0xCBCC: "HANGUL",
        0xCBCD: "HANGUL",
        0xCBCE: "HANGUL",
        0xCBCF: "HANGUL",
        0xCBD0: "HANGUL",
        0xCBD1: "HANGUL",
        0xCBD2: "HANGUL",
        0xCBD3: "HANGUL",
        0xCBD4: "HANGUL",
        0xCBD5: "HANGUL",
        0xCBD6: "HANGUL",
        0xCBD7: "HANGUL",
        0xCBD8: "HANGUL",
        0xCBD9: "HANGUL",
        0xCBDA: "HANGUL",
        0xCBDB: "HANGUL",
        0xCBDC: "HANGUL",
        0xCBDD: "HANGUL",
        0xCBDE: "HANGUL",
        0xCBDF: "HANGUL",
        0xCBE0: "HANGUL",
        0xCBE1: "HANGUL",
        0xCBE2: "HANGUL",
        0xCBE3: "HANGUL",
        0xCBE4: "HANGUL",
        0xCBE5: "HANGUL",
        0xCBE6: "HANGUL",
        0xCBE7: "HANGUL",
        0xCBE8: "HANGUL",
        0xCBE9: "HANGUL",
        0xCBEA: "HANGUL",
        0xCBEB: "HANGUL",
        0xCBEC: "HANGUL",
        0xCBED: "HANGUL",
        0xCBEE: "HANGUL",
        0xCBEF: "HANGUL",
        0xCBF0: "HANGUL",
        0xCBF1: "HANGUL",
        0xCBF2: "HANGUL",
        0xCBF3: "HANGUL",
        0xCBF4: "HANGUL",
        0xCBF5: "HANGUL",
        0xCBF6: "HANGUL",
        0xCBF7: "HANGUL",
        0xCBF8: "HANGUL",
        0xCBF9: "HANGUL",
        0xCBFA: "HANGUL",
        0xCBFB: "HANGUL",
        0xCBFC: "HANGUL",
        0xCBFD: "HANGUL",
        0xCBFE: "HANGUL",
        0xCBFF: "HANGUL",
        0xCC00: "HANGUL",
        0xCC01: "HANGUL",
        0xCC02: "HANGUL",
        0xCC03: "HANGUL",
        0xCC04: "HANGUL",
        0xCC05: "HANGUL",
        0xCC06: "HANGUL",
        0xCC07: "HANGUL",
        0xCC08: "HANGUL",
        0xCC09: "HANGUL",
        0xCC0A: "HANGUL",
        0xCC0B: "HANGUL",
        0xCC0C: "HANGUL",
        0xCC0D: "HANGUL",
        0xCC0E: "HANGUL",
        0xCC0F: "HANGUL",
        0xCC10: "HANGUL",
        0xCC11: "HANGUL",
        0xCC12: "HANGUL",
        0xCC13: "HANGUL",
        0xCC14: "HANGUL",
        0xCC15: "HANGUL",
        0xCC16: "HANGUL",
        0xCC17: "HANGUL",
        0xCC18: "HANGUL",
        0xCC19: "HANGUL",
        0xCC1A: "HANGUL",
        0xCC1B: "HANGUL",
        0xCC1C: "HANGUL",
        0xCC1D: "HANGUL",
        0xCC1E: "HANGUL",
        0xCC1F: "HANGUL",
        0xCC20: "HANGUL",
        0xCC21: "HANGUL",
        0xCC22: "HANGUL",
        0xCC23: "HANGUL",
        0xCC24: "HANGUL",
        0xCC25: "HANGUL",
        0xCC26: "HANGUL",
        0xCC27: "HANGUL",
        0xCC28: "HANGUL",
        0xCC29: "HANGUL",
        0xCC2A: "HANGUL",
        0xCC2B: "HANGUL",
        0xCC2C: "HANGUL",
        0xCC2D: "HANGUL",
        0xCC2E: "HANGUL",
        0xCC2F: "HANGUL",
        0xCC30: "HANGUL",
        0xCC31: "HANGUL",
        0xCC32: "HANGUL",
        0xCC33: "HANGUL",
        0xCC34: "HANGUL",
        0xCC35: "HANGUL",
        0xCC36: "HANGUL",
        0xCC37: "HANGUL",
        0xCC38: "HANGUL",
        0xCC39: "HANGUL",
        0xCC3A: "HANGUL",
        0xCC3B: "HANGUL",
        0xCC3C: "HANGUL",
        0xCC3D: "HANGUL",
        0xCC3E: "HANGUL",
        0xCC3F: "HANGUL",
        0xCC40: "HANGUL",
        0xCC41: "HANGUL",
        0xCC42: "HANGUL",
        0xCC43: "HANGUL",
        0xCC44: "HANGUL",
        0xCC45: "HANGUL",
        0xCC46: "HANGUL",
        0xCC47: "HANGUL",
        0xCC48: "HANGUL",
        0xCC49: "HANGUL",
        0xCC4A: "HANGUL",
        0xCC4B: "HANGUL",
        0xCC4C: "HANGUL",
        0xCC4D: "HANGUL",
        0xCC4E: "HANGUL",
        0xCC4F: "HANGUL",
        0xCC50: "HANGUL",
        0xCC51: "HANGUL",
        0xCC52: "HANGUL",
        0xCC53: "HANGUL",
        0xCC54: "HANGUL",
        0xCC55: "HANGUL",
        0xCC56: "HANGUL",
        0xCC57: "HANGUL",
        0xCC58: "HANGUL",
        0xCC59: "HANGUL",
        0xCC5A: "HANGUL",
        0xCC5B: "HANGUL",
        0xCC5C: "HANGUL",
        0xCC5D: "HANGUL",
        0xCC5E: "HANGUL",
        0xCC5F: "HANGUL",
        0xCC60: "HANGUL",
        0xCC61: "HANGUL",
        0xCC62: "HANGUL",
        0xCC63: "HANGUL",
        0xCC64: "HANGUL",
        0xCC65: "HANGUL",
        0xCC66: "HANGUL",
        0xCC67: "HANGUL",
        0xCC68: "HANGUL",
        0xCC69: "HANGUL",
        0xCC6A: "HANGUL",
        0xCC6B: "HANGUL",
        0xCC6C: "HANGUL",
        0xCC6D: "HANGUL",
        0xCC6E: "HANGUL",
        0xCC6F: "HANGUL",
        0xCC70: "HANGUL",
        0xCC71: "HANGUL",
        0xCC72: "HANGUL",
        0xCC73: "HANGUL",
        0xCC74: "HANGUL",
        0xCC75: "HANGUL",
        0xCC76: "HANGUL",
        0xCC77: "HANGUL",
        0xCC78: "HANGUL",
        0xCC79: "HANGUL",
        0xCC7A: "HANGUL",
        0xCC7B: "HANGUL",
        0xCC7C: "HANGUL",
        0xCC7D: "HANGUL",
        0xCC7E: "HANGUL",
        0xCC7F: "HANGUL",
        0xCC80: "HANGUL",
        0xCC81: "HANGUL",
        0xCC82: "HANGUL",
        0xCC83: "HANGUL",
        0xCC84: "HANGUL",
        0xCC85: "HANGUL",
        0xCC86: "HANGUL",
        0xCC87: "HANGUL",
        0xCC88: "HANGUL",
        0xCC89: "HANGUL",
        0xCC8A: "HANGUL",
        0xCC8B: "HANGUL",
        0xCC8C: "HANGUL",
        0xCC8D: "HANGUL",
        0xCC8E: "HANGUL",
        0xCC8F: "HANGUL",
        0xCC90: "HANGUL",
        0xCC91: "HANGUL",
        0xCC92: "HANGUL",
        0xCC93: "HANGUL",
        0xCC94: "HANGUL",
        0xCC95: "HANGUL",
        0xCC96: "HANGUL",
        0xCC97: "HANGUL",
        0xCC98: "HANGUL",
        0xCC99: "HANGUL",
        0xCC9A: "HANGUL",
        0xCC9B: "HANGUL",
        0xCC9C: "HANGUL",
        0xCC9D: "HANGUL",
        0xCC9E: "HANGUL",
        0xCC9F: "HANGUL",
        0xCCA0: "HANGUL",
        0xCCA1: "HANGUL",
        0xCCA2: "HANGUL",
        0xCCA3: "HANGUL",
        0xCCA4: "HANGUL",
        0xCCA5: "HANGUL",
        0xCCA6: "HANGUL",
        0xCCA7: "HANGUL",
        0xCCA8: "HANGUL",
        0xCCA9: "HANGUL",
        0xCCAA: "HANGUL",
        0xCCAB: "HANGUL",
        0xCCAC: "HANGUL",
        0xCCAD: "HANGUL",
        0xCCAE: "HANGUL",
        0xCCAF: "HANGUL",
        0xCCB0: "HANGUL",
        0xCCB1: "HANGUL",
        0xCCB2: "HANGUL",
        0xCCB3: "HANGUL",
        0xCCB4: "HANGUL",
        0xCCB5: "HANGUL",
        0xCCB6: "HANGUL",
        0xCCB7: "HANGUL",
        0xCCB8: "HANGUL",
        0xCCB9: "HANGUL",
        0xCCBA: "HANGUL",
        0xCCBB: "HANGUL",
        0xCCBC: "HANGUL",
        0xCCBD: "HANGUL",
        0xCCBE: "HANGUL",
        0xCCBF: "HANGUL",
        0xCCC0: "HANGUL",
        0xCCC1: "HANGUL",
        0xCCC2: "HANGUL",
        0xCCC3: "HANGUL",
        0xCCC4: "HANGUL",
        0xCCC5: "HANGUL",
        0xCCC6: "HANGUL",
        0xCCC7: "HANGUL",
        0xCCC8: "HANGUL",
        0xCCC9: "HANGUL",
        0xCCCA: "HANGUL",
        0xCCCB: "HANGUL",
        0xCCCC: "HANGUL",
        0xCCCD: "HANGUL",
        0xCCCE: "HANGUL",
        0xCCCF: "HANGUL",
        0xCCD0: "HANGUL",
        0xCCD1: "HANGUL",
        0xCCD2: "HANGUL",
        0xCCD3: "HANGUL",
        0xCCD4: "HANGUL",
        0xCCD5: "HANGUL",
        0xCCD6: "HANGUL",
        0xCCD7: "HANGUL",
        0xCCD8: "HANGUL",
        0xCCD9: "HANGUL",
        0xCCDA: "HANGUL",
        0xCCDB: "HANGUL",
        0xCCDC: "HANGUL",
        0xCCDD: "HANGUL",
        0xCCDE: "HANGUL",
        0xCCDF: "HANGUL",
        0xCCE0: "HANGUL",
        0xCCE1: "HANGUL",
        0xCCE2: "HANGUL",
        0xCCE3: "HANGUL",
        0xCCE4: "HANGUL",
        0xCCE5: "HANGUL",
        0xCCE6: "HANGUL",
        0xCCE7: "HANGUL",
        0xCCE8: "HANGUL",
        0xCCE9: "HANGUL",
        0xCCEA: "HANGUL",
        0xCCEB: "HANGUL",
        0xCCEC: "HANGUL",
        0xCCED: "HANGUL",
        0xCCEE: "HANGUL",
        0xCCEF: "HANGUL",
        0xCCF0: "HANGUL",
        0xCCF1: "HANGUL",
        0xCCF2: "HANGUL",
        0xCCF3: "HANGUL",
        0xCCF4: "HANGUL",
        0xCCF5: "HANGUL",
        0xCCF6: "HANGUL",
        0xCCF7: "HANGUL",
        0xCCF8: "HANGUL",
        0xCCF9: "HANGUL",
        0xCCFA: "HANGUL",
        0xCCFB: "HANGUL",
        0xCCFC: "HANGUL",
        0xCCFD: "HANGUL",
        0xCCFE: "HANGUL",
        0xCCFF: "HANGUL",
        0xCD00: "HANGUL",
        0xCD01: "HANGUL",
        0xCD02: "HANGUL",
        0xCD03: "HANGUL",
        0xCD04: "HANGUL",
        0xCD05: "HANGUL",
        0xCD06: "HANGUL",
        0xCD07: "HANGUL",
        0xCD08: "HANGUL",
        0xCD09: "HANGUL",
        0xCD0A: "HANGUL",
        0xCD0B: "HANGUL",
        0xCD0C: "HANGUL",
        0xCD0D: "HANGUL",
        0xCD0E: "HANGUL",
        0xCD0F: "HANGUL",
        0xCD10: "HANGUL",
        0xCD11: "HANGUL",
        0xCD12: "HANGUL",
        0xCD13: "HANGUL",
        0xCD14: "HANGUL",
        0xCD15: "HANGUL",
        0xCD16: "HANGUL",
        0xCD17: "HANGUL",
        0xCD18: "HANGUL",
        0xCD19: "HANGUL",
        0xCD1A: "HANGUL",
        0xCD1B: "HANGUL",
        0xCD1C: "HANGUL",
        0xCD1D: "HANGUL",
        0xCD1E: "HANGUL",
        0xCD1F: "HANGUL",
        0xCD20: "HANGUL",
        0xCD21: "HANGUL",
        0xCD22: "HANGUL",
        0xCD23: "HANGUL",
        0xCD24: "HANGUL",
        0xCD25: "HANGUL",
        0xCD26: "HANGUL",
        0xCD27: "HANGUL",
        0xCD28: "HANGUL",
        0xCD29: "HANGUL",
        0xCD2A: "HANGUL",
        0xCD2B: "HANGUL",
        0xCD2C: "HANGUL",
        0xCD2D: "HANGUL",
        0xCD2E: "HANGUL",
        0xCD2F: "HANGUL",
        0xCD30: "HANGUL",
        0xCD31: "HANGUL",
        0xCD32: "HANGUL",
        0xCD33: "HANGUL",
        0xCD34: "HANGUL",
        0xCD35: "HANGUL",
        0xCD36: "HANGUL",
        0xCD37: "HANGUL",
        0xCD38: "HANGUL",
        0xCD39: "HANGUL",
        0xCD3A: "HANGUL",
        0xCD3B: "HANGUL",
        0xCD3C: "HANGUL",
        0xCD3D: "HANGUL",
        0xCD3E: "HANGUL",
        0xCD3F: "HANGUL",
        0xCD40: "HANGUL",
        0xCD41: "HANGUL",
        0xCD42: "HANGUL",
        0xCD43: "HANGUL",
        0xCD44: "HANGUL",
        0xCD45: "HANGUL",
        0xCD46: "HANGUL",
        0xCD47: "HANGUL",
        0xCD48: "HANGUL",
        0xCD49: "HANGUL",
        0xCD4A: "HANGUL",
        0xCD4B: "HANGUL",
        0xCD4C: "HANGUL",
        0xCD4D: "HANGUL",
        0xCD4E: "HANGUL",
        0xCD4F: "HANGUL",
        0xCD50: "HANGUL",
        0xCD51: "HANGUL",
        0xCD52: "HANGUL",
        0xCD53: "HANGUL",
        0xCD54: "HANGUL",
        0xCD55: "HANGUL",
        0xCD56: "HANGUL",
        0xCD57: "HANGUL",
        0xCD58: "HANGUL",
        0xCD59: "HANGUL",
        0xCD5A: "HANGUL",
        0xCD5B: "HANGUL",
        0xCD5C: "HANGUL",
        0xCD5D: "HANGUL",
        0xCD5E: "HANGUL",
        0xCD5F: "HANGUL",
        0xCD60: "HANGUL",
        0xCD61: "HANGUL",
        0xCD62: "HANGUL",
        0xCD63: "HANGUL",
        0xCD64: "HANGUL",
        0xCD65: "HANGUL",
        0xCD66: "HANGUL",
        0xCD67: "HANGUL",
        0xCD68: "HANGUL",
        0xCD69: "HANGUL",
        0xCD6A: "HANGUL",
        0xCD6B: "HANGUL",
        0xCD6C: "HANGUL",
        0xCD6D: "HANGUL",
        0xCD6E: "HANGUL",
        0xCD6F: "HANGUL",
        0xCD70: "HANGUL",
        0xCD71: "HANGUL",
        0xCD72: "HANGUL",
        0xCD73: "HANGUL",
        0xCD74: "HANGUL",
        0xCD75: "HANGUL",
        0xCD76: "HANGUL",
        0xCD77: "HANGUL",
        0xCD78: "HANGUL",
        0xCD79: "HANGUL",
        0xCD7A: "HANGUL",
        0xCD7B: "HANGUL",
        0xCD7C: "HANGUL",
        0xCD7D: "HANGUL",
        0xCD7E: "HANGUL",
        0xCD7F: "HANGUL",
        0xCD80: "HANGUL",
        0xCD81: "HANGUL",
        0xCD82: "HANGUL",
        0xCD83: "HANGUL",
        0xCD84: "HANGUL",
        0xCD85: "HANGUL",
        0xCD86: "HANGUL",
        0xCD87: "HANGUL",
        0xCD88: "HANGUL",
        0xCD89: "HANGUL",
        0xCD8A: "HANGUL",
        0xCD8B: "HANGUL",
        0xCD8C: "HANGUL",
        0xCD8D: "HANGUL",
        0xCD8E: "HANGUL",
        0xCD8F: "HANGUL",
        0xCD90: "HANGUL",
        0xCD91: "HANGUL",
        0xCD92: "HANGUL",
        0xCD93: "HANGUL",
        0xCD94: "HANGUL",
        0xCD95: "HANGUL",
        0xCD96: "HANGUL",
        0xCD97: "HANGUL",
        0xCD98: "HANGUL",
        0xCD99: "HANGUL",
        0xCD9A: "HANGUL",
        0xCD9B: "HANGUL",
        0xCD9C: "HANGUL",
        0xCD9D: "HANGUL",
        0xCD9E: "HANGUL",
        0xCD9F: "HANGUL",
        0xCDA0: "HANGUL",
        0xCDA1: "HANGUL",
        0xCDA2: "HANGUL",
        0xCDA3: "HANGUL",
        0xCDA4: "HANGUL",
        0xCDA5: "HANGUL",
        0xCDA6: "HANGUL",
        0xCDA7: "HANGUL",
        0xCDA8: "HANGUL",
        0xCDA9: "HANGUL",
        0xCDAA: "HANGUL",
        0xCDAB: "HANGUL",
        0xCDAC: "HANGUL",
        0xCDAD: "HANGUL",
        0xCDAE: "HANGUL",
        0xCDAF: "HANGUL",
        0xCDB0: "HANGUL",
        0xCDB1: "HANGUL",
        0xCDB2: "HANGUL",
        0xCDB3: "HANGUL",
        0xCDB4: "HANGUL",
        0xCDB5: "HANGUL",
        0xCDB6: "HANGUL",
        0xCDB7: "HANGUL",
        0xCDB8: "HANGUL",
        0xCDB9: "HANGUL",
        0xCDBA: "HANGUL",
        0xCDBB: "HANGUL",
        0xCDBC: "HANGUL",
        0xCDBD: "HANGUL",
        0xCDBE: "HANGUL",
        0xCDBF: "HANGUL",
        0xCDC0: "HANGUL",
        0xCDC1: "HANGUL",
        0xCDC2: "HANGUL",
        0xCDC3: "HANGUL",
        0xCDC4: "HANGUL",
        0xCDC5: "HANGUL",
        0xCDC6: "HANGUL",
        0xCDC7: "HANGUL",
        0xCDC8: "HANGUL",
        0xCDC9: "HANGUL",
        0xCDCA: "HANGUL",
        0xCDCB: "HANGUL",
        0xCDCC: "HANGUL",
        0xCDCD: "HANGUL",
        0xCDCE: "HANGUL",
        0xCDCF: "HANGUL",
        0xCDD0: "HANGUL",
        0xCDD1: "HANGUL",
        0xCDD2: "HANGUL",
        0xCDD3: "HANGUL",
        0xCDD4: "HANGUL",
        0xCDD5: "HANGUL",
        0xCDD6: "HANGUL",
        0xCDD7: "HANGUL",
        0xCDD8: "HANGUL",
        0xCDD9: "HANGUL",
        0xCDDA: "HANGUL",
        0xCDDB: "HANGUL",
        0xCDDC: "HANGUL",
        0xCDDD: "HANGUL",
        0xCDDE: "HANGUL",
        0xCDDF: "HANGUL",
        0xCDE0: "HANGUL",
        0xCDE1: "HANGUL",
        0xCDE2: "HANGUL",
        0xCDE3: "HANGUL",
        0xCDE4: "HANGUL",
        0xCDE5: "HANGUL",
        0xCDE6: "HANGUL",
        0xCDE7: "HANGUL",
        0xCDE8: "HANGUL",
        0xCDE9: "HANGUL",
        0xCDEA: "HANGUL",
        0xCDEB: "HANGUL",
        0xCDEC: "HANGUL",
        0xCDED: "HANGUL",
        0xCDEE: "HANGUL",
        0xCDEF: "HANGUL",
        0xCDF0: "HANGUL",
        0xCDF1: "HANGUL",
        0xCDF2: "HANGUL",
        0xCDF3: "HANGUL",
        0xCDF4: "HANGUL",
        0xCDF5: "HANGUL",
        0xCDF6: "HANGUL",
        0xCDF7: "HANGUL",
        0xCDF8: "HANGUL",
        0xCDF9: "HANGUL",
        0xCDFA: "HANGUL",
        0xCDFB: "HANGUL",
        0xCDFC: "HANGUL",
        0xCDFD: "HANGUL",
        0xCDFE: "HANGUL",
        0xCDFF: "HANGUL",
        0xCE00: "HANGUL",
        0xCE01: "HANGUL",
        0xCE02: "HANGUL",
        0xCE03: "HANGUL",
        0xCE04: "HANGUL",
        0xCE05: "HANGUL",
        0xCE06: "HANGUL",
        0xCE07: "HANGUL",
        0xCE08: "HANGUL",
        0xCE09: "HANGUL",
        0xCE0A: "HANGUL",
        0xCE0B: "HANGUL",
        0xCE0C: "HANGUL",
        0xCE0D: "HANGUL",
        0xCE0E: "HANGUL",
        0xCE0F: "HANGUL",
        0xCE10: "HANGUL",
        0xCE11: "HANGUL",
        0xCE12: "HANGUL",
        0xCE13: "HANGUL",
        0xCE14: "HANGUL",
        0xCE15: "HANGUL",
        0xCE16: "HANGUL",
        0xCE17: "HANGUL",
        0xCE18: "HANGUL",
        0xCE19: "HANGUL",
        0xCE1A: "HANGUL",
        0xCE1B: "HANGUL",
        0xCE1C: "HANGUL",
        0xCE1D: "HANGUL",
        0xCE1E: "HANGUL",
        0xCE1F: "HANGUL",
        0xCE20: "HANGUL",
        0xCE21: "HANGUL",
        0xCE22: "HANGUL",
        0xCE23: "HANGUL",
        0xCE24: "HANGUL",
        0xCE25: "HANGUL",
        0xCE26: "HANGUL",
        0xCE27: "HANGUL",
        0xCE28: "HANGUL",
        0xCE29: "HANGUL",
        0xCE2A: "HANGUL",
        0xCE2B: "HANGUL",
        0xCE2C: "HANGUL",
        0xCE2D: "HANGUL",
        0xCE2E: "HANGUL",
        0xCE2F: "HANGUL",
        0xCE30: "HANGUL",
        0xCE31: "HANGUL",
        0xCE32: "HANGUL",
        0xCE33: "HANGUL",
        0xCE34: "HANGUL",
        0xCE35: "HANGUL",
        0xCE36: "HANGUL",
        0xCE37: "HANGUL",
        0xCE38: "HANGUL",
        0xCE39: "HANGUL",
        0xCE3A: "HANGUL",
        0xCE3B: "HANGUL",
        0xCE3C: "HANGUL",
        0xCE3D: "HANGUL",
        0xCE3E: "HANGUL",
        0xCE3F: "HANGUL",
        0xCE40: "HANGUL",
        0xCE41: "HANGUL",
        0xCE42: "HANGUL",
        0xCE43: "HANGUL",
        0xCE44: "HANGUL",
        0xCE45: "HANGUL",
        0xCE46: "HANGUL",
        0xCE47: "HANGUL",
        0xCE48: "HANGUL",
        0xCE49: "HANGUL",
        0xCE4A: "HANGUL",
        0xCE4B: "HANGUL",
        0xCE4C: "HANGUL",
        0xCE4D: "HANGUL",
        0xCE4E: "HANGUL",
        0xCE4F: "HANGUL",
        0xCE50: "HANGUL",
        0xCE51: "HANGUL",
        0xCE52: "HANGUL",
        0xCE53: "HANGUL",
        0xCE54: "HANGUL",
        0xCE55: "HANGUL",
        0xCE56: "HANGUL",
        0xCE57: "HANGUL",
        0xCE58: "HANGUL",
        0xCE59: "HANGUL",
        0xCE5A: "HANGUL",
        0xCE5B: "HANGUL",
        0xCE5C: "HANGUL",
        0xCE5D: "HANGUL",
        0xCE5E: "HANGUL",
        0xCE5F: "HANGUL",
        0xCE60: "HANGUL",
        0xCE61: "HANGUL",
        0xCE62: "HANGUL",
        0xCE63: "HANGUL",
        0xCE64: "HANGUL",
        0xCE65: "HANGUL",
        0xCE66: "HANGUL",
        0xCE67: "HANGUL",
        0xCE68: "HANGUL",
        0xCE69: "HANGUL",
        0xCE6A: "HANGUL",
        0xCE6B: "HANGUL",
        0xCE6C: "HANGUL",
        0xCE6D: "HANGUL",
        0xCE6E: "HANGUL",
        0xCE6F: "HANGUL",
        0xCE70: "HANGUL",
        0xCE71: "HANGUL",
        0xCE72: "HANGUL",
        0xCE73: "HANGUL",
        0xCE74: "HANGUL",
        0xCE75: "HANGUL",
        0xCE76: "HANGUL",
        0xCE77: "HANGUL",
        0xCE78: "HANGUL",
        0xCE79: "HANGUL",
        0xCE7A: "HANGUL",
        0xCE7B: "HANGUL",
        0xCE7C: "HANGUL",
        0xCE7D: "HANGUL",
        0xCE7E: "HANGUL",
        0xCE7F: "HANGUL",
        0xCE80: "HANGUL",
        0xCE81: "HANGUL",
        0xCE82: "HANGUL",
        0xCE83: "HANGUL",
        0xCE84: "HANGUL",
        0xCE85: "HANGUL",
        0xCE86: "HANGUL",
        0xCE87: "HANGUL",
        0xCE88: "HANGUL",
        0xCE89: "HANGUL",
        0xCE8A: "HANGUL",
        0xCE8B: "HANGUL",
        0xCE8C: "HANGUL",
        0xCE8D: "HANGUL",
        0xCE8E: "HANGUL",
        0xCE8F: "HANGUL",
        0xCE90: "HANGUL",
        0xCE91: "HANGUL",
        0xCE92: "HANGUL",
        0xCE93: "HANGUL",
        0xCE94: "HANGUL",
        0xCE95: "HANGUL",
        0xCE96: "HANGUL",
        0xCE97: "HANGUL",
        0xCE98: "HANGUL",
        0xCE99: "HANGUL",
        0xCE9A: "HANGUL",
        0xCE9B: "HANGUL",
        0xCE9C: "HANGUL",
        0xCE9D: "HANGUL",
        0xCE9E: "HANGUL",
        0xCE9F: "HANGUL",
        0xCEA0: "HANGUL",
        0xCEA1: "HANGUL",
        0xCEA2: "HANGUL",
        0xCEA3: "HANGUL",
        0xCEA4: "HANGUL",
        0xCEA5: "HANGUL",
        0xCEA6: "HANGUL",
        0xCEA7: "HANGUL",
        0xCEA8: "HANGUL",
        0xCEA9: "HANGUL",
        0xCEAA: "HANGUL",
        0xCEAB: "HANGUL",
        0xCEAC: "HANGUL",
        0xCEAD: "HANGUL",
        0xCEAE: "HANGUL",
        0xCEAF: "HANGUL",
        0xCEB0: "HANGUL",
        0xCEB1: "HANGUL",
        0xCEB2: "HANGUL",
        0xCEB3: "HANGUL",
        0xCEB4: "HANGUL",
        0xCEB5: "HANGUL",
        0xCEB6: "HANGUL",
        0xCEB7: "HANGUL",
        0xCEB8: "HANGUL",
        0xCEB9: "HANGUL",
        0xCEBA: "HANGUL",
        0xCEBB: "HANGUL",
        0xCEBC: "HANGUL",
        0xCEBD: "HANGUL",
        0xCEBE: "HANGUL",
        0xCEBF: "HANGUL",
        0xCEC0: "HANGUL",
        0xCEC1: "HANGUL",
        0xCEC2: "HANGUL",
        0xCEC3: "HANGUL",
        0xCEC4: "HANGUL",
        0xCEC5: "HANGUL",
        0xCEC6: "HANGUL",
        0xCEC7: "HANGUL",
        0xCEC8: "HANGUL",
        0xCEC9: "HANGUL",
        0xCECA: "HANGUL",
        0xCECB: "HANGUL",
        0xCECC: "HANGUL",
        0xCECD: "HANGUL",
        0xCECE: "HANGUL",
        0xCECF: "HANGUL",
        0xCED0: "HANGUL",
        0xCED1: "HANGUL",
        0xCED2: "HANGUL",
        0xCED3: "HANGUL",
        0xCED4: "HANGUL",
        0xCED5: "HANGUL",
        0xCED6: "HANGUL",
        0xCED7: "HANGUL",
        0xCED8: "HANGUL",
        0xCED9: "HANGUL",
        0xCEDA: "HANGUL",
        0xCEDB: "HANGUL",
        0xCEDC: "HANGUL",
        0xCEDD: "HANGUL",
        0xCEDE: "HANGUL",
        0xCEDF: "HANGUL",
        0xCEE0: "HANGUL",
        0xCEE1: "HANGUL",
        0xCEE2: "HANGUL",
        0xCEE3: "HANGUL",
        0xCEE4: "HANGUL",
        0xCEE5: "HANGUL",
        0xCEE6: "HANGUL",
        0xCEE7: "HANGUL",
        0xCEE8: "HANGUL",
        0xCEE9: "HANGUL",
        0xCEEA: "HANGUL",
        0xCEEB: "HANGUL",
        0xCEEC: "HANGUL",
        0xCEED: "HANGUL",
        0xCEEE: "HANGUL",
        0xCEEF: "HANGUL",
        0xCEF0: "HANGUL",
        0xCEF1: "HANGUL",
        0xCEF2: "HANGUL",
        0xCEF3: "HANGUL",
        0xCEF4: "HANGUL",
        0xCEF5: "HANGUL",
        0xCEF6: "HANGUL",
        0xCEF7: "HANGUL",
        0xCEF8: "HANGUL",
        0xCEF9: "HANGUL",
        0xCEFA: "HANGUL",
        0xCEFB: "HANGUL",
        0xCEFC: "HANGUL",
        0xCEFD: "HANGUL",
        0xCEFE: "HANGUL",
        0xCEFF: "HANGUL",
        0xCF00: "HANGUL",
        0xCF01: "HANGUL",
        0xCF02: "HANGUL",
        0xCF03: "HANGUL",
        0xCF04: "HANGUL",
        0xCF05: "HANGUL",
        0xCF06: "HANGUL",
        0xCF07: "HANGUL",
        0xCF08: "HANGUL",
        0xCF09: "HANGUL",
        0xCF0A: "HANGUL",
        0xCF0B: "HANGUL",
        0xCF0C: "HANGUL",
        0xCF0D: "HANGUL",
        0xCF0E: "HANGUL",
        0xCF0F: "HANGUL",
        0xCF10: "HANGUL",
        0xCF11: "HANGUL",
        0xCF12: "HANGUL",
        0xCF13: "HANGUL",
        0xCF14: "HANGUL",
        0xCF15: "HANGUL",
        0xCF16: "HANGUL",
        0xCF17: "HANGUL",
        0xCF18: "HANGUL",
        0xCF19: "HANGUL",
        0xCF1A: "HANGUL",
        0xCF1B: "HANGUL",
        0xCF1C: "HANGUL",
        0xCF1D: "HANGUL",
        0xCF1E: "HANGUL",
        0xCF1F: "HANGUL",
        0xCF20: "HANGUL",
        0xCF21: "HANGUL",
        0xCF22: "HANGUL",
        0xCF23: "HANGUL",
        0xCF24: "HANGUL",
        0xCF25: "HANGUL",
        0xCF26: "HANGUL",
        0xCF27: "HANGUL",
        0xCF28: "HANGUL",
        0xCF29: "HANGUL",
        0xCF2A: "HANGUL",
        0xCF2B: "HANGUL",
        0xCF2C: "HANGUL",
        0xCF2D: "HANGUL",
        0xCF2E: "HANGUL",
        0xCF2F: "HANGUL",
        0xCF30: "HANGUL",
        0xCF31: "HANGUL",
        0xCF32: "HANGUL",
        0xCF33: "HANGUL",
        0xCF34: "HANGUL",
        0xCF35: "HANGUL",
        0xCF36: "HANGUL",
        0xCF37: "HANGUL",
        0xCF38: "HANGUL",
        0xCF39: "HANGUL",
        0xCF3A: "HANGUL",
        0xCF3B: "HANGUL",
        0xCF3C: "HANGUL",
        0xCF3D: "HANGUL",
        0xCF3E: "HANGUL",
        0xCF3F: "HANGUL",
        0xCF40: "HANGUL",
        0xCF41: "HANGUL",
        0xCF42: "HANGUL",
        0xCF43: "HANGUL",
        0xCF44: "HANGUL",
        0xCF45: "HANGUL",
        0xCF46: "HANGUL",
        0xCF47: "HANGUL",
        0xCF48: "HANGUL",
        0xCF49: "HANGUL",
        0xCF4A: "HANGUL",
        0xCF4B: "HANGUL",
        0xCF4C: "HANGUL",
        0xCF4D: "HANGUL",
        0xCF4E: "HANGUL",
        0xCF4F: "HANGUL",
        0xCF50: "HANGUL",
        0xCF51: "HANGUL",
        0xCF52: "HANGUL",
        0xCF53: "HANGUL",
        0xCF54: "HANGUL",
        0xCF55: "HANGUL",
        0xCF56: "HANGUL",
        0xCF57: "HANGUL",
        0xCF58: "HANGUL",
        0xCF59: "HANGUL",
        0xCF5A: "HANGUL",
        0xCF5B: "HANGUL",
        0xCF5C: "HANGUL",
        0xCF5D: "HANGUL",
        0xCF5E: "HANGUL",
        0xCF5F: "HANGUL",
        0xCF60: "HANGUL",
        0xCF61: "HANGUL",
        0xCF62: "HANGUL",
        0xCF63: "HANGUL",
        0xCF64: "HANGUL",
        0xCF65: "HANGUL",
        0xCF66: "HANGUL",
        0xCF67: "HANGUL",
        0xCF68: "HANGUL",
        0xCF69: "HANGUL",
        0xCF6A: "HANGUL",
        0xCF6B: "HANGUL",
        0xCF6C: "HANGUL",
        0xCF6D: "HANGUL",
        0xCF6E: "HANGUL",
        0xCF6F: "HANGUL",
        0xCF70: "HANGUL",
        0xCF71: "HANGUL",
        0xCF72: "HANGUL",
        0xCF73: "HANGUL",
        0xCF74: "HANGUL",
        0xCF75: "HANGUL",
        0xCF76: "HANGUL",
        0xCF77: "HANGUL",
        0xCF78: "HANGUL",
        0xCF79: "HANGUL",
        0xCF7A: "HANGUL",
        0xCF7B: "HANGUL",
        0xCF7C: "HANGUL",
        0xCF7D: "HANGUL",
        0xCF7E: "HANGUL",
        0xCF7F: "HANGUL",
        0xCF80: "HANGUL",
        0xCF81: "HANGUL",
        0xCF82: "HANGUL",
        0xCF83: "HANGUL",
        0xCF84: "HANGUL",
        0xCF85: "HANGUL",
        0xCF86: "HANGUL",
        0xCF87: "HANGUL",
        0xCF88: "HANGUL",
        0xCF89: "HANGUL",
        0xCF8A: "HANGUL",
        0xCF8B: "HANGUL",
        0xCF8C: "HANGUL",
        0xCF8D: "HANGUL",
        0xCF8E: "HANGUL",
        0xCF8F: "HANGUL",
        0xCF90: "HANGUL",
        0xCF91: "HANGUL",
        0xCF92: "HANGUL",
        0xCF93: "HANGUL",
        0xCF94: "HANGUL",
        0xCF95: "HANGUL",
        0xCF96: "HANGUL",
        0xCF97: "HANGUL",
        0xCF98: "HANGUL",
        0xCF99: "HANGUL",
        0xCF9A: "HANGUL",
        0xCF9B: "HANGUL",
        0xCF9C: "HANGUL",
        0xCF9D: "HANGUL",
        0xCF9E: "HANGUL",
        0xCF9F: "HANGUL",
        0xCFA0: "HANGUL",
        0xCFA1: "HANGUL",
        0xCFA2: "HANGUL",
        0xCFA3: "HANGUL",
        0xCFA4: "HANGUL",
        0xCFA5: "HANGUL",
        0xCFA6: "HANGUL",
        0xCFA7: "HANGUL",
        0xCFA8: "HANGUL",
        0xCFA9: "HANGUL",
        0xCFAA: "HANGUL",
        0xCFAB: "HANGUL",
        0xCFAC: "HANGUL",
        0xCFAD: "HANGUL",
        0xCFAE: "HANGUL",
        0xCFAF: "HANGUL",
        0xCFB0: "HANGUL",
        0xCFB1: "HANGUL",
        0xCFB2: "HANGUL",
        0xCFB3: "HANGUL",
        0xCFB4: "HANGUL",
        0xCFB5: "HANGUL",
        0xCFB6: "HANGUL",
        0xCFB7: "HANGUL",
        0xCFB8: "HANGUL",
        0xCFB9: "HANGUL",
        0xCFBA: "HANGUL",
        0xCFBB: "HANGUL",
        0xCFBC: "HANGUL",
        0xCFBD: "HANGUL",
        0xCFBE: "HANGUL",
        0xCFBF: "HANGUL",
        0xCFC0: "HANGUL",
        0xCFC1: "HANGUL",
        0xCFC2: "HANGUL",
        0xCFC3: "HANGUL",
        0xCFC4: "HANGUL",
        0xCFC5: "HANGUL",
        0xCFC6: "HANGUL",
        0xCFC7: "HANGUL",
        0xCFC8: "HANGUL",
        0xCFC9: "HANGUL",
        0xCFCA: "HANGUL",
        0xCFCB: "HANGUL",
        0xCFCC: "HANGUL",
        0xCFCD: "HANGUL",
        0xCFCE: "HANGUL",
        0xCFCF: "HANGUL",
        0xCFD0: "HANGUL",
        0xCFD1: "HANGUL",
        0xCFD2: "HANGUL",
        0xCFD3: "HANGUL",
        0xCFD4: "HANGUL",
        0xCFD5: "HANGUL",
        0xCFD6: "HANGUL",
        0xCFD7: "HANGUL",
        0xCFD8: "HANGUL",
        0xCFD9: "HANGUL",
        0xCFDA: "HANGUL",
        0xCFDB: "HANGUL",
        0xCFDC: "HANGUL",
        0xCFDD: "HANGUL",
        0xCFDE: "HANGUL",
        0xCFDF: "HANGUL",
        0xCFE0: "HANGUL",
        0xCFE1: "HANGUL",
        0xCFE2: "HANGUL",
        0xCFE3: "HANGUL",
        0xCFE4: "HANGUL",
        0xCFE5: "HANGUL",
        0xCFE6: "HANGUL",
        0xCFE7: "HANGUL",
        0xCFE8: "HANGUL",
        0xCFE9: "HANGUL",
        0xCFEA: "HANGUL",
        0xCFEB: "HANGUL",
        0xCFEC: "HANGUL",
        0xCFED: "HANGUL",
        0xCFEE: "HANGUL",
        0xCFEF: "HANGUL",
        0xCFF0: "HANGUL",
        0xCFF1: "HANGUL",
        0xCFF2: "HANGUL",
        0xCFF3: "HANGUL",
        0xCFF4: "HANGUL",
        0xCFF5: "HANGUL",
        0xCFF6: "HANGUL",
        0xCFF7: "HANGUL",
        0xCFF8: "HANGUL",
        0xCFF9: "HANGUL",
        0xCFFA: "HANGUL",
        0xCFFB: "HANGUL",
        0xCFFC: "HANGUL",
        0xCFFD: "HANGUL",
        0xCFFE: "HANGUL",
        0xCFFF: "HANGUL",
        0xD000: "HANGUL",
        0xD001: "HANGUL",
        0xD002: "HANGUL",
        0xD003: "HANGUL",
        0xD004: "HANGUL",
        0xD005: "HANGUL",
        0xD006: "HANGUL",
        0xD007: "HANGUL",
        0xD008: "HANGUL",
        0xD009: "HANGUL",
        0xD00A: "HANGUL",
        0xD00B: "HANGUL",
        0xD00C: "HANGUL",
        0xD00D: "HANGUL",
        0xD00E: "HANGUL",
        0xD00F: "HANGUL",
        0xD010: "HANGUL",
        0xD011: "HANGUL",
        0xD012: "HANGUL",
        0xD013: "HANGUL",
        0xD014: "HANGUL",
        0xD015: "HANGUL",
        0xD016: "HANGUL",
        0xD017: "HANGUL",
        0xD018: "HANGUL",
        0xD019: "HANGUL",
        0xD01A: "HANGUL",
        0xD01B: "HANGUL",
        0xD01C: "HANGUL",
        0xD01D: "HANGUL",
        0xD01E: "HANGUL",
        0xD01F: "HANGUL",
        0xD020: "HANGUL",
        0xD021: "HANGUL",
        0xD022: "HANGUL",
        0xD023: "HANGUL",
        0xD024: "HANGUL",
        0xD025: "HANGUL",
        0xD026: "HANGUL",
        0xD027: "HANGUL",
        0xD028: "HANGUL",
        0xD029: "HANGUL",
        0xD02A: "HANGUL",
        0xD02B: "HANGUL",
        0xD02C: "HANGUL",
        0xD02D: "HANGUL",
        0xD02E: "HANGUL",
        0xD02F: "HANGUL",
        0xD030: "HANGUL",
        0xD031: "HANGUL",
        0xD032: "HANGUL",
        0xD033: "HANGUL",
        0xD034: "HANGUL",
        0xD035: "HANGUL",
        0xD036: "HANGUL",
        0xD037: "HANGUL",
        0xD038: "HANGUL",
        0xD039: "HANGUL",
        0xD03A: "HANGUL",
        0xD03B: "HANGUL",
        0xD03C: "HANGUL",
        0xD03D: "HANGUL",
        0xD03E: "HANGUL",
        0xD03F: "HANGUL",
        0xD040: "HANGUL",
        0xD041: "HANGUL",
        0xD042: "HANGUL",
        0xD043: "HANGUL",
        0xD044: "HANGUL",
        0xD045: "HANGUL",
        0xD046: "HANGUL",
        0xD047: "HANGUL",
        0xD048: "HANGUL",
        0xD049: "HANGUL",
        0xD04A: "HANGUL",
        0xD04B: "HANGUL",
        0xD04C: "HANGUL",
        0xD04D: "HANGUL",
        0xD04E: "HANGUL",
        0xD04F: "HANGUL",
        0xD050: "HANGUL",
        0xD051: "HANGUL",
        0xD052: "HANGUL",
        0xD053: "HANGUL",
        0xD054: "HANGUL",
        0xD055: "HANGUL",
        0xD056: "HANGUL",
        0xD057: "HANGUL",
        0xD058: "HANGUL",
        0xD059: "HANGUL",
        0xD05A: "HANGUL",
        0xD05B: "HANGUL",
        0xD05C: "HANGUL",
        0xD05D: "HANGUL",
        0xD05E: "HANGUL",
        0xD05F: "HANGUL",
        0xD060: "HANGUL",
        0xD061: "HANGUL",
        0xD062: "HANGUL",
        0xD063: "HANGUL",
        0xD064: "HANGUL",
        0xD065: "HANGUL",
        0xD066: "HANGUL",
        0xD067: "HANGUL",
        0xD068: "HANGUL",
        0xD069: "HANGUL",
        0xD06A: "HANGUL",
        0xD06B: "HANGUL",
        0xD06C: "HANGUL",
        0xD06D: "HANGUL",
        0xD06E: "HANGUL",
        0xD06F: "HANGUL",
        0xD070: "HANGUL",
        0xD071: "HANGUL",
        0xD072: "HANGUL",
        0xD073: "HANGUL",
        0xD074: "HANGUL",
        0xD075: "HANGUL",
        0xD076: "HANGUL",
        0xD077: "HANGUL",
        0xD078: "HANGUL",
        0xD079: "HANGUL",
        0xD07A: "HANGUL",
        0xD07B: "HANGUL",
        0xD07C: "HANGUL",
        0xD07D: "HANGUL",
        0xD07E: "HANGUL",
        0xD07F: "HANGUL",
        0xD080: "HANGUL",
        0xD081: "HANGUL",
        0xD082: "HANGUL",
        0xD083: "HANGUL",
        0xD084: "HANGUL",
        0xD085: "HANGUL",
        0xD086: "HANGUL",
        0xD087: "HANGUL",
        0xD088: "HANGUL",
        0xD089: "HANGUL",
        0xD08A: "HANGUL",
        0xD08B: "HANGUL",
        0xD08C: "HANGUL",
        0xD08D: "HANGUL",
        0xD08E: "HANGUL",
        0xD08F: "HANGUL",
        0xD090: "HANGUL",
        0xD091: "HANGUL",
        0xD092: "HANGUL",
        0xD093: "HANGUL",
        0xD094: "HANGUL",
        0xD095: "HANGUL",
        0xD096: "HANGUL",
        0xD097: "HANGUL",
        0xD098: "HANGUL",
        0xD099: "HANGUL",
        0xD09A: "HANGUL",
        0xD09B: "HANGUL",
        0xD09C: "HANGUL",
        0xD09D: "HANGUL",
        0xD09E: "HANGUL",
        0xD09F: "HANGUL",
        0xD0A0: "HANGUL",
        0xD0A1: "HANGUL",
        0xD0A2: "HANGUL",
        0xD0A3: "HANGUL",
        0xD0A4: "HANGUL",
        0xD0A5: "HANGUL",
        0xD0A6: "HANGUL",
        0xD0A7: "HANGUL",
        0xD0A8: "HANGUL",
        0xD0A9: "HANGUL",
        0xD0AA: "HANGUL",
        0xD0AB: "HANGUL",
        0xD0AC: "HANGUL",
        0xD0AD: "HANGUL",
        0xD0AE: "HANGUL",
        0xD0AF: "HANGUL",
        0xD0B0: "HANGUL",
        0xD0B1: "HANGUL",
        0xD0B2: "HANGUL",
        0xD0B3: "HANGUL",
        0xD0B4: "HANGUL",
        0xD0B5: "HANGUL",
        0xD0B6: "HANGUL",
        0xD0B7: "HANGUL",
        0xD0B8: "HANGUL",
        0xD0B9: "HANGUL",
        0xD0BA: "HANGUL",
        0xD0BB: "HANGUL",
        0xD0BC: "HANGUL",
        0xD0BD: "HANGUL",
        0xD0BE: "HANGUL",
        0xD0BF: "HANGUL",
        0xD0C0: "HANGUL",
        0xD0C1: "HANGUL",
        0xD0C2: "HANGUL",
        0xD0C3: "HANGUL",
        0xD0C4: "HANGUL",
        0xD0C5: "HANGUL",
        0xD0C6: "HANGUL",
        0xD0C7: "HANGUL",
        0xD0C8: "HANGUL",
        0xD0C9: "HANGUL",
        0xD0CA: "HANGUL",
        0xD0CB: "HANGUL",
        0xD0CC: "HANGUL",
        0xD0CD: "HANGUL",
        0xD0CE: "HANGUL",
        0xD0CF: "HANGUL",
        0xD0D0: "HANGUL",
        0xD0D1: "HANGUL",
        0xD0D2: "HANGUL",
        0xD0D3: "HANGUL",
        0xD0D4: "HANGUL",
        0xD0D5: "HANGUL",
        0xD0D6: "HANGUL",
        0xD0D7: "HANGUL",
        0xD0D8: "HANGUL",
        0xD0D9: "HANGUL",
        0xD0DA: "HANGUL",
        0xD0DB: "HANGUL",
        0xD0DC: "HANGUL",
        0xD0DD: "HANGUL",
        0xD0DE: "HANGUL",
        0xD0DF: "HANGUL",
        0xD0E0: "HANGUL",
        0xD0E1: "HANGUL",
        0xD0E2: "HANGUL",
        0xD0E3: "HANGUL",
        0xD0E4: "HANGUL",
        0xD0E5: "HANGUL",
        0xD0E6: "HANGUL",
        0xD0E7: "HANGUL",
        0xD0E8: "HANGUL",
        0xD0E9: "HANGUL",
        0xD0EA: "HANGUL",
        0xD0EB: "HANGUL",
        0xD0EC: "HANGUL",
        0xD0ED: "HANGUL",
        0xD0EE: "HANGUL",
        0xD0EF: "HANGUL",
        0xD0F0: "HANGUL",
        0xD0F1: "HANGUL",
        0xD0F2: "HANGUL",
        0xD0F3: "HANGUL",
        0xD0F4: "HANGUL",
        0xD0F5: "HANGUL",
        0xD0F6: "HANGUL",
        0xD0F7: "HANGUL",
        0xD0F8: "HANGUL",
        0xD0F9: "HANGUL",
        0xD0FA: "HANGUL",
        0xD0FB: "HANGUL",
        0xD0FC: "HANGUL",
        0xD0FD: "HANGUL",
        0xD0FE: "HANGUL",
        0xD0FF: "HANGUL",
        0xD100: "HANGUL",
        0xD101: "HANGUL",
        0xD102: "HANGUL",
        0xD103: "HANGUL",
        0xD104: "HANGUL",
        0xD105: "HANGUL",
        0xD106: "HANGUL",
        0xD107: "HANGUL",
        0xD108: "HANGUL",
        0xD109: "HANGUL",
        0xD10A: "HANGUL",
        0xD10B: "HANGUL",
        0xD10C: "HANGUL",
        0xD10D: "HANGUL",
        0xD10E: "HANGUL",
        0xD10F: "HANGUL",
        0xD110: "HANGUL",
        0xD111: "HANGUL",
        0xD112: "HANGUL",
        0xD113: "HANGUL",
        0xD114: "HANGUL",
        0xD115: "HANGUL",
        0xD116: "HANGUL",
        0xD117: "HANGUL",
        0xD118: "HANGUL",
        0xD119: "HANGUL",
        0xD11A: "HANGUL",
        0xD11B: "HANGUL",
        0xD11C: "HANGUL",
        0xD11D: "HANGUL",
        0xD11E: "HANGUL",
        0xD11F: "HANGUL",
        0xD120: "HANGUL",
        0xD121: "HANGUL",
        0xD122: "HANGUL",
        0xD123: "HANGUL",
        0xD124: "HANGUL",
        0xD125: "HANGUL",
        0xD126: "HANGUL",
        0xD127: "HANGUL",
        0xD128: "HANGUL",
        0xD129: "HANGUL",
        0xD12A: "HANGUL",
        0xD12B: "HANGUL",
        0xD12C: "HANGUL",
        0xD12D: "HANGUL",
        0xD12E: "HANGUL",
        0xD12F: "HANGUL",
        0xD130: "HANGUL",
        0xD131: "HANGUL",
        0xD132: "HANGUL",
        0xD133: "HANGUL",
        0xD134: "HANGUL",
        0xD135: "HANGUL",
        0xD136: "HANGUL",
        0xD137: "HANGUL",
        0xD138: "HANGUL",
        0xD139: "HANGUL",
        0xD13A: "HANGUL",
        0xD13B: "HANGUL",
        0xD13C: "HANGUL",
        0xD13D: "HANGUL",
        0xD13E: "HANGUL",
        0xD13F: "HANGUL",
        0xD140: "HANGUL",
        0xD141: "HANGUL",
        0xD142: "HANGUL",
        0xD143: "HANGUL",
        0xD144: "HANGUL",
        0xD145: "HANGUL",
        0xD146: "HANGUL",
        0xD147: "HANGUL",
        0xD148: "HANGUL",
        0xD149: "HANGUL",
        0xD14A: "HANGUL",
        0xD14B: "HANGUL",
        0xD14C: "HANGUL",
        0xD14D: "HANGUL",
        0xD14E: "HANGUL",
        0xD14F: "HANGUL",
        0xD150: "HANGUL",
        0xD151: "HANGUL",
        0xD152: "HANGUL",
        0xD153: "HANGUL",
        0xD154: "HANGUL",
        0xD155: "HANGUL",
        0xD156: "HANGUL",
        0xD157: "HANGUL",
        0xD158: "HANGUL",
        0xD159: "HANGUL",
        0xD15A: "HANGUL",
        0xD15B: "HANGUL",
        0xD15C: "HANGUL",
        0xD15D: "HANGUL",
        0xD15E: "HANGUL",
        0xD15F: "HANGUL",
        0xD160: "HANGUL",
        0xD161: "HANGUL",
        0xD162: "HANGUL",
        0xD163: "HANGUL",
        0xD164: "HANGUL",
        0xD165: "HANGUL",
        0xD166: "HANGUL",
        0xD167: "HANGUL",
        0xD168: "HANGUL",
        0xD169: "HANGUL",
        0xD16A: "HANGUL",
        0xD16B: "HANGUL",
        0xD16C: "HANGUL",
        0xD16D: "HANGUL",
        0xD16E: "HANGUL",
        0xD16F: "HANGUL",
        0xD170: "HANGUL",
        0xD171: "HANGUL",
        0xD172: "HANGUL",
        0xD173: "HANGUL",
        0xD174: "HANGUL",
        0xD175: "HANGUL",
        0xD176: "HANGUL",
        0xD177: "HANGUL",
        0xD178: "HANGUL",
        0xD179: "HANGUL",
        0xD17A: "HANGUL",
        0xD17B: "HANGUL",
        0xD17C: "HANGUL",
        0xD17D: "HANGUL",
        0xD17E: "HANGUL",
        0xD17F: "HANGUL",
        0xD180: "HANGUL",
        0xD181: "HANGUL",
        0xD182: "HANGUL",
        0xD183: "HANGUL",
        0xD184: "HANGUL",
        0xD185: "HANGUL",
        0xD186: "HANGUL",
        0xD187: "HANGUL",
        0xD188: "HANGUL",
        0xD189: "HANGUL",
        0xD18A: "HANGUL",
        0xD18B: "HANGUL",
        0xD18C: "HANGUL",
        0xD18D: "HANGUL",
        0xD18E: "HANGUL",
        0xD18F: "HANGUL",
        0xD190: "HANGUL",
        0xD191: "HANGUL",
        0xD192: "HANGUL",
        0xD193: "HANGUL",
        0xD194: "HANGUL",
        0xD195: "HANGUL",
        0xD196: "HANGUL",
        0xD197: "HANGUL",
        0xD198: "HANGUL",
        0xD199: "HANGUL",
        0xD19A: "HANGUL",
        0xD19B: "HANGUL",
        0xD19C: "HANGUL",
        0xD19D: "HANGUL",
        0xD19E: "HANGUL",
        0xD19F: "HANGUL",
        0xD1A0: "HANGUL",
        0xD1A1: "HANGUL",
        0xD1A2: "HANGUL",
        0xD1A3: "HANGUL",
        0xD1A4: "HANGUL",
        0xD1A5: "HANGUL",
        0xD1A6: "HANGUL",
        0xD1A7: "HANGUL",
        0xD1A8: "HANGUL",
        0xD1A9: "HANGUL",
        0xD1AA: "HANGUL",
        0xD1AB: "HANGUL",
        0xD1AC: "HANGUL",
        0xD1AD: "HANGUL",
        0xD1AE: "HANGUL",
        0xD1AF: "HANGUL",
        0xD1B0: "HANGUL",
        0xD1B1: "HANGUL",
        0xD1B2: "HANGUL",
        0xD1B3: "HANGUL",
        0xD1B4: "HANGUL",
        0xD1B5: "HANGUL",
        0xD1B6: "HANGUL",
        0xD1B7: "HANGUL",
        0xD1B8: "HANGUL",
        0xD1B9: "HANGUL",
        0xD1BA: "HANGUL",
        0xD1BB: "HANGUL",
        0xD1BC: "HANGUL",
        0xD1BD: "HANGUL",
        0xD1BE: "HANGUL",
        0xD1BF: "HANGUL",
        0xD1C0: "HANGUL",
        0xD1C1: "HANGUL",
        0xD1C2: "HANGUL",
        0xD1C3: "HANGUL",
        0xD1C4: "HANGUL",
        0xD1C5: "HANGUL",
        0xD1C6: "HANGUL",
        0xD1C7: "HANGUL",
        0xD1C8: "HANGUL",
        0xD1C9: "HANGUL",
        0xD1CA: "HANGUL",
        0xD1CB: "HANGUL",
        0xD1CC: "HANGUL",
        0xD1CD: "HANGUL",
        0xD1CE: "HANGUL",
        0xD1CF: "HANGUL",
        0xD1D0: "HANGUL",
        0xD1D1: "HANGUL",
        0xD1D2: "HANGUL",
        0xD1D3: "HANGUL",
        0xD1D4: "HANGUL",
        0xD1D5: "HANGUL",
        0xD1D6: "HANGUL",
        0xD1D7: "HANGUL",
        0xD1D8: "HANGUL",
        0xD1D9: "HANGUL",
        0xD1DA: "HANGUL",
        0xD1DB: "HANGUL",
        0xD1DC: "HANGUL",
        0xD1DD: "HANGUL",
        0xD1DE: "HANGUL",
        0xD1DF: "HANGUL",
        0xD1E0: "HANGUL",
        0xD1E1: "HANGUL",
        0xD1E2: "HANGUL",
        0xD1E3: "HANGUL",
        0xD1E4: "HANGUL",
        0xD1E5: "HANGUL",
        0xD1E6: "HANGUL",
        0xD1E7: "HANGUL",
        0xD1E8: "HANGUL",
        0xD1E9: "HANGUL",
        0xD1EA: "HANGUL",
        0xD1EB: "HANGUL",
        0xD1EC: "HANGUL",
        0xD1ED: "HANGUL",
        0xD1EE: "HANGUL",
        0xD1EF: "HANGUL",
        0xD1F0: "HANGUL",
        0xD1F1: "HANGUL",
        0xD1F2: "HANGUL",
        0xD1F3: "HANGUL",
        0xD1F4: "HANGUL",
        0xD1F5: "HANGUL",
        0xD1F6: "HANGUL",
        0xD1F7: "HANGUL",
        0xD1F8: "HANGUL",
        0xD1F9: "HANGUL",
        0xD1FA: "HANGUL",
        0xD1FB: "HANGUL",
        0xD1FC: "HANGUL",
        0xD1FD: "HANGUL",
        0xD1FE: "HANGUL",
        0xD1FF: "HANGUL",
        0xD200: "HANGUL",
        0xD201: "HANGUL",
        0xD202: "HANGUL",
        0xD203: "HANGUL",
        0xD204: "HANGUL",
        0xD205: "HANGUL",
        0xD206: "HANGUL",
        0xD207: "HANGUL",
        0xD208: "HANGUL",
        0xD209: "HANGUL",
        0xD20A: "HANGUL",
        0xD20B: "HANGUL",
        0xD20C: "HANGUL",
        0xD20D: "HANGUL",
        0xD20E: "HANGUL",
        0xD20F: "HANGUL",
        0xD210: "HANGUL",
        0xD211: "HANGUL",
        0xD212: "HANGUL",
        0xD213: "HANGUL",
        0xD214: "HANGUL",
        0xD215: "HANGUL",
        0xD216: "HANGUL",
        0xD217: "HANGUL",
        0xD218: "HANGUL",
        0xD219: "HANGUL",
        0xD21A: "HANGUL",
        0xD21B: "HANGUL",
        0xD21C: "HANGUL",
        0xD21D: "HANGUL",
        0xD21E: "HANGUL",
        0xD21F: "HANGUL",
        0xD220: "HANGUL",
        0xD221: "HANGUL",
        0xD222: "HANGUL",
        0xD223: "HANGUL",
        0xD224: "HANGUL",
        0xD225: "HANGUL",
        0xD226: "HANGUL",
        0xD227: "HANGUL",
        0xD228: "HANGUL",
        0xD229: "HANGUL",
        0xD22A: "HANGUL",
        0xD22B: "HANGUL",
        0xD22C: "HANGUL",
        0xD22D: "HANGUL",
        0xD22E: "HANGUL",
        0xD22F: "HANGUL",
        0xD230: "HANGUL",
        0xD231: "HANGUL",
        0xD232: "HANGUL",
        0xD233: "HANGUL",
        0xD234: "HANGUL",
        0xD235: "HANGUL",
        0xD236: "HANGUL",
        0xD237: "HANGUL",
        0xD238: "HANGUL",
        0xD239: "HANGUL",
        0xD23A: "HANGUL",
        0xD23B: "HANGUL",
        0xD23C: "HANGUL",
        0xD23D: "HANGUL",
        0xD23E: "HANGUL",
        0xD23F: "HANGUL",
        0xD240: "HANGUL",
        0xD241: "HANGUL",
        0xD242: "HANGUL",
        0xD243: "HANGUL",
        0xD244: "HANGUL",
        0xD245: "HANGUL",
        0xD246: "HANGUL",
        0xD247: "HANGUL",
        0xD248: "HANGUL",
        0xD249: "HANGUL",
        0xD24A: "HANGUL",
        0xD24B: "HANGUL",
        0xD24C: "HANGUL",
        0xD24D: "HANGUL",
        0xD24E: "HANGUL",
        0xD24F: "HANGUL",
        0xD250: "HANGUL",
        0xD251: "HANGUL",
        0xD252: "HANGUL",
        0xD253: "HANGUL",
        0xD254: "HANGUL",
        0xD255: "HANGUL",
        0xD256: "HANGUL",
        0xD257: "HANGUL",
        0xD258: "HANGUL",
        0xD259: "HANGUL",
        0xD25A: "HANGUL",
        0xD25B: "HANGUL",
        0xD25C: "HANGUL",
        0xD25D: "HANGUL",
        0xD25E: "HANGUL",
        0xD25F: "HANGUL",
        0xD260: "HANGUL",
        0xD261: "HANGUL",
        0xD262: "HANGUL",
        0xD263: "HANGUL",
        0xD264: "HANGUL",
        0xD265: "HANGUL",
        0xD266: "HANGUL",
        0xD267: "HANGUL",
        0xD268: "HANGUL",
        0xD269: "HANGUL",
        0xD26A: "HANGUL",
        0xD26B: "HANGUL",
        0xD26C: "HANGUL",
        0xD26D: "HANGUL",
        0xD26E: "HANGUL",
        0xD26F: "HANGUL",
        0xD270: "HANGUL",
        0xD271: "HANGUL",
        0xD272: "HANGUL",
        0xD273: "HANGUL",
        0xD274: "HANGUL",
        0xD275: "HANGUL",
        0xD276: "HANGUL",
        0xD277: "HANGUL",
        0xD278: "HANGUL",
        0xD279: "HANGUL",
        0xD27A: "HANGUL",
        0xD27B: "HANGUL",
        0xD27C: "HANGUL",
        0xD27D: "HANGUL",
        0xD27E: "HANGUL",
        0xD27F: "HANGUL",
        0xD280: "HANGUL",
        0xD281: "HANGUL",
        0xD282: "HANGUL",
        0xD283: "HANGUL",
        0xD284: "HANGUL",
        0xD285: "HANGUL",
        0xD286: "HANGUL",
        0xD287: "HANGUL",
        0xD288: "HANGUL",
        0xD289: "HANGUL",
        0xD28A: "HANGUL",
        0xD28B: "HANGUL",
        0xD28C: "HANGUL",
        0xD28D: "HANGUL",
        0xD28E: "HANGUL",
        0xD28F: "HANGUL",
        0xD290: "HANGUL",
        0xD291: "HANGUL",
        0xD292: "HANGUL",
        0xD293: "HANGUL",
        0xD294: "HANGUL",
        0xD295: "HANGUL",
        0xD296: "HANGUL",
        0xD297: "HANGUL",
        0xD298: "HANGUL",
        0xD299: "HANGUL",
        0xD29A: "HANGUL",
        0xD29B: "HANGUL",
        0xD29C: "HANGUL",
        0xD29D: "HANGUL",
        0xD29E: "HANGUL",
        0xD29F: "HANGUL",
        0xD2A0: "HANGUL",
        0xD2A1: "HANGUL",
        0xD2A2: "HANGUL",
        0xD2A3: "HANGUL",
        0xD2A4: "HANGUL",
        0xD2A5: "HANGUL",
        0xD2A6: "HANGUL",
        0xD2A7: "HANGUL",
        0xD2A8: "HANGUL",
        0xD2A9: "HANGUL",
        0xD2AA: "HANGUL",
        0xD2AB: "HANGUL",
        0xD2AC: "HANGUL",
        0xD2AD: "HANGUL",
        0xD2AE: "HANGUL",
        0xD2AF: "HANGUL",
        0xD2B0: "HANGUL",
        0xD2B1: "HANGUL",
        0xD2B2: "HANGUL",
        0xD2B3: "HANGUL",
        0xD2B4: "HANGUL",
        0xD2B5: "HANGUL",
        0xD2B6: "HANGUL",
        0xD2B7: "HANGUL",
        0xD2B8: "HANGUL",
        0xD2B9: "HANGUL",
        0xD2BA: "HANGUL",
        0xD2BB: "HANGUL",
        0xD2BC: "HANGUL",
        0xD2BD: "HANGUL",
        0xD2BE: "HANGUL",
        0xD2BF: "HANGUL",
        0xD2C0: "HANGUL",
        0xD2C1: "HANGUL",
        0xD2C2: "HANGUL",
        0xD2C3: "HANGUL",
        0xD2C4: "HANGUL",
        0xD2C5: "HANGUL",
        0xD2C6: "HANGUL",
        0xD2C7: "HANGUL",
        0xD2C8: "HANGUL",
        0xD2C9: "HANGUL",
        0xD2CA: "HANGUL",
        0xD2CB: "HANGUL",
        0xD2CC: "HANGUL",
        0xD2CD: "HANGUL",
        0xD2CE: "HANGUL",
        0xD2CF: "HANGUL",
        0xD2D0: "HANGUL",
        0xD2D1: "HANGUL",
        0xD2D2: "HANGUL",
        0xD2D3: "HANGUL",
        0xD2D4: "HANGUL",
        0xD2D5: "HANGUL",
        0xD2D6: "HANGUL",
        0xD2D7: "HANGUL",
        0xD2D8: "HANGUL",
        0xD2D9: "HANGUL",
        0xD2DA: "HANGUL",
        0xD2DB: "HANGUL",
        0xD2DC: "HANGUL",
        0xD2DD: "HANGUL",
        0xD2DE: "HANGUL",
        0xD2DF: "HANGUL",
        0xD2E0: "HANGUL",
        0xD2E1: "HANGUL",
        0xD2E2: "HANGUL",
        0xD2E3: "HANGUL",
        0xD2E4: "HANGUL",
        0xD2E5: "HANGUL",
        0xD2E6: "HANGUL",
        0xD2E7: "HANGUL",
        0xD2E8: "HANGUL",
        0xD2E9: "HANGUL",
        0xD2EA: "HANGUL",
        0xD2EB: "HANGUL",
        0xD2EC: "HANGUL",
        0xD2ED: "HANGUL",
        0xD2EE: "HANGUL",
        0xD2EF: "HANGUL",
        0xD2F0: "HANGUL",
        0xD2F1: "HANGUL",
        0xD2F2: "HANGUL",
        0xD2F3: "HANGUL",
        0xD2F4: "HANGUL",
        0xD2F5: "HANGUL",
        0xD2F6: "HANGUL",
        0xD2F7: "HANGUL",
        0xD2F8: "HANGUL",
        0xD2F9: "HANGUL",
        0xD2FA: "HANGUL",
        0xD2FB: "HANGUL",
        0xD2FC: "HANGUL",
        0xD2FD: "HANGUL",
        0xD2FE: "HANGUL",
        0xD2FF: "HANGUL",
        0xD300: "HANGUL",
        0xD301: "HANGUL",
        0xD302: "HANGUL",
        0xD303: "HANGUL",
        0xD304: "HANGUL",
        0xD305: "HANGUL",
        0xD306: "HANGUL",
        0xD307: "HANGUL",
        0xD308: "HANGUL",
        0xD309: "HANGUL",
        0xD30A: "HANGUL",
        0xD30B: "HANGUL",
        0xD30C: "HANGUL",
        0xD30D: "HANGUL",
        0xD30E: "HANGUL",
        0xD30F: "HANGUL",
        0xD310: "HANGUL",
        0xD311: "HANGUL",
        0xD312: "HANGUL",
        0xD313: "HANGUL",
        0xD314: "HANGUL",
        0xD315: "HANGUL",
        0xD316: "HANGUL",
        0xD317: "HANGUL",
        0xD318: "HANGUL",
        0xD319: "HANGUL",
        0xD31A: "HANGUL",
        0xD31B: "HANGUL",
        0xD31C: "HANGUL",
        0xD31D: "HANGUL",
        0xD31E: "HANGUL",
        0xD31F: "HANGUL",
        0xD320: "HANGUL",
        0xD321: "HANGUL",
        0xD322: "HANGUL",
        0xD323: "HANGUL",
        0xD324: "HANGUL",
        0xD325: "HANGUL",
        0xD326: "HANGUL",
        0xD327: "HANGUL",
        0xD328: "HANGUL",
        0xD329: "HANGUL",
        0xD32A: "HANGUL",
        0xD32B: "HANGUL",
        0xD32C: "HANGUL",
        0xD32D: "HANGUL",
        0xD32E: "HANGUL",
        0xD32F: "HANGUL",
        0xD330: "HANGUL",
        0xD331: "HANGUL",
        0xD332: "HANGUL",
        0xD333: "HANGUL",
        0xD334: "HANGUL",
        0xD335: "HANGUL",
        0xD336: "HANGUL",
        0xD337: "HANGUL",
        0xD338: "HANGUL",
        0xD339: "HANGUL",
        0xD33A: "HANGUL",
        0xD33B: "HANGUL",
        0xD33C: "HANGUL",
        0xD33D: "HANGUL",
        0xD33E: "HANGUL",
        0xD33F: "HANGUL",
        0xD340: "HANGUL",
        0xD341: "HANGUL",
        0xD342: "HANGUL",
        0xD343: "HANGUL",
        0xD344: "HANGUL",
        0xD345: "HANGUL",
        0xD346: "HANGUL",
        0xD347: "HANGUL",
        0xD348: "HANGUL",
        0xD349: "HANGUL",
        0xD34A: "HANGUL",
        0xD34B: "HANGUL",
        0xD34C: "HANGUL",
        0xD34D: "HANGUL",
        0xD34E: "HANGUL",
        0xD34F: "HANGUL",
        0xD350: "HANGUL",
        0xD351: "HANGUL",
        0xD352: "HANGUL",
        0xD353: "HANGUL",
        0xD354: "HANGUL",
        0xD355: "HANGUL",
        0xD356: "HANGUL",
        0xD357: "HANGUL",
        0xD358: "HANGUL",
        0xD359: "HANGUL",
        0xD35A: "HANGUL",
        0xD35B: "HANGUL",
        0xD35C: "HANGUL",
        0xD35D: "HANGUL",
        0xD35E: "HANGUL",
        0xD35F: "HANGUL",
        0xD360: "HANGUL",
        0xD361: "HANGUL",
        0xD362: "HANGUL",
        0xD363: "HANGUL",
        0xD364: "HANGUL",
        0xD365: "HANGUL",
        0xD366: "HANGUL",
        0xD367: "HANGUL",
        0xD368: "HANGUL",
        0xD369: "HANGUL",
        0xD36A: "HANGUL",
        0xD36B: "HANGUL",
        0xD36C: "HANGUL",
        0xD36D: "HANGUL",
        0xD36E: "HANGUL",
        0xD36F: "HANGUL",
        0xD370: "HANGUL",
        0xD371: "HANGUL",
        0xD372: "HANGUL",
        0xD373: "HANGUL",
        0xD374: "HANGUL",
        0xD375: "HANGUL",
        0xD376: "HANGUL",
        0xD377: "HANGUL",
        0xD378: "HANGUL",
        0xD379: "HANGUL",
        0xD37A: "HANGUL",
        0xD37B: "HANGUL",
        0xD37C: "HANGUL",
        0xD37D: "HANGUL",
        0xD37E: "HANGUL",
        0xD37F: "HANGUL",
        0xD380: "HANGUL",
        0xD381: "HANGUL",
        0xD382: "HANGUL",
        0xD383: "HANGUL",
        0xD384: "HANGUL",
        0xD385: "HANGUL",
        0xD386: "HANGUL",
        0xD387: "HANGUL",
        0xD388: "HANGUL",
        0xD389: "HANGUL",
        0xD38A: "HANGUL",
        0xD38B: "HANGUL",
        0xD38C: "HANGUL",
        0xD38D: "HANGUL",
        0xD38E: "HANGUL",
        0xD38F: "HANGUL",
        0xD390: "HANGUL",
        0xD391: "HANGUL",
        0xD392: "HANGUL",
        0xD393: "HANGUL",
        0xD394: "HANGUL",
        0xD395: "HANGUL",
        0xD396: "HANGUL",
        0xD397: "HANGUL",
        0xD398: "HANGUL",
        0xD399: "HANGUL",
        0xD39A: "HANGUL",
        0xD39B: "HANGUL",
        0xD39C: "HANGUL",
        0xD39D: "HANGUL",
        0xD39E: "HANGUL",
        0xD39F: "HANGUL",
        0xD3A0: "HANGUL",
        0xD3A1: "HANGUL",
        0xD3A2: "HANGUL",
        0xD3A3: "HANGUL",
        0xD3A4: "HANGUL",
        0xD3A5: "HANGUL",
        0xD3A6: "HANGUL",
        0xD3A7: "HANGUL",
        0xD3A8: "HANGUL",
        0xD3A9: "HANGUL",
        0xD3AA: "HANGUL",
        0xD3AB: "HANGUL",
        0xD3AC: "HANGUL",
        0xD3AD: "HANGUL",
        0xD3AE: "HANGUL",
        0xD3AF: "HANGUL",
        0xD3B0: "HANGUL",
        0xD3B1: "HANGUL",
        0xD3B2: "HANGUL",
        0xD3B3: "HANGUL",
        0xD3B4: "HANGUL",
        0xD3B5: "HANGUL",
        0xD3B6: "HANGUL",
        0xD3B7: "HANGUL",
        0xD3B8: "HANGUL",
        0xD3B9: "HANGUL",
        0xD3BA: "HANGUL",
        0xD3BB: "HANGUL",
        0xD3BC: "HANGUL",
        0xD3BD: "HANGUL",
        0xD3BE: "HANGUL",
        0xD3BF: "HANGUL",
        0xD3C0: "HANGUL",
        0xD3C1: "HANGUL",
        0xD3C2: "HANGUL",
        0xD3C3: "HANGUL",
        0xD3C4: "HANGUL",
        0xD3C5: "HANGUL",
        0xD3C6: "HANGUL",
        0xD3C7: "HANGUL",
        0xD3C8: "HANGUL",
        0xD3C9: "HANGUL",
        0xD3CA: "HANGUL",
        0xD3CB: "HANGUL",
        0xD3CC: "HANGUL",
        0xD3CD: "HANGUL",
        0xD3CE: "HANGUL",
        0xD3CF: "HANGUL",
        0xD3D0: "HANGUL",
        0xD3D1: "HANGUL",
        0xD3D2: "HANGUL",
        0xD3D3: "HANGUL",
        0xD3D4: "HANGUL",
        0xD3D5: "HANGUL",
        0xD3D6: "HANGUL",
        0xD3D7: "HANGUL",
        0xD3D8: "HANGUL",
        0xD3D9: "HANGUL",
        0xD3DA: "HANGUL",
        0xD3DB: "HANGUL",
        0xD3DC: "HANGUL",
        0xD3DD: "HANGUL",
        0xD3DE: "HANGUL",
        0xD3DF: "HANGUL",
        0xD3E0: "HANGUL",
        0xD3E1: "HANGUL",
        0xD3E2: "HANGUL",
        0xD3E3: "HANGUL",
        0xD3E4: "HANGUL",
        0xD3E5: "HANGUL",
        0xD3E6: "HANGUL",
        0xD3E7: "HANGUL",
        0xD3E8: "HANGUL",
        0xD3E9: "HANGUL",
        0xD3EA: "HANGUL",
        0xD3EB: "HANGUL",
        0xD3EC: "HANGUL",
        0xD3ED: "HANGUL",
        0xD3EE: "HANGUL",
        0xD3EF: "HANGUL",
        0xD3F0: "HANGUL",
        0xD3F1: "HANGUL",
        0xD3F2: "HANGUL",
        0xD3F3: "HANGUL",
        0xD3F4: "HANGUL",
        0xD3F5: "HANGUL",
        0xD3F6: "HANGUL",
        0xD3F7: "HANGUL",
        0xD3F8: "HANGUL",
        0xD3F9: "HANGUL",
        0xD3FA: "HANGUL",
        0xD3FB: "HANGUL",
        0xD3FC: "HANGUL",
        0xD3FD: "HANGUL",
        0xD3FE: "HANGUL",
        0xD3FF: "HANGUL",
        0xD400: "HANGUL",
        0xD401: "HANGUL",
        0xD402: "HANGUL",
        0xD403: "HANGUL",
        0xD404: "HANGUL",
        0xD405: "HANGUL",
        0xD406: "HANGUL",
        0xD407: "HANGUL",
        0xD408: "HANGUL",
        0xD409: "HANGUL",
        0xD40A: "HANGUL",
        0xD40B: "HANGUL",
        0xD40C: "HANGUL",
        0xD40D: "HANGUL",
        0xD40E: "HANGUL",
        0xD40F: "HANGUL",
        0xD410: "HANGUL",
        0xD411: "HANGUL",
        0xD412: "HANGUL",
        0xD413: "HANGUL",
        0xD414: "HANGUL",
        0xD415: "HANGUL",
        0xD416: "HANGUL",
        0xD417: "HANGUL",
        0xD418: "HANGUL",
        0xD419: "HANGUL",
        0xD41A: "HANGUL",
        0xD41B: "HANGUL",
        0xD41C: "HANGUL",
        0xD41D: "HANGUL",
        0xD41E: "HANGUL",
        0xD41F: "HANGUL",
        0xD420: "HANGUL",
        0xD421: "HANGUL",
        0xD422: "HANGUL",
        0xD423: "HANGUL",
        0xD424: "HANGUL",
        0xD425: "HANGUL",
        0xD426: "HANGUL",
        0xD427: "HANGUL",
        0xD428: "HANGUL",
        0xD429: "HANGUL",
        0xD42A: "HANGUL",
        0xD42B: "HANGUL",
        0xD42C: "HANGUL",
        0xD42D: "HANGUL",
        0xD42E: "HANGUL",
        0xD42F: "HANGUL",
        0xD430: "HANGUL",
        0xD431: "HANGUL",
        0xD432: "HANGUL",
        0xD433: "HANGUL",
        0xD434: "HANGUL",
        0xD435: "HANGUL",
        0xD436: "HANGUL",
        0xD437: "HANGUL",
        0xD438: "HANGUL",
        0xD439: "HANGUL",
        0xD43A: "HANGUL",
        0xD43B: "HANGUL",
        0xD43C: "HANGUL",
        0xD43D: "HANGUL",
        0xD43E: "HANGUL",
        0xD43F: "HANGUL",
        0xD440: "HANGUL",
        0xD441: "HANGUL",
        0xD442: "HANGUL",
        0xD443: "HANGUL",
        0xD444: "HANGUL",
        0xD445: "HANGUL",
        0xD446: "HANGUL",
        0xD447: "HANGUL",
        0xD448: "HANGUL",
        0xD449: "HANGUL",
        0xD44A: "HANGUL",
        0xD44B: "HANGUL",
        0xD44C: "HANGUL",
        0xD44D: "HANGUL",
        0xD44E: "HANGUL",
        0xD44F: "HANGUL",
        0xD450: "HANGUL",
        0xD451: "HANGUL",
        0xD452: "HANGUL",
        0xD453: "HANGUL",
        0xD454: "HANGUL",
        0xD455: "HANGUL",
        0xD456: "HANGUL",
        0xD457: "HANGUL",
        0xD458: "HANGUL",
        0xD459: "HANGUL",
        0xD45A: "HANGUL",
        0xD45B: "HANGUL",
        0xD45C: "HANGUL",
        0xD45D: "HANGUL",
        0xD45E: "HANGUL",
        0xD45F: "HANGUL",
        0xD460: "HANGUL",
        0xD461: "HANGUL",
        0xD462: "HANGUL",
        0xD463: "HANGUL",
        0xD464: "HANGUL",
        0xD465: "HANGUL",
        0xD466: "HANGUL",
        0xD467: "HANGUL",
        0xD468: "HANGUL",
        0xD469: "HANGUL",
        0xD46A: "HANGUL",
        0xD46B: "HANGUL",
        0xD46C: "HANGUL",
        0xD46D: "HANGUL",
        0xD46E: "HANGUL",
        0xD46F: "HANGUL",
        0xD470: "HANGUL",
        0xD471: "HANGUL",
        0xD472: "HANGUL",
        0xD473: "HANGUL",
        0xD474: "HANGUL",
        0xD475: "HANGUL",
        0xD476: "HANGUL",
        0xD477: "HANGUL",
        0xD478: "HANGUL",
        0xD479: "HANGUL",
        0xD47A: "HANGUL",
        0xD47B: "HANGUL",
        0xD47C: "HANGUL",
        0xD47D: "HANGUL",
        0xD47E: "HANGUL",
        0xD47F: "HANGUL",
        0xD480: "HANGUL",
        0xD481: "HANGUL",
        0xD482: "HANGUL",
        0xD483: "HANGUL",
        0xD484: "HANGUL",
        0xD485: "HANGUL",
        0xD486: "HANGUL",
        0xD487: "HANGUL",
        0xD488: "HANGUL",
        0xD489: "HANGUL",
        0xD48A: "HANGUL",
        0xD48B: "HANGUL",
        0xD48C: "HANGUL",
        0xD48D: "HANGUL",
        0xD48E: "HANGUL",
        0xD48F: "HANGUL",
        0xD490: "HANGUL",
        0xD491: "HANGUL",
        0xD492: "HANGUL",
        0xD493: "HANGUL",
        0xD494: "HANGUL",
        0xD495: "HANGUL",
        0xD496: "HANGUL",
        0xD497: "HANGUL",
        0xD498: "HANGUL",
        0xD499: "HANGUL",
        0xD49A: "HANGUL",
        0xD49B: "HANGUL",
        0xD49C: "HANGUL",
        0xD49D: "HANGUL",
        0xD49E: "HANGUL",
        0xD49F: "HANGUL",
        0xD4A0: "HANGUL",
        0xD4A1: "HANGUL",
        0xD4A2: "HANGUL",
        0xD4A3: "HANGUL",
        0xD4A4: "HANGUL",
        0xD4A5: "HANGUL",
        0xD4A6: "HANGUL",
        0xD4A7: "HANGUL",
        0xD4A8: "HANGUL",
        0xD4A9: "HANGUL",
        0xD4AA: "HANGUL",
        0xD4AB: "HANGUL",
        0xD4AC: "HANGUL",
        0xD4AD: "HANGUL",
        0xD4AE: "HANGUL",
        0xD4AF: "HANGUL",
        0xD4B0: "HANGUL",
        0xD4B1: "HANGUL",
        0xD4B2: "HANGUL",
        0xD4B3: "HANGUL",
        0xD4B4: "HANGUL",
        0xD4B5: "HANGUL",
        0xD4B6: "HANGUL",
        0xD4B7: "HANGUL",
        0xD4B8: "HANGUL",
        0xD4B9: "HANGUL",
        0xD4BA: "HANGUL",
        0xD4BB: "HANGUL",
        0xD4BC: "HANGUL",
        0xD4BD: "HANGUL",
        0xD4BE: "HANGUL",
        0xD4BF: "HANGUL",
        0xD4C0: "HANGUL",
        0xD4C1: "HANGUL",
        0xD4C2: "HANGUL",
        0xD4C3: "HANGUL",
        0xD4C4: "HANGUL",
        0xD4C5: "HANGUL",
        0xD4C6: "HANGUL",
        0xD4C7: "HANGUL",
        0xD4C8: "HANGUL",
        0xD4C9: "HANGUL",
        0xD4CA: "HANGUL",
        0xD4CB: "HANGUL",
        0xD4CC: "HANGUL",
        0xD4CD: "HANGUL",
        0xD4CE: "HANGUL",
        0xD4CF: "HANGUL",
        0xD4D0: "HANGUL",
        0xD4D1: "HANGUL",
        0xD4D2: "HANGUL",
        0xD4D3: "HANGUL",
        0xD4D4: "HANGUL",
        0xD4D5: "HANGUL",
        0xD4D6: "HANGUL",
        0xD4D7: "HANGUL",
        0xD4D8: "HANGUL",
        0xD4D9: "HANGUL",
        0xD4DA: "HANGUL",
        0xD4DB: "HANGUL",
        0xD4DC: "HANGUL",
        0xD4DD: "HANGUL",
        0xD4DE: "HANGUL",
        0xD4DF: "HANGUL",
        0xD4E0: "HANGUL",
        0xD4E1: "HANGUL",
        0xD4E2: "HANGUL",
        0xD4E3: "HANGUL",
        0xD4E4: "HANGUL",
        0xD4E5: "HANGUL",
        0xD4E6: "HANGUL",
        0xD4E7: "HANGUL",
        0xD4E8: "HANGUL",
        0xD4E9: "HANGUL",
        0xD4EA: "HANGUL",
        0xD4EB: "HANGUL",
        0xD4EC: "HANGUL",
        0xD4ED: "HANGUL",
        0xD4EE: "HANGUL",
        0xD4EF: "HANGUL",
        0xD4F0: "HANGUL",
        0xD4F1: "HANGUL",
        0xD4F2: "HANGUL",
        0xD4F3: "HANGUL",
        0xD4F4: "HANGUL",
        0xD4F5: "HANGUL",
        0xD4F6: "HANGUL",
        0xD4F7: "HANGUL",
        0xD4F8: "HANGUL",
        0xD4F9: "HANGUL",
        0xD4FA: "HANGUL",
        0xD4FB: "HANGUL",
        0xD4FC: "HANGUL",
        0xD4FD: "HANGUL",
        0xD4FE: "HANGUL",
        0xD4FF: "HANGUL",
        0xD500: "HANGUL",
        0xD501: "HANGUL",
        0xD502: "HANGUL",
        0xD503: "HANGUL",
        0xD504: "HANGUL",
        0xD505: "HANGUL",
        0xD506: "HANGUL",
        0xD507: "HANGUL",
        0xD508: "HANGUL",
        0xD509: "HANGUL",
        0xD50A: "HANGUL",
        0xD50B: "HANGUL",
        0xD50C: "HANGUL",
        0xD50D: "HANGUL",
        0xD50E: "HANGUL",
        0xD50F: "HANGUL",
        0xD510: "HANGUL",
        0xD511: "HANGUL",
        0xD512: "HANGUL",
        0xD513: "HANGUL",
        0xD514: "HANGUL",
        0xD515: "HANGUL",
        0xD516: "HANGUL",
        0xD517: "HANGUL",
        0xD518: "HANGUL",
        0xD519: "HANGUL",
        0xD51A: "HANGUL",
        0xD51B: "HANGUL",
        0xD51C: "HANGUL",
        0xD51D: "HANGUL",
        0xD51E: "HANGUL",
        0xD51F: "HANGUL",
        0xD520: "HANGUL",
        0xD521: "HANGUL",
        0xD522: "HANGUL",
        0xD523: "HANGUL",
        0xD524: "HANGUL",
        0xD525: "HANGUL",
        0xD526: "HANGUL",
        0xD527: "HANGUL",
        0xD528: "HANGUL",
        0xD529: "HANGUL",
        0xD52A: "HANGUL",
        0xD52B: "HANGUL",
        0xD52C: "HANGUL",
        0xD52D: "HANGUL",
        0xD52E: "HANGUL",
        0xD52F: "HANGUL",
        0xD530: "HANGUL",
        0xD531: "HANGUL",
        0xD532: "HANGUL",
        0xD533: "HANGUL",
        0xD534: "HANGUL",
        0xD535: "HANGUL",
        0xD536: "HANGUL",
        0xD537: "HANGUL",
        0xD538: "HANGUL",
        0xD539: "HANGUL",
        0xD53A: "HANGUL",
        0xD53B: "HANGUL",
        0xD53C: "HANGUL",
        0xD53D: "HANGUL",
        0xD53E: "HANGUL",
        0xD53F: "HANGUL",
        0xD540: "HANGUL",
        0xD541: "HANGUL",
        0xD542: "HANGUL",
        0xD543: "HANGUL",
        0xD544: "HANGUL",
        0xD545: "HANGUL",
        0xD546: "HANGUL",
        0xD547: "HANGUL",
        0xD548: "HANGUL",
        0xD549: "HANGUL",
        0xD54A: "HANGUL",
        0xD54B: "HANGUL",
        0xD54C: "HANGUL",
        0xD54D: "HANGUL",
        0xD54E: "HANGUL",
        0xD54F: "HANGUL",
        0xD550: "HANGUL",
        0xD551: "HANGUL",
        0xD552: "HANGUL",
        0xD553: "HANGUL",
        0xD554: "HANGUL",
        0xD555: "HANGUL",
        0xD556: "HANGUL",
        0xD557: "HANGUL",
        0xD558: "HANGUL",
        0xD559: "HANGUL",
        0xD55A: "HANGUL",
        0xD55B: "HANGUL",
        0xD55C: "HANGUL",
        0xD55D: "HANGUL",
        0xD55E: "HANGUL",
        0xD55F: "HANGUL",
        0xD560: "HANGUL",
        0xD561: "HANGUL",
        0xD562: "HANGUL",
        0xD563: "HANGUL",
        0xD564: "HANGUL",
        0xD565: "HANGUL",
        0xD566: "HANGUL",
        0xD567: "HANGUL",
        0xD568: "HANGUL",
        0xD569: "HANGUL",
        0xD56A: "HANGUL",
        0xD56B: "HANGUL",
        0xD56C: "HANGUL",
        0xD56D: "HANGUL",
        0xD56E: "HANGUL",
        0xD56F: "HANGUL",
        0xD570: "HANGUL",
        0xD571: "HANGUL",
        0xD572: "HANGUL",
        0xD573: "HANGUL",
        0xD574: "HANGUL",
        0xD575: "HANGUL",
        0xD576: "HANGUL",
        0xD577: "HANGUL",
        0xD578: "HANGUL",
        0xD579: "HANGUL",
        0xD57A: "HANGUL",
        0xD57B: "HANGUL",
        0xD57C: "HANGUL",
        0xD57D: "HANGUL",
        0xD57E: "HANGUL",
        0xD57F: "HANGUL",
        0xD580: "HANGUL",
        0xD581: "HANGUL",
        0xD582: "HANGUL",
        0xD583: "HANGUL",
        0xD584: "HANGUL",
        0xD585: "HANGUL",
        0xD586: "HANGUL",
        0xD587: "HANGUL",
        0xD588: "HANGUL",
        0xD589: "HANGUL",
        0xD58A: "HANGUL",
        0xD58B: "HANGUL",
        0xD58C: "HANGUL",
        0xD58D: "HANGUL",
        0xD58E: "HANGUL",
        0xD58F: "HANGUL",
        0xD590: "HANGUL",
        0xD591: "HANGUL",
        0xD592: "HANGUL",
        0xD593: "HANGUL",
        0xD594: "HANGUL",
        0xD595: "HANGUL",
        0xD596: "HANGUL",
        0xD597: "HANGUL",
        0xD598: "HANGUL",
        0xD599: "HANGUL",
        0xD59A: "HANGUL",
        0xD59B: "HANGUL",
        0xD59C: "HANGUL",
        0xD59D: "HANGUL",
        0xD59E: "HANGUL",
        0xD59F: "HANGUL",
        0xD5A0: "HANGUL",
        0xD5A1: "HANGUL",
        0xD5A2: "HANGUL",
        0xD5A3: "HANGUL",
        0xD5A4: "HANGUL",
        0xD5A5: "HANGUL",
        0xD5A6: "HANGUL",
        0xD5A7: "HANGUL",
        0xD5A8: "HANGUL",
        0xD5A9: "HANGUL",
        0xD5AA: "HANGUL",
        0xD5AB: "HANGUL",
        0xD5AC: "HANGUL",
        0xD5AD: "HANGUL",
        0xD5AE: "HANGUL",
        0xD5AF: "HANGUL",
        0xD5B0: "HANGUL",
        0xD5B1: "HANGUL",
        0xD5B2: "HANGUL",
        0xD5B3: "HANGUL",
        0xD5B4: "HANGUL",
        0xD5B5: "HANGUL",
        0xD5B6: "HANGUL",
        0xD5B7: "HANGUL",
        0xD5B8: "HANGUL",
        0xD5B9: "HANGUL",
        0xD5BA: "HANGUL",
        0xD5BB: "HANGUL",
        0xD5BC: "HANGUL",
        0xD5BD: "HANGUL",
        0xD5BE: "HANGUL",
        0xD5BF: "HANGUL",
        0xD5C0: "HANGUL",
        0xD5C1: "HANGUL",
        0xD5C2: "HANGUL",
        0xD5C3: "HANGUL",
        0xD5C4: "HANGUL",
        0xD5C5: "HANGUL",
        0xD5C6: "HANGUL",
        0xD5C7: "HANGUL",
        0xD5C8: "HANGUL",
        0xD5C9: "HANGUL",
        0xD5CA: "HANGUL",
        0xD5CB: "HANGUL",
        0xD5CC: "HANGUL",
        0xD5CD: "HANGUL",
        0xD5CE: "HANGUL",
        0xD5CF: "HANGUL",
        0xD5D0: "HANGUL",
        0xD5D1: "HANGUL",
        0xD5D2: "HANGUL",
        0xD5D3: "HANGUL",
        0xD5D4: "HANGUL",
        0xD5D5: "HANGUL",
        0xD5D6: "HANGUL",
        0xD5D7: "HANGUL",
        0xD5D8: "HANGUL",
        0xD5D9: "HANGUL",
        0xD5DA: "HANGUL",
        0xD5DB: "HANGUL",
        0xD5DC: "HANGUL",
        0xD5DD: "HANGUL",
        0xD5DE: "HANGUL",
        0xD5DF: "HANGUL",
        0xD5E0: "HANGUL",
        0xD5E1: "HANGUL",
        0xD5E2: "HANGUL",
        0xD5E3: "HANGUL",
        0xD5E4: "HANGUL",
        0xD5E5: "HANGUL",
        0xD5E6: "HANGUL",
        0xD5E7: "HANGUL",
        0xD5E8: "HANGUL",
        0xD5E9: "HANGUL",
        0xD5EA: "HANGUL",
        0xD5EB: "HANGUL",
        0xD5EC: "HANGUL",
        0xD5ED: "HANGUL",
        0xD5EE: "HANGUL",
        0xD5EF: "HANGUL",
        0xD5F0: "HANGUL",
        0xD5F1: "HANGUL",
        0xD5F2: "HANGUL",
        0xD5F3: "HANGUL",
        0xD5F4: "HANGUL",
        0xD5F5: "HANGUL",
        0xD5F6: "HANGUL",
        0xD5F7: "HANGUL",
        0xD5F8: "HANGUL",
        0xD5F9: "HANGUL",
        0xD5FA: "HANGUL",
        0xD5FB: "HANGUL",
        0xD5FC: "HANGUL",
        0xD5FD: "HANGUL",
        0xD5FE: "HANGUL",
        0xD5FF: "HANGUL",
        0xD600: "HANGUL",
        0xD601: "HANGUL",
        0xD602: "HANGUL",
        0xD603: "HANGUL",
        0xD604: "HANGUL",
        0xD605: "HANGUL",
        0xD606: "HANGUL",
        0xD607: "HANGUL",
        0xD608: "HANGUL",
        0xD609: "HANGUL",
        0xD60A: "HANGUL",
        0xD60B: "HANGUL",
        0xD60C: "HANGUL",
        0xD60D: "HANGUL",
        0xD60E: "HANGUL",
        0xD60F: "HANGUL",
        0xD610: "HANGUL",
        0xD611: "HANGUL",
        0xD612: "HANGUL",
        0xD613: "HANGUL",
        0xD614: "HANGUL",
        0xD615: "HANGUL",
        0xD616: "HANGUL",
        0xD617: "HANGUL",
        0xD618: "HANGUL",
        0xD619: "HANGUL",
        0xD61A: "HANGUL",
        0xD61B: "HANGUL",
        0xD61C: "HANGUL",
        0xD61D: "HANGUL",
        0xD61E: "HANGUL",
        0xD61F: "HANGUL",
        0xD620: "HANGUL",
        0xD621: "HANGUL",
        0xD622: "HANGUL",
        0xD623: "HANGUL",
        0xD624: "HANGUL",
        0xD625: "HANGUL",
        0xD626: "HANGUL",
        0xD627: "HANGUL",
        0xD628: "HANGUL",
        0xD629: "HANGUL",
        0xD62A: "HANGUL",
        0xD62B: "HANGUL",
        0xD62C: "HANGUL",
        0xD62D: "HANGUL",
        0xD62E: "HANGUL",
        0xD62F: "HANGUL",
        0xD630: "HANGUL",
        0xD631: "HANGUL",
        0xD632: "HANGUL",
        0xD633: "HANGUL",
        0xD634: "HANGUL",
        0xD635: "HANGUL",
        0xD636: "HANGUL",
        0xD637: "HANGUL",
        0xD638: "HANGUL",
        0xD639: "HANGUL",
        0xD63A: "HANGUL",
        0xD63B: "HANGUL",
        0xD63C: "HANGUL",
        0xD63D: "HANGUL",
        0xD63E: "HANGUL",
        0xD63F: "HANGUL",
        0xD640: "HANGUL",
        0xD641: "HANGUL",
        0xD642: "HANGUL",
        0xD643: "HANGUL",
        0xD644: "HANGUL",
        0xD645: "HANGUL",
        0xD646: "HANGUL",
        0xD647: "HANGUL",
        0xD648: "HANGUL",
        0xD649: "HANGUL",
        0xD64A: "HANGUL",
        0xD64B: "HANGUL",
        0xD64C: "HANGUL",
        0xD64D: "HANGUL",
        0xD64E: "HANGUL",
        0xD64F: "HANGUL",
        0xD650: "HANGUL",
        0xD651: "HANGUL",
        0xD652: "HANGUL",
        0xD653: "HANGUL",
        0xD654: "HANGUL",
        0xD655: "HANGUL",
        0xD656: "HANGUL",
        0xD657: "HANGUL",
        0xD658: "HANGUL",
        0xD659: "HANGUL",
        0xD65A: "HANGUL",
        0xD65B: "HANGUL",
        0xD65C: "HANGUL",
        0xD65D: "HANGUL",
        0xD65E: "HANGUL",
        0xD65F: "HANGUL",
        0xD660: "HANGUL",
        0xD661: "HANGUL",
        0xD662: "HANGUL",
        0xD663: "HANGUL",
        0xD664: "HANGUL",
        0xD665: "HANGUL",
        0xD666: "HANGUL",
        0xD667: "HANGUL",
        0xD668: "HANGUL",
        0xD669: "HANGUL",
        0xD66A: "HANGUL",
        0xD66B: "HANGUL",
        0xD66C: "HANGUL",
        0xD66D: "HANGUL",
        0xD66E: "HANGUL",
        0xD66F: "HANGUL",
        0xD670: "HANGUL",
        0xD671: "HANGUL",
        0xD672: "HANGUL",
        0xD673: "HANGUL",
        0xD674: "HANGUL",
        0xD675: "HANGUL",
        0xD676: "HANGUL",
        0xD677: "HANGUL",
        0xD678: "HANGUL",
        0xD679: "HANGUL",
        0xD67A: "HANGUL",
        0xD67B: "HANGUL",
        0xD67C: "HANGUL",
        0xD67D: "HANGUL",
        0xD67E: "HANGUL",
        0xD67F: "HANGUL",
        0xD680: "HANGUL",
        0xD681: "HANGUL",
        0xD682: "HANGUL",
        0xD683: "HANGUL",
        0xD684: "HANGUL",
        0xD685: "HANGUL",
        0xD686: "HANGUL",
        0xD687: "HANGUL",
        0xD688: "HANGUL",
        0xD689: "HANGUL",
        0xD68A: "HANGUL",
        0xD68B: "HANGUL",
        0xD68C: "HANGUL",
        0xD68D: "HANGUL",
        0xD68E: "HANGUL",
        0xD68F: "HANGUL",
        0xD690: "HANGUL",
        0xD691: "HANGUL",
        0xD692: "HANGUL",
        0xD693: "HANGUL",
        0xD694: "HANGUL",
        0xD695: "HANGUL",
        0xD696: "HANGUL",
        0xD697: "HANGUL",
        0xD698: "HANGUL",
        0xD699: "HANGUL",
        0xD69A: "HANGUL",
        0xD69B: "HANGUL",
        0xD69C: "HANGUL",
        0xD69D: "HANGUL",
        0xD69E: "HANGUL",
        0xD69F: "HANGUL",
        0xD6A0: "HANGUL",
        0xD6A1: "HANGUL",
        0xD6A2: "HANGUL",
        0xD6A3: "HANGUL",
        0xD6A4: "HANGUL",
        0xD6A5: "HANGUL",
        0xD6A6: "HANGUL",
        0xD6A7: "HANGUL",
        0xD6A8: "HANGUL",
        0xD6A9: "HANGUL",
        0xD6AA: "HANGUL",
        0xD6AB: "HANGUL",
        0xD6AC: "HANGUL",
        0xD6AD: "HANGUL",
        0xD6AE: "HANGUL",
        0xD6AF: "HANGUL",
        0xD6B0: "HANGUL",
        0xD6B1: "HANGUL",
        0xD6B2: "HANGUL",
        0xD6B3: "HANGUL",
        0xD6B4: "HANGUL",
        0xD6B5: "HANGUL",
        0xD6B6: "HANGUL",
        0xD6B7: "HANGUL",
        0xD6B8: "HANGUL",
        0xD6B9: "HANGUL",
        0xD6BA: "HANGUL",
        0xD6BB: "HANGUL",
        0xD6BC: "HANGUL",
        0xD6BD: "HANGUL",
        0xD6BE: "HANGUL",
        0xD6BF: "HANGUL",
        0xD6C0: "HANGUL",
        0xD6C1: "HANGUL",
        0xD6C2: "HANGUL",
        0xD6C3: "HANGUL",
        0xD6C4: "HANGUL",
        0xD6C5: "HANGUL",
        0xD6C6: "HANGUL",
        0xD6C7: "HANGUL",
        0xD6C8: "HANGUL",
        0xD6C9: "HANGUL",
        0xD6CA: "HANGUL",
        0xD6CB: "HANGUL",
        0xD6CC: "HANGUL",
        0xD6CD: "HANGUL",
        0xD6CE: "HANGUL",
        0xD6CF: "HANGUL",
        0xD6D0: "HANGUL",
        0xD6D1: "HANGUL",
        0xD6D2: "HANGUL",
        0xD6D3: "HANGUL",
        0xD6D4: "HANGUL",
        0xD6D5: "HANGUL",
        0xD6D6: "HANGUL",
        0xD6D7: "HANGUL",
        0xD6D8: "HANGUL",
        0xD6D9: "HANGUL",
        0xD6DA: "HANGUL",
        0xD6DB: "HANGUL",
        0xD6DC: "HANGUL",
        0xD6DD: "HANGUL",
        0xD6DE: "HANGUL",
        0xD6DF: "HANGUL",
        0xD6E0: "HANGUL",
        0xD6E1: "HANGUL",
        0xD6E2: "HANGUL",
        0xD6E3: "HANGUL",
        0xD6E4: "HANGUL",
        0xD6E5: "HANGUL",
        0xD6E6: "HANGUL",
        0xD6E7: "HANGUL",
        0xD6E8: "HANGUL",
        0xD6E9: "HANGUL",
        0xD6EA: "HANGUL",
        0xD6EB: "HANGUL",
        0xD6EC: "HANGUL",
        0xD6ED: "HANGUL",
        0xD6EE: "HANGUL",
        0xD6EF: "HANGUL",
        0xD6F0: "HANGUL",
        0xD6F1: "HANGUL",
        0xD6F2: "HANGUL",
        0xD6F3: "HANGUL",
        0xD6F4: "HANGUL",
        0xD6F5: "HANGUL",
        0xD6F6: "HANGUL",
        0xD6F7: "HANGUL",
        0xD6F8: "HANGUL",
        0xD6F9: "HANGUL",
        0xD6FA: "HANGUL",
        0xD6FB: "HANGUL",
        0xD6FC: "HANGUL",
        0xD6FD: "HANGUL",
        0xD6FE: "HANGUL",
        0xD6FF: "HANGUL",
        0xD700: "HANGUL",
        0xD701: "HANGUL",
        0xD702: "HANGUL",
        0xD703: "HANGUL",
        0xD704: "HANGUL",
        0xD705: "HANGUL",
        0xD706: "HANGUL",
        0xD707: "HANGUL",
        0xD708: "HANGUL",
        0xD709: "HANGUL",
        0xD70A: "HANGUL",
        0xD70B: "HANGUL",
        0xD70C: "HANGUL",
        0xD70D: "HANGUL",
        0xD70E: "HANGUL",
        0xD70F: "HANGUL",
        0xD710: "HANGUL",
        0xD711: "HANGUL",
        0xD712: "HANGUL",
        0xD713: "HANGUL",
        0xD714: "HANGUL",
        0xD715: "HANGUL",
        0xD716: "HANGUL",
        0xD717: "HANGUL",
        0xD718: "HANGUL",
        0xD719: "HANGUL",
        0xD71A: "HANGUL",
        0xD71B: "HANGUL",
        0xD71C: "HANGUL",
        0xD71D: "HANGUL",
        0xD71E: "HANGUL",
        0xD71F: "HANGUL",
        0xD720: "HANGUL",
        0xD721: "HANGUL",
        0xD722: "HANGUL",
        0xD723: "HANGUL",
        0xD724: "HANGUL",
        0xD725: "HANGUL",
        0xD726: "HANGUL",
        0xD727: "HANGUL",
        0xD728: "HANGUL",
        0xD729: "HANGUL",
        0xD72A: "HANGUL",
        0xD72B: "HANGUL",
        0xD72C: "HANGUL",
        0xD72D: "HANGUL",
        0xD72E: "HANGUL",
        0xD72F: "HANGUL",
        0xD730: "HANGUL",
        0xD731: "HANGUL",
        0xD732: "HANGUL",
        0xD733: "HANGUL",
        0xD734: "HANGUL",
        0xD735: "HANGUL",
        0xD736: "HANGUL",
        0xD737: "HANGUL",
        0xD738: "HANGUL",
        0xD739: "HANGUL",
        0xD73A: "HANGUL",
        0xD73B: "HANGUL",
        0xD73C: "HANGUL",
        0xD73D: "HANGUL",
        0xD73E: "HANGUL",
        0xD73F: "HANGUL",
        0xD740: "HANGUL",
        0xD741: "HANGUL",
        0xD742: "HANGUL",
        0xD743: "HANGUL",
        0xD744: "HANGUL",
        0xD745: "HANGUL",
        0xD746: "HANGUL",
        0xD747: "HANGUL",
        0xD748: "HANGUL",
        0xD749: "HANGUL",
        0xD74A: "HANGUL",
        0xD74B: "HANGUL",
        0xD74C: "HANGUL",
        0xD74D: "HANGUL",
        0xD74E: "HANGUL",
        0xD74F: "HANGUL",
        0xD750: "HANGUL",
        0xD751: "HANGUL",
        0xD752: "HANGUL",
        0xD753: "HANGUL",
        0xD754: "HANGUL",
        0xD755: "HANGUL",
        0xD756: "HANGUL",
        0xD757: "HANGUL",
        0xD758: "HANGUL",
        0xD759: "HANGUL",
        0xD75A: "HANGUL",
        0xD75B: "HANGUL",
        0xD75C: "HANGUL",
        0xD75D: "HANGUL",
        0xD75E: "HANGUL",
        0xD75F: "HANGUL",
        0xD760: "HANGUL",
        0xD761: "HANGUL",
        0xD762: "HANGUL",
        0xD763: "HANGUL",
        0xD764: "HANGUL",
        0xD765: "HANGUL",
        0xD766: "HANGUL",
        0xD767: "HANGUL",
        0xD768: "HANGUL",
        0xD769: "HANGUL",
        0xD76A: "HANGUL",
        0xD76B: "HANGUL",
        0xD76C: "HANGUL",
        0xD76D: "HANGUL",
        0xD76E: "HANGUL",
        0xD76F: "HANGUL",
        0xD770: "HANGUL",
        0xD771: "HANGUL",
        0xD772: "HANGUL",
        0xD773: "HANGUL",
        0xD774: "HANGUL",
        0xD775: "HANGUL",
        0xD776: "HANGUL",
        0xD777: "HANGUL",
        0xD778: "HANGUL",
        0xD779: "HANGUL",
        0xD77A: "HANGUL",
        0xD77B: "HANGUL",
        0xD77C: "HANGUL",
        0xD77D: "HANGUL",
        0xD77E: "HANGUL",
        0xD77F: "HANGUL",
        0xD780: "HANGUL",
        0xD781: "HANGUL",
        0xD782: "HANGUL",
        0xD783: "HANGUL",
        0xD784: "HANGUL",
        0xD785: "HANGUL",
        0xD786: "HANGUL",
        0xD787: "HANGUL",
        0xD788: "HANGUL",
        0xD789: "HANGUL",
        0xD78A: "HANGUL",
        0xD78B: "HANGUL",
        0xD78C: "HANGUL",
        0xD78D: "HANGUL",
        0xD78E: "HANGUL",
        0xD78F: "HANGUL",
        0xD790: "HANGUL",
        0xD791: "HANGUL",
        0xD792: "HANGUL",
        0xD793: "HANGUL",
        0xD794: "HANGUL",
        0xD795: "HANGUL",
        0xD796: "HANGUL",
        0xD797: "HANGUL",
        0xD798: "HANGUL",
        0xD799: "HANGUL",
        0xD79A: "HANGUL",
        0xD79B: "HANGUL",
        0xD79C: "HANGUL",
        0xD79D: "HANGUL",
        0xD79E: "HANGUL",
        0xD79F: "HANGUL",
        0xD7A0: "HANGUL",
        0xD7A1: "HANGUL",
        0xD7A2: "HANGUL",
        0xD7A3: "HANGUL",
        0x1100: "HANGUL",
        0x1101: "HANGUL",
        0x1102: "HANGUL",
        0x1103: "HANGUL",
        0x1104: "HANGUL",
        0x1105: "HANGUL",
        0x1106: "HANGUL",
        0x1107: "HANGUL",
        0x1108: "HANGUL",
        0x1109: "HANGUL",
        0x110A: "HANGUL",
        0x110B: "HANGUL",
        0x110C: "HANGUL",
        0x110D: "HANGUL",
        0x110E: "HANGUL",
        0x110F: "HANGUL",
        0x1110: "HANGUL",
        0x1111: "HANGUL",
        0x1112: "HANGUL",
        0x1113: "HANGUL",
        0x1114: "HANGUL",
        0x1115: "HANGUL",
        0x1116: "HANGUL",
        0x1117: "HANGUL",
        0x1118: "HANGUL",
        0x1119: "HANGUL",
        0x111A: "HANGUL",
        0x111B: "HANGUL",
        0x111C: "HANGUL",
        0x111D: "HANGUL",
        0x111E: "HANGUL",
        0x111F: "HANGUL",
        0x1120: "HANGUL",
        0x1121: "HANGUL",
        0x1122: "HANGUL",
        0x1123: "HANGUL",
        0x1124: "HANGUL",
        0x1125: "HANGUL",
        0x1126: "HANGUL",
        0x1127: "HANGUL",
        0x1128: "HANGUL",
        0x1129: "HANGUL",
        0x112A: "HANGUL",
        0x112B: "HANGUL",
        0x112C: "HANGUL",
        0x112D: "HANGUL",
        0x112E: "HANGUL",
        0x112F: "HANGUL",
        0x1130: "HANGUL",
        0x1131: "HANGUL",
        0x1132: "HANGUL",
        0x1133: "HANGUL",
        0x1134: "HANGUL",
        0x1135: "HANGUL",
        0x1136: "HANGUL",
        0x1137: "HANGUL",
        0x1138: "HANGUL",
        0x1139: "HANGUL",
        0x113A: "HANGUL",
        0x113B: "HANGUL",
        0x113C: "HANGUL",
        0x113D: "HANGUL",
        0x113E: "HANGUL",
        0x113F: "HANGUL",
        0x1140: "HANGUL",
        0x1141: "HANGUL",
        0x1142: "HANGUL",
        0x1143: "HANGUL",
        0x1144: "HANGUL",
        0x1145: "HANGUL",
        0x1146: "HANGUL",
        0x1147: "HANGUL",
        0x1148: "HANGUL",
        0x1149: "HANGUL",
        0x114A: "HANGUL",
        0x114B: "HANGUL",
        0x114C: "HANGUL",
        0x114D: "HANGUL",
        0x114E: "HANGUL",
        0x114F: "HANGUL",
        0x1150: "HANGUL",
        0x1151: "HANGUL",
        0x1152: "HANGUL",
        0x1153: "HANGUL",
        0x1154: "HANGUL",
        0x1155: "HANGUL",
        0x1156: "HANGUL",
        0x1157: "HANGUL",
        0x1158: "HANGUL",
        0x1159: "HANGUL",
        0x115A: "HANGUL",
        0x115B: "HANGUL",
        0x115C: "HANGUL",
        0x115D: "HANGUL",
        0x115E: "HANGUL",
        0x115F: "HANGUL",
        0x1160: "HANGUL",
        0x1161: "HANGUL",
        0x1162: "HANGUL",
        0x1163: "HANGUL",
        0x1164: "HANGUL",
        0x1165: "HANGUL",
        0x1166: "HANGUL",
        0x1167: "HANGUL",
        0x1168: "HANGUL",
        0x1169: "HANGUL",
        0x116A: "HANGUL",
        0x116B: "HANGUL",
        0x116C: "HANGUL",
        0x116D: "HANGUL",
        0x116E: "HANGUL",
        0x116F: "HANGUL",
        0x1170: "HANGUL",
        0x1171: "HANGUL",
        0x1172: "HANGUL",
        0x1173: "HANGUL",
        0x1174: "HANGUL",
        0x1175: "HANGUL",
        0x1176: "HANGUL",
        0x1177: "HANGUL",
        0x1178: "HANGUL",
        0x1179: "HANGUL",
        0x117A: "HANGUL",
        0x117B: "HANGUL",
        0x117C: "HANGUL",
        0x117D: "HANGUL",
        0x117E: "HANGUL",
        0x117F: "HANGUL",
        0x1180: "HANGUL",
        0x1181: "HANGUL",
        0x1182: "HANGUL",
        0x1183: "HANGUL",
        0x1184: "HANGUL",
        0x1185: "HANGUL",
        0x1186: "HANGUL",
        0x1187: "HANGUL",
        0x1188: "HANGUL",
        0x1189: "HANGUL",
        0x118A: "HANGUL",
        0x118B: "HANGUL",
        0x118C: "HANGUL",
        0x118D: "HANGUL",
        0x118E: "HANGUL",
        0x118F: "HANGUL",
        0x1190: "HANGUL",
        0x1191: "HANGUL",
        0x1192: "HANGUL",
        0x1193: "HANGUL",
        0x1194: "HANGUL",
        0x1195: "HANGUL",
        0x1196: "HANGUL",
        0x1197: "HANGUL",
        0x1198: "HANGUL",
        0x1199: "HANGUL",
        0x119A: "HANGUL",
        0x119B: "HANGUL",
        0x119C: "HANGUL",
        0x119D: "HANGUL",
        0x119E: "HANGUL",
        0x119F: "HANGUL",
        0x11A0: "HANGUL",
        0x11A1: "HANGUL",
        0x11A2: "HANGUL",
        0x11A3: "HANGUL",
        0x11A4: "HANGUL",
        0x11A5: "HANGUL",
        0x11A6: "HANGUL",
        0x11A7: "HANGUL",
        0x11A8: "HANGUL",
        0x11A9: "HANGUL",
        0x11AA: "HANGUL",
        0x11AB: "HANGUL",
        0x11AC: "HANGUL",
        0x11AD: "HANGUL",
        0x11AE: "HANGUL",
        0x11AF: "HANGUL",
        0x11B0: "HANGUL",
        0x11B1: "HANGUL",
        0x11B2: "HANGUL",
        0x11B3: "HANGUL",
        0x11B4: "HANGUL",
        0x11B5: "HANGUL",
        0x11B6: "HANGUL",
        0x11B7: "HANGUL",
        0x11B8: "HANGUL",
        0x11B9: "HANGUL",
        0x11BA: "HANGUL",
        0x11BB: "HANGUL",
        0x11BC: "HANGUL",
        0x11BD: "HANGUL",
        0x11BE: "HANGUL",
        0x11BF: "HANGUL",
        0x11C0: "HANGUL",
        0x11C1: "HANGUL",
        0x11C2: "HANGUL",
        0x11C3: "HANGUL",
        0x11C4: "HANGUL",
        0x11C5: "HANGUL",
        0x11C6: "HANGUL",
        0x11C7: "HANGUL",
        0x11C8: "HANGUL",
        0x11C9: "HANGUL",
        0x11CA: "HANGUL",
        0x11CB: "HANGUL",
        0x11CC: "HANGUL",
        0x11CD: "HANGUL",
        0x11CE: "HANGUL",
        0x11CF: "HANGUL",
        0x11D0: "HANGUL",
        0x11D1: "HANGUL",
        0x11D2: "HANGUL",
        0x11D3: "HANGUL",
        0x11D4: "HANGUL",
        0x11D5: "HANGUL",
        0x11D6: "HANGUL",
        0x11D7: "HANGUL",
        0x11D8: "HANGUL",
        0x11D9: "HANGUL",
        0x11DA: "HANGUL",
        0x11DB: "HANGUL",
        0x11DC: "HANGUL",
        0x11DD: "HANGUL",
        0x11DE: "HANGUL",
        0x11DF: "HANGUL",
        0x11E0: "HANGUL",
        0x11E1: "HANGUL",
        0x11E2: "HANGUL",
        0x11E3: "HANGUL",
        0x11E4: "HANGUL",
        0x11E5: "HANGUL",
        0x11E6: "HANGUL",
        0x11E7: "HANGUL",
        0x11E8: "HANGUL",
        0x11E9: "HANGUL",
        0x11EA: "HANGUL",
        0x11EB: "HANGUL",
        0x11EC: "HANGUL",
        0x11ED: "HANGUL",
        0x11EE: "HANGUL",
        0x11EF: "HANGUL",
        0x11F0: "HANGUL",
        0x11F1: "HANGUL",
        0x11F2: "HANGUL",
        0x11F3: "HANGUL",
        0x11F4: "HANGUL",
        0x11F5: "HANGUL",
        0x11F6: "HANGUL",
        0x11F7: "HANGUL",
        0x11F8: "HANGUL",
        0x11F9: "HANGUL",
        0x11FA: "HANGUL",
        0x11FB: "HANGUL",
        0x11FC: "HANGUL",
        0x11FD: "HANGUL",
        0x11FE: "HANGUL",
        0x11FF: "HANGUL",
        0x3130: "HANGUL",
        0x3131: "HANGUL",
        0x3132: "HANGUL",
        0x3133: "HANGUL",
        0x3134: "HANGUL",
        0x3135: "HANGUL",
        0x3136: "HANGUL",
        0x3137: "HANGUL",
        0x3138: "HANGUL",
        0x3139: "HANGUL",
        0x313A: "HANGUL",
        0x313B: "HANGUL",
        0x313C: "HANGUL",
        0x313D: "HANGUL",
        0x313E: "HANGUL",
        0x313F: "HANGUL",
        0x3140: "HANGUL",
        0x3141: "HANGUL",
        0x3142: "HANGUL",
        0x3143: "HANGUL",
        0x3144: "HANGUL",
        0x3145: "HANGUL",
        0x3146: "HANGUL",
        0x3147: "HANGUL",
        0x3148: "HANGUL",
        0x3149: "HANGUL",
        0x314A: "HANGUL",
        0x314B: "HANGUL",
        0x314C: "HANGUL",
        0x314D: "HANGUL",
        0x314E: "HANGUL",
        0x314F: "HANGUL",
        0x3150: "HANGUL",
        0x3151: "HANGUL",
        0x3152: "HANGUL",
        0x3153: "HANGUL",
        0x3154: "HANGUL",
        0x3155: "HANGUL",
        0x3156: "HANGUL",
        0x3157: "HANGUL",
        0x3158: "HANGUL",
        0x3159: "HANGUL",
        0x315A: "HANGUL",
        0x315B: "HANGUL",
        0x315C: "HANGUL",
        0x315D: "HANGUL",
        0x315E: "HANGUL",
        0x315F: "HANGUL",
        0x3160: "HANGUL",
        0x3161: "HANGUL",
        0x3162: "HANGUL",
        0x3163: "HANGUL",
        0x3164: "HANGUL",
        0x3165: "HANGUL",
        0x3166: "HANGUL",
        0x3167: "HANGUL",
        0x3168: "HANGUL",
        0x3169: "HANGUL",
        0x316A: "HANGUL",
        0x316B: "HANGUL",
        0x316C: "HANGUL",
        0x316D: "HANGUL",
        0x316E: "HANGUL",
        0x316F: "HANGUL",
        0x3170: "HANGUL",
        0x3171: "HANGUL",
        0x3172: "HANGUL",
        0x3173: "HANGUL",
        0x3174: "HANGUL",
        0x3175: "HANGUL",
        0x3176: "HANGUL",
        0x3177: "HANGUL",
        0x3178: "HANGUL",
        0x3179: "HANGUL",
        0x317A: "HANGUL",
        0x317B: "HANGUL",
        0x317C: "HANGUL",
        0x317D: "HANGUL",
        0x317E: "HANGUL",
        0x317F: "HANGUL",
        0x3180: "HANGUL",
        0x3181: "HANGUL",
        0x3182: "HANGUL",
        0x3183: "HANGUL",
        0x3184: "HANGUL",
        0x3185: "HANGUL",
        0x3186: "HANGUL",
        0x3187: "HANGUL",
        0x3188: "HANGUL",
        0x3189: "HANGUL",
        0x318A: "HANGUL",
        0x318B: "HANGUL",
        0x318C: "HANGUL",
        0x318D: "HANGUL",
        0x318E: "HANGUL",
        0x318F: "HANGUL",
        0x3041: "HIRAGANA",
        0x3042: "HIRAGANA",
        0x3043: "HIRAGANA",
        0x3044: "HIRAGANA",
        0x3045: "HIRAGANA",
        0x3046: "HIRAGANA",
        0x3047: "HIRAGANA",
        0x3048: "HIRAGANA",
        0x3049: "HIRAGANA",
        0x304A: "HIRAGANA",
        0x304B: "HIRAGANA",
        0x304C: "HIRAGANA",
        0x304D: "HIRAGANA",
        0x304E: "HIRAGANA",
        0x304F: "HIRAGANA",
        0x3050: "HIRAGANA",
        0x3051: "HIRAGANA",
        0x3052: "HIRAGANA",
        0x3053: "HIRAGANA",
        0x3054: "HIRAGANA",
        0x3055: "HIRAGANA",
        0x3056: "HIRAGANA",
        0x3057: "HIRAGANA",
        0x3058: "HIRAGANA",
        0x3059: "HIRAGANA",
        0x305A: "HIRAGANA",
        0x305B: "HIRAGANA",
        0x305C: "HIRAGANA",
        0x305D: "HIRAGANA",
        0x305E: "HIRAGANA",
        0x305F: "HIRAGANA",
        0x3060: "HIRAGANA",
        0x3061: "HIRAGANA",
        0x3062: "HIRAGANA",
        0x3063: "HIRAGANA",
        0x3064: "HIRAGANA",
        0x3065: "HIRAGANA",
        0x3066: "HIRAGANA",
        0x3067: "HIRAGANA",
        0x3068: "HIRAGANA",
        0x3069: "HIRAGANA",
        0x306A: "HIRAGANA",
        0x306B: "HIRAGANA",
        0x306C: "HIRAGANA",
        0x306D: "HIRAGANA",
        0x306E: "HIRAGANA",
        0x306F: "HIRAGANA",
        0x3070: "HIRAGANA",
        0x3071: "HIRAGANA",
        0x3072: "HIRAGANA",
        0x3073: "HIRAGANA",
        0x3074: "HIRAGANA",
        0x3075: "HIRAGANA",
        0x3076: "HIRAGANA",
        0x3077: "HIRAGANA",
        0x3078: "HIRAGANA",
        0x3079: "HIRAGANA",
        0x307A: "HIRAGANA",
        0x307B: "HIRAGANA",
        0x307C: "HIRAGANA",
        0x307D: "HIRAGANA",
        0x307E: "HIRAGANA",
        0x307F: "HIRAGANA",
        0x3080: "HIRAGANA",
        0x3081: "HIRAGANA",
        0x3082: "HIRAGANA",
        0x3083: "HIRAGANA",
        0x3084: "HIRAGANA",
        0x3085: "HIRAGANA",
        0x3086: "HIRAGANA",
        0x3087: "HIRAGANA",
        0x3088: "HIRAGANA",
        0x3089: "HIRAGANA",
        0x308A: "HIRAGANA",
        0x308B: "HIRAGANA",
        0x308C: "HIRAGANA",
        0x308D: "HIRAGANA",
        0x308E: "HIRAGANA",
        0x308F: "HIRAGANA",
        0x3090: "HIRAGANA",
        0x3091: "HIRAGANA",
        0x3092: "HIRAGANA",
        0x3093: "HIRAGANA",
        0x3094: "HIRAGANA",
        0x3095: "HIRAGANA",
        0x3096: "HIRAGANA",
        0x3097: "HIRAGANA",
        0x3098: "HIRAGANA",
        0x3099: "HIRAGANA",
        0x309A: "HIRAGANA",
        0x309B: "HIRAGANA",
        0x309C: "HIRAGANA",
        0x309D: "HIRAGANA",
        0x309E: "HIRAGANA",
        0x309F: "HIRAGANA",
        0x30A1: "KATAKANA",
        0x30A2: "KATAKANA",
        0x30A3: "KATAKANA",
        0x30A4: "KATAKANA",
        0x30A5: "KATAKANA",
        0x30A6: "KATAKANA",
        0x30A7: "KATAKANA",
        0x30A8: "KATAKANA",
        0x30A9: "KATAKANA",
        0x30AA: "KATAKANA",
        0x30AB: "KATAKANA",
        0x30AC: "KATAKANA",
        0x30AD: "KATAKANA",
        0x30AE: "KATAKANA",
        0x30AF: "KATAKANA",
        0x30B0: "KATAKANA",
        0x30B1: "KATAKANA",
        0x30B2: "KATAKANA",
        0x30B3: "KATAKANA",
        0x30B4: "KATAKANA",
        0x30B5: "KATAKANA",
        0x30B6: "KATAKANA",
        0x30B7: "KATAKANA",
        0x30B8: "KATAKANA",
        0x30B9: "KATAKANA",
        0x30BA: "KATAKANA",
        0x30BB: "KATAKANA",
        0x30BC: "KATAKANA",
        0x30BD: "KATAKANA",
        0x30BE: "KATAKANA",
        0x30BF: "KATAKANA",
        0x30C0: "KATAKANA",
        0x30C1: "KATAKANA",
        0x30C2: "KATAKANA",
        0x30C3: "KATAKANA",
        0x30C4: "KATAKANA",
        0x30C5: "KATAKANA",
        0x30C6: "KATAKANA",
        0x30C7: "KATAKANA",
        0x30C8: "KATAKANA",
        0x30C9: "KATAKANA",
        0x30CA: "KATAKANA",
        0x30CB: "KATAKANA",
        0x30CC: "KATAKANA",
        0x30CD: "KATAKANA",
        0x30CE: "KATAKANA",
        0x30CF: "KATAKANA",
        0x30D0: "KATAKANA",
        0x30D1: "KATAKANA",
        0x30D2: "KATAKANA",
        0x30D3: "KATAKANA",
        0x30D4: "KATAKANA",
        0x30D5: "KATAKANA",
        0x30D6: "KATAKANA",
        0x30D7: "KATAKANA",
        0x30D8: "KATAKANA",
        0x30D9: "KATAKANA",
        0x30DA: "KATAKANA",
        0x30DB: "KATAKANA",
        0x30DC: "KATAKANA",
        0x30DD: "KATAKANA",
        0x30DE: "KATAKANA",
        0x30DF: "KATAKANA",
        0x30E0: "KATAKANA",
        0x30E1: "KATAKANA",
        0x30E2: "KATAKANA",
        0x30E3: "KATAKANA",
        0x30E4: "KATAKANA",
        0x30E5: "KATAKANA",
        0x30E6: "KATAKANA",
        0x30E7: "KATAKANA",
        0x30E8: "KATAKANA",
        0x30E9: "KATAKANA",
        0x30EA: "KATAKANA",
        0x30EB: "KATAKANA",
        0x30EC: "KATAKANA",
        0x30ED: "KATAKANA",
        0x30EE: "KATAKANA",
        0x30EF: "KATAKANA",
        0x30F0: "KATAKANA",
        0x30F1: "KATAKANA",
        0x30F2: "KATAKANA",
        0x30F3: "KATAKANA",
        0x30F4: "KATAKANA",
        0x30F5: "KATAKANA",
        0x30F6: "KATAKANA",
        0x30F7: "KATAKANA",
        0x30F8: "KATAKANA",
        0x30F9: "KATAKANA",
        0x30FA: "KATAKANA",
        0x30FB: "KATAKANA",
        0x30FD: "KATAKANA",
        0x30FE: "KATAKANA",
        0x30FF: "KATAKANA",
        0x31F0: "KATAKANA",
        0x31F1: "KATAKANA",
        0x31F2: "KATAKANA",
        0x31F3: "KATAKANA",
        0x31F4: "KATAKANA",
        0x31F5: "KATAKANA",
        0x31F6: "KATAKANA",
        0x31F7: "KATAKANA",
        0x31F8: "KATAKANA",
        0x31F9: "KATAKANA",
        0x31FA: "KATAKANA",
        0x31FB: "KATAKANA",
        0x31FC: "KATAKANA",
        0x31FD: "KATAKANA",
        0x31FE: "KATAKANA",
        0x31FF: "KATAKANA",
        0x30FC: "KATAKANA",
        0xFF66: "KATAKANA",
        0xFF67: "KATAKANA",
        0xFF68: "KATAKANA",
        0xFF69: "KATAKANA",
        0xFF6A: "KATAKANA",
        0xFF6B: "KATAKANA",
        0xFF6C: "KATAKANA",
        0xFF6D: "KATAKANA",
        0xFF6E: "KATAKANA",
        0xFF6F: "KATAKANA",
        0xFF70: "KATAKANA",
        0xFF71: "KATAKANA",
        0xFF72: "KATAKANA",
        0xFF73: "KATAKANA",
        0xFF74: "KATAKANA",
        0xFF75: "KATAKANA",
        0xFF76: "KATAKANA",
        0xFF77: "KATAKANA",
        0xFF78: "KATAKANA",
        0xFF79: "KATAKANA",
        0xFF7A: "KATAKANA",
        0xFF7B: "KATAKANA",
        0xFF7C: "KATAKANA",
        0xFF7D: "KATAKANA",
        0xFF7E: "KATAKANA",
        0xFF7F: "KATAKANA",
        0xFF80: "KATAKANA",
        0xFF81: "KATAKANA",
        0xFF82: "KATAKANA",
        0xFF83: "KATAKANA",
        0xFF84: "KATAKANA",
        0xFF85: "KATAKANA",
        0xFF86: "KATAKANA",
        0xFF87: "KATAKANA",
        0xFF88: "KATAKANA",
        0xFF89: "KATAKANA",
        0xFF8A: "KATAKANA",
        0xFF8B: "KATAKANA",
        0xFF8C: "KATAKANA",
        0xFF8D: "KATAKANA",
        0xFF8E: "KATAKANA",
        0xFF8F: "KATAKANA",
        0xFF90: "KATAKANA",
        0xFF91: "KATAKANA",
        0xFF92: "KATAKANA",
        0xFF93: "KATAKANA",
        0xFF94: "KATAKANA",
        0xFF95: "KATAKANA",
        0xFF96: "KATAKANA",
        0xFF97: "KATAKANA",
        0xFF98: "KATAKANA",
        0xFF99: "KATAKANA",
        0xFF9A: "KATAKANA",
        0xFF9B: "KATAKANA",
        0xFF9C: "KATAKANA",
        0xFF9D: "KATAKANA",
        0xFF9E: "KATAKANA",
        0xFF9F: "KATAKANA",
        0x2E80: "HANJA",
        0x2E81: "HANJA",
        0x2E82: "HANJA",
        0x2E83: "HANJA",
        0x2E84: "HANJA",
        0x2E85: "HANJA",
        0x2E86: "HANJA",
        0x2E87: "HANJA",
        0x2E88: "HANJA",
        0x2E89: "HANJA",
        0x2E8A: "HANJA",
        0x2E8B: "HANJA",
        0x2E8C: "HANJA",
        0x2E8D: "HANJA",
        0x2E8E: "HANJA",
        0x2E8F: "HANJA",
        0x2E90: "HANJA",
        0x2E91: "HANJA",
        0x2E92: "HANJA",
        0x2E93: "HANJA",
        0x2E94: "HANJA",
        0x2E95: "HANJA",
        0x2E96: "HANJA",
        0x2E97: "HANJA",
        0x2E98: "HANJA",
        0x2E99: "HANJA",
        0x2E9A: "HANJA",
        0x2E9B: "HANJA",
        0x2E9C: "HANJA",
        0x2E9D: "HANJA",
        0x2E9E: "HANJA",
        0x2E9F: "HANJA",
        0x2EA0: "HANJA",
        0x2EA1: "HANJA",
        0x2EA2: "HANJA",
        0x2EA3: "HANJA",
        0x2EA4: "HANJA",
        0x2EA5: "HANJA",
        0x2EA6: "HANJA",
        0x2EA7: "HANJA",
        0x2EA8: "HANJA",
        0x2EA9: "HANJA",
        0x2EAA: "HANJA",
        0x2EAB: "HANJA",
        0x2EAC: "HANJA",
        0x2EAD: "HANJA",
        0x2EAE: "HANJA",
        0x2EAF: "HANJA",
        0x2EB0: "HANJA",
        0x2EB1: "HANJA",
        0x2EB2: "HANJA",
        0x2EB3: "HANJA",
        0x2EB4: "HANJA",
        0x2EB5: "HANJA",
        0x2EB6: "HANJA",
        0x2EB7: "HANJA",
        0x2EB8: "HANJA",
        0x2EB9: "HANJA",
        0x2EBA: "HANJA",
        0x2EBB: "HANJA",
        0x2EBC: "HANJA",
        0x2EBD: "HANJA",
        0x2EBE: "HANJA",
        0x2EBF: "HANJA",
        0x2EC0: "HANJA",
        0x2EC1: "HANJA",
        0x2EC2: "HANJA",
        0x2EC3: "HANJA",
        0x2EC4: "HANJA",
        0x2EC5: "HANJA",
        0x2EC6: "HANJA",
        0x2EC7: "HANJA",
        0x2EC8: "HANJA",
        0x2EC9: "HANJA",
        0x2ECA: "HANJA",
        0x2ECB: "HANJA",
        0x2ECC: "HANJA",
        0x2ECD: "HANJA",
        0x2ECE: "HANJA",
        0x2ECF: "HANJA",
        0x2ED0: "HANJA",
        0x2ED1: "HANJA",
        0x2ED2: "HANJA",
        0x2ED3: "HANJA",
        0x2ED4: "HANJA",
        0x2ED5: "HANJA",
        0x2ED6: "HANJA",
        0x2ED7: "HANJA",
        0x2ED8: "HANJA",
        0x2ED9: "HANJA",
        0x2EDA: "HANJA",
        0x2EDB: "HANJA",
        0x2EDC: "HANJA",
        0x2EDD: "HANJA",
        0x2EDE: "HANJA",
        0x2EDF: "HANJA",
        0x2EE0: "HANJA",
        0x2EE1: "HANJA",
        0x2EE2: "HANJA",
        0x2EE3: "HANJA",
        0x2EE4: "HANJA",
        0x2EE5: "HANJA",
        0x2EE6: "HANJA",
        0x2EE7: "HANJA",
        0x2EE8: "HANJA",
        0x2EE9: "HANJA",
        0x2EEA: "HANJA",
        0x2EEB: "HANJA",
        0x2EEC: "HANJA",
        0x2EED: "HANJA",
        0x2EEE: "HANJA",
        0x2EEF: "HANJA",
        0x2EF0: "HANJA",
        0x2EF1: "HANJA",
        0x2EF2: "HANJA",
        0x2EF3: "HANJA",
        0x3400: "HANJA",
        0x3401: "HANJA",
        0x3402: "HANJA",
        0x3403: "HANJA",
        0x3404: "HANJA",
        0x3405: "HANJA",
        0x3406: "HANJA",
        0x3407: "HANJA",
        0x3408: "HANJA",
        0x3409: "HANJA",
        0x340A: "HANJA",
        0x340B: "HANJA",
        0x340C: "HANJA",
        0x340D: "HANJA",
        0x340E: "HANJA",
        0x340F: "HANJA",
        0x3410: "HANJA",
        0x3411: "HANJA",
        0x3412: "HANJA",
        0x3413: "HANJA",
        0x3414: "HANJA",
        0x3415: "HANJA",
        0x3416: "HANJA",
        0x3417: "HANJA",
        0x3418: "HANJA",
        0x3419: "HANJA",
        0x341A: "HANJA",
        0x341B: "HANJA",
        0x341C: "HANJA",
        0x341D: "HANJA",
        0x341E: "HANJA",
        0x341F: "HANJA",
        0x3420: "HANJA",
        0x3421: "HANJA",
        0x3422: "HANJA",
        0x3423: "HANJA",
        0x3424: "HANJA",
        0x3425: "HANJA",
        0x3426: "HANJA",
        0x3427: "HANJA",
        0x3428: "HANJA",
        0x3429: "HANJA",
        0x342A: "HANJA",
        0x342B: "HANJA",
        0x342C: "HANJA",
        0x342D: "HANJA",
        0x342E: "HANJA",
        0x342F: "HANJA",
        0x3430: "HANJA",
        0x3431: "HANJA",
        0x3432: "HANJA",
        0x3433: "HANJA",
        0x3434: "HANJA",
        0x3435: "HANJA",
        0x3436: "HANJA",
        0x3437: "HANJA",
        0x3438: "HANJA",
        0x3439: "HANJA",
        0x343A: "HANJA",
        0x343B: "HANJA",
        0x343C: "HANJA",
        0x343D: "HANJA",
        0x343E: "HANJA",
        0x343F: "HANJA",
        0x3440: "HANJA",
        0x3441: "HANJA",
        0x3442: "HANJA",
        0x3443: "HANJA",
        0x3444: "HANJA",
        0x3445: "HANJA",
        0x3446: "HANJA",
        0x3447: "HANJA",
        0x3448: "HANJA",
        0x3449: "HANJA",
        0x344A: "HANJA",
        0x344B: "HANJA",
        0x344C: "HANJA",
        0x344D: "HANJA",
        0x344E: "HANJA",
        0x344F: "HANJA",
        0x3450: "HANJA",
        0x3451: "HANJA",
        0x3452: "HANJA",
        0x3453: "HANJA",
        0x3454: "HANJA",
        0x3455: "HANJA",
        0x3456: "HANJA",
        0x3457: "HANJA",
        0x3458: "HANJA",
        0x3459: "HANJA",
        0x345A: "HANJA",
        0x345B: "HANJA",
        0x345C: "HANJA",
        0x345D: "HANJA",
        0x345E: "HANJA",
        0x345F: "HANJA",
        0x3460: "HANJA",
        0x3461: "HANJA",
        0x3462: "HANJA",
        0x3463: "HANJA",
        0x3464: "HANJA",
        0x3465: "HANJA",
        0x3466: "HANJA",
        0x3467: "HANJA",
        0x3468: "HANJA",
        0x3469: "HANJA",
        0x346A: "HANJA",
        0x346B: "HANJA",
        0x346C: "HANJA",
        0x346D: "HANJA",
        0x346E: "HANJA",
        0x346F: "HANJA",
        0x3470: "HANJA",
        0x3471: "HANJA",
        0x3472: "HANJA",
        0x3473: "HANJA",
        0x3474: "HANJA",
        0x3475: "HANJA",
        0x3476: "HANJA",
        0x3477: "HANJA",
        0x3478: "HANJA",
        0x3479: "HANJA",
        0x347A: "HANJA",
        0x347B: "HANJA",
        0x347C: "HANJA",
        0x347D: "HANJA",
        0x347E: "HANJA",
        0x347F: "HANJA",
        0x3480: "HANJA",
        0x3481: "HANJA",
        0x3482: "HANJA",
        0x3483: "HANJA",
        0x3484: "HANJA",
        0x3485: "HANJA",
        0x3486: "HANJA",
        0x3487: "HANJA",
        0x3488: "HANJA",
        0x3489: "HANJA",
        0x348A: "HANJA",
        0x348B: "HANJA",
        0x348C: "HANJA",
        0x348D: "HANJA",
        0x348E: "HANJA",
        0x348F: "HANJA",
        0x3490: "HANJA",
        0x3491: "HANJA",
        0x3492: "HANJA",
        0x3493: "HANJA",
        0x3494: "HANJA",
        0x3495: "HANJA",
        0x3496: "HANJA",
        0x3497: "HANJA",
        0x3498: "HANJA",
        0x3499: "HANJA",
        0x349A: "HANJA",
        0x349B: "HANJA",
        0x349C: "HANJA",
        0x349D: "HANJA",
        0x349E: "HANJA",
        0x349F: "HANJA",
        0x34A0: "HANJA",
        0x34A1: "HANJA",
        0x34A2: "HANJA",
        0x34A3: "HANJA",
        0x34A4: "HANJA",
        0x34A5: "HANJA",
        0x34A6: "HANJA",
        0x34A7: "HANJA",
        0x34A8: "HANJA",
        0x34A9: "HANJA",
        0x34AA: "HANJA",
        0x34AB: "HANJA",
        0x34AC: "HANJA",
        0x34AD: "HANJA",
        0x34AE: "HANJA",
        0x34AF: "HANJA",
        0x34B0: "HANJA",
        0x34B1: "HANJA",
        0x34B2: "HANJA",
        0x34B3: "HANJA",
        0x34B4: "HANJA",
        0x34B5: "HANJA",
        0x34B6: "HANJA",
        0x34B7: "HANJA",
        0x34B8: "HANJA",
        0x34B9: "HANJA",
        0x34BA: "HANJA",
        0x34BB: "HANJA",
        0x34BC: "HANJA",
        0x34BD: "HANJA",
        0x34BE: "HANJA",
        0x34BF: "HANJA",
        0x34C0: "HANJA",
        0x34C1: "HANJA",
        0x34C2: "HANJA",
        0x34C3: "HANJA",
        0x34C4: "HANJA",
        0x34C5: "HANJA",
        0x34C6: "HANJA",
        0x34C7: "HANJA",
        0x34C8: "HANJA",
        0x34C9: "HANJA",
        0x34CA: "HANJA",
        0x34CB: "HANJA",
        0x34CC: "HANJA",
        0x34CD: "HANJA",
        0x34CE: "HANJA",
        0x34CF: "HANJA",
        0x34D0: "HANJA",
        0x34D1: "HANJA",
        0x34D2: "HANJA",
        0x34D3: "HANJA",
        0x34D4: "HANJA",
        0x34D5: "HANJA",
        0x34D6: "HANJA",
        0x34D7: "HANJA",
        0x34D8: "HANJA",
        0x34D9: "HANJA",
        0x34DA: "HANJA",
        0x34DB: "HANJA",
        0x34DC: "HANJA",
        0x34DD: "HANJA",
        0x34DE: "HANJA",
        0x34DF: "HANJA",
        0x34E0: "HANJA",
        0x34E1: "HANJA",
        0x34E2: "HANJA",
        0x34E3: "HANJA",
        0x34E4: "HANJA",
        0x34E5: "HANJA",
        0x34E6: "HANJA",
        0x34E7: "HANJA",
        0x34E8: "HANJA",
        0x34E9: "HANJA",
        0x34EA: "HANJA",
        0x34EB: "HANJA",
        0x34EC: "HANJA",
        0x34ED: "HANJA",
        0x34EE: "HANJA",
        0x34EF: "HANJA",
        0x34F0: "HANJA",
        0x34F1: "HANJA",
        0x34F2: "HANJA",
        0x34F3: "HANJA",
        0x34F4: "HANJA",
        0x34F5: "HANJA",
        0x34F6: "HANJA",
        0x34F7: "HANJA",
        0x34F8: "HANJA",
        0x34F9: "HANJA",
        0x34FA: "HANJA",
        0x34FB: "HANJA",
        0x34FC: "HANJA",
        0x34FD: "HANJA",
        0x34FE: "HANJA",
        0x34FF: "HANJA",
        0x3500: "HANJA",
        0x3501: "HANJA",
        0x3502: "HANJA",
        0x3503: "HANJA",
        0x3504: "HANJA",
        0x3505: "HANJA",
        0x3506: "HANJA",
        0x3507: "HANJA",
        0x3508: "HANJA",
        0x3509: "HANJA",
        0x350A: "HANJA",
        0x350B: "HANJA",
        0x350C: "HANJA",
        0x350D: "HANJA",
        0x350E: "HANJA",
        0x350F: "HANJA",
        0x3510: "HANJA",
        0x3511: "HANJA",
        0x3512: "HANJA",
        0x3513: "HANJA",
        0x3514: "HANJA",
        0x3515: "HANJA",
        0x3516: "HANJA",
        0x3517: "HANJA",
        0x3518: "HANJA",
        0x3519: "HANJA",
        0x351A: "HANJA",
        0x351B: "HANJA",
        0x351C: "HANJA",
        0x351D: "HANJA",
        0x351E: "HANJA",
        0x351F: "HANJA",
        0x3520: "HANJA",
        0x3521: "HANJA",
        0x3522: "HANJA",
        0x3523: "HANJA",
        0x3524: "HANJA",
        0x3525: "HANJA",
        0x3526: "HANJA",
        0x3527: "HANJA",
        0x3528: "HANJA",
        0x3529: "HANJA",
        0x352A: "HANJA",
        0x352B: "HANJA",
        0x352C: "HANJA",
        0x352D: "HANJA",
        0x352E: "HANJA",
        0x352F: "HANJA",
        0x3530: "HANJA",
        0x3531: "HANJA",
        0x3532: "HANJA",
        0x3533: "HANJA",
        0x3534: "HANJA",
        0x3535: "HANJA",
        0x3536: "HANJA",
        0x3537: "HANJA",
        0x3538: "HANJA",
        0x3539: "HANJA",
        0x353A: "HANJA",
        0x353B: "HANJA",
        0x353C: "HANJA",
        0x353D: "HANJA",
        0x353E: "HANJA",
        0x353F: "HANJA",
        0x3540: "HANJA",
        0x3541: "HANJA",
        0x3542: "HANJA",
        0x3543: "HANJA",
        0x3544: "HANJA",
        0x3545: "HANJA",
        0x3546: "HANJA",
        0x3547: "HANJA",
        0x3548: "HANJA",
        0x3549: "HANJA",
        0x354A: "HANJA",
        0x354B: "HANJA",
        0x354C: "HANJA",
        0x354D: "HANJA",
        0x354E: "HANJA",
        0x354F: "HANJA",
        0x3550: "HANJA",
        0x3551: "HANJA",
        0x3552: "HANJA",
        0x3553: "HANJA",
        0x3554: "HANJA",
        0x3555: "HANJA",
        0x3556: "HANJA",
        0x3557: "HANJA",
        0x3558: "HANJA",
        0x3559: "HANJA",
        0x355A: "HANJA",
        0x355B: "HANJA",
        0x355C: "HANJA",
        0x355D: "HANJA",
        0x355E: "HANJA",
        0x355F: "HANJA",
        0x3560: "HANJA",
        0x3561: "HANJA",
        0x3562: "HANJA",
        0x3563: "HANJA",
        0x3564: "HANJA",
        0x3565: "HANJA",
        0x3566: "HANJA",
        0x3567: "HANJA",
        0x3568: "HANJA",
        0x3569: "HANJA",
        0x356A: "HANJA",
        0x356B: "HANJA",
        0x356C: "HANJA",
        0x356D: "HANJA",
        0x356E: "HANJA",
        0x356F: "HANJA",
        0x3570: "HANJA",
        0x3571: "HANJA",
        0x3572: "HANJA",
        0x3573: "HANJA",
        0x3574: "HANJA",
        0x3575: "HANJA",
        0x3576: "HANJA",
        0x3577: "HANJA",
        0x3578: "HANJA",
        0x3579: "HANJA",
        0x357A: "HANJA",
        0x357B: "HANJA",
        0x357C: "HANJA",
        0x357D: "HANJA",
        0x357E: "HANJA",
        0x357F: "HANJA",
        0x3580: "HANJA",
        0x3581: "HANJA",
        0x3582: "HANJA",
        0x3583: "HANJA",
        0x3584: "HANJA",
        0x3585: "HANJA",
        0x3586: "HANJA",
        0x3587: "HANJA",
        0x3588: "HANJA",
        0x3589: "HANJA",
        0x358A: "HANJA",
        0x358B: "HANJA",
        0x358C: "HANJA",
        0x358D: "HANJA",
        0x358E: "HANJA",
        0x358F: "HANJA",
        0x3590: "HANJA",
        0x3591: "HANJA",
        0x3592: "HANJA",
        0x3593: "HANJA",
        0x3594: "HANJA",
        0x3595: "HANJA",
        0x3596: "HANJA",
        0x3597: "HANJA",
        0x3598: "HANJA",
        0x3599: "HANJA",
        0x359A: "HANJA",
        0x359B: "HANJA",
        0x359C: "HANJA",
        0x359D: "HANJA",
        0x359E: "HANJA",
        0x359F: "HANJA",
        0x35A0: "HANJA",
        0x35A1: "HANJA",
        0x35A2: "HANJA",
        0x35A3: "HANJA",
        0x35A4: "HANJA",
        0x35A5: "HANJA",
        0x35A6: "HANJA",
        0x35A7: "HANJA",
        0x35A8: "HANJA",
        0x35A9: "HANJA",
        0x35AA: "HANJA",
        0x35AB: "HANJA",
        0x35AC: "HANJA",
        0x35AD: "HANJA",
        0x35AE: "HANJA",
        0x35AF: "HANJA",
        0x35B0: "HANJA",
        0x35B1: "HANJA",
        0x35B2: "HANJA",
        0x35B3: "HANJA",
        0x35B4: "HANJA",
        0x35B5: "HANJA",
        0x35B6: "HANJA",
        0x35B7: "HANJA",
        0x35B8: "HANJA",
        0x35B9: "HANJA",
        0x35BA: "HANJA",
        0x35BB: "HANJA",
        0x35BC: "HANJA",
        0x35BD: "HANJA",
        0x35BE: "HANJA",
        0x35BF: "HANJA",
        0x35C0: "HANJA",
        0x35C1: "HANJA",
        0x35C2: "HANJA",
        0x35C3: "HANJA",
        0x35C4: "HANJA",
        0x35C5: "HANJA",
        0x35C6: "HANJA",
        0x35C7: "HANJA",
        0x35C8: "HANJA",
        0x35C9: "HANJA",
        0x35CA: "HANJA",
        0x35CB: "HANJA",
        0x35CC: "HANJA",
        0x35CD: "HANJA",
        0x35CE: "HANJA",
        0x35CF: "HANJA",
        0x35D0: "HANJA",
        0x35D1: "HANJA",
        0x35D2: "HANJA",
        0x35D3: "HANJA",
        0x35D4: "HANJA",
        0x35D5: "HANJA",
        0x35D6: "HANJA",
        0x35D7: "HANJA",
        0x35D8: "HANJA",
        0x35D9: "HANJA",
        0x35DA: "HANJA",
        0x35DB: "HANJA",
        0x35DC: "HANJA",
        0x35DD: "HANJA",
        0x35DE: "HANJA",
        0x35DF: "HANJA",
        0x35E0: "HANJA",
        0x35E1: "HANJA",
        0x35E2: "HANJA",
        0x35E3: "HANJA",
        0x35E4: "HANJA",
        0x35E5: "HANJA",
        0x35E6: "HANJA",
        0x35E7: "HANJA",
        0x35E8: "HANJA",
        0x35E9: "HANJA",
        0x35EA: "HANJA",
        0x35EB: "HANJA",
        0x35EC: "HANJA",
        0x35ED: "HANJA",
        0x35EE: "HANJA",
        0x35EF: "HANJA",
        0x35F0: "HANJA",
        0x35F1: "HANJA",
        0x35F2: "HANJA",
        0x35F3: "HANJA",
        0x35F4: "HANJA",
        0x35F5: "HANJA",
        0x35F6: "HANJA",
        0x35F7: "HANJA",
        0x35F8: "HANJA",
        0x35F9: "HANJA",
        0x35FA: "HANJA",
        0x35FB: "HANJA",
        0x35FC: "HANJA",
        0x35FD: "HANJA",
        0x35FE: "HANJA",
        0x35FF: "HANJA",
        0x3600: "HANJA",
        0x3601: "HANJA",
        0x3602: "HANJA",
        0x3603: "HANJA",
        0x3604: "HANJA",
        0x3605: "HANJA",
        0x3606: "HANJA",
        0x3607: "HANJA",
        0x3608: "HANJA",
        0x3609: "HANJA",
        0x360A: "HANJA",
        0x360B: "HANJA",
        0x360C: "HANJA",
        0x360D: "HANJA",
        0x360E: "HANJA",
        0x360F: "HANJA",
        0x3610: "HANJA",
        0x3611: "HANJA",
        0x3612: "HANJA",
        0x3613: "HANJA",
        0x3614: "HANJA",
        0x3615: "HANJA",
        0x3616: "HANJA",
        0x3617: "HANJA",
        0x3618: "HANJA",
        0x3619: "HANJA",
        0x361A: "HANJA",
        0x361B: "HANJA",
        0x361C: "HANJA",
        0x361D: "HANJA",
        0x361E: "HANJA",
        0x361F: "HANJA",
        0x3620: "HANJA",
        0x3621: "HANJA",
        0x3622: "HANJA",
        0x3623: "HANJA",
        0x3624: "HANJA",
        0x3625: "HANJA",
        0x3626: "HANJA",
        0x3627: "HANJA",
        0x3628: "HANJA",
        0x3629: "HANJA",
        0x362A: "HANJA",
        0x362B: "HANJA",
        0x362C: "HANJA",
        0x362D: "HANJA",
        0x362E: "HANJA",
        0x362F: "HANJA",
        0x3630: "HANJA",
        0x3631: "HANJA",
        0x3632: "HANJA",
        0x3633: "HANJA",
        0x3634: "HANJA",
        0x3635: "HANJA",
        0x3636: "HANJA",
        0x3637: "HANJA",
        0x3638: "HANJA",
        0x3639: "HANJA",
        0x363A: "HANJA",
        0x363B: "HANJA",
        0x363C: "HANJA",
        0x363D: "HANJA",
        0x363E: "HANJA",
        0x363F: "HANJA",
        0x3640: "HANJA",
        0x3641: "HANJA",
        0x3642: "HANJA",
        0x3643: "HANJA",
        0x3644: "HANJA",
        0x3645: "HANJA",
        0x3646: "HANJA",
        0x3647: "HANJA",
        0x3648: "HANJA",
        0x3649: "HANJA",
        0x364A: "HANJA",
        0x364B: "HANJA",
        0x364C: "HANJA",
        0x364D: "HANJA",
        0x364E: "HANJA",
        0x364F: "HANJA",
        0x3650: "HANJA",
        0x3651: "HANJA",
        0x3652: "HANJA",
        0x3653: "HANJA",
        0x3654: "HANJA",
        0x3655: "HANJA",
        0x3656: "HANJA",
        0x3657: "HANJA",
        0x3658: "HANJA",
        0x3659: "HANJA",
        0x365A: "HANJA",
        0x365B: "HANJA",
        0x365C: "HANJA",
        0x365D: "HANJA",
        0x365E: "HANJA",
        0x365F: "HANJA",
        0x3660: "HANJA",
        0x3661: "HANJA",
        0x3662: "HANJA",
        0x3663: "HANJA",
        0x3664: "HANJA",
        0x3665: "HANJA",
        0x3666: "HANJA",
        0x3667: "HANJA",
        0x3668: "HANJA",
        0x3669: "HANJA",
        0x366A: "HANJA",
        0x366B: "HANJA",
        0x366C: "HANJA",
        0x366D: "HANJA",
        0x366E: "HANJA",
        0x366F: "HANJA",
        0x3670: "HANJA",
        0x3671: "HANJA",
        0x3672: "HANJA",
        0x3673: "HANJA",
        0x3674: "HANJA",
        0x3675: "HANJA",
        0x3676: "HANJA",
        0x3677: "HANJA",
        0x3678: "HANJA",
        0x3679: "HANJA",
        0x367A: "HANJA",
        0x367B: "HANJA",
        0x367C: "HANJA",
        0x367D: "HANJA",
        0x367E: "HANJA",
        0x367F: "HANJA",
        0x3680: "HANJA",
        0x3681: "HANJA",
        0x3682: "HANJA",
        0x3683: "HANJA",
        0x3684: "HANJA",
        0x3685: "HANJA",
        0x3686: "HANJA",
        0x3687: "HANJA",
        0x3688: "HANJA",
        0x3689: "HANJA",
        0x368A: "HANJA",
        0x368B: "HANJA",
        0x368C: "HANJA",
        0x368D: "HANJA",
        0x368E: "HANJA",
        0x368F: "HANJA",
        0x3690: "HANJA",
        0x3691: "HANJA",
        0x3692: "HANJA",
        0x3693: "HANJA",
        0x3694: "HANJA",
        0x3695: "HANJA",
        0x3696: "HANJA",
        0x3697: "HANJA",
        0x3698: "HANJA",
        0x3699: "HANJA",
        0x369A: "HANJA",
        0x369B: "HANJA",
        0x369C: "HANJA",
        0x369D: "HANJA",
        0x369E: "HANJA",
        0x369F: "HANJA",
        0x36A0: "HANJA",
        0x36A1: "HANJA",
        0x36A2: "HANJA",
        0x36A3: "HANJA",
        0x36A4: "HANJA",
        0x36A5: "HANJA",
        0x36A6: "HANJA",
        0x36A7: "HANJA",
        0x36A8: "HANJA",
        0x36A9: "HANJA",
        0x36AA: "HANJA",
        0x36AB: "HANJA",
        0x36AC: "HANJA",
        0x36AD: "HANJA",
        0x36AE: "HANJA",
        0x36AF: "HANJA",
        0x36B0: "HANJA",
        0x36B1: "HANJA",
        0x36B2: "HANJA",
        0x36B3: "HANJA",
        0x36B4: "HANJA",
        0x36B5: "HANJA",
        0x36B6: "HANJA",
        0x36B7: "HANJA",
        0x36B8: "HANJA",
        0x36B9: "HANJA",
        0x36BA: "HANJA",
        0x36BB: "HANJA",
        0x36BC: "HANJA",
        0x36BD: "HANJA",
        0x36BE: "HANJA",
        0x36BF: "HANJA",
        0x36C0: "HANJA",
        0x36C1: "HANJA",
        0x36C2: "HANJA",
        0x36C3: "HANJA",
        0x36C4: "HANJA",
        0x36C5: "HANJA",
        0x36C6: "HANJA",
        0x36C7: "HANJA",
        0x36C8: "HANJA",
        0x36C9: "HANJA",
        0x36CA: "HANJA",
        0x36CB: "HANJA",
        0x36CC: "HANJA",
        0x36CD: "HANJA",
        0x36CE: "HANJA",
        0x36CF: "HANJA",
        0x36D0: "HANJA",
        0x36D1: "HANJA",
        0x36D2: "HANJA",
        0x36D3: "HANJA",
        0x36D4: "HANJA",
        0x36D5: "HANJA",
        0x36D6: "HANJA",
        0x36D7: "HANJA",
        0x36D8: "HANJA",
        0x36D9: "HANJA",
        0x36DA: "HANJA",
        0x36DB: "HANJA",
        0x36DC: "HANJA",
        0x36DD: "HANJA",
        0x36DE: "HANJA",
        0x36DF: "HANJA",
        0x36E0: "HANJA",
        0x36E1: "HANJA",
        0x36E2: "HANJA",
        0x36E3: "HANJA",
        0x36E4: "HANJA",
        0x36E5: "HANJA",
        0x36E6: "HANJA",
        0x36E7: "HANJA",
        0x36E8: "HANJA",
        0x36E9: "HANJA",
        0x36EA: "HANJA",
        0x36EB: "HANJA",
        0x36EC: "HANJA",
        0x36ED: "HANJA",
        0x36EE: "HANJA",
        0x36EF: "HANJA",
        0x36F0: "HANJA",
        0x36F1: "HANJA",
        0x36F2: "HANJA",
        0x36F3: "HANJA",
        0x36F4: "HANJA",
        0x36F5: "HANJA",
        0x36F6: "HANJA",
        0x36F7: "HANJA",
        0x36F8: "HANJA",
        0x36F9: "HANJA",
        0x36FA: "HANJA",
        0x36FB: "HANJA",
        0x36FC: "HANJA",
        0x36FD: "HANJA",
        0x36FE: "HANJA",
        0x36FF: "HANJA",
        0x3700: "HANJA",
        0x3701: "HANJA",
        0x3702: "HANJA",
        0x3703: "HANJA",
        0x3704: "HANJA",
        0x3705: "HANJA",
        0x3706: "HANJA",
        0x3707: "HANJA",
        0x3708: "HANJA",
        0x3709: "HANJA",
        0x370A: "HANJA",
        0x370B: "HANJA",
        0x370C: "HANJA",
        0x370D: "HANJA",
        0x370E: "HANJA",
        0x370F: "HANJA",
        0x3710: "HANJA",
        0x3711: "HANJA",
        0x3712: "HANJA",
        0x3713: "HANJA",
        0x3714: "HANJA",
        0x3715: "HANJA",
        0x3716: "HANJA",
        0x3717: "HANJA",
        0x3718: "HANJA",
        0x3719: "HANJA",
        0x371A: "HANJA",
        0x371B: "HANJA",
        0x371C: "HANJA",
        0x371D: "HANJA",
        0x371E: "HANJA",
        0x371F: "HANJA",
        0x3720: "HANJA",
        0x3721: "HANJA",
        0x3722: "HANJA",
        0x3723: "HANJA",
        0x3724: "HANJA",
        0x3725: "HANJA",
        0x3726: "HANJA",
        0x3727: "HANJA",
        0x3728: "HANJA",
        0x3729: "HANJA",
        0x372A: "HANJA",
        0x372B: "HANJA",
        0x372C: "HANJA",
        0x372D: "HANJA",
        0x372E: "HANJA",
        0x372F: "HANJA",
        0x3730: "HANJA",
        0x3731: "HANJA",
        0x3732: "HANJA",
        0x3733: "HANJA",
        0x3734: "HANJA",
        0x3735: "HANJA",
        0x3736: "HANJA",
        0x3737: "HANJA",
        0x3738: "HANJA",
        0x3739: "HANJA",
        0x373A: "HANJA",
        0x373B: "HANJA",
        0x373C: "HANJA",
        0x373D: "HANJA",
        0x373E: "HANJA",
        0x373F: "HANJA",
        0x3740: "HANJA",
        0x3741: "HANJA",
        0x3742: "HANJA",
        0x3743: "HANJA",
        0x3744: "HANJA",
        0x3745: "HANJA",
        0x3746: "HANJA",
        0x3747: "HANJA",
        0x3748: "HANJA",
        0x3749: "HANJA",
        0x374A: "HANJA",
        0x374B: "HANJA",
        0x374C: "HANJA",
        0x374D: "HANJA",
        0x374E: "HANJA",
        0x374F: "HANJA",
        0x3750: "HANJA",
        0x3751: "HANJA",
        0x3752: "HANJA",
        0x3753: "HANJA",
        0x3754: "HANJA",
        0x3755: "HANJA",
        0x3756: "HANJA",
        0x3757: "HANJA",
        0x3758: "HANJA",
        0x3759: "HANJA",
        0x375A: "HANJA",
        0x375B: "HANJA",
        0x375C: "HANJA",
        0x375D: "HANJA",
        0x375E: "HANJA",
        0x375F: "HANJA",
        0x3760: "HANJA",
        0x3761: "HANJA",
        0x3762: "HANJA",
        0x3763: "HANJA",
        0x3764: "HANJA",
        0x3765: "HANJA",
        0x3766: "HANJA",
        0x3767: "HANJA",
        0x3768: "HANJA",
        0x3769: "HANJA",
        0x376A: "HANJA",
        0x376B: "HANJA",
        0x376C: "HANJA",
        0x376D: "HANJA",
        0x376E: "HANJA",
        0x376F: "HANJA",
        0x3770: "HANJA",
        0x3771: "HANJA",
        0x3772: "HANJA",
        0x3773: "HANJA",
        0x3774: "HANJA",
        0x3775: "HANJA",
        0x3776: "HANJA",
        0x3777: "HANJA",
        0x3778: "HANJA",
        0x3779: "HANJA",
        0x377A: "HANJA",
        0x377B: "HANJA",
        0x377C: "HANJA",
        0x377D: "HANJA",
        0x377E: "HANJA",
        0x377F: "HANJA",
        0x3780: "HANJA",
        0x3781: "HANJA",
        0x3782: "HANJA",
        0x3783: "HANJA",
        0x3784: "HANJA",
        0x3785: "HANJA",
        0x3786: "HANJA",
        0x3787: "HANJA",
        0x3788: "HANJA",
        0x3789: "HANJA",
        0x378A: "HANJA",
        0x378B: "HANJA",
        0x378C: "HANJA",
        0x378D: "HANJA",
        0x378E: "HANJA",
        0x378F: "HANJA",
        0x3790: "HANJA",
        0x3791: "HANJA",
        0x3792: "HANJA",
        0x3793: "HANJA",
        0x3794: "HANJA",
        0x3795: "HANJA",
        0x3796: "HANJA",
        0x3797: "HANJA",
        0x3798: "HANJA",
        0x3799: "HANJA",
        0x379A: "HANJA",
        0x379B: "HANJA",
        0x379C: "HANJA",
        0x379D: "HANJA",
        0x379E: "HANJA",
        0x379F: "HANJA",
        0x37A0: "HANJA",
        0x37A1: "HANJA",
        0x37A2: "HANJA",
        0x37A3: "HANJA",
        0x37A4: "HANJA",
        0x37A5: "HANJA",
        0x37A6: "HANJA",
        0x37A7: "HANJA",
        0x37A8: "HANJA",
        0x37A9: "HANJA",
        0x37AA: "HANJA",
        0x37AB: "HANJA",
        0x37AC: "HANJA",
        0x37AD: "HANJA",
        0x37AE: "HANJA",
        0x37AF: "HANJA",
        0x37B0: "HANJA",
        0x37B1: "HANJA",
        0x37B2: "HANJA",
        0x37B3: "HANJA",
        0x37B4: "HANJA",
        0x37B5: "HANJA",
        0x37B6: "HANJA",
        0x37B7: "HANJA",
        0x37B8: "HANJA",
        0x37B9: "HANJA",
        0x37BA: "HANJA",
        0x37BB: "HANJA",
        0x37BC: "HANJA",
        0x37BD: "HANJA",
        0x37BE: "HANJA",
        0x37BF: "HANJA",
        0x37C0: "HANJA",
        0x37C1: "HANJA",
        0x37C2: "HANJA",
        0x37C3: "HANJA",
        0x37C4: "HANJA",
        0x37C5: "HANJA",
        0x37C6: "HANJA",
        0x37C7: "HANJA",
        0x37C8: "HANJA",
        0x37C9: "HANJA",
        0x37CA: "HANJA",
        0x37CB: "HANJA",
        0x37CC: "HANJA",
        0x37CD: "HANJA",
        0x37CE: "HANJA",
        0x37CF: "HANJA",
        0x37D0: "HANJA",
        0x37D1: "HANJA",
        0x37D2: "HANJA",
        0x37D3: "HANJA",
        0x37D4: "HANJA",
        0x37D5: "HANJA",
        0x37D6: "HANJA",
        0x37D7: "HANJA",
        0x37D8: "HANJA",
        0x37D9: "HANJA",
        0x37DA: "HANJA",
        0x37DB: "HANJA",
        0x37DC: "HANJA",
        0x37DD: "HANJA",
        0x37DE: "HANJA",
        0x37DF: "HANJA",
        0x37E0: "HANJA",
        0x37E1: "HANJA",
        0x37E2: "HANJA",
        0x37E3: "HANJA",
        0x37E4: "HANJA",
        0x37E5: "HANJA",
        0x37E6: "HANJA",
        0x37E7: "HANJA",
        0x37E8: "HANJA",
        0x37E9: "HANJA",
        0x37EA: "HANJA",
        0x37EB: "HANJA",
        0x37EC: "HANJA",
        0x37ED: "HANJA",
        0x37EE: "HANJA",
        0x37EF: "HANJA",
        0x37F0: "HANJA",
        0x37F1: "HANJA",
        0x37F2: "HANJA",
        0x37F3: "HANJA",
        0x37F4: "HANJA",
        0x37F5: "HANJA",
        0x37F6: "HANJA",
        0x37F7: "HANJA",
        0x37F8: "HANJA",
        0x37F9: "HANJA",
        0x37FA: "HANJA",
        0x37FB: "HANJA",
        0x37FC: "HANJA",
        0x37FD: "HANJA",
        0x37FE: "HANJA",
        0x37FF: "HANJA",
        0x3800: "HANJA",
        0x3801: "HANJA",
        0x3802: "HANJA",
        0x3803: "HANJA",
        0x3804: "HANJA",
        0x3805: "HANJA",
        0x3806: "HANJA",
        0x3807: "HANJA",
        0x3808: "HANJA",
        0x3809: "HANJA",
        0x380A: "HANJA",
        0x380B: "HANJA",
        0x380C: "HANJA",
        0x380D: "HANJA",
        0x380E: "HANJA",
        0x380F: "HANJA",
        0x3810: "HANJA",
        0x3811: "HANJA",
        0x3812: "HANJA",
        0x3813: "HANJA",
        0x3814: "HANJA",
        0x3815: "HANJA",
        0x3816: "HANJA",
        0x3817: "HANJA",
        0x3818: "HANJA",
        0x3819: "HANJA",
        0x381A: "HANJA",
        0x381B: "HANJA",
        0x381C: "HANJA",
        0x381D: "HANJA",
        0x381E: "HANJA",
        0x381F: "HANJA",
        0x3820: "HANJA",
        0x3821: "HANJA",
        0x3822: "HANJA",
        0x3823: "HANJA",
        0x3824: "HANJA",
        0x3825: "HANJA",
        0x3826: "HANJA",
        0x3827: "HANJA",
        0x3828: "HANJA",
        0x3829: "HANJA",
        0x382A: "HANJA",
        0x382B: "HANJA",
        0x382C: "HANJA",
        0x382D: "HANJA",
        0x382E: "HANJA",
        0x382F: "HANJA",
        0x3830: "HANJA",
        0x3831: "HANJA",
        0x3832: "HANJA",
        0x3833: "HANJA",
        0x3834: "HANJA",
        0x3835: "HANJA",
        0x3836: "HANJA",
        0x3837: "HANJA",
        0x3838: "HANJA",
        0x3839: "HANJA",
        0x383A: "HANJA",
        0x383B: "HANJA",
        0x383C: "HANJA",
        0x383D: "HANJA",
        0x383E: "HANJA",
        0x383F: "HANJA",
        0x3840: "HANJA",
        0x3841: "HANJA",
        0x3842: "HANJA",
        0x3843: "HANJA",
        0x3844: "HANJA",
        0x3845: "HANJA",
        0x3846: "HANJA",
        0x3847: "HANJA",
        0x3848: "HANJA",
        0x3849: "HANJA",
        0x384A: "HANJA",
        0x384B: "HANJA",
        0x384C: "HANJA",
        0x384D: "HANJA",
        0x384E: "HANJA",
        0x384F: "HANJA",
        0x3850: "HANJA",
        0x3851: "HANJA",
        0x3852: "HANJA",
        0x3853: "HANJA",
        0x3854: "HANJA",
        0x3855: "HANJA",
        0x3856: "HANJA",
        0x3857: "HANJA",
        0x3858: "HANJA",
        0x3859: "HANJA",
        0x385A: "HANJA",
        0x385B: "HANJA",
        0x385C: "HANJA",
        0x385D: "HANJA",
        0x385E: "HANJA",
        0x385F: "HANJA",
        0x3860: "HANJA",
        0x3861: "HANJA",
        0x3862: "HANJA",
        0x3863: "HANJA",
        0x3864: "HANJA",
        0x3865: "HANJA",
        0x3866: "HANJA",
        0x3867: "HANJA",
        0x3868: "HANJA",
        0x3869: "HANJA",
        0x386A: "HANJA",
        0x386B: "HANJA",
        0x386C: "HANJA",
        0x386D: "HANJA",
        0x386E: "HANJA",
        0x386F: "HANJA",
        0x3870: "HANJA",
        0x3871: "HANJA",
        0x3872: "HANJA",
        0x3873: "HANJA",
        0x3874: "HANJA",
        0x3875: "HANJA",
        0x3876: "HANJA",
        0x3877: "HANJA",
        0x3878: "HANJA",
        0x3879: "HANJA",
        0x387A: "HANJA",
        0x387B: "HANJA",
        0x387C: "HANJA",
        0x387D: "HANJA",
        0x387E: "HANJA",
        0x387F: "HANJA",
        0x3880: "HANJA",
        0x3881: "HANJA",
        0x3882: "HANJA",
        0x3883: "HANJA",
        0x3884: "HANJA",
        0x3885: "HANJA",
        0x3886: "HANJA",
        0x3887: "HANJA",
        0x3888: "HANJA",
        0x3889: "HANJA",
        0x388A: "HANJA",
        0x388B: "HANJA",
        0x388C: "HANJA",
        0x388D: "HANJA",
        0x388E: "HANJA",
        0x388F: "HANJA",
        0x3890: "HANJA",
        0x3891: "HANJA",
        0x3892: "HANJA",
        0x3893: "HANJA",
        0x3894: "HANJA",
        0x3895: "HANJA",
        0x3896: "HANJA",
        0x3897: "HANJA",
        0x3898: "HANJA",
        0x3899: "HANJA",
        0x389A: "HANJA",
        0x389B: "HANJA",
        0x389C: "HANJA",
        0x389D: "HANJA",
        0x389E: "HANJA",
        0x389F: "HANJA",
        0x38A0: "HANJA",
        0x38A1: "HANJA",
        0x38A2: "HANJA",
        0x38A3: "HANJA",
        0x38A4: "HANJA",
        0x38A5: "HANJA",
        0x38A6: "HANJA",
        0x38A7: "HANJA",
        0x38A8: "HANJA",
        0x38A9: "HANJA",
        0x38AA: "HANJA",
        0x38AB: "HANJA",
        0x38AC: "HANJA",
        0x38AD: "HANJA",
        0x38AE: "HANJA",
        0x38AF: "HANJA",
        0x38B0: "HANJA",
        0x38B1: "HANJA",
        0x38B2: "HANJA",
        0x38B3: "HANJA",
        0x38B4: "HANJA",
        0x38B5: "HANJA",
        0x38B6: "HANJA",
        0x38B7: "HANJA",
        0x38B8: "HANJA",
        0x38B9: "HANJA",
        0x38BA: "HANJA",
        0x38BB: "HANJA",
        0x38BC: "HANJA",
        0x38BD: "HANJA",
        0x38BE: "HANJA",
        0x38BF: "HANJA",
        0x38C0: "HANJA",
        0x38C1: "HANJA",
        0x38C2: "HANJA",
        0x38C3: "HANJA",
        0x38C4: "HANJA",
        0x38C5: "HANJA",
        0x38C6: "HANJA",
        0x38C7: "HANJA",
        0x38C8: "HANJA",
        0x38C9: "HANJA",
        0x38CA: "HANJA",
        0x38CB: "HANJA",
        0x38CC: "HANJA",
        0x38CD: "HANJA",
        0x38CE: "HANJA",
        0x38CF: "HANJA",
        0x38D0: "HANJA",
        0x38D1: "HANJA",
        0x38D2: "HANJA",
        0x38D3: "HANJA",
        0x38D4: "HANJA",
        0x38D5: "HANJA",
        0x38D6: "HANJA",
        0x38D7: "HANJA",
        0x38D8: "HANJA",
        0x38D9: "HANJA",
        0x38DA: "HANJA",
        0x38DB: "HANJA",
        0x38DC: "HANJA",
        0x38DD: "HANJA",
        0x38DE: "HANJA",
        0x38DF: "HANJA",
        0x38E0: "HANJA",
        0x38E1: "HANJA",
        0x38E2: "HANJA",
        0x38E3: "HANJA",
        0x38E4: "HANJA",
        0x38E5: "HANJA",
        0x38E6: "HANJA",
        0x38E7: "HANJA",
        0x38E8: "HANJA",
        0x38E9: "HANJA",
        0x38EA: "HANJA",
        0x38EB: "HANJA",
        0x38EC: "HANJA",
        0x38ED: "HANJA",
        0x38EE: "HANJA",
        0x38EF: "HANJA",
        0x38F0: "HANJA",
        0x38F1: "HANJA",
        0x38F2: "HANJA",
        0x38F3: "HANJA",
        0x38F4: "HANJA",
        0x38F5: "HANJA",
        0x38F6: "HANJA",
        0x38F7: "HANJA",
        0x38F8: "HANJA",
        0x38F9: "HANJA",
        0x38FA: "HANJA",
        0x38FB: "HANJA",
        0x38FC: "HANJA",
        0x38FD: "HANJA",
        0x38FE: "HANJA",
        0x38FF: "HANJA",
        0x3900: "HANJA",
        0x3901: "HANJA",
        0x3902: "HANJA",
        0x3903: "HANJA",
        0x3904: "HANJA",
        0x3905: "HANJA",
        0x3906: "HANJA",
        0x3907: "HANJA",
        0x3908: "HANJA",
        0x3909: "HANJA",
        0x390A: "HANJA",
        0x390B: "HANJA",
        0x390C: "HANJA",
        0x390D: "HANJA",
        0x390E: "HANJA",
        0x390F: "HANJA",
        0x3910: "HANJA",
        0x3911: "HANJA",
        0x3912: "HANJA",
        0x3913: "HANJA",
        0x3914: "HANJA",
        0x3915: "HANJA",
        0x3916: "HANJA",
        0x3917: "HANJA",
        0x3918: "HANJA",
        0x3919: "HANJA",
        0x391A: "HANJA",
        0x391B: "HANJA",
        0x391C: "HANJA",
        0x391D: "HANJA",
        0x391E: "HANJA",
        0x391F: "HANJA",
        0x3920: "HANJA",
        0x3921: "HANJA",
        0x3922: "HANJA",
        0x3923: "HANJA",
        0x3924: "HANJA",
        0x3925: "HANJA",
        0x3926: "HANJA",
        0x3927: "HANJA",
        0x3928: "HANJA",
        0x3929: "HANJA",
        0x392A: "HANJA",
        0x392B: "HANJA",
        0x392C: "HANJA",
        0x392D: "HANJA",
        0x392E: "HANJA",
        0x392F: "HANJA",
        0x3930: "HANJA",
        0x3931: "HANJA",
        0x3932: "HANJA",
        0x3933: "HANJA",
        0x3934: "HANJA",
        0x3935: "HANJA",
        0x3936: "HANJA",
        0x3937: "HANJA",
        0x3938: "HANJA",
        0x3939: "HANJA",
        0x393A: "HANJA",
        0x393B: "HANJA",
        0x393C: "HANJA",
        0x393D: "HANJA",
        0x393E: "HANJA",
        0x393F: "HANJA",
        0x3940: "HANJA",
        0x3941: "HANJA",
        0x3942: "HANJA",
        0x3943: "HANJA",
        0x3944: "HANJA",
        0x3945: "HANJA",
        0x3946: "HANJA",
        0x3947: "HANJA",
        0x3948: "HANJA",
        0x3949: "HANJA",
        0x394A: "HANJA",
        0x394B: "HANJA",
        0x394C: "HANJA",
        0x394D: "HANJA",
        0x394E: "HANJA",
        0x394F: "HANJA",
        0x3950: "HANJA",
        0x3951: "HANJA",
        0x3952: "HANJA",
        0x3953: "HANJA",
        0x3954: "HANJA",
        0x3955: "HANJA",
        0x3956: "HANJA",
        0x3957: "HANJA",
        0x3958: "HANJA",
        0x3959: "HANJA",
        0x395A: "HANJA",
        0x395B: "HANJA",
        0x395C: "HANJA",
        0x395D: "HANJA",
        0x395E: "HANJA",
        0x395F: "HANJA",
        0x3960: "HANJA",
        0x3961: "HANJA",
        0x3962: "HANJA",
        0x3963: "HANJA",
        0x3964: "HANJA",
        0x3965: "HANJA",
        0x3966: "HANJA",
        0x3967: "HANJA",
        0x3968: "HANJA",
        0x3969: "HANJA",
        0x396A: "HANJA",
        0x396B: "HANJA",
        0x396C: "HANJA",
        0x396D: "HANJA",
        0x396E: "HANJA",
        0x396F: "HANJA",
        0x3970: "HANJA",
        0x3971: "HANJA",
        0x3972: "HANJA",
        0x3973: "HANJA",
        0x3974: "HANJA",
        0x3975: "HANJA",
        0x3976: "HANJA",
        0x3977: "HANJA",
        0x3978: "HANJA",
        0x3979: "HANJA",
        0x397A: "HANJA",
        0x397B: "HANJA",
        0x397C: "HANJA",
        0x397D: "HANJA",
        0x397E: "HANJA",
        0x397F: "HANJA",
        0x3980: "HANJA",
        0x3981: "HANJA",
        0x3982: "HANJA",
        0x3983: "HANJA",
        0x3984: "HANJA",
        0x3985: "HANJA",
        0x3986: "HANJA",
        0x3987: "HANJA",
        0x3988: "HANJA",
        0x3989: "HANJA",
        0x398A: "HANJA",
        0x398B: "HANJA",
        0x398C: "HANJA",
        0x398D: "HANJA",
        0x398E: "HANJA",
        0x398F: "HANJA",
        0x3990: "HANJA",
        0x3991: "HANJA",
        0x3992: "HANJA",
        0x3993: "HANJA",
        0x3994: "HANJA",
        0x3995: "HANJA",
        0x3996: "HANJA",
        0x3997: "HANJA",
        0x3998: "HANJA",
        0x3999: "HANJA",
        0x399A: "HANJA",
        0x399B: "HANJA",
        0x399C: "HANJA",
        0x399D: "HANJA",
        0x399E: "HANJA",
        0x399F: "HANJA",
        0x39A0: "HANJA",
        0x39A1: "HANJA",
        0x39A2: "HANJA",
        0x39A3: "HANJA",
        0x39A4: "HANJA",
        0x39A5: "HANJA",
        0x39A6: "HANJA",
        0x39A7: "HANJA",
        0x39A8: "HANJA",
        0x39A9: "HANJA",
        0x39AA: "HANJA",
        0x39AB: "HANJA",
        0x39AC: "HANJA",
        0x39AD: "HANJA",
        0x39AE: "HANJA",
        0x39AF: "HANJA",
        0x39B0: "HANJA",
        0x39B1: "HANJA",
        0x39B2: "HANJA",
        0x39B3: "HANJA",
        0x39B4: "HANJA",
        0x39B5: "HANJA",
        0x39B6: "HANJA",
        0x39B7: "HANJA",
        0x39B8: "HANJA",
        0x39B9: "HANJA",
        0x39BA: "HANJA",
        0x39BB: "HANJA",
        0x39BC: "HANJA",
        0x39BD: "HANJA",
        0x39BE: "HANJA",
        0x39BF: "HANJA",
        0x39C0: "HANJA",
        0x39C1: "HANJA",
        0x39C2: "HANJA",
        0x39C3: "HANJA",
        0x39C4: "HANJA",
        0x39C5: "HANJA",
        0x39C6: "HANJA",
        0x39C7: "HANJA",
        0x39C8: "HANJA",
        0x39C9: "HANJA",
        0x39CA: "HANJA",
        0x39CB: "HANJA",
        0x39CC: "HANJA",
        0x39CD: "HANJA",
        0x39CE: "HANJA",
        0x39CF: "HANJA",
        0x39D0: "HANJA",
        0x39D1: "HANJA",
        0x39D2: "HANJA",
        0x39D3: "HANJA",
        0x39D4: "HANJA",
        0x39D5: "HANJA",
        0x39D6: "HANJA",
        0x39D7: "HANJA",
        0x39D8: "HANJA",
        0x39D9: "HANJA",
        0x39DA: "HANJA",
        0x39DB: "HANJA",
        0x39DC: "HANJA",
        0x39DD: "HANJA",
        0x39DE: "HANJA",
        0x39DF: "HANJA",
        0x39E0: "HANJA",
        0x39E1: "HANJA",
        0x39E2: "HANJA",
        0x39E3: "HANJA",
        0x39E4: "HANJA",
        0x39E5: "HANJA",
        0x39E6: "HANJA",
        0x39E7: "HANJA",
        0x39E8: "HANJA",
        0x39E9: "HANJA",
        0x39EA: "HANJA",
        0x39EB: "HANJA",
        0x39EC: "HANJA",
        0x39ED: "HANJA",
        0x39EE: "HANJA",
        0x39EF: "HANJA",
        0x39F0: "HANJA",
        0x39F1: "HANJA",
        0x39F2: "HANJA",
        0x39F3: "HANJA",
        0x39F4: "HANJA",
        0x39F5: "HANJA",
        0x39F6: "HANJA",
        0x39F7: "HANJA",
        0x39F8: "HANJA",
        0x39F9: "HANJA",
        0x39FA: "HANJA",
        0x39FB: "HANJA",
        0x39FC: "HANJA",
        0x39FD: "HANJA",
        0x39FE: "HANJA",
        0x39FF: "HANJA",
        0x3A00: "HANJA",
        0x3A01: "HANJA",
        0x3A02: "HANJA",
        0x3A03: "HANJA",
        0x3A04: "HANJA",
        0x3A05: "HANJA",
        0x3A06: "HANJA",
        0x3A07: "HANJA",
        0x3A08: "HANJA",
        0x3A09: "HANJA",
        0x3A0A: "HANJA",
        0x3A0B: "HANJA",
        0x3A0C: "HANJA",
        0x3A0D: "HANJA",
        0x3A0E: "HANJA",
        0x3A0F: "HANJA",
        0x3A10: "HANJA",
        0x3A11: "HANJA",
        0x3A12: "HANJA",
        0x3A13: "HANJA",
        0x3A14: "HANJA",
        0x3A15: "HANJA",
        0x3A16: "HANJA",
        0x3A17: "HANJA",
        0x3A18: "HANJA",
        0x3A19: "HANJA",
        0x3A1A: "HANJA",
        0x3A1B: "HANJA",
        0x3A1C: "HANJA",
        0x3A1D: "HANJA",
        0x3A1E: "HANJA",
        0x3A1F: "HANJA",
        0x3A20: "HANJA",
        0x3A21: "HANJA",
        0x3A22: "HANJA",
        0x3A23: "HANJA",
        0x3A24: "HANJA",
        0x3A25: "HANJA",
        0x3A26: "HANJA",
        0x3A27: "HANJA",
        0x3A28: "HANJA",
        0x3A29: "HANJA",
        0x3A2A: "HANJA",
        0x3A2B: "HANJA",
        0x3A2C: "HANJA",
        0x3A2D: "HANJA",
        0x3A2E: "HANJA",
        0x3A2F: "HANJA",
        0x3A30: "HANJA",
        0x3A31: "HANJA",
        0x3A32: "HANJA",
        0x3A33: "HANJA",
        0x3A34: "HANJA",
        0x3A35: "HANJA",
        0x3A36: "HANJA",
        0x3A37: "HANJA",
        0x3A38: "HANJA",
        0x3A39: "HANJA",
        0x3A3A: "HANJA",
        0x3A3B: "HANJA",
        0x3A3C: "HANJA",
        0x3A3D: "HANJA",
        0x3A3E: "HANJA",
        0x3A3F: "HANJA",
        0x3A40: "HANJA",
        0x3A41: "HANJA",
        0x3A42: "HANJA",
        0x3A43: "HANJA",
        0x3A44: "HANJA",
        0x3A45: "HANJA",
        0x3A46: "HANJA",
        0x3A47: "HANJA",
        0x3A48: "HANJA",
        0x3A49: "HANJA",
        0x3A4A: "HANJA",
        0x3A4B: "HANJA",
        0x3A4C: "HANJA",
        0x3A4D: "HANJA",
        0x3A4E: "HANJA",
        0x3A4F: "HANJA",
        0x3A50: "HANJA",
        0x3A51: "HANJA",
        0x3A52: "HANJA",
        0x3A53: "HANJA",
        0x3A54: "HANJA",
        0x3A55: "HANJA",
        0x3A56: "HANJA",
        0x3A57: "HANJA",
        0x3A58: "HANJA",
        0x3A59: "HANJA",
        0x3A5A: "HANJA",
        0x3A5B: "HANJA",
        0x3A5C: "HANJA",
        0x3A5D: "HANJA",
        0x3A5E: "HANJA",
        0x3A5F: "HANJA",
        0x3A60: "HANJA",
        0x3A61: "HANJA",
        0x3A62: "HANJA",
        0x3A63: "HANJA",
        0x3A64: "HANJA",
        0x3A65: "HANJA",
        0x3A66: "HANJA",
        0x3A67: "HANJA",
        0x3A68: "HANJA",
        0x3A69: "HANJA",
        0x3A6A: "HANJA",
        0x3A6B: "HANJA",
        0x3A6C: "HANJA",
        0x3A6D: "HANJA",
        0x3A6E: "HANJA",
        0x3A6F: "HANJA",
        0x3A70: "HANJA",
        0x3A71: "HANJA",
        0x3A72: "HANJA",
        0x3A73: "HANJA",
        0x3A74: "HANJA",
        0x3A75: "HANJA",
        0x3A76: "HANJA",
        0x3A77: "HANJA",
        0x3A78: "HANJA",
        0x3A79: "HANJA",
        0x3A7A: "HANJA",
        0x3A7B: "HANJA",
        0x3A7C: "HANJA",
        0x3A7D: "HANJA",
        0x3A7E: "HANJA",
        0x3A7F: "HANJA",
        0x3A80: "HANJA",
        0x3A81: "HANJA",
        0x3A82: "HANJA",
        0x3A83: "HANJA",
        0x3A84: "HANJA",
        0x3A85: "HANJA",
        0x3A86: "HANJA",
        0x3A87: "HANJA",
        0x3A88: "HANJA",
        0x3A89: "HANJA",
        0x3A8A: "HANJA",
        0x3A8B: "HANJA",
        0x3A8C: "HANJA",
        0x3A8D: "HANJA",
        0x3A8E: "HANJA",
        0x3A8F: "HANJA",
        0x3A90: "HANJA",
        0x3A91: "HANJA",
        0x3A92: "HANJA",
        0x3A93: "HANJA",
        0x3A94: "HANJA",
        0x3A95: "HANJA",
        0x3A96: "HANJA",
        0x3A97: "HANJA",
        0x3A98: "HANJA",
        0x3A99: "HANJA",
        0x3A9A: "HANJA",
        0x3A9B: "HANJA",
        0x3A9C: "HANJA",
        0x3A9D: "HANJA",
        0x3A9E: "HANJA",
        0x3A9F: "HANJA",
        0x3AA0: "HANJA",
        0x3AA1: "HANJA",
        0x3AA2: "HANJA",
        0x3AA3: "HANJA",
        0x3AA4: "HANJA",
        0x3AA5: "HANJA",
        0x3AA6: "HANJA",
        0x3AA7: "HANJA",
        0x3AA8: "HANJA",
        0x3AA9: "HANJA",
        0x3AAA: "HANJA",
        0x3AAB: "HANJA",
        0x3AAC: "HANJA",
        0x3AAD: "HANJA",
        0x3AAE: "HANJA",
        0x3AAF: "HANJA",
        0x3AB0: "HANJA",
        0x3AB1: "HANJA",
        0x3AB2: "HANJA",
        0x3AB3: "HANJA",
        0x3AB4: "HANJA",
        0x3AB5: "HANJA",
        0x3AB6: "HANJA",
        0x3AB7: "HANJA",
        0x3AB8: "HANJA",
        0x3AB9: "HANJA",
        0x3ABA: "HANJA",
        0x3ABB: "HANJA",
        0x3ABC: "HANJA",
        0x3ABD: "HANJA",
        0x3ABE: "HANJA",
        0x3ABF: "HANJA",
        0x3AC0: "HANJA",
        0x3AC1: "HANJA",
        0x3AC2: "HANJA",
        0x3AC3: "HANJA",
        0x3AC4: "HANJA",
        0x3AC5: "HANJA",
        0x3AC6: "HANJA",
        0x3AC7: "HANJA",
        0x3AC8: "HANJA",
        0x3AC9: "HANJA",
        0x3ACA: "HANJA",
        0x3ACB: "HANJA",
        0x3ACC: "HANJA",
        0x3ACD: "HANJA",
        0x3ACE: "HANJA",
        0x3ACF: "HANJA",
        0x3AD0: "HANJA",
        0x3AD1: "HANJA",
        0x3AD2: "HANJA",
        0x3AD3: "HANJA",
        0x3AD4: "HANJA",
        0x3AD5: "HANJA",
        0x3AD6: "HANJA",
        0x3AD7: "HANJA",
        0x3AD8: "HANJA",
        0x3AD9: "HANJA",
        0x3ADA: "HANJA",
        0x3ADB: "HANJA",
        0x3ADC: "HANJA",
        0x3ADD: "HANJA",
        0x3ADE: "HANJA",
        0x3ADF: "HANJA",
        0x3AE0: "HANJA",
        0x3AE1: "HANJA",
        0x3AE2: "HANJA",
        0x3AE3: "HANJA",
        0x3AE4: "HANJA",
        0x3AE5: "HANJA",
        0x3AE6: "HANJA",
        0x3AE7: "HANJA",
        0x3AE8: "HANJA",
        0x3AE9: "HANJA",
        0x3AEA: "HANJA",
        0x3AEB: "HANJA",
        0x3AEC: "HANJA",
        0x3AED: "HANJA",
        0x3AEE: "HANJA",
        0x3AEF: "HANJA",
        0x3AF0: "HANJA",
        0x3AF1: "HANJA",
        0x3AF2: "HANJA",
        0x3AF3: "HANJA",
        0x3AF4: "HANJA",
        0x3AF5: "HANJA",
        0x3AF6: "HANJA",
        0x3AF7: "HANJA",
        0x3AF8: "HANJA",
        0x3AF9: "HANJA",
        0x3AFA: "HANJA",
        0x3AFB: "HANJA",
        0x3AFC: "HANJA",
        0x3AFD: "HANJA",
        0x3AFE: "HANJA",
        0x3AFF: "HANJA",
        0x3B00: "HANJA",
        0x3B01: "HANJA",
        0x3B02: "HANJA",
        0x3B03: "HANJA",
        0x3B04: "HANJA",
        0x3B05: "HANJA",
        0x3B06: "HANJA",
        0x3B07: "HANJA",
        0x3B08: "HANJA",
        0x3B09: "HANJA",
        0x3B0A: "HANJA",
        0x3B0B: "HANJA",
        0x3B0C: "HANJA",
        0x3B0D: "HANJA",
        0x3B0E: "HANJA",
        0x3B0F: "HANJA",
        0x3B10: "HANJA",
        0x3B11: "HANJA",
        0x3B12: "HANJA",
        0x3B13: "HANJA",
        0x3B14: "HANJA",
        0x3B15: "HANJA",
        0x3B16: "HANJA",
        0x3B17: "HANJA",
        0x3B18: "HANJA",
        0x3B19: "HANJA",
        0x3B1A: "HANJA",
        0x3B1B: "HANJA",
        0x3B1C: "HANJA",
        0x3B1D: "HANJA",
        0x3B1E: "HANJA",
        0x3B1F: "HANJA",
        0x3B20: "HANJA",
        0x3B21: "HANJA",
        0x3B22: "HANJA",
        0x3B23: "HANJA",
        0x3B24: "HANJA",
        0x3B25: "HANJA",
        0x3B26: "HANJA",
        0x3B27: "HANJA",
        0x3B28: "HANJA",
        0x3B29: "HANJA",
        0x3B2A: "HANJA",
        0x3B2B: "HANJA",
        0x3B2C: "HANJA",
        0x3B2D: "HANJA",
        0x3B2E: "HANJA",
        0x3B2F: "HANJA",
        0x3B30: "HANJA",
        0x3B31: "HANJA",
        0x3B32: "HANJA",
        0x3B33: "HANJA",
        0x3B34: "HANJA",
        0x3B35: "HANJA",
        0x3B36: "HANJA",
        0x3B37: "HANJA",
        0x3B38: "HANJA",
        0x3B39: "HANJA",
        0x3B3A: "HANJA",
        0x3B3B: "HANJA",
        0x3B3C: "HANJA",
        0x3B3D: "HANJA",
        0x3B3E: "HANJA",
        0x3B3F: "HANJA",
        0x3B40: "HANJA",
        0x3B41: "HANJA",
        0x3B42: "HANJA",
        0x3B43: "HANJA",
        0x3B44: "HANJA",
        0x3B45: "HANJA",
        0x3B46: "HANJA",
        0x3B47: "HANJA",
        0x3B48: "HANJA",
        0x3B49: "HANJA",
        0x3B4A: "HANJA",
        0x3B4B: "HANJA",
        0x3B4C: "HANJA",
        0x3B4D: "HANJA",
        0x3B4E: "HANJA",
        0x3B4F: "HANJA",
        0x3B50: "HANJA",
        0x3B51: "HANJA",
        0x3B52: "HANJA",
        0x3B53: "HANJA",
        0x3B54: "HANJA",
        0x3B55: "HANJA",
        0x3B56: "HANJA",
        0x3B57: "HANJA",
        0x3B58: "HANJA",
        0x3B59: "HANJA",
        0x3B5A: "HANJA",
        0x3B5B: "HANJA",
        0x3B5C: "HANJA",
        0x3B5D: "HANJA",
        0x3B5E: "HANJA",
        0x3B5F: "HANJA",
        0x3B60: "HANJA",
        0x3B61: "HANJA",
        0x3B62: "HANJA",
        0x3B63: "HANJA",
        0x3B64: "HANJA",
        0x3B65: "HANJA",
        0x3B66: "HANJA",
        0x3B67: "HANJA",
        0x3B68: "HANJA",
        0x3B69: "HANJA",
        0x3B6A: "HANJA",
        0x3B6B: "HANJA",
        0x3B6C: "HANJA",
        0x3B6D: "HANJA",
        0x3B6E: "HANJA",
        0x3B6F: "HANJA",
        0x3B70: "HANJA",
        0x3B71: "HANJA",
        0x3B72: "HANJA",
        0x3B73: "HANJA",
        0x3B74: "HANJA",
        0x3B75: "HANJA",
        0x3B76: "HANJA",
        0x3B77: "HANJA",
        0x3B78: "HANJA",
        0x3B79: "HANJA",
        0x3B7A: "HANJA",
        0x3B7B: "HANJA",
        0x3B7C: "HANJA",
        0x3B7D: "HANJA",
        0x3B7E: "HANJA",
        0x3B7F: "HANJA",
        0x3B80: "HANJA",
        0x3B81: "HANJA",
        0x3B82: "HANJA",
        0x3B83: "HANJA",
        0x3B84: "HANJA",
        0x3B85: "HANJA",
        0x3B86: "HANJA",
        0x3B87: "HANJA",
        0x3B88: "HANJA",
        0x3B89: "HANJA",
        0x3B8A: "HANJA",
        0x3B8B: "HANJA",
        0x3B8C: "HANJA",
        0x3B8D: "HANJA",
        0x3B8E: "HANJA",
        0x3B8F: "HANJA",
        0x3B90: "HANJA",
        0x3B91: "HANJA",
        0x3B92: "HANJA",
        0x3B93: "HANJA",
        0x3B94: "HANJA",
        0x3B95: "HANJA",
        0x3B96: "HANJA",
        0x3B97: "HANJA",
        0x3B98: "HANJA",
        0x3B99: "HANJA",
        0x3B9A: "HANJA",
        0x3B9B: "HANJA",
        0x3B9C: "HANJA",
        0x3B9D: "HANJA",
        0x3B9E: "HANJA",
        0x3B9F: "HANJA",
        0x3BA0: "HANJA",
        0x3BA1: "HANJA",
        0x3BA2: "HANJA",
        0x3BA3: "HANJA",
        0x3BA4: "HANJA",
        0x3BA5: "HANJA",
        0x3BA6: "HANJA",
        0x3BA7: "HANJA",
        0x3BA8: "HANJA",
        0x3BA9: "HANJA",
        0x3BAA: "HANJA",
        0x3BAB: "HANJA",
        0x3BAC: "HANJA",
        0x3BAD: "HANJA",
        0x3BAE: "HANJA",
        0x3BAF: "HANJA",
        0x3BB0: "HANJA",
        0x3BB1: "HANJA",
        0x3BB2: "HANJA",
        0x3BB3: "HANJA",
        0x3BB4: "HANJA",
        0x3BB5: "HANJA",
        0x3BB6: "HANJA",
        0x3BB7: "HANJA",
        0x3BB8: "HANJA",
        0x3BB9: "HANJA",
        0x3BBA: "HANJA",
        0x3BBB: "HANJA",
        0x3BBC: "HANJA",
        0x3BBD: "HANJA",
        0x3BBE: "HANJA",
        0x3BBF: "HANJA",
        0x3BC0: "HANJA",
        0x3BC1: "HANJA",
        0x3BC2: "HANJA",
        0x3BC3: "HANJA",
        0x3BC4: "HANJA",
        0x3BC5: "HANJA",
        0x3BC6: "HANJA",
        0x3BC7: "HANJA",
        0x3BC8: "HANJA",
        0x3BC9: "HANJA",
        0x3BCA: "HANJA",
        0x3BCB: "HANJA",
        0x3BCC: "HANJA",
        0x3BCD: "HANJA",
        0x3BCE: "HANJA",
        0x3BCF: "HANJA",
        0x3BD0: "HANJA",
        0x3BD1: "HANJA",
        0x3BD2: "HANJA",
        0x3BD3: "HANJA",
        0x3BD4: "HANJA",
        0x3BD5: "HANJA",
        0x3BD6: "HANJA",
        0x3BD7: "HANJA",
        0x3BD8: "HANJA",
        0x3BD9: "HANJA",
        0x3BDA: "HANJA",
        0x3BDB: "HANJA",
        0x3BDC: "HANJA",
        0x3BDD: "HANJA",
        0x3BDE: "HANJA",
        0x3BDF: "HANJA",
        0x3BE0: "HANJA",
        0x3BE1: "HANJA",
        0x3BE2: "HANJA",
        0x3BE3: "HANJA",
        0x3BE4: "HANJA",
        0x3BE5: "HANJA",
        0x3BE6: "HANJA",
        0x3BE7: "HANJA",
        0x3BE8: "HANJA",
        0x3BE9: "HANJA",
        0x3BEA: "HANJA",
        0x3BEB: "HANJA",
        0x3BEC: "HANJA",
        0x3BED: "HANJA",
        0x3BEE: "HANJA",
        0x3BEF: "HANJA",
        0x3BF0: "HANJA",
        0x3BF1: "HANJA",
        0x3BF2: "HANJA",
        0x3BF3: "HANJA",
        0x3BF4: "HANJA",
        0x3BF5: "HANJA",
        0x3BF6: "HANJA",
        0x3BF7: "HANJA",
        0x3BF8: "HANJA",
        0x3BF9: "HANJA",
        0x3BFA: "HANJA",
        0x3BFB: "HANJA",
        0x3BFC: "HANJA",
        0x3BFD: "HANJA",
        0x3BFE: "HANJA",
        0x3BFF: "HANJA",
        0x3C00: "HANJA",
        0x3C01: "HANJA",
        0x3C02: "HANJA",
        0x3C03: "HANJA",
        0x3C04: "HANJA",
        0x3C05: "HANJA",
        0x3C06: "HANJA",
        0x3C07: "HANJA",
        0x3C08: "HANJA",
        0x3C09: "HANJA",
        0x3C0A: "HANJA",
        0x3C0B: "HANJA",
        0x3C0C: "HANJA",
        0x3C0D: "HANJA",
        0x3C0E: "HANJA",
        0x3C0F: "HANJA",
        0x3C10: "HANJA",
        0x3C11: "HANJA",
        0x3C12: "HANJA",
        0x3C13: "HANJA",
        0x3C14: "HANJA",
        0x3C15: "HANJA",
        0x3C16: "HANJA",
        0x3C17: "HANJA",
        0x3C18: "HANJA",
        0x3C19: "HANJA",
        0x3C1A: "HANJA",
        0x3C1B: "HANJA",
        0x3C1C: "HANJA",
        0x3C1D: "HANJA",
        0x3C1E: "HANJA",
        0x3C1F: "HANJA",
        0x3C20: "HANJA",
        0x3C21: "HANJA",
        0x3C22: "HANJA",
        0x3C23: "HANJA",
        0x3C24: "HANJA",
        0x3C25: "HANJA",
        0x3C26: "HANJA",
        0x3C27: "HANJA",
        0x3C28: "HANJA",
        0x3C29: "HANJA",
        0x3C2A: "HANJA",
        0x3C2B: "HANJA",
        0x3C2C: "HANJA",
        0x3C2D: "HANJA",
        0x3C2E: "HANJA",
        0x3C2F: "HANJA",
        0x3C30: "HANJA",
        0x3C31: "HANJA",
        0x3C32: "HANJA",
        0x3C33: "HANJA",
        0x3C34: "HANJA",
        0x3C35: "HANJA",
        0x3C36: "HANJA",
        0x3C37: "HANJA",
        0x3C38: "HANJA",
        0x3C39: "HANJA",
        0x3C3A: "HANJA",
        0x3C3B: "HANJA",
        0x3C3C: "HANJA",
        0x3C3D: "HANJA",
        0x3C3E: "HANJA",
        0x3C3F: "HANJA",
        0x3C40: "HANJA",
        0x3C41: "HANJA",
        0x3C42: "HANJA",
        0x3C43: "HANJA",
        0x3C44: "HANJA",
        0x3C45: "HANJA",
        0x3C46: "HANJA",
        0x3C47: "HANJA",
        0x3C48: "HANJA",
        0x3C49: "HANJA",
        0x3C4A: "HANJA",
        0x3C4B: "HANJA",
        0x3C4C: "HANJA",
        0x3C4D: "HANJA",
        0x3C4E: "HANJA",
        0x3C4F: "HANJA",
        0x3C50: "HANJA",
        0x3C51: "HANJA",
        0x3C52: "HANJA",
        0x3C53: "HANJA",
        0x3C54: "HANJA",
        0x3C55: "HANJA",
        0x3C56: "HANJA",
        0x3C57: "HANJA",
        0x3C58: "HANJA",
        0x3C59: "HANJA",
        0x3C5A: "HANJA",
        0x3C5B: "HANJA",
        0x3C5C: "HANJA",
        0x3C5D: "HANJA",
        0x3C5E: "HANJA",
        0x3C5F: "HANJA",
        0x3C60: "HANJA",
        0x3C61: "HANJA",
        0x3C62: "HANJA",
        0x3C63: "HANJA",
        0x3C64: "HANJA",
        0x3C65: "HANJA",
        0x3C66: "HANJA",
        0x3C67: "HANJA",
        0x3C68: "HANJA",
        0x3C69: "HANJA",
        0x3C6A: "HANJA",
        0x3C6B: "HANJA",
        0x3C6C: "HANJA",
        0x3C6D: "HANJA",
        0x3C6E: "HANJA",
        0x3C6F: "HANJA",
        0x3C70: "HANJA",
        0x3C71: "HANJA",
        0x3C72: "HANJA",
        0x3C73: "HANJA",
        0x3C74: "HANJA",
        0x3C75: "HANJA",
        0x3C76: "HANJA",
        0x3C77: "HANJA",
        0x3C78: "HANJA",
        0x3C79: "HANJA",
        0x3C7A: "HANJA",
        0x3C7B: "HANJA",
        0x3C7C: "HANJA",
        0x3C7D: "HANJA",
        0x3C7E: "HANJA",
        0x3C7F: "HANJA",
        0x3C80: "HANJA",
        0x3C81: "HANJA",
        0x3C82: "HANJA",
        0x3C83: "HANJA",
        0x3C84: "HANJA",
        0x3C85: "HANJA",
        0x3C86: "HANJA",
        0x3C87: "HANJA",
        0x3C88: "HANJA",
        0x3C89: "HANJA",
        0x3C8A: "HANJA",
        0x3C8B: "HANJA",
        0x3C8C: "HANJA",
        0x3C8D: "HANJA",
        0x3C8E: "HANJA",
        0x3C8F: "HANJA",
        0x3C90: "HANJA",
        0x3C91: "HANJA",
        0x3C92: "HANJA",
        0x3C93: "HANJA",
        0x3C94: "HANJA",
        0x3C95: "HANJA",
        0x3C96: "HANJA",
        0x3C97: "HANJA",
        0x3C98: "HANJA",
        0x3C99: "HANJA",
        0x3C9A: "HANJA",
        0x3C9B: "HANJA",
        0x3C9C: "HANJA",
        0x3C9D: "HANJA",
        0x3C9E: "HANJA",
        0x3C9F: "HANJA",
        0x3CA0: "HANJA",
        0x3CA1: "HANJA",
        0x3CA2: "HANJA",
        0x3CA3: "HANJA",
        0x3CA4: "HANJA",
        0x3CA5: "HANJA",
        0x3CA6: "HANJA",
        0x3CA7: "HANJA",
        0x3CA8: "HANJA",
        0x3CA9: "HANJA",
        0x3CAA: "HANJA",
        0x3CAB: "HANJA",
        0x3CAC: "HANJA",
        0x3CAD: "HANJA",
        0x3CAE: "HANJA",
        0x3CAF: "HANJA",
        0x3CB0: "HANJA",
        0x3CB1: "HANJA",
        0x3CB2: "HANJA",
        0x3CB3: "HANJA",
        0x3CB4: "HANJA",
        0x3CB5: "HANJA",
        0x3CB6: "HANJA",
        0x3CB7: "HANJA",
        0x3CB8: "HANJA",
        0x3CB9: "HANJA",
        0x3CBA: "HANJA",
        0x3CBB: "HANJA",
        0x3CBC: "HANJA",
        0x3CBD: "HANJA",
        0x3CBE: "HANJA",
        0x3CBF: "HANJA",
        0x3CC0: "HANJA",
        0x3CC1: "HANJA",
        0x3CC2: "HANJA",
        0x3CC3: "HANJA",
        0x3CC4: "HANJA",
        0x3CC5: "HANJA",
        0x3CC6: "HANJA",
        0x3CC7: "HANJA",
        0x3CC8: "HANJA",
        0x3CC9: "HANJA",
        0x3CCA: "HANJA",
        0x3CCB: "HANJA",
        0x3CCC: "HANJA",
        0x3CCD: "HANJA",
        0x3CCE: "HANJA",
        0x3CCF: "HANJA",
        0x3CD0: "HANJA",
        0x3CD1: "HANJA",
        0x3CD2: "HANJA",
        0x3CD3: "HANJA",
        0x3CD4: "HANJA",
        0x3CD5: "HANJA",
        0x3CD6: "HANJA",
        0x3CD7: "HANJA",
        0x3CD8: "HANJA",
        0x3CD9: "HANJA",
        0x3CDA: "HANJA",
        0x3CDB: "HANJA",
        0x3CDC: "HANJA",
        0x3CDD: "HANJA",
        0x3CDE: "HANJA",
        0x3CDF: "HANJA",
        0x3CE0: "HANJA",
        0x3CE1: "HANJA",
        0x3CE2: "HANJA",
        0x3CE3: "HANJA",
        0x3CE4: "HANJA",
        0x3CE5: "HANJA",
        0x3CE6: "HANJA",
        0x3CE7: "HANJA",
        0x3CE8: "HANJA",
        0x3CE9: "HANJA",
        0x3CEA: "HANJA",
        0x3CEB: "HANJA",
        0x3CEC: "HANJA",
        0x3CED: "HANJA",
        0x3CEE: "HANJA",
        0x3CEF: "HANJA",
        0x3CF0: "HANJA",
        0x3CF1: "HANJA",
        0x3CF2: "HANJA",
        0x3CF3: "HANJA",
        0x3CF4: "HANJA",
        0x3CF5: "HANJA",
        0x3CF6: "HANJA",
        0x3CF7: "HANJA",
        0x3CF8: "HANJA",
        0x3CF9: "HANJA",
        0x3CFA: "HANJA",
        0x3CFB: "HANJA",
        0x3CFC: "HANJA",
        0x3CFD: "HANJA",
        0x3CFE: "HANJA",
        0x3CFF: "HANJA",
        0x3D00: "HANJA",
        0x3D01: "HANJA",
        0x3D02: "HANJA",
        0x3D03: "HANJA",
        0x3D04: "HANJA",
        0x3D05: "HANJA",
        0x3D06: "HANJA",
        0x3D07: "HANJA",
        0x3D08: "HANJA",
        0x3D09: "HANJA",
        0x3D0A: "HANJA",
        0x3D0B: "HANJA",
        0x3D0C: "HANJA",
        0x3D0D: "HANJA",
        0x3D0E: "HANJA",
        0x3D0F: "HANJA",
        0x3D10: "HANJA",
        0x3D11: "HANJA",
        0x3D12: "HANJA",
        0x3D13: "HANJA",
        0x3D14: "HANJA",
        0x3D15: "HANJA",
        0x3D16: "HANJA",
        0x3D17: "HANJA",
        0x3D18: "HANJA",
        0x3D19: "HANJA",
        0x3D1A: "HANJA",
        0x3D1B: "HANJA",
        0x3D1C: "HANJA",
        0x3D1D: "HANJA",
        0x3D1E: "HANJA",
        0x3D1F: "HANJA",
        0x3D20: "HANJA",
        0x3D21: "HANJA",
        0x3D22: "HANJA",
        0x3D23: "HANJA",
        0x3D24: "HANJA",
        0x3D25: "HANJA",
        0x3D26: "HANJA",
        0x3D27: "HANJA",
        0x3D28: "HANJA",
        0x3D29: "HANJA",
        0x3D2A: "HANJA",
        0x3D2B: "HANJA",
        0x3D2C: "HANJA",
        0x3D2D: "HANJA",
        0x3D2E: "HANJA",
        0x3D2F: "HANJA",
        0x3D30: "HANJA",
        0x3D31: "HANJA",
        0x3D32: "HANJA",
        0x3D33: "HANJA",
        0x3D34: "HANJA",
        0x3D35: "HANJA",
        0x3D36: "HANJA",
        0x3D37: "HANJA",
        0x3D38: "HANJA",
        0x3D39: "HANJA",
        0x3D3A: "HANJA",
        0x3D3B: "HANJA",
        0x3D3C: "HANJA",
        0x3D3D: "HANJA",
        0x3D3E: "HANJA",
        0x3D3F: "HANJA",
        0x3D40: "HANJA",
        0x3D41: "HANJA",
        0x3D42: "HANJA",
        0x3D43: "HANJA",
        0x3D44: "HANJA",
        0x3D45: "HANJA",
        0x3D46: "HANJA",
        0x3D47: "HANJA",
        0x3D48: "HANJA",
        0x3D49: "HANJA",
        0x3D4A: "HANJA",
        0x3D4B: "HANJA",
        0x3D4C: "HANJA",
        0x3D4D: "HANJA",
        0x3D4E: "HANJA",
        0x3D4F: "HANJA",
        0x3D50: "HANJA",
        0x3D51: "HANJA",
        0x3D52: "HANJA",
        0x3D53: "HANJA",
        0x3D54: "HANJA",
        0x3D55: "HANJA",
        0x3D56: "HANJA",
        0x3D57: "HANJA",
        0x3D58: "HANJA",
        0x3D59: "HANJA",
        0x3D5A: "HANJA",
        0x3D5B: "HANJA",
        0x3D5C: "HANJA",
        0x3D5D: "HANJA",
        0x3D5E: "HANJA",
        0x3D5F: "HANJA",
        0x3D60: "HANJA",
        0x3D61: "HANJA",
        0x3D62: "HANJA",
        0x3D63: "HANJA",
        0x3D64: "HANJA",
        0x3D65: "HANJA",
        0x3D66: "HANJA",
        0x3D67: "HANJA",
        0x3D68: "HANJA",
        0x3D69: "HANJA",
        0x3D6A: "HANJA",
        0x3D6B: "HANJA",
        0x3D6C: "HANJA",
        0x3D6D: "HANJA",
        0x3D6E: "HANJA",
        0x3D6F: "HANJA",
        0x3D70: "HANJA",
        0x3D71: "HANJA",
        0x3D72: "HANJA",
        0x3D73: "HANJA",
        0x3D74: "HANJA",
        0x3D75: "HANJA",
        0x3D76: "HANJA",
        0x3D77: "HANJA",
        0x3D78: "HANJA",
        0x3D79: "HANJA",
        0x3D7A: "HANJA",
        0x3D7B: "HANJA",
        0x3D7C: "HANJA",
        0x3D7D: "HANJA",
        0x3D7E: "HANJA",
        0x3D7F: "HANJA",
        0x3D80: "HANJA",
        0x3D81: "HANJA",
        0x3D82: "HANJA",
        0x3D83: "HANJA",
        0x3D84: "HANJA",
        0x3D85: "HANJA",
        0x3D86: "HANJA",
        0x3D87: "HANJA",
        0x3D88: "HANJA",
        0x3D89: "HANJA",
        0x3D8A: "HANJA",
        0x3D8B: "HANJA",
        0x3D8C: "HANJA",
        0x3D8D: "HANJA",
        0x3D8E: "HANJA",
        0x3D8F: "HANJA",
        0x3D90: "HANJA",
        0x3D91: "HANJA",
        0x3D92: "HANJA",
        0x3D93: "HANJA",
        0x3D94: "HANJA",
        0x3D95: "HANJA",
        0x3D96: "HANJA",
        0x3D97: "HANJA",
        0x3D98: "HANJA",
        0x3D99: "HANJA",
        0x3D9A: "HANJA",
        0x3D9B: "HANJA",
        0x3D9C: "HANJA",
        0x3D9D: "HANJA",
        0x3D9E: "HANJA",
        0x3D9F: "HANJA",
        0x3DA0: "HANJA",
        0x3DA1: "HANJA",
        0x3DA2: "HANJA",
        0x3DA3: "HANJA",
        0x3DA4: "HANJA",
        0x3DA5: "HANJA",
        0x3DA6: "HANJA",
        0x3DA7: "HANJA",
        0x3DA8: "HANJA",
        0x3DA9: "HANJA",
        0x3DAA: "HANJA",
        0x3DAB: "HANJA",
        0x3DAC: "HANJA",
        0x3DAD: "HANJA",
        0x3DAE: "HANJA",
        0x3DAF: "HANJA",
        0x3DB0: "HANJA",
        0x3DB1: "HANJA",
        0x3DB2: "HANJA",
        0x3DB3: "HANJA",
        0x3DB4: "HANJA",
        0x3DB5: "HANJA",
        0x3DB6: "HANJA",
        0x3DB7: "HANJA",
        0x3DB8: "HANJA",
        0x3DB9: "HANJA",
        0x3DBA: "HANJA",
        0x3DBB: "HANJA",
        0x3DBC: "HANJA",
        0x3DBD: "HANJA",
        0x3DBE: "HANJA",
        0x3DBF: "HANJA",
        0x3DC0: "HANJA",
        0x3DC1: "HANJA",
        0x3DC2: "HANJA",
        0x3DC3: "HANJA",
        0x3DC4: "HANJA",
        0x3DC5: "HANJA",
        0x3DC6: "HANJA",
        0x3DC7: "HANJA",
        0x3DC8: "HANJA",
        0x3DC9: "HANJA",
        0x3DCA: "HANJA",
        0x3DCB: "HANJA",
        0x3DCC: "HANJA",
        0x3DCD: "HANJA",
        0x3DCE: "HANJA",
        0x3DCF: "HANJA",
        0x3DD0: "HANJA",
        0x3DD1: "HANJA",
        0x3DD2: "HANJA",
        0x3DD3: "HANJA",
        0x3DD4: "HANJA",
        0x3DD5: "HANJA",
        0x3DD6: "HANJA",
        0x3DD7: "HANJA",
        0x3DD8: "HANJA",
        0x3DD9: "HANJA",
        0x3DDA: "HANJA",
        0x3DDB: "HANJA",
        0x3DDC: "HANJA",
        0x3DDD: "HANJA",
        0x3DDE: "HANJA",
        0x3DDF: "HANJA",
        0x3DE0: "HANJA",
        0x3DE1: "HANJA",
        0x3DE2: "HANJA",
        0x3DE3: "HANJA",
        0x3DE4: "HANJA",
        0x3DE5: "HANJA",
        0x3DE6: "HANJA",
        0x3DE7: "HANJA",
        0x3DE8: "HANJA",
        0x3DE9: "HANJA",
        0x3DEA: "HANJA",
        0x3DEB: "HANJA",
        0x3DEC: "HANJA",
        0x3DED: "HANJA",
        0x3DEE: "HANJA",
        0x3DEF: "HANJA",
        0x3DF0: "HANJA",
        0x3DF1: "HANJA",
        0x3DF2: "HANJA",
        0x3DF3: "HANJA",
        0x3DF4: "HANJA",
        0x3DF5: "HANJA",
        0x3DF6: "HANJA",
        0x3DF7: "HANJA",
        0x3DF8: "HANJA",
        0x3DF9: "HANJA",
        0x3DFA: "HANJA",
        0x3DFB: "HANJA",
        0x3DFC: "HANJA",
        0x3DFD: "HANJA",
        0x3DFE: "HANJA",
        0x3DFF: "HANJA",
        0x3E00: "HANJA",
        0x3E01: "HANJA",
        0x3E02: "HANJA",
        0x3E03: "HANJA",
        0x3E04: "HANJA",
        0x3E05: "HANJA",
        0x3E06: "HANJA",
        0x3E07: "HANJA",
        0x3E08: "HANJA",
        0x3E09: "HANJA",
        0x3E0A: "HANJA",
        0x3E0B: "HANJA",
        0x3E0C: "HANJA",
        0x3E0D: "HANJA",
        0x3E0E: "HANJA",
        0x3E0F: "HANJA",
        0x3E10: "HANJA",
        0x3E11: "HANJA",
        0x3E12: "HANJA",
        0x3E13: "HANJA",
        0x3E14: "HANJA",
        0x3E15: "HANJA",
        0x3E16: "HANJA",
        0x3E17: "HANJA",
        0x3E18: "HANJA",
        0x3E19: "HANJA",
        0x3E1A: "HANJA",
        0x3E1B: "HANJA",
        0x3E1C: "HANJA",
        0x3E1D: "HANJA",
        0x3E1E: "HANJA",
        0x3E1F: "HANJA",
        0x3E20: "HANJA",
        0x3E21: "HANJA",
        0x3E22: "HANJA",
        0x3E23: "HANJA",
        0x3E24: "HANJA",
        0x3E25: "HANJA",
        0x3E26: "HANJA",
        0x3E27: "HANJA",
        0x3E28: "HANJA",
        0x3E29: "HANJA",
        0x3E2A: "HANJA",
        0x3E2B: "HANJA",
        0x3E2C: "HANJA",
        0x3E2D: "HANJA",
        0x3E2E: "HANJA",
        0x3E2F: "HANJA",
        0x3E30: "HANJA",
        0x3E31: "HANJA",
        0x3E32: "HANJA",
        0x3E33: "HANJA",
        0x3E34: "HANJA",
        0x3E35: "HANJA",
        0x3E36: "HANJA",
        0x3E37: "HANJA",
        0x3E38: "HANJA",
        0x3E39: "HANJA",
        0x3E3A: "HANJA",
        0x3E3B: "HANJA",
        0x3E3C: "HANJA",
        0x3E3D: "HANJA",
        0x3E3E: "HANJA",
        0x3E3F: "HANJA",
        0x3E40: "HANJA",
        0x3E41: "HANJA",
        0x3E42: "HANJA",
        0x3E43: "HANJA",
        0x3E44: "HANJA",
        0x3E45: "HANJA",
        0x3E46: "HANJA",
        0x3E47: "HANJA",
        0x3E48: "HANJA",
        0x3E49: "HANJA",
        0x3E4A: "HANJA",
        0x3E4B: "HANJA",
        0x3E4C: "HANJA",
        0x3E4D: "HANJA",
        0x3E4E: "HANJA",
        0x3E4F: "HANJA",
        0x3E50: "HANJA",
        0x3E51: "HANJA",
        0x3E52: "HANJA",
        0x3E53: "HANJA",
        0x3E54: "HANJA",
        0x3E55: "HANJA",
        0x3E56: "HANJA",
        0x3E57: "HANJA",
        0x3E58: "HANJA",
        0x3E59: "HANJA",
        0x3E5A: "HANJA",
        0x3E5B: "HANJA",
        0x3E5C: "HANJA",
        0x3E5D: "HANJA",
        0x3E5E: "HANJA",
        0x3E5F: "HANJA",
        0x3E60: "HANJA",
        0x3E61: "HANJA",
        0x3E62: "HANJA",
        0x3E63: "HANJA",
        0x3E64: "HANJA",
        0x3E65: "HANJA",
        0x3E66: "HANJA",
        0x3E67: "HANJA",
        0x3E68: "HANJA",
        0x3E69: "HANJA",
        0x3E6A: "HANJA",
        0x3E6B: "HANJA",
        0x3E6C: "HANJA",
        0x3E6D: "HANJA",
        0x3E6E: "HANJA",
        0x3E6F: "HANJA",
        0x3E70: "HANJA",
        0x3E71: "HANJA",
        0x3E72: "HANJA",
        0x3E73: "HANJA",
        0x3E74: "HANJA",
        0x3E75: "HANJA",
        0x3E76: "HANJA",
        0x3E77: "HANJA",
        0x3E78: "HANJA",
        0x3E79: "HANJA",
        0x3E7A: "HANJA",
        0x3E7B: "HANJA",
        0x3E7C: "HANJA",
        0x3E7D: "HANJA",
        0x3E7E: "HANJA",
        0x3E7F: "HANJA",
        0x3E80: "HANJA",
        0x3E81: "HANJA",
        0x3E82: "HANJA",
        0x3E83: "HANJA",
        0x3E84: "HANJA",
        0x3E85: "HANJA",
        0x3E86: "HANJA",
        0x3E87: "HANJA",
        0x3E88: "HANJA",
        0x3E89: "HANJA",
        0x3E8A: "HANJA",
        0x3E8B: "HANJA",
        0x3E8C: "HANJA",
        0x3E8D: "HANJA",
        0x3E8E: "HANJA",
        0x3E8F: "HANJA",
        0x3E90: "HANJA",
        0x3E91: "HANJA",
        0x3E92: "HANJA",
        0x3E93: "HANJA",
        0x3E94: "HANJA",
        0x3E95: "HANJA",
        0x3E96: "HANJA",
        0x3E97: "HANJA",
        0x3E98: "HANJA",
        0x3E99: "HANJA",
        0x3E9A: "HANJA",
        0x3E9B: "HANJA",
        0x3E9C: "HANJA",
        0x3E9D: "HANJA",
        0x3E9E: "HANJA",
        0x3E9F: "HANJA",
        0x3EA0: "HANJA",
        0x3EA1: "HANJA",
        0x3EA2: "HANJA",
        0x3EA3: "HANJA",
        0x3EA4: "HANJA",
        0x3EA5: "HANJA",
        0x3EA6: "HANJA",
        0x3EA7: "HANJA",
        0x3EA8: "HANJA",
        0x3EA9: "HANJA",
        0x3EAA: "HANJA",
        0x3EAB: "HANJA",
        0x3EAC: "HANJA",
        0x3EAD: "HANJA",
        0x3EAE: "HANJA",
        0x3EAF: "HANJA",
        0x3EB0: "HANJA",
        0x3EB1: "HANJA",
        0x3EB2: "HANJA",
        0x3EB3: "HANJA",
        0x3EB4: "HANJA",
        0x3EB5: "HANJA",
        0x3EB6: "HANJA",
        0x3EB7: "HANJA",
        0x3EB8: "HANJA",
        0x3EB9: "HANJA",
        0x3EBA: "HANJA",
        0x3EBB: "HANJA",
        0x3EBC: "HANJA",
        0x3EBD: "HANJA",
        0x3EBE: "HANJA",
        0x3EBF: "HANJA",
        0x3EC0: "HANJA",
        0x3EC1: "HANJA",
        0x3EC2: "HANJA",
        0x3EC3: "HANJA",
        0x3EC4: "HANJA",
        0x3EC5: "HANJA",
        0x3EC6: "HANJA",
        0x3EC7: "HANJA",
        0x3EC8: "HANJA",
        0x3EC9: "HANJA",
        0x3ECA: "HANJA",
        0x3ECB: "HANJA",
        0x3ECC: "HANJA",
        0x3ECD: "HANJA",
        0x3ECE: "HANJA",
        0x3ECF: "HANJA",
        0x3ED0: "HANJA",
        0x3ED1: "HANJA",
        0x3ED2: "HANJA",
        0x3ED3: "HANJA",
        0x3ED4: "HANJA",
        0x3ED5: "HANJA",
        0x3ED6: "HANJA",
        0x3ED7: "HANJA",
        0x3ED8: "HANJA",
        0x3ED9: "HANJA",
        0x3EDA: "HANJA",
        0x3EDB: "HANJA",
        0x3EDC: "HANJA",
        0x3EDD: "HANJA",
        0x3EDE: "HANJA",
        0x3EDF: "HANJA",
        0x3EE0: "HANJA",
        0x3EE1: "HANJA",
        0x3EE2: "HANJA",
        0x3EE3: "HANJA",
        0x3EE4: "HANJA",
        0x3EE5: "HANJA",
        0x3EE6: "HANJA",
        0x3EE7: "HANJA",
        0x3EE8: "HANJA",
        0x3EE9: "HANJA",
        0x3EEA: "HANJA",
        0x3EEB: "HANJA",
        0x3EEC: "HANJA",
        0x3EED: "HANJA",
        0x3EEE: "HANJA",
        0x3EEF: "HANJA",
        0x3EF0: "HANJA",
        0x3EF1: "HANJA",
        0x3EF2: "HANJA",
        0x3EF3: "HANJA",
        0x3EF4: "HANJA",
        0x3EF5: "HANJA",
        0x3EF6: "HANJA",
        0x3EF7: "HANJA",
        0x3EF8: "HANJA",
        0x3EF9: "HANJA",
        0x3EFA: "HANJA",
        0x3EFB: "HANJA",
        0x3EFC: "HANJA",
        0x3EFD: "HANJA",
        0x3EFE: "HANJA",
        0x3EFF: "HANJA",
        0x3F00: "HANJA",
        0x3F01: "HANJA",
        0x3F02: "HANJA",
        0x3F03: "HANJA",
        0x3F04: "HANJA",
        0x3F05: "HANJA",
        0x3F06: "HANJA",
        0x3F07: "HANJA",
        0x3F08: "HANJA",
        0x3F09: "HANJA",
        0x3F0A: "HANJA",
        0x3F0B: "HANJA",
        0x3F0C: "HANJA",
        0x3F0D: "HANJA",
        0x3F0E: "HANJA",
        0x3F0F: "HANJA",
        0x3F10: "HANJA",
        0x3F11: "HANJA",
        0x3F12: "HANJA",
        0x3F13: "HANJA",
        0x3F14: "HANJA",
        0x3F15: "HANJA",
        0x3F16: "HANJA",
        0x3F17: "HANJA",
        0x3F18: "HANJA",
        0x3F19: "HANJA",
        0x3F1A: "HANJA",
        0x3F1B: "HANJA",
        0x3F1C: "HANJA",
        0x3F1D: "HANJA",
        0x3F1E: "HANJA",
        0x3F1F: "HANJA",
        0x3F20: "HANJA",
        0x3F21: "HANJA",
        0x3F22: "HANJA",
        0x3F23: "HANJA",
        0x3F24: "HANJA",
        0x3F25: "HANJA",
        0x3F26: "HANJA",
        0x3F27: "HANJA",
        0x3F28: "HANJA",
        0x3F29: "HANJA",
        0x3F2A: "HANJA",
        0x3F2B: "HANJA",
        0x3F2C: "HANJA",
        0x3F2D: "HANJA",
        0x3F2E: "HANJA",
        0x3F2F: "HANJA",
        0x3F30: "HANJA",
        0x3F31: "HANJA",
        0x3F32: "HANJA",
        0x3F33: "HANJA",
        0x3F34: "HANJA",
        0x3F35: "HANJA",
        0x3F36: "HANJA",
        0x3F37: "HANJA",
        0x3F38: "HANJA",
        0x3F39: "HANJA",
        0x3F3A: "HANJA",
        0x3F3B: "HANJA",
        0x3F3C: "HANJA",
        0x3F3D: "HANJA",
        0x3F3E: "HANJA",
        0x3F3F: "HANJA",
        0x3F40: "HANJA",
        0x3F41: "HANJA",
        0x3F42: "HANJA",
        0x3F43: "HANJA",
        0x3F44: "HANJA",
        0x3F45: "HANJA",
        0x3F46: "HANJA",
        0x3F47: "HANJA",
        0x3F48: "HANJA",
        0x3F49: "HANJA",
        0x3F4A: "HANJA",
        0x3F4B: "HANJA",
        0x3F4C: "HANJA",
        0x3F4D: "HANJA",
        0x3F4E: "HANJA",
        0x3F4F: "HANJA",
        0x3F50: "HANJA",
        0x3F51: "HANJA",
        0x3F52: "HANJA",
        0x3F53: "HANJA",
        0x3F54: "HANJA",
        0x3F55: "HANJA",
        0x3F56: "HANJA",
        0x3F57: "HANJA",
        0x3F58: "HANJA",
        0x3F59: "HANJA",
        0x3F5A: "HANJA",
        0x3F5B: "HANJA",
        0x3F5C: "HANJA",
        0x3F5D: "HANJA",
        0x3F5E: "HANJA",
        0x3F5F: "HANJA",
        0x3F60: "HANJA",
        0x3F61: "HANJA",
        0x3F62: "HANJA",
        0x3F63: "HANJA",
        0x3F64: "HANJA",
        0x3F65: "HANJA",
        0x3F66: "HANJA",
        0x3F67: "HANJA",
        0x3F68: "HANJA",
        0x3F69: "HANJA",
        0x3F6A: "HANJA",
        0x3F6B: "HANJA",
        0x3F6C: "HANJA",
        0x3F6D: "HANJA",
        0x3F6E: "HANJA",
        0x3F6F: "HANJA",
        0x3F70: "HANJA",
        0x3F71: "HANJA",
        0x3F72: "HANJA",
        0x3F73: "HANJA",
        0x3F74: "HANJA",
        0x3F75: "HANJA",
        0x3F76: "HANJA",
        0x3F77: "HANJA",
        0x3F78: "HANJA",
        0x3F79: "HANJA",
        0x3F7A: "HANJA",
        0x3F7B: "HANJA",
        0x3F7C: "HANJA",
        0x3F7D: "HANJA",
        0x3F7E: "HANJA",
        0x3F7F: "HANJA",
        0x3F80: "HANJA",
        0x3F81: "HANJA",
        0x3F82: "HANJA",
        0x3F83: "HANJA",
        0x3F84: "HANJA",
        0x3F85: "HANJA",
        0x3F86: "HANJA",
        0x3F87: "HANJA",
        0x3F88: "HANJA",
        0x3F89: "HANJA",
        0x3F8A: "HANJA",
        0x3F8B: "HANJA",
        0x3F8C: "HANJA",
        0x3F8D: "HANJA",
        0x3F8E: "HANJA",
        0x3F8F: "HANJA",
        0x3F90: "HANJA",
        0x3F91: "HANJA",
        0x3F92: "HANJA",
        0x3F93: "HANJA",
        0x3F94: "HANJA",
        0x3F95: "HANJA",
        0x3F96: "HANJA",
        0x3F97: "HANJA",
        0x3F98: "HANJA",
        0x3F99: "HANJA",
        0x3F9A: "HANJA",
        0x3F9B: "HANJA",
        0x3F9C: "HANJA",
        0x3F9D: "HANJA",
        0x3F9E: "HANJA",
        0x3F9F: "HANJA",
        0x3FA0: "HANJA",
        0x3FA1: "HANJA",
        0x3FA2: "HANJA",
        0x3FA3: "HANJA",
        0x3FA4: "HANJA",
        0x3FA5: "HANJA",
        0x3FA6: "HANJA",
        0x3FA7: "HANJA",
        0x3FA8: "HANJA",
        0x3FA9: "HANJA",
        0x3FAA: "HANJA",
        0x3FAB: "HANJA",
        0x3FAC: "HANJA",
        0x3FAD: "HANJA",
        0x3FAE: "HANJA",
        0x3FAF: "HANJA",
        0x3FB0: "HANJA",
        0x3FB1: "HANJA",
        0x3FB2: "HANJA",
        0x3FB3: "HANJA",
        0x3FB4: "HANJA",
        0x3FB5: "HANJA",
        0x3FB6: "HANJA",
        0x3FB7: "HANJA",
        0x3FB8: "HANJA",
        0x3FB9: "HANJA",
        0x3FBA: "HANJA",
        0x3FBB: "HANJA",
        0x3FBC: "HANJA",
        0x3FBD: "HANJA",
        0x3FBE: "HANJA",
        0x3FBF: "HANJA",
        0x3FC0: "HANJA",
        0x3FC1: "HANJA",
        0x3FC2: "HANJA",
        0x3FC3: "HANJA",
        0x3FC4: "HANJA",
        0x3FC5: "HANJA",
        0x3FC6: "HANJA",
        0x3FC7: "HANJA",
        0x3FC8: "HANJA",
        0x3FC9: "HANJA",
        0x3FCA: "HANJA",
        0x3FCB: "HANJA",
        0x3FCC: "HANJA",
        0x3FCD: "HANJA",
        0x3FCE: "HANJA",
        0x3FCF: "HANJA",
        0x3FD0: "HANJA",
        0x3FD1: "HANJA",
        0x3FD2: "HANJA",
        0x3FD3: "HANJA",
        0x3FD4: "HANJA",
        0x3FD5: "HANJA",
        0x3FD6: "HANJA",
        0x3FD7: "HANJA",
        0x3FD8: "HANJA",
        0x3FD9: "HANJA",
        0x3FDA: "HANJA",
        0x3FDB: "HANJA",
        0x3FDC: "HANJA",
        0x3FDD: "HANJA",
        0x3FDE: "HANJA",
        0x3FDF: "HANJA",
        0x3FE0: "HANJA",
        0x3FE1: "HANJA",
        0x3FE2: "HANJA",
        0x3FE3: "HANJA",
        0x3FE4: "HANJA",
        0x3FE5: "HANJA",
        0x3FE6: "HANJA",
        0x3FE7: "HANJA",
        0x3FE8: "HANJA",
        0x3FE9: "HANJA",
        0x3FEA: "HANJA",
        0x3FEB: "HANJA",
        0x3FEC: "HANJA",
        0x3FED: "HANJA",
        0x3FEE: "HANJA",
        0x3FEF: "HANJA",
        0x3FF0: "HANJA",
        0x3FF1: "HANJA",
        0x3FF2: "HANJA",
        0x3FF3: "HANJA",
        0x3FF4: "HANJA",
        0x3FF5: "HANJA",
        0x3FF6: "HANJA",
        0x3FF7: "HANJA",
        0x3FF8: "HANJA",
        0x3FF9: "HANJA",
        0x3FFA: "HANJA",
        0x3FFB: "HANJA",
        0x3FFC: "HANJA",
        0x3FFD: "HANJA",
        0x3FFE: "HANJA",
        0x3FFF: "HANJA",
        0x4000: "HANJA",
        0x4001: "HANJA",
        0x4002: "HANJA",
        0x4003: "HANJA",
        0x4004: "HANJA",
        0x4005: "HANJA",
        0x4006: "HANJA",
        0x4007: "HANJA",
        0x4008: "HANJA",
        0x4009: "HANJA",
        0x400A: "HANJA",
        0x400B: "HANJA",
        0x400C: "HANJA",
        0x400D: "HANJA",
        0x400E: "HANJA",
        0x400F: "HANJA",
        0x4010: "HANJA",
        0x4011: "HANJA",
        0x4012: "HANJA",
        0x4013: "HANJA",
        0x4014: "HANJA",
        0x4015: "HANJA",
        0x4016: "HANJA",
        0x4017: "HANJA",
        0x4018: "HANJA",
        0x4019: "HANJA",
        0x401A: "HANJA",
        0x401B: "HANJA",
        0x401C: "HANJA",
        0x401D: "HANJA",
        0x401E: "HANJA",
        0x401F: "HANJA",
        0x4020: "HANJA",
        0x4021: "HANJA",
        0x4022: "HANJA",
        0x4023: "HANJA",
        0x4024: "HANJA",
        0x4025: "HANJA",
        0x4026: "HANJA",
        0x4027: "HANJA",
        0x4028: "HANJA",
        0x4029: "HANJA",
        0x402A: "HANJA",
        0x402B: "HANJA",
        0x402C: "HANJA",
        0x402D: "HANJA",
        0x402E: "HANJA",
        0x402F: "HANJA",
        0x4030: "HANJA",
        0x4031: "HANJA",
        0x4032: "HANJA",
        0x4033: "HANJA",
        0x4034: "HANJA",
        0x4035: "HANJA",
        0x4036: "HANJA",
        0x4037: "HANJA",
        0x4038: "HANJA",
        0x4039: "HANJA",
        0x403A: "HANJA",
        0x403B: "HANJA",
        0x403C: "HANJA",
        0x403D: "HANJA",
        0x403E: "HANJA",
        0x403F: "HANJA",
        0x4040: "HANJA",
        0x4041: "HANJA",
        0x4042: "HANJA",
        0x4043: "HANJA",
        0x4044: "HANJA",
        0x4045: "HANJA",
        0x4046: "HANJA",
        0x4047: "HANJA",
        0x4048: "HANJA",
        0x4049: "HANJA",
        0x404A: "HANJA",
        0x404B: "HANJA",
        0x404C: "HANJA",
        0x404D: "HANJA",
        0x404E: "HANJA",
        0x404F: "HANJA",
        0x4050: "HANJA",
        0x4051: "HANJA",
        0x4052: "HANJA",
        0x4053: "HANJA",
        0x4054: "HANJA",
        0x4055: "HANJA",
        0x4056: "HANJA",
        0x4057: "HANJA",
        0x4058: "HANJA",
        0x4059: "HANJA",
        0x405A: "HANJA",
        0x405B: "HANJA",
        0x405C: "HANJA",
        0x405D: "HANJA",
        0x405E: "HANJA",
        0x405F: "HANJA",
        0x4060: "HANJA",
        0x4061: "HANJA",
        0x4062: "HANJA",
        0x4063: "HANJA",
        0x4064: "HANJA",
        0x4065: "HANJA",
        0x4066: "HANJA",
        0x4067: "HANJA",
        0x4068: "HANJA",
        0x4069: "HANJA",
        0x406A: "HANJA",
        0x406B: "HANJA",
        0x406C: "HANJA",
        0x406D: "HANJA",
        0x406E: "HANJA",
        0x406F: "HANJA",
        0x4070: "HANJA",
        0x4071: "HANJA",
        0x4072: "HANJA",
        0x4073: "HANJA",
        0x4074: "HANJA",
        0x4075: "HANJA",
        0x4076: "HANJA",
        0x4077: "HANJA",
        0x4078: "HANJA",
        0x4079: "HANJA",
        0x407A: "HANJA",
        0x407B: "HANJA",
        0x407C: "HANJA",
        0x407D: "HANJA",
        0x407E: "HANJA",
        0x407F: "HANJA",
        0x4080: "HANJA",
        0x4081: "HANJA",
        0x4082: "HANJA",
        0x4083: "HANJA",
        0x4084: "HANJA",
        0x4085: "HANJA",
        0x4086: "HANJA",
        0x4087: "HANJA",
        0x4088: "HANJA",
        0x4089: "HANJA",
        0x408A: "HANJA",
        0x408B: "HANJA",
        0x408C: "HANJA",
        0x408D: "HANJA",
        0x408E: "HANJA",
        0x408F: "HANJA",
        0x4090: "HANJA",
        0x4091: "HANJA",
        0x4092: "HANJA",
        0x4093: "HANJA",
        0x4094: "HANJA",
        0x4095: "HANJA",
        0x4096: "HANJA",
        0x4097: "HANJA",
        0x4098: "HANJA",
        0x4099: "HANJA",
        0x409A: "HANJA",
        0x409B: "HANJA",
        0x409C: "HANJA",
        0x409D: "HANJA",
        0x409E: "HANJA",
        0x409F: "HANJA",
        0x40A0: "HANJA",
        0x40A1: "HANJA",
        0x40A2: "HANJA",
        0x40A3: "HANJA",
        0x40A4: "HANJA",
        0x40A5: "HANJA",
        0x40A6: "HANJA",
        0x40A7: "HANJA",
        0x40A8: "HANJA",
        0x40A9: "HANJA",
        0x40AA: "HANJA",
        0x40AB: "HANJA",
        0x40AC: "HANJA",
        0x40AD: "HANJA",
        0x40AE: "HANJA",
        0x40AF: "HANJA",
        0x40B0: "HANJA",
        0x40B1: "HANJA",
        0x40B2: "HANJA",
        0x40B3: "HANJA",
        0x40B4: "HANJA",
        0x40B5: "HANJA",
        0x40B6: "HANJA",
        0x40B7: "HANJA",
        0x40B8: "HANJA",
        0x40B9: "HANJA",
        0x40BA: "HANJA",
        0x40BB: "HANJA",
        0x40BC: "HANJA",
        0x40BD: "HANJA",
        0x40BE: "HANJA",
        0x40BF: "HANJA",
        0x40C0: "HANJA",
        0x40C1: "HANJA",
        0x40C2: "HANJA",
        0x40C3: "HANJA",
        0x40C4: "HANJA",
        0x40C5: "HANJA",
        0x40C6: "HANJA",
        0x40C7: "HANJA",
        0x40C8: "HANJA",
        0x40C9: "HANJA",
        0x40CA: "HANJA",
        0x40CB: "HANJA",
        0x40CC: "HANJA",
        0x40CD: "HANJA",
        0x40CE: "HANJA",
        0x40CF: "HANJA",
        0x40D0: "HANJA",
        0x40D1: "HANJA",
        0x40D2: "HANJA",
        0x40D3: "HANJA",
        0x40D4: "HANJA",
        0x40D5: "HANJA",
        0x40D6: "HANJA",
        0x40D7: "HANJA",
        0x40D8: "HANJA",
        0x40D9: "HANJA",
        0x40DA: "HANJA",
        0x40DB: "HANJA",
        0x40DC: "HANJA",
        0x40DD: "HANJA",
        0x40DE: "HANJA",
        0x40DF: "HANJA",
        0x40E0: "HANJA",
        0x40E1: "HANJA",
        0x40E2: "HANJA",
        0x40E3: "HANJA",
        0x40E4: "HANJA",
        0x40E5: "HANJA",
        0x40E6: "HANJA",
        0x40E7: "HANJA",
        0x40E8: "HANJA",
        0x40E9: "HANJA",
        0x40EA: "HANJA",
        0x40EB: "HANJA",
        0x40EC: "HANJA",
        0x40ED: "HANJA",
        0x40EE: "HANJA",
        0x40EF: "HANJA",
        0x40F0: "HANJA",
        0x40F1: "HANJA",
        0x40F2: "HANJA",
        0x40F3: "HANJA",
        0x40F4: "HANJA",
        0x40F5: "HANJA",
        0x40F6: "HANJA",
        0x40F7: "HANJA",
        0x40F8: "HANJA",
        0x40F9: "HANJA",
        0x40FA: "HANJA",
        0x40FB: "HANJA",
        0x40FC: "HANJA",
        0x40FD: "HANJA",
        0x40FE: "HANJA",
        0x40FF: "HANJA",
        0x4100: "HANJA",
        0x4101: "HANJA",
        0x4102: "HANJA",
        0x4103: "HANJA",
        0x4104: "HANJA",
        0x4105: "HANJA",
        0x4106: "HANJA",
        0x4107: "HANJA",
        0x4108: "HANJA",
        0x4109: "HANJA",
        0x410A: "HANJA",
        0x410B: "HANJA",
        0x410C: "HANJA",
        0x410D: "HANJA",
        0x410E: "HANJA",
        0x410F: "HANJA",
        0x4110: "HANJA",
        0x4111: "HANJA",
        0x4112: "HANJA",
        0x4113: "HANJA",
        0x4114: "HANJA",
        0x4115: "HANJA",
        0x4116: "HANJA",
        0x4117: "HANJA",
        0x4118: "HANJA",
        0x4119: "HANJA",
        0x411A: "HANJA",
        0x411B: "HANJA",
        0x411C: "HANJA",
        0x411D: "HANJA",
        0x411E: "HANJA",
        0x411F: "HANJA",
        0x4120: "HANJA",
        0x4121: "HANJA",
        0x4122: "HANJA",
        0x4123: "HANJA",
        0x4124: "HANJA",
        0x4125: "HANJA",
        0x4126: "HANJA",
        0x4127: "HANJA",
        0x4128: "HANJA",
        0x4129: "HANJA",
        0x412A: "HANJA",
        0x412B: "HANJA",
        0x412C: "HANJA",
        0x412D: "HANJA",
        0x412E: "HANJA",
        0x412F: "HANJA",
        0x4130: "HANJA",
        0x4131: "HANJA",
        0x4132: "HANJA",
        0x4133: "HANJA",
        0x4134: "HANJA",
        0x4135: "HANJA",
        0x4136: "HANJA",
        0x4137: "HANJA",
        0x4138: "HANJA",
        0x4139: "HANJA",
        0x413A: "HANJA",
        0x413B: "HANJA",
        0x413C: "HANJA",
        0x413D: "HANJA",
        0x413E: "HANJA",
        0x413F: "HANJA",
        0x4140: "HANJA",
        0x4141: "HANJA",
        0x4142: "HANJA",
        0x4143: "HANJA",
        0x4144: "HANJA",
        0x4145: "HANJA",
        0x4146: "HANJA",
        0x4147: "HANJA",
        0x4148: "HANJA",
        0x4149: "HANJA",
        0x414A: "HANJA",
        0x414B: "HANJA",
        0x414C: "HANJA",
        0x414D: "HANJA",
        0x414E: "HANJA",
        0x414F: "HANJA",
        0x4150: "HANJA",
        0x4151: "HANJA",
        0x4152: "HANJA",
        0x4153: "HANJA",
        0x4154: "HANJA",
        0x4155: "HANJA",
        0x4156: "HANJA",
        0x4157: "HANJA",
        0x4158: "HANJA",
        0x4159: "HANJA",
        0x415A: "HANJA",
        0x415B: "HANJA",
        0x415C: "HANJA",
        0x415D: "HANJA",
        0x415E: "HANJA",
        0x415F: "HANJA",
        0x4160: "HANJA",
        0x4161: "HANJA",
        0x4162: "HANJA",
        0x4163: "HANJA",
        0x4164: "HANJA",
        0x4165: "HANJA",
        0x4166: "HANJA",
        0x4167: "HANJA",
        0x4168: "HANJA",
        0x4169: "HANJA",
        0x416A: "HANJA",
        0x416B: "HANJA",
        0x416C: "HANJA",
        0x416D: "HANJA",
        0x416E: "HANJA",
        0x416F: "HANJA",
        0x4170: "HANJA",
        0x4171: "HANJA",
        0x4172: "HANJA",
        0x4173: "HANJA",
        0x4174: "HANJA",
        0x4175: "HANJA",
        0x4176: "HANJA",
        0x4177: "HANJA",
        0x4178: "HANJA",
        0x4179: "HANJA",
        0x417A: "HANJA",
        0x417B: "HANJA",
        0x417C: "HANJA",
        0x417D: "HANJA",
        0x417E: "HANJA",
        0x417F: "HANJA",
        0x4180: "HANJA",
        0x4181: "HANJA",
        0x4182: "HANJA",
        0x4183: "HANJA",
        0x4184: "HANJA",
        0x4185: "HANJA",
        0x4186: "HANJA",
        0x4187: "HANJA",
        0x4188: "HANJA",
        0x4189: "HANJA",
        0x418A: "HANJA",
        0x418B: "HANJA",
        0x418C: "HANJA",
        0x418D: "HANJA",
        0x418E: "HANJA",
        0x418F: "HANJA",
        0x4190: "HANJA",
        0x4191: "HANJA",
        0x4192: "HANJA",
        0x4193: "HANJA",
        0x4194: "HANJA",
        0x4195: "HANJA",
        0x4196: "HANJA",
        0x4197: "HANJA",
        0x4198: "HANJA",
        0x4199: "HANJA",
        0x419A: "HANJA",
        0x419B: "HANJA",
        0x419C: "HANJA",
        0x419D: "HANJA",
        0x419E: "HANJA",
        0x419F: "HANJA",
        0x41A0: "HANJA",
        0x41A1: "HANJA",
        0x41A2: "HANJA",
        0x41A3: "HANJA",
        0x41A4: "HANJA",
        0x41A5: "HANJA",
        0x41A6: "HANJA",
        0x41A7: "HANJA",
        0x41A8: "HANJA",
        0x41A9: "HANJA",
        0x41AA: "HANJA",
        0x41AB: "HANJA",
        0x41AC: "HANJA",
        0x41AD: "HANJA",
        0x41AE: "HANJA",
        0x41AF: "HANJA",
        0x41B0: "HANJA",
        0x41B1: "HANJA",
        0x41B2: "HANJA",
        0x41B3: "HANJA",
        0x41B4: "HANJA",
        0x41B5: "HANJA",
        0x41B6: "HANJA",
        0x41B7: "HANJA",
        0x41B8: "HANJA",
        0x41B9: "HANJA",
        0x41BA: "HANJA",
        0x41BB: "HANJA",
        0x41BC: "HANJA",
        0x41BD: "HANJA",
        0x41BE: "HANJA",
        0x41BF: "HANJA",
        0x41C0: "HANJA",
        0x41C1: "HANJA",
        0x41C2: "HANJA",
        0x41C3: "HANJA",
        0x41C4: "HANJA",
        0x41C5: "HANJA",
        0x41C6: "HANJA",
        0x41C7: "HANJA",
        0x41C8: "HANJA",
        0x41C9: "HANJA",
        0x41CA: "HANJA",
        0x41CB: "HANJA",
        0x41CC: "HANJA",
        0x41CD: "HANJA",
        0x41CE: "HANJA",
        0x41CF: "HANJA",
        0x41D0: "HANJA",
        0x41D1: "HANJA",
        0x41D2: "HANJA",
        0x41D3: "HANJA",
        0x41D4: "HANJA",
        0x41D5: "HANJA",
        0x41D6: "HANJA",
        0x41D7: "HANJA",
        0x41D8: "HANJA",
        0x41D9: "HANJA",
        0x41DA: "HANJA",
        0x41DB: "HANJA",
        0x41DC: "HANJA",
        0x41DD: "HANJA",
        0x41DE: "HANJA",
        0x41DF: "HANJA",
        0x41E0: "HANJA",
        0x41E1: "HANJA",
        0x41E2: "HANJA",
        0x41E3: "HANJA",
        0x41E4: "HANJA",
        0x41E5: "HANJA",
        0x41E6: "HANJA",
        0x41E7: "HANJA",
        0x41E8: "HANJA",
        0x41E9: "HANJA",
        0x41EA: "HANJA",
        0x41EB: "HANJA",
        0x41EC: "HANJA",
        0x41ED: "HANJA",
        0x41EE: "HANJA",
        0x41EF: "HANJA",
        0x41F0: "HANJA",
        0x41F1: "HANJA",
        0x41F2: "HANJA",
        0x41F3: "HANJA",
        0x41F4: "HANJA",
        0x41F5: "HANJA",
        0x41F6: "HANJA",
        0x41F7: "HANJA",
        0x41F8: "HANJA",
        0x41F9: "HANJA",
        0x41FA: "HANJA",
        0x41FB: "HANJA",
        0x41FC: "HANJA",
        0x41FD: "HANJA",
        0x41FE: "HANJA",
        0x41FF: "HANJA",
        0x4200: "HANJA",
        0x4201: "HANJA",
        0x4202: "HANJA",
        0x4203: "HANJA",
        0x4204: "HANJA",
        0x4205: "HANJA",
        0x4206: "HANJA",
        0x4207: "HANJA",
        0x4208: "HANJA",
        0x4209: "HANJA",
        0x420A: "HANJA",
        0x420B: "HANJA",
        0x420C: "HANJA",
        0x420D: "HANJA",
        0x420E: "HANJA",
        0x420F: "HANJA",
        0x4210: "HANJA",
        0x4211: "HANJA",
        0x4212: "HANJA",
        0x4213: "HANJA",
        0x4214: "HANJA",
        0x4215: "HANJA",
        0x4216: "HANJA",
        0x4217: "HANJA",
        0x4218: "HANJA",
        0x4219: "HANJA",
        0x421A: "HANJA",
        0x421B: "HANJA",
        0x421C: "HANJA",
        0x421D: "HANJA",
        0x421E: "HANJA",
        0x421F: "HANJA",
        0x4220: "HANJA",
        0x4221: "HANJA",
        0x4222: "HANJA",
        0x4223: "HANJA",
        0x4224: "HANJA",
        0x4225: "HANJA",
        0x4226: "HANJA",
        0x4227: "HANJA",
        0x4228: "HANJA",
        0x4229: "HANJA",
        0x422A: "HANJA",
        0x422B: "HANJA",
        0x422C: "HANJA",
        0x422D: "HANJA",
        0x422E: "HANJA",
        0x422F: "HANJA",
        0x4230: "HANJA",
        0x4231: "HANJA",
        0x4232: "HANJA",
        0x4233: "HANJA",
        0x4234: "HANJA",
        0x4235: "HANJA",
        0x4236: "HANJA",
        0x4237: "HANJA",
        0x4238: "HANJA",
        0x4239: "HANJA",
        0x423A: "HANJA",
        0x423B: "HANJA",
        0x423C: "HANJA",
        0x423D: "HANJA",
        0x423E: "HANJA",
        0x423F: "HANJA",
        0x4240: "HANJA",
        0x4241: "HANJA",
        0x4242: "HANJA",
        0x4243: "HANJA",
        0x4244: "HANJA",
        0x4245: "HANJA",
        0x4246: "HANJA",
        0x4247: "HANJA",
        0x4248: "HANJA",
        0x4249: "HANJA",
        0x424A: "HANJA",
        0x424B: "HANJA",
        0x424C: "HANJA",
        0x424D: "HANJA",
        0x424E: "HANJA",
        0x424F: "HANJA",
        0x4250: "HANJA",
        0x4251: "HANJA",
        0x4252: "HANJA",
        0x4253: "HANJA",
        0x4254: "HANJA",
        0x4255: "HANJA",
        0x4256: "HANJA",
        0x4257: "HANJA",
        0x4258: "HANJA",
        0x4259: "HANJA",
        0x425A: "HANJA",
        0x425B: "HANJA",
        0x425C: "HANJA",
        0x425D: "HANJA",
        0x425E: "HANJA",
        0x425F: "HANJA",
        0x4260: "HANJA",
        0x4261: "HANJA",
        0x4262: "HANJA",
        0x4263: "HANJA",
        0x4264: "HANJA",
        0x4265: "HANJA",
        0x4266: "HANJA",
        0x4267: "HANJA",
        0x4268: "HANJA",
        0x4269: "HANJA",
        0x426A: "HANJA",
        0x426B: "HANJA",
        0x426C: "HANJA",
        0x426D: "HANJA",
        0x426E: "HANJA",
        0x426F: "HANJA",
        0x4270: "HANJA",
        0x4271: "HANJA",
        0x4272: "HANJA",
        0x4273: "HANJA",
        0x4274: "HANJA",
        0x4275: "HANJA",
        0x4276: "HANJA",
        0x4277: "HANJA",
        0x4278: "HANJA",
        0x4279: "HANJA",
        0x427A: "HANJA",
        0x427B: "HANJA",
        0x427C: "HANJA",
        0x427D: "HANJA",
        0x427E: "HANJA",
        0x427F: "HANJA",
        0x4280: "HANJA",
        0x4281: "HANJA",
        0x4282: "HANJA",
        0x4283: "HANJA",
        0x4284: "HANJA",
        0x4285: "HANJA",
        0x4286: "HANJA",
        0x4287: "HANJA",
        0x4288: "HANJA",
        0x4289: "HANJA",
        0x428A: "HANJA",
        0x428B: "HANJA",
        0x428C: "HANJA",
        0x428D: "HANJA",
        0x428E: "HANJA",
        0x428F: "HANJA",
        0x4290: "HANJA",
        0x4291: "HANJA",
        0x4292: "HANJA",
        0x4293: "HANJA",
        0x4294: "HANJA",
        0x4295: "HANJA",
        0x4296: "HANJA",
        0x4297: "HANJA",
        0x4298: "HANJA",
        0x4299: "HANJA",
        0x429A: "HANJA",
        0x429B: "HANJA",
        0x429C: "HANJA",
        0x429D: "HANJA",
        0x429E: "HANJA",
        0x429F: "HANJA",
        0x42A0: "HANJA",
        0x42A1: "HANJA",
        0x42A2: "HANJA",
        0x42A3: "HANJA",
        0x42A4: "HANJA",
        0x42A5: "HANJA",
        0x42A6: "HANJA",
        0x42A7: "HANJA",
        0x42A8: "HANJA",
        0x42A9: "HANJA",
        0x42AA: "HANJA",
        0x42AB: "HANJA",
        0x42AC: "HANJA",
        0x42AD: "HANJA",
        0x42AE: "HANJA",
        0x42AF: "HANJA",
        0x42B0: "HANJA",
        0x42B1: "HANJA",
        0x42B2: "HANJA",
        0x42B3: "HANJA",
        0x42B4: "HANJA",
        0x42B5: "HANJA",
        0x42B6: "HANJA",
        0x42B7: "HANJA",
        0x42B8: "HANJA",
        0x42B9: "HANJA",
        0x42BA: "HANJA",
        0x42BB: "HANJA",
        0x42BC: "HANJA",
        0x42BD: "HANJA",
        0x42BE: "HANJA",
        0x42BF: "HANJA",
        0x42C0: "HANJA",
        0x42C1: "HANJA",
        0x42C2: "HANJA",
        0x42C3: "HANJA",
        0x42C4: "HANJA",
        0x42C5: "HANJA",
        0x42C6: "HANJA",
        0x42C7: "HANJA",
        0x42C8: "HANJA",
        0x42C9: "HANJA",
        0x42CA: "HANJA",
        0x42CB: "HANJA",
        0x42CC: "HANJA",
        0x42CD: "HANJA",
        0x42CE: "HANJA",
        0x42CF: "HANJA",
        0x42D0: "HANJA",
        0x42D1: "HANJA",
        0x42D2: "HANJA",
        0x42D3: "HANJA",
        0x42D4: "HANJA",
        0x42D5: "HANJA",
        0x42D6: "HANJA",
        0x42D7: "HANJA",
        0x42D8: "HANJA",
        0x42D9: "HANJA",
        0x42DA: "HANJA",
        0x42DB: "HANJA",
        0x42DC: "HANJA",
        0x42DD: "HANJA",
        0x42DE: "HANJA",
        0x42DF: "HANJA",
        0x42E0: "HANJA",
        0x42E1: "HANJA",
        0x42E2: "HANJA",
        0x42E3: "HANJA",
        0x42E4: "HANJA",
        0x42E5: "HANJA",
        0x42E6: "HANJA",
        0x42E7: "HANJA",
        0x42E8: "HANJA",
        0x42E9: "HANJA",
        0x42EA: "HANJA",
        0x42EB: "HANJA",
        0x42EC: "HANJA",
        0x42ED: "HANJA",
        0x42EE: "HANJA",
        0x42EF: "HANJA",
        0x42F0: "HANJA",
        0x42F1: "HANJA",
        0x42F2: "HANJA",
        0x42F3: "HANJA",
        0x42F4: "HANJA",
        0x42F5: "HANJA",
        0x42F6: "HANJA",
        0x42F7: "HANJA",
        0x42F8: "HANJA",
        0x42F9: "HANJA",
        0x42FA: "HANJA",
        0x42FB: "HANJA",
        0x42FC: "HANJA",
        0x42FD: "HANJA",
        0x42FE: "HANJA",
        0x42FF: "HANJA",
        0x4300: "HANJA",
        0x4301: "HANJA",
        0x4302: "HANJA",
        0x4303: "HANJA",
        0x4304: "HANJA",
        0x4305: "HANJA",
        0x4306: "HANJA",
        0x4307: "HANJA",
        0x4308: "HANJA",
        0x4309: "HANJA",
        0x430A: "HANJA",
        0x430B: "HANJA",
        0x430C: "HANJA",
        0x430D: "HANJA",
        0x430E: "HANJA",
        0x430F: "HANJA",
        0x4310: "HANJA",
        0x4311: "HANJA",
        0x4312: "HANJA",
        0x4313: "HANJA",
        0x4314: "HANJA",
        0x4315: "HANJA",
        0x4316: "HANJA",
        0x4317: "HANJA",
        0x4318: "HANJA",
        0x4319: "HANJA",
        0x431A: "HANJA",
        0x431B: "HANJA",
        0x431C: "HANJA",
        0x431D: "HANJA",
        0x431E: "HANJA",
        0x431F: "HANJA",
        0x4320: "HANJA",
        0x4321: "HANJA",
        0x4322: "HANJA",
        0x4323: "HANJA",
        0x4324: "HANJA",
        0x4325: "HANJA",
        0x4326: "HANJA",
        0x4327: "HANJA",
        0x4328: "HANJA",
        0x4329: "HANJA",
        0x432A: "HANJA",
        0x432B: "HANJA",
        0x432C: "HANJA",
        0x432D: "HANJA",
        0x432E: "HANJA",
        0x432F: "HANJA",
        0x4330: "HANJA",
        0x4331: "HANJA",
        0x4332: "HANJA",
        0x4333: "HANJA",
        0x4334: "HANJA",
        0x4335: "HANJA",
        0x4336: "HANJA",
        0x4337: "HANJA",
        0x4338: "HANJA",
        0x4339: "HANJA",
        0x433A: "HANJA",
        0x433B: "HANJA",
        0x433C: "HANJA",
        0x433D: "HANJA",
        0x433E: "HANJA",
        0x433F: "HANJA",
        0x4340: "HANJA",
        0x4341: "HANJA",
        0x4342: "HANJA",
        0x4343: "HANJA",
        0x4344: "HANJA",
        0x4345: "HANJA",
        0x4346: "HANJA",
        0x4347: "HANJA",
        0x4348: "HANJA",
        0x4349: "HANJA",
        0x434A: "HANJA",
        0x434B: "HANJA",
        0x434C: "HANJA",
        0x434D: "HANJA",
        0x434E: "HANJA",
        0x434F: "HANJA",
        0x4350: "HANJA",
        0x4351: "HANJA",
        0x4352: "HANJA",
        0x4353: "HANJA",
        0x4354: "HANJA",
        0x4355: "HANJA",
        0x4356: "HANJA",
        0x4357: "HANJA",
        0x4358: "HANJA",
        0x4359: "HANJA",
        0x435A: "HANJA",
        0x435B: "HANJA",
        0x435C: "HANJA",
        0x435D: "HANJA",
        0x435E: "HANJA",
        0x435F: "HANJA",
        0x4360: "HANJA",
        0x4361: "HANJA",
        0x4362: "HANJA",
        0x4363: "HANJA",
        0x4364: "HANJA",
        0x4365: "HANJA",
        0x4366: "HANJA",
        0x4367: "HANJA",
        0x4368: "HANJA",
        0x4369: "HANJA",
        0x436A: "HANJA",
        0x436B: "HANJA",
        0x436C: "HANJA",
        0x436D: "HANJA",
        0x436E: "HANJA",
        0x436F: "HANJA",
        0x4370: "HANJA",
        0x4371: "HANJA",
        0x4372: "HANJA",
        0x4373: "HANJA",
        0x4374: "HANJA",
        0x4375: "HANJA",
        0x4376: "HANJA",
        0x4377: "HANJA",
        0x4378: "HANJA",
        0x4379: "HANJA",
        0x437A: "HANJA",
        0x437B: "HANJA",
        0x437C: "HANJA",
        0x437D: "HANJA",
        0x437E: "HANJA",
        0x437F: "HANJA",
        0x4380: "HANJA",
        0x4381: "HANJA",
        0x4382: "HANJA",
        0x4383: "HANJA",
        0x4384: "HANJA",
        0x4385: "HANJA",
        0x4386: "HANJA",
        0x4387: "HANJA",
        0x4388: "HANJA",
        0x4389: "HANJA",
        0x438A: "HANJA",
        0x438B: "HANJA",
        0x438C: "HANJA",
        0x438D: "HANJA",
        0x438E: "HANJA",
        0x438F: "HANJA",
        0x4390: "HANJA",
        0x4391: "HANJA",
        0x4392: "HANJA",
        0x4393: "HANJA",
        0x4394: "HANJA",
        0x4395: "HANJA",
        0x4396: "HANJA",
        0x4397: "HANJA",
        0x4398: "HANJA",
        0x4399: "HANJA",
        0x439A: "HANJA",
        0x439B: "HANJA",
        0x439C: "HANJA",
        0x439D: "HANJA",
        0x439E: "HANJA",
        0x439F: "HANJA",
        0x43A0: "HANJA",
        0x43A1: "HANJA",
        0x43A2: "HANJA",
        0x43A3: "HANJA",
        0x43A4: "HANJA",
        0x43A5: "HANJA",
        0x43A6: "HANJA",
        0x43A7: "HANJA",
        0x43A8: "HANJA",
        0x43A9: "HANJA",
        0x43AA: "HANJA",
        0x43AB: "HANJA",
        0x43AC: "HANJA",
        0x43AD: "HANJA",
        0x43AE: "HANJA",
        0x43AF: "HANJA",
        0x43B0: "HANJA",
        0x43B1: "HANJA",
        0x43B2: "HANJA",
        0x43B3: "HANJA",
        0x43B4: "HANJA",
        0x43B5: "HANJA",
        0x43B6: "HANJA",
        0x43B7: "HANJA",
        0x43B8: "HANJA",
        0x43B9: "HANJA",
        0x43BA: "HANJA",
        0x43BB: "HANJA",
        0x43BC: "HANJA",
        0x43BD: "HANJA",
        0x43BE: "HANJA",
        0x43BF: "HANJA",
        0x43C0: "HANJA",
        0x43C1: "HANJA",
        0x43C2: "HANJA",
        0x43C3: "HANJA",
        0x43C4: "HANJA",
        0x43C5: "HANJA",
        0x43C6: "HANJA",
        0x43C7: "HANJA",
        0x43C8: "HANJA",
        0x43C9: "HANJA",
        0x43CA: "HANJA",
        0x43CB: "HANJA",
        0x43CC: "HANJA",
        0x43CD: "HANJA",
        0x43CE: "HANJA",
        0x43CF: "HANJA",
        0x43D0: "HANJA",
        0x43D1: "HANJA",
        0x43D2: "HANJA",
        0x43D3: "HANJA",
        0x43D4: "HANJA",
        0x43D5: "HANJA",
        0x43D6: "HANJA",
        0x43D7: "HANJA",
        0x43D8: "HANJA",
        0x43D9: "HANJA",
        0x43DA: "HANJA",
        0x43DB: "HANJA",
        0x43DC: "HANJA",
        0x43DD: "HANJA",
        0x43DE: "HANJA",
        0x43DF: "HANJA",
        0x43E0: "HANJA",
        0x43E1: "HANJA",
        0x43E2: "HANJA",
        0x43E3: "HANJA",
        0x43E4: "HANJA",
        0x43E5: "HANJA",
        0x43E6: "HANJA",
        0x43E7: "HANJA",
        0x43E8: "HANJA",
        0x43E9: "HANJA",
        0x43EA: "HANJA",
        0x43EB: "HANJA",
        0x43EC: "HANJA",
        0x43ED: "HANJA",
        0x43EE: "HANJA",
        0x43EF: "HANJA",
        0x43F0: "HANJA",
        0x43F1: "HANJA",
        0x43F2: "HANJA",
        0x43F3: "HANJA",
        0x43F4: "HANJA",
        0x43F5: "HANJA",
        0x43F6: "HANJA",
        0x43F7: "HANJA",
        0x43F8: "HANJA",
        0x43F9: "HANJA",
        0x43FA: "HANJA",
        0x43FB: "HANJA",
        0x43FC: "HANJA",
        0x43FD: "HANJA",
        0x43FE: "HANJA",
        0x43FF: "HANJA",
        0x4400: "HANJA",
        0x4401: "HANJA",
        0x4402: "HANJA",
        0x4403: "HANJA",
        0x4404: "HANJA",
        0x4405: "HANJA",
        0x4406: "HANJA",
        0x4407: "HANJA",
        0x4408: "HANJA",
        0x4409: "HANJA",
        0x440A: "HANJA",
        0x440B: "HANJA",
        0x440C: "HANJA",
        0x440D: "HANJA",
        0x440E: "HANJA",
        0x440F: "HANJA",
        0x4410: "HANJA",
        0x4411: "HANJA",
        0x4412: "HANJA",
        0x4413: "HANJA",
        0x4414: "HANJA",
        0x4415: "HANJA",
        0x4416: "HANJA",
        0x4417: "HANJA",
        0x4418: "HANJA",
        0x4419: "HANJA",
        0x441A: "HANJA",
        0x441B: "HANJA",
        0x441C: "HANJA",
        0x441D: "HANJA",
        0x441E: "HANJA",
        0x441F: "HANJA",
        0x4420: "HANJA",
        0x4421: "HANJA",
        0x4422: "HANJA",
        0x4423: "HANJA",
        0x4424: "HANJA",
        0x4425: "HANJA",
        0x4426: "HANJA",
        0x4427: "HANJA",
        0x4428: "HANJA",
        0x4429: "HANJA",
        0x442A: "HANJA",
        0x442B: "HANJA",
        0x442C: "HANJA",
        0x442D: "HANJA",
        0x442E: "HANJA",
        0x442F: "HANJA",
        0x4430: "HANJA",
        0x4431: "HANJA",
        0x4432: "HANJA",
        0x4433: "HANJA",
        0x4434: "HANJA",
        0x4435: "HANJA",
        0x4436: "HANJA",
        0x4437: "HANJA",
        0x4438: "HANJA",
        0x4439: "HANJA",
        0x443A: "HANJA",
        0x443B: "HANJA",
        0x443C: "HANJA",
        0x443D: "HANJA",
        0x443E: "HANJA",
        0x443F: "HANJA",
        0x4440: "HANJA",
        0x4441: "HANJA",
        0x4442: "HANJA",
        0x4443: "HANJA",
        0x4444: "HANJA",
        0x4445: "HANJA",
        0x4446: "HANJA",
        0x4447: "HANJA",
        0x4448: "HANJA",
        0x4449: "HANJA",
        0x444A: "HANJA",
        0x444B: "HANJA",
        0x444C: "HANJA",
        0x444D: "HANJA",
        0x444E: "HANJA",
        0x444F: "HANJA",
        0x4450: "HANJA",
        0x4451: "HANJA",
        0x4452: "HANJA",
        0x4453: "HANJA",
        0x4454: "HANJA",
        0x4455: "HANJA",
        0x4456: "HANJA",
        0x4457: "HANJA",
        0x4458: "HANJA",
        0x4459: "HANJA",
        0x445A: "HANJA",
        0x445B: "HANJA",
        0x445C: "HANJA",
        0x445D: "HANJA",
        0x445E: "HANJA",
        0x445F: "HANJA",
        0x4460: "HANJA",
        0x4461: "HANJA",
        0x4462: "HANJA",
        0x4463: "HANJA",
        0x4464: "HANJA",
        0x4465: "HANJA",
        0x4466: "HANJA",
        0x4467: "HANJA",
        0x4468: "HANJA",
        0x4469: "HANJA",
        0x446A: "HANJA",
        0x446B: "HANJA",
        0x446C: "HANJA",
        0x446D: "HANJA",
        0x446E: "HANJA",
        0x446F: "HANJA",
        0x4470: "HANJA",
        0x4471: "HANJA",
        0x4472: "HANJA",
        0x4473: "HANJA",
        0x4474: "HANJA",
        0x4475: "HANJA",
        0x4476: "HANJA",
        0x4477: "HANJA",
        0x4478: "HANJA",
        0x4479: "HANJA",
        0x447A: "HANJA",
        0x447B: "HANJA",
        0x447C: "HANJA",
        0x447D: "HANJA",
        0x447E: "HANJA",
        0x447F: "HANJA",
        0x4480: "HANJA",
        0x4481: "HANJA",
        0x4482: "HANJA",
        0x4483: "HANJA",
        0x4484: "HANJA",
        0x4485: "HANJA",
        0x4486: "HANJA",
        0x4487: "HANJA",
        0x4488: "HANJA",
        0x4489: "HANJA",
        0x448A: "HANJA",
        0x448B: "HANJA",
        0x448C: "HANJA",
        0x448D: "HANJA",
        0x448E: "HANJA",
        0x448F: "HANJA",
        0x4490: "HANJA",
        0x4491: "HANJA",
        0x4492: "HANJA",
        0x4493: "HANJA",
        0x4494: "HANJA",
        0x4495: "HANJA",
        0x4496: "HANJA",
        0x4497: "HANJA",
        0x4498: "HANJA",
        0x4499: "HANJA",
        0x449A: "HANJA",
        0x449B: "HANJA",
        0x449C: "HANJA",
        0x449D: "HANJA",
        0x449E: "HANJA",
        0x449F: "HANJA",
        0x44A0: "HANJA",
        0x44A1: "HANJA",
        0x44A2: "HANJA",
        0x44A3: "HANJA",
        0x44A4: "HANJA",
        0x44A5: "HANJA",
        0x44A6: "HANJA",
        0x44A7: "HANJA",
        0x44A8: "HANJA",
        0x44A9: "HANJA",
        0x44AA: "HANJA",
        0x44AB: "HANJA",
        0x44AC: "HANJA",
        0x44AD: "HANJA",
        0x44AE: "HANJA",
        0x44AF: "HANJA",
        0x44B0: "HANJA",
        0x44B1: "HANJA",
        0x44B2: "HANJA",
        0x44B3: "HANJA",
        0x44B4: "HANJA",
        0x44B5: "HANJA",
        0x44B6: "HANJA",
        0x44B7: "HANJA",
        0x44B8: "HANJA",
        0x44B9: "HANJA",
        0x44BA: "HANJA",
        0x44BB: "HANJA",
        0x44BC: "HANJA",
        0x44BD: "HANJA",
        0x44BE: "HANJA",
        0x44BF: "HANJA",
        0x44C0: "HANJA",
        0x44C1: "HANJA",
        0x44C2: "HANJA",
        0x44C3: "HANJA",
        0x44C4: "HANJA",
        0x44C5: "HANJA",
        0x44C6: "HANJA",
        0x44C7: "HANJA",
        0x44C8: "HANJA",
        0x44C9: "HANJA",
        0x44CA: "HANJA",
        0x44CB: "HANJA",
        0x44CC: "HANJA",
        0x44CD: "HANJA",
        0x44CE: "HANJA",
        0x44CF: "HANJA",
        0x44D0: "HANJA",
        0x44D1: "HANJA",
        0x44D2: "HANJA",
        0x44D3: "HANJA",
        0x44D4: "HANJA",
        0x44D5: "HANJA",
        0x44D6: "HANJA",
        0x44D7: "HANJA",
        0x44D8: "HANJA",
        0x44D9: "HANJA",
        0x44DA: "HANJA",
        0x44DB: "HANJA",
        0x44DC: "HANJA",
        0x44DD: "HANJA",
        0x44DE: "HANJA",
        0x44DF: "HANJA",
        0x44E0: "HANJA",
        0x44E1: "HANJA",
        0x44E2: "HANJA",
        0x44E3: "HANJA",
        0x44E4: "HANJA",
        0x44E5: "HANJA",
        0x44E6: "HANJA",
        0x44E7: "HANJA",
        0x44E8: "HANJA",
        0x44E9: "HANJA",
        0x44EA: "HANJA",
        0x44EB: "HANJA",
        0x44EC: "HANJA",
        0x44ED: "HANJA",
        0x44EE: "HANJA",
        0x44EF: "HANJA",
        0x44F0: "HANJA",
        0x44F1: "HANJA",
        0x44F2: "HANJA",
        0x44F3: "HANJA",
        0x44F4: "HANJA",
        0x44F5: "HANJA",
        0x44F6: "HANJA",
        0x44F7: "HANJA",
        0x44F8: "HANJA",
        0x44F9: "HANJA",
        0x44FA: "HANJA",
        0x44FB: "HANJA",
        0x44FC: "HANJA",
        0x44FD: "HANJA",
        0x44FE: "HANJA",
        0x44FF: "HANJA",
        0x4500: "HANJA",
        0x4501: "HANJA",
        0x4502: "HANJA",
        0x4503: "HANJA",
        0x4504: "HANJA",
        0x4505: "HANJA",
        0x4506: "HANJA",
        0x4507: "HANJA",
        0x4508: "HANJA",
        0x4509: "HANJA",
        0x450A: "HANJA",
        0x450B: "HANJA",
        0x450C: "HANJA",
        0x450D: "HANJA",
        0x450E: "HANJA",
        0x450F: "HANJA",
        0x4510: "HANJA",
        0x4511: "HANJA",
        0x4512: "HANJA",
        0x4513: "HANJA",
        0x4514: "HANJA",
        0x4515: "HANJA",
        0x4516: "HANJA",
        0x4517: "HANJA",
        0x4518: "HANJA",
        0x4519: "HANJA",
        0x451A: "HANJA",
        0x451B: "HANJA",
        0x451C: "HANJA",
        0x451D: "HANJA",
        0x451E: "HANJA",
        0x451F: "HANJA",
        0x4520: "HANJA",
        0x4521: "HANJA",
        0x4522: "HANJA",
        0x4523: "HANJA",
        0x4524: "HANJA",
        0x4525: "HANJA",
        0x4526: "HANJA",
        0x4527: "HANJA",
        0x4528: "HANJA",
        0x4529: "HANJA",
        0x452A: "HANJA",
        0x452B: "HANJA",
        0x452C: "HANJA",
        0x452D: "HANJA",
        0x452E: "HANJA",
        0x452F: "HANJA",
        0x4530: "HANJA",
        0x4531: "HANJA",
        0x4532: "HANJA",
        0x4533: "HANJA",
        0x4534: "HANJA",
        0x4535: "HANJA",
        0x4536: "HANJA",
        0x4537: "HANJA",
        0x4538: "HANJA",
        0x4539: "HANJA",
        0x453A: "HANJA",
        0x453B: "HANJA",
        0x453C: "HANJA",
        0x453D: "HANJA",
        0x453E: "HANJA",
        0x453F: "HANJA",
        0x4540: "HANJA",
        0x4541: "HANJA",
        0x4542: "HANJA",
        0x4543: "HANJA",
        0x4544: "HANJA",
        0x4545: "HANJA",
        0x4546: "HANJA",
        0x4547: "HANJA",
        0x4548: "HANJA",
        0x4549: "HANJA",
        0x454A: "HANJA",
        0x454B: "HANJA",
        0x454C: "HANJA",
        0x454D: "HANJA",
        0x454E: "HANJA",
        0x454F: "HANJA",
        0x4550: "HANJA",
        0x4551: "HANJA",
        0x4552: "HANJA",
        0x4553: "HANJA",
        0x4554: "HANJA",
        0x4555: "HANJA",
        0x4556: "HANJA",
        0x4557: "HANJA",
        0x4558: "HANJA",
        0x4559: "HANJA",
        0x455A: "HANJA",
        0x455B: "HANJA",
        0x455C: "HANJA",
        0x455D: "HANJA",
        0x455E: "HANJA",
        0x455F: "HANJA",
        0x4560: "HANJA",
        0x4561: "HANJA",
        0x4562: "HANJA",
        0x4563: "HANJA",
        0x4564: "HANJA",
        0x4565: "HANJA",
        0x4566: "HANJA",
        0x4567: "HANJA",
        0x4568: "HANJA",
        0x4569: "HANJA",
        0x456A: "HANJA",
        0x456B: "HANJA",
        0x456C: "HANJA",
        0x456D: "HANJA",
        0x456E: "HANJA",
        0x456F: "HANJA",
        0x4570: "HANJA",
        0x4571: "HANJA",
        0x4572: "HANJA",
        0x4573: "HANJA",
        0x4574: "HANJA",
        0x4575: "HANJA",
        0x4576: "HANJA",
        0x4577: "HANJA",
        0x4578: "HANJA",
        0x4579: "HANJA",
        0x457A: "HANJA",
        0x457B: "HANJA",
        0x457C: "HANJA",
        0x457D: "HANJA",
        0x457E: "HANJA",
        0x457F: "HANJA",
        0x4580: "HANJA",
        0x4581: "HANJA",
        0x4582: "HANJA",
        0x4583: "HANJA",
        0x4584: "HANJA",
        0x4585: "HANJA",
        0x4586: "HANJA",
        0x4587: "HANJA",
        0x4588: "HANJA",
        0x4589: "HANJA",
        0x458A: "HANJA",
        0x458B: "HANJA",
        0x458C: "HANJA",
        0x458D: "HANJA",
        0x458E: "HANJA",
        0x458F: "HANJA",
        0x4590: "HANJA",
        0x4591: "HANJA",
        0x4592: "HANJA",
        0x4593: "HANJA",
        0x4594: "HANJA",
        0x4595: "HANJA",
        0x4596: "HANJA",
        0x4597: "HANJA",
        0x4598: "HANJA",
        0x4599: "HANJA",
        0x459A: "HANJA",
        0x459B: "HANJA",
        0x459C: "HANJA",
        0x459D: "HANJA",
        0x459E: "HANJA",
        0x459F: "HANJA",
        0x45A0: "HANJA",
        0x45A1: "HANJA",
        0x45A2: "HANJA",
        0x45A3: "HANJA",
        0x45A4: "HANJA",
        0x45A5: "HANJA",
        0x45A6: "HANJA",
        0x45A7: "HANJA",
        0x45A8: "HANJA",
        0x45A9: "HANJA",
        0x45AA: "HANJA",
        0x45AB: "HANJA",
        0x45AC: "HANJA",
        0x45AD: "HANJA",
        0x45AE: "HANJA",
        0x45AF: "HANJA",
        0x45B0: "HANJA",
        0x45B1: "HANJA",
        0x45B2: "HANJA",
        0x45B3: "HANJA",
        0x45B4: "HANJA",
        0x45B5: "HANJA",
        0x45B6: "HANJA",
        0x45B7: "HANJA",
        0x45B8: "HANJA",
        0x45B9: "HANJA",
        0x45BA: "HANJA",
        0x45BB: "HANJA",
        0x45BC: "HANJA",
        0x45BD: "HANJA",
        0x45BE: "HANJA",
        0x45BF: "HANJA",
        0x45C0: "HANJA",
        0x45C1: "HANJA",
        0x45C2: "HANJA",
        0x45C3: "HANJA",
        0x45C4: "HANJA",
        0x45C5: "HANJA",
        0x45C6: "HANJA",
        0x45C7: "HANJA",
        0x45C8: "HANJA",
        0x45C9: "HANJA",
        0x45CA: "HANJA",
        0x45CB: "HANJA",
        0x45CC: "HANJA",
        0x45CD: "HANJA",
        0x45CE: "HANJA",
        0x45CF: "HANJA",
        0x45D0: "HANJA",
        0x45D1: "HANJA",
        0x45D2: "HANJA",
        0x45D3: "HANJA",
        0x45D4: "HANJA",
        0x45D5: "HANJA",
        0x45D6: "HANJA",
        0x45D7: "HANJA",
        0x45D8: "HANJA",
        0x45D9: "HANJA",
        0x45DA: "HANJA",
        0x45DB: "HANJA",
        0x45DC: "HANJA",
        0x45DD: "HANJA",
        0x45DE: "HANJA",
        0x45DF: "HANJA",
        0x45E0: "HANJA",
        0x45E1: "HANJA",
        0x45E2: "HANJA",
        0x45E3: "HANJA",
        0x45E4: "HANJA",
        0x45E5: "HANJA",
        0x45E6: "HANJA",
        0x45E7: "HANJA",
        0x45E8: "HANJA",
        0x45E9: "HANJA",
        0x45EA: "HANJA",
        0x45EB: "HANJA",
        0x45EC: "HANJA",
        0x45ED: "HANJA",
        0x45EE: "HANJA",
        0x45EF: "HANJA",
        0x45F0: "HANJA",
        0x45F1: "HANJA",
        0x45F2: "HANJA",
        0x45F3: "HANJA",
        0x45F4: "HANJA",
        0x45F5: "HANJA",
        0x45F6: "HANJA",
        0x45F7: "HANJA",
        0x45F8: "HANJA",
        0x45F9: "HANJA",
        0x45FA: "HANJA",
        0x45FB: "HANJA",
        0x45FC: "HANJA",
        0x45FD: "HANJA",
        0x45FE: "HANJA",
        0x45FF: "HANJA",
        0x4600: "HANJA",
        0x4601: "HANJA",
        0x4602: "HANJA",
        0x4603: "HANJA",
        0x4604: "HANJA",
        0x4605: "HANJA",
        0x4606: "HANJA",
        0x4607: "HANJA",
        0x4608: "HANJA",
        0x4609: "HANJA",
        0x460A: "HANJA",
        0x460B: "HANJA",
        0x460C: "HANJA",
        0x460D: "HANJA",
        0x460E: "HANJA",
        0x460F: "HANJA",
        0x4610: "HANJA",
        0x4611: "HANJA",
        0x4612: "HANJA",
        0x4613: "HANJA",
        0x4614: "HANJA",
        0x4615: "HANJA",
        0x4616: "HANJA",
        0x4617: "HANJA",
        0x4618: "HANJA",
        0x4619: "HANJA",
        0x461A: "HANJA",
        0x461B: "HANJA",
        0x461C: "HANJA",
        0x461D: "HANJA",
        0x461E: "HANJA",
        0x461F: "HANJA",
        0x4620: "HANJA",
        0x4621: "HANJA",
        0x4622: "HANJA",
        0x4623: "HANJA",
        0x4624: "HANJA",
        0x4625: "HANJA",
        0x4626: "HANJA",
        0x4627: "HANJA",
        0x4628: "HANJA",
        0x4629: "HANJA",
        0x462A: "HANJA",
        0x462B: "HANJA",
        0x462C: "HANJA",
        0x462D: "HANJA",
        0x462E: "HANJA",
        0x462F: "HANJA",
        0x4630: "HANJA",
        0x4631: "HANJA",
        0x4632: "HANJA",
        0x4633: "HANJA",
        0x4634: "HANJA",
        0x4635: "HANJA",
        0x4636: "HANJA",
        0x4637: "HANJA",
        0x4638: "HANJA",
        0x4639: "HANJA",
        0x463A: "HANJA",
        0x463B: "HANJA",
        0x463C: "HANJA",
        0x463D: "HANJA",
        0x463E: "HANJA",
        0x463F: "HANJA",
        0x4640: "HANJA",
        0x4641: "HANJA",
        0x4642: "HANJA",
        0x4643: "HANJA",
        0x4644: "HANJA",
        0x4645: "HANJA",
        0x4646: "HANJA",
        0x4647: "HANJA",
        0x4648: "HANJA",
        0x4649: "HANJA",
        0x464A: "HANJA",
        0x464B: "HANJA",
        0x464C: "HANJA",
        0x464D: "HANJA",
        0x464E: "HANJA",
        0x464F: "HANJA",
        0x4650: "HANJA",
        0x4651: "HANJA",
        0x4652: "HANJA",
        0x4653: "HANJA",
        0x4654: "HANJA",
        0x4655: "HANJA",
        0x4656: "HANJA",
        0x4657: "HANJA",
        0x4658: "HANJA",
        0x4659: "HANJA",
        0x465A: "HANJA",
        0x465B: "HANJA",
        0x465C: "HANJA",
        0x465D: "HANJA",
        0x465E: "HANJA",
        0x465F: "HANJA",
        0x4660: "HANJA",
        0x4661: "HANJA",
        0x4662: "HANJA",
        0x4663: "HANJA",
        0x4664: "HANJA",
        0x4665: "HANJA",
        0x4666: "HANJA",
        0x4667: "HANJA",
        0x4668: "HANJA",
        0x4669: "HANJA",
        0x466A: "HANJA",
        0x466B: "HANJA",
        0x466C: "HANJA",
        0x466D: "HANJA",
        0x466E: "HANJA",
        0x466F: "HANJA",
        0x4670: "HANJA",
        0x4671: "HANJA",
        0x4672: "HANJA",
        0x4673: "HANJA",
        0x4674: "HANJA",
        0x4675: "HANJA",
        0x4676: "HANJA",
        0x4677: "HANJA",
        0x4678: "HANJA",
        0x4679: "HANJA",
        0x467A: "HANJA",
        0x467B: "HANJA",
        0x467C: "HANJA",
        0x467D: "HANJA",
        0x467E: "HANJA",
        0x467F: "HANJA",
        0x4680: "HANJA",
        0x4681: "HANJA",
        0x4682: "HANJA",
        0x4683: "HANJA",
        0x4684: "HANJA",
        0x4685: "HANJA",
        0x4686: "HANJA",
        0x4687: "HANJA",
        0x4688: "HANJA",
        0x4689: "HANJA",
        0x468A: "HANJA",
        0x468B: "HANJA",
        0x468C: "HANJA",
        0x468D: "HANJA",
        0x468E: "HANJA",
        0x468F: "HANJA",
        0x4690: "HANJA",
        0x4691: "HANJA",
        0x4692: "HANJA",
        0x4693: "HANJA",
        0x4694: "HANJA",
        0x4695: "HANJA",
        0x4696: "HANJA",
        0x4697: "HANJA",
        0x4698: "HANJA",
        0x4699: "HANJA",
        0x469A: "HANJA",
        0x469B: "HANJA",
        0x469C: "HANJA",
        0x469D: "HANJA",
        0x469E: "HANJA",
        0x469F: "HANJA",
        0x46A0: "HANJA",
        0x46A1: "HANJA",
        0x46A2: "HANJA",
        0x46A3: "HANJA",
        0x46A4: "HANJA",
        0x46A5: "HANJA",
        0x46A6: "HANJA",
        0x46A7: "HANJA",
        0x46A8: "HANJA",
        0x46A9: "HANJA",
        0x46AA: "HANJA",
        0x46AB: "HANJA",
        0x46AC: "HANJA",
        0x46AD: "HANJA",
        0x46AE: "HANJA",
        0x46AF: "HANJA",
        0x46B0: "HANJA",
        0x46B1: "HANJA",
        0x46B2: "HANJA",
        0x46B3: "HANJA",
        0x46B4: "HANJA",
        0x46B5: "HANJA",
        0x46B6: "HANJA",
        0x46B7: "HANJA",
        0x46B8: "HANJA",
        0x46B9: "HANJA",
        0x46BA: "HANJA",
        0x46BB: "HANJA",
        0x46BC: "HANJA",
        0x46BD: "HANJA",
        0x46BE: "HANJA",
        0x46BF: "HANJA",
        0x46C0: "HANJA",
        0x46C1: "HANJA",
        0x46C2: "HANJA",
        0x46C3: "HANJA",
        0x46C4: "HANJA",
        0x46C5: "HANJA",
        0x46C6: "HANJA",
        0x46C7: "HANJA",
        0x46C8: "HANJA",
        0x46C9: "HANJA",
        0x46CA: "HANJA",
        0x46CB: "HANJA",
        0x46CC: "HANJA",
        0x46CD: "HANJA",
        0x46CE: "HANJA",
        0x46CF: "HANJA",
        0x46D0: "HANJA",
        0x46D1: "HANJA",
        0x46D2: "HANJA",
        0x46D3: "HANJA",
        0x46D4: "HANJA",
        0x46D5: "HANJA",
        0x46D6: "HANJA",
        0x46D7: "HANJA",
        0x46D8: "HANJA",
        0x46D9: "HANJA",
        0x46DA: "HANJA",
        0x46DB: "HANJA",
        0x46DC: "HANJA",
        0x46DD: "HANJA",
        0x46DE: "HANJA",
        0x46DF: "HANJA",
        0x46E0: "HANJA",
        0x46E1: "HANJA",
        0x46E2: "HANJA",
        0x46E3: "HANJA",
        0x46E4: "HANJA",
        0x46E5: "HANJA",
        0x46E6: "HANJA",
        0x46E7: "HANJA",
        0x46E8: "HANJA",
        0x46E9: "HANJA",
        0x46EA: "HANJA",
        0x46EB: "HANJA",
        0x46EC: "HANJA",
        0x46ED: "HANJA",
        0x46EE: "HANJA",
        0x46EF: "HANJA",
        0x46F0: "HANJA",
        0x46F1: "HANJA",
        0x46F2: "HANJA",
        0x46F3: "HANJA",
        0x46F4: "HANJA",
        0x46F5: "HANJA",
        0x46F6: "HANJA",
        0x46F7: "HANJA",
        0x46F8: "HANJA",
        0x46F9: "HANJA",
        0x46FA: "HANJA",
        0x46FB: "HANJA",
        0x46FC: "HANJA",
        0x46FD: "HANJA",
        0x46FE: "HANJA",
        0x46FF: "HANJA",
        0x4700: "HANJA",
        0x4701: "HANJA",
        0x4702: "HANJA",
        0x4703: "HANJA",
        0x4704: "HANJA",
        0x4705: "HANJA",
        0x4706: "HANJA",
        0x4707: "HANJA",
        0x4708: "HANJA",
        0x4709: "HANJA",
        0x470A: "HANJA",
        0x470B: "HANJA",
        0x470C: "HANJA",
        0x470D: "HANJA",
        0x470E: "HANJA",
        0x470F: "HANJA",
        0x4710: "HANJA",
        0x4711: "HANJA",
        0x4712: "HANJA",
        0x4713: "HANJA",
        0x4714: "HANJA",
        0x4715: "HANJA",
        0x4716: "HANJA",
        0x4717: "HANJA",
        0x4718: "HANJA",
        0x4719: "HANJA",
        0x471A: "HANJA",
        0x471B: "HANJA",
        0x471C: "HANJA",
        0x471D: "HANJA",
        0x471E: "HANJA",
        0x471F: "HANJA",
        0x4720: "HANJA",
        0x4721: "HANJA",
        0x4722: "HANJA",
        0x4723: "HANJA",
        0x4724: "HANJA",
        0x4725: "HANJA",
        0x4726: "HANJA",
        0x4727: "HANJA",
        0x4728: "HANJA",
        0x4729: "HANJA",
        0x472A: "HANJA",
        0x472B: "HANJA",
        0x472C: "HANJA",
        0x472D: "HANJA",
        0x472E: "HANJA",
        0x472F: "HANJA",
        0x4730: "HANJA",
        0x4731: "HANJA",
        0x4732: "HANJA",
        0x4733: "HANJA",
        0x4734: "HANJA",
        0x4735: "HANJA",
        0x4736: "HANJA",
        0x4737: "HANJA",
        0x4738: "HANJA",
        0x4739: "HANJA",
        0x473A: "HANJA",
        0x473B: "HANJA",
        0x473C: "HANJA",
        0x473D: "HANJA",
        0x473E: "HANJA",
        0x473F: "HANJA",
        0x4740: "HANJA",
        0x4741: "HANJA",
        0x4742: "HANJA",
        0x4743: "HANJA",
        0x4744: "HANJA",
        0x4745: "HANJA",
        0x4746: "HANJA",
        0x4747: "HANJA",
        0x4748: "HANJA",
        0x4749: "HANJA",
        0x474A: "HANJA",
        0x474B: "HANJA",
        0x474C: "HANJA",
        0x474D: "HANJA",
        0x474E: "HANJA",
        0x474F: "HANJA",
        0x4750: "HANJA",
        0x4751: "HANJA",
        0x4752: "HANJA",
        0x4753: "HANJA",
        0x4754: "HANJA",
        0x4755: "HANJA",
        0x4756: "HANJA",
        0x4757: "HANJA",
        0x4758: "HANJA",
        0x4759: "HANJA",
        0x475A: "HANJA",
        0x475B: "HANJA",
        0x475C: "HANJA",
        0x475D: "HANJA",
        0x475E: "HANJA",
        0x475F: "HANJA",
        0x4760: "HANJA",
        0x4761: "HANJA",
        0x4762: "HANJA",
        0x4763: "HANJA",
        0x4764: "HANJA",
        0x4765: "HANJA",
        0x4766: "HANJA",
        0x4767: "HANJA",
        0x4768: "HANJA",
        0x4769: "HANJA",
        0x476A: "HANJA",
        0x476B: "HANJA",
        0x476C: "HANJA",
        0x476D: "HANJA",
        0x476E: "HANJA",
        0x476F: "HANJA",
        0x4770: "HANJA",
        0x4771: "HANJA",
        0x4772: "HANJA",
        0x4773: "HANJA",
        0x4774: "HANJA",
        0x4775: "HANJA",
        0x4776: "HANJA",
        0x4777: "HANJA",
        0x4778: "HANJA",
        0x4779: "HANJA",
        0x477A: "HANJA",
        0x477B: "HANJA",
        0x477C: "HANJA",
        0x477D: "HANJA",
        0x477E: "HANJA",
        0x477F: "HANJA",
        0x4780: "HANJA",
        0x4781: "HANJA",
        0x4782: "HANJA",
        0x4783: "HANJA",
        0x4784: "HANJA",
        0x4785: "HANJA",
        0x4786: "HANJA",
        0x4787: "HANJA",
        0x4788: "HANJA",
        0x4789: "HANJA",
        0x478A: "HANJA",
        0x478B: "HANJA",
        0x478C: "HANJA",
        0x478D: "HANJA",
        0x478E: "HANJA",
        0x478F: "HANJA",
        0x4790: "HANJA",
        0x4791: "HANJA",
        0x4792: "HANJA",
        0x4793: "HANJA",
        0x4794: "HANJA",
        0x4795: "HANJA",
        0x4796: "HANJA",
        0x4797: "HANJA",
        0x4798: "HANJA",
        0x4799: "HANJA",
        0x479A: "HANJA",
        0x479B: "HANJA",
        0x479C: "HANJA",
        0x479D: "HANJA",
        0x479E: "HANJA",
        0x479F: "HANJA",
        0x47A0: "HANJA",
        0x47A1: "HANJA",
        0x47A2: "HANJA",
        0x47A3: "HANJA",
        0x47A4: "HANJA",
        0x47A5: "HANJA",
        0x47A6: "HANJA",
        0x47A7: "HANJA",
        0x47A8: "HANJA",
        0x47A9: "HANJA",
        0x47AA: "HANJA",
        0x47AB: "HANJA",
        0x47AC: "HANJA",
        0x47AD: "HANJA",
        0x47AE: "HANJA",
        0x47AF: "HANJA",
        0x47B0: "HANJA",
        0x47B1: "HANJA",
        0x47B2: "HANJA",
        0x47B3: "HANJA",
        0x47B4: "HANJA",
        0x47B5: "HANJA",
        0x47B6: "HANJA",
        0x47B7: "HANJA",
        0x47B8: "HANJA",
        0x47B9: "HANJA",
        0x47BA: "HANJA",
        0x47BB: "HANJA",
        0x47BC: "HANJA",
        0x47BD: "HANJA",
        0x47BE: "HANJA",
        0x47BF: "HANJA",
        0x47C0: "HANJA",
        0x47C1: "HANJA",
        0x47C2: "HANJA",
        0x47C3: "HANJA",
        0x47C4: "HANJA",
        0x47C5: "HANJA",
        0x47C6: "HANJA",
        0x47C7: "HANJA",
        0x47C8: "HANJA",
        0x47C9: "HANJA",
        0x47CA: "HANJA",
        0x47CB: "HANJA",
        0x47CC: "HANJA",
        0x47CD: "HANJA",
        0x47CE: "HANJA",
        0x47CF: "HANJA",
        0x47D0: "HANJA",
        0x47D1: "HANJA",
        0x47D2: "HANJA",
        0x47D3: "HANJA",
        0x47D4: "HANJA",
        0x47D5: "HANJA",
        0x47D6: "HANJA",
        0x47D7: "HANJA",
        0x47D8: "HANJA",
        0x47D9: "HANJA",
        0x47DA: "HANJA",
        0x47DB: "HANJA",
        0x47DC: "HANJA",
        0x47DD: "HANJA",
        0x47DE: "HANJA",
        0x47DF: "HANJA",
        0x47E0: "HANJA",
        0x47E1: "HANJA",
        0x47E2: "HANJA",
        0x47E3: "HANJA",
        0x47E4: "HANJA",
        0x47E5: "HANJA",
        0x47E6: "HANJA",
        0x47E7: "HANJA",
        0x47E8: "HANJA",
        0x47E9: "HANJA",
        0x47EA: "HANJA",
        0x47EB: "HANJA",
        0x47EC: "HANJA",
        0x47ED: "HANJA",
        0x47EE: "HANJA",
        0x47EF: "HANJA",
        0x47F0: "HANJA",
        0x47F1: "HANJA",
        0x47F2: "HANJA",
        0x47F3: "HANJA",
        0x47F4: "HANJA",
        0x47F5: "HANJA",
        0x47F6: "HANJA",
        0x47F7: "HANJA",
        0x47F8: "HANJA",
        0x47F9: "HANJA",
        0x47FA: "HANJA",
        0x47FB: "HANJA",
        0x47FC: "HANJA",
        0x47FD: "HANJA",
        0x47FE: "HANJA",
        0x47FF: "HANJA",
        0x4800: "HANJA",
        0x4801: "HANJA",
        0x4802: "HANJA",
        0x4803: "HANJA",
        0x4804: "HANJA",
        0x4805: "HANJA",
        0x4806: "HANJA",
        0x4807: "HANJA",
        0x4808: "HANJA",
        0x4809: "HANJA",
        0x480A: "HANJA",
        0x480B: "HANJA",
        0x480C: "HANJA",
        0x480D: "HANJA",
        0x480E: "HANJA",
        0x480F: "HANJA",
        0x4810: "HANJA",
        0x4811: "HANJA",
        0x4812: "HANJA",
        0x4813: "HANJA",
        0x4814: "HANJA",
        0x4815: "HANJA",
        0x4816: "HANJA",
        0x4817: "HANJA",
        0x4818: "HANJA",
        0x4819: "HANJA",
        0x481A: "HANJA",
        0x481B: "HANJA",
        0x481C: "HANJA",
        0x481D: "HANJA",
        0x481E: "HANJA",
        0x481F: "HANJA",
        0x4820: "HANJA",
        0x4821: "HANJA",
        0x4822: "HANJA",
        0x4823: "HANJA",
        0x4824: "HANJA",
        0x4825: "HANJA",
        0x4826: "HANJA",
        0x4827: "HANJA",
        0x4828: "HANJA",
        0x4829: "HANJA",
        0x482A: "HANJA",
        0x482B: "HANJA",
        0x482C: "HANJA",
        0x482D: "HANJA",
        0x482E: "HANJA",
        0x482F: "HANJA",
        0x4830: "HANJA",
        0x4831: "HANJA",
        0x4832: "HANJA",
        0x4833: "HANJA",
        0x4834: "HANJA",
        0x4835: "HANJA",
        0x4836: "HANJA",
        0x4837: "HANJA",
        0x4838: "HANJA",
        0x4839: "HANJA",
        0x483A: "HANJA",
        0x483B: "HANJA",
        0x483C: "HANJA",
        0x483D: "HANJA",
        0x483E: "HANJA",
        0x483F: "HANJA",
        0x4840: "HANJA",
        0x4841: "HANJA",
        0x4842: "HANJA",
        0x4843: "HANJA",
        0x4844: "HANJA",
        0x4845: "HANJA",
        0x4846: "HANJA",
        0x4847: "HANJA",
        0x4848: "HANJA",
        0x4849: "HANJA",
        0x484A: "HANJA",
        0x484B: "HANJA",
        0x484C: "HANJA",
        0x484D: "HANJA",
        0x484E: "HANJA",
        0x484F: "HANJA",
        0x4850: "HANJA",
        0x4851: "HANJA",
        0x4852: "HANJA",
        0x4853: "HANJA",
        0x4854: "HANJA",
        0x4855: "HANJA",
        0x4856: "HANJA",
        0x4857: "HANJA",
        0x4858: "HANJA",
        0x4859: "HANJA",
        0x485A: "HANJA",
        0x485B: "HANJA",
        0x485C: "HANJA",
        0x485D: "HANJA",
        0x485E: "HANJA",
        0x485F: "HANJA",
        0x4860: "HANJA",
        0x4861: "HANJA",
        0x4862: "HANJA",
        0x4863: "HANJA",
        0x4864: "HANJA",
        0x4865: "HANJA",
        0x4866: "HANJA",
        0x4867: "HANJA",
        0x4868: "HANJA",
        0x4869: "HANJA",
        0x486A: "HANJA",
        0x486B: "HANJA",
        0x486C: "HANJA",
        0x486D: "HANJA",
        0x486E: "HANJA",
        0x486F: "HANJA",
        0x4870: "HANJA",
        0x4871: "HANJA",
        0x4872: "HANJA",
        0x4873: "HANJA",
        0x4874: "HANJA",
        0x4875: "HANJA",
        0x4876: "HANJA",
        0x4877: "HANJA",
        0x4878: "HANJA",
        0x4879: "HANJA",
        0x487A: "HANJA",
        0x487B: "HANJA",
        0x487C: "HANJA",
        0x487D: "HANJA",
        0x487E: "HANJA",
        0x487F: "HANJA",
        0x4880: "HANJA",
        0x4881: "HANJA",
        0x4882: "HANJA",
        0x4883: "HANJA",
        0x4884: "HANJA",
        0x4885: "HANJA",
        0x4886: "HANJA",
        0x4887: "HANJA",
        0x4888: "HANJA",
        0x4889: "HANJA",
        0x488A: "HANJA",
        0x488B: "HANJA",
        0x488C: "HANJA",
        0x488D: "HANJA",
        0x488E: "HANJA",
        0x488F: "HANJA",
        0x4890: "HANJA",
        0x4891: "HANJA",
        0x4892: "HANJA",
        0x4893: "HANJA",
        0x4894: "HANJA",
        0x4895: "HANJA",
        0x4896: "HANJA",
        0x4897: "HANJA",
        0x4898: "HANJA",
        0x4899: "HANJA",
        0x489A: "HANJA",
        0x489B: "HANJA",
        0x489C: "HANJA",
        0x489D: "HANJA",
        0x489E: "HANJA",
        0x489F: "HANJA",
        0x48A0: "HANJA",
        0x48A1: "HANJA",
        0x48A2: "HANJA",
        0x48A3: "HANJA",
        0x48A4: "HANJA",
        0x48A5: "HANJA",
        0x48A6: "HANJA",
        0x48A7: "HANJA",
        0x48A8: "HANJA",
        0x48A9: "HANJA",
        0x48AA: "HANJA",
        0x48AB: "HANJA",
        0x48AC: "HANJA",
        0x48AD: "HANJA",
        0x48AE: "HANJA",
        0x48AF: "HANJA",
        0x48B0: "HANJA",
        0x48B1: "HANJA",
        0x48B2: "HANJA",
        0x48B3: "HANJA",
        0x48B4: "HANJA",
        0x48B5: "HANJA",
        0x48B6: "HANJA",
        0x48B7: "HANJA",
        0x48B8: "HANJA",
        0x48B9: "HANJA",
        0x48BA: "HANJA",
        0x48BB: "HANJA",
        0x48BC: "HANJA",
        0x48BD: "HANJA",
        0x48BE: "HANJA",
        0x48BF: "HANJA",
        0x48C0: "HANJA",
        0x48C1: "HANJA",
        0x48C2: "HANJA",
        0x48C3: "HANJA",
        0x48C4: "HANJA",
        0x48C5: "HANJA",
        0x48C6: "HANJA",
        0x48C7: "HANJA",
        0x48C8: "HANJA",
        0x48C9: "HANJA",
        0x48CA: "HANJA",
        0x48CB: "HANJA",
        0x48CC: "HANJA",
        0x48CD: "HANJA",
        0x48CE: "HANJA",
        0x48CF: "HANJA",
        0x48D0: "HANJA",
        0x48D1: "HANJA",
        0x48D2: "HANJA",
        0x48D3: "HANJA",
        0x48D4: "HANJA",
        0x48D5: "HANJA",
        0x48D6: "HANJA",
        0x48D7: "HANJA",
        0x48D8: "HANJA",
        0x48D9: "HANJA",
        0x48DA: "HANJA",
        0x48DB: "HANJA",
        0x48DC: "HANJA",
        0x48DD: "HANJA",
        0x48DE: "HANJA",
        0x48DF: "HANJA",
        0x48E0: "HANJA",
        0x48E1: "HANJA",
        0x48E2: "HANJA",
        0x48E3: "HANJA",
        0x48E4: "HANJA",
        0x48E5: "HANJA",
        0x48E6: "HANJA",
        0x48E7: "HANJA",
        0x48E8: "HANJA",
        0x48E9: "HANJA",
        0x48EA: "HANJA",
        0x48EB: "HANJA",
        0x48EC: "HANJA",
        0x48ED: "HANJA",
        0x48EE: "HANJA",
        0x48EF: "HANJA",
        0x48F0: "HANJA",
        0x48F1: "HANJA",
        0x48F2: "HANJA",
        0x48F3: "HANJA",
        0x48F4: "HANJA",
        0x48F5: "HANJA",
        0x48F6: "HANJA",
        0x48F7: "HANJA",
        0x48F8: "HANJA",
        0x48F9: "HANJA",
        0x48FA: "HANJA",
        0x48FB: "HANJA",
        0x48FC: "HANJA",
        0x48FD: "HANJA",
        0x48FE: "HANJA",
        0x48FF: "HANJA",
        0x4900: "HANJA",
        0x4901: "HANJA",
        0x4902: "HANJA",
        0x4903: "HANJA",
        0x4904: "HANJA",
        0x4905: "HANJA",
        0x4906: "HANJA",
        0x4907: "HANJA",
        0x4908: "HANJA",
        0x4909: "HANJA",
        0x490A: "HANJA",
        0x490B: "HANJA",
        0x490C: "HANJA",
        0x490D: "HANJA",
        0x490E: "HANJA",
        0x490F: "HANJA",
        0x4910: "HANJA",
        0x4911: "HANJA",
        0x4912: "HANJA",
        0x4913: "HANJA",
        0x4914: "HANJA",
        0x4915: "HANJA",
        0x4916: "HANJA",
        0x4917: "HANJA",
        0x4918: "HANJA",
        0x4919: "HANJA",
        0x491A: "HANJA",
        0x491B: "HANJA",
        0x491C: "HANJA",
        0x491D: "HANJA",
        0x491E: "HANJA",
        0x491F: "HANJA",
        0x4920: "HANJA",
        0x4921: "HANJA",
        0x4922: "HANJA",
        0x4923: "HANJA",
        0x4924: "HANJA",
        0x4925: "HANJA",
        0x4926: "HANJA",
        0x4927: "HANJA",
        0x4928: "HANJA",
        0x4929: "HANJA",
        0x492A: "HANJA",
        0x492B: "HANJA",
        0x492C: "HANJA",
        0x492D: "HANJA",
        0x492E: "HANJA",
        0x492F: "HANJA",
        0x4930: "HANJA",
        0x4931: "HANJA",
        0x4932: "HANJA",
        0x4933: "HANJA",
        0x4934: "HANJA",
        0x4935: "HANJA",
        0x4936: "HANJA",
        0x4937: "HANJA",
        0x4938: "HANJA",
        0x4939: "HANJA",
        0x493A: "HANJA",
        0x493B: "HANJA",
        0x493C: "HANJA",
        0x493D: "HANJA",
        0x493E: "HANJA",
        0x493F: "HANJA",
        0x4940: "HANJA",
        0x4941: "HANJA",
        0x4942: "HANJA",
        0x4943: "HANJA",
        0x4944: "HANJA",
        0x4945: "HANJA",
        0x4946: "HANJA",
        0x4947: "HANJA",
        0x4948: "HANJA",
        0x4949: "HANJA",
        0x494A: "HANJA",
        0x494B: "HANJA",
        0x494C: "HANJA",
        0x494D: "HANJA",
        0x494E: "HANJA",
        0x494F: "HANJA",
        0x4950: "HANJA",
        0x4951: "HANJA",
        0x4952: "HANJA",
        0x4953: "HANJA",
        0x4954: "HANJA",
        0x4955: "HANJA",
        0x4956: "HANJA",
        0x4957: "HANJA",
        0x4958: "HANJA",
        0x4959: "HANJA",
        0x495A: "HANJA",
        0x495B: "HANJA",
        0x495C: "HANJA",
        0x495D: "HANJA",
        0x495E: "HANJA",
        0x495F: "HANJA",
        0x4960: "HANJA",
        0x4961: "HANJA",
        0x4962: "HANJA",
        0x4963: "HANJA",
        0x4964: "HANJA",
        0x4965: "HANJA",
        0x4966: "HANJA",
        0x4967: "HANJA",
        0x4968: "HANJA",
        0x4969: "HANJA",
        0x496A: "HANJA",
        0x496B: "HANJA",
        0x496C: "HANJA",
        0x496D: "HANJA",
        0x496E: "HANJA",
        0x496F: "HANJA",
        0x4970: "HANJA",
        0x4971: "HANJA",
        0x4972: "HANJA",
        0x4973: "HANJA",
        0x4974: "HANJA",
        0x4975: "HANJA",
        0x4976: "HANJA",
        0x4977: "HANJA",
        0x4978: "HANJA",
        0x4979: "HANJA",
        0x497A: "HANJA",
        0x497B: "HANJA",
        0x497C: "HANJA",
        0x497D: "HANJA",
        0x497E: "HANJA",
        0x497F: "HANJA",
        0x4980: "HANJA",
        0x4981: "HANJA",
        0x4982: "HANJA",
        0x4983: "HANJA",
        0x4984: "HANJA",
        0x4985: "HANJA",
        0x4986: "HANJA",
        0x4987: "HANJA",
        0x4988: "HANJA",
        0x4989: "HANJA",
        0x498A: "HANJA",
        0x498B: "HANJA",
        0x498C: "HANJA",
        0x498D: "HANJA",
        0x498E: "HANJA",
        0x498F: "HANJA",
        0x4990: "HANJA",
        0x4991: "HANJA",
        0x4992: "HANJA",
        0x4993: "HANJA",
        0x4994: "HANJA",
        0x4995: "HANJA",
        0x4996: "HANJA",
        0x4997: "HANJA",
        0x4998: "HANJA",
        0x4999: "HANJA",
        0x499A: "HANJA",
        0x499B: "HANJA",
        0x499C: "HANJA",
        0x499D: "HANJA",
        0x499E: "HANJA",
        0x499F: "HANJA",
        0x49A0: "HANJA",
        0x49A1: "HANJA",
        0x49A2: "HANJA",
        0x49A3: "HANJA",
        0x49A4: "HANJA",
        0x49A5: "HANJA",
        0x49A6: "HANJA",
        0x49A7: "HANJA",
        0x49A8: "HANJA",
        0x49A9: "HANJA",
        0x49AA: "HANJA",
        0x49AB: "HANJA",
        0x49AC: "HANJA",
        0x49AD: "HANJA",
        0x49AE: "HANJA",
        0x49AF: "HANJA",
        0x49B0: "HANJA",
        0x49B1: "HANJA",
        0x49B2: "HANJA",
        0x49B3: "HANJA",
        0x49B4: "HANJA",
        0x49B5: "HANJA",
        0x49B6: "HANJA",
        0x49B7: "HANJA",
        0x49B8: "HANJA",
        0x49B9: "HANJA",
        0x49BA: "HANJA",
        0x49BB: "HANJA",
        0x49BC: "HANJA",
        0x49BD: "HANJA",
        0x49BE: "HANJA",
        0x49BF: "HANJA",
        0x49C0: "HANJA",
        0x49C1: "HANJA",
        0x49C2: "HANJA",
        0x49C3: "HANJA",
        0x49C4: "HANJA",
        0x49C5: "HANJA",
        0x49C6: "HANJA",
        0x49C7: "HANJA",
        0x49C8: "HANJA",
        0x49C9: "HANJA",
        0x49CA: "HANJA",
        0x49CB: "HANJA",
        0x49CC: "HANJA",
        0x49CD: "HANJA",
        0x49CE: "HANJA",
        0x49CF: "HANJA",
        0x49D0: "HANJA",
        0x49D1: "HANJA",
        0x49D2: "HANJA",
        0x49D3: "HANJA",
        0x49D4: "HANJA",
        0x49D5: "HANJA",
        0x49D6: "HANJA",
        0x49D7: "HANJA",
        0x49D8: "HANJA",
        0x49D9: "HANJA",
        0x49DA: "HANJA",
        0x49DB: "HANJA",
        0x49DC: "HANJA",
        0x49DD: "HANJA",
        0x49DE: "HANJA",
        0x49DF: "HANJA",
        0x49E0: "HANJA",
        0x49E1: "HANJA",
        0x49E2: "HANJA",
        0x49E3: "HANJA",
        0x49E4: "HANJA",
        0x49E5: "HANJA",
        0x49E6: "HANJA",
        0x49E7: "HANJA",
        0x49E8: "HANJA",
        0x49E9: "HANJA",
        0x49EA: "HANJA",
        0x49EB: "HANJA",
        0x49EC: "HANJA",
        0x49ED: "HANJA",
        0x49EE: "HANJA",
        0x49EF: "HANJA",
        0x49F0: "HANJA",
        0x49F1: "HANJA",
        0x49F2: "HANJA",
        0x49F3: "HANJA",
        0x49F4: "HANJA",
        0x49F5: "HANJA",
        0x49F6: "HANJA",
        0x49F7: "HANJA",
        0x49F8: "HANJA",
        0x49F9: "HANJA",
        0x49FA: "HANJA",
        0x49FB: "HANJA",
        0x49FC: "HANJA",
        0x49FD: "HANJA",
        0x49FE: "HANJA",
        0x49FF: "HANJA",
        0x4A00: "HANJA",
        0x4A01: "HANJA",
        0x4A02: "HANJA",
        0x4A03: "HANJA",
        0x4A04: "HANJA",
        0x4A05: "HANJA",
        0x4A06: "HANJA",
        0x4A07: "HANJA",
        0x4A08: "HANJA",
        0x4A09: "HANJA",
        0x4A0A: "HANJA",
        0x4A0B: "HANJA",
        0x4A0C: "HANJA",
        0x4A0D: "HANJA",
        0x4A0E: "HANJA",
        0x4A0F: "HANJA",
        0x4A10: "HANJA",
        0x4A11: "HANJA",
        0x4A12: "HANJA",
        0x4A13: "HANJA",
        0x4A14: "HANJA",
        0x4A15: "HANJA",
        0x4A16: "HANJA",
        0x4A17: "HANJA",
        0x4A18: "HANJA",
        0x4A19: "HANJA",
        0x4A1A: "HANJA",
        0x4A1B: "HANJA",
        0x4A1C: "HANJA",
        0x4A1D: "HANJA",
        0x4A1E: "HANJA",
        0x4A1F: "HANJA",
        0x4A20: "HANJA",
        0x4A21: "HANJA",
        0x4A22: "HANJA",
        0x4A23: "HANJA",
        0x4A24: "HANJA",
        0x4A25: "HANJA",
        0x4A26: "HANJA",
        0x4A27: "HANJA",
        0x4A28: "HANJA",
        0x4A29: "HANJA",
        0x4A2A: "HANJA",
        0x4A2B: "HANJA",
        0x4A2C: "HANJA",
        0x4A2D: "HANJA",
        0x4A2E: "HANJA",
        0x4A2F: "HANJA",
        0x4A30: "HANJA",
        0x4A31: "HANJA",
        0x4A32: "HANJA",
        0x4A33: "HANJA",
        0x4A34: "HANJA",
        0x4A35: "HANJA",
        0x4A36: "HANJA",
        0x4A37: "HANJA",
        0x4A38: "HANJA",
        0x4A39: "HANJA",
        0x4A3A: "HANJA",
        0x4A3B: "HANJA",
        0x4A3C: "HANJA",
        0x4A3D: "HANJA",
        0x4A3E: "HANJA",
        0x4A3F: "HANJA",
        0x4A40: "HANJA",
        0x4A41: "HANJA",
        0x4A42: "HANJA",
        0x4A43: "HANJA",
        0x4A44: "HANJA",
        0x4A45: "HANJA",
        0x4A46: "HANJA",
        0x4A47: "HANJA",
        0x4A48: "HANJA",
        0x4A49: "HANJA",
        0x4A4A: "HANJA",
        0x4A4B: "HANJA",
        0x4A4C: "HANJA",
        0x4A4D: "HANJA",
        0x4A4E: "HANJA",
        0x4A4F: "HANJA",
        0x4A50: "HANJA",
        0x4A51: "HANJA",
        0x4A52: "HANJA",
        0x4A53: "HANJA",
        0x4A54: "HANJA",
        0x4A55: "HANJA",
        0x4A56: "HANJA",
        0x4A57: "HANJA",
        0x4A58: "HANJA",
        0x4A59: "HANJA",
        0x4A5A: "HANJA",
        0x4A5B: "HANJA",
        0x4A5C: "HANJA",
        0x4A5D: "HANJA",
        0x4A5E: "HANJA",
        0x4A5F: "HANJA",
        0x4A60: "HANJA",
        0x4A61: "HANJA",
        0x4A62: "HANJA",
        0x4A63: "HANJA",
        0x4A64: "HANJA",
        0x4A65: "HANJA",
        0x4A66: "HANJA",
        0x4A67: "HANJA",
        0x4A68: "HANJA",
        0x4A69: "HANJA",
        0x4A6A: "HANJA",
        0x4A6B: "HANJA",
        0x4A6C: "HANJA",
        0x4A6D: "HANJA",
        0x4A6E: "HANJA",
        0x4A6F: "HANJA",
        0x4A70: "HANJA",
        0x4A71: "HANJA",
        0x4A72: "HANJA",
        0x4A73: "HANJA",
        0x4A74: "HANJA",
        0x4A75: "HANJA",
        0x4A76: "HANJA",
        0x4A77: "HANJA",
        0x4A78: "HANJA",
        0x4A79: "HANJA",
        0x4A7A: "HANJA",
        0x4A7B: "HANJA",
        0x4A7C: "HANJA",
        0x4A7D: "HANJA",
        0x4A7E: "HANJA",
        0x4A7F: "HANJA",
        0x4A80: "HANJA",
        0x4A81: "HANJA",
        0x4A82: "HANJA",
        0x4A83: "HANJA",
        0x4A84: "HANJA",
        0x4A85: "HANJA",
        0x4A86: "HANJA",
        0x4A87: "HANJA",
        0x4A88: "HANJA",
        0x4A89: "HANJA",
        0x4A8A: "HANJA",
        0x4A8B: "HANJA",
        0x4A8C: "HANJA",
        0x4A8D: "HANJA",
        0x4A8E: "HANJA",
        0x4A8F: "HANJA",
        0x4A90: "HANJA",
        0x4A91: "HANJA",
        0x4A92: "HANJA",
        0x4A93: "HANJA",
        0x4A94: "HANJA",
        0x4A95: "HANJA",
        0x4A96: "HANJA",
        0x4A97: "HANJA",
        0x4A98: "HANJA",
        0x4A99: "HANJA",
        0x4A9A: "HANJA",
        0x4A9B: "HANJA",
        0x4A9C: "HANJA",
        0x4A9D: "HANJA",
        0x4A9E: "HANJA",
        0x4A9F: "HANJA",
        0x4AA0: "HANJA",
        0x4AA1: "HANJA",
        0x4AA2: "HANJA",
        0x4AA3: "HANJA",
        0x4AA4: "HANJA",
        0x4AA5: "HANJA",
        0x4AA6: "HANJA",
        0x4AA7: "HANJA",
        0x4AA8: "HANJA",
        0x4AA9: "HANJA",
        0x4AAA: "HANJA",
        0x4AAB: "HANJA",
        0x4AAC: "HANJA",
        0x4AAD: "HANJA",
        0x4AAE: "HANJA",
        0x4AAF: "HANJA",
        0x4AB0: "HANJA",
        0x4AB1: "HANJA",
        0x4AB2: "HANJA",
        0x4AB3: "HANJA",
        0x4AB4: "HANJA",
        0x4AB5: "HANJA",
        0x4AB6: "HANJA",
        0x4AB7: "HANJA",
        0x4AB8: "HANJA",
        0x4AB9: "HANJA",
        0x4ABA: "HANJA",
        0x4ABB: "HANJA",
        0x4ABC: "HANJA",
        0x4ABD: "HANJA",
        0x4ABE: "HANJA",
        0x4ABF: "HANJA",
        0x4AC0: "HANJA",
        0x4AC1: "HANJA",
        0x4AC2: "HANJA",
        0x4AC3: "HANJA",
        0x4AC4: "HANJA",
        0x4AC5: "HANJA",
        0x4AC6: "HANJA",
        0x4AC7: "HANJA",
        0x4AC8: "HANJA",
        0x4AC9: "HANJA",
        0x4ACA: "HANJA",
        0x4ACB: "HANJA",
        0x4ACC: "HANJA",
        0x4ACD: "HANJA",
        0x4ACE: "HANJA",
        0x4ACF: "HANJA",
        0x4AD0: "HANJA",
        0x4AD1: "HANJA",
        0x4AD2: "HANJA",
        0x4AD3: "HANJA",
        0x4AD4: "HANJA",
        0x4AD5: "HANJA",
        0x4AD6: "HANJA",
        0x4AD7: "HANJA",
        0x4AD8: "HANJA",
        0x4AD9: "HANJA",
        0x4ADA: "HANJA",
        0x4ADB: "HANJA",
        0x4ADC: "HANJA",
        0x4ADD: "HANJA",
        0x4ADE: "HANJA",
        0x4ADF: "HANJA",
        0x4AE0: "HANJA",
        0x4AE1: "HANJA",
        0x4AE2: "HANJA",
        0x4AE3: "HANJA",
        0x4AE4: "HANJA",
        0x4AE5: "HANJA",
        0x4AE6: "HANJA",
        0x4AE7: "HANJA",
        0x4AE8: "HANJA",
        0x4AE9: "HANJA",
        0x4AEA: "HANJA",
        0x4AEB: "HANJA",
        0x4AEC: "HANJA",
        0x4AED: "HANJA",
        0x4AEE: "HANJA",
        0x4AEF: "HANJA",
        0x4AF0: "HANJA",
        0x4AF1: "HANJA",
        0x4AF2: "HANJA",
        0x4AF3: "HANJA",
        0x4AF4: "HANJA",
        0x4AF5: "HANJA",
        0x4AF6: "HANJA",
        0x4AF7: "HANJA",
        0x4AF8: "HANJA",
        0x4AF9: "HANJA",
        0x4AFA: "HANJA",
        0x4AFB: "HANJA",
        0x4AFC: "HANJA",
        0x4AFD: "HANJA",
        0x4AFE: "HANJA",
        0x4AFF: "HANJA",
        0x4B00: "HANJA",
        0x4B01: "HANJA",
        0x4B02: "HANJA",
        0x4B03: "HANJA",
        0x4B04: "HANJA",
        0x4B05: "HANJA",
        0x4B06: "HANJA",
        0x4B07: "HANJA",
        0x4B08: "HANJA",
        0x4B09: "HANJA",
        0x4B0A: "HANJA",
        0x4B0B: "HANJA",
        0x4B0C: "HANJA",
        0x4B0D: "HANJA",
        0x4B0E: "HANJA",
        0x4B0F: "HANJA",
        0x4B10: "HANJA",
        0x4B11: "HANJA",
        0x4B12: "HANJA",
        0x4B13: "HANJA",
        0x4B14: "HANJA",
        0x4B15: "HANJA",
        0x4B16: "HANJA",
        0x4B17: "HANJA",
        0x4B18: "HANJA",
        0x4B19: "HANJA",
        0x4B1A: "HANJA",
        0x4B1B: "HANJA",
        0x4B1C: "HANJA",
        0x4B1D: "HANJA",
        0x4B1E: "HANJA",
        0x4B1F: "HANJA",
        0x4B20: "HANJA",
        0x4B21: "HANJA",
        0x4B22: "HANJA",
        0x4B23: "HANJA",
        0x4B24: "HANJA",
        0x4B25: "HANJA",
        0x4B26: "HANJA",
        0x4B27: "HANJA",
        0x4B28: "HANJA",
        0x4B29: "HANJA",
        0x4B2A: "HANJA",
        0x4B2B: "HANJA",
        0x4B2C: "HANJA",
        0x4B2D: "HANJA",
        0x4B2E: "HANJA",
        0x4B2F: "HANJA",
        0x4B30: "HANJA",
        0x4B31: "HANJA",
        0x4B32: "HANJA",
        0x4B33: "HANJA",
        0x4B34: "HANJA",
        0x4B35: "HANJA",
        0x4B36: "HANJA",
        0x4B37: "HANJA",
        0x4B38: "HANJA",
        0x4B39: "HANJA",
        0x4B3A: "HANJA",
        0x4B3B: "HANJA",
        0x4B3C: "HANJA",
        0x4B3D: "HANJA",
        0x4B3E: "HANJA",
        0x4B3F: "HANJA",
        0x4B40: "HANJA",
        0x4B41: "HANJA",
        0x4B42: "HANJA",
        0x4B43: "HANJA",
        0x4B44: "HANJA",
        0x4B45: "HANJA",
        0x4B46: "HANJA",
        0x4B47: "HANJA",
        0x4B48: "HANJA",
        0x4B49: "HANJA",
        0x4B4A: "HANJA",
        0x4B4B: "HANJA",
        0x4B4C: "HANJA",
        0x4B4D: "HANJA",
        0x4B4E: "HANJA",
        0x4B4F: "HANJA",
        0x4B50: "HANJA",
        0x4B51: "HANJA",
        0x4B52: "HANJA",
        0x4B53: "HANJA",
        0x4B54: "HANJA",
        0x4B55: "HANJA",
        0x4B56: "HANJA",
        0x4B57: "HANJA",
        0x4B58: "HANJA",
        0x4B59: "HANJA",
        0x4B5A: "HANJA",
        0x4B5B: "HANJA",
        0x4B5C: "HANJA",
        0x4B5D: "HANJA",
        0x4B5E: "HANJA",
        0x4B5F: "HANJA",
        0x4B60: "HANJA",
        0x4B61: "HANJA",
        0x4B62: "HANJA",
        0x4B63: "HANJA",
        0x4B64: "HANJA",
        0x4B65: "HANJA",
        0x4B66: "HANJA",
        0x4B67: "HANJA",
        0x4B68: "HANJA",
        0x4B69: "HANJA",
        0x4B6A: "HANJA",
        0x4B6B: "HANJA",
        0x4B6C: "HANJA",
        0x4B6D: "HANJA",
        0x4B6E: "HANJA",
        0x4B6F: "HANJA",
        0x4B70: "HANJA",
        0x4B71: "HANJA",
        0x4B72: "HANJA",
        0x4B73: "HANJA",
        0x4B74: "HANJA",
        0x4B75: "HANJA",
        0x4B76: "HANJA",
        0x4B77: "HANJA",
        0x4B78: "HANJA",
        0x4B79: "HANJA",
        0x4B7A: "HANJA",
        0x4B7B: "HANJA",
        0x4B7C: "HANJA",
        0x4B7D: "HANJA",
        0x4B7E: "HANJA",
        0x4B7F: "HANJA",
        0x4B80: "HANJA",
        0x4B81: "HANJA",
        0x4B82: "HANJA",
        0x4B83: "HANJA",
        0x4B84: "HANJA",
        0x4B85: "HANJA",
        0x4B86: "HANJA",
        0x4B87: "HANJA",
        0x4B88: "HANJA",
        0x4B89: "HANJA",
        0x4B8A: "HANJA",
        0x4B8B: "HANJA",
        0x4B8C: "HANJA",
        0x4B8D: "HANJA",
        0x4B8E: "HANJA",
        0x4B8F: "HANJA",
        0x4B90: "HANJA",
        0x4B91: "HANJA",
        0x4B92: "HANJA",
        0x4B93: "HANJA",
        0x4B94: "HANJA",
        0x4B95: "HANJA",
        0x4B96: "HANJA",
        0x4B97: "HANJA",
        0x4B98: "HANJA",
        0x4B99: "HANJA",
        0x4B9A: "HANJA",
        0x4B9B: "HANJA",
        0x4B9C: "HANJA",
        0x4B9D: "HANJA",
        0x4B9E: "HANJA",
        0x4B9F: "HANJA",
        0x4BA0: "HANJA",
        0x4BA1: "HANJA",
        0x4BA2: "HANJA",
        0x4BA3: "HANJA",
        0x4BA4: "HANJA",
        0x4BA5: "HANJA",
        0x4BA6: "HANJA",
        0x4BA7: "HANJA",
        0x4BA8: "HANJA",
        0x4BA9: "HANJA",
        0x4BAA: "HANJA",
        0x4BAB: "HANJA",
        0x4BAC: "HANJA",
        0x4BAD: "HANJA",
        0x4BAE: "HANJA",
        0x4BAF: "HANJA",
        0x4BB0: "HANJA",
        0x4BB1: "HANJA",
        0x4BB2: "HANJA",
        0x4BB3: "HANJA",
        0x4BB4: "HANJA",
        0x4BB5: "HANJA",
        0x4BB6: "HANJA",
        0x4BB7: "HANJA",
        0x4BB8: "HANJA",
        0x4BB9: "HANJA",
        0x4BBA: "HANJA",
        0x4BBB: "HANJA",
        0x4BBC: "HANJA",
        0x4BBD: "HANJA",
        0x4BBE: "HANJA",
        0x4BBF: "HANJA",
        0x4BC0: "HANJA",
        0x4BC1: "HANJA",
        0x4BC2: "HANJA",
        0x4BC3: "HANJA",
        0x4BC4: "HANJA",
        0x4BC5: "HANJA",
        0x4BC6: "HANJA",
        0x4BC7: "HANJA",
        0x4BC8: "HANJA",
        0x4BC9: "HANJA",
        0x4BCA: "HANJA",
        0x4BCB: "HANJA",
        0x4BCC: "HANJA",
        0x4BCD: "HANJA",
        0x4BCE: "HANJA",
        0x4BCF: "HANJA",
        0x4BD0: "HANJA",
        0x4BD1: "HANJA",
        0x4BD2: "HANJA",
        0x4BD3: "HANJA",
        0x4BD4: "HANJA",
        0x4BD5: "HANJA",
        0x4BD6: "HANJA",
        0x4BD7: "HANJA",
        0x4BD8: "HANJA",
        0x4BD9: "HANJA",
        0x4BDA: "HANJA",
        0x4BDB: "HANJA",
        0x4BDC: "HANJA",
        0x4BDD: "HANJA",
        0x4BDE: "HANJA",
        0x4BDF: "HANJA",
        0x4BE0: "HANJA",
        0x4BE1: "HANJA",
        0x4BE2: "HANJA",
        0x4BE3: "HANJA",
        0x4BE4: "HANJA",
        0x4BE5: "HANJA",
        0x4BE6: "HANJA",
        0x4BE7: "HANJA",
        0x4BE8: "HANJA",
        0x4BE9: "HANJA",
        0x4BEA: "HANJA",
        0x4BEB: "HANJA",
        0x4BEC: "HANJA",
        0x4BED: "HANJA",
        0x4BEE: "HANJA",
        0x4BEF: "HANJA",
        0x4BF0: "HANJA",
        0x4BF1: "HANJA",
        0x4BF2: "HANJA",
        0x4BF3: "HANJA",
        0x4BF4: "HANJA",
        0x4BF5: "HANJA",
        0x4BF6: "HANJA",
        0x4BF7: "HANJA",
        0x4BF8: "HANJA",
        0x4BF9: "HANJA",
        0x4BFA: "HANJA",
        0x4BFB: "HANJA",
        0x4BFC: "HANJA",
        0x4BFD: "HANJA",
        0x4BFE: "HANJA",
        0x4BFF: "HANJA",
        0x4C00: "HANJA",
        0x4C01: "HANJA",
        0x4C02: "HANJA",
        0x4C03: "HANJA",
        0x4C04: "HANJA",
        0x4C05: "HANJA",
        0x4C06: "HANJA",
        0x4C07: "HANJA",
        0x4C08: "HANJA",
        0x4C09: "HANJA",
        0x4C0A: "HANJA",
        0x4C0B: "HANJA",
        0x4C0C: "HANJA",
        0x4C0D: "HANJA",
        0x4C0E: "HANJA",
        0x4C0F: "HANJA",
        0x4C10: "HANJA",
        0x4C11: "HANJA",
        0x4C12: "HANJA",
        0x4C13: "HANJA",
        0x4C14: "HANJA",
        0x4C15: "HANJA",
        0x4C16: "HANJA",
        0x4C17: "HANJA",
        0x4C18: "HANJA",
        0x4C19: "HANJA",
        0x4C1A: "HANJA",
        0x4C1B: "HANJA",
        0x4C1C: "HANJA",
        0x4C1D: "HANJA",
        0x4C1E: "HANJA",
        0x4C1F: "HANJA",
        0x4C20: "HANJA",
        0x4C21: "HANJA",
        0x4C22: "HANJA",
        0x4C23: "HANJA",
        0x4C24: "HANJA",
        0x4C25: "HANJA",
        0x4C26: "HANJA",
        0x4C27: "HANJA",
        0x4C28: "HANJA",
        0x4C29: "HANJA",
        0x4C2A: "HANJA",
        0x4C2B: "HANJA",
        0x4C2C: "HANJA",
        0x4C2D: "HANJA",
        0x4C2E: "HANJA",
        0x4C2F: "HANJA",
        0x4C30: "HANJA",
        0x4C31: "HANJA",
        0x4C32: "HANJA",
        0x4C33: "HANJA",
        0x4C34: "HANJA",
        0x4C35: "HANJA",
        0x4C36: "HANJA",
        0x4C37: "HANJA",
        0x4C38: "HANJA",
        0x4C39: "HANJA",
        0x4C3A: "HANJA",
        0x4C3B: "HANJA",
        0x4C3C: "HANJA",
        0x4C3D: "HANJA",
        0x4C3E: "HANJA",
        0x4C3F: "HANJA",
        0x4C40: "HANJA",
        0x4C41: "HANJA",
        0x4C42: "HANJA",
        0x4C43: "HANJA",
        0x4C44: "HANJA",
        0x4C45: "HANJA",
        0x4C46: "HANJA",
        0x4C47: "HANJA",
        0x4C48: "HANJA",
        0x4C49: "HANJA",
        0x4C4A: "HANJA",
        0x4C4B: "HANJA",
        0x4C4C: "HANJA",
        0x4C4D: "HANJA",
        0x4C4E: "HANJA",
        0x4C4F: "HANJA",
        0x4C50: "HANJA",
        0x4C51: "HANJA",
        0x4C52: "HANJA",
        0x4C53: "HANJA",
        0x4C54: "HANJA",
        0x4C55: "HANJA",
        0x4C56: "HANJA",
        0x4C57: "HANJA",
        0x4C58: "HANJA",
        0x4C59: "HANJA",
        0x4C5A: "HANJA",
        0x4C5B: "HANJA",
        0x4C5C: "HANJA",
        0x4C5D: "HANJA",
        0x4C5E: "HANJA",
        0x4C5F: "HANJA",
        0x4C60: "HANJA",
        0x4C61: "HANJA",
        0x4C62: "HANJA",
        0x4C63: "HANJA",
        0x4C64: "HANJA",
        0x4C65: "HANJA",
        0x4C66: "HANJA",
        0x4C67: "HANJA",
        0x4C68: "HANJA",
        0x4C69: "HANJA",
        0x4C6A: "HANJA",
        0x4C6B: "HANJA",
        0x4C6C: "HANJA",
        0x4C6D: "HANJA",
        0x4C6E: "HANJA",
        0x4C6F: "HANJA",
        0x4C70: "HANJA",
        0x4C71: "HANJA",
        0x4C72: "HANJA",
        0x4C73: "HANJA",
        0x4C74: "HANJA",
        0x4C75: "HANJA",
        0x4C76: "HANJA",
        0x4C77: "HANJA",
        0x4C78: "HANJA",
        0x4C79: "HANJA",
        0x4C7A: "HANJA",
        0x4C7B: "HANJA",
        0x4C7C: "HANJA",
        0x4C7D: "HANJA",
        0x4C7E: "HANJA",
        0x4C7F: "HANJA",
        0x4C80: "HANJA",
        0x4C81: "HANJA",
        0x4C82: "HANJA",
        0x4C83: "HANJA",
        0x4C84: "HANJA",
        0x4C85: "HANJA",
        0x4C86: "HANJA",
        0x4C87: "HANJA",
        0x4C88: "HANJA",
        0x4C89: "HANJA",
        0x4C8A: "HANJA",
        0x4C8B: "HANJA",
        0x4C8C: "HANJA",
        0x4C8D: "HANJA",
        0x4C8E: "HANJA",
        0x4C8F: "HANJA",
        0x4C90: "HANJA",
        0x4C91: "HANJA",
        0x4C92: "HANJA",
        0x4C93: "HANJA",
        0x4C94: "HANJA",
        0x4C95: "HANJA",
        0x4C96: "HANJA",
        0x4C97: "HANJA",
        0x4C98: "HANJA",
        0x4C99: "HANJA",
        0x4C9A: "HANJA",
        0x4C9B: "HANJA",
        0x4C9C: "HANJA",
        0x4C9D: "HANJA",
        0x4C9E: "HANJA",
        0x4C9F: "HANJA",
        0x4CA0: "HANJA",
        0x4CA1: "HANJA",
        0x4CA2: "HANJA",
        0x4CA3: "HANJA",
        0x4CA4: "HANJA",
        0x4CA5: "HANJA",
        0x4CA6: "HANJA",
        0x4CA7: "HANJA",
        0x4CA8: "HANJA",
        0x4CA9: "HANJA",
        0x4CAA: "HANJA",
        0x4CAB: "HANJA",
        0x4CAC: "HANJA",
        0x4CAD: "HANJA",
        0x4CAE: "HANJA",
        0x4CAF: "HANJA",
        0x4CB0: "HANJA",
        0x4CB1: "HANJA",
        0x4CB2: "HANJA",
        0x4CB3: "HANJA",
        0x4CB4: "HANJA",
        0x4CB5: "HANJA",
        0x4CB6: "HANJA",
        0x4CB7: "HANJA",
        0x4CB8: "HANJA",
        0x4CB9: "HANJA",
        0x4CBA: "HANJA",
        0x4CBB: "HANJA",
        0x4CBC: "HANJA",
        0x4CBD: "HANJA",
        0x4CBE: "HANJA",
        0x4CBF: "HANJA",
        0x4CC0: "HANJA",
        0x4CC1: "HANJA",
        0x4CC2: "HANJA",
        0x4CC3: "HANJA",
        0x4CC4: "HANJA",
        0x4CC5: "HANJA",
        0x4CC6: "HANJA",
        0x4CC7: "HANJA",
        0x4CC8: "HANJA",
        0x4CC9: "HANJA",
        0x4CCA: "HANJA",
        0x4CCB: "HANJA",
        0x4CCC: "HANJA",
        0x4CCD: "HANJA",
        0x4CCE: "HANJA",
        0x4CCF: "HANJA",
        0x4CD0: "HANJA",
        0x4CD1: "HANJA",
        0x4CD2: "HANJA",
        0x4CD3: "HANJA",
        0x4CD4: "HANJA",
        0x4CD5: "HANJA",
        0x4CD6: "HANJA",
        0x4CD7: "HANJA",
        0x4CD8: "HANJA",
        0x4CD9: "HANJA",
        0x4CDA: "HANJA",
        0x4CDB: "HANJA",
        0x4CDC: "HANJA",
        0x4CDD: "HANJA",
        0x4CDE: "HANJA",
        0x4CDF: "HANJA",
        0x4CE0: "HANJA",
        0x4CE1: "HANJA",
        0x4CE2: "HANJA",
        0x4CE3: "HANJA",
        0x4CE4: "HANJA",
        0x4CE5: "HANJA",
        0x4CE6: "HANJA",
        0x4CE7: "HANJA",
        0x4CE8: "HANJA",
        0x4CE9: "HANJA",
        0x4CEA: "HANJA",
        0x4CEB: "HANJA",
        0x4CEC: "HANJA",
        0x4CED: "HANJA",
        0x4CEE: "HANJA",
        0x4CEF: "HANJA",
        0x4CF0: "HANJA",
        0x4CF1: "HANJA",
        0x4CF2: "HANJA",
        0x4CF3: "HANJA",
        0x4CF4: "HANJA",
        0x4CF5: "HANJA",
        0x4CF6: "HANJA",
        0x4CF7: "HANJA",
        0x4CF8: "HANJA",
        0x4CF9: "HANJA",
        0x4CFA: "HANJA",
        0x4CFB: "HANJA",
        0x4CFC: "HANJA",
        0x4CFD: "HANJA",
        0x4CFE: "HANJA",
        0x4CFF: "HANJA",
        0x4D00: "HANJA",
        0x4D01: "HANJA",
        0x4D02: "HANJA",
        0x4D03: "HANJA",
        0x4D04: "HANJA",
        0x4D05: "HANJA",
        0x4D06: "HANJA",
        0x4D07: "HANJA",
        0x4D08: "HANJA",
        0x4D09: "HANJA",
        0x4D0A: "HANJA",
        0x4D0B: "HANJA",
        0x4D0C: "HANJA",
        0x4D0D: "HANJA",
        0x4D0E: "HANJA",
        0x4D0F: "HANJA",
        0x4D10: "HANJA",
        0x4D11: "HANJA",
        0x4D12: "HANJA",
        0x4D13: "HANJA",
        0x4D14: "HANJA",
        0x4D15: "HANJA",
        0x4D16: "HANJA",
        0x4D17: "HANJA",
        0x4D18: "HANJA",
        0x4D19: "HANJA",
        0x4D1A: "HANJA",
        0x4D1B: "HANJA",
        0x4D1C: "HANJA",
        0x4D1D: "HANJA",
        0x4D1E: "HANJA",
        0x4D1F: "HANJA",
        0x4D20: "HANJA",
        0x4D21: "HANJA",
        0x4D22: "HANJA",
        0x4D23: "HANJA",
        0x4D24: "HANJA",
        0x4D25: "HANJA",
        0x4D26: "HANJA",
        0x4D27: "HANJA",
        0x4D28: "HANJA",
        0x4D29: "HANJA",
        0x4D2A: "HANJA",
        0x4D2B: "HANJA",
        0x4D2C: "HANJA",
        0x4D2D: "HANJA",
        0x4D2E: "HANJA",
        0x4D2F: "HANJA",
        0x4D30: "HANJA",
        0x4D31: "HANJA",
        0x4D32: "HANJA",
        0x4D33: "HANJA",
        0x4D34: "HANJA",
        0x4D35: "HANJA",
        0x4D36: "HANJA",
        0x4D37: "HANJA",
        0x4D38: "HANJA",
        0x4D39: "HANJA",
        0x4D3A: "HANJA",
        0x4D3B: "HANJA",
        0x4D3C: "HANJA",
        0x4D3D: "HANJA",
        0x4D3E: "HANJA",
        0x4D3F: "HANJA",
        0x4D40: "HANJA",
        0x4D41: "HANJA",
        0x4D42: "HANJA",
        0x4D43: "HANJA",
        0x4D44: "HANJA",
        0x4D45: "HANJA",
        0x4D46: "HANJA",
        0x4D47: "HANJA",
        0x4D48: "HANJA",
        0x4D49: "HANJA",
        0x4D4A: "HANJA",
        0x4D4B: "HANJA",
        0x4D4C: "HANJA",
        0x4D4D: "HANJA",
        0x4D4E: "HANJA",
        0x4D4F: "HANJA",
        0x4D50: "HANJA",
        0x4D51: "HANJA",
        0x4D52: "HANJA",
        0x4D53: "HANJA",
        0x4D54: "HANJA",
        0x4D55: "HANJA",
        0x4D56: "HANJA",
        0x4D57: "HANJA",
        0x4D58: "HANJA",
        0x4D59: "HANJA",
        0x4D5A: "HANJA",
        0x4D5B: "HANJA",
        0x4D5C: "HANJA",
        0x4D5D: "HANJA",
        0x4D5E: "HANJA",
        0x4D5F: "HANJA",
        0x4D60: "HANJA",
        0x4D61: "HANJA",
        0x4D62: "HANJA",
        0x4D63: "HANJA",
        0x4D64: "HANJA",
        0x4D65: "HANJA",
        0x4D66: "HANJA",
        0x4D67: "HANJA",
        0x4D68: "HANJA",
        0x4D69: "HANJA",
        0x4D6A: "HANJA",
        0x4D6B: "HANJA",
        0x4D6C: "HANJA",
        0x4D6D: "HANJA",
        0x4D6E: "HANJA",
        0x4D6F: "HANJA",
        0x4D70: "HANJA",
        0x4D71: "HANJA",
        0x4D72: "HANJA",
        0x4D73: "HANJA",
        0x4D74: "HANJA",
        0x4D75: "HANJA",
        0x4D76: "HANJA",
        0x4D77: "HANJA",
        0x4D78: "HANJA",
        0x4D79: "HANJA",
        0x4D7A: "HANJA",
        0x4D7B: "HANJA",
        0x4D7C: "HANJA",
        0x4D7D: "HANJA",
        0x4D7E: "HANJA",
        0x4D7F: "HANJA",
        0x4D80: "HANJA",
        0x4D81: "HANJA",
        0x4D82: "HANJA",
        0x4D83: "HANJA",
        0x4D84: "HANJA",
        0x4D85: "HANJA",
        0x4D86: "HANJA",
        0x4D87: "HANJA",
        0x4D88: "HANJA",
        0x4D89: "HANJA",
        0x4D8A: "HANJA",
        0x4D8B: "HANJA",
        0x4D8C: "HANJA",
        0x4D8D: "HANJA",
        0x4D8E: "HANJA",
        0x4D8F: "HANJA",
        0x4D90: "HANJA",
        0x4D91: "HANJA",
        0x4D92: "HANJA",
        0x4D93: "HANJA",
        0x4D94: "HANJA",
        0x4D95: "HANJA",
        0x4D96: "HANJA",
        0x4D97: "HANJA",
        0x4D98: "HANJA",
        0x4D99: "HANJA",
        0x4D9A: "HANJA",
        0x4D9B: "HANJA",
        0x4D9C: "HANJA",
        0x4D9D: "HANJA",
        0x4D9E: "HANJA",
        0x4D9F: "HANJA",
        0x4DA0: "HANJA",
        0x4DA1: "HANJA",
        0x4DA2: "HANJA",
        0x4DA3: "HANJA",
        0x4DA4: "HANJA",
        0x4DA5: "HANJA",
        0x4DA6: "HANJA",
        0x4DA7: "HANJA",
        0x4DA8: "HANJA",
        0x4DA9: "HANJA",
        0x4DAA: "HANJA",
        0x4DAB: "HANJA",
        0x4DAC: "HANJA",
        0x4DAD: "HANJA",
        0x4DAE: "HANJA",
        0x4DAF: "HANJA",
        0x4DB0: "HANJA",
        0x4DB1: "HANJA",
        0x4DB2: "HANJA",
        0x4DB3: "HANJA",
        0x4DB4: "HANJA",
        0x4DB5: "HANJA",
        0x4E00: "HANJA",
        0x4E01: "HANJA",
        0x4E02: "HANJA",
        0x4E03: "HANJA",
        0x4E04: "HANJA",
        0x4E05: "HANJA",
        0x4E06: "HANJA",
        0x4E07: "HANJA",
        0x4E08: "HANJA",
        0x4E09: "HANJA",
        0x4E0A: "HANJA",
        0x4E0B: "HANJA",
        0x4E0C: "HANJA",
        0x4E0D: "HANJA",
        0x4E0E: "HANJA",
        0x4E0F: "HANJA",
        0x4E10: "HANJA",
        0x4E11: "HANJA",
        0x4E12: "HANJA",
        0x4E13: "HANJA",
        0x4E14: "HANJA",
        0x4E15: "HANJA",
        0x4E16: "HANJA",
        0x4E17: "HANJA",
        0x4E18: "HANJA",
        0x4E19: "HANJA",
        0x4E1A: "HANJA",
        0x4E1B: "HANJA",
        0x4E1C: "HANJA",
        0x4E1D: "HANJA",
        0x4E1E: "HANJA",
        0x4E1F: "HANJA",
        0x4E20: "HANJA",
        0x4E21: "HANJA",
        0x4E22: "HANJA",
        0x4E23: "HANJA",
        0x4E24: "HANJA",
        0x4E25: "HANJA",
        0x4E26: "HANJA",
        0x4E27: "HANJA",
        0x4E28: "HANJA",
        0x4E29: "HANJA",
        0x4E2A: "HANJA",
        0x4E2B: "HANJA",
        0x4E2C: "HANJA",
        0x4E2D: "HANJA",
        0x4E2E: "HANJA",
        0x4E2F: "HANJA",
        0x4E30: "HANJA",
        0x4E31: "HANJA",
        0x4E32: "HANJA",
        0x4E33: "HANJA",
        0x4E34: "HANJA",
        0x4E35: "HANJA",
        0x4E36: "HANJA",
        0x4E37: "HANJA",
        0x4E38: "HANJA",
        0x4E39: "HANJA",
        0x4E3A: "HANJA",
        0x4E3B: "HANJA",
        0x4E3C: "HANJA",
        0x4E3D: "HANJA",
        0x4E3E: "HANJA",
        0x4E3F: "HANJA",
        0x4E40: "HANJA",
        0x4E41: "HANJA",
        0x4E42: "HANJA",
        0x4E43: "HANJA",
        0x4E44: "HANJA",
        0x4E45: "HANJA",
        0x4E46: "HANJA",
        0x4E47: "HANJA",
        0x4E48: "HANJA",
        0x4E49: "HANJA",
        0x4E4A: "HANJA",
        0x4E4B: "HANJA",
        0x4E4C: "HANJA",
        0x4E4D: "HANJA",
        0x4E4E: "HANJA",
        0x4E4F: "HANJA",
        0x4E50: "HANJA",
        0x4E51: "HANJA",
        0x4E52: "HANJA",
        0x4E53: "HANJA",
        0x4E54: "HANJA",
        0x4E55: "HANJA",
        0x4E56: "HANJA",
        0x4E57: "HANJA",
        0x4E58: "HANJA",
        0x4E59: "HANJA",
        0x4E5A: "HANJA",
        0x4E5B: "HANJA",
        0x4E5C: "HANJA",
        0x4E5E: "HANJA",
        0x4E5F: "HANJA",
        0x4E60: "HANJA",
        0x4E61: "HANJA",
        0x4E62: "HANJA",
        0x4E63: "HANJA",
        0x4E64: "HANJA",
        0x4E65: "HANJA",
        0x4E66: "HANJA",
        0x4E67: "HANJA",
        0x4E68: "HANJA",
        0x4E69: "HANJA",
        0x4E6A: "HANJA",
        0x4E6B: "HANJA",
        0x4E6C: "HANJA",
        0x4E6D: "HANJA",
        0x4E6E: "HANJA",
        0x4E6F: "HANJA",
        0x4E70: "HANJA",
        0x4E71: "HANJA",
        0x4E72: "HANJA",
        0x4E73: "HANJA",
        0x4E74: "HANJA",
        0x4E75: "HANJA",
        0x4E76: "HANJA",
        0x4E77: "HANJA",
        0x4E78: "HANJA",
        0x4E79: "HANJA",
        0x4E7A: "HANJA",
        0x4E7B: "HANJA",
        0x4E7C: "HANJA",
        0x4E7D: "HANJA",
        0x4E7E: "HANJA",
        0x4E7F: "HANJA",
        0x4E80: "HANJA",
        0x4E81: "HANJA",
        0x4E82: "HANJA",
        0x4E83: "HANJA",
        0x4E84: "HANJA",
        0x4E85: "HANJA",
        0x4E86: "HANJA",
        0x4E87: "HANJA",
        0x4E88: "HANJA",
        0x4E89: "HANJA",
        0x4E8A: "HANJA",
        0x4E8B: "HANJA",
        0x4E8D: "HANJA",
        0x4E8E: "HANJA",
        0x4E8F: "HANJA",
        0x4E90: "HANJA",
        0x4E91: "HANJA",
        0x4E92: "HANJA",
        0x4E93: "HANJA",
        0x4E95: "HANJA",
        0x4E96: "HANJA",
        0x4E97: "HANJA",
        0x4E98: "HANJA",
        0x4E99: "HANJA",
        0x4E9A: "HANJA",
        0x4E9B: "HANJA",
        0x4E9C: "HANJA",
        0x4E9D: "HANJA",
        0x4E9E: "HANJA",
        0x4E9F: "HANJA",
        0x4EA0: "HANJA",
        0x4EA1: "HANJA",
        0x4EA2: "HANJA",
        0x4EA3: "HANJA",
        0x4EA4: "HANJA",
        0x4EA5: "HANJA",
        0x4EA6: "HANJA",
        0x4EA7: "HANJA",
        0x4EA8: "HANJA",
        0x4EA9: "HANJA",
        0x4EAA: "HANJA",
        0x4EAB: "HANJA",
        0x4EAC: "HANJA",
        0x4EAD: "HANJA",
        0x4EAE: "HANJA",
        0x4EAF: "HANJA",
        0x4EB0: "HANJA",
        0x4EB1: "HANJA",
        0x4EB2: "HANJA",
        0x4EB3: "HANJA",
        0x4EB4: "HANJA",
        0x4EB5: "HANJA",
        0x4EB6: "HANJA",
        0x4EB7: "HANJA",
        0x4EB8: "HANJA",
        0x4EB9: "HANJA",
        0x4EBA: "HANJA",
        0x4EBB: "HANJA",
        0x4EBC: "HANJA",
        0x4EBD: "HANJA",
        0x4EBE: "HANJA",
        0x4EBF: "HANJA",
        0x4EC0: "HANJA",
        0x4EC1: "HANJA",
        0x4EC2: "HANJA",
        0x4EC3: "HANJA",
        0x4EC4: "HANJA",
        0x4EC5: "HANJA",
        0x4EC6: "HANJA",
        0x4EC7: "HANJA",
        0x4EC8: "HANJA",
        0x4EC9: "HANJA",
        0x4ECA: "HANJA",
        0x4ECB: "HANJA",
        0x4ECC: "HANJA",
        0x4ECD: "HANJA",
        0x4ECE: "HANJA",
        0x4ECF: "HANJA",
        0x4ED0: "HANJA",
        0x4ED1: "HANJA",
        0x4ED2: "HANJA",
        0x4ED3: "HANJA",
        0x4ED4: "HANJA",
        0x4ED5: "HANJA",
        0x4ED6: "HANJA",
        0x4ED7: "HANJA",
        0x4ED8: "HANJA",
        0x4ED9: "HANJA",
        0x4EDA: "HANJA",
        0x4EDB: "HANJA",
        0x4EDC: "HANJA",
        0x4EDD: "HANJA",
        0x4EDE: "HANJA",
        0x4EDF: "HANJA",
        0x4EE0: "HANJA",
        0x4EE1: "HANJA",
        0x4EE2: "HANJA",
        0x4EE3: "HANJA",
        0x4EE4: "HANJA",
        0x4EE5: "HANJA",
        0x4EE6: "HANJA",
        0x4EE7: "HANJA",
        0x4EE8: "HANJA",
        0x4EE9: "HANJA",
        0x4EEA: "HANJA",
        0x4EEB: "HANJA",
        0x4EEC: "HANJA",
        0x4EED: "HANJA",
        0x4EEE: "HANJA",
        0x4EEF: "HANJA",
        0x4EF0: "HANJA",
        0x4EF1: "HANJA",
        0x4EF2: "HANJA",
        0x4EF3: "HANJA",
        0x4EF4: "HANJA",
        0x4EF5: "HANJA",
        0x4EF6: "HANJA",
        0x4EF7: "HANJA",
        0x4EF8: "HANJA",
        0x4EF9: "HANJA",
        0x4EFA: "HANJA",
        0x4EFB: "HANJA",
        0x4EFC: "HANJA",
        0x4EFD: "HANJA",
        0x4EFE: "HANJA",
        0x4EFF: "HANJA",
        0x4F00: "HANJA",
        0x4F01: "HANJA",
        0x4F02: "HANJA",
        0x4F03: "HANJA",
        0x4F04: "HANJA",
        0x4F05: "HANJA",
        0x4F06: "HANJA",
        0x4F07: "HANJA",
        0x4F08: "HANJA",
        0x4F09: "HANJA",
        0x4F0A: "HANJA",
        0x4F0B: "HANJA",
        0x4F0C: "HANJA",
        0x4F0D: "HANJA",
        0x4F0E: "HANJA",
        0x4F0F: "HANJA",
        0x4F10: "HANJA",
        0x4F11: "HANJA",
        0x4F12: "HANJA",
        0x4F13: "HANJA",
        0x4F14: "HANJA",
        0x4F15: "HANJA",
        0x4F16: "HANJA",
        0x4F17: "HANJA",
        0x4F18: "HANJA",
        0x4F19: "HANJA",
        0x4F1A: "HANJA",
        0x4F1B: "HANJA",
        0x4F1C: "HANJA",
        0x4F1D: "HANJA",
        0x4F1E: "HANJA",
        0x4F1F: "HANJA",
        0x4F20: "HANJA",
        0x4F21: "HANJA",
        0x4F22: "HANJA",
        0x4F23: "HANJA",
        0x4F24: "HANJA",
        0x4F25: "HANJA",
        0x4F26: "HANJA",
        0x4F27: "HANJA",
        0x4F28: "HANJA",
        0x4F29: "HANJA",
        0x4F2A: "HANJA",
        0x4F2B: "HANJA",
        0x4F2C: "HANJA",
        0x4F2D: "HANJA",
        0x4F2E: "HANJA",
        0x4F2F: "HANJA",
        0x4F30: "HANJA",
        0x4F31: "HANJA",
        0x4F32: "HANJA",
        0x4F33: "HANJA",
        0x4F34: "HANJA",
        0x4F35: "HANJA",
        0x4F36: "HANJA",
        0x4F37: "HANJA",
        0x4F38: "HANJA",
        0x4F39: "HANJA",
        0x4F3A: "HANJA",
        0x4F3B: "HANJA",
        0x4F3C: "HANJA",
        0x4F3D: "HANJA",
        0x4F3E: "HANJA",
        0x4F3F: "HANJA",
        0x4F40: "HANJA",
        0x4F41: "HANJA",
        0x4F42: "HANJA",
        0x4F43: "HANJA",
        0x4F44: "HANJA",
        0x4F45: "HANJA",
        0x4F46: "HANJA",
        0x4F47: "HANJA",
        0x4F48: "HANJA",
        0x4F49: "HANJA",
        0x4F4A: "HANJA",
        0x4F4B: "HANJA",
        0x4F4C: "HANJA",
        0x4F4D: "HANJA",
        0x4F4E: "HANJA",
        0x4F4F: "HANJA",
        0x4F50: "HANJA",
        0x4F51: "HANJA",
        0x4F52: "HANJA",
        0x4F53: "HANJA",
        0x4F54: "HANJA",
        0x4F55: "HANJA",
        0x4F56: "HANJA",
        0x4F57: "HANJA",
        0x4F58: "HANJA",
        0x4F59: "HANJA",
        0x4F5A: "HANJA",
        0x4F5B: "HANJA",
        0x4F5C: "HANJA",
        0x4F5D: "HANJA",
        0x4F5E: "HANJA",
        0x4F5F: "HANJA",
        0x4F60: "HANJA",
        0x4F61: "HANJA",
        0x4F62: "HANJA",
        0x4F63: "HANJA",
        0x4F64: "HANJA",
        0x4F65: "HANJA",
        0x4F66: "HANJA",
        0x4F67: "HANJA",
        0x4F68: "HANJA",
        0x4F69: "HANJA",
        0x4F6A: "HANJA",
        0x4F6B: "HANJA",
        0x4F6C: "HANJA",
        0x4F6D: "HANJA",
        0x4F6E: "HANJA",
        0x4F6F: "HANJA",
        0x4F70: "HANJA",
        0x4F71: "HANJA",
        0x4F72: "HANJA",
        0x4F73: "HANJA",
        0x4F74: "HANJA",
        0x4F75: "HANJA",
        0x4F76: "HANJA",
        0x4F77: "HANJA",
        0x4F78: "HANJA",
        0x4F79: "HANJA",
        0x4F7A: "HANJA",
        0x4F7B: "HANJA",
        0x4F7C: "HANJA",
        0x4F7D: "HANJA",
        0x4F7E: "HANJA",
        0x4F7F: "HANJA",
        0x4F80: "HANJA",
        0x4F81: "HANJA",
        0x4F82: "HANJA",
        0x4F83: "HANJA",
        0x4F84: "HANJA",
        0x4F85: "HANJA",
        0x4F86: "HANJA",
        0x4F87: "HANJA",
        0x4F88: "HANJA",
        0x4F89: "HANJA",
        0x4F8A: "HANJA",
        0x4F8B: "HANJA",
        0x4F8C: "HANJA",
        0x4F8D: "HANJA",
        0x4F8E: "HANJA",
        0x4F8F: "HANJA",
        0x4F90: "HANJA",
        0x4F91: "HANJA",
        0x4F92: "HANJA",
        0x4F93: "HANJA",
        0x4F94: "HANJA",
        0x4F95: "HANJA",
        0x4F96: "HANJA",
        0x4F97: "HANJA",
        0x4F98: "HANJA",
        0x4F99: "HANJA",
        0x4F9A: "HANJA",
        0x4F9B: "HANJA",
        0x4F9C: "HANJA",
        0x4F9D: "HANJA",
        0x4F9E: "HANJA",
        0x4F9F: "HANJA",
        0x4FA0: "HANJA",
        0x4FA1: "HANJA",
        0x4FA2: "HANJA",
        0x4FA3: "HANJA",
        0x4FA4: "HANJA",
        0x4FA5: "HANJA",
        0x4FA6: "HANJA",
        0x4FA7: "HANJA",
        0x4FA8: "HANJA",
        0x4FA9: "HANJA",
        0x4FAA: "HANJA",
        0x4FAB: "HANJA",
        0x4FAC: "HANJA",
        0x4FAD: "HANJA",
        0x4FAE: "HANJA",
        0x4FAF: "HANJA",
        0x4FB0: "HANJA",
        0x4FB1: "HANJA",
        0x4FB2: "HANJA",
        0x4FB3: "HANJA",
        0x4FB4: "HANJA",
        0x4FB5: "HANJA",
        0x4FB6: "HANJA",
        0x4FB7: "HANJA",
        0x4FB8: "HANJA",
        0x4FB9: "HANJA",
        0x4FBA: "HANJA",
        0x4FBB: "HANJA",
        0x4FBC: "HANJA",
        0x4FBD: "HANJA",
        0x4FBE: "HANJA",
        0x4FBF: "HANJA",
        0x4FC0: "HANJA",
        0x4FC1: "HANJA",
        0x4FC2: "HANJA",
        0x4FC3: "HANJA",
        0x4FC4: "HANJA",
        0x4FC5: "HANJA",
        0x4FC6: "HANJA",
        0x4FC7: "HANJA",
        0x4FC8: "HANJA",
        0x4FC9: "HANJA",
        0x4FCA: "HANJA",
        0x4FCB: "HANJA",
        0x4FCC: "HANJA",
        0x4FCD: "HANJA",
        0x4FCE: "HANJA",
        0x4FCF: "HANJA",
        0x4FD0: "HANJA",
        0x4FD1: "HANJA",
        0x4FD2: "HANJA",
        0x4FD3: "HANJA",
        0x4FD4: "HANJA",
        0x4FD5: "HANJA",
        0x4FD6: "HANJA",
        0x4FD7: "HANJA",
        0x4FD8: "HANJA",
        0x4FD9: "HANJA",
        0x4FDA: "HANJA",
        0x4FDB: "HANJA",
        0x4FDC: "HANJA",
        0x4FDD: "HANJA",
        0x4FDE: "HANJA",
        0x4FDF: "HANJA",
        0x4FE0: "HANJA",
        0x4FE1: "HANJA",
        0x4FE2: "HANJA",
        0x4FE3: "HANJA",
        0x4FE4: "HANJA",
        0x4FE5: "HANJA",
        0x4FE6: "HANJA",
        0x4FE7: "HANJA",
        0x4FE8: "HANJA",
        0x4FE9: "HANJA",
        0x4FEA: "HANJA",
        0x4FEB: "HANJA",
        0x4FEC: "HANJA",
        0x4FED: "HANJA",
        0x4FEE: "HANJA",
        0x4FEF: "HANJA",
        0x4FF0: "HANJA",
        0x4FF1: "HANJA",
        0x4FF2: "HANJA",
        0x4FF3: "HANJA",
        0x4FF4: "HANJA",
        0x4FF5: "HANJA",
        0x4FF6: "HANJA",
        0x4FF7: "HANJA",
        0x4FF8: "HANJA",
        0x4FF9: "HANJA",
        0x4FFA: "HANJA",
        0x4FFB: "HANJA",
        0x4FFC: "HANJA",
        0x4FFD: "HANJA",
        0x4FFE: "HANJA",
        0x4FFF: "HANJA",
        0x5000: "HANJA",
        0x5001: "HANJA",
        0x5002: "HANJA",
        0x5003: "HANJA",
        0x5004: "HANJA",
        0x5005: "HANJA",
        0x5006: "HANJA",
        0x5007: "HANJA",
        0x5008: "HANJA",
        0x5009: "HANJA",
        0x500A: "HANJA",
        0x500B: "HANJA",
        0x500C: "HANJA",
        0x500D: "HANJA",
        0x500E: "HANJA",
        0x500F: "HANJA",
        0x5010: "HANJA",
        0x5011: "HANJA",
        0x5012: "HANJA",
        0x5013: "HANJA",
        0x5014: "HANJA",
        0x5015: "HANJA",
        0x5016: "HANJA",
        0x5017: "HANJA",
        0x5018: "HANJA",
        0x5019: "HANJA",
        0x501A: "HANJA",
        0x501B: "HANJA",
        0x501C: "HANJA",
        0x501D: "HANJA",
        0x501E: "HANJA",
        0x501F: "HANJA",
        0x5020: "HANJA",
        0x5021: "HANJA",
        0x5022: "HANJA",
        0x5023: "HANJA",
        0x5024: "HANJA",
        0x5025: "HANJA",
        0x5026: "HANJA",
        0x5027: "HANJA",
        0x5028: "HANJA",
        0x5029: "HANJA",
        0x502A: "HANJA",
        0x502B: "HANJA",
        0x502C: "HANJA",
        0x502D: "HANJA",
        0x502E: "HANJA",
        0x502F: "HANJA",
        0x5030: "HANJA",
        0x5031: "HANJA",
        0x5032: "HANJA",
        0x5033: "HANJA",
        0x5034: "HANJA",
        0x5035: "HANJA",
        0x5036: "HANJA",
        0x5037: "HANJA",
        0x5038: "HANJA",
        0x5039: "HANJA",
        0x503A: "HANJA",
        0x503B: "HANJA",
        0x503C: "HANJA",
        0x503D: "HANJA",
        0x503E: "HANJA",
        0x503F: "HANJA",
        0x5040: "HANJA",
        0x5041: "HANJA",
        0x5042: "HANJA",
        0x5043: "HANJA",
        0x5044: "HANJA",
        0x5045: "HANJA",
        0x5046: "HANJA",
        0x5047: "HANJA",
        0x5048: "HANJA",
        0x5049: "HANJA",
        0x504A: "HANJA",
        0x504B: "HANJA",
        0x504C: "HANJA",
        0x504D: "HANJA",
        0x504E: "HANJA",
        0x504F: "HANJA",
        0x5050: "HANJA",
        0x5051: "HANJA",
        0x5052: "HANJA",
        0x5053: "HANJA",
        0x5054: "HANJA",
        0x5055: "HANJA",
        0x5056: "HANJA",
        0x5057: "HANJA",
        0x5058: "HANJA",
        0x5059: "HANJA",
        0x505A: "HANJA",
        0x505B: "HANJA",
        0x505C: "HANJA",
        0x505D: "HANJA",
        0x505E: "HANJA",
        0x505F: "HANJA",
        0x5060: "HANJA",
        0x5061: "HANJA",
        0x5062: "HANJA",
        0x5063: "HANJA",
        0x5064: "HANJA",
        0x5065: "HANJA",
        0x5066: "HANJA",
        0x5067: "HANJA",
        0x5068: "HANJA",
        0x5069: "HANJA",
        0x506A: "HANJA",
        0x506B: "HANJA",
        0x506C: "HANJA",
        0x506D: "HANJA",
        0x506E: "HANJA",
        0x506F: "HANJA",
        0x5070: "HANJA",
        0x5071: "HANJA",
        0x5072: "HANJA",
        0x5073: "HANJA",
        0x5074: "HANJA",
        0x5075: "HANJA",
        0x5076: "HANJA",
        0x5077: "HANJA",
        0x5078: "HANJA",
        0x5079: "HANJA",
        0x507A: "HANJA",
        0x507B: "HANJA",
        0x507C: "HANJA",
        0x507D: "HANJA",
        0x507E: "HANJA",
        0x507F: "HANJA",
        0x5080: "HANJA",
        0x5081: "HANJA",
        0x5082: "HANJA",
        0x5083: "HANJA",
        0x5084: "HANJA",
        0x5085: "HANJA",
        0x5086: "HANJA",
        0x5087: "HANJA",
        0x5088: "HANJA",
        0x5089: "HANJA",
        0x508A: "HANJA",
        0x508B: "HANJA",
        0x508C: "HANJA",
        0x508D: "HANJA",
        0x508E: "HANJA",
        0x508F: "HANJA",
        0x5090: "HANJA",
        0x5091: "HANJA",
        0x5092: "HANJA",
        0x5093: "HANJA",
        0x5094: "HANJA",
        0x5095: "HANJA",
        0x5096: "HANJA",
        0x5097: "HANJA",
        0x5098: "HANJA",
        0x5099: "HANJA",
        0x509A: "HANJA",
        0x509B: "HANJA",
        0x509C: "HANJA",
        0x509D: "HANJA",
        0x509E: "HANJA",
        0x509F: "HANJA",
        0x50A0: "HANJA",
        0x50A1: "HANJA",
        0x50A2: "HANJA",
        0x50A3: "HANJA",
        0x50A4: "HANJA",
        0x50A5: "HANJA",
        0x50A6: "HANJA",
        0x50A7: "HANJA",
        0x50A8: "HANJA",
        0x50A9: "HANJA",
        0x50AA: "HANJA",
        0x50AB: "HANJA",
        0x50AC: "HANJA",
        0x50AD: "HANJA",
        0x50AE: "HANJA",
        0x50AF: "HANJA",
        0x50B0: "HANJA",
        0x50B1: "HANJA",
        0x50B2: "HANJA",
        0x50B3: "HANJA",
        0x50B4: "HANJA",
        0x50B5: "HANJA",
        0x50B6: "HANJA",
        0x50B7: "HANJA",
        0x50B8: "HANJA",
        0x50B9: "HANJA",
        0x50BA: "HANJA",
        0x50BB: "HANJA",
        0x50BC: "HANJA",
        0x50BD: "HANJA",
        0x50BE: "HANJA",
        0x50BF: "HANJA",
        0x50C0: "HANJA",
        0x50C1: "HANJA",
        0x50C2: "HANJA",
        0x50C3: "HANJA",
        0x50C4: "HANJA",
        0x50C5: "HANJA",
        0x50C6: "HANJA",
        0x50C7: "HANJA",
        0x50C8: "HANJA",
        0x50C9: "HANJA",
        0x50CA: "HANJA",
        0x50CB: "HANJA",
        0x50CC: "HANJA",
        0x50CD: "HANJA",
        0x50CE: "HANJA",
        0x50CF: "HANJA",
        0x50D0: "HANJA",
        0x50D1: "HANJA",
        0x50D2: "HANJA",
        0x50D3: "HANJA",
        0x50D4: "HANJA",
        0x50D5: "HANJA",
        0x50D6: "HANJA",
        0x50D7: "HANJA",
        0x50D8: "HANJA",
        0x50D9: "HANJA",
        0x50DA: "HANJA",
        0x50DB: "HANJA",
        0x50DC: "HANJA",
        0x50DD: "HANJA",
        0x50DE: "HANJA",
        0x50DF: "HANJA",
        0x50E0: "HANJA",
        0x50E1: "HANJA",
        0x50E2: "HANJA",
        0x50E3: "HANJA",
        0x50E4: "HANJA",
        0x50E5: "HANJA",
        0x50E6: "HANJA",
        0x50E7: "HANJA",
        0x50E8: "HANJA",
        0x50E9: "HANJA",
        0x50EA: "HANJA",
        0x50EB: "HANJA",
        0x50EC: "HANJA",
        0x50ED: "HANJA",
        0x50EE: "HANJA",
        0x50EF: "HANJA",
        0x50F0: "HANJA",
        0x50F1: "HANJA",
        0x50F2: "HANJA",
        0x50F3: "HANJA",
        0x50F4: "HANJA",
        0x50F5: "HANJA",
        0x50F6: "HANJA",
        0x50F7: "HANJA",
        0x50F8: "HANJA",
        0x50F9: "HANJA",
        0x50FA: "HANJA",
        0x50FB: "HANJA",
        0x50FC: "HANJA",
        0x50FD: "HANJA",
        0x50FE: "HANJA",
        0x50FF: "HANJA",
        0x5100: "HANJA",
        0x5101: "HANJA",
        0x5102: "HANJA",
        0x5103: "HANJA",
        0x5104: "HANJA",
        0x5105: "HANJA",
        0x5106: "HANJA",
        0x5107: "HANJA",
        0x5108: "HANJA",
        0x5109: "HANJA",
        0x510A: "HANJA",
        0x510B: "HANJA",
        0x510C: "HANJA",
        0x510D: "HANJA",
        0x510E: "HANJA",
        0x510F: "HANJA",
        0x5110: "HANJA",
        0x5111: "HANJA",
        0x5112: "HANJA",
        0x5113: "HANJA",
        0x5114: "HANJA",
        0x5115: "HANJA",
        0x5116: "HANJA",
        0x5117: "HANJA",
        0x5118: "HANJA",
        0x5119: "HANJA",
        0x511A: "HANJA",
        0x511B: "HANJA",
        0x511C: "HANJA",
        0x511D: "HANJA",
        0x511E: "HANJA",
        0x511F: "HANJA",
        0x5120: "HANJA",
        0x5121: "HANJA",
        0x5122: "HANJA",
        0x5123: "HANJA",
        0x5124: "HANJA",
        0x5125: "HANJA",
        0x5126: "HANJA",
        0x5127: "HANJA",
        0x5128: "HANJA",
        0x5129: "HANJA",
        0x512A: "HANJA",
        0x512B: "HANJA",
        0x512C: "HANJA",
        0x512D: "HANJA",
        0x512E: "HANJA",
        0x512F: "HANJA",
        0x5130: "HANJA",
        0x5131: "HANJA",
        0x5132: "HANJA",
        0x5133: "HANJA",
        0x5134: "HANJA",
        0x5135: "HANJA",
        0x5136: "HANJA",
        0x5137: "HANJA",
        0x5138: "HANJA",
        0x5139: "HANJA",
        0x513A: "HANJA",
        0x513B: "HANJA",
        0x513C: "HANJA",
        0x513D: "HANJA",
        0x513E: "HANJA",
        0x513F: "HANJA",
        0x5140: "HANJA",
        0x5141: "HANJA",
        0x5142: "HANJA",
        0x5143: "HANJA",
        0x5144: "HANJA",
        0x5145: "HANJA",
        0x5146: "HANJA",
        0x5147: "HANJA",
        0x5148: "HANJA",
        0x5149: "HANJA",
        0x514A: "HANJA",
        0x514B: "HANJA",
        0x514C: "HANJA",
        0x514D: "HANJA",
        0x514E: "HANJA",
        0x514F: "HANJA",
        0x5150: "HANJA",
        0x5151: "HANJA",
        0x5152: "HANJA",
        0x5153: "HANJA",
        0x5154: "HANJA",
        0x5155: "HANJA",
        0x5156: "HANJA",
        0x5157: "HANJA",
        0x5158: "HANJA",
        0x5159: "HANJA",
        0x515A: "HANJA",
        0x515B: "HANJA",
        0x515C: "HANJA",
        0x515D: "HANJA",
        0x515E: "HANJA",
        0x515F: "HANJA",
        0x5160: "HANJA",
        0x5161: "HANJA",
        0x5162: "HANJA",
        0x5163: "HANJA",
        0x5164: "HANJA",
        0x5165: "HANJA",
        0x5166: "HANJA",
        0x5167: "HANJA",
        0x5168: "HANJA",
        0x5169: "HANJA",
        0x516A: "HANJA",
        0x516B: "HANJA",
        0x516C: "HANJA",
        0x516D: "HANJA",
        0x516E: "HANJA",
        0x516F: "HANJA",
        0x5170: "HANJA",
        0x5171: "HANJA",
        0x5172: "HANJA",
        0x5173: "HANJA",
        0x5174: "HANJA",
        0x5175: "HANJA",
        0x5176: "HANJA",
        0x5177: "HANJA",
        0x5178: "HANJA",
        0x5179: "HANJA",
        0x517A: "HANJA",
        0x517B: "HANJA",
        0x517C: "HANJA",
        0x517D: "HANJA",
        0x517E: "HANJA",
        0x517F: "HANJA",
        0x5180: "HANJA",
        0x5181: "HANJA",
        0x5182: "HANJA",
        0x5183: "HANJA",
        0x5184: "HANJA",
        0x5185: "HANJA",
        0x5186: "HANJA",
        0x5187: "HANJA",
        0x5188: "HANJA",
        0x5189: "HANJA",
        0x518A: "HANJA",
        0x518B: "HANJA",
        0x518C: "HANJA",
        0x518D: "HANJA",
        0x518E: "HANJA",
        0x518F: "HANJA",
        0x5190: "HANJA",
        0x5191: "HANJA",
        0x5192: "HANJA",
        0x5193: "HANJA",
        0x5194: "HANJA",
        0x5195: "HANJA",
        0x5196: "HANJA",
        0x5197: "HANJA",
        0x5198: "HANJA",
        0x5199: "HANJA",
        0x519A: "HANJA",
        0x519B: "HANJA",
        0x519C: "HANJA",
        0x519D: "HANJA",
        0x519E: "HANJA",
        0x519F: "HANJA",
        0x51A0: "HANJA",
        0x51A1: "HANJA",
        0x51A2: "HANJA",
        0x51A3: "HANJA",
        0x51A4: "HANJA",
        0x51A5: "HANJA",
        0x51A6: "HANJA",
        0x51A7: "HANJA",
        0x51A8: "HANJA",
        0x51A9: "HANJA",
        0x51AA: "HANJA",
        0x51AB: "HANJA",
        0x51AC: "HANJA",
        0x51AD: "HANJA",
        0x51AE: "HANJA",
        0x51AF: "HANJA",
        0x51B0: "HANJA",
        0x51B1: "HANJA",
        0x51B2: "HANJA",
        0x51B3: "HANJA",
        0x51B4: "HANJA",
        0x51B5: "HANJA",
        0x51B6: "HANJA",
        0x51B7: "HANJA",
        0x51B8: "HANJA",
        0x51B9: "HANJA",
        0x51BA: "HANJA",
        0x51BB: "HANJA",
        0x51BC: "HANJA",
        0x51BD: "HANJA",
        0x51BE: "HANJA",
        0x51BF: "HANJA",
        0x51C0: "HANJA",
        0x51C1: "HANJA",
        0x51C2: "HANJA",
        0x51C3: "HANJA",
        0x51C4: "HANJA",
        0x51C5: "HANJA",
        0x51C6: "HANJA",
        0x51C7: "HANJA",
        0x51C8: "HANJA",
        0x51C9: "HANJA",
        0x51CA: "HANJA",
        0x51CB: "HANJA",
        0x51CC: "HANJA",
        0x51CD: "HANJA",
        0x51CE: "HANJA",
        0x51CF: "HANJA",
        0x51D0: "HANJA",
        0x51D1: "HANJA",
        0x51D2: "HANJA",
        0x51D3: "HANJA",
        0x51D4: "HANJA",
        0x51D5: "HANJA",
        0x51D6: "HANJA",
        0x51D7: "HANJA",
        0x51D8: "HANJA",
        0x51D9: "HANJA",
        0x51DA: "HANJA",
        0x51DB: "HANJA",
        0x51DC: "HANJA",
        0x51DD: "HANJA",
        0x51DE: "HANJA",
        0x51DF: "HANJA",
        0x51E0: "HANJA",
        0x51E1: "HANJA",
        0x51E2: "HANJA",
        0x51E3: "HANJA",
        0x51E4: "HANJA",
        0x51E5: "HANJA",
        0x51E6: "HANJA",
        0x51E7: "HANJA",
        0x51E8: "HANJA",
        0x51E9: "HANJA",
        0x51EA: "HANJA",
        0x51EB: "HANJA",
        0x51EC: "HANJA",
        0x51ED: "HANJA",
        0x51EE: "HANJA",
        0x51EF: "HANJA",
        0x51F0: "HANJA",
        0x51F1: "HANJA",
        0x51F2: "HANJA",
        0x51F3: "HANJA",
        0x51F4: "HANJA",
        0x51F5: "HANJA",
        0x51F6: "HANJA",
        0x51F7: "HANJA",
        0x51F8: "HANJA",
        0x51F9: "HANJA",
        0x51FA: "HANJA",
        0x51FB: "HANJA",
        0x51FC: "HANJA",
        0x51FD: "HANJA",
        0x51FE: "HANJA",
        0x51FF: "HANJA",
        0x5200: "HANJA",
        0x5201: "HANJA",
        0x5202: "HANJA",
        0x5203: "HANJA",
        0x5204: "HANJA",
        0x5205: "HANJA",
        0x5206: "HANJA",
        0x5207: "HANJA",
        0x5208: "HANJA",
        0x5209: "HANJA",
        0x520A: "HANJA",
        0x520B: "HANJA",
        0x520C: "HANJA",
        0x520D: "HANJA",
        0x520E: "HANJA",
        0x520F: "HANJA",
        0x5210: "HANJA",
        0x5211: "HANJA",
        0x5212: "HANJA",
        0x5213: "HANJA",
        0x5214: "HANJA",
        0x5215: "HANJA",
        0x5216: "HANJA",
        0x5217: "HANJA",
        0x5218: "HANJA",
        0x5219: "HANJA",
        0x521A: "HANJA",
        0x521B: "HANJA",
        0x521C: "HANJA",
        0x521D: "HANJA",
        0x521E: "HANJA",
        0x521F: "HANJA",
        0x5220: "HANJA",
        0x5221: "HANJA",
        0x5222: "HANJA",
        0x5223: "HANJA",
        0x5224: "HANJA",
        0x5225: "HANJA",
        0x5226: "HANJA",
        0x5227: "HANJA",
        0x5228: "HANJA",
        0x5229: "HANJA",
        0x522A: "HANJA",
        0x522B: "HANJA",
        0x522C: "HANJA",
        0x522D: "HANJA",
        0x522E: "HANJA",
        0x522F: "HANJA",
        0x5230: "HANJA",
        0x5231: "HANJA",
        0x5232: "HANJA",
        0x5233: "HANJA",
        0x5234: "HANJA",
        0x5235: "HANJA",
        0x5236: "HANJA",
        0x5237: "HANJA",
        0x5238: "HANJA",
        0x5239: "HANJA",
        0x523A: "HANJA",
        0x523B: "HANJA",
        0x523C: "HANJA",
        0x523D: "HANJA",
        0x523E: "HANJA",
        0x523F: "HANJA",
        0x5240: "HANJA",
        0x5241: "HANJA",
        0x5242: "HANJA",
        0x5243: "HANJA",
        0x5244: "HANJA",
        0x5245: "HANJA",
        0x5246: "HANJA",
        0x5247: "HANJA",
        0x5248: "HANJA",
        0x5249: "HANJA",
        0x524A: "HANJA",
        0x524B: "HANJA",
        0x524C: "HANJA",
        0x524D: "HANJA",
        0x524E: "HANJA",
        0x524F: "HANJA",
        0x5250: "HANJA",
        0x5251: "HANJA",
        0x5252: "HANJA",
        0x5253: "HANJA",
        0x5254: "HANJA",
        0x5255: "HANJA",
        0x5256: "HANJA",
        0x5257: "HANJA",
        0x5258: "HANJA",
        0x5259: "HANJA",
        0x525A: "HANJA",
        0x525B: "HANJA",
        0x525C: "HANJA",
        0x525D: "HANJA",
        0x525E: "HANJA",
        0x525F: "HANJA",
        0x5260: "HANJA",
        0x5261: "HANJA",
        0x5262: "HANJA",
        0x5263: "HANJA",
        0x5264: "HANJA",
        0x5265: "HANJA",
        0x5266: "HANJA",
        0x5267: "HANJA",
        0x5268: "HANJA",
        0x5269: "HANJA",
        0x526A: "HANJA",
        0x526B: "HANJA",
        0x526C: "HANJA",
        0x526D: "HANJA",
        0x526E: "HANJA",
        0x526F: "HANJA",
        0x5270: "HANJA",
        0x5271: "HANJA",
        0x5272: "HANJA",
        0x5273: "HANJA",
        0x5274: "HANJA",
        0x5275: "HANJA",
        0x5276: "HANJA",
        0x5277: "HANJA",
        0x5278: "HANJA",
        0x5279: "HANJA",
        0x527A: "HANJA",
        0x527B: "HANJA",
        0x527C: "HANJA",
        0x527D: "HANJA",
        0x527E: "HANJA",
        0x527F: "HANJA",
        0x5280: "HANJA",
        0x5281: "HANJA",
        0x5282: "HANJA",
        0x5283: "HANJA",
        0x5284: "HANJA",
        0x5285: "HANJA",
        0x5286: "HANJA",
        0x5287: "HANJA",
        0x5288: "HANJA",
        0x5289: "HANJA",
        0x528A: "HANJA",
        0x528B: "HANJA",
        0x528C: "HANJA",
        0x528D: "HANJA",
        0x528E: "HANJA",
        0x528F: "HANJA",
        0x5290: "HANJA",
        0x5291: "HANJA",
        0x5292: "HANJA",
        0x5293: "HANJA",
        0x5294: "HANJA",
        0x5295: "HANJA",
        0x5296: "HANJA",
        0x5297: "HANJA",
        0x5298: "HANJA",
        0x5299: "HANJA",
        0x529A: "HANJA",
        0x529B: "HANJA",
        0x529C: "HANJA",
        0x529D: "HANJA",
        0x529E: "HANJA",
        0x529F: "HANJA",
        0x52A0: "HANJA",
        0x52A1: "HANJA",
        0x52A2: "HANJA",
        0x52A3: "HANJA",
        0x52A4: "HANJA",
        0x52A5: "HANJA",
        0x52A6: "HANJA",
        0x52A7: "HANJA",
        0x52A8: "HANJA",
        0x52A9: "HANJA",
        0x52AA: "HANJA",
        0x52AB: "HANJA",
        0x52AC: "HANJA",
        0x52AD: "HANJA",
        0x52AE: "HANJA",
        0x52AF: "HANJA",
        0x52B0: "HANJA",
        0x52B1: "HANJA",
        0x52B2: "HANJA",
        0x52B3: "HANJA",
        0x52B4: "HANJA",
        0x52B5: "HANJA",
        0x52B6: "HANJA",
        0x52B7: "HANJA",
        0x52B8: "HANJA",
        0x52B9: "HANJA",
        0x52BA: "HANJA",
        0x52BB: "HANJA",
        0x52BC: "HANJA",
        0x52BD: "HANJA",
        0x52BE: "HANJA",
        0x52BF: "HANJA",
        0x52C0: "HANJA",
        0x52C1: "HANJA",
        0x52C2: "HANJA",
        0x52C3: "HANJA",
        0x52C4: "HANJA",
        0x52C5: "HANJA",
        0x52C6: "HANJA",
        0x52C7: "HANJA",
        0x52C8: "HANJA",
        0x52C9: "HANJA",
        0x52CA: "HANJA",
        0x52CB: "HANJA",
        0x52CC: "HANJA",
        0x52CD: "HANJA",
        0x52CE: "HANJA",
        0x52CF: "HANJA",
        0x52D0: "HANJA",
        0x52D1: "HANJA",
        0x52D2: "HANJA",
        0x52D3: "HANJA",
        0x52D4: "HANJA",
        0x52D5: "HANJA",
        0x52D6: "HANJA",
        0x52D7: "HANJA",
        0x52D8: "HANJA",
        0x52D9: "HANJA",
        0x52DA: "HANJA",
        0x52DB: "HANJA",
        0x52DC: "HANJA",
        0x52DD: "HANJA",
        0x52DE: "HANJA",
        0x52DF: "HANJA",
        0x52E0: "HANJA",
        0x52E1: "HANJA",
        0x52E2: "HANJA",
        0x52E3: "HANJA",
        0x52E4: "HANJA",
        0x52E5: "HANJA",
        0x52E6: "HANJA",
        0x52E7: "HANJA",
        0x52E8: "HANJA",
        0x52E9: "HANJA",
        0x52EA: "HANJA",
        0x52EB: "HANJA",
        0x52EC: "HANJA",
        0x52ED: "HANJA",
        0x52EE: "HANJA",
        0x52EF: "HANJA",
        0x52F0: "HANJA",
        0x52F1: "HANJA",
        0x52F2: "HANJA",
        0x52F3: "HANJA",
        0x52F4: "HANJA",
        0x52F5: "HANJA",
        0x52F6: "HANJA",
        0x52F7: "HANJA",
        0x52F8: "HANJA",
        0x52F9: "HANJA",
        0x52FA: "HANJA",
        0x52FB: "HANJA",
        0x52FC: "HANJA",
        0x52FD: "HANJA",
        0x52FE: "HANJA",
        0x52FF: "HANJA",
        0x5300: "HANJA",
        0x5301: "HANJA",
        0x5302: "HANJA",
        0x5303: "HANJA",
        0x5304: "HANJA",
        0x5305: "HANJA",
        0x5306: "HANJA",
        0x5307: "HANJA",
        0x5308: "HANJA",
        0x5309: "HANJA",
        0x530A: "HANJA",
        0x530B: "HANJA",
        0x530C: "HANJA",
        0x530D: "HANJA",
        0x530E: "HANJA",
        0x530F: "HANJA",
        0x5310: "HANJA",
        0x5311: "HANJA",
        0x5312: "HANJA",
        0x5313: "HANJA",
        0x5314: "HANJA",
        0x5315: "HANJA",
        0x5316: "HANJA",
        0x5317: "HANJA",
        0x5318: "HANJA",
        0x5319: "HANJA",
        0x531A: "HANJA",
        0x531B: "HANJA",
        0x531C: "HANJA",
        0x531D: "HANJA",
        0x531E: "HANJA",
        0x531F: "HANJA",
        0x5320: "HANJA",
        0x5321: "HANJA",
        0x5322: "HANJA",
        0x5323: "HANJA",
        0x5324: "HANJA",
        0x5325: "HANJA",
        0x5326: "HANJA",
        0x5327: "HANJA",
        0x5328: "HANJA",
        0x5329: "HANJA",
        0x532A: "HANJA",
        0x532B: "HANJA",
        0x532C: "HANJA",
        0x532D: "HANJA",
        0x532E: "HANJA",
        0x532F: "HANJA",
        0x5330: "HANJA",
        0x5331: "HANJA",
        0x5332: "HANJA",
        0x5333: "HANJA",
        0x5334: "HANJA",
        0x5335: "HANJA",
        0x5336: "HANJA",
        0x5337: "HANJA",
        0x5338: "HANJA",
        0x5339: "HANJA",
        0x533A: "HANJA",
        0x533B: "HANJA",
        0x533C: "HANJA",
        0x533D: "HANJA",
        0x533E: "HANJA",
        0x533F: "HANJA",
        0x5340: "HANJA",
        0x5341: "HANJA",
        0x5342: "HANJA",
        0x5343: "HANJA",
        0x5344: "HANJA",
        0x5345: "HANJA",
        0x5346: "HANJA",
        0x5347: "HANJA",
        0x5348: "HANJA",
        0x5349: "HANJA",
        0x534A: "HANJA",
        0x534B: "HANJA",
        0x534C: "HANJA",
        0x534D: "HANJA",
        0x534E: "HANJA",
        0x534F: "HANJA",
        0x5350: "HANJA",
        0x5351: "HANJA",
        0x5352: "HANJA",
        0x5353: "HANJA",
        0x5354: "HANJA",
        0x5355: "HANJA",
        0x5356: "HANJA",
        0x5357: "HANJA",
        0x5358: "HANJA",
        0x5359: "HANJA",
        0x535A: "HANJA",
        0x535B: "HANJA",
        0x535C: "HANJA",
        0x535D: "HANJA",
        0x535E: "HANJA",
        0x535F: "HANJA",
        0x5360: "HANJA",
        0x5361: "HANJA",
        0x5362: "HANJA",
        0x5363: "HANJA",
        0x5364: "HANJA",
        0x5365: "HANJA",
        0x5366: "HANJA",
        0x5367: "HANJA",
        0x5368: "HANJA",
        0x5369: "HANJA",
        0x536A: "HANJA",
        0x536B: "HANJA",
        0x536C: "HANJA",
        0x536D: "HANJA",
        0x536E: "HANJA",
        0x536F: "HANJA",
        0x5370: "HANJA",
        0x5371: "HANJA",
        0x5372: "HANJA",
        0x5373: "HANJA",
        0x5374: "HANJA",
        0x5375: "HANJA",
        0x5376: "HANJA",
        0x5377: "HANJA",
        0x5378: "HANJA",
        0x5379: "HANJA",
        0x537A: "HANJA",
        0x537B: "HANJA",
        0x537C: "HANJA",
        0x537D: "HANJA",
        0x537E: "HANJA",
        0x537F: "HANJA",
        0x5380: "HANJA",
        0x5381: "HANJA",
        0x5382: "HANJA",
        0x5383: "HANJA",
        0x5384: "HANJA",
        0x5385: "HANJA",
        0x5386: "HANJA",
        0x5387: "HANJA",
        0x5388: "HANJA",
        0x5389: "HANJA",
        0x538A: "HANJA",
        0x538B: "HANJA",
        0x538C: "HANJA",
        0x538D: "HANJA",
        0x538E: "HANJA",
        0x538F: "HANJA",
        0x5390: "HANJA",
        0x5391: "HANJA",
        0x5392: "HANJA",
        0x5393: "HANJA",
        0x5394: "HANJA",
        0x5395: "HANJA",
        0x5396: "HANJA",
        0x5397: "HANJA",
        0x5398: "HANJA",
        0x5399: "HANJA",
        0x539A: "HANJA",
        0x539B: "HANJA",
        0x539C: "HANJA",
        0x539D: "HANJA",
        0x539E: "HANJA",
        0x539F: "HANJA",
        0x53A0: "HANJA",
        0x53A1: "HANJA",
        0x53A2: "HANJA",
        0x53A3: "HANJA",
        0x53A4: "HANJA",
        0x53A5: "HANJA",
        0x53A6: "HANJA",
        0x53A7: "HANJA",
        0x53A8: "HANJA",
        0x53A9: "HANJA",
        0x53AA: "HANJA",
        0x53AB: "HANJA",
        0x53AC: "HANJA",
        0x53AD: "HANJA",
        0x53AE: "HANJA",
        0x53AF: "HANJA",
        0x53B0: "HANJA",
        0x53B1: "HANJA",
        0x53B2: "HANJA",
        0x53B3: "HANJA",
        0x53B4: "HANJA",
        0x53B5: "HANJA",
        0x53B6: "HANJA",
        0x53B7: "HANJA",
        0x53B8: "HANJA",
        0x53B9: "HANJA",
        0x53BA: "HANJA",
        0x53BB: "HANJA",
        0x53BC: "HANJA",
        0x53BD: "HANJA",
        0x53BE: "HANJA",
        0x53BF: "HANJA",
        0x53C0: "HANJA",
        0x53C1: "HANJA",
        0x53C2: "HANJA",
        0x53C3: "HANJA",
        0x53C4: "HANJA",
        0x53C5: "HANJA",
        0x53C6: "HANJA",
        0x53C7: "HANJA",
        0x53C8: "HANJA",
        0x53C9: "HANJA",
        0x53CA: "HANJA",
        0x53CB: "HANJA",
        0x53CC: "HANJA",
        0x53CD: "HANJA",
        0x53CE: "HANJA",
        0x53CF: "HANJA",
        0x53D0: "HANJA",
        0x53D1: "HANJA",
        0x53D2: "HANJA",
        0x53D3: "HANJA",
        0x53D4: "HANJA",
        0x53D5: "HANJA",
        0x53D6: "HANJA",
        0x53D7: "HANJA",
        0x53D8: "HANJA",
        0x53D9: "HANJA",
        0x53DA: "HANJA",
        0x53DB: "HANJA",
        0x53DC: "HANJA",
        0x53DD: "HANJA",
        0x53DE: "HANJA",
        0x53DF: "HANJA",
        0x53E0: "HANJA",
        0x53E1: "HANJA",
        0x53E2: "HANJA",
        0x53E3: "HANJA",
        0x53E4: "HANJA",
        0x53E5: "HANJA",
        0x53E6: "HANJA",
        0x53E7: "HANJA",
        0x53E8: "HANJA",
        0x53E9: "HANJA",
        0x53EA: "HANJA",
        0x53EB: "HANJA",
        0x53EC: "HANJA",
        0x53ED: "HANJA",
        0x53EE: "HANJA",
        0x53EF: "HANJA",
        0x53F0: "HANJA",
        0x53F1: "HANJA",
        0x53F2: "HANJA",
        0x53F3: "HANJA",
        0x53F4: "HANJA",
        0x53F5: "HANJA",
        0x53F6: "HANJA",
        0x53F7: "HANJA",
        0x53F8: "HANJA",
        0x53F9: "HANJA",
        0x53FA: "HANJA",
        0x53FB: "HANJA",
        0x53FC: "HANJA",
        0x53FD: "HANJA",
        0x53FE: "HANJA",
        0x53FF: "HANJA",
        0x5400: "HANJA",
        0x5401: "HANJA",
        0x5402: "HANJA",
        0x5403: "HANJA",
        0x5404: "HANJA",
        0x5405: "HANJA",
        0x5406: "HANJA",
        0x5407: "HANJA",
        0x5408: "HANJA",
        0x5409: "HANJA",
        0x540A: "HANJA",
        0x540B: "HANJA",
        0x540C: "HANJA",
        0x540D: "HANJA",
        0x540E: "HANJA",
        0x540F: "HANJA",
        0x5410: "HANJA",
        0x5411: "HANJA",
        0x5412: "HANJA",
        0x5413: "HANJA",
        0x5414: "HANJA",
        0x5415: "HANJA",
        0x5416: "HANJA",
        0x5417: "HANJA",
        0x5418: "HANJA",
        0x5419: "HANJA",
        0x541A: "HANJA",
        0x541B: "HANJA",
        0x541C: "HANJA",
        0x541D: "HANJA",
        0x541E: "HANJA",
        0x541F: "HANJA",
        0x5420: "HANJA",
        0x5421: "HANJA",
        0x5422: "HANJA",
        0x5423: "HANJA",
        0x5424: "HANJA",
        0x5425: "HANJA",
        0x5426: "HANJA",
        0x5427: "HANJA",
        0x5428: "HANJA",
        0x5429: "HANJA",
        0x542A: "HANJA",
        0x542B: "HANJA",
        0x542C: "HANJA",
        0x542D: "HANJA",
        0x542E: "HANJA",
        0x542F: "HANJA",
        0x5430: "HANJA",
        0x5431: "HANJA",
        0x5432: "HANJA",
        0x5433: "HANJA",
        0x5434: "HANJA",
        0x5435: "HANJA",
        0x5436: "HANJA",
        0x5437: "HANJA",
        0x5438: "HANJA",
        0x5439: "HANJA",
        0x543A: "HANJA",
        0x543B: "HANJA",
        0x543C: "HANJA",
        0x543D: "HANJA",
        0x543E: "HANJA",
        0x543F: "HANJA",
        0x5440: "HANJA",
        0x5441: "HANJA",
        0x5442: "HANJA",
        0x5443: "HANJA",
        0x5444: "HANJA",
        0x5445: "HANJA",
        0x5446: "HANJA",
        0x5447: "HANJA",
        0x5448: "HANJA",
        0x5449: "HANJA",
        0x544A: "HANJA",
        0x544B: "HANJA",
        0x544C: "HANJA",
        0x544D: "HANJA",
        0x544E: "HANJA",
        0x544F: "HANJA",
        0x5450: "HANJA",
        0x5451: "HANJA",
        0x5452: "HANJA",
        0x5453: "HANJA",
        0x5454: "HANJA",
        0x5455: "HANJA",
        0x5456: "HANJA",
        0x5457: "HANJA",
        0x5458: "HANJA",
        0x5459: "HANJA",
        0x545A: "HANJA",
        0x545B: "HANJA",
        0x545C: "HANJA",
        0x545D: "HANJA",
        0x545E: "HANJA",
        0x545F: "HANJA",
        0x5460: "HANJA",
        0x5461: "HANJA",
        0x5462: "HANJA",
        0x5463: "HANJA",
        0x5464: "HANJA",
        0x5465: "HANJA",
        0x5466: "HANJA",
        0x5467: "HANJA",
        0x5468: "HANJA",
        0x5469: "HANJA",
        0x546A: "HANJA",
        0x546B: "HANJA",
        0x546C: "HANJA",
        0x546D: "HANJA",
        0x546E: "HANJA",
        0x546F: "HANJA",
        0x5470: "HANJA",
        0x5471: "HANJA",
        0x5472: "HANJA",
        0x5473: "HANJA",
        0x5474: "HANJA",
        0x5475: "HANJA",
        0x5476: "HANJA",
        0x5477: "HANJA",
        0x5478: "HANJA",
        0x5479: "HANJA",
        0x547A: "HANJA",
        0x547B: "HANJA",
        0x547C: "HANJA",
        0x547D: "HANJA",
        0x547E: "HANJA",
        0x547F: "HANJA",
        0x5480: "HANJA",
        0x5481: "HANJA",
        0x5482: "HANJA",
        0x5483: "HANJA",
        0x5484: "HANJA",
        0x5485: "HANJA",
        0x5486: "HANJA",
        0x5487: "HANJA",
        0x5488: "HANJA",
        0x5489: "HANJA",
        0x548A: "HANJA",
        0x548B: "HANJA",
        0x548C: "HANJA",
        0x548D: "HANJA",
        0x548E: "HANJA",
        0x548F: "HANJA",
        0x5490: "HANJA",
        0x5491: "HANJA",
        0x5492: "HANJA",
        0x5493: "HANJA",
        0x5494: "HANJA",
        0x5495: "HANJA",
        0x5496: "HANJA",
        0x5497: "HANJA",
        0x5498: "HANJA",
        0x5499: "HANJA",
        0x549A: "HANJA",
        0x549B: "HANJA",
        0x549C: "HANJA",
        0x549D: "HANJA",
        0x549E: "HANJA",
        0x549F: "HANJA",
        0x54A0: "HANJA",
        0x54A1: "HANJA",
        0x54A2: "HANJA",
        0x54A3: "HANJA",
        0x54A4: "HANJA",
        0x54A5: "HANJA",
        0x54A6: "HANJA",
        0x54A7: "HANJA",
        0x54A8: "HANJA",
        0x54A9: "HANJA",
        0x54AA: "HANJA",
        0x54AB: "HANJA",
        0x54AC: "HANJA",
        0x54AD: "HANJA",
        0x54AE: "HANJA",
        0x54AF: "HANJA",
        0x54B0: "HANJA",
        0x54B1: "HANJA",
        0x54B2: "HANJA",
        0x54B3: "HANJA",
        0x54B4: "HANJA",
        0x54B5: "HANJA",
        0x54B6: "HANJA",
        0x54B7: "HANJA",
        0x54B8: "HANJA",
        0x54B9: "HANJA",
        0x54BA: "HANJA",
        0x54BB: "HANJA",
        0x54BC: "HANJA",
        0x54BD: "HANJA",
        0x54BE: "HANJA",
        0x54BF: "HANJA",
        0x54C0: "HANJA",
        0x54C1: "HANJA",
        0x54C2: "HANJA",
        0x54C3: "HANJA",
        0x54C4: "HANJA",
        0x54C5: "HANJA",
        0x54C6: "HANJA",
        0x54C7: "HANJA",
        0x54C8: "HANJA",
        0x54C9: "HANJA",
        0x54CA: "HANJA",
        0x54CB: "HANJA",
        0x54CC: "HANJA",
        0x54CD: "HANJA",
        0x54CE: "HANJA",
        0x54CF: "HANJA",
        0x54D0: "HANJA",
        0x54D1: "HANJA",
        0x54D2: "HANJA",
        0x54D3: "HANJA",
        0x54D4: "HANJA",
        0x54D5: "HANJA",
        0x54D6: "HANJA",
        0x54D7: "HANJA",
        0x54D8: "HANJA",
        0x54D9: "HANJA",
        0x54DA: "HANJA",
        0x54DB: "HANJA",
        0x54DC: "HANJA",
        0x54DD: "HANJA",
        0x54DE: "HANJA",
        0x54DF: "HANJA",
        0x54E0: "HANJA",
        0x54E1: "HANJA",
        0x54E2: "HANJA",
        0x54E3: "HANJA",
        0x54E4: "HANJA",
        0x54E5: "HANJA",
        0x54E6: "HANJA",
        0x54E7: "HANJA",
        0x54E8: "HANJA",
        0x54E9: "HANJA",
        0x54EA: "HANJA",
        0x54EB: "HANJA",
        0x54EC: "HANJA",
        0x54ED: "HANJA",
        0x54EE: "HANJA",
        0x54EF: "HANJA",
        0x54F0: "HANJA",
        0x54F1: "HANJA",
        0x54F2: "HANJA",
        0x54F3: "HANJA",
        0x54F4: "HANJA",
        0x54F5: "HANJA",
        0x54F6: "HANJA",
        0x54F7: "HANJA",
        0x54F8: "HANJA",
        0x54F9: "HANJA",
        0x54FA: "HANJA",
        0x54FB: "HANJA",
        0x54FC: "HANJA",
        0x54FD: "HANJA",
        0x54FE: "HANJA",
        0x54FF: "HANJA",
        0x5500: "HANJA",
        0x5501: "HANJA",
        0x5502: "HANJA",
        0x5503: "HANJA",
        0x5504: "HANJA",
        0x5505: "HANJA",
        0x5506: "HANJA",
        0x5507: "HANJA",
        0x5508: "HANJA",
        0x5509: "HANJA",
        0x550A: "HANJA",
        0x550B: "HANJA",
        0x550C: "HANJA",
        0x550D: "HANJA",
        0x550E: "HANJA",
        0x550F: "HANJA",
        0x5510: "HANJA",
        0x5511: "HANJA",
        0x5512: "HANJA",
        0x5513: "HANJA",
        0x5514: "HANJA",
        0x5515: "HANJA",
        0x5516: "HANJA",
        0x5517: "HANJA",
        0x5518: "HANJA",
        0x5519: "HANJA",
        0x551A: "HANJA",
        0x551B: "HANJA",
        0x551C: "HANJA",
        0x551D: "HANJA",
        0x551E: "HANJA",
        0x551F: "HANJA",
        0x5520: "HANJA",
        0x5521: "HANJA",
        0x5522: "HANJA",
        0x5523: "HANJA",
        0x5524: "HANJA",
        0x5525: "HANJA",
        0x5526: "HANJA",
        0x5527: "HANJA",
        0x5528: "HANJA",
        0x5529: "HANJA",
        0x552A: "HANJA",
        0x552B: "HANJA",
        0x552C: "HANJA",
        0x552D: "HANJA",
        0x552E: "HANJA",
        0x552F: "HANJA",
        0x5530: "HANJA",
        0x5531: "HANJA",
        0x5532: "HANJA",
        0x5533: "HANJA",
        0x5534: "HANJA",
        0x5535: "HANJA",
        0x5536: "HANJA",
        0x5537: "HANJA",
        0x5538: "HANJA",
        0x5539: "HANJA",
        0x553A: "HANJA",
        0x553B: "HANJA",
        0x553C: "HANJA",
        0x553D: "HANJA",
        0x553E: "HANJA",
        0x553F: "HANJA",
        0x5540: "HANJA",
        0x5541: "HANJA",
        0x5542: "HANJA",
        0x5543: "HANJA",
        0x5544: "HANJA",
        0x5545: "HANJA",
        0x5546: "HANJA",
        0x5547: "HANJA",
        0x5548: "HANJA",
        0x5549: "HANJA",
        0x554A: "HANJA",
        0x554B: "HANJA",
        0x554C: "HANJA",
        0x554D: "HANJA",
        0x554E: "HANJA",
        0x554F: "HANJA",
        0x5550: "HANJA",
        0x5551: "HANJA",
        0x5552: "HANJA",
        0x5553: "HANJA",
        0x5554: "HANJA",
        0x5555: "HANJA",
        0x5556: "HANJA",
        0x5557: "HANJA",
        0x5558: "HANJA",
        0x5559: "HANJA",
        0x555A: "HANJA",
        0x555B: "HANJA",
        0x555C: "HANJA",
        0x555D: "HANJA",
        0x555E: "HANJA",
        0x555F: "HANJA",
        0x5560: "HANJA",
        0x5561: "HANJA",
        0x5562: "HANJA",
        0x5563: "HANJA",
        0x5564: "HANJA",
        0x5565: "HANJA",
        0x5566: "HANJA",
        0x5567: "HANJA",
        0x5568: "HANJA",
        0x5569: "HANJA",
        0x556A: "HANJA",
        0x556B: "HANJA",
        0x556C: "HANJA",
        0x556D: "HANJA",
        0x556E: "HANJA",
        0x556F: "HANJA",
        0x5570: "HANJA",
        0x5571: "HANJA",
        0x5572: "HANJA",
        0x5573: "HANJA",
        0x5574: "HANJA",
        0x5575: "HANJA",
        0x5576: "HANJA",
        0x5577: "HANJA",
        0x5578: "HANJA",
        0x5579: "HANJA",
        0x557A: "HANJA",
        0x557B: "HANJA",
        0x557C: "HANJA",
        0x557D: "HANJA",
        0x557E: "HANJA",
        0x557F: "HANJA",
        0x5580: "HANJA",
        0x5581: "HANJA",
        0x5582: "HANJA",
        0x5583: "HANJA",
        0x5584: "HANJA",
        0x5585: "HANJA",
        0x5586: "HANJA",
        0x5587: "HANJA",
        0x5588: "HANJA",
        0x5589: "HANJA",
        0x558A: "HANJA",
        0x558B: "HANJA",
        0x558C: "HANJA",
        0x558D: "HANJA",
        0x558E: "HANJA",
        0x558F: "HANJA",
        0x5590: "HANJA",
        0x5591: "HANJA",
        0x5592: "HANJA",
        0x5593: "HANJA",
        0x5594: "HANJA",
        0x5595: "HANJA",
        0x5596: "HANJA",
        0x5597: "HANJA",
        0x5598: "HANJA",
        0x5599: "HANJA",
        0x559A: "HANJA",
        0x559B: "HANJA",
        0x559C: "HANJA",
        0x559D: "HANJA",
        0x559E: "HANJA",
        0x559F: "HANJA",
        0x55A0: "HANJA",
        0x55A1: "HANJA",
        0x55A2: "HANJA",
        0x55A3: "HANJA",
        0x55A4: "HANJA",
        0x55A5: "HANJA",
        0x55A6: "HANJA",
        0x55A7: "HANJA",
        0x55A8: "HANJA",
        0x55A9: "HANJA",
        0x55AA: "HANJA",
        0x55AB: "HANJA",
        0x55AC: "HANJA",
        0x55AD: "HANJA",
        0x55AE: "HANJA",
        0x55AF: "HANJA",
        0x55B0: "HANJA",
        0x55B1: "HANJA",
        0x55B2: "HANJA",
        0x55B3: "HANJA",
        0x55B4: "HANJA",
        0x55B5: "HANJA",
        0x55B6: "HANJA",
        0x55B7: "HANJA",
        0x55B8: "HANJA",
        0x55B9: "HANJA",
        0x55BA: "HANJA",
        0x55BB: "HANJA",
        0x55BC: "HANJA",
        0x55BD: "HANJA",
        0x55BE: "HANJA",
        0x55BF: "HANJA",
        0x55C0: "HANJA",
        0x55C1: "HANJA",
        0x55C2: "HANJA",
        0x55C3: "HANJA",
        0x55C4: "HANJA",
        0x55C5: "HANJA",
        0x55C6: "HANJA",
        0x55C7: "HANJA",
        0x55C8: "HANJA",
        0x55C9: "HANJA",
        0x55CA: "HANJA",
        0x55CB: "HANJA",
        0x55CC: "HANJA",
        0x55CD: "HANJA",
        0x55CE: "HANJA",
        0x55CF: "HANJA",
        0x55D0: "HANJA",
        0x55D1: "HANJA",
        0x55D2: "HANJA",
        0x55D3: "HANJA",
        0x55D4: "HANJA",
        0x55D5: "HANJA",
        0x55D6: "HANJA",
        0x55D7: "HANJA",
        0x55D8: "HANJA",
        0x55D9: "HANJA",
        0x55DA: "HANJA",
        0x55DB: "HANJA",
        0x55DC: "HANJA",
        0x55DD: "HANJA",
        0x55DE: "HANJA",
        0x55DF: "HANJA",
        0x55E0: "HANJA",
        0x55E1: "HANJA",
        0x55E2: "HANJA",
        0x55E3: "HANJA",
        0x55E4: "HANJA",
        0x55E5: "HANJA",
        0x55E6: "HANJA",
        0x55E7: "HANJA",
        0x55E8: "HANJA",
        0x55E9: "HANJA",
        0x55EA: "HANJA",
        0x55EB: "HANJA",
        0x55EC: "HANJA",
        0x55ED: "HANJA",
        0x55EE: "HANJA",
        0x55EF: "HANJA",
        0x55F0: "HANJA",
        0x55F1: "HANJA",
        0x55F2: "HANJA",
        0x55F3: "HANJA",
        0x55F4: "HANJA",
        0x55F5: "HANJA",
        0x55F6: "HANJA",
        0x55F7: "HANJA",
        0x55F8: "HANJA",
        0x55F9: "HANJA",
        0x55FA: "HANJA",
        0x55FB: "HANJA",
        0x55FC: "HANJA",
        0x55FD: "HANJA",
        0x55FE: "HANJA",
        0x55FF: "HANJA",
        0x5600: "HANJA",
        0x5601: "HANJA",
        0x5602: "HANJA",
        0x5603: "HANJA",
        0x5604: "HANJA",
        0x5605: "HANJA",
        0x5606: "HANJA",
        0x5607: "HANJA",
        0x5608: "HANJA",
        0x5609: "HANJA",
        0x560A: "HANJA",
        0x560B: "HANJA",
        0x560C: "HANJA",
        0x560D: "HANJA",
        0x560E: "HANJA",
        0x560F: "HANJA",
        0x5610: "HANJA",
        0x5611: "HANJA",
        0x5612: "HANJA",
        0x5613: "HANJA",
        0x5614: "HANJA",
        0x5615: "HANJA",
        0x5616: "HANJA",
        0x5617: "HANJA",
        0x5618: "HANJA",
        0x5619: "HANJA",
        0x561A: "HANJA",
        0x561B: "HANJA",
        0x561C: "HANJA",
        0x561D: "HANJA",
        0x561E: "HANJA",
        0x561F: "HANJA",
        0x5620: "HANJA",
        0x5621: "HANJA",
        0x5622: "HANJA",
        0x5623: "HANJA",
        0x5624: "HANJA",
        0x5625: "HANJA",
        0x5626: "HANJA",
        0x5627: "HANJA",
        0x5628: "HANJA",
        0x5629: "HANJA",
        0x562A: "HANJA",
        0x562B: "HANJA",
        0x562C: "HANJA",
        0x562D: "HANJA",
        0x562E: "HANJA",
        0x562F: "HANJA",
        0x5630: "HANJA",
        0x5631: "HANJA",
        0x5632: "HANJA",
        0x5633: "HANJA",
        0x5634: "HANJA",
        0x5635: "HANJA",
        0x5636: "HANJA",
        0x5637: "HANJA",
        0x5638: "HANJA",
        0x5639: "HANJA",
        0x563A: "HANJA",
        0x563B: "HANJA",
        0x563C: "HANJA",
        0x563D: "HANJA",
        0x563E: "HANJA",
        0x563F: "HANJA",
        0x5640: "HANJA",
        0x5641: "HANJA",
        0x5642: "HANJA",
        0x5643: "HANJA",
        0x5644: "HANJA",
        0x5645: "HANJA",
        0x5646: "HANJA",
        0x5647: "HANJA",
        0x5648: "HANJA",
        0x5649: "HANJA",
        0x564A: "HANJA",
        0x564B: "HANJA",
        0x564C: "HANJA",
        0x564D: "HANJA",
        0x564E: "HANJA",
        0x564F: "HANJA",
        0x5650: "HANJA",
        0x5651: "HANJA",
        0x5652: "HANJA",
        0x5653: "HANJA",
        0x5654: "HANJA",
        0x5655: "HANJA",
        0x5656: "HANJA",
        0x5657: "HANJA",
        0x5658: "HANJA",
        0x5659: "HANJA",
        0x565A: "HANJA",
        0x565B: "HANJA",
        0x565C: "HANJA",
        0x565D: "HANJA",
        0x565E: "HANJA",
        0x565F: "HANJA",
        0x5660: "HANJA",
        0x5661: "HANJA",
        0x5662: "HANJA",
        0x5663: "HANJA",
        0x5664: "HANJA",
        0x5665: "HANJA",
        0x5666: "HANJA",
        0x5667: "HANJA",
        0x5668: "HANJA",
        0x5669: "HANJA",
        0x566A: "HANJA",
        0x566B: "HANJA",
        0x566C: "HANJA",
        0x566D: "HANJA",
        0x566E: "HANJA",
        0x566F: "HANJA",
        0x5670: "HANJA",
        0x5671: "HANJA",
        0x5672: "HANJA",
        0x5673: "HANJA",
        0x5674: "HANJA",
        0x5675: "HANJA",
        0x5676: "HANJA",
        0x5677: "HANJA",
        0x5678: "HANJA",
        0x5679: "HANJA",
        0x567A: "HANJA",
        0x567B: "HANJA",
        0x567C: "HANJA",
        0x567D: "HANJA",
        0x567E: "HANJA",
        0x567F: "HANJA",
        0x5680: "HANJA",
        0x5681: "HANJA",
        0x5682: "HANJA",
        0x5683: "HANJA",
        0x5684: "HANJA",
        0x5685: "HANJA",
        0x5686: "HANJA",
        0x5687: "HANJA",
        0x5688: "HANJA",
        0x5689: "HANJA",
        0x568A: "HANJA",
        0x568B: "HANJA",
        0x568C: "HANJA",
        0x568D: "HANJA",
        0x568E: "HANJA",
        0x568F: "HANJA",
        0x5690: "HANJA",
        0x5691: "HANJA",
        0x5692: "HANJA",
        0x5693: "HANJA",
        0x5694: "HANJA",
        0x5695: "HANJA",
        0x5696: "HANJA",
        0x5697: "HANJA",
        0x5698: "HANJA",
        0x5699: "HANJA",
        0x569A: "HANJA",
        0x569B: "HANJA",
        0x569C: "HANJA",
        0x569D: "HANJA",
        0x569E: "HANJA",
        0x569F: "HANJA",
        0x56A0: "HANJA",
        0x56A1: "HANJA",
        0x56A2: "HANJA",
        0x56A3: "HANJA",
        0x56A4: "HANJA",
        0x56A5: "HANJA",
        0x56A6: "HANJA",
        0x56A7: "HANJA",
        0x56A8: "HANJA",
        0x56A9: "HANJA",
        0x56AA: "HANJA",
        0x56AB: "HANJA",
        0x56AC: "HANJA",
        0x56AD: "HANJA",
        0x56AE: "HANJA",
        0x56AF: "HANJA",
        0x56B0: "HANJA",
        0x56B1: "HANJA",
        0x56B2: "HANJA",
        0x56B3: "HANJA",
        0x56B4: "HANJA",
        0x56B5: "HANJA",
        0x56B6: "HANJA",
        0x56B7: "HANJA",
        0x56B8: "HANJA",
        0x56B9: "HANJA",
        0x56BA: "HANJA",
        0x56BB: "HANJA",
        0x56BC: "HANJA",
        0x56BD: "HANJA",
        0x56BE: "HANJA",
        0x56BF: "HANJA",
        0x56C0: "HANJA",
        0x56C1: "HANJA",
        0x56C2: "HANJA",
        0x56C3: "HANJA",
        0x56C4: "HANJA",
        0x56C5: "HANJA",
        0x56C6: "HANJA",
        0x56C7: "HANJA",
        0x56C8: "HANJA",
        0x56C9: "HANJA",
        0x56CA: "HANJA",
        0x56CB: "HANJA",
        0x56CC: "HANJA",
        0x56CD: "HANJA",
        0x56CE: "HANJA",
        0x56CF: "HANJA",
        0x56D0: "HANJA",
        0x56D1: "HANJA",
        0x56D2: "HANJA",
        0x56D3: "HANJA",
        0x56D4: "HANJA",
        0x56D5: "HANJA",
        0x56D6: "HANJA",
        0x56D7: "HANJA",
        0x56D8: "HANJA",
        0x56D9: "HANJA",
        0x56DA: "HANJA",
        0x56DB: "HANJA",
        0x56DC: "HANJA",
        0x56DD: "HANJA",
        0x56DE: "HANJA",
        0x56DF: "HANJA",
        0x56E0: "HANJA",
        0x56E1: "HANJA",
        0x56E2: "HANJA",
        0x56E3: "HANJA",
        0x56E4: "HANJA",
        0x56E5: "HANJA",
        0x56E6: "HANJA",
        0x56E7: "HANJA",
        0x56E8: "HANJA",
        0x56E9: "HANJA",
        0x56EA: "HANJA",
        0x56EB: "HANJA",
        0x56EC: "HANJA",
        0x56ED: "HANJA",
        0x56EE: "HANJA",
        0x56EF: "HANJA",
        0x56F0: "HANJA",
        0x56F1: "HANJA",
        0x56F2: "HANJA",
        0x56F3: "HANJA",
        0x56F4: "HANJA",
        0x56F5: "HANJA",
        0x56F6: "HANJA",
        0x56F7: "HANJA",
        0x56F8: "HANJA",
        0x56F9: "HANJA",
        0x56FA: "HANJA",
        0x56FB: "HANJA",
        0x56FC: "HANJA",
        0x56FD: "HANJA",
        0x56FE: "HANJA",
        0x56FF: "HANJA",
        0x5700: "HANJA",
        0x5701: "HANJA",
        0x5702: "HANJA",
        0x5703: "HANJA",
        0x5704: "HANJA",
        0x5705: "HANJA",
        0x5706: "HANJA",
        0x5707: "HANJA",
        0x5708: "HANJA",
        0x5709: "HANJA",
        0x570A: "HANJA",
        0x570B: "HANJA",
        0x570C: "HANJA",
        0x570D: "HANJA",
        0x570E: "HANJA",
        0x570F: "HANJA",
        0x5710: "HANJA",
        0x5711: "HANJA",
        0x5712: "HANJA",
        0x5713: "HANJA",
        0x5714: "HANJA",
        0x5715: "HANJA",
        0x5716: "HANJA",
        0x5717: "HANJA",
        0x5718: "HANJA",
        0x5719: "HANJA",
        0x571A: "HANJA",
        0x571B: "HANJA",
        0x571C: "HANJA",
        0x571D: "HANJA",
        0x571E: "HANJA",
        0x571F: "HANJA",
        0x5720: "HANJA",
        0x5721: "HANJA",
        0x5722: "HANJA",
        0x5723: "HANJA",
        0x5724: "HANJA",
        0x5725: "HANJA",
        0x5726: "HANJA",
        0x5727: "HANJA",
        0x5728: "HANJA",
        0x5729: "HANJA",
        0x572A: "HANJA",
        0x572B: "HANJA",
        0x572C: "HANJA",
        0x572D: "HANJA",
        0x572E: "HANJA",
        0x572F: "HANJA",
        0x5730: "HANJA",
        0x5731: "HANJA",
        0x5732: "HANJA",
        0x5733: "HANJA",
        0x5734: "HANJA",
        0x5735: "HANJA",
        0x5736: "HANJA",
        0x5737: "HANJA",
        0x5738: "HANJA",
        0x5739: "HANJA",
        0x573A: "HANJA",
        0x573B: "HANJA",
        0x573C: "HANJA",
        0x573D: "HANJA",
        0x573E: "HANJA",
        0x573F: "HANJA",
        0x5740: "HANJA",
        0x5741: "HANJA",
        0x5742: "HANJA",
        0x5743: "HANJA",
        0x5744: "HANJA",
        0x5745: "HANJA",
        0x5746: "HANJA",
        0x5747: "HANJA",
        0x5748: "HANJA",
        0x5749: "HANJA",
        0x574A: "HANJA",
        0x574B: "HANJA",
        0x574C: "HANJA",
        0x574D: "HANJA",
        0x574E: "HANJA",
        0x574F: "HANJA",
        0x5750: "HANJA",
        0x5751: "HANJA",
        0x5752: "HANJA",
        0x5753: "HANJA",
        0x5754: "HANJA",
        0x5755: "HANJA",
        0x5756: "HANJA",
        0x5757: "HANJA",
        0x5758: "HANJA",
        0x5759: "HANJA",
        0x575A: "HANJA",
        0x575B: "HANJA",
        0x575C: "HANJA",
        0x575D: "HANJA",
        0x575E: "HANJA",
        0x575F: "HANJA",
        0x5760: "HANJA",
        0x5761: "HANJA",
        0x5762: "HANJA",
        0x5763: "HANJA",
        0x5764: "HANJA",
        0x5765: "HANJA",
        0x5766: "HANJA",
        0x5767: "HANJA",
        0x5768: "HANJA",
        0x5769: "HANJA",
        0x576A: "HANJA",
        0x576B: "HANJA",
        0x576C: "HANJA",
        0x576D: "HANJA",
        0x576E: "HANJA",
        0x576F: "HANJA",
        0x5770: "HANJA",
        0x5771: "HANJA",
        0x5772: "HANJA",
        0x5773: "HANJA",
        0x5774: "HANJA",
        0x5775: "HANJA",
        0x5776: "HANJA",
        0x5777: "HANJA",
        0x5778: "HANJA",
        0x5779: "HANJA",
        0x577A: "HANJA",
        0x577B: "HANJA",
        0x577C: "HANJA",
        0x577D: "HANJA",
        0x577E: "HANJA",
        0x577F: "HANJA",
        0x5780: "HANJA",
        0x5781: "HANJA",
        0x5782: "HANJA",
        0x5783: "HANJA",
        0x5784: "HANJA",
        0x5785: "HANJA",
        0x5786: "HANJA",
        0x5787: "HANJA",
        0x5788: "HANJA",
        0x5789: "HANJA",
        0x578A: "HANJA",
        0x578B: "HANJA",
        0x578C: "HANJA",
        0x578D: "HANJA",
        0x578E: "HANJA",
        0x578F: "HANJA",
        0x5790: "HANJA",
        0x5791: "HANJA",
        0x5792: "HANJA",
        0x5793: "HANJA",
        0x5794: "HANJA",
        0x5795: "HANJA",
        0x5796: "HANJA",
        0x5797: "HANJA",
        0x5798: "HANJA",
        0x5799: "HANJA",
        0x579A: "HANJA",
        0x579B: "HANJA",
        0x579C: "HANJA",
        0x579D: "HANJA",
        0x579E: "HANJA",
        0x579F: "HANJA",
        0x57A0: "HANJA",
        0x57A1: "HANJA",
        0x57A2: "HANJA",
        0x57A3: "HANJA",
        0x57A4: "HANJA",
        0x57A5: "HANJA",
        0x57A6: "HANJA",
        0x57A7: "HANJA",
        0x57A8: "HANJA",
        0x57A9: "HANJA",
        0x57AA: "HANJA",
        0x57AB: "HANJA",
        0x57AC: "HANJA",
        0x57AD: "HANJA",
        0x57AE: "HANJA",
        0x57AF: "HANJA",
        0x57B0: "HANJA",
        0x57B1: "HANJA",
        0x57B2: "HANJA",
        0x57B3: "HANJA",
        0x57B4: "HANJA",
        0x57B5: "HANJA",
        0x57B6: "HANJA",
        0x57B7: "HANJA",
        0x57B8: "HANJA",
        0x57B9: "HANJA",
        0x57BA: "HANJA",
        0x57BB: "HANJA",
        0x57BC: "HANJA",
        0x57BD: "HANJA",
        0x57BE: "HANJA",
        0x57BF: "HANJA",
        0x57C0: "HANJA",
        0x57C1: "HANJA",
        0x57C2: "HANJA",
        0x57C3: "HANJA",
        0x57C4: "HANJA",
        0x57C5: "HANJA",
        0x57C6: "HANJA",
        0x57C7: "HANJA",
        0x57C8: "HANJA",
        0x57C9: "HANJA",
        0x57CA: "HANJA",
        0x57CB: "HANJA",
        0x57CC: "HANJA",
        0x57CD: "HANJA",
        0x57CE: "HANJA",
        0x57CF: "HANJA",
        0x57D0: "HANJA",
        0x57D1: "HANJA",
        0x57D2: "HANJA",
        0x57D3: "HANJA",
        0x57D4: "HANJA",
        0x57D5: "HANJA",
        0x57D6: "HANJA",
        0x57D7: "HANJA",
        0x57D8: "HANJA",
        0x57D9: "HANJA",
        0x57DA: "HANJA",
        0x57DB: "HANJA",
        0x57DC: "HANJA",
        0x57DD: "HANJA",
        0x57DE: "HANJA",
        0x57DF: "HANJA",
        0x57E0: "HANJA",
        0x57E1: "HANJA",
        0x57E2: "HANJA",
        0x57E3: "HANJA",
        0x57E4: "HANJA",
        0x57E5: "HANJA",
        0x57E6: "HANJA",
        0x57E7: "HANJA",
        0x57E8: "HANJA",
        0x57E9: "HANJA",
        0x57EA: "HANJA",
        0x57EB: "HANJA",
        0x57EC: "HANJA",
        0x57ED: "HANJA",
        0x57EE: "HANJA",
        0x57EF: "HANJA",
        0x57F0: "HANJA",
        0x57F1: "HANJA",
        0x57F2: "HANJA",
        0x57F3: "HANJA",
        0x57F4: "HANJA",
        0x57F5: "HANJA",
        0x57F6: "HANJA",
        0x57F7: "HANJA",
        0x57F8: "HANJA",
        0x57F9: "HANJA",
        0x57FA: "HANJA",
        0x57FB: "HANJA",
        0x57FC: "HANJA",
        0x57FD: "HANJA",
        0x57FE: "HANJA",
        0x57FF: "HANJA",
        0x5800: "HANJA",
        0x5801: "HANJA",
        0x5802: "HANJA",
        0x5803: "HANJA",
        0x5804: "HANJA",
        0x5805: "HANJA",
        0x5806: "HANJA",
        0x5807: "HANJA",
        0x5808: "HANJA",
        0x5809: "HANJA",
        0x580A: "HANJA",
        0x580B: "HANJA",
        0x580C: "HANJA",
        0x580D: "HANJA",
        0x580E: "HANJA",
        0x580F: "HANJA",
        0x5810: "HANJA",
        0x5811: "HANJA",
        0x5812: "HANJA",
        0x5813: "HANJA",
        0x5814: "HANJA",
        0x5815: "HANJA",
        0x5816: "HANJA",
        0x5817: "HANJA",
        0x5818: "HANJA",
        0x5819: "HANJA",
        0x581A: "HANJA",
        0x581B: "HANJA",
        0x581C: "HANJA",
        0x581D: "HANJA",
        0x581E: "HANJA",
        0x581F: "HANJA",
        0x5820: "HANJA",
        0x5821: "HANJA",
        0x5822: "HANJA",
        0x5823: "HANJA",
        0x5824: "HANJA",
        0x5825: "HANJA",
        0x5826: "HANJA",
        0x5827: "HANJA",
        0x5828: "HANJA",
        0x5829: "HANJA",
        0x582A: "HANJA",
        0x582B: "HANJA",
        0x582C: "HANJA",
        0x582D: "HANJA",
        0x582E: "HANJA",
        0x582F: "HANJA",
        0x5830: "HANJA",
        0x5831: "HANJA",
        0x5832: "HANJA",
        0x5833: "HANJA",
        0x5834: "HANJA",
        0x5835: "HANJA",
        0x5836: "HANJA",
        0x5837: "HANJA",
        0x5838: "HANJA",
        0x5839: "HANJA",
        0x583A: "HANJA",
        0x583B: "HANJA",
        0x583C: "HANJA",
        0x583D: "HANJA",
        0x583E: "HANJA",
        0x583F: "HANJA",
        0x5840: "HANJA",
        0x5841: "HANJA",
        0x5842: "HANJA",
        0x5843: "HANJA",
        0x5844: "HANJA",
        0x5845: "HANJA",
        0x5846: "HANJA",
        0x5847: "HANJA",
        0x5848: "HANJA",
        0x5849: "HANJA",
        0x584A: "HANJA",
        0x584B: "HANJA",
        0x584C: "HANJA",
        0x584D: "HANJA",
        0x584E: "HANJA",
        0x584F: "HANJA",
        0x5850: "HANJA",
        0x5851: "HANJA",
        0x5852: "HANJA",
        0x5853: "HANJA",
        0x5854: "HANJA",
        0x5855: "HANJA",
        0x5856: "HANJA",
        0x5857: "HANJA",
        0x5858: "HANJA",
        0x5859: "HANJA",
        0x585A: "HANJA",
        0x585B: "HANJA",
        0x585C: "HANJA",
        0x585D: "HANJA",
        0x585E: "HANJA",
        0x585F: "HANJA",
        0x5860: "HANJA",
        0x5861: "HANJA",
        0x5862: "HANJA",
        0x5863: "HANJA",
        0x5864: "HANJA",
        0x5865: "HANJA",
        0x5866: "HANJA",
        0x5867: "HANJA",
        0x5868: "HANJA",
        0x5869: "HANJA",
        0x586A: "HANJA",
        0x586B: "HANJA",
        0x586C: "HANJA",
        0x586D: "HANJA",
        0x586E: "HANJA",
        0x586F: "HANJA",
        0x5870: "HANJA",
        0x5871: "HANJA",
        0x5872: "HANJA",
        0x5873: "HANJA",
        0x5874: "HANJA",
        0x5875: "HANJA",
        0x5876: "HANJA",
        0x5877: "HANJA",
        0x5878: "HANJA",
        0x5879: "HANJA",
        0x587A: "HANJA",
        0x587B: "HANJA",
        0x587C: "HANJA",
        0x587D: "HANJA",
        0x587E: "HANJA",
        0x587F: "HANJA",
        0x5880: "HANJA",
        0x5881: "HANJA",
        0x5882: "HANJA",
        0x5883: "HANJA",
        0x5884: "HANJA",
        0x5885: "HANJA",
        0x5886: "HANJA",
        0x5887: "HANJA",
        0x5888: "HANJA",
        0x5889: "HANJA",
        0x588A: "HANJA",
        0x588B: "HANJA",
        0x588C: "HANJA",
        0x588D: "HANJA",
        0x588E: "HANJA",
        0x588F: "HANJA",
        0x5890: "HANJA",
        0x5891: "HANJA",
        0x5892: "HANJA",
        0x5893: "HANJA",
        0x5894: "HANJA",
        0x5895: "HANJA",
        0x5896: "HANJA",
        0x5897: "HANJA",
        0x5898: "HANJA",
        0x5899: "HANJA",
        0x589A: "HANJA",
        0x589B: "HANJA",
        0x589C: "HANJA",
        0x589D: "HANJA",
        0x589E: "HANJA",
        0x589F: "HANJA",
        0x58A0: "HANJA",
        0x58A1: "HANJA",
        0x58A2: "HANJA",
        0x58A3: "HANJA",
        0x58A4: "HANJA",
        0x58A5: "HANJA",
        0x58A6: "HANJA",
        0x58A7: "HANJA",
        0x58A8: "HANJA",
        0x58A9: "HANJA",
        0x58AA: "HANJA",
        0x58AB: "HANJA",
        0x58AC: "HANJA",
        0x58AD: "HANJA",
        0x58AE: "HANJA",
        0x58AF: "HANJA",
        0x58B0: "HANJA",
        0x58B1: "HANJA",
        0x58B2: "HANJA",
        0x58B3: "HANJA",
        0x58B4: "HANJA",
        0x58B5: "HANJA",
        0x58B6: "HANJA",
        0x58B7: "HANJA",
        0x58B8: "HANJA",
        0x58B9: "HANJA",
        0x58BA: "HANJA",
        0x58BB: "HANJA",
        0x58BC: "HANJA",
        0x58BD: "HANJA",
        0x58BE: "HANJA",
        0x58BF: "HANJA",
        0x58C0: "HANJA",
        0x58C1: "HANJA",
        0x58C2: "HANJA",
        0x58C3: "HANJA",
        0x58C4: "HANJA",
        0x58C5: "HANJA",
        0x58C6: "HANJA",
        0x58C7: "HANJA",
        0x58C8: "HANJA",
        0x58C9: "HANJA",
        0x58CA: "HANJA",
        0x58CB: "HANJA",
        0x58CC: "HANJA",
        0x58CD: "HANJA",
        0x58CE: "HANJA",
        0x58CF: "HANJA",
        0x58D0: "HANJA",
        0x58D1: "HANJA",
        0x58D2: "HANJA",
        0x58D3: "HANJA",
        0x58D4: "HANJA",
        0x58D5: "HANJA",
        0x58D6: "HANJA",
        0x58D7: "HANJA",
        0x58D8: "HANJA",
        0x58D9: "HANJA",
        0x58DA: "HANJA",
        0x58DB: "HANJA",
        0x58DC: "HANJA",
        0x58DD: "HANJA",
        0x58DE: "HANJA",
        0x58DF: "HANJA",
        0x58E0: "HANJA",
        0x58E1: "HANJA",
        0x58E2: "HANJA",
        0x58E3: "HANJA",
        0x58E4: "HANJA",
        0x58E5: "HANJA",
        0x58E6: "HANJA",
        0x58E7: "HANJA",
        0x58E8: "HANJA",
        0x58E9: "HANJA",
        0x58EA: "HANJA",
        0x58EB: "HANJA",
        0x58EC: "HANJA",
        0x58ED: "HANJA",
        0x58EE: "HANJA",
        0x58EF: "HANJA",
        0x58F0: "HANJA",
        0x58F1: "HANJA",
        0x58F2: "HANJA",
        0x58F3: "HANJA",
        0x58F4: "HANJA",
        0x58F5: "HANJA",
        0x58F6: "HANJA",
        0x58F7: "HANJA",
        0x58F8: "HANJA",
        0x58F9: "HANJA",
        0x58FA: "HANJA",
        0x58FB: "HANJA",
        0x58FC: "HANJA",
        0x58FD: "HANJA",
        0x58FE: "HANJA",
        0x58FF: "HANJA",
        0x5900: "HANJA",
        0x5901: "HANJA",
        0x5902: "HANJA",
        0x5903: "HANJA",
        0x5904: "HANJA",
        0x5905: "HANJA",
        0x5906: "HANJA",
        0x5907: "HANJA",
        0x5908: "HANJA",
        0x5909: "HANJA",
        0x590A: "HANJA",
        0x590B: "HANJA",
        0x590C: "HANJA",
        0x590D: "HANJA",
        0x590E: "HANJA",
        0x590F: "HANJA",
        0x5910: "HANJA",
        0x5911: "HANJA",
        0x5912: "HANJA",
        0x5913: "HANJA",
        0x5914: "HANJA",
        0x5915: "HANJA",
        0x5916: "HANJA",
        0x5917: "HANJA",
        0x5918: "HANJA",
        0x5919: "HANJA",
        0x591A: "HANJA",
        0x591B: "HANJA",
        0x591C: "HANJA",
        0x591D: "HANJA",
        0x591E: "HANJA",
        0x591F: "HANJA",
        0x5920: "HANJA",
        0x5921: "HANJA",
        0x5922: "HANJA",
        0x5923: "HANJA",
        0x5924: "HANJA",
        0x5925: "HANJA",
        0x5926: "HANJA",
        0x5927: "HANJA",
        0x5928: "HANJA",
        0x5929: "HANJA",
        0x592A: "HANJA",
        0x592B: "HANJA",
        0x592C: "HANJA",
        0x592D: "HANJA",
        0x592E: "HANJA",
        0x592F: "HANJA",
        0x5930: "HANJA",
        0x5931: "HANJA",
        0x5932: "HANJA",
        0x5933: "HANJA",
        0x5934: "HANJA",
        0x5935: "HANJA",
        0x5936: "HANJA",
        0x5937: "HANJA",
        0x5938: "HANJA",
        0x5939: "HANJA",
        0x593A: "HANJA",
        0x593B: "HANJA",
        0x593C: "HANJA",
        0x593D: "HANJA",
        0x593E: "HANJA",
        0x593F: "HANJA",
        0x5940: "HANJA",
        0x5941: "HANJA",
        0x5942: "HANJA",
        0x5943: "HANJA",
        0x5944: "HANJA",
        0x5945: "HANJA",
        0x5946: "HANJA",
        0x5947: "HANJA",
        0x5948: "HANJA",
        0x5949: "HANJA",
        0x594A: "HANJA",
        0x594B: "HANJA",
        0x594C: "HANJA",
        0x594D: "HANJA",
        0x594E: "HANJA",
        0x594F: "HANJA",
        0x5950: "HANJA",
        0x5951: "HANJA",
        0x5952: "HANJA",
        0x5953: "HANJA",
        0x5954: "HANJA",
        0x5955: "HANJA",
        0x5956: "HANJA",
        0x5957: "HANJA",
        0x5958: "HANJA",
        0x5959: "HANJA",
        0x595A: "HANJA",
        0x595B: "HANJA",
        0x595C: "HANJA",
        0x595D: "HANJA",
        0x595E: "HANJA",
        0x595F: "HANJA",
        0x5960: "HANJA",
        0x5961: "HANJA",
        0x5962: "HANJA",
        0x5963: "HANJA",
        0x5964: "HANJA",
        0x5965: "HANJA",
        0x5966: "HANJA",
        0x5967: "HANJA",
        0x5968: "HANJA",
        0x5969: "HANJA",
        0x596A: "HANJA",
        0x596B: "HANJA",
        0x596C: "HANJA",
        0x596D: "HANJA",
        0x596E: "HANJA",
        0x596F: "HANJA",
        0x5970: "HANJA",
        0x5971: "HANJA",
        0x5972: "HANJA",
        0x5973: "HANJA",
        0x5974: "HANJA",
        0x5975: "HANJA",
        0x5976: "HANJA",
        0x5977: "HANJA",
        0x5978: "HANJA",
        0x5979: "HANJA",
        0x597A: "HANJA",
        0x597B: "HANJA",
        0x597C: "HANJA",
        0x597D: "HANJA",
        0x597E: "HANJA",
        0x597F: "HANJA",
        0x5980: "HANJA",
        0x5981: "HANJA",
        0x5982: "HANJA",
        0x5983: "HANJA",
        0x5984: "HANJA",
        0x5985: "HANJA",
        0x5986: "HANJA",
        0x5987: "HANJA",
        0x5988: "HANJA",
        0x5989: "HANJA",
        0x598A: "HANJA",
        0x598B: "HANJA",
        0x598C: "HANJA",
        0x598D: "HANJA",
        0x598E: "HANJA",
        0x598F: "HANJA",
        0x5990: "HANJA",
        0x5991: "HANJA",
        0x5992: "HANJA",
        0x5993: "HANJA",
        0x5994: "HANJA",
        0x5995: "HANJA",
        0x5996: "HANJA",
        0x5997: "HANJA",
        0x5998: "HANJA",
        0x5999: "HANJA",
        0x599A: "HANJA",
        0x599B: "HANJA",
        0x599C: "HANJA",
        0x599D: "HANJA",
        0x599E: "HANJA",
        0x599F: "HANJA",
        0x59A0: "HANJA",
        0x59A1: "HANJA",
        0x59A2: "HANJA",
        0x59A3: "HANJA",
        0x59A4: "HANJA",
        0x59A5: "HANJA",
        0x59A6: "HANJA",
        0x59A7: "HANJA",
        0x59A8: "HANJA",
        0x59A9: "HANJA",
        0x59AA: "HANJA",
        0x59AB: "HANJA",
        0x59AC: "HANJA",
        0x59AD: "HANJA",
        0x59AE: "HANJA",
        0x59AF: "HANJA",
        0x59B0: "HANJA",
        0x59B1: "HANJA",
        0x59B2: "HANJA",
        0x59B3: "HANJA",
        0x59B4: "HANJA",
        0x59B5: "HANJA",
        0x59B6: "HANJA",
        0x59B7: "HANJA",
        0x59B8: "HANJA",
        0x59B9: "HANJA",
        0x59BA: "HANJA",
        0x59BB: "HANJA",
        0x59BC: "HANJA",
        0x59BD: "HANJA",
        0x59BE: "HANJA",
        0x59BF: "HANJA",
        0x59C0: "HANJA",
        0x59C1: "HANJA",
        0x59C2: "HANJA",
        0x59C3: "HANJA",
        0x59C4: "HANJA",
        0x59C5: "HANJA",
        0x59C6: "HANJA",
        0x59C7: "HANJA",
        0x59C8: "HANJA",
        0x59C9: "HANJA",
        0x59CA: "HANJA",
        0x59CB: "HANJA",
        0x59CC: "HANJA",
        0x59CD: "HANJA",
        0x59CE: "HANJA",
        0x59CF: "HANJA",
        0x59D0: "HANJA",
        0x59D1: "HANJA",
        0x59D2: "HANJA",
        0x59D3: "HANJA",
        0x59D4: "HANJA",
        0x59D5: "HANJA",
        0x59D6: "HANJA",
        0x59D7: "HANJA",
        0x59D8: "HANJA",
        0x59D9: "HANJA",
        0x59DA: "HANJA",
        0x59DB: "HANJA",
        0x59DC: "HANJA",
        0x59DD: "HANJA",
        0x59DE: "HANJA",
        0x59DF: "HANJA",
        0x59E0: "HANJA",
        0x59E1: "HANJA",
        0x59E2: "HANJA",
        0x59E3: "HANJA",
        0x59E4: "HANJA",
        0x59E5: "HANJA",
        0x59E6: "HANJA",
        0x59E7: "HANJA",
        0x59E8: "HANJA",
        0x59E9: "HANJA",
        0x59EA: "HANJA",
        0x59EB: "HANJA",
        0x59EC: "HANJA",
        0x59ED: "HANJA",
        0x59EE: "HANJA",
        0x59EF: "HANJA",
        0x59F0: "HANJA",
        0x59F1: "HANJA",
        0x59F2: "HANJA",
        0x59F3: "HANJA",
        0x59F4: "HANJA",
        0x59F5: "HANJA",
        0x59F6: "HANJA",
        0x59F7: "HANJA",
        0x59F8: "HANJA",
        0x59F9: "HANJA",
        0x59FA: "HANJA",
        0x59FB: "HANJA",
        0x59FC: "HANJA",
        0x59FD: "HANJA",
        0x59FE: "HANJA",
        0x59FF: "HANJA",
        0x5A00: "HANJA",
        0x5A01: "HANJA",
        0x5A02: "HANJA",
        0x5A03: "HANJA",
        0x5A04: "HANJA",
        0x5A05: "HANJA",
        0x5A06: "HANJA",
        0x5A07: "HANJA",
        0x5A08: "HANJA",
        0x5A09: "HANJA",
        0x5A0A: "HANJA",
        0x5A0B: "HANJA",
        0x5A0C: "HANJA",
        0x5A0D: "HANJA",
        0x5A0E: "HANJA",
        0x5A0F: "HANJA",
        0x5A10: "HANJA",
        0x5A11: "HANJA",
        0x5A12: "HANJA",
        0x5A13: "HANJA",
        0x5A14: "HANJA",
        0x5A15: "HANJA",
        0x5A16: "HANJA",
        0x5A17: "HANJA",
        0x5A18: "HANJA",
        0x5A19: "HANJA",
        0x5A1A: "HANJA",
        0x5A1B: "HANJA",
        0x5A1C: "HANJA",
        0x5A1D: "HANJA",
        0x5A1E: "HANJA",
        0x5A1F: "HANJA",
        0x5A20: "HANJA",
        0x5A21: "HANJA",
        0x5A22: "HANJA",
        0x5A23: "HANJA",
        0x5A24: "HANJA",
        0x5A25: "HANJA",
        0x5A26: "HANJA",
        0x5A27: "HANJA",
        0x5A28: "HANJA",
        0x5A29: "HANJA",
        0x5A2A: "HANJA",
        0x5A2B: "HANJA",
        0x5A2C: "HANJA",
        0x5A2D: "HANJA",
        0x5A2E: "HANJA",
        0x5A2F: "HANJA",
        0x5A30: "HANJA",
        0x5A31: "HANJA",
        0x5A32: "HANJA",
        0x5A33: "HANJA",
        0x5A34: "HANJA",
        0x5A35: "HANJA",
        0x5A36: "HANJA",
        0x5A37: "HANJA",
        0x5A38: "HANJA",
        0x5A39: "HANJA",
        0x5A3A: "HANJA",
        0x5A3B: "HANJA",
        0x5A3C: "HANJA",
        0x5A3D: "HANJA",
        0x5A3E: "HANJA",
        0x5A3F: "HANJA",
        0x5A40: "HANJA",
        0x5A41: "HANJA",
        0x5A42: "HANJA",
        0x5A43: "HANJA",
        0x5A44: "HANJA",
        0x5A45: "HANJA",
        0x5A46: "HANJA",
        0x5A47: "HANJA",
        0x5A48: "HANJA",
        0x5A49: "HANJA",
        0x5A4A: "HANJA",
        0x5A4B: "HANJA",
        0x5A4C: "HANJA",
        0x5A4D: "HANJA",
        0x5A4E: "HANJA",
        0x5A4F: "HANJA",
        0x5A50: "HANJA",
        0x5A51: "HANJA",
        0x5A52: "HANJA",
        0x5A53: "HANJA",
        0x5A54: "HANJA",
        0x5A55: "HANJA",
        0x5A56: "HANJA",
        0x5A57: "HANJA",
        0x5A58: "HANJA",
        0x5A59: "HANJA",
        0x5A5A: "HANJA",
        0x5A5B: "HANJA",
        0x5A5C: "HANJA",
        0x5A5D: "HANJA",
        0x5A5E: "HANJA",
        0x5A5F: "HANJA",
        0x5A60: "HANJA",
        0x5A61: "HANJA",
        0x5A62: "HANJA",
        0x5A63: "HANJA",
        0x5A64: "HANJA",
        0x5A65: "HANJA",
        0x5A66: "HANJA",
        0x5A67: "HANJA",
        0x5A68: "HANJA",
        0x5A69: "HANJA",
        0x5A6A: "HANJA",
        0x5A6B: "HANJA",
        0x5A6C: "HANJA",
        0x5A6D: "HANJA",
        0x5A6E: "HANJA",
        0x5A6F: "HANJA",
        0x5A70: "HANJA",
        0x5A71: "HANJA",
        0x5A72: "HANJA",
        0x5A73: "HANJA",
        0x5A74: "HANJA",
        0x5A75: "HANJA",
        0x5A76: "HANJA",
        0x5A77: "HANJA",
        0x5A78: "HANJA",
        0x5A79: "HANJA",
        0x5A7A: "HANJA",
        0x5A7B: "HANJA",
        0x5A7C: "HANJA",
        0x5A7D: "HANJA",
        0x5A7E: "HANJA",
        0x5A7F: "HANJA",
        0x5A80: "HANJA",
        0x5A81: "HANJA",
        0x5A82: "HANJA",
        0x5A83: "HANJA",
        0x5A84: "HANJA",
        0x5A85: "HANJA",
        0x5A86: "HANJA",
        0x5A87: "HANJA",
        0x5A88: "HANJA",
        0x5A89: "HANJA",
        0x5A8A: "HANJA",
        0x5A8B: "HANJA",
        0x5A8C: "HANJA",
        0x5A8D: "HANJA",
        0x5A8E: "HANJA",
        0x5A8F: "HANJA",
        0x5A90: "HANJA",
        0x5A91: "HANJA",
        0x5A92: "HANJA",
        0x5A93: "HANJA",
        0x5A94: "HANJA",
        0x5A95: "HANJA",
        0x5A96: "HANJA",
        0x5A97: "HANJA",
        0x5A98: "HANJA",
        0x5A99: "HANJA",
        0x5A9A: "HANJA",
        0x5A9B: "HANJA",
        0x5A9C: "HANJA",
        0x5A9D: "HANJA",
        0x5A9E: "HANJA",
        0x5A9F: "HANJA",
        0x5AA0: "HANJA",
        0x5AA1: "HANJA",
        0x5AA2: "HANJA",
        0x5AA3: "HANJA",
        0x5AA4: "HANJA",
        0x5AA5: "HANJA",
        0x5AA6: "HANJA",
        0x5AA7: "HANJA",
        0x5AA8: "HANJA",
        0x5AA9: "HANJA",
        0x5AAA: "HANJA",
        0x5AAB: "HANJA",
        0x5AAC: "HANJA",
        0x5AAD: "HANJA",
        0x5AAE: "HANJA",
        0x5AAF: "HANJA",
        0x5AB0: "HANJA",
        0x5AB1: "HANJA",
        0x5AB2: "HANJA",
        0x5AB3: "HANJA",
        0x5AB4: "HANJA",
        0x5AB5: "HANJA",
        0x5AB6: "HANJA",
        0x5AB7: "HANJA",
        0x5AB8: "HANJA",
        0x5AB9: "HANJA",
        0x5ABA: "HANJA",
        0x5ABB: "HANJA",
        0x5ABC: "HANJA",
        0x5ABD: "HANJA",
        0x5ABE: "HANJA",
        0x5ABF: "HANJA",
        0x5AC0: "HANJA",
        0x5AC1: "HANJA",
        0x5AC2: "HANJA",
        0x5AC3: "HANJA",
        0x5AC4: "HANJA",
        0x5AC5: "HANJA",
        0x5AC6: "HANJA",
        0x5AC7: "HANJA",
        0x5AC8: "HANJA",
        0x5AC9: "HANJA",
        0x5ACA: "HANJA",
        0x5ACB: "HANJA",
        0x5ACC: "HANJA",
        0x5ACD: "HANJA",
        0x5ACE: "HANJA",
        0x5ACF: "HANJA",
        0x5AD0: "HANJA",
        0x5AD1: "HANJA",
        0x5AD2: "HANJA",
        0x5AD3: "HANJA",
        0x5AD4: "HANJA",
        0x5AD5: "HANJA",
        0x5AD6: "HANJA",
        0x5AD7: "HANJA",
        0x5AD8: "HANJA",
        0x5AD9: "HANJA",
        0x5ADA: "HANJA",
        0x5ADB: "HANJA",
        0x5ADC: "HANJA",
        0x5ADD: "HANJA",
        0x5ADE: "HANJA",
        0x5ADF: "HANJA",
        0x5AE0: "HANJA",
        0x5AE1: "HANJA",
        0x5AE2: "HANJA",
        0x5AE3: "HANJA",
        0x5AE4: "HANJA",
        0x5AE5: "HANJA",
        0x5AE6: "HANJA",
        0x5AE7: "HANJA",
        0x5AE8: "HANJA",
        0x5AE9: "HANJA",
        0x5AEA: "HANJA",
        0x5AEB: "HANJA",
        0x5AEC: "HANJA",
        0x5AED: "HANJA",
        0x5AEE: "HANJA",
        0x5AEF: "HANJA",
        0x5AF0: "HANJA",
        0x5AF1: "HANJA",
        0x5AF2: "HANJA",
        0x5AF3: "HANJA",
        0x5AF4: "HANJA",
        0x5AF5: "HANJA",
        0x5AF6: "HANJA",
        0x5AF7: "HANJA",
        0x5AF8: "HANJA",
        0x5AF9: "HANJA",
        0x5AFA: "HANJA",
        0x5AFB: "HANJA",
        0x5AFC: "HANJA",
        0x5AFD: "HANJA",
        0x5AFE: "HANJA",
        0x5AFF: "HANJA",
        0x5B00: "HANJA",
        0x5B01: "HANJA",
        0x5B02: "HANJA",
        0x5B03: "HANJA",
        0x5B04: "HANJA",
        0x5B05: "HANJA",
        0x5B06: "HANJA",
        0x5B07: "HANJA",
        0x5B08: "HANJA",
        0x5B09: "HANJA",
        0x5B0A: "HANJA",
        0x5B0B: "HANJA",
        0x5B0C: "HANJA",
        0x5B0D: "HANJA",
        0x5B0E: "HANJA",
        0x5B0F: "HANJA",
        0x5B10: "HANJA",
        0x5B11: "HANJA",
        0x5B12: "HANJA",
        0x5B13: "HANJA",
        0x5B14: "HANJA",
        0x5B15: "HANJA",
        0x5B16: "HANJA",
        0x5B17: "HANJA",
        0x5B18: "HANJA",
        0x5B19: "HANJA",
        0x5B1A: "HANJA",
        0x5B1B: "HANJA",
        0x5B1C: "HANJA",
        0x5B1D: "HANJA",
        0x5B1E: "HANJA",
        0x5B1F: "HANJA",
        0x5B20: "HANJA",
        0x5B21: "HANJA",
        0x5B22: "HANJA",
        0x5B23: "HANJA",
        0x5B24: "HANJA",
        0x5B25: "HANJA",
        0x5B26: "HANJA",
        0x5B27: "HANJA",
        0x5B28: "HANJA",
        0x5B29: "HANJA",
        0x5B2A: "HANJA",
        0x5B2B: "HANJA",
        0x5B2C: "HANJA",
        0x5B2D: "HANJA",
        0x5B2E: "HANJA",
        0x5B2F: "HANJA",
        0x5B30: "HANJA",
        0x5B31: "HANJA",
        0x5B32: "HANJA",
        0x5B33: "HANJA",
        0x5B34: "HANJA",
        0x5B35: "HANJA",
        0x5B36: "HANJA",
        0x5B37: "HANJA",
        0x5B38: "HANJA",
        0x5B39: "HANJA",
        0x5B3A: "HANJA",
        0x5B3B: "HANJA",
        0x5B3C: "HANJA",
        0x5B3D: "HANJA",
        0x5B3E: "HANJA",
        0x5B3F: "HANJA",
        0x5B40: "HANJA",
        0x5B41: "HANJA",
        0x5B42: "HANJA",
        0x5B43: "HANJA",
        0x5B44: "HANJA",
        0x5B45: "HANJA",
        0x5B46: "HANJA",
        0x5B47: "HANJA",
        0x5B48: "HANJA",
        0x5B49: "HANJA",
        0x5B4A: "HANJA",
        0x5B4B: "HANJA",
        0x5B4C: "HANJA",
        0x5B4D: "HANJA",
        0x5B4E: "HANJA",
        0x5B4F: "HANJA",
        0x5B50: "HANJA",
        0x5B51: "HANJA",
        0x5B52: "HANJA",
        0x5B53: "HANJA",
        0x5B54: "HANJA",
        0x5B55: "HANJA",
        0x5B56: "HANJA",
        0x5B57: "HANJA",
        0x5B58: "HANJA",
        0x5B59: "HANJA",
        0x5B5A: "HANJA",
        0x5B5B: "HANJA",
        0x5B5C: "HANJA",
        0x5B5D: "HANJA",
        0x5B5E: "HANJA",
        0x5B5F: "HANJA",
        0x5B60: "HANJA",
        0x5B61: "HANJA",
        0x5B62: "HANJA",
        0x5B63: "HANJA",
        0x5B64: "HANJA",
        0x5B65: "HANJA",
        0x5B66: "HANJA",
        0x5B67: "HANJA",
        0x5B68: "HANJA",
        0x5B69: "HANJA",
        0x5B6A: "HANJA",
        0x5B6B: "HANJA",
        0x5B6C: "HANJA",
        0x5B6D: "HANJA",
        0x5B6E: "HANJA",
        0x5B6F: "HANJA",
        0x5B70: "HANJA",
        0x5B71: "HANJA",
        0x5B72: "HANJA",
        0x5B73: "HANJA",
        0x5B74: "HANJA",
        0x5B75: "HANJA",
        0x5B76: "HANJA",
        0x5B77: "HANJA",
        0x5B78: "HANJA",
        0x5B79: "HANJA",
        0x5B7A: "HANJA",
        0x5B7B: "HANJA",
        0x5B7C: "HANJA",
        0x5B7D: "HANJA",
        0x5B7E: "HANJA",
        0x5B7F: "HANJA",
        0x5B80: "HANJA",
        0x5B81: "HANJA",
        0x5B82: "HANJA",
        0x5B83: "HANJA",
        0x5B84: "HANJA",
        0x5B85: "HANJA",
        0x5B86: "HANJA",
        0x5B87: "HANJA",
        0x5B88: "HANJA",
        0x5B89: "HANJA",
        0x5B8A: "HANJA",
        0x5B8B: "HANJA",
        0x5B8C: "HANJA",
        0x5B8D: "HANJA",
        0x5B8E: "HANJA",
        0x5B8F: "HANJA",
        0x5B90: "HANJA",
        0x5B91: "HANJA",
        0x5B92: "HANJA",
        0x5B93: "HANJA",
        0x5B94: "HANJA",
        0x5B95: "HANJA",
        0x5B96: "HANJA",
        0x5B97: "HANJA",
        0x5B98: "HANJA",
        0x5B99: "HANJA",
        0x5B9A: "HANJA",
        0x5B9B: "HANJA",
        0x5B9C: "HANJA",
        0x5B9D: "HANJA",
        0x5B9E: "HANJA",
        0x5B9F: "HANJA",
        0x5BA0: "HANJA",
        0x5BA1: "HANJA",
        0x5BA2: "HANJA",
        0x5BA3: "HANJA",
        0x5BA4: "HANJA",
        0x5BA5: "HANJA",
        0x5BA6: "HANJA",
        0x5BA7: "HANJA",
        0x5BA8: "HANJA",
        0x5BA9: "HANJA",
        0x5BAA: "HANJA",
        0x5BAB: "HANJA",
        0x5BAC: "HANJA",
        0x5BAD: "HANJA",
        0x5BAE: "HANJA",
        0x5BAF: "HANJA",
        0x5BB0: "HANJA",
        0x5BB1: "HANJA",
        0x5BB2: "HANJA",
        0x5BB3: "HANJA",
        0x5BB4: "HANJA",
        0x5BB5: "HANJA",
        0x5BB6: "HANJA",
        0x5BB7: "HANJA",
        0x5BB8: "HANJA",
        0x5BB9: "HANJA",
        0x5BBA: "HANJA",
        0x5BBB: "HANJA",
        0x5BBC: "HANJA",
        0x5BBD: "HANJA",
        0x5BBE: "HANJA",
        0x5BBF: "HANJA",
        0x5BC0: "HANJA",
        0x5BC1: "HANJA",
        0x5BC2: "HANJA",
        0x5BC3: "HANJA",
        0x5BC4: "HANJA",
        0x5BC5: "HANJA",
        0x5BC6: "HANJA",
        0x5BC7: "HANJA",
        0x5BC8: "HANJA",
        0x5BC9: "HANJA",
        0x5BCA: "HANJA",
        0x5BCB: "HANJA",
        0x5BCC: "HANJA",
        0x5BCD: "HANJA",
        0x5BCE: "HANJA",
        0x5BCF: "HANJA",
        0x5BD0: "HANJA",
        0x5BD1: "HANJA",
        0x5BD2: "HANJA",
        0x5BD3: "HANJA",
        0x5BD4: "HANJA",
        0x5BD5: "HANJA",
        0x5BD6: "HANJA",
        0x5BD7: "HANJA",
        0x5BD8: "HANJA",
        0x5BD9: "HANJA",
        0x5BDA: "HANJA",
        0x5BDB: "HANJA",
        0x5BDC: "HANJA",
        0x5BDD: "HANJA",
        0x5BDE: "HANJA",
        0x5BDF: "HANJA",
        0x5BE0: "HANJA",
        0x5BE1: "HANJA",
        0x5BE2: "HANJA",
        0x5BE3: "HANJA",
        0x5BE4: "HANJA",
        0x5BE5: "HANJA",
        0x5BE6: "HANJA",
        0x5BE7: "HANJA",
        0x5BE8: "HANJA",
        0x5BE9: "HANJA",
        0x5BEA: "HANJA",
        0x5BEB: "HANJA",
        0x5BEC: "HANJA",
        0x5BED: "HANJA",
        0x5BEE: "HANJA",
        0x5BEF: "HANJA",
        0x5BF0: "HANJA",
        0x5BF1: "HANJA",
        0x5BF2: "HANJA",
        0x5BF3: "HANJA",
        0x5BF4: "HANJA",
        0x5BF5: "HANJA",
        0x5BF6: "HANJA",
        0x5BF7: "HANJA",
        0x5BF8: "HANJA",
        0x5BF9: "HANJA",
        0x5BFA: "HANJA",
        0x5BFB: "HANJA",
        0x5BFC: "HANJA",
        0x5BFD: "HANJA",
        0x5BFE: "HANJA",
        0x5BFF: "HANJA",
        0x5C00: "HANJA",
        0x5C01: "HANJA",
        0x5C02: "HANJA",
        0x5C03: "HANJA",
        0x5C04: "HANJA",
        0x5C05: "HANJA",
        0x5C06: "HANJA",
        0x5C07: "HANJA",
        0x5C08: "HANJA",
        0x5C09: "HANJA",
        0x5C0A: "HANJA",
        0x5C0B: "HANJA",
        0x5C0C: "HANJA",
        0x5C0D: "HANJA",
        0x5C0E: "HANJA",
        0x5C0F: "HANJA",
        0x5C10: "HANJA",
        0x5C11: "HANJA",
        0x5C12: "HANJA",
        0x5C13: "HANJA",
        0x5C14: "HANJA",
        0x5C15: "HANJA",
        0x5C16: "HANJA",
        0x5C17: "HANJA",
        0x5C18: "HANJA",
        0x5C19: "HANJA",
        0x5C1A: "HANJA",
        0x5C1B: "HANJA",
        0x5C1C: "HANJA",
        0x5C1D: "HANJA",
        0x5C1E: "HANJA",
        0x5C1F: "HANJA",
        0x5C20: "HANJA",
        0x5C21: "HANJA",
        0x5C22: "HANJA",
        0x5C23: "HANJA",
        0x5C24: "HANJA",
        0x5C25: "HANJA",
        0x5C26: "HANJA",
        0x5C27: "HANJA",
        0x5C28: "HANJA",
        0x5C29: "HANJA",
        0x5C2A: "HANJA",
        0x5C2B: "HANJA",
        0x5C2C: "HANJA",
        0x5C2D: "HANJA",
        0x5C2E: "HANJA",
        0x5C2F: "HANJA",
        0x5C30: "HANJA",
        0x5C31: "HANJA",
        0x5C32: "HANJA",
        0x5C33: "HANJA",
        0x5C34: "HANJA",
        0x5C35: "HANJA",
        0x5C36: "HANJA",
        0x5C37: "HANJA",
        0x5C38: "HANJA",
        0x5C39: "HANJA",
        0x5C3A: "HANJA",
        0x5C3B: "HANJA",
        0x5C3C: "HANJA",
        0x5C3D: "HANJA",
        0x5C3E: "HANJA",
        0x5C3F: "HANJA",
        0x5C40: "HANJA",
        0x5C41: "HANJA",
        0x5C42: "HANJA",
        0x5C43: "HANJA",
        0x5C44: "HANJA",
        0x5C45: "HANJA",
        0x5C46: "HANJA",
        0x5C47: "HANJA",
        0x5C48: "HANJA",
        0x5C49: "HANJA",
        0x5C4A: "HANJA",
        0x5C4B: "HANJA",
        0x5C4C: "HANJA",
        0x5C4D: "HANJA",
        0x5C4E: "HANJA",
        0x5C4F: "HANJA",
        0x5C50: "HANJA",
        0x5C51: "HANJA",
        0x5C52: "HANJA",
        0x5C53: "HANJA",
        0x5C54: "HANJA",
        0x5C55: "HANJA",
        0x5C56: "HANJA",
        0x5C57: "HANJA",
        0x5C58: "HANJA",
        0x5C59: "HANJA",
        0x5C5A: "HANJA",
        0x5C5B: "HANJA",
        0x5C5C: "HANJA",
        0x5C5D: "HANJA",
        0x5C5E: "HANJA",
        0x5C5F: "HANJA",
        0x5C60: "HANJA",
        0x5C61: "HANJA",
        0x5C62: "HANJA",
        0x5C63: "HANJA",
        0x5C64: "HANJA",
        0x5C65: "HANJA",
        0x5C66: "HANJA",
        0x5C67: "HANJA",
        0x5C68: "HANJA",
        0x5C69: "HANJA",
        0x5C6A: "HANJA",
        0x5C6B: "HANJA",
        0x5C6C: "HANJA",
        0x5C6D: "HANJA",
        0x5C6E: "HANJA",
        0x5C6F: "HANJA",
        0x5C70: "HANJA",
        0x5C71: "HANJA",
        0x5C72: "HANJA",
        0x5C73: "HANJA",
        0x5C74: "HANJA",
        0x5C75: "HANJA",
        0x5C76: "HANJA",
        0x5C77: "HANJA",
        0x5C78: "HANJA",
        0x5C79: "HANJA",
        0x5C7A: "HANJA",
        0x5C7B: "HANJA",
        0x5C7C: "HANJA",
        0x5C7D: "HANJA",
        0x5C7E: "HANJA",
        0x5C7F: "HANJA",
        0x5C80: "HANJA",
        0x5C81: "HANJA",
        0x5C82: "HANJA",
        0x5C83: "HANJA",
        0x5C84: "HANJA",
        0x5C85: "HANJA",
        0x5C86: "HANJA",
        0x5C87: "HANJA",
        0x5C88: "HANJA",
        0x5C89: "HANJA",
        0x5C8A: "HANJA",
        0x5C8B: "HANJA",
        0x5C8C: "HANJA",
        0x5C8D: "HANJA",
        0x5C8E: "HANJA",
        0x5C8F: "HANJA",
        0x5C90: "HANJA",
        0x5C91: "HANJA",
        0x5C92: "HANJA",
        0x5C93: "HANJA",
        0x5C94: "HANJA",
        0x5C95: "HANJA",
        0x5C96: "HANJA",
        0x5C97: "HANJA",
        0x5C98: "HANJA",
        0x5C99: "HANJA",
        0x5C9A: "HANJA",
        0x5C9B: "HANJA",
        0x5C9C: "HANJA",
        0x5C9D: "HANJA",
        0x5C9E: "HANJA",
        0x5C9F: "HANJA",
        0x5CA0: "HANJA",
        0x5CA1: "HANJA",
        0x5CA2: "HANJA",
        0x5CA3: "HANJA",
        0x5CA4: "HANJA",
        0x5CA5: "HANJA",
        0x5CA6: "HANJA",
        0x5CA7: "HANJA",
        0x5CA8: "HANJA",
        0x5CA9: "HANJA",
        0x5CAA: "HANJA",
        0x5CAB: "HANJA",
        0x5CAC: "HANJA",
        0x5CAD: "HANJA",
        0x5CAE: "HANJA",
        0x5CAF: "HANJA",
        0x5CB0: "HANJA",
        0x5CB1: "HANJA",
        0x5CB2: "HANJA",
        0x5CB3: "HANJA",
        0x5CB4: "HANJA",
        0x5CB5: "HANJA",
        0x5CB6: "HANJA",
        0x5CB7: "HANJA",
        0x5CB8: "HANJA",
        0x5CB9: "HANJA",
        0x5CBA: "HANJA",
        0x5CBB: "HANJA",
        0x5CBC: "HANJA",
        0x5CBD: "HANJA",
        0x5CBE: "HANJA",
        0x5CBF: "HANJA",
        0x5CC0: "HANJA",
        0x5CC1: "HANJA",
        0x5CC2: "HANJA",
        0x5CC3: "HANJA",
        0x5CC4: "HANJA",
        0x5CC5: "HANJA",
        0x5CC6: "HANJA",
        0x5CC7: "HANJA",
        0x5CC8: "HANJA",
        0x5CC9: "HANJA",
        0x5CCA: "HANJA",
        0x5CCB: "HANJA",
        0x5CCC: "HANJA",
        0x5CCD: "HANJA",
        0x5CCE: "HANJA",
        0x5CCF: "HANJA",
        0x5CD0: "HANJA",
        0x5CD1: "HANJA",
        0x5CD2: "HANJA",
        0x5CD3: "HANJA",
        0x5CD4: "HANJA",
        0x5CD5: "HANJA",
        0x5CD6: "HANJA",
        0x5CD7: "HANJA",
        0x5CD8: "HANJA",
        0x5CD9: "HANJA",
        0x5CDA: "HANJA",
        0x5CDB: "HANJA",
        0x5CDC: "HANJA",
        0x5CDD: "HANJA",
        0x5CDE: "HANJA",
        0x5CDF: "HANJA",
        0x5CE0: "HANJA",
        0x5CE1: "HANJA",
        0x5CE2: "HANJA",
        0x5CE3: "HANJA",
        0x5CE4: "HANJA",
        0x5CE5: "HANJA",
        0x5CE6: "HANJA",
        0x5CE7: "HANJA",
        0x5CE8: "HANJA",
        0x5CE9: "HANJA",
        0x5CEA: "HANJA",
        0x5CEB: "HANJA",
        0x5CEC: "HANJA",
        0x5CED: "HANJA",
        0x5CEE: "HANJA",
        0x5CEF: "HANJA",
        0x5CF0: "HANJA",
        0x5CF1: "HANJA",
        0x5CF2: "HANJA",
        0x5CF3: "HANJA",
        0x5CF4: "HANJA",
        0x5CF5: "HANJA",
        0x5CF6: "HANJA",
        0x5CF7: "HANJA",
        0x5CF8: "HANJA",
        0x5CF9: "HANJA",
        0x5CFA: "HANJA",
        0x5CFB: "HANJA",
        0x5CFC: "HANJA",
        0x5CFD: "HANJA",
        0x5CFE: "HANJA",
        0x5CFF: "HANJA",
        0x5D00: "HANJA",
        0x5D01: "HANJA",
        0x5D02: "HANJA",
        0x5D03: "HANJA",
        0x5D04: "HANJA",
        0x5D05: "HANJA",
        0x5D06: "HANJA",
        0x5D07: "HANJA",
        0x5D08: "HANJA",
        0x5D09: "HANJA",
        0x5D0A: "HANJA",
        0x5D0B: "HANJA",
        0x5D0C: "HANJA",
        0x5D0D: "HANJA",
        0x5D0E: "HANJA",
        0x5D0F: "HANJA",
        0x5D10: "HANJA",
        0x5D11: "HANJA",
        0x5D12: "HANJA",
        0x5D13: "HANJA",
        0x5D14: "HANJA",
        0x5D15: "HANJA",
        0x5D16: "HANJA",
        0x5D17: "HANJA",
        0x5D18: "HANJA",
        0x5D19: "HANJA",
        0x5D1A: "HANJA",
        0x5D1B: "HANJA",
        0x5D1C: "HANJA",
        0x5D1D: "HANJA",
        0x5D1E: "HANJA",
        0x5D1F: "HANJA",
        0x5D20: "HANJA",
        0x5D21: "HANJA",
        0x5D22: "HANJA",
        0x5D23: "HANJA",
        0x5D24: "HANJA",
        0x5D25: "HANJA",
        0x5D26: "HANJA",
        0x5D27: "HANJA",
        0x5D28: "HANJA",
        0x5D29: "HANJA",
        0x5D2A: "HANJA",
        0x5D2B: "HANJA",
        0x5D2C: "HANJA",
        0x5D2D: "HANJA",
        0x5D2E: "HANJA",
        0x5D2F: "HANJA",
        0x5D30: "HANJA",
        0x5D31: "HANJA",
        0x5D32: "HANJA",
        0x5D33: "HANJA",
        0x5D34: "HANJA",
        0x5D35: "HANJA",
        0x5D36: "HANJA",
        0x5D37: "HANJA",
        0x5D38: "HANJA",
        0x5D39: "HANJA",
        0x5D3A: "HANJA",
        0x5D3B: "HANJA",
        0x5D3C: "HANJA",
        0x5D3D: "HANJA",
        0x5D3E: "HANJA",
        0x5D3F: "HANJA",
        0x5D40: "HANJA",
        0x5D41: "HANJA",
        0x5D42: "HANJA",
        0x5D43: "HANJA",
        0x5D44: "HANJA",
        0x5D45: "HANJA",
        0x5D46: "HANJA",
        0x5D47: "HANJA",
        0x5D48: "HANJA",
        0x5D49: "HANJA",
        0x5D4A: "HANJA",
        0x5D4B: "HANJA",
        0x5D4C: "HANJA",
        0x5D4D: "HANJA",
        0x5D4E: "HANJA",
        0x5D4F: "HANJA",
        0x5D50: "HANJA",
        0x5D51: "HANJA",
        0x5D52: "HANJA",
        0x5D53: "HANJA",
        0x5D54: "HANJA",
        0x5D55: "HANJA",
        0x5D56: "HANJA",
        0x5D57: "HANJA",
        0x5D58: "HANJA",
        0x5D59: "HANJA",
        0x5D5A: "HANJA",
        0x5D5B: "HANJA",
        0x5D5C: "HANJA",
        0x5D5D: "HANJA",
        0x5D5E: "HANJA",
        0x5D5F: "HANJA",
        0x5D60: "HANJA",
        0x5D61: "HANJA",
        0x5D62: "HANJA",
        0x5D63: "HANJA",
        0x5D64: "HANJA",
        0x5D65: "HANJA",
        0x5D66: "HANJA",
        0x5D67: "HANJA",
        0x5D68: "HANJA",
        0x5D69: "HANJA",
        0x5D6A: "HANJA",
        0x5D6B: "HANJA",
        0x5D6C: "HANJA",
        0x5D6D: "HANJA",
        0x5D6E: "HANJA",
        0x5D6F: "HANJA",
        0x5D70: "HANJA",
        0x5D71: "HANJA",
        0x5D72: "HANJA",
        0x5D73: "HANJA",
        0x5D74: "HANJA",
        0x5D75: "HANJA",
        0x5D76: "HANJA",
        0x5D77: "HANJA",
        0x5D78: "HANJA",
        0x5D79: "HANJA",
        0x5D7A: "HANJA",
        0x5D7B: "HANJA",
        0x5D7C: "HANJA",
        0x5D7D: "HANJA",
        0x5D7E: "HANJA",
        0x5D7F: "HANJA",
        0x5D80: "HANJA",
        0x5D81: "HANJA",
        0x5D82: "HANJA",
        0x5D83: "HANJA",
        0x5D84: "HANJA",
        0x5D85: "HANJA",
        0x5D86: "HANJA",
        0x5D87: "HANJA",
        0x5D88: "HANJA",
        0x5D89: "HANJA",
        0x5D8A: "HANJA",
        0x5D8B: "HANJA",
        0x5D8C: "HANJA",
        0x5D8D: "HANJA",
        0x5D8E: "HANJA",
        0x5D8F: "HANJA",
        0x5D90: "HANJA",
        0x5D91: "HANJA",
        0x5D92: "HANJA",
        0x5D93: "HANJA",
        0x5D94: "HANJA",
        0x5D95: "HANJA",
        0x5D96: "HANJA",
        0x5D97: "HANJA",
        0x5D98: "HANJA",
        0x5D99: "HANJA",
        0x5D9A: "HANJA",
        0x5D9B: "HANJA",
        0x5D9C: "HANJA",
        0x5D9D: "HANJA",
        0x5D9E: "HANJA",
        0x5D9F: "HANJA",
        0x5DA0: "HANJA",
        0x5DA1: "HANJA",
        0x5DA2: "HANJA",
        0x5DA3: "HANJA",
        0x5DA4: "HANJA",
        0x5DA5: "HANJA",
        0x5DA6: "HANJA",
        0x5DA7: "HANJA",
        0x5DA8: "HANJA",
        0x5DA9: "HANJA",
        0x5DAA: "HANJA",
        0x5DAB: "HANJA",
        0x5DAC: "HANJA",
        0x5DAD: "HANJA",
        0x5DAE: "HANJA",
        0x5DAF: "HANJA",
        0x5DB0: "HANJA",
        0x5DB1: "HANJA",
        0x5DB2: "HANJA",
        0x5DB3: "HANJA",
        0x5DB4: "HANJA",
        0x5DB5: "HANJA",
        0x5DB6: "HANJA",
        0x5DB7: "HANJA",
        0x5DB8: "HANJA",
        0x5DB9: "HANJA",
        0x5DBA: "HANJA",
        0x5DBB: "HANJA",
        0x5DBC: "HANJA",
        0x5DBD: "HANJA",
        0x5DBE: "HANJA",
        0x5DBF: "HANJA",
        0x5DC0: "HANJA",
        0x5DC1: "HANJA",
        0x5DC2: "HANJA",
        0x5DC3: "HANJA",
        0x5DC4: "HANJA",
        0x5DC5: "HANJA",
        0x5DC6: "HANJA",
        0x5DC7: "HANJA",
        0x5DC8: "HANJA",
        0x5DC9: "HANJA",
        0x5DCA: "HANJA",
        0x5DCB: "HANJA",
        0x5DCC: "HANJA",
        0x5DCD: "HANJA",
        0x5DCE: "HANJA",
        0x5DCF: "HANJA",
        0x5DD0: "HANJA",
        0x5DD1: "HANJA",
        0x5DD2: "HANJA",
        0x5DD3: "HANJA",
        0x5DD4: "HANJA",
        0x5DD5: "HANJA",
        0x5DD6: "HANJA",
        0x5DD7: "HANJA",
        0x5DD8: "HANJA",
        0x5DD9: "HANJA",
        0x5DDA: "HANJA",
        0x5DDB: "HANJA",
        0x5DDC: "HANJA",
        0x5DDD: "HANJA",
        0x5DDE: "HANJA",
        0x5DDF: "HANJA",
        0x5DE0: "HANJA",
        0x5DE1: "HANJA",
        0x5DE2: "HANJA",
        0x5DE3: "HANJA",
        0x5DE4: "HANJA",
        0x5DE5: "HANJA",
        0x5DE6: "HANJA",
        0x5DE7: "HANJA",
        0x5DE8: "HANJA",
        0x5DE9: "HANJA",
        0x5DEA: "HANJA",
        0x5DEB: "HANJA",
        0x5DEC: "HANJA",
        0x5DED: "HANJA",
        0x5DEE: "HANJA",
        0x5DEF: "HANJA",
        0x5DF0: "HANJA",
        0x5DF1: "HANJA",
        0x5DF2: "HANJA",
        0x5DF3: "HANJA",
        0x5DF4: "HANJA",
        0x5DF5: "HANJA",
        0x5DF6: "HANJA",
        0x5DF7: "HANJA",
        0x5DF8: "HANJA",
        0x5DF9: "HANJA",
        0x5DFA: "HANJA",
        0x5DFB: "HANJA",
        0x5DFC: "HANJA",
        0x5DFD: "HANJA",
        0x5DFE: "HANJA",
        0x5DFF: "HANJA",
        0x5E00: "HANJA",
        0x5E01: "HANJA",
        0x5E02: "HANJA",
        0x5E03: "HANJA",
        0x5E04: "HANJA",
        0x5E05: "HANJA",
        0x5E06: "HANJA",
        0x5E07: "HANJA",
        0x5E08: "HANJA",
        0x5E09: "HANJA",
        0x5E0A: "HANJA",
        0x5E0B: "HANJA",
        0x5E0C: "HANJA",
        0x5E0D: "HANJA",
        0x5E0E: "HANJA",
        0x5E0F: "HANJA",
        0x5E10: "HANJA",
        0x5E11: "HANJA",
        0x5E12: "HANJA",
        0x5E13: "HANJA",
        0x5E14: "HANJA",
        0x5E15: "HANJA",
        0x5E16: "HANJA",
        0x5E17: "HANJA",
        0x5E18: "HANJA",
        0x5E19: "HANJA",
        0x5E1A: "HANJA",
        0x5E1B: "HANJA",
        0x5E1C: "HANJA",
        0x5E1D: "HANJA",
        0x5E1E: "HANJA",
        0x5E1F: "HANJA",
        0x5E20: "HANJA",
        0x5E21: "HANJA",
        0x5E22: "HANJA",
        0x5E23: "HANJA",
        0x5E24: "HANJA",
        0x5E25: "HANJA",
        0x5E26: "HANJA",
        0x5E27: "HANJA",
        0x5E28: "HANJA",
        0x5E29: "HANJA",
        0x5E2A: "HANJA",
        0x5E2B: "HANJA",
        0x5E2C: "HANJA",
        0x5E2D: "HANJA",
        0x5E2E: "HANJA",
        0x5E2F: "HANJA",
        0x5E30: "HANJA",
        0x5E31: "HANJA",
        0x5E32: "HANJA",
        0x5E33: "HANJA",
        0x5E34: "HANJA",
        0x5E35: "HANJA",
        0x5E36: "HANJA",
        0x5E37: "HANJA",
        0x5E38: "HANJA",
        0x5E39: "HANJA",
        0x5E3A: "HANJA",
        0x5E3B: "HANJA",
        0x5E3C: "HANJA",
        0x5E3D: "HANJA",
        0x5E3E: "HANJA",
        0x5E3F: "HANJA",
        0x5E40: "HANJA",
        0x5E41: "HANJA",
        0x5E42: "HANJA",
        0x5E43: "HANJA",
        0x5E44: "HANJA",
        0x5E45: "HANJA",
        0x5E46: "HANJA",
        0x5E47: "HANJA",
        0x5E48: "HANJA",
        0x5E49: "HANJA",
        0x5E4A: "HANJA",
        0x5E4B: "HANJA",
        0x5E4C: "HANJA",
        0x5E4D: "HANJA",
        0x5E4E: "HANJA",
        0x5E4F: "HANJA",
        0x5E50: "HANJA",
        0x5E51: "HANJA",
        0x5E52: "HANJA",
        0x5E53: "HANJA",
        0x5E54: "HANJA",
        0x5E55: "HANJA",
        0x5E56: "HANJA",
        0x5E57: "HANJA",
        0x5E58: "HANJA",
        0x5E59: "HANJA",
        0x5E5A: "HANJA",
        0x5E5B: "HANJA",
        0x5E5C: "HANJA",
        0x5E5D: "HANJA",
        0x5E5E: "HANJA",
        0x5E5F: "HANJA",
        0x5E60: "HANJA",
        0x5E61: "HANJA",
        0x5E62: "HANJA",
        0x5E63: "HANJA",
        0x5E64: "HANJA",
        0x5E65: "HANJA",
        0x5E66: "HANJA",
        0x5E67: "HANJA",
        0x5E68: "HANJA",
        0x5E69: "HANJA",
        0x5E6A: "HANJA",
        0x5E6B: "HANJA",
        0x5E6C: "HANJA",
        0x5E6D: "HANJA",
        0x5E6E: "HANJA",
        0x5E6F: "HANJA",
        0x5E70: "HANJA",
        0x5E71: "HANJA",
        0x5E72: "HANJA",
        0x5E73: "HANJA",
        0x5E74: "HANJA",
        0x5E75: "HANJA",
        0x5E76: "HANJA",
        0x5E77: "HANJA",
        0x5E78: "HANJA",
        0x5E79: "HANJA",
        0x5E7A: "HANJA",
        0x5E7B: "HANJA",
        0x5E7C: "HANJA",
        0x5E7D: "HANJA",
        0x5E7E: "HANJA",
        0x5E7F: "HANJA",
        0x5E80: "HANJA",
        0x5E81: "HANJA",
        0x5E82: "HANJA",
        0x5E83: "HANJA",
        0x5E84: "HANJA",
        0x5E85: "HANJA",
        0x5E86: "HANJA",
        0x5E87: "HANJA",
        0x5E88: "HANJA",
        0x5E89: "HANJA",
        0x5E8A: "HANJA",
        0x5E8B: "HANJA",
        0x5E8C: "HANJA",
        0x5E8D: "HANJA",
        0x5E8E: "HANJA",
        0x5E8F: "HANJA",
        0x5E90: "HANJA",
        0x5E91: "HANJA",
        0x5E92: "HANJA",
        0x5E93: "HANJA",
        0x5E94: "HANJA",
        0x5E95: "HANJA",
        0x5E96: "HANJA",
        0x5E97: "HANJA",
        0x5E98: "HANJA",
        0x5E99: "HANJA",
        0x5E9A: "HANJA",
        0x5E9B: "HANJA",
        0x5E9C: "HANJA",
        0x5E9D: "HANJA",
        0x5E9E: "HANJA",
        0x5E9F: "HANJA",
        0x5EA0: "HANJA",
        0x5EA1: "HANJA",
        0x5EA2: "HANJA",
        0x5EA3: "HANJA",
        0x5EA4: "HANJA",
        0x5EA5: "HANJA",
        0x5EA6: "HANJA",
        0x5EA7: "HANJA",
        0x5EA8: "HANJA",
        0x5EA9: "HANJA",
        0x5EAA: "HANJA",
        0x5EAB: "HANJA",
        0x5EAC: "HANJA",
        0x5EAD: "HANJA",
        0x5EAE: "HANJA",
        0x5EAF: "HANJA",
        0x5EB0: "HANJA",
        0x5EB1: "HANJA",
        0x5EB2: "HANJA",
        0x5EB3: "HANJA",
        0x5EB4: "HANJA",
        0x5EB5: "HANJA",
        0x5EB6: "HANJA",
        0x5EB7: "HANJA",
        0x5EB8: "HANJA",
        0x5EB9: "HANJA",
        0x5EBA: "HANJA",
        0x5EBB: "HANJA",
        0x5EBC: "HANJA",
        0x5EBD: "HANJA",
        0x5EBE: "HANJA",
        0x5EBF: "HANJA",
        0x5EC0: "HANJA",
        0x5EC1: "HANJA",
        0x5EC2: "HANJA",
        0x5EC3: "HANJA",
        0x5EC4: "HANJA",
        0x5EC5: "HANJA",
        0x5EC6: "HANJA",
        0x5EC7: "HANJA",
        0x5EC8: "HANJA",
        0x5EC9: "HANJA",
        0x5ECA: "HANJA",
        0x5ECB: "HANJA",
        0x5ECC: "HANJA",
        0x5ECD: "HANJA",
        0x5ECE: "HANJA",
        0x5ECF: "HANJA",
        0x5ED0: "HANJA",
        0x5ED1: "HANJA",
        0x5ED2: "HANJA",
        0x5ED3: "HANJA",
        0x5ED4: "HANJA",
        0x5ED5: "HANJA",
        0x5ED6: "HANJA",
        0x5ED7: "HANJA",
        0x5ED8: "HANJA",
        0x5ED9: "HANJA",
        0x5EDA: "HANJA",
        0x5EDB: "HANJA",
        0x5EDC: "HANJA",
        0x5EDD: "HANJA",
        0x5EDE: "HANJA",
        0x5EDF: "HANJA",
        0x5EE0: "HANJA",
        0x5EE1: "HANJA",
        0x5EE2: "HANJA",
        0x5EE3: "HANJA",
        0x5EE4: "HANJA",
        0x5EE5: "HANJA",
        0x5EE6: "HANJA",
        0x5EE7: "HANJA",
        0x5EE8: "HANJA",
        0x5EE9: "HANJA",
        0x5EEA: "HANJA",
        0x5EEB: "HANJA",
        0x5EEC: "HANJA",
        0x5EED: "HANJA",
        0x5EEE: "HANJA",
        0x5EEF: "HANJA",
        0x5EF0: "HANJA",
        0x5EF1: "HANJA",
        0x5EF2: "HANJA",
        0x5EF3: "HANJA",
        0x5EF4: "HANJA",
        0x5EF5: "HANJA",
        0x5EF6: "HANJA",
        0x5EF7: "HANJA",
        0x5EF8: "HANJA",
        0x5EF9: "HANJA",
        0x5EFA: "HANJA",
        0x5EFB: "HANJA",
        0x5EFC: "HANJA",
        0x5EFD: "HANJA",
        0x5EFE: "HANJA",
        0x5EFF: "HANJA",
        0x5F00: "HANJA",
        0x5F01: "HANJA",
        0x5F02: "HANJA",
        0x5F03: "HANJA",
        0x5F04: "HANJA",
        0x5F05: "HANJA",
        0x5F06: "HANJA",
        0x5F07: "HANJA",
        0x5F08: "HANJA",
        0x5F09: "HANJA",
        0x5F0A: "HANJA",
        0x5F0B: "HANJA",
        0x5F0C: "HANJA",
        0x5F0D: "HANJA",
        0x5F0E: "HANJA",
        0x5F0F: "HANJA",
        0x5F10: "HANJA",
        0x5F11: "HANJA",
        0x5F12: "HANJA",
        0x5F13: "HANJA",
        0x5F14: "HANJA",
        0x5F15: "HANJA",
        0x5F16: "HANJA",
        0x5F17: "HANJA",
        0x5F18: "HANJA",
        0x5F19: "HANJA",
        0x5F1A: "HANJA",
        0x5F1B: "HANJA",
        0x5F1C: "HANJA",
        0x5F1D: "HANJA",
        0x5F1E: "HANJA",
        0x5F1F: "HANJA",
        0x5F20: "HANJA",
        0x5F21: "HANJA",
        0x5F22: "HANJA",
        0x5F23: "HANJA",
        0x5F24: "HANJA",
        0x5F25: "HANJA",
        0x5F26: "HANJA",
        0x5F27: "HANJA",
        0x5F28: "HANJA",
        0x5F29: "HANJA",
        0x5F2A: "HANJA",
        0x5F2B: "HANJA",
        0x5F2C: "HANJA",
        0x5F2D: "HANJA",
        0x5F2E: "HANJA",
        0x5F2F: "HANJA",
        0x5F30: "HANJA",
        0x5F31: "HANJA",
        0x5F32: "HANJA",
        0x5F33: "HANJA",
        0x5F34: "HANJA",
        0x5F35: "HANJA",
        0x5F36: "HANJA",
        0x5F37: "HANJA",
        0x5F38: "HANJA",
        0x5F39: "HANJA",
        0x5F3A: "HANJA",
        0x5F3B: "HANJA",
        0x5F3C: "HANJA",
        0x5F3D: "HANJA",
        0x5F3E: "HANJA",
        0x5F3F: "HANJA",
        0x5F40: "HANJA",
        0x5F41: "HANJA",
        0x5F42: "HANJA",
        0x5F43: "HANJA",
        0x5F44: "HANJA",
        0x5F45: "HANJA",
        0x5F46: "HANJA",
        0x5F47: "HANJA",
        0x5F48: "HANJA",
        0x5F49: "HANJA",
        0x5F4A: "HANJA",
        0x5F4B: "HANJA",
        0x5F4C: "HANJA",
        0x5F4D: "HANJA",
        0x5F4E: "HANJA",
        0x5F4F: "HANJA",
        0x5F50: "HANJA",
        0x5F51: "HANJA",
        0x5F52: "HANJA",
        0x5F53: "HANJA",
        0x5F54: "HANJA",
        0x5F55: "HANJA",
        0x5F56: "HANJA",
        0x5F57: "HANJA",
        0x5F58: "HANJA",
        0x5F59: "HANJA",
        0x5F5A: "HANJA",
        0x5F5B: "HANJA",
        0x5F5C: "HANJA",
        0x5F5D: "HANJA",
        0x5F5E: "HANJA",
        0x5F5F: "HANJA",
        0x5F60: "HANJA",
        0x5F61: "HANJA",
        0x5F62: "HANJA",
        0x5F63: "HANJA",
        0x5F64: "HANJA",
        0x5F65: "HANJA",
        0x5F66: "HANJA",
        0x5F67: "HANJA",
        0x5F68: "HANJA",
        0x5F69: "HANJA",
        0x5F6A: "HANJA",
        0x5F6B: "HANJA",
        0x5F6C: "HANJA",
        0x5F6D: "HANJA",
        0x5F6E: "HANJA",
        0x5F6F: "HANJA",
        0x5F70: "HANJA",
        0x5F71: "HANJA",
        0x5F72: "HANJA",
        0x5F73: "HANJA",
        0x5F74: "HANJA",
        0x5F75: "HANJA",
        0x5F76: "HANJA",
        0x5F77: "HANJA",
        0x5F78: "HANJA",
        0x5F79: "HANJA",
        0x5F7A: "HANJA",
        0x5F7B: "HANJA",
        0x5F7C: "HANJA",
        0x5F7D: "HANJA",
        0x5F7E: "HANJA",
        0x5F7F: "HANJA",
        0x5F80: "HANJA",
        0x5F81: "HANJA",
        0x5F82: "HANJA",
        0x5F83: "HANJA",
        0x5F84: "HANJA",
        0x5F85: "HANJA",
        0x5F86: "HANJA",
        0x5F87: "HANJA",
        0x5F88: "HANJA",
        0x5F89: "HANJA",
        0x5F8A: "HANJA",
        0x5F8B: "HANJA",
        0x5F8C: "HANJA",
        0x5F8D: "HANJA",
        0x5F8E: "HANJA",
        0x5F8F: "HANJA",
        0x5F90: "HANJA",
        0x5F91: "HANJA",
        0x5F92: "HANJA",
        0x5F93: "HANJA",
        0x5F94: "HANJA",
        0x5F95: "HANJA",
        0x5F96: "HANJA",
        0x5F97: "HANJA",
        0x5F98: "HANJA",
        0x5F99: "HANJA",
        0x5F9A: "HANJA",
        0x5F9B: "HANJA",
        0x5F9C: "HANJA",
        0x5F9D: "HANJA",
        0x5F9E: "HANJA",
        0x5F9F: "HANJA",
        0x5FA0: "HANJA",
        0x5FA1: "HANJA",
        0x5FA2: "HANJA",
        0x5FA3: "HANJA",
        0x5FA4: "HANJA",
        0x5FA5: "HANJA",
        0x5FA6: "HANJA",
        0x5FA7: "HANJA",
        0x5FA8: "HANJA",
        0x5FA9: "HANJA",
        0x5FAA: "HANJA",
        0x5FAB: "HANJA",
        0x5FAC: "HANJA",
        0x5FAD: "HANJA",
        0x5FAE: "HANJA",
        0x5FAF: "HANJA",
        0x5FB0: "HANJA",
        0x5FB1: "HANJA",
        0x5FB2: "HANJA",
        0x5FB3: "HANJA",
        0x5FB4: "HANJA",
        0x5FB5: "HANJA",
        0x5FB6: "HANJA",
        0x5FB7: "HANJA",
        0x5FB8: "HANJA",
        0x5FB9: "HANJA",
        0x5FBA: "HANJA",
        0x5FBB: "HANJA",
        0x5FBC: "HANJA",
        0x5FBD: "HANJA",
        0x5FBE: "HANJA",
        0x5FBF: "HANJA",
        0x5FC0: "HANJA",
        0x5FC1: "HANJA",
        0x5FC2: "HANJA",
        0x5FC3: "HANJA",
        0x5FC4: "HANJA",
        0x5FC5: "HANJA",
        0x5FC6: "HANJA",
        0x5FC7: "HANJA",
        0x5FC8: "HANJA",
        0x5FC9: "HANJA",
        0x5FCA: "HANJA",
        0x5FCB: "HANJA",
        0x5FCC: "HANJA",
        0x5FCD: "HANJA",
        0x5FCE: "HANJA",
        0x5FCF: "HANJA",
        0x5FD0: "HANJA",
        0x5FD1: "HANJA",
        0x5FD2: "HANJA",
        0x5FD3: "HANJA",
        0x5FD4: "HANJA",
        0x5FD5: "HANJA",
        0x5FD6: "HANJA",
        0x5FD7: "HANJA",
        0x5FD8: "HANJA",
        0x5FD9: "HANJA",
        0x5FDA: "HANJA",
        0x5FDB: "HANJA",
        0x5FDC: "HANJA",
        0x5FDD: "HANJA",
        0x5FDE: "HANJA",
        0x5FDF: "HANJA",
        0x5FE0: "HANJA",
        0x5FE1: "HANJA",
        0x5FE2: "HANJA",
        0x5FE3: "HANJA",
        0x5FE4: "HANJA",
        0x5FE5: "HANJA",
        0x5FE6: "HANJA",
        0x5FE7: "HANJA",
        0x5FE8: "HANJA",
        0x5FE9: "HANJA",
        0x5FEA: "HANJA",
        0x5FEB: "HANJA",
        0x5FEC: "HANJA",
        0x5FED: "HANJA",
        0x5FEE: "HANJA",
        0x5FEF: "HANJA",
        0x5FF0: "HANJA",
        0x5FF1: "HANJA",
        0x5FF2: "HANJA",
        0x5FF3: "HANJA",
        0x5FF4: "HANJA",
        0x5FF5: "HANJA",
        0x5FF6: "HANJA",
        0x5FF7: "HANJA",
        0x5FF8: "HANJA",
        0x5FF9: "HANJA",
        0x5FFA: "HANJA",
        0x5FFB: "HANJA",
        0x5FFC: "HANJA",
        0x5FFD: "HANJA",
        0x5FFE: "HANJA",
        0x5FFF: "HANJA",
        0x6000: "HANJA",
        0x6001: "HANJA",
        0x6002: "HANJA",
        0x6003: "HANJA",
        0x6004: "HANJA",
        0x6005: "HANJA",
        0x6006: "HANJA",
        0x6007: "HANJA",
        0x6008: "HANJA",
        0x6009: "HANJA",
        0x600A: "HANJA",
        0x600B: "HANJA",
        0x600C: "HANJA",
        0x600D: "HANJA",
        0x600E: "HANJA",
        0x600F: "HANJA",
        0x6010: "HANJA",
        0x6011: "HANJA",
        0x6012: "HANJA",
        0x6013: "HANJA",
        0x6014: "HANJA",
        0x6015: "HANJA",
        0x6016: "HANJA",
        0x6017: "HANJA",
        0x6018: "HANJA",
        0x6019: "HANJA",
        0x601A: "HANJA",
        0x601B: "HANJA",
        0x601C: "HANJA",
        0x601D: "HANJA",
        0x601E: "HANJA",
        0x601F: "HANJA",
        0x6020: "HANJA",
        0x6021: "HANJA",
        0x6022: "HANJA",
        0x6023: "HANJA",
        0x6024: "HANJA",
        0x6025: "HANJA",
        0x6026: "HANJA",
        0x6027: "HANJA",
        0x6028: "HANJA",
        0x6029: "HANJA",
        0x602A: "HANJA",
        0x602B: "HANJA",
        0x602C: "HANJA",
        0x602D: "HANJA",
        0x602E: "HANJA",
        0x602F: "HANJA",
        0x6030: "HANJA",
        0x6031: "HANJA",
        0x6032: "HANJA",
        0x6033: "HANJA",
        0x6034: "HANJA",
        0x6035: "HANJA",
        0x6036: "HANJA",
        0x6037: "HANJA",
        0x6038: "HANJA",
        0x6039: "HANJA",
        0x603A: "HANJA",
        0x603B: "HANJA",
        0x603C: "HANJA",
        0x603D: "HANJA",
        0x603E: "HANJA",
        0x603F: "HANJA",
        0x6040: "HANJA",
        0x6041: "HANJA",
        0x6042: "HANJA",
        0x6043: "HANJA",
        0x6044: "HANJA",
        0x6045: "HANJA",
        0x6046: "HANJA",
        0x6047: "HANJA",
        0x6048: "HANJA",
        0x6049: "HANJA",
        0x604A: "HANJA",
        0x604B: "HANJA",
        0x604C: "HANJA",
        0x604D: "HANJA",
        0x604E: "HANJA",
        0x604F: "HANJA",
        0x6050: "HANJA",
        0x6051: "HANJA",
        0x6052: "HANJA",
        0x6053: "HANJA",
        0x6054: "HANJA",
        0x6055: "HANJA",
        0x6056: "HANJA",
        0x6057: "HANJA",
        0x6058: "HANJA",
        0x6059: "HANJA",
        0x605A: "HANJA",
        0x605B: "HANJA",
        0x605C: "HANJA",
        0x605D: "HANJA",
        0x605E: "HANJA",
        0x605F: "HANJA",
        0x6060: "HANJA",
        0x6061: "HANJA",
        0x6062: "HANJA",
        0x6063: "HANJA",
        0x6064: "HANJA",
        0x6065: "HANJA",
        0x6066: "HANJA",
        0x6067: "HANJA",
        0x6068: "HANJA",
        0x6069: "HANJA",
        0x606A: "HANJA",
        0x606B: "HANJA",
        0x606C: "HANJA",
        0x606D: "HANJA",
        0x606E: "HANJA",
        0x606F: "HANJA",
        0x6070: "HANJA",
        0x6071: "HANJA",
        0x6072: "HANJA",
        0x6073: "HANJA",
        0x6074: "HANJA",
        0x6075: "HANJA",
        0x6076: "HANJA",
        0x6077: "HANJA",
        0x6078: "HANJA",
        0x6079: "HANJA",
        0x607A: "HANJA",
        0x607B: "HANJA",
        0x607C: "HANJA",
        0x607D: "HANJA",
        0x607E: "HANJA",
        0x607F: "HANJA",
        0x6080: "HANJA",
        0x6081: "HANJA",
        0x6082: "HANJA",
        0x6083: "HANJA",
        0x6084: "HANJA",
        0x6085: "HANJA",
        0x6086: "HANJA",
        0x6087: "HANJA",
        0x6088: "HANJA",
        0x6089: "HANJA",
        0x608A: "HANJA",
        0x608B: "HANJA",
        0x608C: "HANJA",
        0x608D: "HANJA",
        0x608E: "HANJA",
        0x608F: "HANJA",
        0x6090: "HANJA",
        0x6091: "HANJA",
        0x6092: "HANJA",
        0x6093: "HANJA",
        0x6094: "HANJA",
        0x6095: "HANJA",
        0x6096: "HANJA",
        0x6097: "HANJA",
        0x6098: "HANJA",
        0x6099: "HANJA",
        0x609A: "HANJA",
        0x609B: "HANJA",
        0x609C: "HANJA",
        0x609D: "HANJA",
        0x609E: "HANJA",
        0x609F: "HANJA",
        0x60A0: "HANJA",
        0x60A1: "HANJA",
        0x60A2: "HANJA",
        0x60A3: "HANJA",
        0x60A4: "HANJA",
        0x60A5: "HANJA",
        0x60A6: "HANJA",
        0x60A7: "HANJA",
        0x60A8: "HANJA",
        0x60A9: "HANJA",
        0x60AA: "HANJA",
        0x60AB: "HANJA",
        0x60AC: "HANJA",
        0x60AD: "HANJA",
        0x60AE: "HANJA",
        0x60AF: "HANJA",
        0x60B0: "HANJA",
        0x60B1: "HANJA",
        0x60B2: "HANJA",
        0x60B3: "HANJA",
        0x60B4: "HANJA",
        0x60B5: "HANJA",
        0x60B6: "HANJA",
        0x60B7: "HANJA",
        0x60B8: "HANJA",
        0x60B9: "HANJA",
        0x60BA: "HANJA",
        0x60BB: "HANJA",
        0x60BC: "HANJA",
        0x60BD: "HANJA",
        0x60BE: "HANJA",
        0x60BF: "HANJA",
        0x60C0: "HANJA",
        0x60C1: "HANJA",
        0x60C2: "HANJA",
        0x60C3: "HANJA",
        0x60C4: "HANJA",
        0x60C5: "HANJA",
        0x60C6: "HANJA",
        0x60C7: "HANJA",
        0x60C8: "HANJA",
        0x60C9: "HANJA",
        0x60CA: "HANJA",
        0x60CB: "HANJA",
        0x60CC: "HANJA",
        0x60CD: "HANJA",
        0x60CE: "HANJA",
        0x60CF: "HANJA",
        0x60D0: "HANJA",
        0x60D1: "HANJA",
        0x60D2: "HANJA",
        0x60D3: "HANJA",
        0x60D4: "HANJA",
        0x60D5: "HANJA",
        0x60D6: "HANJA",
        0x60D7: "HANJA",
        0x60D8: "HANJA",
        0x60D9: "HANJA",
        0x60DA: "HANJA",
        0x60DB: "HANJA",
        0x60DC: "HANJA",
        0x60DD: "HANJA",
        0x60DE: "HANJA",
        0x60DF: "HANJA",
        0x60E0: "HANJA",
        0x60E1: "HANJA",
        0x60E2: "HANJA",
        0x60E3: "HANJA",
        0x60E4: "HANJA",
        0x60E5: "HANJA",
        0x60E6: "HANJA",
        0x60E7: "HANJA",
        0x60E8: "HANJA",
        0x60E9: "HANJA",
        0x60EA: "HANJA",
        0x60EB: "HANJA",
        0x60EC: "HANJA",
        0x60ED: "HANJA",
        0x60EE: "HANJA",
        0x60EF: "HANJA",
        0x60F0: "HANJA",
        0x60F1: "HANJA",
        0x60F2: "HANJA",
        0x60F3: "HANJA",
        0x60F4: "HANJA",
        0x60F5: "HANJA",
        0x60F6: "HANJA",
        0x60F7: "HANJA",
        0x60F8: "HANJA",
        0x60F9: "HANJA",
        0x60FA: "HANJA",
        0x60FB: "HANJA",
        0x60FC: "HANJA",
        0x60FD: "HANJA",
        0x60FE: "HANJA",
        0x60FF: "HANJA",
        0x6100: "HANJA",
        0x6101: "HANJA",
        0x6102: "HANJA",
        0x6103: "HANJA",
        0x6104: "HANJA",
        0x6105: "HANJA",
        0x6106: "HANJA",
        0x6107: "HANJA",
        0x6108: "HANJA",
        0x6109: "HANJA",
        0x610A: "HANJA",
        0x610B: "HANJA",
        0x610C: "HANJA",
        0x610D: "HANJA",
        0x610E: "HANJA",
        0x610F: "HANJA",
        0x6110: "HANJA",
        0x6111: "HANJA",
        0x6112: "HANJA",
        0x6113: "HANJA",
        0x6114: "HANJA",
        0x6115: "HANJA",
        0x6116: "HANJA",
        0x6117: "HANJA",
        0x6118: "HANJA",
        0x6119: "HANJA",
        0x611A: "HANJA",
        0x611B: "HANJA",
        0x611C: "HANJA",
        0x611D: "HANJA",
        0x611E: "HANJA",
        0x611F: "HANJA",
        0x6120: "HANJA",
        0x6121: "HANJA",
        0x6122: "HANJA",
        0x6123: "HANJA",
        0x6124: "HANJA",
        0x6125: "HANJA",
        0x6126: "HANJA",
        0x6127: "HANJA",
        0x6128: "HANJA",
        0x6129: "HANJA",
        0x612A: "HANJA",
        0x612B: "HANJA",
        0x612C: "HANJA",
        0x612D: "HANJA",
        0x612E: "HANJA",
        0x612F: "HANJA",
        0x6130: "HANJA",
        0x6131: "HANJA",
        0x6132: "HANJA",
        0x6133: "HANJA",
        0x6134: "HANJA",
        0x6135: "HANJA",
        0x6136: "HANJA",
        0x6137: "HANJA",
        0x6138: "HANJA",
        0x6139: "HANJA",
        0x613A: "HANJA",
        0x613B: "HANJA",
        0x613C: "HANJA",
        0x613D: "HANJA",
        0x613E: "HANJA",
        0x613F: "HANJA",
        0x6140: "HANJA",
        0x6141: "HANJA",
        0x6142: "HANJA",
        0x6143: "HANJA",
        0x6144: "HANJA",
        0x6145: "HANJA",
        0x6146: "HANJA",
        0x6147: "HANJA",
        0x6148: "HANJA",
        0x6149: "HANJA",
        0x614A: "HANJA",
        0x614B: "HANJA",
        0x614C: "HANJA",
        0x614D: "HANJA",
        0x614E: "HANJA",
        0x614F: "HANJA",
        0x6150: "HANJA",
        0x6151: "HANJA",
        0x6152: "HANJA",
        0x6153: "HANJA",
        0x6154: "HANJA",
        0x6155: "HANJA",
        0x6156: "HANJA",
        0x6157: "HANJA",
        0x6158: "HANJA",
        0x6159: "HANJA",
        0x615A: "HANJA",
        0x615B: "HANJA",
        0x615C: "HANJA",
        0x615D: "HANJA",
        0x615E: "HANJA",
        0x615F: "HANJA",
        0x6160: "HANJA",
        0x6161: "HANJA",
        0x6162: "HANJA",
        0x6163: "HANJA",
        0x6164: "HANJA",
        0x6165: "HANJA",
        0x6166: "HANJA",
        0x6167: "HANJA",
        0x6168: "HANJA",
        0x6169: "HANJA",
        0x616A: "HANJA",
        0x616B: "HANJA",
        0x616C: "HANJA",
        0x616D: "HANJA",
        0x616E: "HANJA",
        0x616F: "HANJA",
        0x6170: "HANJA",
        0x6171: "HANJA",
        0x6172: "HANJA",
        0x6173: "HANJA",
        0x6174: "HANJA",
        0x6175: "HANJA",
        0x6176: "HANJA",
        0x6177: "HANJA",
        0x6178: "HANJA",
        0x6179: "HANJA",
        0x617A: "HANJA",
        0x617B: "HANJA",
        0x617C: "HANJA",
        0x617D: "HANJA",
        0x617E: "HANJA",
        0x617F: "HANJA",
        0x6180: "HANJA",
        0x6181: "HANJA",
        0x6182: "HANJA",
        0x6183: "HANJA",
        0x6184: "HANJA",
        0x6185: "HANJA",
        0x6186: "HANJA",
        0x6187: "HANJA",
        0x6188: "HANJA",
        0x6189: "HANJA",
        0x618A: "HANJA",
        0x618B: "HANJA",
        0x618C: "HANJA",
        0x618D: "HANJA",
        0x618E: "HANJA",
        0x618F: "HANJA",
        0x6190: "HANJA",
        0x6191: "HANJA",
        0x6192: "HANJA",
        0x6193: "HANJA",
        0x6194: "HANJA",
        0x6195: "HANJA",
        0x6196: "HANJA",
        0x6197: "HANJA",
        0x6198: "HANJA",
        0x6199: "HANJA",
        0x619A: "HANJA",
        0x619B: "HANJA",
        0x619C: "HANJA",
        0x619D: "HANJA",
        0x619E: "HANJA",
        0x619F: "HANJA",
        0x61A0: "HANJA",
        0x61A1: "HANJA",
        0x61A2: "HANJA",
        0x61A3: "HANJA",
        0x61A4: "HANJA",
        0x61A5: "HANJA",
        0x61A6: "HANJA",
        0x61A7: "HANJA",
        0x61A8: "HANJA",
        0x61A9: "HANJA",
        0x61AA: "HANJA",
        0x61AB: "HANJA",
        0x61AC: "HANJA",
        0x61AD: "HANJA",
        0x61AE: "HANJA",
        0x61AF: "HANJA",
        0x61B0: "HANJA",
        0x61B1: "HANJA",
        0x61B2: "HANJA",
        0x61B3: "HANJA",
        0x61B4: "HANJA",
        0x61B5: "HANJA",
        0x61B6: "HANJA",
        0x61B7: "HANJA",
        0x61B8: "HANJA",
        0x61B9: "HANJA",
        0x61BA: "HANJA",
        0x61BB: "HANJA",
        0x61BC: "HANJA",
        0x61BD: "HANJA",
        0x61BE: "HANJA",
        0x61BF: "HANJA",
        0x61C0: "HANJA",
        0x61C1: "HANJA",
        0x61C2: "HANJA",
        0x61C3: "HANJA",
        0x61C4: "HANJA",
        0x61C5: "HANJA",
        0x61C6: "HANJA",
        0x61C7: "HANJA",
        0x61C8: "HANJA",
        0x61C9: "HANJA",
        0x61CA: "HANJA",
        0x61CB: "HANJA",
        0x61CC: "HANJA",
        0x61CD: "HANJA",
        0x61CE: "HANJA",
        0x61CF: "HANJA",
        0x61D0: "HANJA",
        0x61D1: "HANJA",
        0x61D2: "HANJA",
        0x61D3: "HANJA",
        0x61D4: "HANJA",
        0x61D5: "HANJA",
        0x61D6: "HANJA",
        0x61D7: "HANJA",
        0x61D8: "HANJA",
        0x61D9: "HANJA",
        0x61DA: "HANJA",
        0x61DB: "HANJA",
        0x61DC: "HANJA",
        0x61DD: "HANJA",
        0x61DE: "HANJA",
        0x61DF: "HANJA",
        0x61E0: "HANJA",
        0x61E1: "HANJA",
        0x61E2: "HANJA",
        0x61E3: "HANJA",
        0x61E4: "HANJA",
        0x61E5: "HANJA",
        0x61E6: "HANJA",
        0x61E7: "HANJA",
        0x61E8: "HANJA",
        0x61E9: "HANJA",
        0x61EA: "HANJA",
        0x61EB: "HANJA",
        0x61EC: "HANJA",
        0x61ED: "HANJA",
        0x61EE: "HANJA",
        0x61EF: "HANJA",
        0x61F0: "HANJA",
        0x61F1: "HANJA",
        0x61F2: "HANJA",
        0x61F3: "HANJA",
        0x61F4: "HANJA",
        0x61F5: "HANJA",
        0x61F6: "HANJA",
        0x61F7: "HANJA",
        0x61F8: "HANJA",
        0x61F9: "HANJA",
        0x61FA: "HANJA",
        0x61FB: "HANJA",
        0x61FC: "HANJA",
        0x61FD: "HANJA",
        0x61FE: "HANJA",
        0x61FF: "HANJA",
        0x6200: "HANJA",
        0x6201: "HANJA",
        0x6202: "HANJA",
        0x6203: "HANJA",
        0x6204: "HANJA",
        0x6205: "HANJA",
        0x6206: "HANJA",
        0x6207: "HANJA",
        0x6208: "HANJA",
        0x6209: "HANJA",
        0x620A: "HANJA",
        0x620B: "HANJA",
        0x620C: "HANJA",
        0x620D: "HANJA",
        0x620E: "HANJA",
        0x620F: "HANJA",
        0x6210: "HANJA",
        0x6211: "HANJA",
        0x6212: "HANJA",
        0x6213: "HANJA",
        0x6214: "HANJA",
        0x6215: "HANJA",
        0x6216: "HANJA",
        0x6217: "HANJA",
        0x6218: "HANJA",
        0x6219: "HANJA",
        0x621A: "HANJA",
        0x621B: "HANJA",
        0x621C: "HANJA",
        0x621D: "HANJA",
        0x621E: "HANJA",
        0x621F: "HANJA",
        0x6220: "HANJA",
        0x6221: "HANJA",
        0x6222: "HANJA",
        0x6223: "HANJA",
        0x6224: "HANJA",
        0x6225: "HANJA",
        0x6226: "HANJA",
        0x6227: "HANJA",
        0x6228: "HANJA",
        0x6229: "HANJA",
        0x622A: "HANJA",
        0x622B: "HANJA",
        0x622C: "HANJA",
        0x622D: "HANJA",
        0x622E: "HANJA",
        0x622F: "HANJA",
        0x6230: "HANJA",
        0x6231: "HANJA",
        0x6232: "HANJA",
        0x6233: "HANJA",
        0x6234: "HANJA",
        0x6235: "HANJA",
        0x6236: "HANJA",
        0x6237: "HANJA",
        0x6238: "HANJA",
        0x6239: "HANJA",
        0x623A: "HANJA",
        0x623B: "HANJA",
        0x623C: "HANJA",
        0x623D: "HANJA",
        0x623E: "HANJA",
        0x623F: "HANJA",
        0x6240: "HANJA",
        0x6241: "HANJA",
        0x6242: "HANJA",
        0x6243: "HANJA",
        0x6244: "HANJA",
        0x6245: "HANJA",
        0x6246: "HANJA",
        0x6247: "HANJA",
        0x6248: "HANJA",
        0x6249: "HANJA",
        0x624A: "HANJA",
        0x624B: "HANJA",
        0x624C: "HANJA",
        0x624D: "HANJA",
        0x624E: "HANJA",
        0x624F: "HANJA",
        0x6250: "HANJA",
        0x6251: "HANJA",
        0x6252: "HANJA",
        0x6253: "HANJA",
        0x6254: "HANJA",
        0x6255: "HANJA",
        0x6256: "HANJA",
        0x6257: "HANJA",
        0x6258: "HANJA",
        0x6259: "HANJA",
        0x625A: "HANJA",
        0x625B: "HANJA",
        0x625C: "HANJA",
        0x625D: "HANJA",
        0x625E: "HANJA",
        0x625F: "HANJA",
        0x6260: "HANJA",
        0x6261: "HANJA",
        0x6262: "HANJA",
        0x6263: "HANJA",
        0x6264: "HANJA",
        0x6265: "HANJA",
        0x6266: "HANJA",
        0x6267: "HANJA",
        0x6268: "HANJA",
        0x6269: "HANJA",
        0x626A: "HANJA",
        0x626B: "HANJA",
        0x626C: "HANJA",
        0x626D: "HANJA",
        0x626E: "HANJA",
        0x626F: "HANJA",
        0x6270: "HANJA",
        0x6271: "HANJA",
        0x6272: "HANJA",
        0x6273: "HANJA",
        0x6274: "HANJA",
        0x6275: "HANJA",
        0x6276: "HANJA",
        0x6277: "HANJA",
        0x6278: "HANJA",
        0x6279: "HANJA",
        0x627A: "HANJA",
        0x627B: "HANJA",
        0x627C: "HANJA",
        0x627D: "HANJA",
        0x627E: "HANJA",
        0x627F: "HANJA",
        0x6280: "HANJA",
        0x6281: "HANJA",
        0x6282: "HANJA",
        0x6283: "HANJA",
        0x6284: "HANJA",
        0x6285: "HANJA",
        0x6286: "HANJA",
        0x6287: "HANJA",
        0x6288: "HANJA",
        0x6289: "HANJA",
        0x628A: "HANJA",
        0x628B: "HANJA",
        0x628C: "HANJA",
        0x628D: "HANJA",
        0x628E: "HANJA",
        0x628F: "HANJA",
        0x6290: "HANJA",
        0x6291: "HANJA",
        0x6292: "HANJA",
        0x6293: "HANJA",
        0x6294: "HANJA",
        0x6295: "HANJA",
        0x6296: "HANJA",
        0x6297: "HANJA",
        0x6298: "HANJA",
        0x6299: "HANJA",
        0x629A: "HANJA",
        0x629B: "HANJA",
        0x629C: "HANJA",
        0x629D: "HANJA",
        0x629E: "HANJA",
        0x629F: "HANJA",
        0x62A0: "HANJA",
        0x62A1: "HANJA",
        0x62A2: "HANJA",
        0x62A3: "HANJA",
        0x62A4: "HANJA",
        0x62A5: "HANJA",
        0x62A6: "HANJA",
        0x62A7: "HANJA",
        0x62A8: "HANJA",
        0x62A9: "HANJA",
        0x62AA: "HANJA",
        0x62AB: "HANJA",
        0x62AC: "HANJA",
        0x62AD: "HANJA",
        0x62AE: "HANJA",
        0x62AF: "HANJA",
        0x62B0: "HANJA",
        0x62B1: "HANJA",
        0x62B2: "HANJA",
        0x62B3: "HANJA",
        0x62B4: "HANJA",
        0x62B5: "HANJA",
        0x62B6: "HANJA",
        0x62B7: "HANJA",
        0x62B8: "HANJA",
        0x62B9: "HANJA",
        0x62BA: "HANJA",
        0x62BB: "HANJA",
        0x62BC: "HANJA",
        0x62BD: "HANJA",
        0x62BE: "HANJA",
        0x62BF: "HANJA",
        0x62C0: "HANJA",
        0x62C1: "HANJA",
        0x62C2: "HANJA",
        0x62C3: "HANJA",
        0x62C4: "HANJA",
        0x62C5: "HANJA",
        0x62C6: "HANJA",
        0x62C7: "HANJA",
        0x62C8: "HANJA",
        0x62C9: "HANJA",
        0x62CA: "HANJA",
        0x62CB: "HANJA",
        0x62CC: "HANJA",
        0x62CD: "HANJA",
        0x62CE: "HANJA",
        0x62CF: "HANJA",
        0x62D0: "HANJA",
        0x62D1: "HANJA",
        0x62D2: "HANJA",
        0x62D3: "HANJA",
        0x62D4: "HANJA",
        0x62D5: "HANJA",
        0x62D6: "HANJA",
        0x62D7: "HANJA",
        0x62D8: "HANJA",
        0x62D9: "HANJA",
        0x62DA: "HANJA",
        0x62DB: "HANJA",
        0x62DC: "HANJA",
        0x62DD: "HANJA",
        0x62DE: "HANJA",
        0x62DF: "HANJA",
        0x62E0: "HANJA",
        0x62E1: "HANJA",
        0x62E2: "HANJA",
        0x62E3: "HANJA",
        0x62E4: "HANJA",
        0x62E5: "HANJA",
        0x62E6: "HANJA",
        0x62E7: "HANJA",
        0x62E8: "HANJA",
        0x62E9: "HANJA",
        0x62EA: "HANJA",
        0x62EB: "HANJA",
        0x62EC: "HANJA",
        0x62ED: "HANJA",
        0x62EE: "HANJA",
        0x62EF: "HANJA",
        0x62F0: "HANJA",
        0x62F1: "HANJA",
        0x62F2: "HANJA",
        0x62F3: "HANJA",
        0x62F4: "HANJA",
        0x62F5: "HANJA",
        0x62F6: "HANJA",
        0x62F7: "HANJA",
        0x62F8: "HANJA",
        0x62F9: "HANJA",
        0x62FA: "HANJA",
        0x62FB: "HANJA",
        0x62FC: "HANJA",
        0x62FD: "HANJA",
        0x62FE: "HANJA",
        0x62FF: "HANJA",
        0x6300: "HANJA",
        0x6301: "HANJA",
        0x6302: "HANJA",
        0x6303: "HANJA",
        0x6304: "HANJA",
        0x6305: "HANJA",
        0x6306: "HANJA",
        0x6307: "HANJA",
        0x6308: "HANJA",
        0x6309: "HANJA",
        0x630A: "HANJA",
        0x630B: "HANJA",
        0x630C: "HANJA",
        0x630D: "HANJA",
        0x630E: "HANJA",
        0x630F: "HANJA",
        0x6310: "HANJA",
        0x6311: "HANJA",
        0x6312: "HANJA",
        0x6313: "HANJA",
        0x6314: "HANJA",
        0x6315: "HANJA",
        0x6316: "HANJA",
        0x6317: "HANJA",
        0x6318: "HANJA",
        0x6319: "HANJA",
        0x631A: "HANJA",
        0x631B: "HANJA",
        0x631C: "HANJA",
        0x631D: "HANJA",
        0x631E: "HANJA",
        0x631F: "HANJA",
        0x6320: "HANJA",
        0x6321: "HANJA",
        0x6322: "HANJA",
        0x6323: "HANJA",
        0x6324: "HANJA",
        0x6325: "HANJA",
        0x6326: "HANJA",
        0x6327: "HANJA",
        0x6328: "HANJA",
        0x6329: "HANJA",
        0x632A: "HANJA",
        0x632B: "HANJA",
        0x632C: "HANJA",
        0x632D: "HANJA",
        0x632E: "HANJA",
        0x632F: "HANJA",
        0x6330: "HANJA",
        0x6331: "HANJA",
        0x6332: "HANJA",
        0x6333: "HANJA",
        0x6334: "HANJA",
        0x6335: "HANJA",
        0x6336: "HANJA",
        0x6337: "HANJA",
        0x6338: "HANJA",
        0x6339: "HANJA",
        0x633A: "HANJA",
        0x633B: "HANJA",
        0x633C: "HANJA",
        0x633D: "HANJA",
        0x633E: "HANJA",
        0x633F: "HANJA",
        0x6340: "HANJA",
        0x6341: "HANJA",
        0x6342: "HANJA",
        0x6343: "HANJA",
        0x6344: "HANJA",
        0x6345: "HANJA",
        0x6346: "HANJA",
        0x6347: "HANJA",
        0x6348: "HANJA",
        0x6349: "HANJA",
        0x634A: "HANJA",
        0x634B: "HANJA",
        0x634C: "HANJA",
        0x634D: "HANJA",
        0x634E: "HANJA",
        0x634F: "HANJA",
        0x6350: "HANJA",
        0x6351: "HANJA",
        0x6352: "HANJA",
        0x6353: "HANJA",
        0x6354: "HANJA",
        0x6355: "HANJA",
        0x6356: "HANJA",
        0x6357: "HANJA",
        0x6358: "HANJA",
        0x6359: "HANJA",
        0x635A: "HANJA",
        0x635B: "HANJA",
        0x635C: "HANJA",
        0x635D: "HANJA",
        0x635E: "HANJA",
        0x635F: "HANJA",
        0x6360: "HANJA",
        0x6361: "HANJA",
        0x6362: "HANJA",
        0x6363: "HANJA",
        0x6364: "HANJA",
        0x6365: "HANJA",
        0x6366: "HANJA",
        0x6367: "HANJA",
        0x6368: "HANJA",
        0x6369: "HANJA",
        0x636A: "HANJA",
        0x636B: "HANJA",
        0x636C: "HANJA",
        0x636D: "HANJA",
        0x636E: "HANJA",
        0x636F: "HANJA",
        0x6370: "HANJA",
        0x6371: "HANJA",
        0x6372: "HANJA",
        0x6373: "HANJA",
        0x6374: "HANJA",
        0x6375: "HANJA",
        0x6376: "HANJA",
        0x6377: "HANJA",
        0x6378: "HANJA",
        0x6379: "HANJA",
        0x637A: "HANJA",
        0x637B: "HANJA",
        0x637C: "HANJA",
        0x637D: "HANJA",
        0x637E: "HANJA",
        0x637F: "HANJA",
        0x6380: "HANJA",
        0x6381: "HANJA",
        0x6382: "HANJA",
        0x6383: "HANJA",
        0x6384: "HANJA",
        0x6385: "HANJA",
        0x6386: "HANJA",
        0x6387: "HANJA",
        0x6388: "HANJA",
        0x6389: "HANJA",
        0x638A: "HANJA",
        0x638B: "HANJA",
        0x638C: "HANJA",
        0x638D: "HANJA",
        0x638E: "HANJA",
        0x638F: "HANJA",
        0x6390: "HANJA",
        0x6391: "HANJA",
        0x6392: "HANJA",
        0x6393: "HANJA",
        0x6394: "HANJA",
        0x6395: "HANJA",
        0x6396: "HANJA",
        0x6397: "HANJA",
        0x6398: "HANJA",
        0x6399: "HANJA",
        0x639A: "HANJA",
        0x639B: "HANJA",
        0x639C: "HANJA",
        0x639D: "HANJA",
        0x639E: "HANJA",
        0x639F: "HANJA",
        0x63A0: "HANJA",
        0x63A1: "HANJA",
        0x63A2: "HANJA",
        0x63A3: "HANJA",
        0x63A4: "HANJA",
        0x63A5: "HANJA",
        0x63A6: "HANJA",
        0x63A7: "HANJA",
        0x63A8: "HANJA",
        0x63A9: "HANJA",
        0x63AA: "HANJA",
        0x63AB: "HANJA",
        0x63AC: "HANJA",
        0x63AD: "HANJA",
        0x63AE: "HANJA",
        0x63AF: "HANJA",
        0x63B0: "HANJA",
        0x63B1: "HANJA",
        0x63B2: "HANJA",
        0x63B3: "HANJA",
        0x63B4: "HANJA",
        0x63B5: "HANJA",
        0x63B6: "HANJA",
        0x63B7: "HANJA",
        0x63B8: "HANJA",
        0x63B9: "HANJA",
        0x63BA: "HANJA",
        0x63BB: "HANJA",
        0x63BC: "HANJA",
        0x63BD: "HANJA",
        0x63BE: "HANJA",
        0x63BF: "HANJA",
        0x63C0: "HANJA",
        0x63C1: "HANJA",
        0x63C2: "HANJA",
        0x63C3: "HANJA",
        0x63C4: "HANJA",
        0x63C5: "HANJA",
        0x63C6: "HANJA",
        0x63C7: "HANJA",
        0x63C8: "HANJA",
        0x63C9: "HANJA",
        0x63CA: "HANJA",
        0x63CB: "HANJA",
        0x63CC: "HANJA",
        0x63CD: "HANJA",
        0x63CE: "HANJA",
        0x63CF: "HANJA",
        0x63D0: "HANJA",
        0x63D1: "HANJA",
        0x63D2: "HANJA",
        0x63D3: "HANJA",
        0x63D4: "HANJA",
        0x63D5: "HANJA",
        0x63D6: "HANJA",
        0x63D7: "HANJA",
        0x63D8: "HANJA",
        0x63D9: "HANJA",
        0x63DA: "HANJA",
        0x63DB: "HANJA",
        0x63DC: "HANJA",
        0x63DD: "HANJA",
        0x63DE: "HANJA",
        0x63DF: "HANJA",
        0x63E0: "HANJA",
        0x63E1: "HANJA",
        0x63E2: "HANJA",
        0x63E3: "HANJA",
        0x63E4: "HANJA",
        0x63E5: "HANJA",
        0x63E6: "HANJA",
        0x63E7: "HANJA",
        0x63E8: "HANJA",
        0x63E9: "HANJA",
        0x63EA: "HANJA",
        0x63EB: "HANJA",
        0x63EC: "HANJA",
        0x63ED: "HANJA",
        0x63EE: "HANJA",
        0x63EF: "HANJA",
        0x63F0: "HANJA",
        0x63F1: "HANJA",
        0x63F2: "HANJA",
        0x63F3: "HANJA",
        0x63F4: "HANJA",
        0x63F5: "HANJA",
        0x63F6: "HANJA",
        0x63F7: "HANJA",
        0x63F8: "HANJA",
        0x63F9: "HANJA",
        0x63FA: "HANJA",
        0x63FB: "HANJA",
        0x63FC: "HANJA",
        0x63FD: "HANJA",
        0x63FE: "HANJA",
        0x63FF: "HANJA",
        0x6400: "HANJA",
        0x6401: "HANJA",
        0x6402: "HANJA",
        0x6403: "HANJA",
        0x6404: "HANJA",
        0x6405: "HANJA",
        0x6406: "HANJA",
        0x6407: "HANJA",
        0x6408: "HANJA",
        0x6409: "HANJA",
        0x640A: "HANJA",
        0x640B: "HANJA",
        0x640C: "HANJA",
        0x640D: "HANJA",
        0x640E: "HANJA",
        0x640F: "HANJA",
        0x6410: "HANJA",
        0x6411: "HANJA",
        0x6412: "HANJA",
        0x6413: "HANJA",
        0x6414: "HANJA",
        0x6415: "HANJA",
        0x6416: "HANJA",
        0x6417: "HANJA",
        0x6418: "HANJA",
        0x6419: "HANJA",
        0x641A: "HANJA",
        0x641B: "HANJA",
        0x641C: "HANJA",
        0x641D: "HANJA",
        0x641E: "HANJA",
        0x641F: "HANJA",
        0x6420: "HANJA",
        0x6421: "HANJA",
        0x6422: "HANJA",
        0x6423: "HANJA",
        0x6424: "HANJA",
        0x6425: "HANJA",
        0x6426: "HANJA",
        0x6427: "HANJA",
        0x6428: "HANJA",
        0x6429: "HANJA",
        0x642A: "HANJA",
        0x642B: "HANJA",
        0x642C: "HANJA",
        0x642D: "HANJA",
        0x642E: "HANJA",
        0x642F: "HANJA",
        0x6430: "HANJA",
        0x6431: "HANJA",
        0x6432: "HANJA",
        0x6433: "HANJA",
        0x6434: "HANJA",
        0x6435: "HANJA",
        0x6436: "HANJA",
        0x6437: "HANJA",
        0x6438: "HANJA",
        0x6439: "HANJA",
        0x643A: "HANJA",
        0x643B: "HANJA",
        0x643C: "HANJA",
        0x643D: "HANJA",
        0x643E: "HANJA",
        0x643F: "HANJA",
        0x6440: "HANJA",
        0x6441: "HANJA",
        0x6442: "HANJA",
        0x6443: "HANJA",
        0x6444: "HANJA",
        0x6445: "HANJA",
        0x6446: "HANJA",
        0x6447: "HANJA",
        0x6448: "HANJA",
        0x6449: "HANJA",
        0x644A: "HANJA",
        0x644B: "HANJA",
        0x644C: "HANJA",
        0x644D: "HANJA",
        0x644E: "HANJA",
        0x644F: "HANJA",
        0x6450: "HANJA",
        0x6451: "HANJA",
        0x6452: "HANJA",
        0x6453: "HANJA",
        0x6454: "HANJA",
        0x6455: "HANJA",
        0x6456: "HANJA",
        0x6457: "HANJA",
        0x6458: "HANJA",
        0x6459: "HANJA",
        0x645A: "HANJA",
        0x645B: "HANJA",
        0x645C: "HANJA",
        0x645D: "HANJA",
        0x645E: "HANJA",
        0x645F: "HANJA",
        0x6460: "HANJA",
        0x6461: "HANJA",
        0x6462: "HANJA",
        0x6463: "HANJA",
        0x6464: "HANJA",
        0x6465: "HANJA",
        0x6466: "HANJA",
        0x6467: "HANJA",
        0x6468: "HANJA",
        0x6469: "HANJA",
        0x646A: "HANJA",
        0x646B: "HANJA",
        0x646C: "HANJA",
        0x646D: "HANJA",
        0x646E: "HANJA",
        0x646F: "HANJA",
        0x6470: "HANJA",
        0x6471: "HANJA",
        0x6472: "HANJA",
        0x6473: "HANJA",
        0x6474: "HANJA",
        0x6475: "HANJA",
        0x6476: "HANJA",
        0x6477: "HANJA",
        0x6478: "HANJA",
        0x6479: "HANJA",
        0x647A: "HANJA",
        0x647B: "HANJA",
        0x647C: "HANJA",
        0x647D: "HANJA",
        0x647E: "HANJA",
        0x647F: "HANJA",
        0x6480: "HANJA",
        0x6481: "HANJA",
        0x6482: "HANJA",
        0x6483: "HANJA",
        0x6484: "HANJA",
        0x6485: "HANJA",
        0x6486: "HANJA",
        0x6487: "HANJA",
        0x6488: "HANJA",
        0x6489: "HANJA",
        0x648A: "HANJA",
        0x648B: "HANJA",
        0x648C: "HANJA",
        0x648D: "HANJA",
        0x648E: "HANJA",
        0x648F: "HANJA",
        0x6490: "HANJA",
        0x6491: "HANJA",
        0x6492: "HANJA",
        0x6493: "HANJA",
        0x6494: "HANJA",
        0x6495: "HANJA",
        0x6496: "HANJA",
        0x6497: "HANJA",
        0x6498: "HANJA",
        0x6499: "HANJA",
        0x649A: "HANJA",
        0x649B: "HANJA",
        0x649C: "HANJA",
        0x649D: "HANJA",
        0x649E: "HANJA",
        0x649F: "HANJA",
        0x64A0: "HANJA",
        0x64A1: "HANJA",
        0x64A2: "HANJA",
        0x64A3: "HANJA",
        0x64A4: "HANJA",
        0x64A5: "HANJA",
        0x64A6: "HANJA",
        0x64A7: "HANJA",
        0x64A8: "HANJA",
        0x64A9: "HANJA",
        0x64AA: "HANJA",
        0x64AB: "HANJA",
        0x64AC: "HANJA",
        0x64AD: "HANJA",
        0x64AE: "HANJA",
        0x64AF: "HANJA",
        0x64B0: "HANJA",
        0x64B1: "HANJA",
        0x64B2: "HANJA",
        0x64B3: "HANJA",
        0x64B4: "HANJA",
        0x64B5: "HANJA",
        0x64B6: "HANJA",
        0x64B7: "HANJA",
        0x64B8: "HANJA",
        0x64B9: "HANJA",
        0x64BA: "HANJA",
        0x64BB: "HANJA",
        0x64BC: "HANJA",
        0x64BD: "HANJA",
        0x64BE: "HANJA",
        0x64BF: "HANJA",
        0x64C0: "HANJA",
        0x64C1: "HANJA",
        0x64C2: "HANJA",
        0x64C3: "HANJA",
        0x64C4: "HANJA",
        0x64C5: "HANJA",
        0x64C6: "HANJA",
        0x64C7: "HANJA",
        0x64C8: "HANJA",
        0x64C9: "HANJA",
        0x64CA: "HANJA",
        0x64CB: "HANJA",
        0x64CC: "HANJA",
        0x64CD: "HANJA",
        0x64CE: "HANJA",
        0x64CF: "HANJA",
        0x64D0: "HANJA",
        0x64D1: "HANJA",
        0x64D2: "HANJA",
        0x64D3: "HANJA",
        0x64D4: "HANJA",
        0x64D5: "HANJA",
        0x64D6: "HANJA",
        0x64D7: "HANJA",
        0x64D8: "HANJA",
        0x64D9: "HANJA",
        0x64DA: "HANJA",
        0x64DB: "HANJA",
        0x64DC: "HANJA",
        0x64DD: "HANJA",
        0x64DE: "HANJA",
        0x64DF: "HANJA",
        0x64E0: "HANJA",
        0x64E1: "HANJA",
        0x64E2: "HANJA",
        0x64E3: "HANJA",
        0x64E4: "HANJA",
        0x64E5: "HANJA",
        0x64E6: "HANJA",
        0x64E7: "HANJA",
        0x64E8: "HANJA",
        0x64E9: "HANJA",
        0x64EA: "HANJA",
        0x64EB: "HANJA",
        0x64EC: "HANJA",
        0x64ED: "HANJA",
        0x64EE: "HANJA",
        0x64EF: "HANJA",
        0x64F0: "HANJA",
        0x64F1: "HANJA",
        0x64F2: "HANJA",
        0x64F3: "HANJA",
        0x64F4: "HANJA",
        0x64F5: "HANJA",
        0x64F6: "HANJA",
        0x64F7: "HANJA",
        0x64F8: "HANJA",
        0x64F9: "HANJA",
        0x64FA: "HANJA",
        0x64FB: "HANJA",
        0x64FC: "HANJA",
        0x64FD: "HANJA",
        0x64FE: "HANJA",
        0x64FF: "HANJA",
        0x6500: "HANJA",
        0x6501: "HANJA",
        0x6502: "HANJA",
        0x6503: "HANJA",
        0x6504: "HANJA",
        0x6505: "HANJA",
        0x6506: "HANJA",
        0x6507: "HANJA",
        0x6508: "HANJA",
        0x6509: "HANJA",
        0x650A: "HANJA",
        0x650B: "HANJA",
        0x650C: "HANJA",
        0x650D: "HANJA",
        0x650E: "HANJA",
        0x650F: "HANJA",
        0x6510: "HANJA",
        0x6511: "HANJA",
        0x6512: "HANJA",
        0x6513: "HANJA",
        0x6514: "HANJA",
        0x6515: "HANJA",
        0x6516: "HANJA",
        0x6517: "HANJA",
        0x6518: "HANJA",
        0x6519: "HANJA",
        0x651A: "HANJA",
        0x651B: "HANJA",
        0x651C: "HANJA",
        0x651D: "HANJA",
        0x651E: "HANJA",
        0x651F: "HANJA",
        0x6520: "HANJA",
        0x6521: "HANJA",
        0x6522: "HANJA",
        0x6523: "HANJA",
        0x6524: "HANJA",
        0x6525: "HANJA",
        0x6526: "HANJA",
        0x6527: "HANJA",
        0x6528: "HANJA",
        0x6529: "HANJA",
        0x652A: "HANJA",
        0x652B: "HANJA",
        0x652C: "HANJA",
        0x652D: "HANJA",
        0x652E: "HANJA",
        0x652F: "HANJA",
        0x6530: "HANJA",
        0x6531: "HANJA",
        0x6532: "HANJA",
        0x6533: "HANJA",
        0x6534: "HANJA",
        0x6535: "HANJA",
        0x6536: "HANJA",
        0x6537: "HANJA",
        0x6538: "HANJA",
        0x6539: "HANJA",
        0x653A: "HANJA",
        0x653B: "HANJA",
        0x653C: "HANJA",
        0x653D: "HANJA",
        0x653E: "HANJA",
        0x653F: "HANJA",
        0x6540: "HANJA",
        0x6541: "HANJA",
        0x6542: "HANJA",
        0x6543: "HANJA",
        0x6544: "HANJA",
        0x6545: "HANJA",
        0x6546: "HANJA",
        0x6547: "HANJA",
        0x6548: "HANJA",
        0x6549: "HANJA",
        0x654A: "HANJA",
        0x654B: "HANJA",
        0x654C: "HANJA",
        0x654D: "HANJA",
        0x654E: "HANJA",
        0x654F: "HANJA",
        0x6550: "HANJA",
        0x6551: "HANJA",
        0x6552: "HANJA",
        0x6553: "HANJA",
        0x6554: "HANJA",
        0x6555: "HANJA",
        0x6556: "HANJA",
        0x6557: "HANJA",
        0x6558: "HANJA",
        0x6559: "HANJA",
        0x655A: "HANJA",
        0x655B: "HANJA",
        0x655C: "HANJA",
        0x655D: "HANJA",
        0x655E: "HANJA",
        0x655F: "HANJA",
        0x6560: "HANJA",
        0x6561: "HANJA",
        0x6562: "HANJA",
        0x6563: "HANJA",
        0x6564: "HANJA",
        0x6565: "HANJA",
        0x6566: "HANJA",
        0x6567: "HANJA",
        0x6568: "HANJA",
        0x6569: "HANJA",
        0x656A: "HANJA",
        0x656B: "HANJA",
        0x656C: "HANJA",
        0x656D: "HANJA",
        0x656E: "HANJA",
        0x656F: "HANJA",
        0x6570: "HANJA",
        0x6571: "HANJA",
        0x6572: "HANJA",
        0x6573: "HANJA",
        0x6574: "HANJA",
        0x6575: "HANJA",
        0x6576: "HANJA",
        0x6577: "HANJA",
        0x6578: "HANJA",
        0x6579: "HANJA",
        0x657A: "HANJA",
        0x657B: "HANJA",
        0x657C: "HANJA",
        0x657D: "HANJA",
        0x657E: "HANJA",
        0x657F: "HANJA",
        0x6580: "HANJA",
        0x6581: "HANJA",
        0x6582: "HANJA",
        0x6583: "HANJA",
        0x6584: "HANJA",
        0x6585: "HANJA",
        0x6586: "HANJA",
        0x6587: "HANJA",
        0x6588: "HANJA",
        0x6589: "HANJA",
        0x658A: "HANJA",
        0x658B: "HANJA",
        0x658C: "HANJA",
        0x658D: "HANJA",
        0x658E: "HANJA",
        0x658F: "HANJA",
        0x6590: "HANJA",
        0x6591: "HANJA",
        0x6592: "HANJA",
        0x6593: "HANJA",
        0x6594: "HANJA",
        0x6595: "HANJA",
        0x6596: "HANJA",
        0x6597: "HANJA",
        0x6598: "HANJA",
        0x6599: "HANJA",
        0x659A: "HANJA",
        0x659B: "HANJA",
        0x659C: "HANJA",
        0x659D: "HANJA",
        0x659E: "HANJA",
        0x659F: "HANJA",
        0x65A0: "HANJA",
        0x65A1: "HANJA",
        0x65A2: "HANJA",
        0x65A3: "HANJA",
        0x65A4: "HANJA",
        0x65A5: "HANJA",
        0x65A6: "HANJA",
        0x65A7: "HANJA",
        0x65A8: "HANJA",
        0x65A9: "HANJA",
        0x65AA: "HANJA",
        0x65AB: "HANJA",
        0x65AC: "HANJA",
        0x65AD: "HANJA",
        0x65AE: "HANJA",
        0x65AF: "HANJA",
        0x65B0: "HANJA",
        0x65B1: "HANJA",
        0x65B2: "HANJA",
        0x65B3: "HANJA",
        0x65B4: "HANJA",
        0x65B5: "HANJA",
        0x65B6: "HANJA",
        0x65B7: "HANJA",
        0x65B8: "HANJA",
        0x65B9: "HANJA",
        0x65BA: "HANJA",
        0x65BB: "HANJA",
        0x65BC: "HANJA",
        0x65BD: "HANJA",
        0x65BE: "HANJA",
        0x65BF: "HANJA",
        0x65C0: "HANJA",
        0x65C1: "HANJA",
        0x65C2: "HANJA",
        0x65C3: "HANJA",
        0x65C4: "HANJA",
        0x65C5: "HANJA",
        0x65C6: "HANJA",
        0x65C7: "HANJA",
        0x65C8: "HANJA",
        0x65C9: "HANJA",
        0x65CA: "HANJA",
        0x65CB: "HANJA",
        0x65CC: "HANJA",
        0x65CD: "HANJA",
        0x65CE: "HANJA",
        0x65CF: "HANJA",
        0x65D0: "HANJA",
        0x65D1: "HANJA",
        0x65D2: "HANJA",
        0x65D3: "HANJA",
        0x65D4: "HANJA",
        0x65D5: "HANJA",
        0x65D6: "HANJA",
        0x65D7: "HANJA",
        0x65D8: "HANJA",
        0x65D9: "HANJA",
        0x65DA: "HANJA",
        0x65DB: "HANJA",
        0x65DC: "HANJA",
        0x65DD: "HANJA",
        0x65DE: "HANJA",
        0x65DF: "HANJA",
        0x65E0: "HANJA",
        0x65E1: "HANJA",
        0x65E2: "HANJA",
        0x65E3: "HANJA",
        0x65E4: "HANJA",
        0x65E5: "HANJA",
        0x65E6: "HANJA",
        0x65E7: "HANJA",
        0x65E8: "HANJA",
        0x65E9: "HANJA",
        0x65EA: "HANJA",
        0x65EB: "HANJA",
        0x65EC: "HANJA",
        0x65ED: "HANJA",
        0x65EE: "HANJA",
        0x65EF: "HANJA",
        0x65F0: "HANJA",
        0x65F1: "HANJA",
        0x65F2: "HANJA",
        0x65F3: "HANJA",
        0x65F4: "HANJA",
        0x65F5: "HANJA",
        0x65F6: "HANJA",
        0x65F7: "HANJA",
        0x65F8: "HANJA",
        0x65F9: "HANJA",
        0x65FA: "HANJA",
        0x65FB: "HANJA",
        0x65FC: "HANJA",
        0x65FD: "HANJA",
        0x65FE: "HANJA",
        0x65FF: "HANJA",
        0x6600: "HANJA",
        0x6601: "HANJA",
        0x6602: "HANJA",
        0x6603: "HANJA",
        0x6604: "HANJA",
        0x6605: "HANJA",
        0x6606: "HANJA",
        0x6607: "HANJA",
        0x6608: "HANJA",
        0x6609: "HANJA",
        0x660A: "HANJA",
        0x660B: "HANJA",
        0x660C: "HANJA",
        0x660D: "HANJA",
        0x660E: "HANJA",
        0x660F: "HANJA",
        0x6610: "HANJA",
        0x6611: "HANJA",
        0x6612: "HANJA",
        0x6613: "HANJA",
        0x6614: "HANJA",
        0x6615: "HANJA",
        0x6616: "HANJA",
        0x6617: "HANJA",
        0x6618: "HANJA",
        0x6619: "HANJA",
        0x661A: "HANJA",
        0x661B: "HANJA",
        0x661C: "HANJA",
        0x661D: "HANJA",
        0x661E: "HANJA",
        0x661F: "HANJA",
        0x6620: "HANJA",
        0x6621: "HANJA",
        0x6622: "HANJA",
        0x6623: "HANJA",
        0x6624: "HANJA",
        0x6625: "HANJA",
        0x6626: "HANJA",
        0x6627: "HANJA",
        0x6628: "HANJA",
        0x6629: "HANJA",
        0x662A: "HANJA",
        0x662B: "HANJA",
        0x662C: "HANJA",
        0x662D: "HANJA",
        0x662E: "HANJA",
        0x662F: "HANJA",
        0x6630: "HANJA",
        0x6631: "HANJA",
        0x6632: "HANJA",
        0x6633: "HANJA",
        0x6634: "HANJA",
        0x6635: "HANJA",
        0x6636: "HANJA",
        0x6637: "HANJA",
        0x6638: "HANJA",
        0x6639: "HANJA",
        0x663A: "HANJA",
        0x663B: "HANJA",
        0x663C: "HANJA",
        0x663D: "HANJA",
        0x663E: "HANJA",
        0x663F: "HANJA",
        0x6640: "HANJA",
        0x6641: "HANJA",
        0x6642: "HANJA",
        0x6643: "HANJA",
        0x6644: "HANJA",
        0x6645: "HANJA",
        0x6646: "HANJA",
        0x6647: "HANJA",
        0x6648: "HANJA",
        0x6649: "HANJA",
        0x664A: "HANJA",
        0x664B: "HANJA",
        0x664C: "HANJA",
        0x664D: "HANJA",
        0x664E: "HANJA",
        0x664F: "HANJA",
        0x6650: "HANJA",
        0x6651: "HANJA",
        0x6652: "HANJA",
        0x6653: "HANJA",
        0x6654: "HANJA",
        0x6655: "HANJA",
        0x6656: "HANJA",
        0x6657: "HANJA",
        0x6658: "HANJA",
        0x6659: "HANJA",
        0x665A: "HANJA",
        0x665B: "HANJA",
        0x665C: "HANJA",
        0x665D: "HANJA",
        0x665E: "HANJA",
        0x665F: "HANJA",
        0x6660: "HANJA",
        0x6661: "HANJA",
        0x6662: "HANJA",
        0x6663: "HANJA",
        0x6664: "HANJA",
        0x6665: "HANJA",
        0x6666: "HANJA",
        0x6667: "HANJA",
        0x6668: "HANJA",
        0x6669: "HANJA",
        0x666A: "HANJA",
        0x666B: "HANJA",
        0x666C: "HANJA",
        0x666D: "HANJA",
        0x666E: "HANJA",
        0x666F: "HANJA",
        0x6670: "HANJA",
        0x6671: "HANJA",
        0x6672: "HANJA",
        0x6673: "HANJA",
        0x6674: "HANJA",
        0x6675: "HANJA",
        0x6676: "HANJA",
        0x6677: "HANJA",
        0x6678: "HANJA",
        0x6679: "HANJA",
        0x667A: "HANJA",
        0x667B: "HANJA",
        0x667C: "HANJA",
        0x667D: "HANJA",
        0x667E: "HANJA",
        0x667F: "HANJA",
        0x6680: "HANJA",
        0x6681: "HANJA",
        0x6682: "HANJA",
        0x6683: "HANJA",
        0x6684: "HANJA",
        0x6685: "HANJA",
        0x6686: "HANJA",
        0x6687: "HANJA",
        0x6688: "HANJA",
        0x6689: "HANJA",
        0x668A: "HANJA",
        0x668B: "HANJA",
        0x668C: "HANJA",
        0x668D: "HANJA",
        0x668E: "HANJA",
        0x668F: "HANJA",
        0x6690: "HANJA",
        0x6691: "HANJA",
        0x6692: "HANJA",
        0x6693: "HANJA",
        0x6694: "HANJA",
        0x6695: "HANJA",
        0x6696: "HANJA",
        0x6697: "HANJA",
        0x6698: "HANJA",
        0x6699: "HANJA",
        0x669A: "HANJA",
        0x669B: "HANJA",
        0x669C: "HANJA",
        0x669D: "HANJA",
        0x669E: "HANJA",
        0x669F: "HANJA",
        0x66A0: "HANJA",
        0x66A1: "HANJA",
        0x66A2: "HANJA",
        0x66A3: "HANJA",
        0x66A4: "HANJA",
        0x66A5: "HANJA",
        0x66A6: "HANJA",
        0x66A7: "HANJA",
        0x66A8: "HANJA",
        0x66A9: "HANJA",
        0x66AA: "HANJA",
        0x66AB: "HANJA",
        0x66AC: "HANJA",
        0x66AD: "HANJA",
        0x66AE: "HANJA",
        0x66AF: "HANJA",
        0x66B0: "HANJA",
        0x66B1: "HANJA",
        0x66B2: "HANJA",
        0x66B3: "HANJA",
        0x66B4: "HANJA",
        0x66B5: "HANJA",
        0x66B6: "HANJA",
        0x66B7: "HANJA",
        0x66B8: "HANJA",
        0x66B9: "HANJA",
        0x66BA: "HANJA",
        0x66BB: "HANJA",
        0x66BC: "HANJA",
        0x66BD: "HANJA",
        0x66BE: "HANJA",
        0x66BF: "HANJA",
        0x66C0: "HANJA",
        0x66C1: "HANJA",
        0x66C2: "HANJA",
        0x66C3: "HANJA",
        0x66C4: "HANJA",
        0x66C5: "HANJA",
        0x66C6: "HANJA",
        0x66C7: "HANJA",
        0x66C8: "HANJA",
        0x66C9: "HANJA",
        0x66CA: "HANJA",
        0x66CB: "HANJA",
        0x66CC: "HANJA",
        0x66CD: "HANJA",
        0x66CE: "HANJA",
        0x66CF: "HANJA",
        0x66D0: "HANJA",
        0x66D1: "HANJA",
        0x66D2: "HANJA",
        0x66D3: "HANJA",
        0x66D4: "HANJA",
        0x66D5: "HANJA",
        0x66D6: "HANJA",
        0x66D7: "HANJA",
        0x66D8: "HANJA",
        0x66D9: "HANJA",
        0x66DA: "HANJA",
        0x66DB: "HANJA",
        0x66DC: "HANJA",
        0x66DD: "HANJA",
        0x66DE: "HANJA",
        0x66DF: "HANJA",
        0x66E0: "HANJA",
        0x66E1: "HANJA",
        0x66E2: "HANJA",
        0x66E3: "HANJA",
        0x66E4: "HANJA",
        0x66E5: "HANJA",
        0x66E6: "HANJA",
        0x66E7: "HANJA",
        0x66E8: "HANJA",
        0x66E9: "HANJA",
        0x66EA: "HANJA",
        0x66EB: "HANJA",
        0x66EC: "HANJA",
        0x66ED: "HANJA",
        0x66EE: "HANJA",
        0x66EF: "HANJA",
        0x66F0: "HANJA",
        0x66F1: "HANJA",
        0x66F2: "HANJA",
        0x66F3: "HANJA",
        0x66F4: "HANJA",
        0x66F5: "HANJA",
        0x66F6: "HANJA",
        0x66F7: "HANJA",
        0x66F8: "HANJA",
        0x66F9: "HANJA",
        0x66FA: "HANJA",
        0x66FB: "HANJA",
        0x66FC: "HANJA",
        0x66FD: "HANJA",
        0x66FE: "HANJA",
        0x66FF: "HANJA",
        0x6700: "HANJA",
        0x6701: "HANJA",
        0x6702: "HANJA",
        0x6703: "HANJA",
        0x6704: "HANJA",
        0x6705: "HANJA",
        0x6706: "HANJA",
        0x6707: "HANJA",
        0x6708: "HANJA",
        0x6709: "HANJA",
        0x670A: "HANJA",
        0x670B: "HANJA",
        0x670C: "HANJA",
        0x670D: "HANJA",
        0x670E: "HANJA",
        0x670F: "HANJA",
        0x6710: "HANJA",
        0x6711: "HANJA",
        0x6712: "HANJA",
        0x6713: "HANJA",
        0x6714: "HANJA",
        0x6715: "HANJA",
        0x6716: "HANJA",
        0x6717: "HANJA",
        0x6718: "HANJA",
        0x6719: "HANJA",
        0x671A: "HANJA",
        0x671B: "HANJA",
        0x671C: "HANJA",
        0x671D: "HANJA",
        0x671E: "HANJA",
        0x671F: "HANJA",
        0x6720: "HANJA",
        0x6721: "HANJA",
        0x6722: "HANJA",
        0x6723: "HANJA",
        0x6724: "HANJA",
        0x6725: "HANJA",
        0x6726: "HANJA",
        0x6727: "HANJA",
        0x6728: "HANJA",
        0x6729: "HANJA",
        0x672A: "HANJA",
        0x672B: "HANJA",
        0x672C: "HANJA",
        0x672D: "HANJA",
        0x672E: "HANJA",
        0x672F: "HANJA",
        0x6730: "HANJA",
        0x6731: "HANJA",
        0x6732: "HANJA",
        0x6733: "HANJA",
        0x6734: "HANJA",
        0x6735: "HANJA",
        0x6736: "HANJA",
        0x6737: "HANJA",
        0x6738: "HANJA",
        0x6739: "HANJA",
        0x673A: "HANJA",
        0x673B: "HANJA",
        0x673C: "HANJA",
        0x673D: "HANJA",
        0x673E: "HANJA",
        0x673F: "HANJA",
        0x6740: "HANJA",
        0x6741: "HANJA",
        0x6742: "HANJA",
        0x6743: "HANJA",
        0x6744: "HANJA",
        0x6745: "HANJA",
        0x6746: "HANJA",
        0x6747: "HANJA",
        0x6748: "HANJA",
        0x6749: "HANJA",
        0x674A: "HANJA",
        0x674B: "HANJA",
        0x674C: "HANJA",
        0x674D: "HANJA",
        0x674E: "HANJA",
        0x674F: "HANJA",
        0x6750: "HANJA",
        0x6751: "HANJA",
        0x6752: "HANJA",
        0x6753: "HANJA",
        0x6754: "HANJA",
        0x6755: "HANJA",
        0x6756: "HANJA",
        0x6757: "HANJA",
        0x6758: "HANJA",
        0x6759: "HANJA",
        0x675A: "HANJA",
        0x675B: "HANJA",
        0x675C: "HANJA",
        0x675D: "HANJA",
        0x675E: "HANJA",
        0x675F: "HANJA",
        0x6760: "HANJA",
        0x6761: "HANJA",
        0x6762: "HANJA",
        0x6763: "HANJA",
        0x6764: "HANJA",
        0x6765: "HANJA",
        0x6766: "HANJA",
        0x6767: "HANJA",
        0x6768: "HANJA",
        0x6769: "HANJA",
        0x676A: "HANJA",
        0x676B: "HANJA",
        0x676C: "HANJA",
        0x676D: "HANJA",
        0x676E: "HANJA",
        0x676F: "HANJA",
        0x6770: "HANJA",
        0x6771: "HANJA",
        0x6772: "HANJA",
        0x6773: "HANJA",
        0x6774: "HANJA",
        0x6775: "HANJA",
        0x6776: "HANJA",
        0x6777: "HANJA",
        0x6778: "HANJA",
        0x6779: "HANJA",
        0x677A: "HANJA",
        0x677B: "HANJA",
        0x677C: "HANJA",
        0x677D: "HANJA",
        0x677E: "HANJA",
        0x677F: "HANJA",
        0x6780: "HANJA",
        0x6781: "HANJA",
        0x6782: "HANJA",
        0x6783: "HANJA",
        0x6784: "HANJA",
        0x6785: "HANJA",
        0x6786: "HANJA",
        0x6787: "HANJA",
        0x6788: "HANJA",
        0x6789: "HANJA",
        0x678A: "HANJA",
        0x678B: "HANJA",
        0x678C: "HANJA",
        0x678D: "HANJA",
        0x678E: "HANJA",
        0x678F: "HANJA",
        0x6790: "HANJA",
        0x6791: "HANJA",
        0x6792: "HANJA",
        0x6793: "HANJA",
        0x6794: "HANJA",
        0x6795: "HANJA",
        0x6796: "HANJA",
        0x6797: "HANJA",
        0x6798: "HANJA",
        0x6799: "HANJA",
        0x679A: "HANJA",
        0x679B: "HANJA",
        0x679C: "HANJA",
        0x679D: "HANJA",
        0x679E: "HANJA",
        0x679F: "HANJA",
        0x67A0: "HANJA",
        0x67A1: "HANJA",
        0x67A2: "HANJA",
        0x67A3: "HANJA",
        0x67A4: "HANJA",
        0x67A5: "HANJA",
        0x67A6: "HANJA",
        0x67A7: "HANJA",
        0x67A8: "HANJA",
        0x67A9: "HANJA",
        0x67AA: "HANJA",
        0x67AB: "HANJA",
        0x67AC: "HANJA",
        0x67AD: "HANJA",
        0x67AE: "HANJA",
        0x67AF: "HANJA",
        0x67B0: "HANJA",
        0x67B1: "HANJA",
        0x67B2: "HANJA",
        0x67B3: "HANJA",
        0x67B4: "HANJA",
        0x67B5: "HANJA",
        0x67B6: "HANJA",
        0x67B7: "HANJA",
        0x67B8: "HANJA",
        0x67B9: "HANJA",
        0x67BA: "HANJA",
        0x67BB: "HANJA",
        0x67BC: "HANJA",
        0x67BD: "HANJA",
        0x67BE: "HANJA",
        0x67BF: "HANJA",
        0x67C0: "HANJA",
        0x67C1: "HANJA",
        0x67C2: "HANJA",
        0x67C3: "HANJA",
        0x67C4: "HANJA",
        0x67C5: "HANJA",
        0x67C6: "HANJA",
        0x67C7: "HANJA",
        0x67C8: "HANJA",
        0x67C9: "HANJA",
        0x67CA: "HANJA",
        0x67CB: "HANJA",
        0x67CC: "HANJA",
        0x67CD: "HANJA",
        0x67CE: "HANJA",
        0x67CF: "HANJA",
        0x67D0: "HANJA",
        0x67D1: "HANJA",
        0x67D2: "HANJA",
        0x67D3: "HANJA",
        0x67D4: "HANJA",
        0x67D5: "HANJA",
        0x67D6: "HANJA",
        0x67D7: "HANJA",
        0x67D8: "HANJA",
        0x67D9: "HANJA",
        0x67DA: "HANJA",
        0x67DB: "HANJA",
        0x67DC: "HANJA",
        0x67DD: "HANJA",
        0x67DE: "HANJA",
        0x67DF: "HANJA",
        0x67E0: "HANJA",
        0x67E1: "HANJA",
        0x67E2: "HANJA",
        0x67E3: "HANJA",
        0x67E4: "HANJA",
        0x67E5: "HANJA",
        0x67E6: "HANJA",
        0x67E7: "HANJA",
        0x67E8: "HANJA",
        0x67E9: "HANJA",
        0x67EA: "HANJA",
        0x67EB: "HANJA",
        0x67EC: "HANJA",
        0x67ED: "HANJA",
        0x67EE: "HANJA",
        0x67EF: "HANJA",
        0x67F0: "HANJA",
        0x67F1: "HANJA",
        0x67F2: "HANJA",
        0x67F3: "HANJA",
        0x67F4: "HANJA",
        0x67F5: "HANJA",
        0x67F6: "HANJA",
        0x67F7: "HANJA",
        0x67F8: "HANJA",
        0x67F9: "HANJA",
        0x67FA: "HANJA",
        0x67FB: "HANJA",
        0x67FC: "HANJA",
        0x67FD: "HANJA",
        0x67FE: "HANJA",
        0x67FF: "HANJA",
        0x6800: "HANJA",
        0x6801: "HANJA",
        0x6802: "HANJA",
        0x6803: "HANJA",
        0x6804: "HANJA",
        0x6805: "HANJA",
        0x6806: "HANJA",
        0x6807: "HANJA",
        0x6808: "HANJA",
        0x6809: "HANJA",
        0x680A: "HANJA",
        0x680B: "HANJA",
        0x680C: "HANJA",
        0x680D: "HANJA",
        0x680E: "HANJA",
        0x680F: "HANJA",
        0x6810: "HANJA",
        0x6811: "HANJA",
        0x6812: "HANJA",
        0x6813: "HANJA",
        0x6814: "HANJA",
        0x6815: "HANJA",
        0x6816: "HANJA",
        0x6817: "HANJA",
        0x6818: "HANJA",
        0x6819: "HANJA",
        0x681A: "HANJA",
        0x681B: "HANJA",
        0x681C: "HANJA",
        0x681D: "HANJA",
        0x681E: "HANJA",
        0x681F: "HANJA",
        0x6820: "HANJA",
        0x6821: "HANJA",
        0x6822: "HANJA",
        0x6823: "HANJA",
        0x6824: "HANJA",
        0x6825: "HANJA",
        0x6826: "HANJA",
        0x6827: "HANJA",
        0x6828: "HANJA",
        0x6829: "HANJA",
        0x682A: "HANJA",
        0x682B: "HANJA",
        0x682C: "HANJA",
        0x682D: "HANJA",
        0x682E: "HANJA",
        0x682F: "HANJA",
        0x6830: "HANJA",
        0x6831: "HANJA",
        0x6832: "HANJA",
        0x6833: "HANJA",
        0x6834: "HANJA",
        0x6835: "HANJA",
        0x6836: "HANJA",
        0x6837: "HANJA",
        0x6838: "HANJA",
        0x6839: "HANJA",
        0x683A: "HANJA",
        0x683B: "HANJA",
        0x683C: "HANJA",
        0x683D: "HANJA",
        0x683E: "HANJA",
        0x683F: "HANJA",
        0x6840: "HANJA",
        0x6841: "HANJA",
        0x6842: "HANJA",
        0x6843: "HANJA",
        0x6844: "HANJA",
        0x6845: "HANJA",
        0x6846: "HANJA",
        0x6847: "HANJA",
        0x6848: "HANJA",
        0x6849: "HANJA",
        0x684A: "HANJA",
        0x684B: "HANJA",
        0x684C: "HANJA",
        0x684D: "HANJA",
        0x684E: "HANJA",
        0x684F: "HANJA",
        0x6850: "HANJA",
        0x6851: "HANJA",
        0x6852: "HANJA",
        0x6853: "HANJA",
        0x6854: "HANJA",
        0x6855: "HANJA",
        0x6856: "HANJA",
        0x6857: "HANJA",
        0x6858: "HANJA",
        0x6859: "HANJA",
        0x685A: "HANJA",
        0x685B: "HANJA",
        0x685C: "HANJA",
        0x685D: "HANJA",
        0x685E: "HANJA",
        0x685F: "HANJA",
        0x6860: "HANJA",
        0x6861: "HANJA",
        0x6862: "HANJA",
        0x6863: "HANJA",
        0x6864: "HANJA",
        0x6865: "HANJA",
        0x6866: "HANJA",
        0x6867: "HANJA",
        0x6868: "HANJA",
        0x6869: "HANJA",
        0x686A: "HANJA",
        0x686B: "HANJA",
        0x686C: "HANJA",
        0x686D: "HANJA",
        0x686E: "HANJA",
        0x686F: "HANJA",
        0x6870: "HANJA",
        0x6871: "HANJA",
        0x6872: "HANJA",
        0x6873: "HANJA",
        0x6874: "HANJA",
        0x6875: "HANJA",
        0x6876: "HANJA",
        0x6877: "HANJA",
        0x6878: "HANJA",
        0x6879: "HANJA",
        0x687A: "HANJA",
        0x687B: "HANJA",
        0x687C: "HANJA",
        0x687D: "HANJA",
        0x687E: "HANJA",
        0x687F: "HANJA",
        0x6880: "HANJA",
        0x6881: "HANJA",
        0x6882: "HANJA",
        0x6883: "HANJA",
        0x6884: "HANJA",
        0x6885: "HANJA",
        0x6886: "HANJA",
        0x6887: "HANJA",
        0x6888: "HANJA",
        0x6889: "HANJA",
        0x688A: "HANJA",
        0x688B: "HANJA",
        0x688C: "HANJA",
        0x688D: "HANJA",
        0x688E: "HANJA",
        0x688F: "HANJA",
        0x6890: "HANJA",
        0x6891: "HANJA",
        0x6892: "HANJA",
        0x6893: "HANJA",
        0x6894: "HANJA",
        0x6895: "HANJA",
        0x6896: "HANJA",
        0x6897: "HANJA",
        0x6898: "HANJA",
        0x6899: "HANJA",
        0x689A: "HANJA",
        0x689B: "HANJA",
        0x689C: "HANJA",
        0x689D: "HANJA",
        0x689E: "HANJA",
        0x689F: "HANJA",
        0x68A0: "HANJA",
        0x68A1: "HANJA",
        0x68A2: "HANJA",
        0x68A3: "HANJA",
        0x68A4: "HANJA",
        0x68A5: "HANJA",
        0x68A6: "HANJA",
        0x68A7: "HANJA",
        0x68A8: "HANJA",
        0x68A9: "HANJA",
        0x68AA: "HANJA",
        0x68AB: "HANJA",
        0x68AC: "HANJA",
        0x68AD: "HANJA",
        0x68AE: "HANJA",
        0x68AF: "HANJA",
        0x68B0: "HANJA",
        0x68B1: "HANJA",
        0x68B2: "HANJA",
        0x68B3: "HANJA",
        0x68B4: "HANJA",
        0x68B5: "HANJA",
        0x68B6: "HANJA",
        0x68B7: "HANJA",
        0x68B8: "HANJA",
        0x68B9: "HANJA",
        0x68BA: "HANJA",
        0x68BB: "HANJA",
        0x68BC: "HANJA",
        0x68BD: "HANJA",
        0x68BE: "HANJA",
        0x68BF: "HANJA",
        0x68C0: "HANJA",
        0x68C1: "HANJA",
        0x68C2: "HANJA",
        0x68C3: "HANJA",
        0x68C4: "HANJA",
        0x68C5: "HANJA",
        0x68C6: "HANJA",
        0x68C7: "HANJA",
        0x68C8: "HANJA",
        0x68C9: "HANJA",
        0x68CA: "HANJA",
        0x68CB: "HANJA",
        0x68CC: "HANJA",
        0x68CD: "HANJA",
        0x68CE: "HANJA",
        0x68CF: "HANJA",
        0x68D0: "HANJA",
        0x68D1: "HANJA",
        0x68D2: "HANJA",
        0x68D3: "HANJA",
        0x68D4: "HANJA",
        0x68D5: "HANJA",
        0x68D6: "HANJA",
        0x68D7: "HANJA",
        0x68D8: "HANJA",
        0x68D9: "HANJA",
        0x68DA: "HANJA",
        0x68DB: "HANJA",
        0x68DC: "HANJA",
        0x68DD: "HANJA",
        0x68DE: "HANJA",
        0x68DF: "HANJA",
        0x68E0: "HANJA",
        0x68E1: "HANJA",
        0x68E2: "HANJA",
        0x68E3: "HANJA",
        0x68E4: "HANJA",
        0x68E5: "HANJA",
        0x68E6: "HANJA",
        0x68E7: "HANJA",
        0x68E8: "HANJA",
        0x68E9: "HANJA",
        0x68EA: "HANJA",
        0x68EB: "HANJA",
        0x68EC: "HANJA",
        0x68ED: "HANJA",
        0x68EE: "HANJA",
        0x68EF: "HANJA",
        0x68F0: "HANJA",
        0x68F1: "HANJA",
        0x68F2: "HANJA",
        0x68F3: "HANJA",
        0x68F4: "HANJA",
        0x68F5: "HANJA",
        0x68F6: "HANJA",
        0x68F7: "HANJA",
        0x68F8: "HANJA",
        0x68F9: "HANJA",
        0x68FA: "HANJA",
        0x68FB: "HANJA",
        0x68FC: "HANJA",
        0x68FD: "HANJA",
        0x68FE: "HANJA",
        0x68FF: "HANJA",
        0x6900: "HANJA",
        0x6901: "HANJA",
        0x6902: "HANJA",
        0x6903: "HANJA",
        0x6904: "HANJA",
        0x6905: "HANJA",
        0x6906: "HANJA",
        0x6907: "HANJA",
        0x6908: "HANJA",
        0x6909: "HANJA",
        0x690A: "HANJA",
        0x690B: "HANJA",
        0x690C: "HANJA",
        0x690D: "HANJA",
        0x690E: "HANJA",
        0x690F: "HANJA",
        0x6910: "HANJA",
        0x6911: "HANJA",
        0x6912: "HANJA",
        0x6913: "HANJA",
        0x6914: "HANJA",
        0x6915: "HANJA",
        0x6916: "HANJA",
        0x6917: "HANJA",
        0x6918: "HANJA",
        0x6919: "HANJA",
        0x691A: "HANJA",
        0x691B: "HANJA",
        0x691C: "HANJA",
        0x691D: "HANJA",
        0x691E: "HANJA",
        0x691F: "HANJA",
        0x6920: "HANJA",
        0x6921: "HANJA",
        0x6922: "HANJA",
        0x6923: "HANJA",
        0x6924: "HANJA",
        0x6925: "HANJA",
        0x6926: "HANJA",
        0x6927: "HANJA",
        0x6928: "HANJA",
        0x6929: "HANJA",
        0x692A: "HANJA",
        0x692B: "HANJA",
        0x692C: "HANJA",
        0x692D: "HANJA",
        0x692E: "HANJA",
        0x692F: "HANJA",
        0x6930: "HANJA",
        0x6931: "HANJA",
        0x6932: "HANJA",
        0x6933: "HANJA",
        0x6934: "HANJA",
        0x6935: "HANJA",
        0x6936: "HANJA",
        0x6937: "HANJA",
        0x6938: "HANJA",
        0x6939: "HANJA",
        0x693A: "HANJA",
        0x693B: "HANJA",
        0x693C: "HANJA",
        0x693D: "HANJA",
        0x693E: "HANJA",
        0x693F: "HANJA",
        0x6940: "HANJA",
        0x6941: "HANJA",
        0x6942: "HANJA",
        0x6943: "HANJA",
        0x6944: "HANJA",
        0x6945: "HANJA",
        0x6946: "HANJA",
        0x6947: "HANJA",
        0x6948: "HANJA",
        0x6949: "HANJA",
        0x694A: "HANJA",
        0x694B: "HANJA",
        0x694C: "HANJA",
        0x694D: "HANJA",
        0x694E: "HANJA",
        0x694F: "HANJA",
        0x6950: "HANJA",
        0x6951: "HANJA",
        0x6952: "HANJA",
        0x6953: "HANJA",
        0x6954: "HANJA",
        0x6955: "HANJA",
        0x6956: "HANJA",
        0x6957: "HANJA",
        0x6958: "HANJA",
        0x6959: "HANJA",
        0x695A: "HANJA",
        0x695B: "HANJA",
        0x695C: "HANJA",
        0x695D: "HANJA",
        0x695E: "HANJA",
        0x695F: "HANJA",
        0x6960: "HANJA",
        0x6961: "HANJA",
        0x6962: "HANJA",
        0x6963: "HANJA",
        0x6964: "HANJA",
        0x6965: "HANJA",
        0x6966: "HANJA",
        0x6967: "HANJA",
        0x6968: "HANJA",
        0x6969: "HANJA",
        0x696A: "HANJA",
        0x696B: "HANJA",
        0x696C: "HANJA",
        0x696D: "HANJA",
        0x696E: "HANJA",
        0x696F: "HANJA",
        0x6970: "HANJA",
        0x6971: "HANJA",
        0x6972: "HANJA",
        0x6973: "HANJA",
        0x6974: "HANJA",
        0x6975: "HANJA",
        0x6976: "HANJA",
        0x6977: "HANJA",
        0x6978: "HANJA",
        0x6979: "HANJA",
        0x697A: "HANJA",
        0x697B: "HANJA",
        0x697C: "HANJA",
        0x697D: "HANJA",
        0x697E: "HANJA",
        0x697F: "HANJA",
        0x6980: "HANJA",
        0x6981: "HANJA",
        0x6982: "HANJA",
        0x6983: "HANJA",
        0x6984: "HANJA",
        0x6985: "HANJA",
        0x6986: "HANJA",
        0x6987: "HANJA",
        0x6988: "HANJA",
        0x6989: "HANJA",
        0x698A: "HANJA",
        0x698B: "HANJA",
        0x698C: "HANJA",
        0x698D: "HANJA",
        0x698E: "HANJA",
        0x698F: "HANJA",
        0x6990: "HANJA",
        0x6991: "HANJA",
        0x6992: "HANJA",
        0x6993: "HANJA",
        0x6994: "HANJA",
        0x6995: "HANJA",
        0x6996: "HANJA",
        0x6997: "HANJA",
        0x6998: "HANJA",
        0x6999: "HANJA",
        0x699A: "HANJA",
        0x699B: "HANJA",
        0x699C: "HANJA",
        0x699D: "HANJA",
        0x699E: "HANJA",
        0x699F: "HANJA",
        0x69A0: "HANJA",
        0x69A1: "HANJA",
        0x69A2: "HANJA",
        0x69A3: "HANJA",
        0x69A4: "HANJA",
        0x69A5: "HANJA",
        0x69A6: "HANJA",
        0x69A7: "HANJA",
        0x69A8: "HANJA",
        0x69A9: "HANJA",
        0x69AA: "HANJA",
        0x69AB: "HANJA",
        0x69AC: "HANJA",
        0x69AD: "HANJA",
        0x69AE: "HANJA",
        0x69AF: "HANJA",
        0x69B0: "HANJA",
        0x69B1: "HANJA",
        0x69B2: "HANJA",
        0x69B3: "HANJA",
        0x69B4: "HANJA",
        0x69B5: "HANJA",
        0x69B6: "HANJA",
        0x69B7: "HANJA",
        0x69B8: "HANJA",
        0x69B9: "HANJA",
        0x69BA: "HANJA",
        0x69BB: "HANJA",
        0x69BC: "HANJA",
        0x69BD: "HANJA",
        0x69BE: "HANJA",
        0x69BF: "HANJA",
        0x69C0: "HANJA",
        0x69C1: "HANJA",
        0x69C2: "HANJA",
        0x69C3: "HANJA",
        0x69C4: "HANJA",
        0x69C5: "HANJA",
        0x69C6: "HANJA",
        0x69C7: "HANJA",
        0x69C8: "HANJA",
        0x69C9: "HANJA",
        0x69CA: "HANJA",
        0x69CB: "HANJA",
        0x69CC: "HANJA",
        0x69CD: "HANJA",
        0x69CE: "HANJA",
        0x69CF: "HANJA",
        0x69D0: "HANJA",
        0x69D1: "HANJA",
        0x69D2: "HANJA",
        0x69D3: "HANJA",
        0x69D4: "HANJA",
        0x69D5: "HANJA",
        0x69D6: "HANJA",
        0x69D7: "HANJA",
        0x69D8: "HANJA",
        0x69D9: "HANJA",
        0x69DA: "HANJA",
        0x69DB: "HANJA",
        0x69DC: "HANJA",
        0x69DD: "HANJA",
        0x69DE: "HANJA",
        0x69DF: "HANJA",
        0x69E0: "HANJA",
        0x69E1: "HANJA",
        0x69E2: "HANJA",
        0x69E3: "HANJA",
        0x69E4: "HANJA",
        0x69E5: "HANJA",
        0x69E6: "HANJA",
        0x69E7: "HANJA",
        0x69E8: "HANJA",
        0x69E9: "HANJA",
        0x69EA: "HANJA",
        0x69EB: "HANJA",
        0x69EC: "HANJA",
        0x69ED: "HANJA",
        0x69EE: "HANJA",
        0x69EF: "HANJA",
        0x69F0: "HANJA",
        0x69F1: "HANJA",
        0x69F2: "HANJA",
        0x69F3: "HANJA",
        0x69F4: "HANJA",
        0x69F5: "HANJA",
        0x69F6: "HANJA",
        0x69F7: "HANJA",
        0x69F8: "HANJA",
        0x69F9: "HANJA",
        0x69FA: "HANJA",
        0x69FB: "HANJA",
        0x69FC: "HANJA",
        0x69FD: "HANJA",
        0x69FE: "HANJA",
        0x69FF: "HANJA",
        0x6A00: "HANJA",
        0x6A01: "HANJA",
        0x6A02: "HANJA",
        0x6A03: "HANJA",
        0x6A04: "HANJA",
        0x6A05: "HANJA",
        0x6A06: "HANJA",
        0x6A07: "HANJA",
        0x6A08: "HANJA",
        0x6A09: "HANJA",
        0x6A0A: "HANJA",
        0x6A0B: "HANJA",
        0x6A0C: "HANJA",
        0x6A0D: "HANJA",
        0x6A0E: "HANJA",
        0x6A0F: "HANJA",
        0x6A10: "HANJA",
        0x6A11: "HANJA",
        0x6A12: "HANJA",
        0x6A13: "HANJA",
        0x6A14: "HANJA",
        0x6A15: "HANJA",
        0x6A16: "HANJA",
        0x6A17: "HANJA",
        0x6A18: "HANJA",
        0x6A19: "HANJA",
        0x6A1A: "HANJA",
        0x6A1B: "HANJA",
        0x6A1C: "HANJA",
        0x6A1D: "HANJA",
        0x6A1E: "HANJA",
        0x6A1F: "HANJA",
        0x6A20: "HANJA",
        0x6A21: "HANJA",
        0x6A22: "HANJA",
        0x6A23: "HANJA",
        0x6A24: "HANJA",
        0x6A25: "HANJA",
        0x6A26: "HANJA",
        0x6A27: "HANJA",
        0x6A28: "HANJA",
        0x6A29: "HANJA",
        0x6A2A: "HANJA",
        0x6A2B: "HANJA",
        0x6A2C: "HANJA",
        0x6A2D: "HANJA",
        0x6A2E: "HANJA",
        0x6A2F: "HANJA",
        0x6A30: "HANJA",
        0x6A31: "HANJA",
        0x6A32: "HANJA",
        0x6A33: "HANJA",
        0x6A34: "HANJA",
        0x6A35: "HANJA",
        0x6A36: "HANJA",
        0x6A37: "HANJA",
        0x6A38: "HANJA",
        0x6A39: "HANJA",
        0x6A3A: "HANJA",
        0x6A3B: "HANJA",
        0x6A3C: "HANJA",
        0x6A3D: "HANJA",
        0x6A3E: "HANJA",
        0x6A3F: "HANJA",
        0x6A40: "HANJA",
        0x6A41: "HANJA",
        0x6A42: "HANJA",
        0x6A43: "HANJA",
        0x6A44: "HANJA",
        0x6A45: "HANJA",
        0x6A46: "HANJA",
        0x6A47: "HANJA",
        0x6A48: "HANJA",
        0x6A49: "HANJA",
        0x6A4A: "HANJA",
        0x6A4B: "HANJA",
        0x6A4C: "HANJA",
        0x6A4D: "HANJA",
        0x6A4E: "HANJA",
        0x6A4F: "HANJA",
        0x6A50: "HANJA",
        0x6A51: "HANJA",
        0x6A52: "HANJA",
        0x6A53: "HANJA",
        0x6A54: "HANJA",
        0x6A55: "HANJA",
        0x6A56: "HANJA",
        0x6A57: "HANJA",
        0x6A58: "HANJA",
        0x6A59: "HANJA",
        0x6A5A: "HANJA",
        0x6A5B: "HANJA",
        0x6A5C: "HANJA",
        0x6A5D: "HANJA",
        0x6A5E: "HANJA",
        0x6A5F: "HANJA",
        0x6A60: "HANJA",
        0x6A61: "HANJA",
        0x6A62: "HANJA",
        0x6A63: "HANJA",
        0x6A64: "HANJA",
        0x6A65: "HANJA",
        0x6A66: "HANJA",
        0x6A67: "HANJA",
        0x6A68: "HANJA",
        0x6A69: "HANJA",
        0x6A6A: "HANJA",
        0x6A6B: "HANJA",
        0x6A6C: "HANJA",
        0x6A6D: "HANJA",
        0x6A6E: "HANJA",
        0x6A6F: "HANJA",
        0x6A70: "HANJA",
        0x6A71: "HANJA",
        0x6A72: "HANJA",
        0x6A73: "HANJA",
        0x6A74: "HANJA",
        0x6A75: "HANJA",
        0x6A76: "HANJA",
        0x6A77: "HANJA",
        0x6A78: "HANJA",
        0x6A79: "HANJA",
        0x6A7A: "HANJA",
        0x6A7B: "HANJA",
        0x6A7C: "HANJA",
        0x6A7D: "HANJA",
        0x6A7E: "HANJA",
        0x6A7F: "HANJA",
        0x6A80: "HANJA",
        0x6A81: "HANJA",
        0x6A82: "HANJA",
        0x6A83: "HANJA",
        0x6A84: "HANJA",
        0x6A85: "HANJA",
        0x6A86: "HANJA",
        0x6A87: "HANJA",
        0x6A88: "HANJA",
        0x6A89: "HANJA",
        0x6A8A: "HANJA",
        0x6A8B: "HANJA",
        0x6A8C: "HANJA",
        0x6A8D: "HANJA",
        0x6A8E: "HANJA",
        0x6A8F: "HANJA",
        0x6A90: "HANJA",
        0x6A91: "HANJA",
        0x6A92: "HANJA",
        0x6A93: "HANJA",
        0x6A94: "HANJA",
        0x6A95: "HANJA",
        0x6A96: "HANJA",
        0x6A97: "HANJA",
        0x6A98: "HANJA",
        0x6A99: "HANJA",
        0x6A9A: "HANJA",
        0x6A9B: "HANJA",
        0x6A9C: "HANJA",
        0x6A9D: "HANJA",
        0x6A9E: "HANJA",
        0x6A9F: "HANJA",
        0x6AA0: "HANJA",
        0x6AA1: "HANJA",
        0x6AA2: "HANJA",
        0x6AA3: "HANJA",
        0x6AA4: "HANJA",
        0x6AA5: "HANJA",
        0x6AA6: "HANJA",
        0x6AA7: "HANJA",
        0x6AA8: "HANJA",
        0x6AA9: "HANJA",
        0x6AAA: "HANJA",
        0x6AAB: "HANJA",
        0x6AAC: "HANJA",
        0x6AAD: "HANJA",
        0x6AAE: "HANJA",
        0x6AAF: "HANJA",
        0x6AB0: "HANJA",
        0x6AB1: "HANJA",
        0x6AB2: "HANJA",
        0x6AB3: "HANJA",
        0x6AB4: "HANJA",
        0x6AB5: "HANJA",
        0x6AB6: "HANJA",
        0x6AB7: "HANJA",
        0x6AB8: "HANJA",
        0x6AB9: "HANJA",
        0x6ABA: "HANJA",
        0x6ABB: "HANJA",
        0x6ABC: "HANJA",
        0x6ABD: "HANJA",
        0x6ABE: "HANJA",
        0x6ABF: "HANJA",
        0x6AC0: "HANJA",
        0x6AC1: "HANJA",
        0x6AC2: "HANJA",
        0x6AC3: "HANJA",
        0x6AC4: "HANJA",
        0x6AC5: "HANJA",
        0x6AC6: "HANJA",
        0x6AC7: "HANJA",
        0x6AC8: "HANJA",
        0x6AC9: "HANJA",
        0x6ACA: "HANJA",
        0x6ACB: "HANJA",
        0x6ACC: "HANJA",
        0x6ACD: "HANJA",
        0x6ACE: "HANJA",
        0x6ACF: "HANJA",
        0x6AD0: "HANJA",
        0x6AD1: "HANJA",
        0x6AD2: "HANJA",
        0x6AD3: "HANJA",
        0x6AD4: "HANJA",
        0x6AD5: "HANJA",
        0x6AD6: "HANJA",
        0x6AD7: "HANJA",
        0x6AD8: "HANJA",
        0x6AD9: "HANJA",
        0x6ADA: "HANJA",
        0x6ADB: "HANJA",
        0x6ADC: "HANJA",
        0x6ADD: "HANJA",
        0x6ADE: "HANJA",
        0x6ADF: "HANJA",
        0x6AE0: "HANJA",
        0x6AE1: "HANJA",
        0x6AE2: "HANJA",
        0x6AE3: "HANJA",
        0x6AE4: "HANJA",
        0x6AE5: "HANJA",
        0x6AE6: "HANJA",
        0x6AE7: "HANJA",
        0x6AE8: "HANJA",
        0x6AE9: "HANJA",
        0x6AEA: "HANJA",
        0x6AEB: "HANJA",
        0x6AEC: "HANJA",
        0x6AED: "HANJA",
        0x6AEE: "HANJA",
        0x6AEF: "HANJA",
        0x6AF0: "HANJA",
        0x6AF1: "HANJA",
        0x6AF2: "HANJA",
        0x6AF3: "HANJA",
        0x6AF4: "HANJA",
        0x6AF5: "HANJA",
        0x6AF6: "HANJA",
        0x6AF7: "HANJA",
        0x6AF8: "HANJA",
        0x6AF9: "HANJA",
        0x6AFA: "HANJA",
        0x6AFB: "HANJA",
        0x6AFC: "HANJA",
        0x6AFD: "HANJA",
        0x6AFE: "HANJA",
        0x6AFF: "HANJA",
        0x6B00: "HANJA",
        0x6B01: "HANJA",
        0x6B02: "HANJA",
        0x6B03: "HANJA",
        0x6B04: "HANJA",
        0x6B05: "HANJA",
        0x6B06: "HANJA",
        0x6B07: "HANJA",
        0x6B08: "HANJA",
        0x6B09: "HANJA",
        0x6B0A: "HANJA",
        0x6B0B: "HANJA",
        0x6B0C: "HANJA",
        0x6B0D: "HANJA",
        0x6B0E: "HANJA",
        0x6B0F: "HANJA",
        0x6B10: "HANJA",
        0x6B11: "HANJA",
        0x6B12: "HANJA",
        0x6B13: "HANJA",
        0x6B14: "HANJA",
        0x6B15: "HANJA",
        0x6B16: "HANJA",
        0x6B17: "HANJA",
        0x6B18: "HANJA",
        0x6B19: "HANJA",
        0x6B1A: "HANJA",
        0x6B1B: "HANJA",
        0x6B1C: "HANJA",
        0x6B1D: "HANJA",
        0x6B1E: "HANJA",
        0x6B1F: "HANJA",
        0x6B20: "HANJA",
        0x6B21: "HANJA",
        0x6B22: "HANJA",
        0x6B23: "HANJA",
        0x6B24: "HANJA",
        0x6B25: "HANJA",
        0x6B26: "HANJA",
        0x6B27: "HANJA",
        0x6B28: "HANJA",
        0x6B29: "HANJA",
        0x6B2A: "HANJA",
        0x6B2B: "HANJA",
        0x6B2C: "HANJA",
        0x6B2D: "HANJA",
        0x6B2E: "HANJA",
        0x6B2F: "HANJA",
        0x6B30: "HANJA",
        0x6B31: "HANJA",
        0x6B32: "HANJA",
        0x6B33: "HANJA",
        0x6B34: "HANJA",
        0x6B35: "HANJA",
        0x6B36: "HANJA",
        0x6B37: "HANJA",
        0x6B38: "HANJA",
        0x6B39: "HANJA",
        0x6B3A: "HANJA",
        0x6B3B: "HANJA",
        0x6B3C: "HANJA",
        0x6B3D: "HANJA",
        0x6B3E: "HANJA",
        0x6B3F: "HANJA",
        0x6B40: "HANJA",
        0x6B41: "HANJA",
        0x6B42: "HANJA",
        0x6B43: "HANJA",
        0x6B44: "HANJA",
        0x6B45: "HANJA",
        0x6B46: "HANJA",
        0x6B47: "HANJA",
        0x6B48: "HANJA",
        0x6B49: "HANJA",
        0x6B4A: "HANJA",
        0x6B4B: "HANJA",
        0x6B4C: "HANJA",
        0x6B4D: "HANJA",
        0x6B4E: "HANJA",
        0x6B4F: "HANJA",
        0x6B50: "HANJA",
        0x6B51: "HANJA",
        0x6B52: "HANJA",
        0x6B53: "HANJA",
        0x6B54: "HANJA",
        0x6B55: "HANJA",
        0x6B56: "HANJA",
        0x6B57: "HANJA",
        0x6B58: "HANJA",
        0x6B59: "HANJA",
        0x6B5A: "HANJA",
        0x6B5B: "HANJA",
        0x6B5C: "HANJA",
        0x6B5D: "HANJA",
        0x6B5E: "HANJA",
        0x6B5F: "HANJA",
        0x6B60: "HANJA",
        0x6B61: "HANJA",
        0x6B62: "HANJA",
        0x6B63: "HANJA",
        0x6B64: "HANJA",
        0x6B65: "HANJA",
        0x6B66: "HANJA",
        0x6B67: "HANJA",
        0x6B68: "HANJA",
        0x6B69: "HANJA",
        0x6B6A: "HANJA",
        0x6B6B: "HANJA",
        0x6B6C: "HANJA",
        0x6B6D: "HANJA",
        0x6B6E: "HANJA",
        0x6B6F: "HANJA",
        0x6B70: "HANJA",
        0x6B71: "HANJA",
        0x6B72: "HANJA",
        0x6B73: "HANJA",
        0x6B74: "HANJA",
        0x6B75: "HANJA",
        0x6B76: "HANJA",
        0x6B77: "HANJA",
        0x6B78: "HANJA",
        0x6B79: "HANJA",
        0x6B7A: "HANJA",
        0x6B7B: "HANJA",
        0x6B7C: "HANJA",
        0x6B7D: "HANJA",
        0x6B7E: "HANJA",
        0x6B7F: "HANJA",
        0x6B80: "HANJA",
        0x6B81: "HANJA",
        0x6B82: "HANJA",
        0x6B83: "HANJA",
        0x6B84: "HANJA",
        0x6B85: "HANJA",
        0x6B86: "HANJA",
        0x6B87: "HANJA",
        0x6B88: "HANJA",
        0x6B89: "HANJA",
        0x6B8A: "HANJA",
        0x6B8B: "HANJA",
        0x6B8C: "HANJA",
        0x6B8D: "HANJA",
        0x6B8E: "HANJA",
        0x6B8F: "HANJA",
        0x6B90: "HANJA",
        0x6B91: "HANJA",
        0x6B92: "HANJA",
        0x6B93: "HANJA",
        0x6B94: "HANJA",
        0x6B95: "HANJA",
        0x6B96: "HANJA",
        0x6B97: "HANJA",
        0x6B98: "HANJA",
        0x6B99: "HANJA",
        0x6B9A: "HANJA",
        0x6B9B: "HANJA",
        0x6B9C: "HANJA",
        0x6B9D: "HANJA",
        0x6B9E: "HANJA",
        0x6B9F: "HANJA",
        0x6BA0: "HANJA",
        0x6BA1: "HANJA",
        0x6BA2: "HANJA",
        0x6BA3: "HANJA",
        0x6BA4: "HANJA",
        0x6BA5: "HANJA",
        0x6BA6: "HANJA",
        0x6BA7: "HANJA",
        0x6BA8: "HANJA",
        0x6BA9: "HANJA",
        0x6BAA: "HANJA",
        0x6BAB: "HANJA",
        0x6BAC: "HANJA",
        0x6BAD: "HANJA",
        0x6BAE: "HANJA",
        0x6BAF: "HANJA",
        0x6BB0: "HANJA",
        0x6BB1: "HANJA",
        0x6BB2: "HANJA",
        0x6BB3: "HANJA",
        0x6BB4: "HANJA",
        0x6BB5: "HANJA",
        0x6BB6: "HANJA",
        0x6BB7: "HANJA",
        0x6BB8: "HANJA",
        0x6BB9: "HANJA",
        0x6BBA: "HANJA",
        0x6BBB: "HANJA",
        0x6BBC: "HANJA",
        0x6BBD: "HANJA",
        0x6BBE: "HANJA",
        0x6BBF: "HANJA",
        0x6BC0: "HANJA",
        0x6BC1: "HANJA",
        0x6BC2: "HANJA",
        0x6BC3: "HANJA",
        0x6BC4: "HANJA",
        0x6BC5: "HANJA",
        0x6BC6: "HANJA",
        0x6BC7: "HANJA",
        0x6BC8: "HANJA",
        0x6BC9: "HANJA",
        0x6BCA: "HANJA",
        0x6BCB: "HANJA",
        0x6BCC: "HANJA",
        0x6BCD: "HANJA",
        0x6BCE: "HANJA",
        0x6BCF: "HANJA",
        0x6BD0: "HANJA",
        0x6BD1: "HANJA",
        0x6BD2: "HANJA",
        0x6BD3: "HANJA",
        0x6BD4: "HANJA",
        0x6BD5: "HANJA",
        0x6BD6: "HANJA",
        0x6BD7: "HANJA",
        0x6BD8: "HANJA",
        0x6BD9: "HANJA",
        0x6BDA: "HANJA",
        0x6BDB: "HANJA",
        0x6BDC: "HANJA",
        0x6BDD: "HANJA",
        0x6BDE: "HANJA",
        0x6BDF: "HANJA",
        0x6BE0: "HANJA",
        0x6BE1: "HANJA",
        0x6BE2: "HANJA",
        0x6BE3: "HANJA",
        0x6BE4: "HANJA",
        0x6BE5: "HANJA",
        0x6BE6: "HANJA",
        0x6BE7: "HANJA",
        0x6BE8: "HANJA",
        0x6BE9: "HANJA",
        0x6BEA: "HANJA",
        0x6BEB: "HANJA",
        0x6BEC: "HANJA",
        0x6BED: "HANJA",
        0x6BEE: "HANJA",
        0x6BEF: "HANJA",
        0x6BF0: "HANJA",
        0x6BF1: "HANJA",
        0x6BF2: "HANJA",
        0x6BF3: "HANJA",
        0x6BF4: "HANJA",
        0x6BF5: "HANJA",
        0x6BF6: "HANJA",
        0x6BF7: "HANJA",
        0x6BF8: "HANJA",
        0x6BF9: "HANJA",
        0x6BFA: "HANJA",
        0x6BFB: "HANJA",
        0x6BFC: "HANJA",
        0x6BFD: "HANJA",
        0x6BFE: "HANJA",
        0x6BFF: "HANJA",
        0x6C00: "HANJA",
        0x6C01: "HANJA",
        0x6C02: "HANJA",
        0x6C03: "HANJA",
        0x6C04: "HANJA",
        0x6C05: "HANJA",
        0x6C06: "HANJA",
        0x6C07: "HANJA",
        0x6C08: "HANJA",
        0x6C09: "HANJA",
        0x6C0A: "HANJA",
        0x6C0B: "HANJA",
        0x6C0C: "HANJA",
        0x6C0D: "HANJA",
        0x6C0E: "HANJA",
        0x6C0F: "HANJA",
        0x6C10: "HANJA",
        0x6C11: "HANJA",
        0x6C12: "HANJA",
        0x6C13: "HANJA",
        0x6C14: "HANJA",
        0x6C15: "HANJA",
        0x6C16: "HANJA",
        0x6C17: "HANJA",
        0x6C18: "HANJA",
        0x6C19: "HANJA",
        0x6C1A: "HANJA",
        0x6C1B: "HANJA",
        0x6C1C: "HANJA",
        0x6C1D: "HANJA",
        0x6C1E: "HANJA",
        0x6C1F: "HANJA",
        0x6C20: "HANJA",
        0x6C21: "HANJA",
        0x6C22: "HANJA",
        0x6C23: "HANJA",
        0x6C24: "HANJA",
        0x6C25: "HANJA",
        0x6C26: "HANJA",
        0x6C27: "HANJA",
        0x6C28: "HANJA",
        0x6C29: "HANJA",
        0x6C2A: "HANJA",
        0x6C2B: "HANJA",
        0x6C2C: "HANJA",
        0x6C2D: "HANJA",
        0x6C2E: "HANJA",
        0x6C2F: "HANJA",
        0x6C30: "HANJA",
        0x6C31: "HANJA",
        0x6C32: "HANJA",
        0x6C33: "HANJA",
        0x6C34: "HANJA",
        0x6C35: "HANJA",
        0x6C36: "HANJA",
        0x6C37: "HANJA",
        0x6C38: "HANJA",
        0x6C39: "HANJA",
        0x6C3A: "HANJA",
        0x6C3B: "HANJA",
        0x6C3C: "HANJA",
        0x6C3D: "HANJA",
        0x6C3E: "HANJA",
        0x6C3F: "HANJA",
        0x6C40: "HANJA",
        0x6C41: "HANJA",
        0x6C42: "HANJA",
        0x6C43: "HANJA",
        0x6C44: "HANJA",
        0x6C45: "HANJA",
        0x6C46: "HANJA",
        0x6C47: "HANJA",
        0x6C48: "HANJA",
        0x6C49: "HANJA",
        0x6C4A: "HANJA",
        0x6C4B: "HANJA",
        0x6C4C: "HANJA",
        0x6C4D: "HANJA",
        0x6C4E: "HANJA",
        0x6C4F: "HANJA",
        0x6C50: "HANJA",
        0x6C51: "HANJA",
        0x6C52: "HANJA",
        0x6C53: "HANJA",
        0x6C54: "HANJA",
        0x6C55: "HANJA",
        0x6C56: "HANJA",
        0x6C57: "HANJA",
        0x6C58: "HANJA",
        0x6C59: "HANJA",
        0x6C5A: "HANJA",
        0x6C5B: "HANJA",
        0x6C5C: "HANJA",
        0x6C5D: "HANJA",
        0x6C5E: "HANJA",
        0x6C5F: "HANJA",
        0x6C60: "HANJA",
        0x6C61: "HANJA",
        0x6C62: "HANJA",
        0x6C63: "HANJA",
        0x6C64: "HANJA",
        0x6C65: "HANJA",
        0x6C66: "HANJA",
        0x6C67: "HANJA",
        0x6C68: "HANJA",
        0x6C69: "HANJA",
        0x6C6A: "HANJA",
        0x6C6B: "HANJA",
        0x6C6C: "HANJA",
        0x6C6D: "HANJA",
        0x6C6E: "HANJA",
        0x6C6F: "HANJA",
        0x6C70: "HANJA",
        0x6C71: "HANJA",
        0x6C72: "HANJA",
        0x6C73: "HANJA",
        0x6C74: "HANJA",
        0x6C75: "HANJA",
        0x6C76: "HANJA",
        0x6C77: "HANJA",
        0x6C78: "HANJA",
        0x6C79: "HANJA",
        0x6C7A: "HANJA",
        0x6C7B: "HANJA",
        0x6C7C: "HANJA",
        0x6C7D: "HANJA",
        0x6C7E: "HANJA",
        0x6C7F: "HANJA",
        0x6C80: "HANJA",
        0x6C81: "HANJA",
        0x6C82: "HANJA",
        0x6C83: "HANJA",
        0x6C84: "HANJA",
        0x6C85: "HANJA",
        0x6C86: "HANJA",
        0x6C87: "HANJA",
        0x6C88: "HANJA",
        0x6C89: "HANJA",
        0x6C8A: "HANJA",
        0x6C8B: "HANJA",
        0x6C8C: "HANJA",
        0x6C8D: "HANJA",
        0x6C8E: "HANJA",
        0x6C8F: "HANJA",
        0x6C90: "HANJA",
        0x6C91: "HANJA",
        0x6C92: "HANJA",
        0x6C93: "HANJA",
        0x6C94: "HANJA",
        0x6C95: "HANJA",
        0x6C96: "HANJA",
        0x6C97: "HANJA",
        0x6C98: "HANJA",
        0x6C99: "HANJA",
        0x6C9A: "HANJA",
        0x6C9B: "HANJA",
        0x6C9C: "HANJA",
        0x6C9D: "HANJA",
        0x6C9E: "HANJA",
        0x6C9F: "HANJA",
        0x6CA0: "HANJA",
        0x6CA1: "HANJA",
        0x6CA2: "HANJA",
        0x6CA3: "HANJA",
        0x6CA4: "HANJA",
        0x6CA5: "HANJA",
        0x6CA6: "HANJA",
        0x6CA7: "HANJA",
        0x6CA8: "HANJA",
        0x6CA9: "HANJA",
        0x6CAA: "HANJA",
        0x6CAB: "HANJA",
        0x6CAC: "HANJA",
        0x6CAD: "HANJA",
        0x6CAE: "HANJA",
        0x6CAF: "HANJA",
        0x6CB0: "HANJA",
        0x6CB1: "HANJA",
        0x6CB2: "HANJA",
        0x6CB3: "HANJA",
        0x6CB4: "HANJA",
        0x6CB5: "HANJA",
        0x6CB6: "HANJA",
        0x6CB7: "HANJA",
        0x6CB8: "HANJA",
        0x6CB9: "HANJA",
        0x6CBA: "HANJA",
        0x6CBB: "HANJA",
        0x6CBC: "HANJA",
        0x6CBD: "HANJA",
        0x6CBE: "HANJA",
        0x6CBF: "HANJA",
        0x6CC0: "HANJA",
        0x6CC1: "HANJA",
        0x6CC2: "HANJA",
        0x6CC3: "HANJA",
        0x6CC4: "HANJA",
        0x6CC5: "HANJA",
        0x6CC6: "HANJA",
        0x6CC7: "HANJA",
        0x6CC8: "HANJA",
        0x6CC9: "HANJA",
        0x6CCA: "HANJA",
        0x6CCB: "HANJA",
        0x6CCC: "HANJA",
        0x6CCD: "HANJA",
        0x6CCE: "HANJA",
        0x6CCF: "HANJA",
        0x6CD0: "HANJA",
        0x6CD1: "HANJA",
        0x6CD2: "HANJA",
        0x6CD3: "HANJA",
        0x6CD4: "HANJA",
        0x6CD5: "HANJA",
        0x6CD6: "HANJA",
        0x6CD7: "HANJA",
        0x6CD8: "HANJA",
        0x6CD9: "HANJA",
        0x6CDA: "HANJA",
        0x6CDB: "HANJA",
        0x6CDC: "HANJA",
        0x6CDD: "HANJA",
        0x6CDE: "HANJA",
        0x6CDF: "HANJA",
        0x6CE0: "HANJA",
        0x6CE1: "HANJA",
        0x6CE2: "HANJA",
        0x6CE3: "HANJA",
        0x6CE4: "HANJA",
        0x6CE5: "HANJA",
        0x6CE6: "HANJA",
        0x6CE7: "HANJA",
        0x6CE8: "HANJA",
        0x6CE9: "HANJA",
        0x6CEA: "HANJA",
        0x6CEB: "HANJA",
        0x6CEC: "HANJA",
        0x6CED: "HANJA",
        0x6CEE: "HANJA",
        0x6CEF: "HANJA",
        0x6CF0: "HANJA",
        0x6CF1: "HANJA",
        0x6CF2: "HANJA",
        0x6CF3: "HANJA",
        0x6CF4: "HANJA",
        0x6CF5: "HANJA",
        0x6CF6: "HANJA",
        0x6CF7: "HANJA",
        0x6CF8: "HANJA",
        0x6CF9: "HANJA",
        0x6CFA: "HANJA",
        0x6CFB: "HANJA",
        0x6CFC: "HANJA",
        0x6CFD: "HANJA",
        0x6CFE: "HANJA",
        0x6CFF: "HANJA",
        0x6D00: "HANJA",
        0x6D01: "HANJA",
        0x6D02: "HANJA",
        0x6D03: "HANJA",
        0x6D04: "HANJA",
        0x6D05: "HANJA",
        0x6D06: "HANJA",
        0x6D07: "HANJA",
        0x6D08: "HANJA",
        0x6D09: "HANJA",
        0x6D0A: "HANJA",
        0x6D0B: "HANJA",
        0x6D0C: "HANJA",
        0x6D0D: "HANJA",
        0x6D0E: "HANJA",
        0x6D0F: "HANJA",
        0x6D10: "HANJA",
        0x6D11: "HANJA",
        0x6D12: "HANJA",
        0x6D13: "HANJA",
        0x6D14: "HANJA",
        0x6D15: "HANJA",
        0x6D16: "HANJA",
        0x6D17: "HANJA",
        0x6D18: "HANJA",
        0x6D19: "HANJA",
        0x6D1A: "HANJA",
        0x6D1B: "HANJA",
        0x6D1C: "HANJA",
        0x6D1D: "HANJA",
        0x6D1E: "HANJA",
        0x6D1F: "HANJA",
        0x6D20: "HANJA",
        0x6D21: "HANJA",
        0x6D22: "HANJA",
        0x6D23: "HANJA",
        0x6D24: "HANJA",
        0x6D25: "HANJA",
        0x6D26: "HANJA",
        0x6D27: "HANJA",
        0x6D28: "HANJA",
        0x6D29: "HANJA",
        0x6D2A: "HANJA",
        0x6D2B: "HANJA",
        0x6D2C: "HANJA",
        0x6D2D: "HANJA",
        0x6D2E: "HANJA",
        0x6D2F: "HANJA",
        0x6D30: "HANJA",
        0x6D31: "HANJA",
        0x6D32: "HANJA",
        0x6D33: "HANJA",
        0x6D34: "HANJA",
        0x6D35: "HANJA",
        0x6D36: "HANJA",
        0x6D37: "HANJA",
        0x6D38: "HANJA",
        0x6D39: "HANJA",
        0x6D3A: "HANJA",
        0x6D3B: "HANJA",
        0x6D3C: "HANJA",
        0x6D3D: "HANJA",
        0x6D3E: "HANJA",
        0x6D3F: "HANJA",
        0x6D40: "HANJA",
        0x6D41: "HANJA",
        0x6D42: "HANJA",
        0x6D43: "HANJA",
        0x6D44: "HANJA",
        0x6D45: "HANJA",
        0x6D46: "HANJA",
        0x6D47: "HANJA",
        0x6D48: "HANJA",
        0x6D49: "HANJA",
        0x6D4A: "HANJA",
        0x6D4B: "HANJA",
        0x6D4C: "HANJA",
        0x6D4D: "HANJA",
        0x6D4E: "HANJA",
        0x6D4F: "HANJA",
        0x6D50: "HANJA",
        0x6D51: "HANJA",
        0x6D52: "HANJA",
        0x6D53: "HANJA",
        0x6D54: "HANJA",
        0x6D55: "HANJA",
        0x6D56: "HANJA",
        0x6D57: "HANJA",
        0x6D58: "HANJA",
        0x6D59: "HANJA",
        0x6D5A: "HANJA",
        0x6D5B: "HANJA",
        0x6D5C: "HANJA",
        0x6D5D: "HANJA",
        0x6D5E: "HANJA",
        0x6D5F: "HANJA",
        0x6D60: "HANJA",
        0x6D61: "HANJA",
        0x6D62: "HANJA",
        0x6D63: "HANJA",
        0x6D64: "HANJA",
        0x6D65: "HANJA",
        0x6D66: "HANJA",
        0x6D67: "HANJA",
        0x6D68: "HANJA",
        0x6D69: "HANJA",
        0x6D6A: "HANJA",
        0x6D6B: "HANJA",
        0x6D6C: "HANJA",
        0x6D6D: "HANJA",
        0x6D6E: "HANJA",
        0x6D6F: "HANJA",
        0x6D70: "HANJA",
        0x6D71: "HANJA",
        0x6D72: "HANJA",
        0x6D73: "HANJA",
        0x6D74: "HANJA",
        0x6D75: "HANJA",
        0x6D76: "HANJA",
        0x6D77: "HANJA",
        0x6D78: "HANJA",
        0x6D79: "HANJA",
        0x6D7A: "HANJA",
        0x6D7B: "HANJA",
        0x6D7C: "HANJA",
        0x6D7D: "HANJA",
        0x6D7E: "HANJA",
        0x6D7F: "HANJA",
        0x6D80: "HANJA",
        0x6D81: "HANJA",
        0x6D82: "HANJA",
        0x6D83: "HANJA",
        0x6D84: "HANJA",
        0x6D85: "HANJA",
        0x6D86: "HANJA",
        0x6D87: "HANJA",
        0x6D88: "HANJA",
        0x6D89: "HANJA",
        0x6D8A: "HANJA",
        0x6D8B: "HANJA",
        0x6D8C: "HANJA",
        0x6D8D: "HANJA",
        0x6D8E: "HANJA",
        0x6D8F: "HANJA",
        0x6D90: "HANJA",
        0x6D91: "HANJA",
        0x6D92: "HANJA",
        0x6D93: "HANJA",
        0x6D94: "HANJA",
        0x6D95: "HANJA",
        0x6D96: "HANJA",
        0x6D97: "HANJA",
        0x6D98: "HANJA",
        0x6D99: "HANJA",
        0x6D9A: "HANJA",
        0x6D9B: "HANJA",
        0x6D9C: "HANJA",
        0x6D9D: "HANJA",
        0x6D9E: "HANJA",
        0x6D9F: "HANJA",
        0x6DA0: "HANJA",
        0x6DA1: "HANJA",
        0x6DA2: "HANJA",
        0x6DA3: "HANJA",
        0x6DA4: "HANJA",
        0x6DA5: "HANJA",
        0x6DA6: "HANJA",
        0x6DA7: "HANJA",
        0x6DA8: "HANJA",
        0x6DA9: "HANJA",
        0x6DAA: "HANJA",
        0x6DAB: "HANJA",
        0x6DAC: "HANJA",
        0x6DAD: "HANJA",
        0x6DAE: "HANJA",
        0x6DAF: "HANJA",
        0x6DB0: "HANJA",
        0x6DB1: "HANJA",
        0x6DB2: "HANJA",
        0x6DB3: "HANJA",
        0x6DB4: "HANJA",
        0x6DB5: "HANJA",
        0x6DB6: "HANJA",
        0x6DB7: "HANJA",
        0x6DB8: "HANJA",
        0x6DB9: "HANJA",
        0x6DBA: "HANJA",
        0x6DBB: "HANJA",
        0x6DBC: "HANJA",
        0x6DBD: "HANJA",
        0x6DBE: "HANJA",
        0x6DBF: "HANJA",
        0x6DC0: "HANJA",
        0x6DC1: "HANJA",
        0x6DC2: "HANJA",
        0x6DC3: "HANJA",
        0x6DC4: "HANJA",
        0x6DC5: "HANJA",
        0x6DC6: "HANJA",
        0x6DC7: "HANJA",
        0x6DC8: "HANJA",
        0x6DC9: "HANJA",
        0x6DCA: "HANJA",
        0x6DCB: "HANJA",
        0x6DCC: "HANJA",
        0x6DCD: "HANJA",
        0x6DCE: "HANJA",
        0x6DCF: "HANJA",
        0x6DD0: "HANJA",
        0x6DD1: "HANJA",
        0x6DD2: "HANJA",
        0x6DD3: "HANJA",
        0x6DD4: "HANJA",
        0x6DD5: "HANJA",
        0x6DD6: "HANJA",
        0x6DD7: "HANJA",
        0x6DD8: "HANJA",
        0x6DD9: "HANJA",
        0x6DDA: "HANJA",
        0x6DDB: "HANJA",
        0x6DDC: "HANJA",
        0x6DDD: "HANJA",
        0x6DDE: "HANJA",
        0x6DDF: "HANJA",
        0x6DE0: "HANJA",
        0x6DE1: "HANJA",
        0x6DE2: "HANJA",
        0x6DE3: "HANJA",
        0x6DE4: "HANJA",
        0x6DE5: "HANJA",
        0x6DE6: "HANJA",
        0x6DE7: "HANJA",
        0x6DE8: "HANJA",
        0x6DE9: "HANJA",
        0x6DEA: "HANJA",
        0x6DEB: "HANJA",
        0x6DEC: "HANJA",
        0x6DED: "HANJA",
        0x6DEE: "HANJA",
        0x6DEF: "HANJA",
        0x6DF0: "HANJA",
        0x6DF1: "HANJA",
        0x6DF2: "HANJA",
        0x6DF3: "HANJA",
        0x6DF4: "HANJA",
        0x6DF5: "HANJA",
        0x6DF6: "HANJA",
        0x6DF7: "HANJA",
        0x6DF8: "HANJA",
        0x6DF9: "HANJA",
        0x6DFA: "HANJA",
        0x6DFB: "HANJA",
        0x6DFC: "HANJA",
        0x6DFD: "HANJA",
        0x6DFE: "HANJA",
        0x6DFF: "HANJA",
        0x6E00: "HANJA",
        0x6E01: "HANJA",
        0x6E02: "HANJA",
        0x6E03: "HANJA",
        0x6E04: "HANJA",
        0x6E05: "HANJA",
        0x6E06: "HANJA",
        0x6E07: "HANJA",
        0x6E08: "HANJA",
        0x6E09: "HANJA",
        0x6E0A: "HANJA",
        0x6E0B: "HANJA",
        0x6E0C: "HANJA",
        0x6E0D: "HANJA",
        0x6E0E: "HANJA",
        0x6E0F: "HANJA",
        0x6E10: "HANJA",
        0x6E11: "HANJA",
        0x6E12: "HANJA",
        0x6E13: "HANJA",
        0x6E14: "HANJA",
        0x6E15: "HANJA",
        0x6E16: "HANJA",
        0x6E17: "HANJA",
        0x6E18: "HANJA",
        0x6E19: "HANJA",
        0x6E1A: "HANJA",
        0x6E1B: "HANJA",
        0x6E1C: "HANJA",
        0x6E1D: "HANJA",
        0x6E1E: "HANJA",
        0x6E1F: "HANJA",
        0x6E20: "HANJA",
        0x6E21: "HANJA",
        0x6E22: "HANJA",
        0x6E23: "HANJA",
        0x6E24: "HANJA",
        0x6E25: "HANJA",
        0x6E26: "HANJA",
        0x6E27: "HANJA",
        0x6E28: "HANJA",
        0x6E29: "HANJA",
        0x6E2A: "HANJA",
        0x6E2B: "HANJA",
        0x6E2C: "HANJA",
        0x6E2D: "HANJA",
        0x6E2E: "HANJA",
        0x6E2F: "HANJA",
        0x6E30: "HANJA",
        0x6E31: "HANJA",
        0x6E32: "HANJA",
        0x6E33: "HANJA",
        0x6E34: "HANJA",
        0x6E35: "HANJA",
        0x6E36: "HANJA",
        0x6E37: "HANJA",
        0x6E38: "HANJA",
        0x6E39: "HANJA",
        0x6E3A: "HANJA",
        0x6E3B: "HANJA",
        0x6E3C: "HANJA",
        0x6E3D: "HANJA",
        0x6E3E: "HANJA",
        0x6E3F: "HANJA",
        0x6E40: "HANJA",
        0x6E41: "HANJA",
        0x6E42: "HANJA",
        0x6E43: "HANJA",
        0x6E44: "HANJA",
        0x6E45: "HANJA",
        0x6E46: "HANJA",
        0x6E47: "HANJA",
        0x6E48: "HANJA",
        0x6E49: "HANJA",
        0x6E4A: "HANJA",
        0x6E4B: "HANJA",
        0x6E4C: "HANJA",
        0x6E4D: "HANJA",
        0x6E4E: "HANJA",
        0x6E4F: "HANJA",
        0x6E50: "HANJA",
        0x6E51: "HANJA",
        0x6E52: "HANJA",
        0x6E53: "HANJA",
        0x6E54: "HANJA",
        0x6E55: "HANJA",
        0x6E56: "HANJA",
        0x6E57: "HANJA",
        0x6E58: "HANJA",
        0x6E59: "HANJA",
        0x6E5A: "HANJA",
        0x6E5B: "HANJA",
        0x6E5C: "HANJA",
        0x6E5D: "HANJA",
        0x6E5E: "HANJA",
        0x6E5F: "HANJA",
        0x6E60: "HANJA",
        0x6E61: "HANJA",
        0x6E62: "HANJA",
        0x6E63: "HANJA",
        0x6E64: "HANJA",
        0x6E65: "HANJA",
        0x6E66: "HANJA",
        0x6E67: "HANJA",
        0x6E68: "HANJA",
        0x6E69: "HANJA",
        0x6E6A: "HANJA",
        0x6E6B: "HANJA",
        0x6E6C: "HANJA",
        0x6E6D: "HANJA",
        0x6E6E: "HANJA",
        0x6E6F: "HANJA",
        0x6E70: "HANJA",
        0x6E71: "HANJA",
        0x6E72: "HANJA",
        0x6E73: "HANJA",
        0x6E74: "HANJA",
        0x6E75: "HANJA",
        0x6E76: "HANJA",
        0x6E77: "HANJA",
        0x6E78: "HANJA",
        0x6E79: "HANJA",
        0x6E7A: "HANJA",
        0x6E7B: "HANJA",
        0x6E7C: "HANJA",
        0x6E7D: "HANJA",
        0x6E7E: "HANJA",
        0x6E7F: "HANJA",
        0x6E80: "HANJA",
        0x6E81: "HANJA",
        0x6E82: "HANJA",
        0x6E83: "HANJA",
        0x6E84: "HANJA",
        0x6E85: "HANJA",
        0x6E86: "HANJA",
        0x6E87: "HANJA",
        0x6E88: "HANJA",
        0x6E89: "HANJA",
        0x6E8A: "HANJA",
        0x6E8B: "HANJA",
        0x6E8C: "HANJA",
        0x6E8D: "HANJA",
        0x6E8E: "HANJA",
        0x6E8F: "HANJA",
        0x6E90: "HANJA",
        0x6E91: "HANJA",
        0x6E92: "HANJA",
        0x6E93: "HANJA",
        0x6E94: "HANJA",
        0x6E95: "HANJA",
        0x6E96: "HANJA",
        0x6E97: "HANJA",
        0x6E98: "HANJA",
        0x6E99: "HANJA",
        0x6E9A: "HANJA",
        0x6E9B: "HANJA",
        0x6E9C: "HANJA",
        0x6E9D: "HANJA",
        0x6E9E: "HANJA",
        0x6E9F: "HANJA",
        0x6EA0: "HANJA",
        0x6EA1: "HANJA",
        0x6EA2: "HANJA",
        0x6EA3: "HANJA",
        0x6EA4: "HANJA",
        0x6EA5: "HANJA",
        0x6EA6: "HANJA",
        0x6EA7: "HANJA",
        0x6EA8: "HANJA",
        0x6EA9: "HANJA",
        0x6EAA: "HANJA",
        0x6EAB: "HANJA",
        0x6EAC: "HANJA",
        0x6EAD: "HANJA",
        0x6EAE: "HANJA",
        0x6EAF: "HANJA",
        0x6EB0: "HANJA",
        0x6EB1: "HANJA",
        0x6EB2: "HANJA",
        0x6EB3: "HANJA",
        0x6EB4: "HANJA",
        0x6EB5: "HANJA",
        0x6EB6: "HANJA",
        0x6EB7: "HANJA",
        0x6EB8: "HANJA",
        0x6EB9: "HANJA",
        0x6EBA: "HANJA",
        0x6EBB: "HANJA",
        0x6EBC: "HANJA",
        0x6EBD: "HANJA",
        0x6EBE: "HANJA",
        0x6EBF: "HANJA",
        0x6EC0: "HANJA",
        0x6EC1: "HANJA",
        0x6EC2: "HANJA",
        0x6EC3: "HANJA",
        0x6EC4: "HANJA",
        0x6EC5: "HANJA",
        0x6EC6: "HANJA",
        0x6EC7: "HANJA",
        0x6EC8: "HANJA",
        0x6EC9: "HANJA",
        0x6ECA: "HANJA",
        0x6ECB: "HANJA",
        0x6ECC: "HANJA",
        0x6ECD: "HANJA",
        0x6ECE: "HANJA",
        0x6ECF: "HANJA",
        0x6ED0: "HANJA",
        0x6ED1: "HANJA",
        0x6ED2: "HANJA",
        0x6ED3: "HANJA",
        0x6ED4: "HANJA",
        0x6ED5: "HANJA",
        0x6ED6: "HANJA",
        0x6ED7: "HANJA",
        0x6ED8: "HANJA",
        0x6ED9: "HANJA",
        0x6EDA: "HANJA",
        0x6EDB: "HANJA",
        0x6EDC: "HANJA",
        0x6EDD: "HANJA",
        0x6EDE: "HANJA",
        0x6EDF: "HANJA",
        0x6EE0: "HANJA",
        0x6EE1: "HANJA",
        0x6EE2: "HANJA",
        0x6EE3: "HANJA",
        0x6EE4: "HANJA",
        0x6EE5: "HANJA",
        0x6EE6: "HANJA",
        0x6EE7: "HANJA",
        0x6EE8: "HANJA",
        0x6EE9: "HANJA",
        0x6EEA: "HANJA",
        0x6EEB: "HANJA",
        0x6EEC: "HANJA",
        0x6EED: "HANJA",
        0x6EEE: "HANJA",
        0x6EEF: "HANJA",
        0x6EF0: "HANJA",
        0x6EF1: "HANJA",
        0x6EF2: "HANJA",
        0x6EF3: "HANJA",
        0x6EF4: "HANJA",
        0x6EF5: "HANJA",
        0x6EF6: "HANJA",
        0x6EF7: "HANJA",
        0x6EF8: "HANJA",
        0x6EF9: "HANJA",
        0x6EFA: "HANJA",
        0x6EFB: "HANJA",
        0x6EFC: "HANJA",
        0x6EFD: "HANJA",
        0x6EFE: "HANJA",
        0x6EFF: "HANJA",
        0x6F00: "HANJA",
        0x6F01: "HANJA",
        0x6F02: "HANJA",
        0x6F03: "HANJA",
        0x6F04: "HANJA",
        0x6F05: "HANJA",
        0x6F06: "HANJA",
        0x6F07: "HANJA",
        0x6F08: "HANJA",
        0x6F09: "HANJA",
        0x6F0A: "HANJA",
        0x6F0B: "HANJA",
        0x6F0C: "HANJA",
        0x6F0D: "HANJA",
        0x6F0E: "HANJA",
        0x6F0F: "HANJA",
        0x6F10: "HANJA",
        0x6F11: "HANJA",
        0x6F12: "HANJA",
        0x6F13: "HANJA",
        0x6F14: "HANJA",
        0x6F15: "HANJA",
        0x6F16: "HANJA",
        0x6F17: "HANJA",
        0x6F18: "HANJA",
        0x6F19: "HANJA",
        0x6F1A: "HANJA",
        0x6F1B: "HANJA",
        0x6F1C: "HANJA",
        0x6F1D: "HANJA",
        0x6F1E: "HANJA",
        0x6F1F: "HANJA",
        0x6F20: "HANJA",
        0x6F21: "HANJA",
        0x6F22: "HANJA",
        0x6F23: "HANJA",
        0x6F24: "HANJA",
        0x6F25: "HANJA",
        0x6F26: "HANJA",
        0x6F27: "HANJA",
        0x6F28: "HANJA",
        0x6F29: "HANJA",
        0x6F2A: "HANJA",
        0x6F2B: "HANJA",
        0x6F2C: "HANJA",
        0x6F2D: "HANJA",
        0x6F2E: "HANJA",
        0x6F2F: "HANJA",
        0x6F30: "HANJA",
        0x6F31: "HANJA",
        0x6F32: "HANJA",
        0x6F33: "HANJA",
        0x6F34: "HANJA",
        0x6F35: "HANJA",
        0x6F36: "HANJA",
        0x6F37: "HANJA",
        0x6F38: "HANJA",
        0x6F39: "HANJA",
        0x6F3A: "HANJA",
        0x6F3B: "HANJA",
        0x6F3C: "HANJA",
        0x6F3D: "HANJA",
        0x6F3E: "HANJA",
        0x6F3F: "HANJA",
        0x6F40: "HANJA",
        0x6F41: "HANJA",
        0x6F42: "HANJA",
        0x6F43: "HANJA",
        0x6F44: "HANJA",
        0x6F45: "HANJA",
        0x6F46: "HANJA",
        0x6F47: "HANJA",
        0x6F48: "HANJA",
        0x6F49: "HANJA",
        0x6F4A: "HANJA",
        0x6F4B: "HANJA",
        0x6F4C: "HANJA",
        0x6F4D: "HANJA",
        0x6F4E: "HANJA",
        0x6F4F: "HANJA",
        0x6F50: "HANJA",
        0x6F51: "HANJA",
        0x6F52: "HANJA",
        0x6F53: "HANJA",
        0x6F54: "HANJA",
        0x6F55: "HANJA",
        0x6F56: "HANJA",
        0x6F57: "HANJA",
        0x6F58: "HANJA",
        0x6F59: "HANJA",
        0x6F5A: "HANJA",
        0x6F5B: "HANJA",
        0x6F5C: "HANJA",
        0x6F5D: "HANJA",
        0x6F5E: "HANJA",
        0x6F5F: "HANJA",
        0x6F60: "HANJA",
        0x6F61: "HANJA",
        0x6F62: "HANJA",
        0x6F63: "HANJA",
        0x6F64: "HANJA",
        0x6F65: "HANJA",
        0x6F66: "HANJA",
        0x6F67: "HANJA",
        0x6F68: "HANJA",
        0x6F69: "HANJA",
        0x6F6A: "HANJA",
        0x6F6B: "HANJA",
        0x6F6C: "HANJA",
        0x6F6D: "HANJA",
        0x6F6E: "HANJA",
        0x6F6F: "HANJA",
        0x6F70: "HANJA",
        0x6F71: "HANJA",
        0x6F72: "HANJA",
        0x6F73: "HANJA",
        0x6F74: "HANJA",
        0x6F75: "HANJA",
        0x6F76: "HANJA",
        0x6F77: "HANJA",
        0x6F78: "HANJA",
        0x6F79: "HANJA",
        0x6F7A: "HANJA",
        0x6F7B: "HANJA",
        0x6F7C: "HANJA",
        0x6F7D: "HANJA",
        0x6F7E: "HANJA",
        0x6F7F: "HANJA",
        0x6F80: "HANJA",
        0x6F81: "HANJA",
        0x6F82: "HANJA",
        0x6F83: "HANJA",
        0x6F84: "HANJA",
        0x6F85: "HANJA",
        0x6F86: "HANJA",
        0x6F87: "HANJA",
        0x6F88: "HANJA",
        0x6F89: "HANJA",
        0x6F8A: "HANJA",
        0x6F8B: "HANJA",
        0x6F8C: "HANJA",
        0x6F8D: "HANJA",
        0x6F8E: "HANJA",
        0x6F8F: "HANJA",
        0x6F90: "HANJA",
        0x6F91: "HANJA",
        0x6F92: "HANJA",
        0x6F93: "HANJA",
        0x6F94: "HANJA",
        0x6F95: "HANJA",
        0x6F96: "HANJA",
        0x6F97: "HANJA",
        0x6F98: "HANJA",
        0x6F99: "HANJA",
        0x6F9A: "HANJA",
        0x6F9B: "HANJA",
        0x6F9C: "HANJA",
        0x6F9D: "HANJA",
        0x6F9E: "HANJA",
        0x6F9F: "HANJA",
        0x6FA0: "HANJA",
        0x6FA1: "HANJA",
        0x6FA2: "HANJA",
        0x6FA3: "HANJA",
        0x6FA4: "HANJA",
        0x6FA5: "HANJA",
        0x6FA6: "HANJA",
        0x6FA7: "HANJA",
        0x6FA8: "HANJA",
        0x6FA9: "HANJA",
        0x6FAA: "HANJA",
        0x6FAB: "HANJA",
        0x6FAC: "HANJA",
        0x6FAD: "HANJA",
        0x6FAE: "HANJA",
        0x6FAF: "HANJA",
        0x6FB0: "HANJA",
        0x6FB1: "HANJA",
        0x6FB2: "HANJA",
        0x6FB3: "HANJA",
        0x6FB4: "HANJA",
        0x6FB5: "HANJA",
        0x6FB6: "HANJA",
        0x6FB7: "HANJA",
        0x6FB8: "HANJA",
        0x6FB9: "HANJA",
        0x6FBA: "HANJA",
        0x6FBB: "HANJA",
        0x6FBC: "HANJA",
        0x6FBD: "HANJA",
        0x6FBE: "HANJA",
        0x6FBF: "HANJA",
        0x6FC0: "HANJA",
        0x6FC1: "HANJA",
        0x6FC2: "HANJA",
        0x6FC3: "HANJA",
        0x6FC4: "HANJA",
        0x6FC5: "HANJA",
        0x6FC6: "HANJA",
        0x6FC7: "HANJA",
        0x6FC8: "HANJA",
        0x6FC9: "HANJA",
        0x6FCA: "HANJA",
        0x6FCB: "HANJA",
        0x6FCC: "HANJA",
        0x6FCD: "HANJA",
        0x6FCE: "HANJA",
        0x6FCF: "HANJA",
        0x6FD0: "HANJA",
        0x6FD1: "HANJA",
        0x6FD2: "HANJA",
        0x6FD3: "HANJA",
        0x6FD4: "HANJA",
        0x6FD5: "HANJA",
        0x6FD6: "HANJA",
        0x6FD7: "HANJA",
        0x6FD8: "HANJA",
        0x6FD9: "HANJA",
        0x6FDA: "HANJA",
        0x6FDB: "HANJA",
        0x6FDC: "HANJA",
        0x6FDD: "HANJA",
        0x6FDE: "HANJA",
        0x6FDF: "HANJA",
        0x6FE0: "HANJA",
        0x6FE1: "HANJA",
        0x6FE2: "HANJA",
        0x6FE3: "HANJA",
        0x6FE4: "HANJA",
        0x6FE5: "HANJA",
        0x6FE6: "HANJA",
        0x6FE7: "HANJA",
        0x6FE8: "HANJA",
        0x6FE9: "HANJA",
        0x6FEA: "HANJA",
        0x6FEB: "HANJA",
        0x6FEC: "HANJA",
        0x6FED: "HANJA",
        0x6FEE: "HANJA",
        0x6FEF: "HANJA",
        0x6FF0: "HANJA",
        0x6FF1: "HANJA",
        0x6FF2: "HANJA",
        0x6FF3: "HANJA",
        0x6FF4: "HANJA",
        0x6FF5: "HANJA",
        0x6FF6: "HANJA",
        0x6FF7: "HANJA",
        0x6FF8: "HANJA",
        0x6FF9: "HANJA",
        0x6FFA: "HANJA",
        0x6FFB: "HANJA",
        0x6FFC: "HANJA",
        0x6FFD: "HANJA",
        0x6FFE: "HANJA",
        0x6FFF: "HANJA",
        0x7000: "HANJA",
        0x7001: "HANJA",
        0x7002: "HANJA",
        0x7003: "HANJA",
        0x7004: "HANJA",
        0x7005: "HANJA",
        0x7006: "HANJA",
        0x7007: "HANJA",
        0x7008: "HANJA",
        0x7009: "HANJA",
        0x700A: "HANJA",
        0x700B: "HANJA",
        0x700C: "HANJA",
        0x700D: "HANJA",
        0x700E: "HANJA",
        0x700F: "HANJA",
        0x7010: "HANJA",
        0x7011: "HANJA",
        0x7012: "HANJA",
        0x7013: "HANJA",
        0x7014: "HANJA",
        0x7015: "HANJA",
        0x7016: "HANJA",
        0x7017: "HANJA",
        0x7018: "HANJA",
        0x7019: "HANJA",
        0x701A: "HANJA",
        0x701B: "HANJA",
        0x701C: "HANJA",
        0x701D: "HANJA",
        0x701E: "HANJA",
        0x701F: "HANJA",
        0x7020: "HANJA",
        0x7021: "HANJA",
        0x7022: "HANJA",
        0x7023: "HANJA",
        0x7024: "HANJA",
        0x7025: "HANJA",
        0x7026: "HANJA",
        0x7027: "HANJA",
        0x7028: "HANJA",
        0x7029: "HANJA",
        0x702A: "HANJA",
        0x702B: "HANJA",
        0x702C: "HANJA",
        0x702D: "HANJA",
        0x702E: "HANJA",
        0x702F: "HANJA",
        0x7030: "HANJA",
        0x7031: "HANJA",
        0x7032: "HANJA",
        0x7033: "HANJA",
        0x7034: "HANJA",
        0x7035: "HANJA",
        0x7036: "HANJA",
        0x7037: "HANJA",
        0x7038: "HANJA",
        0x7039: "HANJA",
        0x703A: "HANJA",
        0x703B: "HANJA",
        0x703C: "HANJA",
        0x703D: "HANJA",
        0x703E: "HANJA",
        0x703F: "HANJA",
        0x7040: "HANJA",
        0x7041: "HANJA",
        0x7042: "HANJA",
        0x7043: "HANJA",
        0x7044: "HANJA",
        0x7045: "HANJA",
        0x7046: "HANJA",
        0x7047: "HANJA",
        0x7048: "HANJA",
        0x7049: "HANJA",
        0x704A: "HANJA",
        0x704B: "HANJA",
        0x704C: "HANJA",
        0x704D: "HANJA",
        0x704E: "HANJA",
        0x704F: "HANJA",
        0x7050: "HANJA",
        0x7051: "HANJA",
        0x7052: "HANJA",
        0x7053: "HANJA",
        0x7054: "HANJA",
        0x7055: "HANJA",
        0x7056: "HANJA",
        0x7057: "HANJA",
        0x7058: "HANJA",
        0x7059: "HANJA",
        0x705A: "HANJA",
        0x705B: "HANJA",
        0x705C: "HANJA",
        0x705D: "HANJA",
        0x705E: "HANJA",
        0x705F: "HANJA",
        0x7060: "HANJA",
        0x7061: "HANJA",
        0x7062: "HANJA",
        0x7063: "HANJA",
        0x7064: "HANJA",
        0x7065: "HANJA",
        0x7066: "HANJA",
        0x7067: "HANJA",
        0x7068: "HANJA",
        0x7069: "HANJA",
        0x706A: "HANJA",
        0x706B: "HANJA",
        0x706C: "HANJA",
        0x706D: "HANJA",
        0x706E: "HANJA",
        0x706F: "HANJA",
        0x7070: "HANJA",
        0x7071: "HANJA",
        0x7072: "HANJA",
        0x7073: "HANJA",
        0x7074: "HANJA",
        0x7075: "HANJA",
        0x7076: "HANJA",
        0x7077: "HANJA",
        0x7078: "HANJA",
        0x7079: "HANJA",
        0x707A: "HANJA",
        0x707B: "HANJA",
        0x707C: "HANJA",
        0x707D: "HANJA",
        0x707E: "HANJA",
        0x707F: "HANJA",
        0x7080: "HANJA",
        0x7081: "HANJA",
        0x7082: "HANJA",
        0x7083: "HANJA",
        0x7084: "HANJA",
        0x7085: "HANJA",
        0x7086: "HANJA",
        0x7087: "HANJA",
        0x7088: "HANJA",
        0x7089: "HANJA",
        0x708A: "HANJA",
        0x708B: "HANJA",
        0x708C: "HANJA",
        0x708D: "HANJA",
        0x708E: "HANJA",
        0x708F: "HANJA",
        0x7090: "HANJA",
        0x7091: "HANJA",
        0x7092: "HANJA",
        0x7093: "HANJA",
        0x7094: "HANJA",
        0x7095: "HANJA",
        0x7096: "HANJA",
        0x7097: "HANJA",
        0x7098: "HANJA",
        0x7099: "HANJA",
        0x709A: "HANJA",
        0x709B: "HANJA",
        0x709C: "HANJA",
        0x709D: "HANJA",
        0x709E: "HANJA",
        0x709F: "HANJA",
        0x70A0: "HANJA",
        0x70A1: "HANJA",
        0x70A2: "HANJA",
        0x70A3: "HANJA",
        0x70A4: "HANJA",
        0x70A5: "HANJA",
        0x70A6: "HANJA",
        0x70A7: "HANJA",
        0x70A8: "HANJA",
        0x70A9: "HANJA",
        0x70AA: "HANJA",
        0x70AB: "HANJA",
        0x70AC: "HANJA",
        0x70AD: "HANJA",
        0x70AE: "HANJA",
        0x70AF: "HANJA",
        0x70B0: "HANJA",
        0x70B1: "HANJA",
        0x70B2: "HANJA",
        0x70B3: "HANJA",
        0x70B4: "HANJA",
        0x70B5: "HANJA",
        0x70B6: "HANJA",
        0x70B7: "HANJA",
        0x70B8: "HANJA",
        0x70B9: "HANJA",
        0x70BA: "HANJA",
        0x70BB: "HANJA",
        0x70BC: "HANJA",
        0x70BD: "HANJA",
        0x70BE: "HANJA",
        0x70BF: "HANJA",
        0x70C0: "HANJA",
        0x70C1: "HANJA",
        0x70C2: "HANJA",
        0x70C3: "HANJA",
        0x70C4: "HANJA",
        0x70C5: "HANJA",
        0x70C6: "HANJA",
        0x70C7: "HANJA",
        0x70C8: "HANJA",
        0x70C9: "HANJA",
        0x70CA: "HANJA",
        0x70CB: "HANJA",
        0x70CC: "HANJA",
        0x70CD: "HANJA",
        0x70CE: "HANJA",
        0x70CF: "HANJA",
        0x70D0: "HANJA",
        0x70D1: "HANJA",
        0x70D2: "HANJA",
        0x70D3: "HANJA",
        0x70D4: "HANJA",
        0x70D5: "HANJA",
        0x70D6: "HANJA",
        0x70D7: "HANJA",
        0x70D8: "HANJA",
        0x70D9: "HANJA",
        0x70DA: "HANJA",
        0x70DB: "HANJA",
        0x70DC: "HANJA",
        0x70DD: "HANJA",
        0x70DE: "HANJA",
        0x70DF: "HANJA",
        0x70E0: "HANJA",
        0x70E1: "HANJA",
        0x70E2: "HANJA",
        0x70E3: "HANJA",
        0x70E4: "HANJA",
        0x70E5: "HANJA",
        0x70E6: "HANJA",
        0x70E7: "HANJA",
        0x70E8: "HANJA",
        0x70E9: "HANJA",
        0x70EA: "HANJA",
        0x70EB: "HANJA",
        0x70EC: "HANJA",
        0x70ED: "HANJA",
        0x70EE: "HANJA",
        0x70EF: "HANJA",
        0x70F0: "HANJA",
        0x70F1: "HANJA",
        0x70F2: "HANJA",
        0x70F3: "HANJA",
        0x70F4: "HANJA",
        0x70F5: "HANJA",
        0x70F6: "HANJA",
        0x70F7: "HANJA",
        0x70F8: "HANJA",
        0x70F9: "HANJA",
        0x70FA: "HANJA",
        0x70FB: "HANJA",
        0x70FC: "HANJA",
        0x70FD: "HANJA",
        0x70FE: "HANJA",
        0x70FF: "HANJA",
        0x7100: "HANJA",
        0x7101: "HANJA",
        0x7102: "HANJA",
        0x7103: "HANJA",
        0x7104: "HANJA",
        0x7105: "HANJA",
        0x7106: "HANJA",
        0x7107: "HANJA",
        0x7108: "HANJA",
        0x7109: "HANJA",
        0x710A: "HANJA",
        0x710B: "HANJA",
        0x710C: "HANJA",
        0x710D: "HANJA",
        0x710E: "HANJA",
        0x710F: "HANJA",
        0x7110: "HANJA",
        0x7111: "HANJA",
        0x7112: "HANJA",
        0x7113: "HANJA",
        0x7114: "HANJA",
        0x7115: "HANJA",
        0x7116: "HANJA",
        0x7117: "HANJA",
        0x7118: "HANJA",
        0x7119: "HANJA",
        0x711A: "HANJA",
        0x711B: "HANJA",
        0x711C: "HANJA",
        0x711D: "HANJA",
        0x711E: "HANJA",
        0x711F: "HANJA",
        0x7120: "HANJA",
        0x7121: "HANJA",
        0x7122: "HANJA",
        0x7123: "HANJA",
        0x7124: "HANJA",
        0x7125: "HANJA",
        0x7126: "HANJA",
        0x7127: "HANJA",
        0x7128: "HANJA",
        0x7129: "HANJA",
        0x712A: "HANJA",
        0x712B: "HANJA",
        0x712C: "HANJA",
        0x712D: "HANJA",
        0x712E: "HANJA",
        0x712F: "HANJA",
        0x7130: "HANJA",
        0x7131: "HANJA",
        0x7132: "HANJA",
        0x7133: "HANJA",
        0x7134: "HANJA",
        0x7135: "HANJA",
        0x7136: "HANJA",
        0x7137: "HANJA",
        0x7138: "HANJA",
        0x7139: "HANJA",
        0x713A: "HANJA",
        0x713B: "HANJA",
        0x713C: "HANJA",
        0x713D: "HANJA",
        0x713E: "HANJA",
        0x713F: "HANJA",
        0x7140: "HANJA",
        0x7141: "HANJA",
        0x7142: "HANJA",
        0x7143: "HANJA",
        0x7144: "HANJA",
        0x7145: "HANJA",
        0x7146: "HANJA",
        0x7147: "HANJA",
        0x7148: "HANJA",
        0x7149: "HANJA",
        0x714A: "HANJA",
        0x714B: "HANJA",
        0x714C: "HANJA",
        0x714D: "HANJA",
        0x714E: "HANJA",
        0x714F: "HANJA",
        0x7150: "HANJA",
        0x7151: "HANJA",
        0x7152: "HANJA",
        0x7153: "HANJA",
        0x7154: "HANJA",
        0x7155: "HANJA",
        0x7156: "HANJA",
        0x7157: "HANJA",
        0x7158: "HANJA",
        0x7159: "HANJA",
        0x715A: "HANJA",
        0x715B: "HANJA",
        0x715C: "HANJA",
        0x715D: "HANJA",
        0x715E: "HANJA",
        0x715F: "HANJA",
        0x7160: "HANJA",
        0x7161: "HANJA",
        0x7162: "HANJA",
        0x7163: "HANJA",
        0x7164: "HANJA",
        0x7165: "HANJA",
        0x7166: "HANJA",
        0x7167: "HANJA",
        0x7168: "HANJA",
        0x7169: "HANJA",
        0x716A: "HANJA",
        0x716B: "HANJA",
        0x716C: "HANJA",
        0x716D: "HANJA",
        0x716E: "HANJA",
        0x716F: "HANJA",
        0x7170: "HANJA",
        0x7171: "HANJA",
        0x7172: "HANJA",
        0x7173: "HANJA",
        0x7174: "HANJA",
        0x7175: "HANJA",
        0x7176: "HANJA",
        0x7177: "HANJA",
        0x7178: "HANJA",
        0x7179: "HANJA",
        0x717A: "HANJA",
        0x717B: "HANJA",
        0x717C: "HANJA",
        0x717D: "HANJA",
        0x717E: "HANJA",
        0x717F: "HANJA",
        0x7180: "HANJA",
        0x7181: "HANJA",
        0x7182: "HANJA",
        0x7183: "HANJA",
        0x7184: "HANJA",
        0x7185: "HANJA",
        0x7186: "HANJA",
        0x7187: "HANJA",
        0x7188: "HANJA",
        0x7189: "HANJA",
        0x718A: "HANJA",
        0x718B: "HANJA",
        0x718C: "HANJA",
        0x718D: "HANJA",
        0x718E: "HANJA",
        0x718F: "HANJA",
        0x7190: "HANJA",
        0x7191: "HANJA",
        0x7192: "HANJA",
        0x7193: "HANJA",
        0x7194: "HANJA",
        0x7195: "HANJA",
        0x7196: "HANJA",
        0x7197: "HANJA",
        0x7198: "HANJA",
        0x7199: "HANJA",
        0x719A: "HANJA",
        0x719B: "HANJA",
        0x719C: "HANJA",
        0x719D: "HANJA",
        0x719E: "HANJA",
        0x719F: "HANJA",
        0x71A0: "HANJA",
        0x71A1: "HANJA",
        0x71A2: "HANJA",
        0x71A3: "HANJA",
        0x71A4: "HANJA",
        0x71A5: "HANJA",
        0x71A6: "HANJA",
        0x71A7: "HANJA",
        0x71A8: "HANJA",
        0x71A9: "HANJA",
        0x71AA: "HANJA",
        0x71AB: "HANJA",
        0x71AC: "HANJA",
        0x71AD: "HANJA",
        0x71AE: "HANJA",
        0x71AF: "HANJA",
        0x71B0: "HANJA",
        0x71B1: "HANJA",
        0x71B2: "HANJA",
        0x71B3: "HANJA",
        0x71B4: "HANJA",
        0x71B5: "HANJA",
        0x71B6: "HANJA",
        0x71B7: "HANJA",
        0x71B8: "HANJA",
        0x71B9: "HANJA",
        0x71BA: "HANJA",
        0x71BB: "HANJA",
        0x71BC: "HANJA",
        0x71BD: "HANJA",
        0x71BE: "HANJA",
        0x71BF: "HANJA",
        0x71C0: "HANJA",
        0x71C1: "HANJA",
        0x71C2: "HANJA",
        0x71C3: "HANJA",
        0x71C4: "HANJA",
        0x71C5: "HANJA",
        0x71C6: "HANJA",
        0x71C7: "HANJA",
        0x71C8: "HANJA",
        0x71C9: "HANJA",
        0x71CA: "HANJA",
        0x71CB: "HANJA",
        0x71CC: "HANJA",
        0x71CD: "HANJA",
        0x71CE: "HANJA",
        0x71CF: "HANJA",
        0x71D0: "HANJA",
        0x71D1: "HANJA",
        0x71D2: "HANJA",
        0x71D3: "HANJA",
        0x71D4: "HANJA",
        0x71D5: "HANJA",
        0x71D6: "HANJA",
        0x71D7: "HANJA",
        0x71D8: "HANJA",
        0x71D9: "HANJA",
        0x71DA: "HANJA",
        0x71DB: "HANJA",
        0x71DC: "HANJA",
        0x71DD: "HANJA",
        0x71DE: "HANJA",
        0x71DF: "HANJA",
        0x71E0: "HANJA",
        0x71E1: "HANJA",
        0x71E2: "HANJA",
        0x71E3: "HANJA",
        0x71E4: "HANJA",
        0x71E5: "HANJA",
        0x71E6: "HANJA",
        0x71E7: "HANJA",
        0x71E8: "HANJA",
        0x71E9: "HANJA",
        0x71EA: "HANJA",
        0x71EB: "HANJA",
        0x71EC: "HANJA",
        0x71ED: "HANJA",
        0x71EE: "HANJA",
        0x71EF: "HANJA",
        0x71F0: "HANJA",
        0x71F1: "HANJA",
        0x71F2: "HANJA",
        0x71F3: "HANJA",
        0x71F4: "HANJA",
        0x71F5: "HANJA",
        0x71F6: "HANJA",
        0x71F7: "HANJA",
        0x71F8: "HANJA",
        0x71F9: "HANJA",
        0x71FA: "HANJA",
        0x71FB: "HANJA",
        0x71FC: "HANJA",
        0x71FD: "HANJA",
        0x71FE: "HANJA",
        0x71FF: "HANJA",
        0x7200: "HANJA",
        0x7201: "HANJA",
        0x7202: "HANJA",
        0x7203: "HANJA",
        0x7204: "HANJA",
        0x7205: "HANJA",
        0x7206: "HANJA",
        0x7207: "HANJA",
        0x7208: "HANJA",
        0x7209: "HANJA",
        0x720A: "HANJA",
        0x720B: "HANJA",
        0x720C: "HANJA",
        0x720D: "HANJA",
        0x720E: "HANJA",
        0x720F: "HANJA",
        0x7210: "HANJA",
        0x7211: "HANJA",
        0x7212: "HANJA",
        0x7213: "HANJA",
        0x7214: "HANJA",
        0x7215: "HANJA",
        0x7216: "HANJA",
        0x7217: "HANJA",
        0x7218: "HANJA",
        0x7219: "HANJA",
        0x721A: "HANJA",
        0x721B: "HANJA",
        0x721C: "HANJA",
        0x721D: "HANJA",
        0x721E: "HANJA",
        0x721F: "HANJA",
        0x7220: "HANJA",
        0x7221: "HANJA",
        0x7222: "HANJA",
        0x7223: "HANJA",
        0x7224: "HANJA",
        0x7225: "HANJA",
        0x7226: "HANJA",
        0x7227: "HANJA",
        0x7228: "HANJA",
        0x7229: "HANJA",
        0x722A: "HANJA",
        0x722B: "HANJA",
        0x722C: "HANJA",
        0x722D: "HANJA",
        0x722E: "HANJA",
        0x722F: "HANJA",
        0x7230: "HANJA",
        0x7231: "HANJA",
        0x7232: "HANJA",
        0x7233: "HANJA",
        0x7234: "HANJA",
        0x7235: "HANJA",
        0x7236: "HANJA",
        0x7237: "HANJA",
        0x7238: "HANJA",
        0x7239: "HANJA",
        0x723A: "HANJA",
        0x723B: "HANJA",
        0x723C: "HANJA",
        0x723D: "HANJA",
        0x723E: "HANJA",
        0x723F: "HANJA",
        0x7240: "HANJA",
        0x7241: "HANJA",
        0x7242: "HANJA",
        0x7243: "HANJA",
        0x7244: "HANJA",
        0x7245: "HANJA",
        0x7246: "HANJA",
        0x7247: "HANJA",
        0x7248: "HANJA",
        0x7249: "HANJA",
        0x724A: "HANJA",
        0x724B: "HANJA",
        0x724C: "HANJA",
        0x724D: "HANJA",
        0x724E: "HANJA",
        0x724F: "HANJA",
        0x7250: "HANJA",
        0x7251: "HANJA",
        0x7252: "HANJA",
        0x7253: "HANJA",
        0x7254: "HANJA",
        0x7255: "HANJA",
        0x7256: "HANJA",
        0x7257: "HANJA",
        0x7258: "HANJA",
        0x7259: "HANJA",
        0x725A: "HANJA",
        0x725B: "HANJA",
        0x725C: "HANJA",
        0x725D: "HANJA",
        0x725E: "HANJA",
        0x725F: "HANJA",
        0x7260: "HANJA",
        0x7261: "HANJA",
        0x7262: "HANJA",
        0x7263: "HANJA",
        0x7264: "HANJA",
        0x7265: "HANJA",
        0x7266: "HANJA",
        0x7267: "HANJA",
        0x7268: "HANJA",
        0x7269: "HANJA",
        0x726A: "HANJA",
        0x726B: "HANJA",
        0x726C: "HANJA",
        0x726D: "HANJA",
        0x726E: "HANJA",
        0x726F: "HANJA",
        0x7270: "HANJA",
        0x7271: "HANJA",
        0x7272: "HANJA",
        0x7273: "HANJA",
        0x7274: "HANJA",
        0x7275: "HANJA",
        0x7276: "HANJA",
        0x7277: "HANJA",
        0x7278: "HANJA",
        0x7279: "HANJA",
        0x727A: "HANJA",
        0x727B: "HANJA",
        0x727C: "HANJA",
        0x727D: "HANJA",
        0x727E: "HANJA",
        0x727F: "HANJA",
        0x7280: "HANJA",
        0x7281: "HANJA",
        0x7282: "HANJA",
        0x7283: "HANJA",
        0x7284: "HANJA",
        0x7285: "HANJA",
        0x7286: "HANJA",
        0x7287: "HANJA",
        0x7288: "HANJA",
        0x7289: "HANJA",
        0x728A: "HANJA",
        0x728B: "HANJA",
        0x728C: "HANJA",
        0x728D: "HANJA",
        0x728E: "HANJA",
        0x728F: "HANJA",
        0x7290: "HANJA",
        0x7291: "HANJA",
        0x7292: "HANJA",
        0x7293: "HANJA",
        0x7294: "HANJA",
        0x7295: "HANJA",
        0x7296: "HANJA",
        0x7297: "HANJA",
        0x7298: "HANJA",
        0x7299: "HANJA",
        0x729A: "HANJA",
        0x729B: "HANJA",
        0x729C: "HANJA",
        0x729D: "HANJA",
        0x729E: "HANJA",
        0x729F: "HANJA",
        0x72A0: "HANJA",
        0x72A1: "HANJA",
        0x72A2: "HANJA",
        0x72A3: "HANJA",
        0x72A4: "HANJA",
        0x72A5: "HANJA",
        0x72A6: "HANJA",
        0x72A7: "HANJA",
        0x72A8: "HANJA",
        0x72A9: "HANJA",
        0x72AA: "HANJA",
        0x72AB: "HANJA",
        0x72AC: "HANJA",
        0x72AD: "HANJA",
        0x72AE: "HANJA",
        0x72AF: "HANJA",
        0x72B0: "HANJA",
        0x72B1: "HANJA",
        0x72B2: "HANJA",
        0x72B3: "HANJA",
        0x72B4: "HANJA",
        0x72B5: "HANJA",
        0x72B6: "HANJA",
        0x72B7: "HANJA",
        0x72B8: "HANJA",
        0x72B9: "HANJA",
        0x72BA: "HANJA",
        0x72BB: "HANJA",
        0x72BC: "HANJA",
        0x72BD: "HANJA",
        0x72BE: "HANJA",
        0x72BF: "HANJA",
        0x72C0: "HANJA",
        0x72C1: "HANJA",
        0x72C2: "HANJA",
        0x72C3: "HANJA",
        0x72C4: "HANJA",
        0x72C5: "HANJA",
        0x72C6: "HANJA",
        0x72C7: "HANJA",
        0x72C8: "HANJA",
        0x72C9: "HANJA",
        0x72CA: "HANJA",
        0x72CB: "HANJA",
        0x72CC: "HANJA",
        0x72CD: "HANJA",
        0x72CE: "HANJA",
        0x72CF: "HANJA",
        0x72D0: "HANJA",
        0x72D1: "HANJA",
        0x72D2: "HANJA",
        0x72D3: "HANJA",
        0x72D4: "HANJA",
        0x72D5: "HANJA",
        0x72D6: "HANJA",
        0x72D7: "HANJA",
        0x72D8: "HANJA",
        0x72D9: "HANJA",
        0x72DA: "HANJA",
        0x72DB: "HANJA",
        0x72DC: "HANJA",
        0x72DD: "HANJA",
        0x72DE: "HANJA",
        0x72DF: "HANJA",
        0x72E0: "HANJA",
        0x72E1: "HANJA",
        0x72E2: "HANJA",
        0x72E3: "HANJA",
        0x72E4: "HANJA",
        0x72E5: "HANJA",
        0x72E6: "HANJA",
        0x72E7: "HANJA",
        0x72E8: "HANJA",
        0x72E9: "HANJA",
        0x72EA: "HANJA",
        0x72EB: "HANJA",
        0x72EC: "HANJA",
        0x72ED: "HANJA",
        0x72EE: "HANJA",
        0x72EF: "HANJA",
        0x72F0: "HANJA",
        0x72F1: "HANJA",
        0x72F2: "HANJA",
        0x72F3: "HANJA",
        0x72F4: "HANJA",
        0x72F5: "HANJA",
        0x72F6: "HANJA",
        0x72F7: "HANJA",
        0x72F8: "HANJA",
        0x72F9: "HANJA",
        0x72FA: "HANJA",
        0x72FB: "HANJA",
        0x72FC: "HANJA",
        0x72FD: "HANJA",
        0x72FE: "HANJA",
        0x72FF: "HANJA",
        0x7300: "HANJA",
        0x7301: "HANJA",
        0x7302: "HANJA",
        0x7303: "HANJA",
        0x7304: "HANJA",
        0x7305: "HANJA",
        0x7306: "HANJA",
        0x7307: "HANJA",
        0x7308: "HANJA",
        0x7309: "HANJA",
        0x730A: "HANJA",
        0x730B: "HANJA",
        0x730C: "HANJA",
        0x730D: "HANJA",
        0x730E: "HANJA",
        0x730F: "HANJA",
        0x7310: "HANJA",
        0x7311: "HANJA",
        0x7312: "HANJA",
        0x7313: "HANJA",
        0x7314: "HANJA",
        0x7315: "HANJA",
        0x7316: "HANJA",
        0x7317: "HANJA",
        0x7318: "HANJA",
        0x7319: "HANJA",
        0x731A: "HANJA",
        0x731B: "HANJA",
        0x731C: "HANJA",
        0x731D: "HANJA",
        0x731E: "HANJA",
        0x731F: "HANJA",
        0x7320: "HANJA",
        0x7321: "HANJA",
        0x7322: "HANJA",
        0x7323: "HANJA",
        0x7324: "HANJA",
        0x7325: "HANJA",
        0x7326: "HANJA",
        0x7327: "HANJA",
        0x7328: "HANJA",
        0x7329: "HANJA",
        0x732A: "HANJA",
        0x732B: "HANJA",
        0x732C: "HANJA",
        0x732D: "HANJA",
        0x732E: "HANJA",
        0x732F: "HANJA",
        0x7330: "HANJA",
        0x7331: "HANJA",
        0x7332: "HANJA",
        0x7333: "HANJA",
        0x7334: "HANJA",
        0x7335: "HANJA",
        0x7336: "HANJA",
        0x7337: "HANJA",
        0x7338: "HANJA",
        0x7339: "HANJA",
        0x733A: "HANJA",
        0x733B: "HANJA",
        0x733C: "HANJA",
        0x733D: "HANJA",
        0x733E: "HANJA",
        0x733F: "HANJA",
        0x7340: "HANJA",
        0x7341: "HANJA",
        0x7342: "HANJA",
        0x7343: "HANJA",
        0x7344: "HANJA",
        0x7345: "HANJA",
        0x7346: "HANJA",
        0x7347: "HANJA",
        0x7348: "HANJA",
        0x7349: "HANJA",
        0x734A: "HANJA",
        0x734B: "HANJA",
        0x734C: "HANJA",
        0x734D: "HANJA",
        0x734E: "HANJA",
        0x734F: "HANJA",
        0x7350: "HANJA",
        0x7351: "HANJA",
        0x7352: "HANJA",
        0x7353: "HANJA",
        0x7354: "HANJA",
        0x7355: "HANJA",
        0x7356: "HANJA",
        0x7357: "HANJA",
        0x7358: "HANJA",
        0x7359: "HANJA",
        0x735A: "HANJA",
        0x735B: "HANJA",
        0x735C: "HANJA",
        0x735D: "HANJA",
        0x735E: "HANJA",
        0x735F: "HANJA",
        0x7360: "HANJA",
        0x7361: "HANJA",
        0x7362: "HANJA",
        0x7363: "HANJA",
        0x7364: "HANJA",
        0x7365: "HANJA",
        0x7366: "HANJA",
        0x7367: "HANJA",
        0x7368: "HANJA",
        0x7369: "HANJA",
        0x736A: "HANJA",
        0x736B: "HANJA",
        0x736C: "HANJA",
        0x736D: "HANJA",
        0x736E: "HANJA",
        0x736F: "HANJA",
        0x7370: "HANJA",
        0x7371: "HANJA",
        0x7372: "HANJA",
        0x7373: "HANJA",
        0x7374: "HANJA",
        0x7375: "HANJA",
        0x7376: "HANJA",
        0x7377: "HANJA",
        0x7378: "HANJA",
        0x7379: "HANJA",
        0x737A: "HANJA",
        0x737B: "HANJA",
        0x737C: "HANJA",
        0x737D: "HANJA",
        0x737E: "HANJA",
        0x737F: "HANJA",
        0x7380: "HANJA",
        0x7381: "HANJA",
        0x7382: "HANJA",
        0x7383: "HANJA",
        0x7384: "HANJA",
        0x7385: "HANJA",
        0x7386: "HANJA",
        0x7387: "HANJA",
        0x7388: "HANJA",
        0x7389: "HANJA",
        0x738A: "HANJA",
        0x738B: "HANJA",
        0x738C: "HANJA",
        0x738D: "HANJA",
        0x738E: "HANJA",
        0x738F: "HANJA",
        0x7390: "HANJA",
        0x7391: "HANJA",
        0x7392: "HANJA",
        0x7393: "HANJA",
        0x7394: "HANJA",
        0x7395: "HANJA",
        0x7396: "HANJA",
        0x7397: "HANJA",
        0x7398: "HANJA",
        0x7399: "HANJA",
        0x739A: "HANJA",
        0x739B: "HANJA",
        0x739C: "HANJA",
        0x739D: "HANJA",
        0x739E: "HANJA",
        0x739F: "HANJA",
        0x73A0: "HANJA",
        0x73A1: "HANJA",
        0x73A2: "HANJA",
        0x73A3: "HANJA",
        0x73A4: "HANJA",
        0x73A5: "HANJA",
        0x73A6: "HANJA",
        0x73A7: "HANJA",
        0x73A8: "HANJA",
        0x73A9: "HANJA",
        0x73AA: "HANJA",
        0x73AB: "HANJA",
        0x73AC: "HANJA",
        0x73AD: "HANJA",
        0x73AE: "HANJA",
        0x73AF: "HANJA",
        0x73B0: "HANJA",
        0x73B1: "HANJA",
        0x73B2: "HANJA",
        0x73B3: "HANJA",
        0x73B4: "HANJA",
        0x73B5: "HANJA",
        0x73B6: "HANJA",
        0x73B7: "HANJA",
        0x73B8: "HANJA",
        0x73B9: "HANJA",
        0x73BA: "HANJA",
        0x73BB: "HANJA",
        0x73BC: "HANJA",
        0x73BD: "HANJA",
        0x73BE: "HANJA",
        0x73BF: "HANJA",
        0x73C0: "HANJA",
        0x73C1: "HANJA",
        0x73C2: "HANJA",
        0x73C3: "HANJA",
        0x73C4: "HANJA",
        0x73C5: "HANJA",
        0x73C6: "HANJA",
        0x73C7: "HANJA",
        0x73C8: "HANJA",
        0x73C9: "HANJA",
        0x73CA: "HANJA",
        0x73CB: "HANJA",
        0x73CC: "HANJA",
        0x73CD: "HANJA",
        0x73CE: "HANJA",
        0x73CF: "HANJA",
        0x73D0: "HANJA",
        0x73D1: "HANJA",
        0x73D2: "HANJA",
        0x73D3: "HANJA",
        0x73D4: "HANJA",
        0x73D5: "HANJA",
        0x73D6: "HANJA",
        0x73D7: "HANJA",
        0x73D8: "HANJA",
        0x73D9: "HANJA",
        0x73DA: "HANJA",
        0x73DB: "HANJA",
        0x73DC: "HANJA",
        0x73DD: "HANJA",
        0x73DE: "HANJA",
        0x73DF: "HANJA",
        0x73E0: "HANJA",
        0x73E1: "HANJA",
        0x73E2: "HANJA",
        0x73E3: "HANJA",
        0x73E4: "HANJA",
        0x73E5: "HANJA",
        0x73E6: "HANJA",
        0x73E7: "HANJA",
        0x73E8: "HANJA",
        0x73E9: "HANJA",
        0x73EA: "HANJA",
        0x73EB: "HANJA",
        0x73EC: "HANJA",
        0x73ED: "HANJA",
        0x73EE: "HANJA",
        0x73EF: "HANJA",
        0x73F0: "HANJA",
        0x73F1: "HANJA",
        0x73F2: "HANJA",
        0x73F3: "HANJA",
        0x73F4: "HANJA",
        0x73F5: "HANJA",
        0x73F6: "HANJA",
        0x73F7: "HANJA",
        0x73F8: "HANJA",
        0x73F9: "HANJA",
        0x73FA: "HANJA",
        0x73FB: "HANJA",
        0x73FC: "HANJA",
        0x73FD: "HANJA",
        0x73FE: "HANJA",
        0x73FF: "HANJA",
        0x7400: "HANJA",
        0x7401: "HANJA",
        0x7402: "HANJA",
        0x7403: "HANJA",
        0x7404: "HANJA",
        0x7405: "HANJA",
        0x7406: "HANJA",
        0x7407: "HANJA",
        0x7408: "HANJA",
        0x7409: "HANJA",
        0x740A: "HANJA",
        0x740B: "HANJA",
        0x740C: "HANJA",
        0x740D: "HANJA",
        0x740E: "HANJA",
        0x740F: "HANJA",
        0x7410: "HANJA",
        0x7411: "HANJA",
        0x7412: "HANJA",
        0x7413: "HANJA",
        0x7414: "HANJA",
        0x7415: "HANJA",
        0x7416: "HANJA",
        0x7417: "HANJA",
        0x7418: "HANJA",
        0x7419: "HANJA",
        0x741A: "HANJA",
        0x741B: "HANJA",
        0x741C: "HANJA",
        0x741D: "HANJA",
        0x741E: "HANJA",
        0x741F: "HANJA",
        0x7420: "HANJA",
        0x7421: "HANJA",
        0x7422: "HANJA",
        0x7423: "HANJA",
        0x7424: "HANJA",
        0x7425: "HANJA",
        0x7426: "HANJA",
        0x7427: "HANJA",
        0x7428: "HANJA",
        0x7429: "HANJA",
        0x742A: "HANJA",
        0x742B: "HANJA",
        0x742C: "HANJA",
        0x742D: "HANJA",
        0x742E: "HANJA",
        0x742F: "HANJA",
        0x7430: "HANJA",
        0x7431: "HANJA",
        0x7432: "HANJA",
        0x7433: "HANJA",
        0x7434: "HANJA",
        0x7435: "HANJA",
        0x7436: "HANJA",
        0x7437: "HANJA",
        0x7438: "HANJA",
        0x7439: "HANJA",
        0x743A: "HANJA",
        0x743B: "HANJA",
        0x743C: "HANJA",
        0x743D: "HANJA",
        0x743E: "HANJA",
        0x743F: "HANJA",
        0x7440: "HANJA",
        0x7441: "HANJA",
        0x7442: "HANJA",
        0x7443: "HANJA",
        0x7444: "HANJA",
        0x7445: "HANJA",
        0x7446: "HANJA",
        0x7447: "HANJA",
        0x7448: "HANJA",
        0x7449: "HANJA",
        0x744A: "HANJA",
        0x744B: "HANJA",
        0x744C: "HANJA",
        0x744D: "HANJA",
        0x744E: "HANJA",
        0x744F: "HANJA",
        0x7450: "HANJA",
        0x7451: "HANJA",
        0x7452: "HANJA",
        0x7453: "HANJA",
        0x7454: "HANJA",
        0x7455: "HANJA",
        0x7456: "HANJA",
        0x7457: "HANJA",
        0x7458: "HANJA",
        0x7459: "HANJA",
        0x745A: "HANJA",
        0x745B: "HANJA",
        0x745C: "HANJA",
        0x745D: "HANJA",
        0x745E: "HANJA",
        0x745F: "HANJA",
        0x7460: "HANJA",
        0x7461: "HANJA",
        0x7462: "HANJA",
        0x7463: "HANJA",
        0x7464: "HANJA",
        0x7465: "HANJA",
        0x7466: "HANJA",
        0x7467: "HANJA",
        0x7468: "HANJA",
        0x7469: "HANJA",
        0x746A: "HANJA",
        0x746B: "HANJA",
        0x746C: "HANJA",
        0x746D: "HANJA",
        0x746E: "HANJA",
        0x746F: "HANJA",
        0x7470: "HANJA",
        0x7471: "HANJA",
        0x7472: "HANJA",
        0x7473: "HANJA",
        0x7474: "HANJA",
        0x7475: "HANJA",
        0x7476: "HANJA",
        0x7477: "HANJA",
        0x7478: "HANJA",
        0x7479: "HANJA",
        0x747A: "HANJA",
        0x747B: "HANJA",
        0x747C: "HANJA",
        0x747D: "HANJA",
        0x747E: "HANJA",
        0x747F: "HANJA",
        0x7480: "HANJA",
        0x7481: "HANJA",
        0x7482: "HANJA",
        0x7483: "HANJA",
        0x7484: "HANJA",
        0x7485: "HANJA",
        0x7486: "HANJA",
        0x7487: "HANJA",
        0x7488: "HANJA",
        0x7489: "HANJA",
        0x748A: "HANJA",
        0x748B: "HANJA",
        0x748C: "HANJA",
        0x748D: "HANJA",
        0x748E: "HANJA",
        0x748F: "HANJA",
        0x7490: "HANJA",
        0x7491: "HANJA",
        0x7492: "HANJA",
        0x7493: "HANJA",
        0x7494: "HANJA",
        0x7495: "HANJA",
        0x7496: "HANJA",
        0x7497: "HANJA",
        0x7498: "HANJA",
        0x7499: "HANJA",
        0x749A: "HANJA",
        0x749B: "HANJA",
        0x749C: "HANJA",
        0x749D: "HANJA",
        0x749E: "HANJA",
        0x749F: "HANJA",
        0x74A0: "HANJA",
        0x74A1: "HANJA",
        0x74A2: "HANJA",
        0x74A3: "HANJA",
        0x74A4: "HANJA",
        0x74A5: "HANJA",
        0x74A6: "HANJA",
        0x74A7: "HANJA",
        0x74A8: "HANJA",
        0x74A9: "HANJA",
        0x74AA: "HANJA",
        0x74AB: "HANJA",
        0x74AC: "HANJA",
        0x74AD: "HANJA",
        0x74AE: "HANJA",
        0x74AF: "HANJA",
        0x74B0: "HANJA",
        0x74B1: "HANJA",
        0x74B2: "HANJA",
        0x74B3: "HANJA",
        0x74B4: "HANJA",
        0x74B5: "HANJA",
        0x74B6: "HANJA",
        0x74B7: "HANJA",
        0x74B8: "HANJA",
        0x74B9: "HANJA",
        0x74BA: "HANJA",
        0x74BB: "HANJA",
        0x74BC: "HANJA",
        0x74BD: "HANJA",
        0x74BE: "HANJA",
        0x74BF: "HANJA",
        0x74C0: "HANJA",
        0x74C1: "HANJA",
        0x74C2: "HANJA",
        0x74C3: "HANJA",
        0x74C4: "HANJA",
        0x74C5: "HANJA",
        0x74C6: "HANJA",
        0x74C7: "HANJA",
        0x74C8: "HANJA",
        0x74C9: "HANJA",
        0x74CA: "HANJA",
        0x74CB: "HANJA",
        0x74CC: "HANJA",
        0x74CD: "HANJA",
        0x74CE: "HANJA",
        0x74CF: "HANJA",
        0x74D0: "HANJA",
        0x74D1: "HANJA",
        0x74D2: "HANJA",
        0x74D3: "HANJA",
        0x74D4: "HANJA",
        0x74D5: "HANJA",
        0x74D6: "HANJA",
        0x74D7: "HANJA",
        0x74D8: "HANJA",
        0x74D9: "HANJA",
        0x74DA: "HANJA",
        0x74DB: "HANJA",
        0x74DC: "HANJA",
        0x74DD: "HANJA",
        0x74DE: "HANJA",
        0x74DF: "HANJA",
        0x74E0: "HANJA",
        0x74E1: "HANJA",
        0x74E2: "HANJA",
        0x74E3: "HANJA",
        0x74E4: "HANJA",
        0x74E5: "HANJA",
        0x74E6: "HANJA",
        0x74E7: "HANJA",
        0x74E8: "HANJA",
        0x74E9: "HANJA",
        0x74EA: "HANJA",
        0x74EB: "HANJA",
        0x74EC: "HANJA",
        0x74ED: "HANJA",
        0x74EE: "HANJA",
        0x74EF: "HANJA",
        0x74F0: "HANJA",
        0x74F1: "HANJA",
        0x74F2: "HANJA",
        0x74F3: "HANJA",
        0x74F4: "HANJA",
        0x74F5: "HANJA",
        0x74F6: "HANJA",
        0x74F7: "HANJA",
        0x74F8: "HANJA",
        0x74F9: "HANJA",
        0x74FA: "HANJA",
        0x74FB: "HANJA",
        0x74FC: "HANJA",
        0x74FD: "HANJA",
        0x74FE: "HANJA",
        0x74FF: "HANJA",
        0x7500: "HANJA",
        0x7501: "HANJA",
        0x7502: "HANJA",
        0x7503: "HANJA",
        0x7504: "HANJA",
        0x7505: "HANJA",
        0x7506: "HANJA",
        0x7507: "HANJA",
        0x7508: "HANJA",
        0x7509: "HANJA",
        0x750A: "HANJA",
        0x750B: "HANJA",
        0x750C: "HANJA",
        0x750D: "HANJA",
        0x750E: "HANJA",
        0x750F: "HANJA",
        0x7510: "HANJA",
        0x7511: "HANJA",
        0x7512: "HANJA",
        0x7513: "HANJA",
        0x7514: "HANJA",
        0x7515: "HANJA",
        0x7516: "HANJA",
        0x7517: "HANJA",
        0x7518: "HANJA",
        0x7519: "HANJA",
        0x751A: "HANJA",
        0x751B: "HANJA",
        0x751C: "HANJA",
        0x751D: "HANJA",
        0x751E: "HANJA",
        0x751F: "HANJA",
        0x7520: "HANJA",
        0x7521: "HANJA",
        0x7522: "HANJA",
        0x7523: "HANJA",
        0x7524: "HANJA",
        0x7525: "HANJA",
        0x7526: "HANJA",
        0x7527: "HANJA",
        0x7528: "HANJA",
        0x7529: "HANJA",
        0x752A: "HANJA",
        0x752B: "HANJA",
        0x752C: "HANJA",
        0x752D: "HANJA",
        0x752E: "HANJA",
        0x752F: "HANJA",
        0x7530: "HANJA",
        0x7531: "HANJA",
        0x7532: "HANJA",
        0x7533: "HANJA",
        0x7534: "HANJA",
        0x7535: "HANJA",
        0x7536: "HANJA",
        0x7537: "HANJA",
        0x7538: "HANJA",
        0x7539: "HANJA",
        0x753A: "HANJA",
        0x753B: "HANJA",
        0x753C: "HANJA",
        0x753D: "HANJA",
        0x753E: "HANJA",
        0x753F: "HANJA",
        0x7540: "HANJA",
        0x7541: "HANJA",
        0x7542: "HANJA",
        0x7543: "HANJA",
        0x7544: "HANJA",
        0x7545: "HANJA",
        0x7546: "HANJA",
        0x7547: "HANJA",
        0x7548: "HANJA",
        0x7549: "HANJA",
        0x754A: "HANJA",
        0x754B: "HANJA",
        0x754C: "HANJA",
        0x754D: "HANJA",
        0x754E: "HANJA",
        0x754F: "HANJA",
        0x7550: "HANJA",
        0x7551: "HANJA",
        0x7552: "HANJA",
        0x7553: "HANJA",
        0x7554: "HANJA",
        0x7555: "HANJA",
        0x7556: "HANJA",
        0x7557: "HANJA",
        0x7558: "HANJA",
        0x7559: "HANJA",
        0x755A: "HANJA",
        0x755B: "HANJA",
        0x755C: "HANJA",
        0x755D: "HANJA",
        0x755E: "HANJA",
        0x755F: "HANJA",
        0x7560: "HANJA",
        0x7561: "HANJA",
        0x7562: "HANJA",
        0x7563: "HANJA",
        0x7564: "HANJA",
        0x7565: "HANJA",
        0x7566: "HANJA",
        0x7567: "HANJA",
        0x7568: "HANJA",
        0x7569: "HANJA",
        0x756A: "HANJA",
        0x756B: "HANJA",
        0x756C: "HANJA",
        0x756D: "HANJA",
        0x756E: "HANJA",
        0x756F: "HANJA",
        0x7570: "HANJA",
        0x7571: "HANJA",
        0x7572: "HANJA",
        0x7573: "HANJA",
        0x7574: "HANJA",
        0x7575: "HANJA",
        0x7576: "HANJA",
        0x7577: "HANJA",
        0x7578: "HANJA",
        0x7579: "HANJA",
        0x757A: "HANJA",
        0x757B: "HANJA",
        0x757C: "HANJA",
        0x757D: "HANJA",
        0x757E: "HANJA",
        0x757F: "HANJA",
        0x7580: "HANJA",
        0x7581: "HANJA",
        0x7582: "HANJA",
        0x7583: "HANJA",
        0x7584: "HANJA",
        0x7585: "HANJA",
        0x7586: "HANJA",
        0x7587: "HANJA",
        0x7588: "HANJA",
        0x7589: "HANJA",
        0x758A: "HANJA",
        0x758B: "HANJA",
        0x758C: "HANJA",
        0x758D: "HANJA",
        0x758E: "HANJA",
        0x758F: "HANJA",
        0x7590: "HANJA",
        0x7591: "HANJA",
        0x7592: "HANJA",
        0x7593: "HANJA",
        0x7594: "HANJA",
        0x7595: "HANJA",
        0x7596: "HANJA",
        0x7597: "HANJA",
        0x7598: "HANJA",
        0x7599: "HANJA",
        0x759A: "HANJA",
        0x759B: "HANJA",
        0x759C: "HANJA",
        0x759D: "HANJA",
        0x759E: "HANJA",
        0x759F: "HANJA",
        0x75A0: "HANJA",
        0x75A1: "HANJA",
        0x75A2: "HANJA",
        0x75A3: "HANJA",
        0x75A4: "HANJA",
        0x75A5: "HANJA",
        0x75A6: "HANJA",
        0x75A7: "HANJA",
        0x75A8: "HANJA",
        0x75A9: "HANJA",
        0x75AA: "HANJA",
        0x75AB: "HANJA",
        0x75AC: "HANJA",
        0x75AD: "HANJA",
        0x75AE: "HANJA",
        0x75AF: "HANJA",
        0x75B0: "HANJA",
        0x75B1: "HANJA",
        0x75B2: "HANJA",
        0x75B3: "HANJA",
        0x75B4: "HANJA",
        0x75B5: "HANJA",
        0x75B6: "HANJA",
        0x75B7: "HANJA",
        0x75B8: "HANJA",
        0x75B9: "HANJA",
        0x75BA: "HANJA",
        0x75BB: "HANJA",
        0x75BC: "HANJA",
        0x75BD: "HANJA",
        0x75BE: "HANJA",
        0x75BF: "HANJA",
        0x75C0: "HANJA",
        0x75C1: "HANJA",
        0x75C2: "HANJA",
        0x75C3: "HANJA",
        0x75C4: "HANJA",
        0x75C5: "HANJA",
        0x75C6: "HANJA",
        0x75C7: "HANJA",
        0x75C8: "HANJA",
        0x75C9: "HANJA",
        0x75CA: "HANJA",
        0x75CB: "HANJA",
        0x75CC: "HANJA",
        0x75CD: "HANJA",
        0x75CE: "HANJA",
        0x75CF: "HANJA",
        0x75D0: "HANJA",
        0x75D1: "HANJA",
        0x75D2: "HANJA",
        0x75D3: "HANJA",
        0x75D4: "HANJA",
        0x75D5: "HANJA",
        0x75D6: "HANJA",
        0x75D7: "HANJA",
        0x75D8: "HANJA",
        0x75D9: "HANJA",
        0x75DA: "HANJA",
        0x75DB: "HANJA",
        0x75DC: "HANJA",
        0x75DD: "HANJA",
        0x75DE: "HANJA",
        0x75DF: "HANJA",
        0x75E0: "HANJA",
        0x75E1: "HANJA",
        0x75E2: "HANJA",
        0x75E3: "HANJA",
        0x75E4: "HANJA",
        0x75E5: "HANJA",
        0x75E6: "HANJA",
        0x75E7: "HANJA",
        0x75E8: "HANJA",
        0x75E9: "HANJA",
        0x75EA: "HANJA",
        0x75EB: "HANJA",
        0x75EC: "HANJA",
        0x75ED: "HANJA",
        0x75EE: "HANJA",
        0x75EF: "HANJA",
        0x75F0: "HANJA",
        0x75F1: "HANJA",
        0x75F2: "HANJA",
        0x75F3: "HANJA",
        0x75F4: "HANJA",
        0x75F5: "HANJA",
        0x75F6: "HANJA",
        0x75F7: "HANJA",
        0x75F8: "HANJA",
        0x75F9: "HANJA",
        0x75FA: "HANJA",
        0x75FB: "HANJA",
        0x75FC: "HANJA",
        0x75FD: "HANJA",
        0x75FE: "HANJA",
        0x75FF: "HANJA",
        0x7600: "HANJA",
        0x7601: "HANJA",
        0x7602: "HANJA",
        0x7603: "HANJA",
        0x7604: "HANJA",
        0x7605: "HANJA",
        0x7606: "HANJA",
        0x7607: "HANJA",
        0x7608: "HANJA",
        0x7609: "HANJA",
        0x760A: "HANJA",
        0x760B: "HANJA",
        0x760C: "HANJA",
        0x760D: "HANJA",
        0x760E: "HANJA",
        0x760F: "HANJA",
        0x7610: "HANJA",
        0x7611: "HANJA",
        0x7612: "HANJA",
        0x7613: "HANJA",
        0x7614: "HANJA",
        0x7615: "HANJA",
        0x7616: "HANJA",
        0x7617: "HANJA",
        0x7618: "HANJA",
        0x7619: "HANJA",
        0x761A: "HANJA",
        0x761B: "HANJA",
        0x761C: "HANJA",
        0x761D: "HANJA",
        0x761E: "HANJA",
        0x761F: "HANJA",
        0x7620: "HANJA",
        0x7621: "HANJA",
        0x7622: "HANJA",
        0x7623: "HANJA",
        0x7624: "HANJA",
        0x7625: "HANJA",
        0x7626: "HANJA",
        0x7627: "HANJA",
        0x7628: "HANJA",
        0x7629: "HANJA",
        0x762A: "HANJA",
        0x762B: "HANJA",
        0x762C: "HANJA",
        0x762D: "HANJA",
        0x762E: "HANJA",
        0x762F: "HANJA",
        0x7630: "HANJA",
        0x7631: "HANJA",
        0x7632: "HANJA",
        0x7633: "HANJA",
        0x7634: "HANJA",
        0x7635: "HANJA",
        0x7636: "HANJA",
        0x7637: "HANJA",
        0x7638: "HANJA",
        0x7639: "HANJA",
        0x763A: "HANJA",
        0x763B: "HANJA",
        0x763C: "HANJA",
        0x763D: "HANJA",
        0x763E: "HANJA",
        0x763F: "HANJA",
        0x7640: "HANJA",
        0x7641: "HANJA",
        0x7642: "HANJA",
        0x7643: "HANJA",
        0x7644: "HANJA",
        0x7645: "HANJA",
        0x7646: "HANJA",
        0x7647: "HANJA",
        0x7648: "HANJA",
        0x7649: "HANJA",
        0x764A: "HANJA",
        0x764B: "HANJA",
        0x764C: "HANJA",
        0x764D: "HANJA",
        0x764E: "HANJA",
        0x764F: "HANJA",
        0x7650: "HANJA",
        0x7651: "HANJA",
        0x7652: "HANJA",
        0x7653: "HANJA",
        0x7654: "HANJA",
        0x7655: "HANJA",
        0x7656: "HANJA",
        0x7657: "HANJA",
        0x7658: "HANJA",
        0x7659: "HANJA",
        0x765A: "HANJA",
        0x765B: "HANJA",
        0x765C: "HANJA",
        0x765D: "HANJA",
        0x765E: "HANJA",
        0x765F: "HANJA",
        0x7660: "HANJA",
        0x7661: "HANJA",
        0x7662: "HANJA",
        0x7663: "HANJA",
        0x7664: "HANJA",
        0x7665: "HANJA",
        0x7666: "HANJA",
        0x7667: "HANJA",
        0x7668: "HANJA",
        0x7669: "HANJA",
        0x766A: "HANJA",
        0x766B: "HANJA",
        0x766C: "HANJA",
        0x766D: "HANJA",
        0x766E: "HANJA",
        0x766F: "HANJA",
        0x7670: "HANJA",
        0x7671: "HANJA",
        0x7672: "HANJA",
        0x7673: "HANJA",
        0x7674: "HANJA",
        0x7675: "HANJA",
        0x7676: "HANJA",
        0x7677: "HANJA",
        0x7678: "HANJA",
        0x7679: "HANJA",
        0x767A: "HANJA",
        0x767B: "HANJA",
        0x767C: "HANJA",
        0x767D: "HANJA",
        0x767E: "HANJA",
        0x767F: "HANJA",
        0x7680: "HANJA",
        0x7681: "HANJA",
        0x7682: "HANJA",
        0x7683: "HANJA",
        0x7684: "HANJA",
        0x7685: "HANJA",
        0x7686: "HANJA",
        0x7687: "HANJA",
        0x7688: "HANJA",
        0x7689: "HANJA",
        0x768A: "HANJA",
        0x768B: "HANJA",
        0x768C: "HANJA",
        0x768D: "HANJA",
        0x768E: "HANJA",
        0x768F: "HANJA",
        0x7690: "HANJA",
        0x7691: "HANJA",
        0x7692: "HANJA",
        0x7693: "HANJA",
        0x7694: "HANJA",
        0x7695: "HANJA",
        0x7696: "HANJA",
        0x7697: "HANJA",
        0x7698: "HANJA",
        0x7699: "HANJA",
        0x769A: "HANJA",
        0x769B: "HANJA",
        0x769C: "HANJA",
        0x769D: "HANJA",
        0x769E: "HANJA",
        0x769F: "HANJA",
        0x76A0: "HANJA",
        0x76A1: "HANJA",
        0x76A2: "HANJA",
        0x76A3: "HANJA",
        0x76A4: "HANJA",
        0x76A5: "HANJA",
        0x76A6: "HANJA",
        0x76A7: "HANJA",
        0x76A8: "HANJA",
        0x76A9: "HANJA",
        0x76AA: "HANJA",
        0x76AB: "HANJA",
        0x76AC: "HANJA",
        0x76AD: "HANJA",
        0x76AE: "HANJA",
        0x76AF: "HANJA",
        0x76B0: "HANJA",
        0x76B1: "HANJA",
        0x76B2: "HANJA",
        0x76B3: "HANJA",
        0x76B4: "HANJA",
        0x76B5: "HANJA",
        0x76B6: "HANJA",
        0x76B7: "HANJA",
        0x76B8: "HANJA",
        0x76B9: "HANJA",
        0x76BA: "HANJA",
        0x76BB: "HANJA",
        0x76BC: "HANJA",
        0x76BD: "HANJA",
        0x76BE: "HANJA",
        0x76BF: "HANJA",
        0x76C0: "HANJA",
        0x76C1: "HANJA",
        0x76C2: "HANJA",
        0x76C3: "HANJA",
        0x76C4: "HANJA",
        0x76C5: "HANJA",
        0x76C6: "HANJA",
        0x76C7: "HANJA",
        0x76C8: "HANJA",
        0x76C9: "HANJA",
        0x76CA: "HANJA",
        0x76CB: "HANJA",
        0x76CC: "HANJA",
        0x76CD: "HANJA",
        0x76CE: "HANJA",
        0x76CF: "HANJA",
        0x76D0: "HANJA",
        0x76D1: "HANJA",
        0x76D2: "HANJA",
        0x76D3: "HANJA",
        0x76D4: "HANJA",
        0x76D5: "HANJA",
        0x76D6: "HANJA",
        0x76D7: "HANJA",
        0x76D8: "HANJA",
        0x76D9: "HANJA",
        0x76DA: "HANJA",
        0x76DB: "HANJA",
        0x76DC: "HANJA",
        0x76DD: "HANJA",
        0x76DE: "HANJA",
        0x76DF: "HANJA",
        0x76E0: "HANJA",
        0x76E1: "HANJA",
        0x76E2: "HANJA",
        0x76E3: "HANJA",
        0x76E4: "HANJA",
        0x76E5: "HANJA",
        0x76E6: "HANJA",
        0x76E7: "HANJA",
        0x76E8: "HANJA",
        0x76E9: "HANJA",
        0x76EA: "HANJA",
        0x76EB: "HANJA",
        0x76EC: "HANJA",
        0x76ED: "HANJA",
        0x76EE: "HANJA",
        0x76EF: "HANJA",
        0x76F0: "HANJA",
        0x76F1: "HANJA",
        0x76F2: "HANJA",
        0x76F3: "HANJA",
        0x76F4: "HANJA",
        0x76F5: "HANJA",
        0x76F6: "HANJA",
        0x76F7: "HANJA",
        0x76F8: "HANJA",
        0x76F9: "HANJA",
        0x76FA: "HANJA",
        0x76FB: "HANJA",
        0x76FC: "HANJA",
        0x76FD: "HANJA",
        0x76FE: "HANJA",
        0x76FF: "HANJA",
        0x7700: "HANJA",
        0x7701: "HANJA",
        0x7702: "HANJA",
        0x7703: "HANJA",
        0x7704: "HANJA",
        0x7705: "HANJA",
        0x7706: "HANJA",
        0x7707: "HANJA",
        0x7708: "HANJA",
        0x7709: "HANJA",
        0x770A: "HANJA",
        0x770B: "HANJA",
        0x770C: "HANJA",
        0x770D: "HANJA",
        0x770E: "HANJA",
        0x770F: "HANJA",
        0x7710: "HANJA",
        0x7711: "HANJA",
        0x7712: "HANJA",
        0x7713: "HANJA",
        0x7714: "HANJA",
        0x7715: "HANJA",
        0x7716: "HANJA",
        0x7717: "HANJA",
        0x7718: "HANJA",
        0x7719: "HANJA",
        0x771A: "HANJA",
        0x771B: "HANJA",
        0x771C: "HANJA",
        0x771D: "HANJA",
        0x771E: "HANJA",
        0x771F: "HANJA",
        0x7720: "HANJA",
        0x7721: "HANJA",
        0x7722: "HANJA",
        0x7723: "HANJA",
        0x7724: "HANJA",
        0x7725: "HANJA",
        0x7726: "HANJA",
        0x7727: "HANJA",
        0x7728: "HANJA",
        0x7729: "HANJA",
        0x772A: "HANJA",
        0x772B: "HANJA",
        0x772C: "HANJA",
        0x772D: "HANJA",
        0x772E: "HANJA",
        0x772F: "HANJA",
        0x7730: "HANJA",
        0x7731: "HANJA",
        0x7732: "HANJA",
        0x7733: "HANJA",
        0x7734: "HANJA",
        0x7735: "HANJA",
        0x7736: "HANJA",
        0x7737: "HANJA",
        0x7738: "HANJA",
        0x7739: "HANJA",
        0x773A: "HANJA",
        0x773B: "HANJA",
        0x773C: "HANJA",
        0x773D: "HANJA",
        0x773E: "HANJA",
        0x773F: "HANJA",
        0x7740: "HANJA",
        0x7741: "HANJA",
        0x7742: "HANJA",
        0x7743: "HANJA",
        0x7744: "HANJA",
        0x7745: "HANJA",
        0x7746: "HANJA",
        0x7747: "HANJA",
        0x7748: "HANJA",
        0x7749: "HANJA",
        0x774A: "HANJA",
        0x774B: "HANJA",
        0x774C: "HANJA",
        0x774D: "HANJA",
        0x774E: "HANJA",
        0x774F: "HANJA",
        0x7750: "HANJA",
        0x7751: "HANJA",
        0x7752: "HANJA",
        0x7753: "HANJA",
        0x7754: "HANJA",
        0x7755: "HANJA",
        0x7756: "HANJA",
        0x7757: "HANJA",
        0x7758: "HANJA",
        0x7759: "HANJA",
        0x775A: "HANJA",
        0x775B: "HANJA",
        0x775C: "HANJA",
        0x775D: "HANJA",
        0x775E: "HANJA",
        0x775F: "HANJA",
        0x7760: "HANJA",
        0x7761: "HANJA",
        0x7762: "HANJA",
        0x7763: "HANJA",
        0x7764: "HANJA",
        0x7765: "HANJA",
        0x7766: "HANJA",
        0x7767: "HANJA",
        0x7768: "HANJA",
        0x7769: "HANJA",
        0x776A: "HANJA",
        0x776B: "HANJA",
        0x776C: "HANJA",
        0x776D: "HANJA",
        0x776E: "HANJA",
        0x776F: "HANJA",
        0x7770: "HANJA",
        0x7771: "HANJA",
        0x7772: "HANJA",
        0x7773: "HANJA",
        0x7774: "HANJA",
        0x7775: "HANJA",
        0x7776: "HANJA",
        0x7777: "HANJA",
        0x7778: "HANJA",
        0x7779: "HANJA",
        0x777A: "HANJA",
        0x777B: "HANJA",
        0x777C: "HANJA",
        0x777D: "HANJA",
        0x777E: "HANJA",
        0x777F: "HANJA",
        0x7780: "HANJA",
        0x7781: "HANJA",
        0x7782: "HANJA",
        0x7783: "HANJA",
        0x7784: "HANJA",
        0x7785: "HANJA",
        0x7786: "HANJA",
        0x7787: "HANJA",
        0x7788: "HANJA",
        0x7789: "HANJA",
        0x778A: "HANJA",
        0x778B: "HANJA",
        0x778C: "HANJA",
        0x778D: "HANJA",
        0x778E: "HANJA",
        0x778F: "HANJA",
        0x7790: "HANJA",
        0x7791: "HANJA",
        0x7792: "HANJA",
        0x7793: "HANJA",
        0x7794: "HANJA",
        0x7795: "HANJA",
        0x7796: "HANJA",
        0x7797: "HANJA",
        0x7798: "HANJA",
        0x7799: "HANJA",
        0x779A: "HANJA",
        0x779B: "HANJA",
        0x779C: "HANJA",
        0x779D: "HANJA",
        0x779E: "HANJA",
        0x779F: "HANJA",
        0x77A0: "HANJA",
        0x77A1: "HANJA",
        0x77A2: "HANJA",
        0x77A3: "HANJA",
        0x77A4: "HANJA",
        0x77A5: "HANJA",
        0x77A6: "HANJA",
        0x77A7: "HANJA",
        0x77A8: "HANJA",
        0x77A9: "HANJA",
        0x77AA: "HANJA",
        0x77AB: "HANJA",
        0x77AC: "HANJA",
        0x77AD: "HANJA",
        0x77AE: "HANJA",
        0x77AF: "HANJA",
        0x77B0: "HANJA",
        0x77B1: "HANJA",
        0x77B2: "HANJA",
        0x77B3: "HANJA",
        0x77B4: "HANJA",
        0x77B5: "HANJA",
        0x77B6: "HANJA",
        0x77B7: "HANJA",
        0x77B8: "HANJA",
        0x77B9: "HANJA",
        0x77BA: "HANJA",
        0x77BB: "HANJA",
        0x77BC: "HANJA",
        0x77BD: "HANJA",
        0x77BE: "HANJA",
        0x77BF: "HANJA",
        0x77C0: "HANJA",
        0x77C1: "HANJA",
        0x77C2: "HANJA",
        0x77C3: "HANJA",
        0x77C4: "HANJA",
        0x77C5: "HANJA",
        0x77C6: "HANJA",
        0x77C7: "HANJA",
        0x77C8: "HANJA",
        0x77C9: "HANJA",
        0x77CA: "HANJA",
        0x77CB: "HANJA",
        0x77CC: "HANJA",
        0x77CD: "HANJA",
        0x77CE: "HANJA",
        0x77CF: "HANJA",
        0x77D0: "HANJA",
        0x77D1: "HANJA",
        0x77D2: "HANJA",
        0x77D3: "HANJA",
        0x77D4: "HANJA",
        0x77D5: "HANJA",
        0x77D6: "HANJA",
        0x77D7: "HANJA",
        0x77D8: "HANJA",
        0x77D9: "HANJA",
        0x77DA: "HANJA",
        0x77DB: "HANJA",
        0x77DC: "HANJA",
        0x77DD: "HANJA",
        0x77DE: "HANJA",
        0x77DF: "HANJA",
        0x77E0: "HANJA",
        0x77E1: "HANJA",
        0x77E2: "HANJA",
        0x77E3: "HANJA",
        0x77E4: "HANJA",
        0x77E5: "HANJA",
        0x77E6: "HANJA",
        0x77E7: "HANJA",
        0x77E8: "HANJA",
        0x77E9: "HANJA",
        0x77EA: "HANJA",
        0x77EB: "HANJA",
        0x77EC: "HANJA",
        0x77ED: "HANJA",
        0x77EE: "HANJA",
        0x77EF: "HANJA",
        0x77F0: "HANJA",
        0x77F1: "HANJA",
        0x77F2: "HANJA",
        0x77F3: "HANJA",
        0x77F4: "HANJA",
        0x77F5: "HANJA",
        0x77F6: "HANJA",
        0x77F7: "HANJA",
        0x77F8: "HANJA",
        0x77F9: "HANJA",
        0x77FA: "HANJA",
        0x77FB: "HANJA",
        0x77FC: "HANJA",
        0x77FD: "HANJA",
        0x77FE: "HANJA",
        0x77FF: "HANJA",
        0x7800: "HANJA",
        0x7801: "HANJA",
        0x7802: "HANJA",
        0x7803: "HANJA",
        0x7804: "HANJA",
        0x7805: "HANJA",
        0x7806: "HANJA",
        0x7807: "HANJA",
        0x7808: "HANJA",
        0x7809: "HANJA",
        0x780A: "HANJA",
        0x780B: "HANJA",
        0x780C: "HANJA",
        0x780D: "HANJA",
        0x780E: "HANJA",
        0x780F: "HANJA",
        0x7810: "HANJA",
        0x7811: "HANJA",
        0x7812: "HANJA",
        0x7813: "HANJA",
        0x7814: "HANJA",
        0x7815: "HANJA",
        0x7816: "HANJA",
        0x7817: "HANJA",
        0x7818: "HANJA",
        0x7819: "HANJA",
        0x781A: "HANJA",
        0x781B: "HANJA",
        0x781C: "HANJA",
        0x781D: "HANJA",
        0x781E: "HANJA",
        0x781F: "HANJA",
        0x7820: "HANJA",
        0x7821: "HANJA",
        0x7822: "HANJA",
        0x7823: "HANJA",
        0x7824: "HANJA",
        0x7825: "HANJA",
        0x7826: "HANJA",
        0x7827: "HANJA",
        0x7828: "HANJA",
        0x7829: "HANJA",
        0x782A: "HANJA",
        0x782B: "HANJA",
        0x782C: "HANJA",
        0x782D: "HANJA",
        0x782E: "HANJA",
        0x782F: "HANJA",
        0x7830: "HANJA",
        0x7831: "HANJA",
        0x7832: "HANJA",
        0x7833: "HANJA",
        0x7834: "HANJA",
        0x7835: "HANJA",
        0x7836: "HANJA",
        0x7837: "HANJA",
        0x7838: "HANJA",
        0x7839: "HANJA",
        0x783A: "HANJA",
        0x783B: "HANJA",
        0x783C: "HANJA",
        0x783D: "HANJA",
        0x783E: "HANJA",
        0x783F: "HANJA",
        0x7840: "HANJA",
        0x7841: "HANJA",
        0x7842: "HANJA",
        0x7843: "HANJA",
        0x7844: "HANJA",
        0x7845: "HANJA",
        0x7846: "HANJA",
        0x7847: "HANJA",
        0x7848: "HANJA",
        0x7849: "HANJA",
        0x784A: "HANJA",
        0x784B: "HANJA",
        0x784C: "HANJA",
        0x784D: "HANJA",
        0x784E: "HANJA",
        0x784F: "HANJA",
        0x7850: "HANJA",
        0x7851: "HANJA",
        0x7852: "HANJA",
        0x7853: "HANJA",
        0x7854: "HANJA",
        0x7855: "HANJA",
        0x7856: "HANJA",
        0x7857: "HANJA",
        0x7858: "HANJA",
        0x7859: "HANJA",
        0x785A: "HANJA",
        0x785B: "HANJA",
        0x785C: "HANJA",
        0x785D: "HANJA",
        0x785E: "HANJA",
        0x785F: "HANJA",
        0x7860: "HANJA",
        0x7861: "HANJA",
        0x7862: "HANJA",
        0x7863: "HANJA",
        0x7864: "HANJA",
        0x7865: "HANJA",
        0x7866: "HANJA",
        0x7867: "HANJA",
        0x7868: "HANJA",
        0x7869: "HANJA",
        0x786A: "HANJA",
        0x786B: "HANJA",
        0x786C: "HANJA",
        0x786D: "HANJA",
        0x786E: "HANJA",
        0x786F: "HANJA",
        0x7870: "HANJA",
        0x7871: "HANJA",
        0x7872: "HANJA",
        0x7873: "HANJA",
        0x7874: "HANJA",
        0x7875: "HANJA",
        0x7876: "HANJA",
        0x7877: "HANJA",
        0x7878: "HANJA",
        0x7879: "HANJA",
        0x787A: "HANJA",
        0x787B: "HANJA",
        0x787C: "HANJA",
        0x787D: "HANJA",
        0x787E: "HANJA",
        0x787F: "HANJA",
        0x7880: "HANJA",
        0x7881: "HANJA",
        0x7882: "HANJA",
        0x7883: "HANJA",
        0x7884: "HANJA",
        0x7885: "HANJA",
        0x7886: "HANJA",
        0x7887: "HANJA",
        0x7888: "HANJA",
        0x7889: "HANJA",
        0x788A: "HANJA",
        0x788B: "HANJA",
        0x788C: "HANJA",
        0x788D: "HANJA",
        0x788E: "HANJA",
        0x788F: "HANJA",
        0x7890: "HANJA",
        0x7891: "HANJA",
        0x7892: "HANJA",
        0x7893: "HANJA",
        0x7894: "HANJA",
        0x7895: "HANJA",
        0x7896: "HANJA",
        0x7897: "HANJA",
        0x7898: "HANJA",
        0x7899: "HANJA",
        0x789A: "HANJA",
        0x789B: "HANJA",
        0x789C: "HANJA",
        0x789D: "HANJA",
        0x789E: "HANJA",
        0x789F: "HANJA",
        0x78A0: "HANJA",
        0x78A1: "HANJA",
        0x78A2: "HANJA",
        0x78A3: "HANJA",
        0x78A4: "HANJA",
        0x78A5: "HANJA",
        0x78A6: "HANJA",
        0x78A7: "HANJA",
        0x78A8: "HANJA",
        0x78A9: "HANJA",
        0x78AA: "HANJA",
        0x78AB: "HANJA",
        0x78AC: "HANJA",
        0x78AD: "HANJA",
        0x78AE: "HANJA",
        0x78AF: "HANJA",
        0x78B0: "HANJA",
        0x78B1: "HANJA",
        0x78B2: "HANJA",
        0x78B3: "HANJA",
        0x78B4: "HANJA",
        0x78B5: "HANJA",
        0x78B6: "HANJA",
        0x78B7: "HANJA",
        0x78B8: "HANJA",
        0x78B9: "HANJA",
        0x78BA: "HANJA",
        0x78BB: "HANJA",
        0x78BC: "HANJA",
        0x78BD: "HANJA",
        0x78BE: "HANJA",
        0x78BF: "HANJA",
        0x78C0: "HANJA",
        0x78C1: "HANJA",
        0x78C2: "HANJA",
        0x78C3: "HANJA",
        0x78C4: "HANJA",
        0x78C5: "HANJA",
        0x78C6: "HANJA",
        0x78C7: "HANJA",
        0x78C8: "HANJA",
        0x78C9: "HANJA",
        0x78CA: "HANJA",
        0x78CB: "HANJA",
        0x78CC: "HANJA",
        0x78CD: "HANJA",
        0x78CE: "HANJA",
        0x78CF: "HANJA",
        0x78D0: "HANJA",
        0x78D1: "HANJA",
        0x78D2: "HANJA",
        0x78D3: "HANJA",
        0x78D4: "HANJA",
        0x78D5: "HANJA",
        0x78D6: "HANJA",
        0x78D7: "HANJA",
        0x78D8: "HANJA",
        0x78D9: "HANJA",
        0x78DA: "HANJA",
        0x78DB: "HANJA",
        0x78DC: "HANJA",
        0x78DD: "HANJA",
        0x78DE: "HANJA",
        0x78DF: "HANJA",
        0x78E0: "HANJA",
        0x78E1: "HANJA",
        0x78E2: "HANJA",
        0x78E3: "HANJA",
        0x78E4: "HANJA",
        0x78E5: "HANJA",
        0x78E6: "HANJA",
        0x78E7: "HANJA",
        0x78E8: "HANJA",
        0x78E9: "HANJA",
        0x78EA: "HANJA",
        0x78EB: "HANJA",
        0x78EC: "HANJA",
        0x78ED: "HANJA",
        0x78EE: "HANJA",
        0x78EF: "HANJA",
        0x78F0: "HANJA",
        0x78F1: "HANJA",
        0x78F2: "HANJA",
        0x78F3: "HANJA",
        0x78F4: "HANJA",
        0x78F5: "HANJA",
        0x78F6: "HANJA",
        0x78F7: "HANJA",
        0x78F8: "HANJA",
        0x78F9: "HANJA",
        0x78FA: "HANJA",
        0x78FB: "HANJA",
        0x78FC: "HANJA",
        0x78FD: "HANJA",
        0x78FE: "HANJA",
        0x78FF: "HANJA",
        0x7900: "HANJA",
        0x7901: "HANJA",
        0x7902: "HANJA",
        0x7903: "HANJA",
        0x7904: "HANJA",
        0x7905: "HANJA",
        0x7906: "HANJA",
        0x7907: "HANJA",
        0x7908: "HANJA",
        0x7909: "HANJA",
        0x790A: "HANJA",
        0x790B: "HANJA",
        0x790C: "HANJA",
        0x790D: "HANJA",
        0x790E: "HANJA",
        0x790F: "HANJA",
        0x7910: "HANJA",
        0x7911: "HANJA",
        0x7912: "HANJA",
        0x7913: "HANJA",
        0x7914: "HANJA",
        0x7915: "HANJA",
        0x7916: "HANJA",
        0x7917: "HANJA",
        0x7918: "HANJA",
        0x7919: "HANJA",
        0x791A: "HANJA",
        0x791B: "HANJA",
        0x791C: "HANJA",
        0x791D: "HANJA",
        0x791E: "HANJA",
        0x791F: "HANJA",
        0x7920: "HANJA",
        0x7921: "HANJA",
        0x7922: "HANJA",
        0x7923: "HANJA",
        0x7924: "HANJA",
        0x7925: "HANJA",
        0x7926: "HANJA",
        0x7927: "HANJA",
        0x7928: "HANJA",
        0x7929: "HANJA",
        0x792A: "HANJA",
        0x792B: "HANJA",
        0x792C: "HANJA",
        0x792D: "HANJA",
        0x792E: "HANJA",
        0x792F: "HANJA",
        0x7930: "HANJA",
        0x7931: "HANJA",
        0x7932: "HANJA",
        0x7933: "HANJA",
        0x7934: "HANJA",
        0x7935: "HANJA",
        0x7936: "HANJA",
        0x7937: "HANJA",
        0x7938: "HANJA",
        0x7939: "HANJA",
        0x793A: "HANJA",
        0x793B: "HANJA",
        0x793C: "HANJA",
        0x793D: "HANJA",
        0x793E: "HANJA",
        0x793F: "HANJA",
        0x7940: "HANJA",
        0x7941: "HANJA",
        0x7942: "HANJA",
        0x7943: "HANJA",
        0x7944: "HANJA",
        0x7945: "HANJA",
        0x7946: "HANJA",
        0x7947: "HANJA",
        0x7948: "HANJA",
        0x7949: "HANJA",
        0x794A: "HANJA",
        0x794B: "HANJA",
        0x794C: "HANJA",
        0x794D: "HANJA",
        0x794E: "HANJA",
        0x794F: "HANJA",
        0x7950: "HANJA",
        0x7951: "HANJA",
        0x7952: "HANJA",
        0x7953: "HANJA",
        0x7954: "HANJA",
        0x7955: "HANJA",
        0x7956: "HANJA",
        0x7957: "HANJA",
        0x7958: "HANJA",
        0x7959: "HANJA",
        0x795A: "HANJA",
        0x795B: "HANJA",
        0x795C: "HANJA",
        0x795D: "HANJA",
        0x795E: "HANJA",
        0x795F: "HANJA",
        0x7960: "HANJA",
        0x7961: "HANJA",
        0x7962: "HANJA",
        0x7963: "HANJA",
        0x7964: "HANJA",
        0x7965: "HANJA",
        0x7966: "HANJA",
        0x7967: "HANJA",
        0x7968: "HANJA",
        0x7969: "HANJA",
        0x796A: "HANJA",
        0x796B: "HANJA",
        0x796C: "HANJA",
        0x796D: "HANJA",
        0x796E: "HANJA",
        0x796F: "HANJA",
        0x7970: "HANJA",
        0x7971: "HANJA",
        0x7972: "HANJA",
        0x7973: "HANJA",
        0x7974: "HANJA",
        0x7975: "HANJA",
        0x7976: "HANJA",
        0x7977: "HANJA",
        0x7978: "HANJA",
        0x7979: "HANJA",
        0x797A: "HANJA",
        0x797B: "HANJA",
        0x797C: "HANJA",
        0x797D: "HANJA",
        0x797E: "HANJA",
        0x797F: "HANJA",
        0x7980: "HANJA",
        0x7981: "HANJA",
        0x7982: "HANJA",
        0x7983: "HANJA",
        0x7984: "HANJA",
        0x7985: "HANJA",
        0x7986: "HANJA",
        0x7987: "HANJA",
        0x7988: "HANJA",
        0x7989: "HANJA",
        0x798A: "HANJA",
        0x798B: "HANJA",
        0x798C: "HANJA",
        0x798D: "HANJA",
        0x798E: "HANJA",
        0x798F: "HANJA",
        0x7990: "HANJA",
        0x7991: "HANJA",
        0x7992: "HANJA",
        0x7993: "HANJA",
        0x7994: "HANJA",
        0x7995: "HANJA",
        0x7996: "HANJA",
        0x7997: "HANJA",
        0x7998: "HANJA",
        0x7999: "HANJA",
        0x799A: "HANJA",
        0x799B: "HANJA",
        0x799C: "HANJA",
        0x799D: "HANJA",
        0x799E: "HANJA",
        0x799F: "HANJA",
        0x79A0: "HANJA",
        0x79A1: "HANJA",
        0x79A2: "HANJA",
        0x79A3: "HANJA",
        0x79A4: "HANJA",
        0x79A5: "HANJA",
        0x79A6: "HANJA",
        0x79A7: "HANJA",
        0x79A8: "HANJA",
        0x79A9: "HANJA",
        0x79AA: "HANJA",
        0x79AB: "HANJA",
        0x79AC: "HANJA",
        0x79AD: "HANJA",
        0x79AE: "HANJA",
        0x79AF: "HANJA",
        0x79B0: "HANJA",
        0x79B1: "HANJA",
        0x79B2: "HANJA",
        0x79B3: "HANJA",
        0x79B4: "HANJA",
        0x79B5: "HANJA",
        0x79B6: "HANJA",
        0x79B7: "HANJA",
        0x79B8: "HANJA",
        0x79B9: "HANJA",
        0x79BA: "HANJA",
        0x79BB: "HANJA",
        0x79BC: "HANJA",
        0x79BD: "HANJA",
        0x79BE: "HANJA",
        0x79BF: "HANJA",
        0x79C0: "HANJA",
        0x79C1: "HANJA",
        0x79C2: "HANJA",
        0x79C3: "HANJA",
        0x79C4: "HANJA",
        0x79C5: "HANJA",
        0x79C6: "HANJA",
        0x79C7: "HANJA",
        0x79C8: "HANJA",
        0x79C9: "HANJA",
        0x79CA: "HANJA",
        0x79CB: "HANJA",
        0x79CC: "HANJA",
        0x79CD: "HANJA",
        0x79CE: "HANJA",
        0x79CF: "HANJA",
        0x79D0: "HANJA",
        0x79D1: "HANJA",
        0x79D2: "HANJA",
        0x79D3: "HANJA",
        0x79D4: "HANJA",
        0x79D5: "HANJA",
        0x79D6: "HANJA",
        0x79D7: "HANJA",
        0x79D8: "HANJA",
        0x79D9: "HANJA",
        0x79DA: "HANJA",
        0x79DB: "HANJA",
        0x79DC: "HANJA",
        0x79DD: "HANJA",
        0x79DE: "HANJA",
        0x79DF: "HANJA",
        0x79E0: "HANJA",
        0x79E1: "HANJA",
        0x79E2: "HANJA",
        0x79E3: "HANJA",
        0x79E4: "HANJA",
        0x79E5: "HANJA",
        0x79E6: "HANJA",
        0x79E7: "HANJA",
        0x79E8: "HANJA",
        0x79E9: "HANJA",
        0x79EA: "HANJA",
        0x79EB: "HANJA",
        0x79EC: "HANJA",
        0x79ED: "HANJA",
        0x79EE: "HANJA",
        0x79EF: "HANJA",
        0x79F0: "HANJA",
        0x79F1: "HANJA",
        0x79F2: "HANJA",
        0x79F3: "HANJA",
        0x79F4: "HANJA",
        0x79F5: "HANJA",
        0x79F6: "HANJA",
        0x79F7: "HANJA",
        0x79F8: "HANJA",
        0x79F9: "HANJA",
        0x79FA: "HANJA",
        0x79FB: "HANJA",
        0x79FC: "HANJA",
        0x79FD: "HANJA",
        0x79FE: "HANJA",
        0x79FF: "HANJA",
        0x7A00: "HANJA",
        0x7A01: "HANJA",
        0x7A02: "HANJA",
        0x7A03: "HANJA",
        0x7A04: "HANJA",
        0x7A05: "HANJA",
        0x7A06: "HANJA",
        0x7A07: "HANJA",
        0x7A08: "HANJA",
        0x7A09: "HANJA",
        0x7A0A: "HANJA",
        0x7A0B: "HANJA",
        0x7A0C: "HANJA",
        0x7A0D: "HANJA",
        0x7A0E: "HANJA",
        0x7A0F: "HANJA",
        0x7A10: "HANJA",
        0x7A11: "HANJA",
        0x7A12: "HANJA",
        0x7A13: "HANJA",
        0x7A14: "HANJA",
        0x7A15: "HANJA",
        0x7A16: "HANJA",
        0x7A17: "HANJA",
        0x7A18: "HANJA",
        0x7A19: "HANJA",
        0x7A1A: "HANJA",
        0x7A1B: "HANJA",
        0x7A1C: "HANJA",
        0x7A1D: "HANJA",
        0x7A1E: "HANJA",
        0x7A1F: "HANJA",
        0x7A20: "HANJA",
        0x7A21: "HANJA",
        0x7A22: "HANJA",
        0x7A23: "HANJA",
        0x7A24: "HANJA",
        0x7A25: "HANJA",
        0x7A26: "HANJA",
        0x7A27: "HANJA",
        0x7A28: "HANJA",
        0x7A29: "HANJA",
        0x7A2A: "HANJA",
        0x7A2B: "HANJA",
        0x7A2C: "HANJA",
        0x7A2D: "HANJA",
        0x7A2E: "HANJA",
        0x7A2F: "HANJA",
        0x7A30: "HANJA",
        0x7A31: "HANJA",
        0x7A32: "HANJA",
        0x7A33: "HANJA",
        0x7A34: "HANJA",
        0x7A35: "HANJA",
        0x7A36: "HANJA",
        0x7A37: "HANJA",
        0x7A38: "HANJA",
        0x7A39: "HANJA",
        0x7A3A: "HANJA",
        0x7A3B: "HANJA",
        0x7A3C: "HANJA",
        0x7A3D: "HANJA",
        0x7A3E: "HANJA",
        0x7A3F: "HANJA",
        0x7A40: "HANJA",
        0x7A41: "HANJA",
        0x7A42: "HANJA",
        0x7A43: "HANJA",
        0x7A44: "HANJA",
        0x7A45: "HANJA",
        0x7A46: "HANJA",
        0x7A47: "HANJA",
        0x7A48: "HANJA",
        0x7A49: "HANJA",
        0x7A4A: "HANJA",
        0x7A4B: "HANJA",
        0x7A4C: "HANJA",
        0x7A4D: "HANJA",
        0x7A4E: "HANJA",
        0x7A4F: "HANJA",
        0x7A50: "HANJA",
        0x7A51: "HANJA",
        0x7A52: "HANJA",
        0x7A53: "HANJA",
        0x7A54: "HANJA",
        0x7A55: "HANJA",
        0x7A56: "HANJA",
        0x7A57: "HANJA",
        0x7A58: "HANJA",
        0x7A59: "HANJA",
        0x7A5A: "HANJA",
        0x7A5B: "HANJA",
        0x7A5C: "HANJA",
        0x7A5D: "HANJA",
        0x7A5E: "HANJA",
        0x7A5F: "HANJA",
        0x7A60: "HANJA",
        0x7A61: "HANJA",
        0x7A62: "HANJA",
        0x7A63: "HANJA",
        0x7A64: "HANJA",
        0x7A65: "HANJA",
        0x7A66: "HANJA",
        0x7A67: "HANJA",
        0x7A68: "HANJA",
        0x7A69: "HANJA",
        0x7A6A: "HANJA",
        0x7A6B: "HANJA",
        0x7A6C: "HANJA",
        0x7A6D: "HANJA",
        0x7A6E: "HANJA",
        0x7A6F: "HANJA",
        0x7A70: "HANJA",
        0x7A71: "HANJA",
        0x7A72: "HANJA",
        0x7A73: "HANJA",
        0x7A74: "HANJA",
        0x7A75: "HANJA",
        0x7A76: "HANJA",
        0x7A77: "HANJA",
        0x7A78: "HANJA",
        0x7A79: "HANJA",
        0x7A7A: "HANJA",
        0x7A7B: "HANJA",
        0x7A7C: "HANJA",
        0x7A7D: "HANJA",
        0x7A7E: "HANJA",
        0x7A7F: "HANJA",
        0x7A80: "HANJA",
        0x7A81: "HANJA",
        0x7A82: "HANJA",
        0x7A83: "HANJA",
        0x7A84: "HANJA",
        0x7A85: "HANJA",
        0x7A86: "HANJA",
        0x7A87: "HANJA",
        0x7A88: "HANJA",
        0x7A89: "HANJA",
        0x7A8A: "HANJA",
        0x7A8B: "HANJA",
        0x7A8C: "HANJA",
        0x7A8D: "HANJA",
        0x7A8E: "HANJA",
        0x7A8F: "HANJA",
        0x7A90: "HANJA",
        0x7A91: "HANJA",
        0x7A92: "HANJA",
        0x7A93: "HANJA",
        0x7A94: "HANJA",
        0x7A95: "HANJA",
        0x7A96: "HANJA",
        0x7A97: "HANJA",
        0x7A98: "HANJA",
        0x7A99: "HANJA",
        0x7A9A: "HANJA",
        0x7A9B: "HANJA",
        0x7A9C: "HANJA",
        0x7A9D: "HANJA",
        0x7A9E: "HANJA",
        0x7A9F: "HANJA",
        0x7AA0: "HANJA",
        0x7AA1: "HANJA",
        0x7AA2: "HANJA",
        0x7AA3: "HANJA",
        0x7AA4: "HANJA",
        0x7AA5: "HANJA",
        0x7AA6: "HANJA",
        0x7AA7: "HANJA",
        0x7AA8: "HANJA",
        0x7AA9: "HANJA",
        0x7AAA: "HANJA",
        0x7AAB: "HANJA",
        0x7AAC: "HANJA",
        0x7AAD: "HANJA",
        0x7AAE: "HANJA",
        0x7AAF: "HANJA",
        0x7AB0: "HANJA",
        0x7AB1: "HANJA",
        0x7AB2: "HANJA",
        0x7AB3: "HANJA",
        0x7AB4: "HANJA",
        0x7AB5: "HANJA",
        0x7AB6: "HANJA",
        0x7AB7: "HANJA",
        0x7AB8: "HANJA",
        0x7AB9: "HANJA",
        0x7ABA: "HANJA",
        0x7ABB: "HANJA",
        0x7ABC: "HANJA",
        0x7ABD: "HANJA",
        0x7ABE: "HANJA",
        0x7ABF: "HANJA",
        0x7AC0: "HANJA",
        0x7AC1: "HANJA",
        0x7AC2: "HANJA",
        0x7AC3: "HANJA",
        0x7AC4: "HANJA",
        0x7AC5: "HANJA",
        0x7AC6: "HANJA",
        0x7AC7: "HANJA",
        0x7AC8: "HANJA",
        0x7AC9: "HANJA",
        0x7ACA: "HANJA",
        0x7ACB: "HANJA",
        0x7ACC: "HANJA",
        0x7ACD: "HANJA",
        0x7ACE: "HANJA",
        0x7ACF: "HANJA",
        0x7AD0: "HANJA",
        0x7AD1: "HANJA",
        0x7AD2: "HANJA",
        0x7AD3: "HANJA",
        0x7AD4: "HANJA",
        0x7AD5: "HANJA",
        0x7AD6: "HANJA",
        0x7AD7: "HANJA",
        0x7AD8: "HANJA",
        0x7AD9: "HANJA",
        0x7ADA: "HANJA",
        0x7ADB: "HANJA",
        0x7ADC: "HANJA",
        0x7ADD: "HANJA",
        0x7ADE: "HANJA",
        0x7ADF: "HANJA",
        0x7AE0: "HANJA",
        0x7AE1: "HANJA",
        0x7AE2: "HANJA",
        0x7AE3: "HANJA",
        0x7AE4: "HANJA",
        0x7AE5: "HANJA",
        0x7AE6: "HANJA",
        0x7AE7: "HANJA",
        0x7AE8: "HANJA",
        0x7AE9: "HANJA",
        0x7AEA: "HANJA",
        0x7AEB: "HANJA",
        0x7AEC: "HANJA",
        0x7AED: "HANJA",
        0x7AEE: "HANJA",
        0x7AEF: "HANJA",
        0x7AF0: "HANJA",
        0x7AF1: "HANJA",
        0x7AF2: "HANJA",
        0x7AF3: "HANJA",
        0x7AF4: "HANJA",
        0x7AF5: "HANJA",
        0x7AF6: "HANJA",
        0x7AF7: "HANJA",
        0x7AF8: "HANJA",
        0x7AF9: "HANJA",
        0x7AFA: "HANJA",
        0x7AFB: "HANJA",
        0x7AFC: "HANJA",
        0x7AFD: "HANJA",
        0x7AFE: "HANJA",
        0x7AFF: "HANJA",
        0x7B00: "HANJA",
        0x7B01: "HANJA",
        0x7B02: "HANJA",
        0x7B03: "HANJA",
        0x7B04: "HANJA",
        0x7B05: "HANJA",
        0x7B06: "HANJA",
        0x7B07: "HANJA",
        0x7B08: "HANJA",
        0x7B09: "HANJA",
        0x7B0A: "HANJA",
        0x7B0B: "HANJA",
        0x7B0C: "HANJA",
        0x7B0D: "HANJA",
        0x7B0E: "HANJA",
        0x7B0F: "HANJA",
        0x7B10: "HANJA",
        0x7B11: "HANJA",
        0x7B12: "HANJA",
        0x7B13: "HANJA",
        0x7B14: "HANJA",
        0x7B15: "HANJA",
        0x7B16: "HANJA",
        0x7B17: "HANJA",
        0x7B18: "HANJA",
        0x7B19: "HANJA",
        0x7B1A: "HANJA",
        0x7B1B: "HANJA",
        0x7B1C: "HANJA",
        0x7B1D: "HANJA",
        0x7B1E: "HANJA",
        0x7B1F: "HANJA",
        0x7B20: "HANJA",
        0x7B21: "HANJA",
        0x7B22: "HANJA",
        0x7B23: "HANJA",
        0x7B24: "HANJA",
        0x7B25: "HANJA",
        0x7B26: "HANJA",
        0x7B27: "HANJA",
        0x7B28: "HANJA",
        0x7B29: "HANJA",
        0x7B2A: "HANJA",
        0x7B2B: "HANJA",
        0x7B2C: "HANJA",
        0x7B2D: "HANJA",
        0x7B2E: "HANJA",
        0x7B2F: "HANJA",
        0x7B30: "HANJA",
        0x7B31: "HANJA",
        0x7B32: "HANJA",
        0x7B33: "HANJA",
        0x7B34: "HANJA",
        0x7B35: "HANJA",
        0x7B36: "HANJA",
        0x7B37: "HANJA",
        0x7B38: "HANJA",
        0x7B39: "HANJA",
        0x7B3A: "HANJA",
        0x7B3B: "HANJA",
        0x7B3C: "HANJA",
        0x7B3D: "HANJA",
        0x7B3E: "HANJA",
        0x7B3F: "HANJA",
        0x7B40: "HANJA",
        0x7B41: "HANJA",
        0x7B42: "HANJA",
        0x7B43: "HANJA",
        0x7B44: "HANJA",
        0x7B45: "HANJA",
        0x7B46: "HANJA",
        0x7B47: "HANJA",
        0x7B48: "HANJA",
        0x7B49: "HANJA",
        0x7B4A: "HANJA",
        0x7B4B: "HANJA",
        0x7B4C: "HANJA",
        0x7B4D: "HANJA",
        0x7B4E: "HANJA",
        0x7B4F: "HANJA",
        0x7B50: "HANJA",
        0x7B51: "HANJA",
        0x7B52: "HANJA",
        0x7B53: "HANJA",
        0x7B54: "HANJA",
        0x7B55: "HANJA",
        0x7B56: "HANJA",
        0x7B57: "HANJA",
        0x7B58: "HANJA",
        0x7B59: "HANJA",
        0x7B5A: "HANJA",
        0x7B5B: "HANJA",
        0x7B5C: "HANJA",
        0x7B5D: "HANJA",
        0x7B5E: "HANJA",
        0x7B5F: "HANJA",
        0x7B60: "HANJA",
        0x7B61: "HANJA",
        0x7B62: "HANJA",
        0x7B63: "HANJA",
        0x7B64: "HANJA",
        0x7B65: "HANJA",
        0x7B66: "HANJA",
        0x7B67: "HANJA",
        0x7B68: "HANJA",
        0x7B69: "HANJA",
        0x7B6A: "HANJA",
        0x7B6B: "HANJA",
        0x7B6C: "HANJA",
        0x7B6D: "HANJA",
        0x7B6E: "HANJA",
        0x7B6F: "HANJA",
        0x7B70: "HANJA",
        0x7B71: "HANJA",
        0x7B72: "HANJA",
        0x7B73: "HANJA",
        0x7B74: "HANJA",
        0x7B75: "HANJA",
        0x7B76: "HANJA",
        0x7B77: "HANJA",
        0x7B78: "HANJA",
        0x7B79: "HANJA",
        0x7B7A: "HANJA",
        0x7B7B: "HANJA",
        0x7B7C: "HANJA",
        0x7B7D: "HANJA",
        0x7B7E: "HANJA",
        0x7B7F: "HANJA",
        0x7B80: "HANJA",
        0x7B81: "HANJA",
        0x7B82: "HANJA",
        0x7B83: "HANJA",
        0x7B84: "HANJA",
        0x7B85: "HANJA",
        0x7B86: "HANJA",
        0x7B87: "HANJA",
        0x7B88: "HANJA",
        0x7B89: "HANJA",
        0x7B8A: "HANJA",
        0x7B8B: "HANJA",
        0x7B8C: "HANJA",
        0x7B8D: "HANJA",
        0x7B8E: "HANJA",
        0x7B8F: "HANJA",
        0x7B90: "HANJA",
        0x7B91: "HANJA",
        0x7B92: "HANJA",
        0x7B93: "HANJA",
        0x7B94: "HANJA",
        0x7B95: "HANJA",
        0x7B96: "HANJA",
        0x7B97: "HANJA",
        0x7B98: "HANJA",
        0x7B99: "HANJA",
        0x7B9A: "HANJA",
        0x7B9B: "HANJA",
        0x7B9C: "HANJA",
        0x7B9D: "HANJA",
        0x7B9E: "HANJA",
        0x7B9F: "HANJA",
        0x7BA0: "HANJA",
        0x7BA1: "HANJA",
        0x7BA2: "HANJA",
        0x7BA3: "HANJA",
        0x7BA4: "HANJA",
        0x7BA5: "HANJA",
        0x7BA6: "HANJA",
        0x7BA7: "HANJA",
        0x7BA8: "HANJA",
        0x7BA9: "HANJA",
        0x7BAA: "HANJA",
        0x7BAB: "HANJA",
        0x7BAC: "HANJA",
        0x7BAD: "HANJA",
        0x7BAE: "HANJA",
        0x7BAF: "HANJA",
        0x7BB0: "HANJA",
        0x7BB1: "HANJA",
        0x7BB2: "HANJA",
        0x7BB3: "HANJA",
        0x7BB4: "HANJA",
        0x7BB5: "HANJA",
        0x7BB6: "HANJA",
        0x7BB7: "HANJA",
        0x7BB8: "HANJA",
        0x7BB9: "HANJA",
        0x7BBA: "HANJA",
        0x7BBB: "HANJA",
        0x7BBC: "HANJA",
        0x7BBD: "HANJA",
        0x7BBE: "HANJA",
        0x7BBF: "HANJA",
        0x7BC0: "HANJA",
        0x7BC1: "HANJA",
        0x7BC2: "HANJA",
        0x7BC3: "HANJA",
        0x7BC4: "HANJA",
        0x7BC5: "HANJA",
        0x7BC6: "HANJA",
        0x7BC7: "HANJA",
        0x7BC8: "HANJA",
        0x7BC9: "HANJA",
        0x7BCA: "HANJA",
        0x7BCB: "HANJA",
        0x7BCC: "HANJA",
        0x7BCD: "HANJA",
        0x7BCE: "HANJA",
        0x7BCF: "HANJA",
        0x7BD0: "HANJA",
        0x7BD1: "HANJA",
        0x7BD2: "HANJA",
        0x7BD3: "HANJA",
        0x7BD4: "HANJA",
        0x7BD5: "HANJA",
        0x7BD6: "HANJA",
        0x7BD7: "HANJA",
        0x7BD8: "HANJA",
        0x7BD9: "HANJA",
        0x7BDA: "HANJA",
        0x7BDB: "HANJA",
        0x7BDC: "HANJA",
        0x7BDD: "HANJA",
        0x7BDE: "HANJA",
        0x7BDF: "HANJA",
        0x7BE0: "HANJA",
        0x7BE1: "HANJA",
        0x7BE2: "HANJA",
        0x7BE3: "HANJA",
        0x7BE4: "HANJA",
        0x7BE5: "HANJA",
        0x7BE6: "HANJA",
        0x7BE7: "HANJA",
        0x7BE8: "HANJA",
        0x7BE9: "HANJA",
        0x7BEA: "HANJA",
        0x7BEB: "HANJA",
        0x7BEC: "HANJA",
        0x7BED: "HANJA",
        0x7BEE: "HANJA",
        0x7BEF: "HANJA",
        0x7BF0: "HANJA",
        0x7BF1: "HANJA",
        0x7BF2: "HANJA",
        0x7BF3: "HANJA",
        0x7BF4: "HANJA",
        0x7BF5: "HANJA",
        0x7BF6: "HANJA",
        0x7BF7: "HANJA",
        0x7BF8: "HANJA",
        0x7BF9: "HANJA",
        0x7BFA: "HANJA",
        0x7BFB: "HANJA",
        0x7BFC: "HANJA",
        0x7BFD: "HANJA",
        0x7BFE: "HANJA",
        0x7BFF: "HANJA",
        0x7C00: "HANJA",
        0x7C01: "HANJA",
        0x7C02: "HANJA",
        0x7C03: "HANJA",
        0x7C04: "HANJA",
        0x7C05: "HANJA",
        0x7C06: "HANJA",
        0x7C07: "HANJA",
        0x7C08: "HANJA",
        0x7C09: "HANJA",
        0x7C0A: "HANJA",
        0x7C0B: "HANJA",
        0x7C0C: "HANJA",
        0x7C0D: "HANJA",
        0x7C0E: "HANJA",
        0x7C0F: "HANJA",
        0x7C10: "HANJA",
        0x7C11: "HANJA",
        0x7C12: "HANJA",
        0x7C13: "HANJA",
        0x7C14: "HANJA",
        0x7C15: "HANJA",
        0x7C16: "HANJA",
        0x7C17: "HANJA",
        0x7C18: "HANJA",
        0x7C19: "HANJA",
        0x7C1A: "HANJA",
        0x7C1B: "HANJA",
        0x7C1C: "HANJA",
        0x7C1D: "HANJA",
        0x7C1E: "HANJA",
        0x7C1F: "HANJA",
        0x7C20: "HANJA",
        0x7C21: "HANJA",
        0x7C22: "HANJA",
        0x7C23: "HANJA",
        0x7C24: "HANJA",
        0x7C25: "HANJA",
        0x7C26: "HANJA",
        0x7C27: "HANJA",
        0x7C28: "HANJA",
        0x7C29: "HANJA",
        0x7C2A: "HANJA",
        0x7C2B: "HANJA",
        0x7C2C: "HANJA",
        0x7C2D: "HANJA",
        0x7C2E: "HANJA",
        0x7C2F: "HANJA",
        0x7C30: "HANJA",
        0x7C31: "HANJA",
        0x7C32: "HANJA",
        0x7C33: "HANJA",
        0x7C34: "HANJA",
        0x7C35: "HANJA",
        0x7C36: "HANJA",
        0x7C37: "HANJA",
        0x7C38: "HANJA",
        0x7C39: "HANJA",
        0x7C3A: "HANJA",
        0x7C3B: "HANJA",
        0x7C3C: "HANJA",
        0x7C3D: "HANJA",
        0x7C3E: "HANJA",
        0x7C3F: "HANJA",
        0x7C40: "HANJA",
        0x7C41: "HANJA",
        0x7C42: "HANJA",
        0x7C43: "HANJA",
        0x7C44: "HANJA",
        0x7C45: "HANJA",
        0x7C46: "HANJA",
        0x7C47: "HANJA",
        0x7C48: "HANJA",
        0x7C49: "HANJA",
        0x7C4A: "HANJA",
        0x7C4B: "HANJA",
        0x7C4C: "HANJA",
        0x7C4D: "HANJA",
        0x7C4E: "HANJA",
        0x7C4F: "HANJA",
        0x7C50: "HANJA",
        0x7C51: "HANJA",
        0x7C52: "HANJA",
        0x7C53: "HANJA",
        0x7C54: "HANJA",
        0x7C55: "HANJA",
        0x7C56: "HANJA",
        0x7C57: "HANJA",
        0x7C58: "HANJA",
        0x7C59: "HANJA",
        0x7C5A: "HANJA",
        0x7C5B: "HANJA",
        0x7C5C: "HANJA",
        0x7C5D: "HANJA",
        0x7C5E: "HANJA",
        0x7C5F: "HANJA",
        0x7C60: "HANJA",
        0x7C61: "HANJA",
        0x7C62: "HANJA",
        0x7C63: "HANJA",
        0x7C64: "HANJA",
        0x7C65: "HANJA",
        0x7C66: "HANJA",
        0x7C67: "HANJA",
        0x7C68: "HANJA",
        0x7C69: "HANJA",
        0x7C6A: "HANJA",
        0x7C6B: "HANJA",
        0x7C6C: "HANJA",
        0x7C6D: "HANJA",
        0x7C6E: "HANJA",
        0x7C6F: "HANJA",
        0x7C70: "HANJA",
        0x7C71: "HANJA",
        0x7C72: "HANJA",
        0x7C73: "HANJA",
        0x7C74: "HANJA",
        0x7C75: "HANJA",
        0x7C76: "HANJA",
        0x7C77: "HANJA",
        0x7C78: "HANJA",
        0x7C79: "HANJA",
        0x7C7A: "HANJA",
        0x7C7B: "HANJA",
        0x7C7C: "HANJA",
        0x7C7D: "HANJA",
        0x7C7E: "HANJA",
        0x7C7F: "HANJA",
        0x7C80: "HANJA",
        0x7C81: "HANJA",
        0x7C82: "HANJA",
        0x7C83: "HANJA",
        0x7C84: "HANJA",
        0x7C85: "HANJA",
        0x7C86: "HANJA",
        0x7C87: "HANJA",
        0x7C88: "HANJA",
        0x7C89: "HANJA",
        0x7C8A: "HANJA",
        0x7C8B: "HANJA",
        0x7C8C: "HANJA",
        0x7C8D: "HANJA",
        0x7C8E: "HANJA",
        0x7C8F: "HANJA",
        0x7C90: "HANJA",
        0x7C91: "HANJA",
        0x7C92: "HANJA",
        0x7C93: "HANJA",
        0x7C94: "HANJA",
        0x7C95: "HANJA",
        0x7C96: "HANJA",
        0x7C97: "HANJA",
        0x7C98: "HANJA",
        0x7C99: "HANJA",
        0x7C9A: "HANJA",
        0x7C9B: "HANJA",
        0x7C9C: "HANJA",
        0x7C9D: "HANJA",
        0x7C9E: "HANJA",
        0x7C9F: "HANJA",
        0x7CA0: "HANJA",
        0x7CA1: "HANJA",
        0x7CA2: "HANJA",
        0x7CA3: "HANJA",
        0x7CA4: "HANJA",
        0x7CA5: "HANJA",
        0x7CA6: "HANJA",
        0x7CA7: "HANJA",
        0x7CA8: "HANJA",
        0x7CA9: "HANJA",
        0x7CAA: "HANJA",
        0x7CAB: "HANJA",
        0x7CAC: "HANJA",
        0x7CAD: "HANJA",
        0x7CAE: "HANJA",
        0x7CAF: "HANJA",
        0x7CB0: "HANJA",
        0x7CB1: "HANJA",
        0x7CB2: "HANJA",
        0x7CB3: "HANJA",
        0x7CB4: "HANJA",
        0x7CB5: "HANJA",
        0x7CB6: "HANJA",
        0x7CB7: "HANJA",
        0x7CB8: "HANJA",
        0x7CB9: "HANJA",
        0x7CBA: "HANJA",
        0x7CBB: "HANJA",
        0x7CBC: "HANJA",
        0x7CBD: "HANJA",
        0x7CBE: "HANJA",
        0x7CBF: "HANJA",
        0x7CC0: "HANJA",
        0x7CC1: "HANJA",
        0x7CC2: "HANJA",
        0x7CC3: "HANJA",
        0x7CC4: "HANJA",
        0x7CC5: "HANJA",
        0x7CC6: "HANJA",
        0x7CC7: "HANJA",
        0x7CC8: "HANJA",
        0x7CC9: "HANJA",
        0x7CCA: "HANJA",
        0x7CCB: "HANJA",
        0x7CCC: "HANJA",
        0x7CCD: "HANJA",
        0x7CCE: "HANJA",
        0x7CCF: "HANJA",
        0x7CD0: "HANJA",
        0x7CD1: "HANJA",
        0x7CD2: "HANJA",
        0x7CD3: "HANJA",
        0x7CD4: "HANJA",
        0x7CD5: "HANJA",
        0x7CD6: "HANJA",
        0x7CD7: "HANJA",
        0x7CD8: "HANJA",
        0x7CD9: "HANJA",
        0x7CDA: "HANJA",
        0x7CDB: "HANJA",
        0x7CDC: "HANJA",
        0x7CDD: "HANJA",
        0x7CDE: "HANJA",
        0x7CDF: "HANJA",
        0x7CE0: "HANJA",
        0x7CE1: "HANJA",
        0x7CE2: "HANJA",
        0x7CE3: "HANJA",
        0x7CE4: "HANJA",
        0x7CE5: "HANJA",
        0x7CE6: "HANJA",
        0x7CE7: "HANJA",
        0x7CE8: "HANJA",
        0x7CE9: "HANJA",
        0x7CEA: "HANJA",
        0x7CEB: "HANJA",
        0x7CEC: "HANJA",
        0x7CED: "HANJA",
        0x7CEE: "HANJA",
        0x7CEF: "HANJA",
        0x7CF0: "HANJA",
        0x7CF1: "HANJA",
        0x7CF2: "HANJA",
        0x7CF3: "HANJA",
        0x7CF4: "HANJA",
        0x7CF5: "HANJA",
        0x7CF6: "HANJA",
        0x7CF7: "HANJA",
        0x7CF8: "HANJA",
        0x7CF9: "HANJA",
        0x7CFA: "HANJA",
        0x7CFB: "HANJA",
        0x7CFC: "HANJA",
        0x7CFD: "HANJA",
        0x7CFE: "HANJA",
        0x7CFF: "HANJA",
        0x7D00: "HANJA",
        0x7D01: "HANJA",
        0x7D02: "HANJA",
        0x7D03: "HANJA",
        0x7D04: "HANJA",
        0x7D05: "HANJA",
        0x7D06: "HANJA",
        0x7D07: "HANJA",
        0x7D08: "HANJA",
        0x7D09: "HANJA",
        0x7D0A: "HANJA",
        0x7D0B: "HANJA",
        0x7D0C: "HANJA",
        0x7D0D: "HANJA",
        0x7D0E: "HANJA",
        0x7D0F: "HANJA",
        0x7D10: "HANJA",
        0x7D11: "HANJA",
        0x7D12: "HANJA",
        0x7D13: "HANJA",
        0x7D14: "HANJA",
        0x7D15: "HANJA",
        0x7D16: "HANJA",
        0x7D17: "HANJA",
        0x7D18: "HANJA",
        0x7D19: "HANJA",
        0x7D1A: "HANJA",
        0x7D1B: "HANJA",
        0x7D1C: "HANJA",
        0x7D1D: "HANJA",
        0x7D1E: "HANJA",
        0x7D1F: "HANJA",
        0x7D20: "HANJA",
        0x7D21: "HANJA",
        0x7D22: "HANJA",
        0x7D23: "HANJA",
        0x7D24: "HANJA",
        0x7D25: "HANJA",
        0x7D26: "HANJA",
        0x7D27: "HANJA",
        0x7D28: "HANJA",
        0x7D29: "HANJA",
        0x7D2A: "HANJA",
        0x7D2B: "HANJA",
        0x7D2C: "HANJA",
        0x7D2D: "HANJA",
        0x7D2E: "HANJA",
        0x7D2F: "HANJA",
        0x7D30: "HANJA",
        0x7D31: "HANJA",
        0x7D32: "HANJA",
        0x7D33: "HANJA",
        0x7D34: "HANJA",
        0x7D35: "HANJA",
        0x7D36: "HANJA",
        0x7D37: "HANJA",
        0x7D38: "HANJA",
        0x7D39: "HANJA",
        0x7D3A: "HANJA",
        0x7D3B: "HANJA",
        0x7D3C: "HANJA",
        0x7D3D: "HANJA",
        0x7D3E: "HANJA",
        0x7D3F: "HANJA",
        0x7D40: "HANJA",
        0x7D41: "HANJA",
        0x7D42: "HANJA",
        0x7D43: "HANJA",
        0x7D44: "HANJA",
        0x7D45: "HANJA",
        0x7D46: "HANJA",
        0x7D47: "HANJA",
        0x7D48: "HANJA",
        0x7D49: "HANJA",
        0x7D4A: "HANJA",
        0x7D4B: "HANJA",
        0x7D4C: "HANJA",
        0x7D4D: "HANJA",
        0x7D4E: "HANJA",
        0x7D4F: "HANJA",
        0x7D50: "HANJA",
        0x7D51: "HANJA",
        0x7D52: "HANJA",
        0x7D53: "HANJA",
        0x7D54: "HANJA",
        0x7D55: "HANJA",
        0x7D56: "HANJA",
        0x7D57: "HANJA",
        0x7D58: "HANJA",
        0x7D59: "HANJA",
        0x7D5A: "HANJA",
        0x7D5B: "HANJA",
        0x7D5C: "HANJA",
        0x7D5D: "HANJA",
        0x7D5E: "HANJA",
        0x7D5F: "HANJA",
        0x7D60: "HANJA",
        0x7D61: "HANJA",
        0x7D62: "HANJA",
        0x7D63: "HANJA",
        0x7D64: "HANJA",
        0x7D65: "HANJA",
        0x7D66: "HANJA",
        0x7D67: "HANJA",
        0x7D68: "HANJA",
        0x7D69: "HANJA",
        0x7D6A: "HANJA",
        0x7D6B: "HANJA",
        0x7D6C: "HANJA",
        0x7D6D: "HANJA",
        0x7D6E: "HANJA",
        0x7D6F: "HANJA",
        0x7D70: "HANJA",
        0x7D71: "HANJA",
        0x7D72: "HANJA",
        0x7D73: "HANJA",
        0x7D74: "HANJA",
        0x7D75: "HANJA",
        0x7D76: "HANJA",
        0x7D77: "HANJA",
        0x7D78: "HANJA",
        0x7D79: "HANJA",
        0x7D7A: "HANJA",
        0x7D7B: "HANJA",
        0x7D7C: "HANJA",
        0x7D7D: "HANJA",
        0x7D7E: "HANJA",
        0x7D7F: "HANJA",
        0x7D80: "HANJA",
        0x7D81: "HANJA",
        0x7D82: "HANJA",
        0x7D83: "HANJA",
        0x7D84: "HANJA",
        0x7D85: "HANJA",
        0x7D86: "HANJA",
        0x7D87: "HANJA",
        0x7D88: "HANJA",
        0x7D89: "HANJA",
        0x7D8A: "HANJA",
        0x7D8B: "HANJA",
        0x7D8C: "HANJA",
        0x7D8D: "HANJA",
        0x7D8E: "HANJA",
        0x7D8F: "HANJA",
        0x7D90: "HANJA",
        0x7D91: "HANJA",
        0x7D92: "HANJA",
        0x7D93: "HANJA",
        0x7D94: "HANJA",
        0x7D95: "HANJA",
        0x7D96: "HANJA",
        0x7D97: "HANJA",
        0x7D98: "HANJA",
        0x7D99: "HANJA",
        0x7D9A: "HANJA",
        0x7D9B: "HANJA",
        0x7D9C: "HANJA",
        0x7D9D: "HANJA",
        0x7D9E: "HANJA",
        0x7D9F: "HANJA",
        0x7DA0: "HANJA",
        0x7DA1: "HANJA",
        0x7DA2: "HANJA",
        0x7DA3: "HANJA",
        0x7DA4: "HANJA",
        0x7DA5: "HANJA",
        0x7DA6: "HANJA",
        0x7DA7: "HANJA",
        0x7DA8: "HANJA",
        0x7DA9: "HANJA",
        0x7DAA: "HANJA",
        0x7DAB: "HANJA",
        0x7DAC: "HANJA",
        0x7DAD: "HANJA",
        0x7DAE: "HANJA",
        0x7DAF: "HANJA",
        0x7DB0: "HANJA",
        0x7DB1: "HANJA",
        0x7DB2: "HANJA",
        0x7DB3: "HANJA",
        0x7DB4: "HANJA",
        0x7DB5: "HANJA",
        0x7DB6: "HANJA",
        0x7DB7: "HANJA",
        0x7DB8: "HANJA",
        0x7DB9: "HANJA",
        0x7DBA: "HANJA",
        0x7DBB: "HANJA",
        0x7DBC: "HANJA",
        0x7DBD: "HANJA",
        0x7DBE: "HANJA",
        0x7DBF: "HANJA",
        0x7DC0: "HANJA",
        0x7DC1: "HANJA",
        0x7DC2: "HANJA",
        0x7DC3: "HANJA",
        0x7DC4: "HANJA",
        0x7DC5: "HANJA",
        0x7DC6: "HANJA",
        0x7DC7: "HANJA",
        0x7DC8: "HANJA",
        0x7DC9: "HANJA",
        0x7DCA: "HANJA",
        0x7DCB: "HANJA",
        0x7DCC: "HANJA",
        0x7DCD: "HANJA",
        0x7DCE: "HANJA",
        0x7DCF: "HANJA",
        0x7DD0: "HANJA",
        0x7DD1: "HANJA",
        0x7DD2: "HANJA",
        0x7DD3: "HANJA",
        0x7DD4: "HANJA",
        0x7DD5: "HANJA",
        0x7DD6: "HANJA",
        0x7DD7: "HANJA",
        0x7DD8: "HANJA",
        0x7DD9: "HANJA",
        0x7DDA: "HANJA",
        0x7DDB: "HANJA",
        0x7DDC: "HANJA",
        0x7DDD: "HANJA",
        0x7DDE: "HANJA",
        0x7DDF: "HANJA",
        0x7DE0: "HANJA",
        0x7DE1: "HANJA",
        0x7DE2: "HANJA",
        0x7DE3: "HANJA",
        0x7DE4: "HANJA",
        0x7DE5: "HANJA",
        0x7DE6: "HANJA",
        0x7DE7: "HANJA",
        0x7DE8: "HANJA",
        0x7DE9: "HANJA",
        0x7DEA: "HANJA",
        0x7DEB: "HANJA",
        0x7DEC: "HANJA",
        0x7DED: "HANJA",
        0x7DEE: "HANJA",
        0x7DEF: "HANJA",
        0x7DF0: "HANJA",
        0x7DF1: "HANJA",
        0x7DF2: "HANJA",
        0x7DF3: "HANJA",
        0x7DF4: "HANJA",
        0x7DF5: "HANJA",
        0x7DF6: "HANJA",
        0x7DF7: "HANJA",
        0x7DF8: "HANJA",
        0x7DF9: "HANJA",
        0x7DFA: "HANJA",
        0x7DFB: "HANJA",
        0x7DFC: "HANJA",
        0x7DFD: "HANJA",
        0x7DFE: "HANJA",
        0x7DFF: "HANJA",
        0x7E00: "HANJA",
        0x7E01: "HANJA",
        0x7E02: "HANJA",
        0x7E03: "HANJA",
        0x7E04: "HANJA",
        0x7E05: "HANJA",
        0x7E06: "HANJA",
        0x7E07: "HANJA",
        0x7E08: "HANJA",
        0x7E09: "HANJA",
        0x7E0A: "HANJA",
        0x7E0B: "HANJA",
        0x7E0C: "HANJA",
        0x7E0D: "HANJA",
        0x7E0E: "HANJA",
        0x7E0F: "HANJA",
        0x7E10: "HANJA",
        0x7E11: "HANJA",
        0x7E12: "HANJA",
        0x7E13: "HANJA",
        0x7E14: "HANJA",
        0x7E15: "HANJA",
        0x7E16: "HANJA",
        0x7E17: "HANJA",
        0x7E18: "HANJA",
        0x7E19: "HANJA",
        0x7E1A: "HANJA",
        0x7E1B: "HANJA",
        0x7E1C: "HANJA",
        0x7E1D: "HANJA",
        0x7E1E: "HANJA",
        0x7E1F: "HANJA",
        0x7E20: "HANJA",
        0x7E21: "HANJA",
        0x7E22: "HANJA",
        0x7E23: "HANJA",
        0x7E24: "HANJA",
        0x7E25: "HANJA",
        0x7E26: "HANJA",
        0x7E27: "HANJA",
        0x7E28: "HANJA",
        0x7E29: "HANJA",
        0x7E2A: "HANJA",
        0x7E2B: "HANJA",
        0x7E2C: "HANJA",
        0x7E2D: "HANJA",
        0x7E2E: "HANJA",
        0x7E2F: "HANJA",
        0x7E30: "HANJA",
        0x7E31: "HANJA",
        0x7E32: "HANJA",
        0x7E33: "HANJA",
        0x7E34: "HANJA",
        0x7E35: "HANJA",
        0x7E36: "HANJA",
        0x7E37: "HANJA",
        0x7E38: "HANJA",
        0x7E39: "HANJA",
        0x7E3A: "HANJA",
        0x7E3B: "HANJA",
        0x7E3C: "HANJA",
        0x7E3D: "HANJA",
        0x7E3E: "HANJA",
        0x7E3F: "HANJA",
        0x7E40: "HANJA",
        0x7E41: "HANJA",
        0x7E42: "HANJA",
        0x7E43: "HANJA",
        0x7E44: "HANJA",
        0x7E45: "HANJA",
        0x7E46: "HANJA",
        0x7E47: "HANJA",
        0x7E48: "HANJA",
        0x7E49: "HANJA",
        0x7E4A: "HANJA",
        0x7E4B: "HANJA",
        0x7E4C: "HANJA",
        0x7E4D: "HANJA",
        0x7E4E: "HANJA",
        0x7E4F: "HANJA",
        0x7E50: "HANJA",
        0x7E51: "HANJA",
        0x7E52: "HANJA",
        0x7E53: "HANJA",
        0x7E54: "HANJA",
        0x7E55: "HANJA",
        0x7E56: "HANJA",
        0x7E57: "HANJA",
        0x7E58: "HANJA",
        0x7E59: "HANJA",
        0x7E5A: "HANJA",
        0x7E5B: "HANJA",
        0x7E5C: "HANJA",
        0x7E5D: "HANJA",
        0x7E5E: "HANJA",
        0x7E5F: "HANJA",
        0x7E60: "HANJA",
        0x7E61: "HANJA",
        0x7E62: "HANJA",
        0x7E63: "HANJA",
        0x7E64: "HANJA",
        0x7E65: "HANJA",
        0x7E66: "HANJA",
        0x7E67: "HANJA",
        0x7E68: "HANJA",
        0x7E69: "HANJA",
        0x7E6A: "HANJA",
        0x7E6B: "HANJA",
        0x7E6C: "HANJA",
        0x7E6D: "HANJA",
        0x7E6E: "HANJA",
        0x7E6F: "HANJA",
        0x7E70: "HANJA",
        0x7E71: "HANJA",
        0x7E72: "HANJA",
        0x7E73: "HANJA",
        0x7E74: "HANJA",
        0x7E75: "HANJA",
        0x7E76: "HANJA",
        0x7E77: "HANJA",
        0x7E78: "HANJA",
        0x7E79: "HANJA",
        0x7E7A: "HANJA",
        0x7E7B: "HANJA",
        0x7E7C: "HANJA",
        0x7E7D: "HANJA",
        0x7E7E: "HANJA",
        0x7E7F: "HANJA",
        0x7E80: "HANJA",
        0x7E81: "HANJA",
        0x7E82: "HANJA",
        0x7E83: "HANJA",
        0x7E84: "HANJA",
        0x7E85: "HANJA",
        0x7E86: "HANJA",
        0x7E87: "HANJA",
        0x7E88: "HANJA",
        0x7E89: "HANJA",
        0x7E8A: "HANJA",
        0x7E8B: "HANJA",
        0x7E8C: "HANJA",
        0x7E8D: "HANJA",
        0x7E8E: "HANJA",
        0x7E8F: "HANJA",
        0x7E90: "HANJA",
        0x7E91: "HANJA",
        0x7E92: "HANJA",
        0x7E93: "HANJA",
        0x7E94: "HANJA",
        0x7E95: "HANJA",
        0x7E96: "HANJA",
        0x7E97: "HANJA",
        0x7E98: "HANJA",
        0x7E99: "HANJA",
        0x7E9A: "HANJA",
        0x7E9B: "HANJA",
        0x7E9C: "HANJA",
        0x7E9D: "HANJA",
        0x7E9E: "HANJA",
        0x7E9F: "HANJA",
        0x7EA0: "HANJA",
        0x7EA1: "HANJA",
        0x7EA2: "HANJA",
        0x7EA3: "HANJA",
        0x7EA4: "HANJA",
        0x7EA5: "HANJA",
        0x7EA6: "HANJA",
        0x7EA7: "HANJA",
        0x7EA8: "HANJA",
        0x7EA9: "HANJA",
        0x7EAA: "HANJA",
        0x7EAB: "HANJA",
        0x7EAC: "HANJA",
        0x7EAD: "HANJA",
        0x7EAE: "HANJA",
        0x7EAF: "HANJA",
        0x7EB0: "HANJA",
        0x7EB1: "HANJA",
        0x7EB2: "HANJA",
        0x7EB3: "HANJA",
        0x7EB4: "HANJA",
        0x7EB5: "HANJA",
        0x7EB6: "HANJA",
        0x7EB7: "HANJA",
        0x7EB8: "HANJA",
        0x7EB9: "HANJA",
        0x7EBA: "HANJA",
        0x7EBB: "HANJA",
        0x7EBC: "HANJA",
        0x7EBD: "HANJA",
        0x7EBE: "HANJA",
        0x7EBF: "HANJA",
        0x7EC0: "HANJA",
        0x7EC1: "HANJA",
        0x7EC2: "HANJA",
        0x7EC3: "HANJA",
        0x7EC4: "HANJA",
        0x7EC5: "HANJA",
        0x7EC6: "HANJA",
        0x7EC7: "HANJA",
        0x7EC8: "HANJA",
        0x7EC9: "HANJA",
        0x7ECA: "HANJA",
        0x7ECB: "HANJA",
        0x7ECC: "HANJA",
        0x7ECD: "HANJA",
        0x7ECE: "HANJA",
        0x7ECF: "HANJA",
        0x7ED0: "HANJA",
        0x7ED1: "HANJA",
        0x7ED2: "HANJA",
        0x7ED3: "HANJA",
        0x7ED4: "HANJA",
        0x7ED5: "HANJA",
        0x7ED6: "HANJA",
        0x7ED7: "HANJA",
        0x7ED8: "HANJA",
        0x7ED9: "HANJA",
        0x7EDA: "HANJA",
        0x7EDB: "HANJA",
        0x7EDC: "HANJA",
        0x7EDD: "HANJA",
        0x7EDE: "HANJA",
        0x7EDF: "HANJA",
        0x7EE0: "HANJA",
        0x7EE1: "HANJA",
        0x7EE2: "HANJA",
        0x7EE3: "HANJA",
        0x7EE4: "HANJA",
        0x7EE5: "HANJA",
        0x7EE6: "HANJA",
        0x7EE7: "HANJA",
        0x7EE8: "HANJA",
        0x7EE9: "HANJA",
        0x7EEA: "HANJA",
        0x7EEB: "HANJA",
        0x7EEC: "HANJA",
        0x7EED: "HANJA",
        0x7EEE: "HANJA",
        0x7EEF: "HANJA",
        0x7EF0: "HANJA",
        0x7EF1: "HANJA",
        0x7EF2: "HANJA",
        0x7EF3: "HANJA",
        0x7EF4: "HANJA",
        0x7EF5: "HANJA",
        0x7EF6: "HANJA",
        0x7EF7: "HANJA",
        0x7EF8: "HANJA",
        0x7EF9: "HANJA",
        0x7EFA: "HANJA",
        0x7EFB: "HANJA",
        0x7EFC: "HANJA",
        0x7EFD: "HANJA",
        0x7EFE: "HANJA",
        0x7EFF: "HANJA",
        0x7F00: "HANJA",
        0x7F01: "HANJA",
        0x7F02: "HANJA",
        0x7F03: "HANJA",
        0x7F04: "HANJA",
        0x7F05: "HANJA",
        0x7F06: "HANJA",
        0x7F07: "HANJA",
        0x7F08: "HANJA",
        0x7F09: "HANJA",
        0x7F0A: "HANJA",
        0x7F0B: "HANJA",
        0x7F0C: "HANJA",
        0x7F0D: "HANJA",
        0x7F0E: "HANJA",
        0x7F0F: "HANJA",
        0x7F10: "HANJA",
        0x7F11: "HANJA",
        0x7F12: "HANJA",
        0x7F13: "HANJA",
        0x7F14: "HANJA",
        0x7F15: "HANJA",
        0x7F16: "HANJA",
        0x7F17: "HANJA",
        0x7F18: "HANJA",
        0x7F19: "HANJA",
        0x7F1A: "HANJA",
        0x7F1B: "HANJA",
        0x7F1C: "HANJA",
        0x7F1D: "HANJA",
        0x7F1E: "HANJA",
        0x7F1F: "HANJA",
        0x7F20: "HANJA",
        0x7F21: "HANJA",
        0x7F22: "HANJA",
        0x7F23: "HANJA",
        0x7F24: "HANJA",
        0x7F25: "HANJA",
        0x7F26: "HANJA",
        0x7F27: "HANJA",
        0x7F28: "HANJA",
        0x7F29: "HANJA",
        0x7F2A: "HANJA",
        0x7F2B: "HANJA",
        0x7F2C: "HANJA",
        0x7F2D: "HANJA",
        0x7F2E: "HANJA",
        0x7F2F: "HANJA",
        0x7F30: "HANJA",
        0x7F31: "HANJA",
        0x7F32: "HANJA",
        0x7F33: "HANJA",
        0x7F34: "HANJA",
        0x7F35: "HANJA",
        0x7F36: "HANJA",
        0x7F37: "HANJA",
        0x7F38: "HANJA",
        0x7F39: "HANJA",
        0x7F3A: "HANJA",
        0x7F3B: "HANJA",
        0x7F3C: "HANJA",
        0x7F3D: "HANJA",
        0x7F3E: "HANJA",
        0x7F3F: "HANJA",
        0x7F40: "HANJA",
        0x7F41: "HANJA",
        0x7F42: "HANJA",
        0x7F43: "HANJA",
        0x7F44: "HANJA",
        0x7F45: "HANJA",
        0x7F46: "HANJA",
        0x7F47: "HANJA",
        0x7F48: "HANJA",
        0x7F49: "HANJA",
        0x7F4A: "HANJA",
        0x7F4B: "HANJA",
        0x7F4C: "HANJA",
        0x7F4D: "HANJA",
        0x7F4E: "HANJA",
        0x7F4F: "HANJA",
        0x7F50: "HANJA",
        0x7F51: "HANJA",
        0x7F52: "HANJA",
        0x7F53: "HANJA",
        0x7F54: "HANJA",
        0x7F55: "HANJA",
        0x7F56: "HANJA",
        0x7F57: "HANJA",
        0x7F58: "HANJA",
        0x7F59: "HANJA",
        0x7F5A: "HANJA",
        0x7F5B: "HANJA",
        0x7F5C: "HANJA",
        0x7F5D: "HANJA",
        0x7F5E: "HANJA",
        0x7F5F: "HANJA",
        0x7F60: "HANJA",
        0x7F61: "HANJA",
        0x7F62: "HANJA",
        0x7F63: "HANJA",
        0x7F64: "HANJA",
        0x7F65: "HANJA",
        0x7F66: "HANJA",
        0x7F67: "HANJA",
        0x7F68: "HANJA",
        0x7F69: "HANJA",
        0x7F6A: "HANJA",
        0x7F6B: "HANJA",
        0x7F6C: "HANJA",
        0x7F6D: "HANJA",
        0x7F6E: "HANJA",
        0x7F6F: "HANJA",
        0x7F70: "HANJA",
        0x7F71: "HANJA",
        0x7F72: "HANJA",
        0x7F73: "HANJA",
        0x7F74: "HANJA",
        0x7F75: "HANJA",
        0x7F76: "HANJA",
        0x7F77: "HANJA",
        0x7F78: "HANJA",
        0x7F79: "HANJA",
        0x7F7A: "HANJA",
        0x7F7B: "HANJA",
        0x7F7C: "HANJA",
        0x7F7D: "HANJA",
        0x7F7E: "HANJA",
        0x7F7F: "HANJA",
        0x7F80: "HANJA",
        0x7F81: "HANJA",
        0x7F82: "HANJA",
        0x7F83: "HANJA",
        0x7F84: "HANJA",
        0x7F85: "HANJA",
        0x7F86: "HANJA",
        0x7F87: "HANJA",
        0x7F88: "HANJA",
        0x7F89: "HANJA",
        0x7F8A: "HANJA",
        0x7F8B: "HANJA",
        0x7F8C: "HANJA",
        0x7F8D: "HANJA",
        0x7F8E: "HANJA",
        0x7F8F: "HANJA",
        0x7F90: "HANJA",
        0x7F91: "HANJA",
        0x7F92: "HANJA",
        0x7F93: "HANJA",
        0x7F94: "HANJA",
        0x7F95: "HANJA",
        0x7F96: "HANJA",
        0x7F97: "HANJA",
        0x7F98: "HANJA",
        0x7F99: "HANJA",
        0x7F9A: "HANJA",
        0x7F9B: "HANJA",
        0x7F9C: "HANJA",
        0x7F9D: "HANJA",
        0x7F9E: "HANJA",
        0x7F9F: "HANJA",
        0x7FA0: "HANJA",
        0x7FA1: "HANJA",
        0x7FA2: "HANJA",
        0x7FA3: "HANJA",
        0x7FA4: "HANJA",
        0x7FA5: "HANJA",
        0x7FA6: "HANJA",
        0x7FA7: "HANJA",
        0x7FA8: "HANJA",
        0x7FA9: "HANJA",
        0x7FAA: "HANJA",
        0x7FAB: "HANJA",
        0x7FAC: "HANJA",
        0x7FAD: "HANJA",
        0x7FAE: "HANJA",
        0x7FAF: "HANJA",
        0x7FB0: "HANJA",
        0x7FB1: "HANJA",
        0x7FB2: "HANJA",
        0x7FB3: "HANJA",
        0x7FB4: "HANJA",
        0x7FB5: "HANJA",
        0x7FB6: "HANJA",
        0x7FB7: "HANJA",
        0x7FB8: "HANJA",
        0x7FB9: "HANJA",
        0x7FBA: "HANJA",
        0x7FBB: "HANJA",
        0x7FBC: "HANJA",
        0x7FBD: "HANJA",
        0x7FBE: "HANJA",
        0x7FBF: "HANJA",
        0x7FC0: "HANJA",
        0x7FC1: "HANJA",
        0x7FC2: "HANJA",
        0x7FC3: "HANJA",
        0x7FC4: "HANJA",
        0x7FC5: "HANJA",
        0x7FC6: "HANJA",
        0x7FC7: "HANJA",
        0x7FC8: "HANJA",
        0x7FC9: "HANJA",
        0x7FCA: "HANJA",
        0x7FCB: "HANJA",
        0x7FCC: "HANJA",
        0x7FCD: "HANJA",
        0x7FCE: "HANJA",
        0x7FCF: "HANJA",
        0x7FD0: "HANJA",
        0x7FD1: "HANJA",
        0x7FD2: "HANJA",
        0x7FD3: "HANJA",
        0x7FD4: "HANJA",
        0x7FD5: "HANJA",
        0x7FD6: "HANJA",
        0x7FD7: "HANJA",
        0x7FD8: "HANJA",
        0x7FD9: "HANJA",
        0x7FDA: "HANJA",
        0x7FDB: "HANJA",
        0x7FDC: "HANJA",
        0x7FDD: "HANJA",
        0x7FDE: "HANJA",
        0x7FDF: "HANJA",
        0x7FE0: "HANJA",
        0x7FE1: "HANJA",
        0x7FE2: "HANJA",
        0x7FE3: "HANJA",
        0x7FE4: "HANJA",
        0x7FE5: "HANJA",
        0x7FE6: "HANJA",
        0x7FE7: "HANJA",
        0x7FE8: "HANJA",
        0x7FE9: "HANJA",
        0x7FEA: "HANJA",
        0x7FEB: "HANJA",
        0x7FEC: "HANJA",
        0x7FED: "HANJA",
        0x7FEE: "HANJA",
        0x7FEF: "HANJA",
        0x7FF0: "HANJA",
        0x7FF1: "HANJA",
        0x7FF2: "HANJA",
        0x7FF3: "HANJA",
        0x7FF4: "HANJA",
        0x7FF5: "HANJA",
        0x7FF6: "HANJA",
        0x7FF7: "HANJA",
        0x7FF8: "HANJA",
        0x7FF9: "HANJA",
        0x7FFA: "HANJA",
        0x7FFB: "HANJA",
        0x7FFC: "HANJA",
        0x7FFD: "HANJA",
        0x7FFE: "HANJA",
        0x7FFF: "HANJA",
        0x8000: "HANJA",
        0x8001: "HANJA",
        0x8002: "HANJA",
        0x8003: "HANJA",
        0x8004: "HANJA",
        0x8005: "HANJA",
        0x8006: "HANJA",
        0x8007: "HANJA",
        0x8008: "HANJA",
        0x8009: "HANJA",
        0x800A: "HANJA",
        0x800B: "HANJA",
        0x800C: "HANJA",
        0x800D: "HANJA",
        0x800E: "HANJA",
        0x800F: "HANJA",
        0x8010: "HANJA",
        0x8011: "HANJA",
        0x8012: "HANJA",
        0x8013: "HANJA",
        0x8014: "HANJA",
        0x8015: "HANJA",
        0x8016: "HANJA",
        0x8017: "HANJA",
        0x8018: "HANJA",
        0x8019: "HANJA",
        0x801A: "HANJA",
        0x801B: "HANJA",
        0x801C: "HANJA",
        0x801D: "HANJA",
        0x801E: "HANJA",
        0x801F: "HANJA",
        0x8020: "HANJA",
        0x8021: "HANJA",
        0x8022: "HANJA",
        0x8023: "HANJA",
        0x8024: "HANJA",
        0x8025: "HANJA",
        0x8026: "HANJA",
        0x8027: "HANJA",
        0x8028: "HANJA",
        0x8029: "HANJA",
        0x802A: "HANJA",
        0x802B: "HANJA",
        0x802C: "HANJA",
        0x802D: "HANJA",
        0x802E: "HANJA",
        0x802F: "HANJA",
        0x8030: "HANJA",
        0x8031: "HANJA",
        0x8032: "HANJA",
        0x8033: "HANJA",
        0x8034: "HANJA",
        0x8035: "HANJA",
        0x8036: "HANJA",
        0x8037: "HANJA",
        0x8038: "HANJA",
        0x8039: "HANJA",
        0x803A: "HANJA",
        0x803B: "HANJA",
        0x803C: "HANJA",
        0x803D: "HANJA",
        0x803E: "HANJA",
        0x803F: "HANJA",
        0x8040: "HANJA",
        0x8041: "HANJA",
        0x8042: "HANJA",
        0x8043: "HANJA",
        0x8044: "HANJA",
        0x8045: "HANJA",
        0x8046: "HANJA",
        0x8047: "HANJA",
        0x8048: "HANJA",
        0x8049: "HANJA",
        0x804A: "HANJA",
        0x804B: "HANJA",
        0x804C: "HANJA",
        0x804D: "HANJA",
        0x804E: "HANJA",
        0x804F: "HANJA",
        0x8050: "HANJA",
        0x8051: "HANJA",
        0x8052: "HANJA",
        0x8053: "HANJA",
        0x8054: "HANJA",
        0x8055: "HANJA",
        0x8056: "HANJA",
        0x8057: "HANJA",
        0x8058: "HANJA",
        0x8059: "HANJA",
        0x805A: "HANJA",
        0x805B: "HANJA",
        0x805C: "HANJA",
        0x805D: "HANJA",
        0x805E: "HANJA",
        0x805F: "HANJA",
        0x8060: "HANJA",
        0x8061: "HANJA",
        0x8062: "HANJA",
        0x8063: "HANJA",
        0x8064: "HANJA",
        0x8065: "HANJA",
        0x8066: "HANJA",
        0x8067: "HANJA",
        0x8068: "HANJA",
        0x8069: "HANJA",
        0x806A: "HANJA",
        0x806B: "HANJA",
        0x806C: "HANJA",
        0x806D: "HANJA",
        0x806E: "HANJA",
        0x806F: "HANJA",
        0x8070: "HANJA",
        0x8071: "HANJA",
        0x8072: "HANJA",
        0x8073: "HANJA",
        0x8074: "HANJA",
        0x8075: "HANJA",
        0x8076: "HANJA",
        0x8077: "HANJA",
        0x8078: "HANJA",
        0x8079: "HANJA",
        0x807A: "HANJA",
        0x807B: "HANJA",
        0x807C: "HANJA",
        0x807D: "HANJA",
        0x807E: "HANJA",
        0x807F: "HANJA",
        0x8080: "HANJA",
        0x8081: "HANJA",
        0x8082: "HANJA",
        0x8083: "HANJA",
        0x8084: "HANJA",
        0x8085: "HANJA",
        0x8086: "HANJA",
        0x8087: "HANJA",
        0x8088: "HANJA",
        0x8089: "HANJA",
        0x808A: "HANJA",
        0x808B: "HANJA",
        0x808C: "HANJA",
        0x808D: "HANJA",
        0x808E: "HANJA",
        0x808F: "HANJA",
        0x8090: "HANJA",
        0x8091: "HANJA",
        0x8092: "HANJA",
        0x8093: "HANJA",
        0x8094: "HANJA",
        0x8095: "HANJA",
        0x8096: "HANJA",
        0x8097: "HANJA",
        0x8098: "HANJA",
        0x8099: "HANJA",
        0x809A: "HANJA",
        0x809B: "HANJA",
        0x809C: "HANJA",
        0x809D: "HANJA",
        0x809E: "HANJA",
        0x809F: "HANJA",
        0x80A0: "HANJA",
        0x80A1: "HANJA",
        0x80A2: "HANJA",
        0x80A3: "HANJA",
        0x80A4: "HANJA",
        0x80A5: "HANJA",
        0x80A6: "HANJA",
        0x80A7: "HANJA",
        0x80A8: "HANJA",
        0x80A9: "HANJA",
        0x80AA: "HANJA",
        0x80AB: "HANJA",
        0x80AC: "HANJA",
        0x80AD: "HANJA",
        0x80AE: "HANJA",
        0x80AF: "HANJA",
        0x80B0: "HANJA",
        0x80B1: "HANJA",
        0x80B2: "HANJA",
        0x80B3: "HANJA",
        0x80B4: "HANJA",
        0x80B5: "HANJA",
        0x80B6: "HANJA",
        0x80B7: "HANJA",
        0x80B8: "HANJA",
        0x80B9: "HANJA",
        0x80BA: "HANJA",
        0x80BB: "HANJA",
        0x80BC: "HANJA",
        0x80BD: "HANJA",
        0x80BE: "HANJA",
        0x80BF: "HANJA",
        0x80C0: "HANJA",
        0x80C1: "HANJA",
        0x80C2: "HANJA",
        0x80C3: "HANJA",
        0x80C4: "HANJA",
        0x80C5: "HANJA",
        0x80C6: "HANJA",
        0x80C7: "HANJA",
        0x80C8: "HANJA",
        0x80C9: "HANJA",
        0x80CA: "HANJA",
        0x80CB: "HANJA",
        0x80CC: "HANJA",
        0x80CD: "HANJA",
        0x80CE: "HANJA",
        0x80CF: "HANJA",
        0x80D0: "HANJA",
        0x80D1: "HANJA",
        0x80D2: "HANJA",
        0x80D3: "HANJA",
        0x80D4: "HANJA",
        0x80D5: "HANJA",
        0x80D6: "HANJA",
        0x80D7: "HANJA",
        0x80D8: "HANJA",
        0x80D9: "HANJA",
        0x80DA: "HANJA",
        0x80DB: "HANJA",
        0x80DC: "HANJA",
        0x80DD: "HANJA",
        0x80DE: "HANJA",
        0x80DF: "HANJA",
        0x80E0: "HANJA",
        0x80E1: "HANJA",
        0x80E2: "HANJA",
        0x80E3: "HANJA",
        0x80E4: "HANJA",
        0x80E5: "HANJA",
        0x80E6: "HANJA",
        0x80E7: "HANJA",
        0x80E8: "HANJA",
        0x80E9: "HANJA",
        0x80EA: "HANJA",
        0x80EB: "HANJA",
        0x80EC: "HANJA",
        0x80ED: "HANJA",
        0x80EE: "HANJA",
        0x80EF: "HANJA",
        0x80F0: "HANJA",
        0x80F1: "HANJA",
        0x80F2: "HANJA",
        0x80F3: "HANJA",
        0x80F4: "HANJA",
        0x80F5: "HANJA",
        0x80F6: "HANJA",
        0x80F7: "HANJA",
        0x80F8: "HANJA",
        0x80F9: "HANJA",
        0x80FA: "HANJA",
        0x80FB: "HANJA",
        0x80FC: "HANJA",
        0x80FD: "HANJA",
        0x80FE: "HANJA",
        0x80FF: "HANJA",
        0x8100: "HANJA",
        0x8101: "HANJA",
        0x8102: "HANJA",
        0x8103: "HANJA",
        0x8104: "HANJA",
        0x8105: "HANJA",
        0x8106: "HANJA",
        0x8107: "HANJA",
        0x8108: "HANJA",
        0x8109: "HANJA",
        0x810A: "HANJA",
        0x810B: "HANJA",
        0x810C: "HANJA",
        0x810D: "HANJA",
        0x810E: "HANJA",
        0x810F: "HANJA",
        0x8110: "HANJA",
        0x8111: "HANJA",
        0x8112: "HANJA",
        0x8113: "HANJA",
        0x8114: "HANJA",
        0x8115: "HANJA",
        0x8116: "HANJA",
        0x8117: "HANJA",
        0x8118: "HANJA",
        0x8119: "HANJA",
        0x811A: "HANJA",
        0x811B: "HANJA",
        0x811C: "HANJA",
        0x811D: "HANJA",
        0x811E: "HANJA",
        0x811F: "HANJA",
        0x8120: "HANJA",
        0x8121: "HANJA",
        0x8122: "HANJA",
        0x8123: "HANJA",
        0x8124: "HANJA",
        0x8125: "HANJA",
        0x8126: "HANJA",
        0x8127: "HANJA",
        0x8128: "HANJA",
        0x8129: "HANJA",
        0x812A: "HANJA",
        0x812B: "HANJA",
        0x812C: "HANJA",
        0x812D: "HANJA",
        0x812E: "HANJA",
        0x812F: "HANJA",
        0x8130: "HANJA",
        0x8131: "HANJA",
        0x8132: "HANJA",
        0x8133: "HANJA",
        0x8134: "HANJA",
        0x8135: "HANJA",
        0x8136: "HANJA",
        0x8137: "HANJA",
        0x8138: "HANJA",
        0x8139: "HANJA",
        0x813A: "HANJA",
        0x813B: "HANJA",
        0x813C: "HANJA",
        0x813D: "HANJA",
        0x813E: "HANJA",
        0x813F: "HANJA",
        0x8140: "HANJA",
        0x8141: "HANJA",
        0x8142: "HANJA",
        0x8143: "HANJA",
        0x8144: "HANJA",
        0x8145: "HANJA",
        0x8146: "HANJA",
        0x8147: "HANJA",
        0x8148: "HANJA",
        0x8149: "HANJA",
        0x814A: "HANJA",
        0x814B: "HANJA",
        0x814C: "HANJA",
        0x814D: "HANJA",
        0x814E: "HANJA",
        0x814F: "HANJA",
        0x8150: "HANJA",
        0x8151: "HANJA",
        0x8152: "HANJA",
        0x8153: "HANJA",
        0x8154: "HANJA",
        0x8155: "HANJA",
        0x8156: "HANJA",
        0x8157: "HANJA",
        0x8158: "HANJA",
        0x8159: "HANJA",
        0x815A: "HANJA",
        0x815B: "HANJA",
        0x815C: "HANJA",
        0x815D: "HANJA",
        0x815E: "HANJA",
        0x815F: "HANJA",
        0x8160: "HANJA",
        0x8161: "HANJA",
        0x8162: "HANJA",
        0x8163: "HANJA",
        0x8164: "HANJA",
        0x8165: "HANJA",
        0x8166: "HANJA",
        0x8167: "HANJA",
        0x8168: "HANJA",
        0x8169: "HANJA",
        0x816A: "HANJA",
        0x816B: "HANJA",
        0x816C: "HANJA",
        0x816D: "HANJA",
        0x816E: "HANJA",
        0x816F: "HANJA",
        0x8170: "HANJA",
        0x8171: "HANJA",
        0x8172: "HANJA",
        0x8173: "HANJA",
        0x8174: "HANJA",
        0x8175: "HANJA",
        0x8176: "HANJA",
        0x8177: "HANJA",
        0x8178: "HANJA",
        0x8179: "HANJA",
        0x817A: "HANJA",
        0x817B: "HANJA",
        0x817C: "HANJA",
        0x817D: "HANJA",
        0x817E: "HANJA",
        0x817F: "HANJA",
        0x8180: "HANJA",
        0x8181: "HANJA",
        0x8182: "HANJA",
        0x8183: "HANJA",
        0x8184: "HANJA",
        0x8185: "HANJA",
        0x8186: "HANJA",
        0x8187: "HANJA",
        0x8188: "HANJA",
        0x8189: "HANJA",
        0x818A: "HANJA",
        0x818B: "HANJA",
        0x818C: "HANJA",
        0x818D: "HANJA",
        0x818E: "HANJA",
        0x818F: "HANJA",
        0x8190: "HANJA",
        0x8191: "HANJA",
        0x8192: "HANJA",
        0x8193: "HANJA",
        0x8194: "HANJA",
        0x8195: "HANJA",
        0x8196: "HANJA",
        0x8197: "HANJA",
        0x8198: "HANJA",
        0x8199: "HANJA",
        0x819A: "HANJA",
        0x819B: "HANJA",
        0x819C: "HANJA",
        0x819D: "HANJA",
        0x819E: "HANJA",
        0x819F: "HANJA",
        0x81A0: "HANJA",
        0x81A1: "HANJA",
        0x81A2: "HANJA",
        0x81A3: "HANJA",
        0x81A4: "HANJA",
        0x81A5: "HANJA",
        0x81A6: "HANJA",
        0x81A7: "HANJA",
        0x81A8: "HANJA",
        0x81A9: "HANJA",
        0x81AA: "HANJA",
        0x81AB: "HANJA",
        0x81AC: "HANJA",
        0x81AD: "HANJA",
        0x81AE: "HANJA",
        0x81AF: "HANJA",
        0x81B0: "HANJA",
        0x81B1: "HANJA",
        0x81B2: "HANJA",
        0x81B3: "HANJA",
        0x81B4: "HANJA",
        0x81B5: "HANJA",
        0x81B6: "HANJA",
        0x81B7: "HANJA",
        0x81B8: "HANJA",
        0x81B9: "HANJA",
        0x81BA: "HANJA",
        0x81BB: "HANJA",
        0x81BC: "HANJA",
        0x81BD: "HANJA",
        0x81BE: "HANJA",
        0x81BF: "HANJA",
        0x81C0: "HANJA",
        0x81C1: "HANJA",
        0x81C2: "HANJA",
        0x81C3: "HANJA",
        0x81C4: "HANJA",
        0x81C5: "HANJA",
        0x81C6: "HANJA",
        0x81C7: "HANJA",
        0x81C8: "HANJA",
        0x81C9: "HANJA",
        0x81CA: "HANJA",
        0x81CB: "HANJA",
        0x81CC: "HANJA",
        0x81CD: "HANJA",
        0x81CE: "HANJA",
        0x81CF: "HANJA",
        0x81D0: "HANJA",
        0x81D1: "HANJA",
        0x81D2: "HANJA",
        0x81D3: "HANJA",
        0x81D4: "HANJA",
        0x81D5: "HANJA",
        0x81D6: "HANJA",
        0x81D7: "HANJA",
        0x81D8: "HANJA",
        0x81D9: "HANJA",
        0x81DA: "HANJA",
        0x81DB: "HANJA",
        0x81DC: "HANJA",
        0x81DD: "HANJA",
        0x81DE: "HANJA",
        0x81DF: "HANJA",
        0x81E0: "HANJA",
        0x81E1: "HANJA",
        0x81E2: "HANJA",
        0x81E3: "HANJA",
        0x81E4: "HANJA",
        0x81E5: "HANJA",
        0x81E6: "HANJA",
        0x81E7: "HANJA",
        0x81E8: "HANJA",
        0x81E9: "HANJA",
        0x81EA: "HANJA",
        0x81EB: "HANJA",
        0x81EC: "HANJA",
        0x81ED: "HANJA",
        0x81EE: "HANJA",
        0x81EF: "HANJA",
        0x81F0: "HANJA",
        0x81F1: "HANJA",
        0x81F2: "HANJA",
        0x81F3: "HANJA",
        0x81F4: "HANJA",
        0x81F5: "HANJA",
        0x81F6: "HANJA",
        0x81F7: "HANJA",
        0x81F8: "HANJA",
        0x81F9: "HANJA",
        0x81FA: "HANJA",
        0x81FB: "HANJA",
        0x81FC: "HANJA",
        0x81FD: "HANJA",
        0x81FE: "HANJA",
        0x81FF: "HANJA",
        0x8200: "HANJA",
        0x8201: "HANJA",
        0x8202: "HANJA",
        0x8203: "HANJA",
        0x8204: "HANJA",
        0x8205: "HANJA",
        0x8206: "HANJA",
        0x8207: "HANJA",
        0x8208: "HANJA",
        0x8209: "HANJA",
        0x820A: "HANJA",
        0x820B: "HANJA",
        0x820C: "HANJA",
        0x820D: "HANJA",
        0x820E: "HANJA",
        0x820F: "HANJA",
        0x8210: "HANJA",
        0x8211: "HANJA",
        0x8212: "HANJA",
        0x8213: "HANJA",
        0x8214: "HANJA",
        0x8215: "HANJA",
        0x8216: "HANJA",
        0x8217: "HANJA",
        0x8218: "HANJA",
        0x8219: "HANJA",
        0x821A: "HANJA",
        0x821B: "HANJA",
        0x821C: "HANJA",
        0x821D: "HANJA",
        0x821E: "HANJA",
        0x821F: "HANJA",
        0x8220: "HANJA",
        0x8221: "HANJA",
        0x8222: "HANJA",
        0x8223: "HANJA",
        0x8224: "HANJA",
        0x8225: "HANJA",
        0x8226: "HANJA",
        0x8227: "HANJA",
        0x8228: "HANJA",
        0x8229: "HANJA",
        0x822A: "HANJA",
        0x822B: "HANJA",
        0x822C: "HANJA",
        0x822D: "HANJA",
        0x822E: "HANJA",
        0x822F: "HANJA",
        0x8230: "HANJA",
        0x8231: "HANJA",
        0x8232: "HANJA",
        0x8233: "HANJA",
        0x8234: "HANJA",
        0x8235: "HANJA",
        0x8236: "HANJA",
        0x8237: "HANJA",
        0x8238: "HANJA",
        0x8239: "HANJA",
        0x823A: "HANJA",
        0x823B: "HANJA",
        0x823C: "HANJA",
        0x823D: "HANJA",
        0x823E: "HANJA",
        0x823F: "HANJA",
        0x8240: "HANJA",
        0x8241: "HANJA",
        0x8242: "HANJA",
        0x8243: "HANJA",
        0x8244: "HANJA",
        0x8245: "HANJA",
        0x8246: "HANJA",
        0x8247: "HANJA",
        0x8248: "HANJA",
        0x8249: "HANJA",
        0x824A: "HANJA",
        0x824B: "HANJA",
        0x824C: "HANJA",
        0x824D: "HANJA",
        0x824E: "HANJA",
        0x824F: "HANJA",
        0x8250: "HANJA",
        0x8251: "HANJA",
        0x8252: "HANJA",
        0x8253: "HANJA",
        0x8254: "HANJA",
        0x8255: "HANJA",
        0x8256: "HANJA",
        0x8257: "HANJA",
        0x8258: "HANJA",
        0x8259: "HANJA",
        0x825A: "HANJA",
        0x825B: "HANJA",
        0x825C: "HANJA",
        0x825D: "HANJA",
        0x825E: "HANJA",
        0x825F: "HANJA",
        0x8260: "HANJA",
        0x8261: "HANJA",
        0x8262: "HANJA",
        0x8263: "HANJA",
        0x8264: "HANJA",
        0x8265: "HANJA",
        0x8266: "HANJA",
        0x8267: "HANJA",
        0x8268: "HANJA",
        0x8269: "HANJA",
        0x826A: "HANJA",
        0x826B: "HANJA",
        0x826C: "HANJA",
        0x826D: "HANJA",
        0x826E: "HANJA",
        0x826F: "HANJA",
        0x8270: "HANJA",
        0x8271: "HANJA",
        0x8272: "HANJA",
        0x8273: "HANJA",
        0x8274: "HANJA",
        0x8275: "HANJA",
        0x8276: "HANJA",
        0x8277: "HANJA",
        0x8278: "HANJA",
        0x8279: "HANJA",
        0x827A: "HANJA",
        0x827B: "HANJA",
        0x827C: "HANJA",
        0x827D: "HANJA",
        0x827E: "HANJA",
        0x827F: "HANJA",
        0x8280: "HANJA",
        0x8281: "HANJA",
        0x8282: "HANJA",
        0x8283: "HANJA",
        0x8284: "HANJA",
        0x8285: "HANJA",
        0x8286: "HANJA",
        0x8287: "HANJA",
        0x8288: "HANJA",
        0x8289: "HANJA",
        0x828A: "HANJA",
        0x828B: "HANJA",
        0x828C: "HANJA",
        0x828D: "HANJA",
        0x828E: "HANJA",
        0x828F: "HANJA",
        0x8290: "HANJA",
        0x8291: "HANJA",
        0x8292: "HANJA",
        0x8293: "HANJA",
        0x8294: "HANJA",
        0x8295: "HANJA",
        0x8296: "HANJA",
        0x8297: "HANJA",
        0x8298: "HANJA",
        0x8299: "HANJA",
        0x829A: "HANJA",
        0x829B: "HANJA",
        0x829C: "HANJA",
        0x829D: "HANJA",
        0x829E: "HANJA",
        0x829F: "HANJA",
        0x82A0: "HANJA",
        0x82A1: "HANJA",
        0x82A2: "HANJA",
        0x82A3: "HANJA",
        0x82A4: "HANJA",
        0x82A5: "HANJA",
        0x82A6: "HANJA",
        0x82A7: "HANJA",
        0x82A8: "HANJA",
        0x82A9: "HANJA",
        0x82AA: "HANJA",
        0x82AB: "HANJA",
        0x82AC: "HANJA",
        0x82AD: "HANJA",
        0x82AE: "HANJA",
        0x82AF: "HANJA",
        0x82B0: "HANJA",
        0x82B1: "HANJA",
        0x82B2: "HANJA",
        0x82B3: "HANJA",
        0x82B4: "HANJA",
        0x82B5: "HANJA",
        0x82B6: "HANJA",
        0x82B7: "HANJA",
        0x82B8: "HANJA",
        0x82B9: "HANJA",
        0x82BA: "HANJA",
        0x82BB: "HANJA",
        0x82BC: "HANJA",
        0x82BD: "HANJA",
        0x82BE: "HANJA",
        0x82BF: "HANJA",
        0x82C0: "HANJA",
        0x82C1: "HANJA",
        0x82C2: "HANJA",
        0x82C3: "HANJA",
        0x82C4: "HANJA",
        0x82C5: "HANJA",
        0x82C6: "HANJA",
        0x82C7: "HANJA",
        0x82C8: "HANJA",
        0x82C9: "HANJA",
        0x82CA: "HANJA",
        0x82CB: "HANJA",
        0x82CC: "HANJA",
        0x82CD: "HANJA",
        0x82CE: "HANJA",
        0x82CF: "HANJA",
        0x82D0: "HANJA",
        0x82D1: "HANJA",
        0x82D2: "HANJA",
        0x82D3: "HANJA",
        0x82D4: "HANJA",
        0x82D5: "HANJA",
        0x82D6: "HANJA",
        0x82D7: "HANJA",
        0x82D8: "HANJA",
        0x82D9: "HANJA",
        0x82DA: "HANJA",
        0x82DB: "HANJA",
        0x82DC: "HANJA",
        0x82DD: "HANJA",
        0x82DE: "HANJA",
        0x82DF: "HANJA",
        0x82E0: "HANJA",
        0x82E1: "HANJA",
        0x82E2: "HANJA",
        0x82E3: "HANJA",
        0x82E4: "HANJA",
        0x82E5: "HANJA",
        0x82E6: "HANJA",
        0x82E7: "HANJA",
        0x82E8: "HANJA",
        0x82E9: "HANJA",
        0x82EA: "HANJA",
        0x82EB: "HANJA",
        0x82EC: "HANJA",
        0x82ED: "HANJA",
        0x82EE: "HANJA",
        0x82EF: "HANJA",
        0x82F0: "HANJA",
        0x82F1: "HANJA",
        0x82F2: "HANJA",
        0x82F3: "HANJA",
        0x82F4: "HANJA",
        0x82F5: "HANJA",
        0x82F6: "HANJA",
        0x82F7: "HANJA",
        0x82F8: "HANJA",
        0x82F9: "HANJA",
        0x82FA: "HANJA",
        0x82FB: "HANJA",
        0x82FC: "HANJA",
        0x82FD: "HANJA",
        0x82FE: "HANJA",
        0x82FF: "HANJA",
        0x8300: "HANJA",
        0x8301: "HANJA",
        0x8302: "HANJA",
        0x8303: "HANJA",
        0x8304: "HANJA",
        0x8305: "HANJA",
        0x8306: "HANJA",
        0x8307: "HANJA",
        0x8308: "HANJA",
        0x8309: "HANJA",
        0x830A: "HANJA",
        0x830B: "HANJA",
        0x830C: "HANJA",
        0x830D: "HANJA",
        0x830E: "HANJA",
        0x830F: "HANJA",
        0x8310: "HANJA",
        0x8311: "HANJA",
        0x8312: "HANJA",
        0x8313: "HANJA",
        0x8314: "HANJA",
        0x8315: "HANJA",
        0x8316: "HANJA",
        0x8317: "HANJA",
        0x8318: "HANJA",
        0x8319: "HANJA",
        0x831A: "HANJA",
        0x831B: "HANJA",
        0x831C: "HANJA",
        0x831D: "HANJA",
        0x831E: "HANJA",
        0x831F: "HANJA",
        0x8320: "HANJA",
        0x8321: "HANJA",
        0x8322: "HANJA",
        0x8323: "HANJA",
        0x8324: "HANJA",
        0x8325: "HANJA",
        0x8326: "HANJA",
        0x8327: "HANJA",
        0x8328: "HANJA",
        0x8329: "HANJA",
        0x832A: "HANJA",
        0x832B: "HANJA",
        0x832C: "HANJA",
        0x832D: "HANJA",
        0x832E: "HANJA",
        0x832F: "HANJA",
        0x8330: "HANJA",
        0x8331: "HANJA",
        0x8332: "HANJA",
        0x8333: "HANJA",
        0x8334: "HANJA",
        0x8335: "HANJA",
        0x8336: "HANJA",
        0x8337: "HANJA",
        0x8338: "HANJA",
        0x8339: "HANJA",
        0x833A: "HANJA",
        0x833B: "HANJA",
        0x833C: "HANJA",
        0x833D: "HANJA",
        0x833E: "HANJA",
        0x833F: "HANJA",
        0x8340: "HANJA",
        0x8341: "HANJA",
        0x8342: "HANJA",
        0x8343: "HANJA",
        0x8344: "HANJA",
        0x8345: "HANJA",
        0x8346: "HANJA",
        0x8347: "HANJA",
        0x8348: "HANJA",
        0x8349: "HANJA",
        0x834A: "HANJA",
        0x834B: "HANJA",
        0x834C: "HANJA",
        0x834D: "HANJA",
        0x834E: "HANJA",
        0x834F: "HANJA",
        0x8350: "HANJA",
        0x8351: "HANJA",
        0x8352: "HANJA",
        0x8353: "HANJA",
        0x8354: "HANJA",
        0x8355: "HANJA",
        0x8356: "HANJA",
        0x8357: "HANJA",
        0x8358: "HANJA",
        0x8359: "HANJA",
        0x835A: "HANJA",
        0x835B: "HANJA",
        0x835C: "HANJA",
        0x835D: "HANJA",
        0x835E: "HANJA",
        0x835F: "HANJA",
        0x8360: "HANJA",
        0x8361: "HANJA",
        0x8362: "HANJA",
        0x8363: "HANJA",
        0x8364: "HANJA",
        0x8365: "HANJA",
        0x8366: "HANJA",
        0x8367: "HANJA",
        0x8368: "HANJA",
        0x8369: "HANJA",
        0x836A: "HANJA",
        0x836B: "HANJA",
        0x836C: "HANJA",
        0x836D: "HANJA",
        0x836E: "HANJA",
        0x836F: "HANJA",
        0x8370: "HANJA",
        0x8371: "HANJA",
        0x8372: "HANJA",
        0x8373: "HANJA",
        0x8374: "HANJA",
        0x8375: "HANJA",
        0x8376: "HANJA",
        0x8377: "HANJA",
        0x8378: "HANJA",
        0x8379: "HANJA",
        0x837A: "HANJA",
        0x837B: "HANJA",
        0x837C: "HANJA",
        0x837D: "HANJA",
        0x837E: "HANJA",
        0x837F: "HANJA",
        0x8380: "HANJA",
        0x8381: "HANJA",
        0x8382: "HANJA",
        0x8383: "HANJA",
        0x8384: "HANJA",
        0x8385: "HANJA",
        0x8386: "HANJA",
        0x8387: "HANJA",
        0x8388: "HANJA",
        0x8389: "HANJA",
        0x838A: "HANJA",
        0x838B: "HANJA",
        0x838C: "HANJA",
        0x838D: "HANJA",
        0x838E: "HANJA",
        0x838F: "HANJA",
        0x8390: "HANJA",
        0x8391: "HANJA",
        0x8392: "HANJA",
        0x8393: "HANJA",
        0x8394: "HANJA",
        0x8395: "HANJA",
        0x8396: "HANJA",
        0x8397: "HANJA",
        0x8398: "HANJA",
        0x8399: "HANJA",
        0x839A: "HANJA",
        0x839B: "HANJA",
        0x839C: "HANJA",
        0x839D: "HANJA",
        0x839E: "HANJA",
        0x839F: "HANJA",
        0x83A0: "HANJA",
        0x83A1: "HANJA",
        0x83A2: "HANJA",
        0x83A3: "HANJA",
        0x83A4: "HANJA",
        0x83A5: "HANJA",
        0x83A6: "HANJA",
        0x83A7: "HANJA",
        0x83A8: "HANJA",
        0x83A9: "HANJA",
        0x83AA: "HANJA",
        0x83AB: "HANJA",
        0x83AC: "HANJA",
        0x83AD: "HANJA",
        0x83AE: "HANJA",
        0x83AF: "HANJA",
        0x83B0: "HANJA",
        0x83B1: "HANJA",
        0x83B2: "HANJA",
        0x83B3: "HANJA",
        0x83B4: "HANJA",
        0x83B5: "HANJA",
        0x83B6: "HANJA",
        0x83B7: "HANJA",
        0x83B8: "HANJA",
        0x83B9: "HANJA",
        0x83BA: "HANJA",
        0x83BB: "HANJA",
        0x83BC: "HANJA",
        0x83BD: "HANJA",
        0x83BE: "HANJA",
        0x83BF: "HANJA",
        0x83C0: "HANJA",
        0x83C1: "HANJA",
        0x83C2: "HANJA",
        0x83C3: "HANJA",
        0x83C4: "HANJA",
        0x83C5: "HANJA",
        0x83C6: "HANJA",
        0x83C7: "HANJA",
        0x83C8: "HANJA",
        0x83C9: "HANJA",
        0x83CA: "HANJA",
        0x83CB: "HANJA",
        0x83CC: "HANJA",
        0x83CD: "HANJA",
        0x83CE: "HANJA",
        0x83CF: "HANJA",
        0x83D0: "HANJA",
        0x83D1: "HANJA",
        0x83D2: "HANJA",
        0x83D3: "HANJA",
        0x83D4: "HANJA",
        0x83D5: "HANJA",
        0x83D6: "HANJA",
        0x83D7: "HANJA",
        0x83D8: "HANJA",
        0x83D9: "HANJA",
        0x83DA: "HANJA",
        0x83DB: "HANJA",
        0x83DC: "HANJA",
        0x83DD: "HANJA",
        0x83DE: "HANJA",
        0x83DF: "HANJA",
        0x83E0: "HANJA",
        0x83E1: "HANJA",
        0x83E2: "HANJA",
        0x83E3: "HANJA",
        0x83E4: "HANJA",
        0x83E5: "HANJA",
        0x83E6: "HANJA",
        0x83E7: "HANJA",
        0x83E8: "HANJA",
        0x83E9: "HANJA",
        0x83EA: "HANJA",
        0x83EB: "HANJA",
        0x83EC: "HANJA",
        0x83ED: "HANJA",
        0x83EE: "HANJA",
        0x83EF: "HANJA",
        0x83F0: "HANJA",
        0x83F1: "HANJA",
        0x83F2: "HANJA",
        0x83F3: "HANJA",
        0x83F4: "HANJA",
        0x83F5: "HANJA",
        0x83F6: "HANJA",
        0x83F7: "HANJA",
        0x83F8: "HANJA",
        0x83F9: "HANJA",
        0x83FA: "HANJA",
        0x83FB: "HANJA",
        0x83FC: "HANJA",
        0x83FD: "HANJA",
        0x83FE: "HANJA",
        0x83FF: "HANJA",
        0x8400: "HANJA",
        0x8401: "HANJA",
        0x8402: "HANJA",
        0x8403: "HANJA",
        0x8404: "HANJA",
        0x8405: "HANJA",
        0x8406: "HANJA",
        0x8407: "HANJA",
        0x8408: "HANJA",
        0x8409: "HANJA",
        0x840A: "HANJA",
        0x840B: "HANJA",
        0x840C: "HANJA",
        0x840D: "HANJA",
        0x840E: "HANJA",
        0x840F: "HANJA",
        0x8410: "HANJA",
        0x8411: "HANJA",
        0x8412: "HANJA",
        0x8413: "HANJA",
        0x8414: "HANJA",
        0x8415: "HANJA",
        0x8416: "HANJA",
        0x8417: "HANJA",
        0x8418: "HANJA",
        0x8419: "HANJA",
        0x841A: "HANJA",
        0x841B: "HANJA",
        0x841C: "HANJA",
        0x841D: "HANJA",
        0x841E: "HANJA",
        0x841F: "HANJA",
        0x8420: "HANJA",
        0x8421: "HANJA",
        0x8422: "HANJA",
        0x8423: "HANJA",
        0x8424: "HANJA",
        0x8425: "HANJA",
        0x8426: "HANJA",
        0x8427: "HANJA",
        0x8428: "HANJA",
        0x8429: "HANJA",
        0x842A: "HANJA",
        0x842B: "HANJA",
        0x842C: "HANJA",
        0x842D: "HANJA",
        0x842E: "HANJA",
        0x842F: "HANJA",
        0x8430: "HANJA",
        0x8431: "HANJA",
        0x8432: "HANJA",
        0x8433: "HANJA",
        0x8434: "HANJA",
        0x8435: "HANJA",
        0x8436: "HANJA",
        0x8437: "HANJA",
        0x8438: "HANJA",
        0x8439: "HANJA",
        0x843A: "HANJA",
        0x843B: "HANJA",
        0x843C: "HANJA",
        0x843D: "HANJA",
        0x843E: "HANJA",
        0x843F: "HANJA",
        0x8440: "HANJA",
        0x8441: "HANJA",
        0x8442: "HANJA",
        0x8443: "HANJA",
        0x8444: "HANJA",
        0x8445: "HANJA",
        0x8446: "HANJA",
        0x8447: "HANJA",
        0x8448: "HANJA",
        0x8449: "HANJA",
        0x844A: "HANJA",
        0x844B: "HANJA",
        0x844C: "HANJA",
        0x844D: "HANJA",
        0x844E: "HANJA",
        0x844F: "HANJA",
        0x8450: "HANJA",
        0x8451: "HANJA",
        0x8452: "HANJA",
        0x8453: "HANJA",
        0x8454: "HANJA",
        0x8455: "HANJA",
        0x8456: "HANJA",
        0x8457: "HANJA",
        0x8458: "HANJA",
        0x8459: "HANJA",
        0x845A: "HANJA",
        0x845B: "HANJA",
        0x845C: "HANJA",
        0x845D: "HANJA",
        0x845E: "HANJA",
        0x845F: "HANJA",
        0x8460: "HANJA",
        0x8461: "HANJA",
        0x8462: "HANJA",
        0x8463: "HANJA",
        0x8464: "HANJA",
        0x8465: "HANJA",
        0x8466: "HANJA",
        0x8467: "HANJA",
        0x8468: "HANJA",
        0x8469: "HANJA",
        0x846A: "HANJA",
        0x846B: "HANJA",
        0x846C: "HANJA",
        0x846D: "HANJA",
        0x846E: "HANJA",
        0x846F: "HANJA",
        0x8470: "HANJA",
        0x8471: "HANJA",
        0x8472: "HANJA",
        0x8473: "HANJA",
        0x8474: "HANJA",
        0x8475: "HANJA",
        0x8476: "HANJA",
        0x8477: "HANJA",
        0x8478: "HANJA",
        0x8479: "HANJA",
        0x847A: "HANJA",
        0x847B: "HANJA",
        0x847C: "HANJA",
        0x847D: "HANJA",
        0x847E: "HANJA",
        0x847F: "HANJA",
        0x8480: "HANJA",
        0x8481: "HANJA",
        0x8482: "HANJA",
        0x8483: "HANJA",
        0x8484: "HANJA",
        0x8485: "HANJA",
        0x8486: "HANJA",
        0x8487: "HANJA",
        0x8488: "HANJA",
        0x8489: "HANJA",
        0x848A: "HANJA",
        0x848B: "HANJA",
        0x848C: "HANJA",
        0x848D: "HANJA",
        0x848E: "HANJA",
        0x848F: "HANJA",
        0x8490: "HANJA",
        0x8491: "HANJA",
        0x8492: "HANJA",
        0x8493: "HANJA",
        0x8494: "HANJA",
        0x8495: "HANJA",
        0x8496: "HANJA",
        0x8497: "HANJA",
        0x8498: "HANJA",
        0x8499: "HANJA",
        0x849A: "HANJA",
        0x849B: "HANJA",
        0x849C: "HANJA",
        0x849D: "HANJA",
        0x849E: "HANJA",
        0x849F: "HANJA",
        0x84A0: "HANJA",
        0x84A1: "HANJA",
        0x84A2: "HANJA",
        0x84A3: "HANJA",
        0x84A4: "HANJA",
        0x84A5: "HANJA",
        0x84A6: "HANJA",
        0x84A7: "HANJA",
        0x84A8: "HANJA",
        0x84A9: "HANJA",
        0x84AA: "HANJA",
        0x84AB: "HANJA",
        0x84AC: "HANJA",
        0x84AD: "HANJA",
        0x84AE: "HANJA",
        0x84AF: "HANJA",
        0x84B0: "HANJA",
        0x84B1: "HANJA",
        0x84B2: "HANJA",
        0x84B3: "HANJA",
        0x84B4: "HANJA",
        0x84B5: "HANJA",
        0x84B6: "HANJA",
        0x84B7: "HANJA",
        0x84B8: "HANJA",
        0x84B9: "HANJA",
        0x84BA: "HANJA",
        0x84BB: "HANJA",
        0x84BC: "HANJA",
        0x84BD: "HANJA",
        0x84BE: "HANJA",
        0x84BF: "HANJA",
        0x84C0: "HANJA",
        0x84C1: "HANJA",
        0x84C2: "HANJA",
        0x84C3: "HANJA",
        0x84C4: "HANJA",
        0x84C5: "HANJA",
        0x84C6: "HANJA",
        0x84C7: "HANJA",
        0x84C8: "HANJA",
        0x84C9: "HANJA",
        0x84CA: "HANJA",
        0x84CB: "HANJA",
        0x84CC: "HANJA",
        0x84CD: "HANJA",
        0x84CE: "HANJA",
        0x84CF: "HANJA",
        0x84D0: "HANJA",
        0x84D1: "HANJA",
        0x84D2: "HANJA",
        0x84D3: "HANJA",
        0x84D4: "HANJA",
        0x84D5: "HANJA",
        0x84D6: "HANJA",
        0x84D7: "HANJA",
        0x84D8: "HANJA",
        0x84D9: "HANJA",
        0x84DA: "HANJA",
        0x84DB: "HANJA",
        0x84DC: "HANJA",
        0x84DD: "HANJA",
        0x84DE: "HANJA",
        0x84DF: "HANJA",
        0x84E0: "HANJA",
        0x84E1: "HANJA",
        0x84E2: "HANJA",
        0x84E3: "HANJA",
        0x84E4: "HANJA",
        0x84E5: "HANJA",
        0x84E6: "HANJA",
        0x84E7: "HANJA",
        0x84E8: "HANJA",
        0x84E9: "HANJA",
        0x84EA: "HANJA",
        0x84EB: "HANJA",
        0x84EC: "HANJA",
        0x84ED: "HANJA",
        0x84EE: "HANJA",
        0x84EF: "HANJA",
        0x84F0: "HANJA",
        0x84F1: "HANJA",
        0x84F2: "HANJA",
        0x84F3: "HANJA",
        0x84F4: "HANJA",
        0x84F5: "HANJA",
        0x84F6: "HANJA",
        0x84F7: "HANJA",
        0x84F8: "HANJA",
        0x84F9: "HANJA",
        0x84FA: "HANJA",
        0x84FB: "HANJA",
        0x84FC: "HANJA",
        0x84FD: "HANJA",
        0x84FE: "HANJA",
        0x84FF: "HANJA",
        0x8500: "HANJA",
        0x8501: "HANJA",
        0x8502: "HANJA",
        0x8503: "HANJA",
        0x8504: "HANJA",
        0x8505: "HANJA",
        0x8506: "HANJA",
        0x8507: "HANJA",
        0x8508: "HANJA",
        0x8509: "HANJA",
        0x850A: "HANJA",
        0x850B: "HANJA",
        0x850C: "HANJA",
        0x850D: "HANJA",
        0x850E: "HANJA",
        0x850F: "HANJA",
        0x8510: "HANJA",
        0x8511: "HANJA",
        0x8512: "HANJA",
        0x8513: "HANJA",
        0x8514: "HANJA",
        0x8515: "HANJA",
        0x8516: "HANJA",
        0x8517: "HANJA",
        0x8518: "HANJA",
        0x8519: "HANJA",
        0x851A: "HANJA",
        0x851B: "HANJA",
        0x851C: "HANJA",
        0x851D: "HANJA",
        0x851E: "HANJA",
        0x851F: "HANJA",
        0x8520: "HANJA",
        0x8521: "HANJA",
        0x8522: "HANJA",
        0x8523: "HANJA",
        0x8524: "HANJA",
        0x8525: "HANJA",
        0x8526: "HANJA",
        0x8527: "HANJA",
        0x8528: "HANJA",
        0x8529: "HANJA",
        0x852A: "HANJA",
        0x852B: "HANJA",
        0x852C: "HANJA",
        0x852D: "HANJA",
        0x852E: "HANJA",
        0x852F: "HANJA",
        0x8530: "HANJA",
        0x8531: "HANJA",
        0x8532: "HANJA",
        0x8533: "HANJA",
        0x8534: "HANJA",
        0x8535: "HANJA",
        0x8536: "HANJA",
        0x8537: "HANJA",
        0x8538: "HANJA",
        0x8539: "HANJA",
        0x853A: "HANJA",
        0x853B: "HANJA",
        0x853C: "HANJA",
        0x853D: "HANJA",
        0x853E: "HANJA",
        0x853F: "HANJA",
        0x8540: "HANJA",
        0x8541: "HANJA",
        0x8542: "HANJA",
        0x8543: "HANJA",
        0x8544: "HANJA",
        0x8545: "HANJA",
        0x8546: "HANJA",
        0x8547: "HANJA",
        0x8548: "HANJA",
        0x8549: "HANJA",
        0x854A: "HANJA",
        0x854B: "HANJA",
        0x854C: "HANJA",
        0x854D: "HANJA",
        0x854E: "HANJA",
        0x854F: "HANJA",
        0x8550: "HANJA",
        0x8551: "HANJA",
        0x8552: "HANJA",
        0x8553: "HANJA",
        0x8554: "HANJA",
        0x8555: "HANJA",
        0x8556: "HANJA",
        0x8557: "HANJA",
        0x8558: "HANJA",
        0x8559: "HANJA",
        0x855A: "HANJA",
        0x855B: "HANJA",
        0x855C: "HANJA",
        0x855D: "HANJA",
        0x855E: "HANJA",
        0x855F: "HANJA",
        0x8560: "HANJA",
        0x8561: "HANJA",
        0x8562: "HANJA",
        0x8563: "HANJA",
        0x8564: "HANJA",
        0x8565: "HANJA",
        0x8566: "HANJA",
        0x8567: "HANJA",
        0x8568: "HANJA",
        0x8569: "HANJA",
        0x856A: "HANJA",
        0x856B: "HANJA",
        0x856C: "HANJA",
        0x856D: "HANJA",
        0x856E: "HANJA",
        0x856F: "HANJA",
        0x8570: "HANJA",
        0x8571: "HANJA",
        0x8572: "HANJA",
        0x8573: "HANJA",
        0x8574: "HANJA",
        0x8575: "HANJA",
        0x8576: "HANJA",
        0x8577: "HANJA",
        0x8578: "HANJA",
        0x8579: "HANJA",
        0x857A: "HANJA",
        0x857B: "HANJA",
        0x857C: "HANJA",
        0x857D: "HANJA",
        0x857E: "HANJA",
        0x857F: "HANJA",
        0x8580: "HANJA",
        0x8581: "HANJA",
        0x8582: "HANJA",
        0x8583: "HANJA",
        0x8584: "HANJA",
        0x8585: "HANJA",
        0x8586: "HANJA",
        0x8587: "HANJA",
        0x8588: "HANJA",
        0x8589: "HANJA",
        0x858A: "HANJA",
        0x858B: "HANJA",
        0x858C: "HANJA",
        0x858D: "HANJA",
        0x858E: "HANJA",
        0x858F: "HANJA",
        0x8590: "HANJA",
        0x8591: "HANJA",
        0x8592: "HANJA",
        0x8593: "HANJA",
        0x8594: "HANJA",
        0x8595: "HANJA",
        0x8596: "HANJA",
        0x8597: "HANJA",
        0x8598: "HANJA",
        0x8599: "HANJA",
        0x859A: "HANJA",
        0x859B: "HANJA",
        0x859C: "HANJA",
        0x859D: "HANJA",
        0x859E: "HANJA",
        0x859F: "HANJA",
        0x85A0: "HANJA",
        0x85A1: "HANJA",
        0x85A2: "HANJA",
        0x85A3: "HANJA",
        0x85A4: "HANJA",
        0x85A5: "HANJA",
        0x85A6: "HANJA",
        0x85A7: "HANJA",
        0x85A8: "HANJA",
        0x85A9: "HANJA",
        0x85AA: "HANJA",
        0x85AB: "HANJA",
        0x85AC: "HANJA",
        0x85AD: "HANJA",
        0x85AE: "HANJA",
        0x85AF: "HANJA",
        0x85B0: "HANJA",
        0x85B1: "HANJA",
        0x85B2: "HANJA",
        0x85B3: "HANJA",
        0x85B4: "HANJA",
        0x85B5: "HANJA",
        0x85B6: "HANJA",
        0x85B7: "HANJA",
        0x85B8: "HANJA",
        0x85B9: "HANJA",
        0x85BA: "HANJA",
        0x85BB: "HANJA",
        0x85BC: "HANJA",
        0x85BD: "HANJA",
        0x85BE: "HANJA",
        0x85BF: "HANJA",
        0x85C0: "HANJA",
        0x85C1: "HANJA",
        0x85C2: "HANJA",
        0x85C3: "HANJA",
        0x85C4: "HANJA",
        0x85C5: "HANJA",
        0x85C6: "HANJA",
        0x85C7: "HANJA",
        0x85C8: "HANJA",
        0x85C9: "HANJA",
        0x85CA: "HANJA",
        0x85CB: "HANJA",
        0x85CC: "HANJA",
        0x85CD: "HANJA",
        0x85CE: "HANJA",
        0x85CF: "HANJA",
        0x85D0: "HANJA",
        0x85D1: "HANJA",
        0x85D2: "HANJA",
        0x85D3: "HANJA",
        0x85D4: "HANJA",
        0x85D5: "HANJA",
        0x85D6: "HANJA",
        0x85D7: "HANJA",
        0x85D8: "HANJA",
        0x85D9: "HANJA",
        0x85DA: "HANJA",
        0x85DB: "HANJA",
        0x85DC: "HANJA",
        0x85DD: "HANJA",
        0x85DE: "HANJA",
        0x85DF: "HANJA",
        0x85E0: "HANJA",
        0x85E1: "HANJA",
        0x85E2: "HANJA",
        0x85E3: "HANJA",
        0x85E4: "HANJA",
        0x85E5: "HANJA",
        0x85E6: "HANJA",
        0x85E7: "HANJA",
        0x85E8: "HANJA",
        0x85E9: "HANJA",
        0x85EA: "HANJA",
        0x85EB: "HANJA",
        0x85EC: "HANJA",
        0x85ED: "HANJA",
        0x85EE: "HANJA",
        0x85EF: "HANJA",
        0x85F0: "HANJA",
        0x85F1: "HANJA",
        0x85F2: "HANJA",
        0x85F3: "HANJA",
        0x85F4: "HANJA",
        0x85F5: "HANJA",
        0x85F6: "HANJA",
        0x85F7: "HANJA",
        0x85F8: "HANJA",
        0x85F9: "HANJA",
        0x85FA: "HANJA",
        0x85FB: "HANJA",
        0x85FC: "HANJA",
        0x85FD: "HANJA",
        0x85FE: "HANJA",
        0x85FF: "HANJA",
        0x8600: "HANJA",
        0x8601: "HANJA",
        0x8602: "HANJA",
        0x8603: "HANJA",
        0x8604: "HANJA",
        0x8605: "HANJA",
        0x8606: "HANJA",
        0x8607: "HANJA",
        0x8608: "HANJA",
        0x8609: "HANJA",
        0x860A: "HANJA",
        0x860B: "HANJA",
        0x860C: "HANJA",
        0x860D: "HANJA",
        0x860E: "HANJA",
        0x860F: "HANJA",
        0x8610: "HANJA",
        0x8611: "HANJA",
        0x8612: "HANJA",
        0x8613: "HANJA",
        0x8614: "HANJA",
        0x8615: "HANJA",
        0x8616: "HANJA",
        0x8617: "HANJA",
        0x8618: "HANJA",
        0x8619: "HANJA",
        0x861A: "HANJA",
        0x861B: "HANJA",
        0x861C: "HANJA",
        0x861D: "HANJA",
        0x861E: "HANJA",
        0x861F: "HANJA",
        0x8620: "HANJA",
        0x8621: "HANJA",
        0x8622: "HANJA",
        0x8623: "HANJA",
        0x8624: "HANJA",
        0x8625: "HANJA",
        0x8626: "HANJA",
        0x8627: "HANJA",
        0x8628: "HANJA",
        0x8629: "HANJA",
        0x862A: "HANJA",
        0x862B: "HANJA",
        0x862C: "HANJA",
        0x862D: "HANJA",
        0x862E: "HANJA",
        0x862F: "HANJA",
        0x8630: "HANJA",
        0x8631: "HANJA",
        0x8632: "HANJA",
        0x8633: "HANJA",
        0x8634: "HANJA",
        0x8635: "HANJA",
        0x8636: "HANJA",
        0x8637: "HANJA",
        0x8638: "HANJA",
        0x8639: "HANJA",
        0x863A: "HANJA",
        0x863B: "HANJA",
        0x863C: "HANJA",
        0x863D: "HANJA",
        0x863E: "HANJA",
        0x863F: "HANJA",
        0x8640: "HANJA",
        0x8641: "HANJA",
        0x8642: "HANJA",
        0x8643: "HANJA",
        0x8644: "HANJA",
        0x8645: "HANJA",
        0x8646: "HANJA",
        0x8647: "HANJA",
        0x8648: "HANJA",
        0x8649: "HANJA",
        0x864A: "HANJA",
        0x864B: "HANJA",
        0x864C: "HANJA",
        0x864D: "HANJA",
        0x864E: "HANJA",
        0x864F: "HANJA",
        0x8650: "HANJA",
        0x8651: "HANJA",
        0x8652: "HANJA",
        0x8653: "HANJA",
        0x8654: "HANJA",
        0x8655: "HANJA",
        0x8656: "HANJA",
        0x8657: "HANJA",
        0x8658: "HANJA",
        0x8659: "HANJA",
        0x865A: "HANJA",
        0x865B: "HANJA",
        0x865C: "HANJA",
        0x865D: "HANJA",
        0x865E: "HANJA",
        0x865F: "HANJA",
        0x8660: "HANJA",
        0x8661: "HANJA",
        0x8662: "HANJA",
        0x8663: "HANJA",
        0x8664: "HANJA",
        0x8665: "HANJA",
        0x8666: "HANJA",
        0x8667: "HANJA",
        0x8668: "HANJA",
        0x8669: "HANJA",
        0x866A: "HANJA",
        0x866B: "HANJA",
        0x866C: "HANJA",
        0x866D: "HANJA",
        0x866E: "HANJA",
        0x866F: "HANJA",
        0x8670: "HANJA",
        0x8671: "HANJA",
        0x8672: "HANJA",
        0x8673: "HANJA",
        0x8674: "HANJA",
        0x8675: "HANJA",
        0x8676: "HANJA",
        0x8677: "HANJA",
        0x8678: "HANJA",
        0x8679: "HANJA",
        0x867A: "HANJA",
        0x867B: "HANJA",
        0x867C: "HANJA",
        0x867D: "HANJA",
        0x867E: "HANJA",
        0x867F: "HANJA",
        0x8680: "HANJA",
        0x8681: "HANJA",
        0x8682: "HANJA",
        0x8683: "HANJA",
        0x8684: "HANJA",
        0x8685: "HANJA",
        0x8686: "HANJA",
        0x8687: "HANJA",
        0x8688: "HANJA",
        0x8689: "HANJA",
        0x868A: "HANJA",
        0x868B: "HANJA",
        0x868C: "HANJA",
        0x868D: "HANJA",
        0x868E: "HANJA",
        0x868F: "HANJA",
        0x8690: "HANJA",
        0x8691: "HANJA",
        0x8692: "HANJA",
        0x8693: "HANJA",
        0x8694: "HANJA",
        0x8695: "HANJA",
        0x8696: "HANJA",
        0x8697: "HANJA",
        0x8698: "HANJA",
        0x8699: "HANJA",
        0x869A: "HANJA",
        0x869B: "HANJA",
        0x869C: "HANJA",
        0x869D: "HANJA",
        0x869E: "HANJA",
        0x869F: "HANJA",
        0x86A0: "HANJA",
        0x86A1: "HANJA",
        0x86A2: "HANJA",
        0x86A3: "HANJA",
        0x86A4: "HANJA",
        0x86A5: "HANJA",
        0x86A6: "HANJA",
        0x86A7: "HANJA",
        0x86A8: "HANJA",
        0x86A9: "HANJA",
        0x86AA: "HANJA",
        0x86AB: "HANJA",
        0x86AC: "HANJA",
        0x86AD: "HANJA",
        0x86AE: "HANJA",
        0x86AF: "HANJA",
        0x86B0: "HANJA",
        0x86B1: "HANJA",
        0x86B2: "HANJA",
        0x86B3: "HANJA",
        0x86B4: "HANJA",
        0x86B5: "HANJA",
        0x86B6: "HANJA",
        0x86B7: "HANJA",
        0x86B8: "HANJA",
        0x86B9: "HANJA",
        0x86BA: "HANJA",
        0x86BB: "HANJA",
        0x86BC: "HANJA",
        0x86BD: "HANJA",
        0x86BE: "HANJA",
        0x86BF: "HANJA",
        0x86C0: "HANJA",
        0x86C1: "HANJA",
        0x86C2: "HANJA",
        0x86C3: "HANJA",
        0x86C4: "HANJA",
        0x86C5: "HANJA",
        0x86C6: "HANJA",
        0x86C7: "HANJA",
        0x86C8: "HANJA",
        0x86C9: "HANJA",
        0x86CA: "HANJA",
        0x86CB: "HANJA",
        0x86CC: "HANJA",
        0x86CD: "HANJA",
        0x86CE: "HANJA",
        0x86CF: "HANJA",
        0x86D0: "HANJA",
        0x86D1: "HANJA",
        0x86D2: "HANJA",
        0x86D3: "HANJA",
        0x86D4: "HANJA",
        0x86D5: "HANJA",
        0x86D6: "HANJA",
        0x86D7: "HANJA",
        0x86D8: "HANJA",
        0x86D9: "HANJA",
        0x86DA: "HANJA",
        0x86DB: "HANJA",
        0x86DC: "HANJA",
        0x86DD: "HANJA",
        0x86DE: "HANJA",
        0x86DF: "HANJA",
        0x86E0: "HANJA",
        0x86E1: "HANJA",
        0x86E2: "HANJA",
        0x86E3: "HANJA",
        0x86E4: "HANJA",
        0x86E5: "HANJA",
        0x86E6: "HANJA",
        0x86E7: "HANJA",
        0x86E8: "HANJA",
        0x86E9: "HANJA",
        0x86EA: "HANJA",
        0x86EB: "HANJA",
        0x86EC: "HANJA",
        0x86ED: "HANJA",
        0x86EE: "HANJA",
        0x86EF: "HANJA",
        0x86F0: "HANJA",
        0x86F1: "HANJA",
        0x86F2: "HANJA",
        0x86F3: "HANJA",
        0x86F4: "HANJA",
        0x86F5: "HANJA",
        0x86F6: "HANJA",
        0x86F7: "HANJA",
        0x86F8: "HANJA",
        0x86F9: "HANJA",
        0x86FA: "HANJA",
        0x86FB: "HANJA",
        0x86FC: "HANJA",
        0x86FD: "HANJA",
        0x86FE: "HANJA",
        0x86FF: "HANJA",
        0x8700: "HANJA",
        0x8701: "HANJA",
        0x8702: "HANJA",
        0x8703: "HANJA",
        0x8704: "HANJA",
        0x8705: "HANJA",
        0x8706: "HANJA",
        0x8707: "HANJA",
        0x8708: "HANJA",
        0x8709: "HANJA",
        0x870A: "HANJA",
        0x870B: "HANJA",
        0x870C: "HANJA",
        0x870D: "HANJA",
        0x870E: "HANJA",
        0x870F: "HANJA",
        0x8710: "HANJA",
        0x8711: "HANJA",
        0x8712: "HANJA",
        0x8713: "HANJA",
        0x8714: "HANJA",
        0x8715: "HANJA",
        0x8716: "HANJA",
        0x8717: "HANJA",
        0x8718: "HANJA",
        0x8719: "HANJA",
        0x871A: "HANJA",
        0x871B: "HANJA",
        0x871C: "HANJA",
        0x871D: "HANJA",
        0x871E: "HANJA",
        0x871F: "HANJA",
        0x8720: "HANJA",
        0x8721: "HANJA",
        0x8722: "HANJA",
        0x8723: "HANJA",
        0x8724: "HANJA",
        0x8725: "HANJA",
        0x8726: "HANJA",
        0x8727: "HANJA",
        0x8728: "HANJA",
        0x8729: "HANJA",
        0x872A: "HANJA",
        0x872B: "HANJA",
        0x872C: "HANJA",
        0x872D: "HANJA",
        0x872E: "HANJA",
        0x872F: "HANJA",
        0x8730: "HANJA",
        0x8731: "HANJA",
        0x8732: "HANJA",
        0x8733: "HANJA",
        0x8734: "HANJA",
        0x8735: "HANJA",
        0x8736: "HANJA",
        0x8737: "HANJA",
        0x8738: "HANJA",
        0x8739: "HANJA",
        0x873A: "HANJA",
        0x873B: "HANJA",
        0x873C: "HANJA",
        0x873D: "HANJA",
        0x873E: "HANJA",
        0x873F: "HANJA",
        0x8740: "HANJA",
        0x8741: "HANJA",
        0x8742: "HANJA",
        0x8743: "HANJA",
        0x8744: "HANJA",
        0x8745: "HANJA",
        0x8746: "HANJA",
        0x8747: "HANJA",
        0x8748: "HANJA",
        0x8749: "HANJA",
        0x874A: "HANJA",
        0x874B: "HANJA",
        0x874C: "HANJA",
        0x874D: "HANJA",
        0x874E: "HANJA",
        0x874F: "HANJA",
        0x8750: "HANJA",
        0x8751: "HANJA",
        0x8752: "HANJA",
        0x8753: "HANJA",
        0x8754: "HANJA",
        0x8755: "HANJA",
        0x8756: "HANJA",
        0x8757: "HANJA",
        0x8758: "HANJA",
        0x8759: "HANJA",
        0x875A: "HANJA",
        0x875B: "HANJA",
        0x875C: "HANJA",
        0x875D: "HANJA",
        0x875E: "HANJA",
        0x875F: "HANJA",
        0x8760: "HANJA",
        0x8761: "HANJA",
        0x8762: "HANJA",
        0x8763: "HANJA",
        0x8764: "HANJA",
        0x8765: "HANJA",
        0x8766: "HANJA",
        0x8767: "HANJA",
        0x8768: "HANJA",
        0x8769: "HANJA",
        0x876A: "HANJA",
        0x876B: "HANJA",
        0x876C: "HANJA",
        0x876D: "HANJA",
        0x876E: "HANJA",
        0x876F: "HANJA",
        0x8770: "HANJA",
        0x8771: "HANJA",
        0x8772: "HANJA",
        0x8773: "HANJA",
        0x8774: "HANJA",
        0x8775: "HANJA",
        0x8776: "HANJA",
        0x8777: "HANJA",
        0x8778: "HANJA",
        0x8779: "HANJA",
        0x877A: "HANJA",
        0x877B: "HANJA",
        0x877C: "HANJA",
        0x877D: "HANJA",
        0x877E: "HANJA",
        0x877F: "HANJA",
        0x8780: "HANJA",
        0x8781: "HANJA",
        0x8782: "HANJA",
        0x8783: "HANJA",
        0x8784: "HANJA",
        0x8785: "HANJA",
        0x8786: "HANJA",
        0x8787: "HANJA",
        0x8788: "HANJA",
        0x8789: "HANJA",
        0x878A: "HANJA",
        0x878B: "HANJA",
        0x878C: "HANJA",
        0x878D: "HANJA",
        0x878E: "HANJA",
        0x878F: "HANJA",
        0x8790: "HANJA",
        0x8791: "HANJA",
        0x8792: "HANJA",
        0x8793: "HANJA",
        0x8794: "HANJA",
        0x8795: "HANJA",
        0x8796: "HANJA",
        0x8797: "HANJA",
        0x8798: "HANJA",
        0x8799: "HANJA",
        0x879A: "HANJA",
        0x879B: "HANJA",
        0x879C: "HANJA",
        0x879D: "HANJA",
        0x879E: "HANJA",
        0x879F: "HANJA",
        0x87A0: "HANJA",
        0x87A1: "HANJA",
        0x87A2: "HANJA",
        0x87A3: "HANJA",
        0x87A4: "HANJA",
        0x87A5: "HANJA",
        0x87A6: "HANJA",
        0x87A7: "HANJA",
        0x87A8: "HANJA",
        0x87A9: "HANJA",
        0x87AA: "HANJA",
        0x87AB: "HANJA",
        0x87AC: "HANJA",
        0x87AD: "HANJA",
        0x87AE: "HANJA",
        0x87AF: "HANJA",
        0x87B0: "HANJA",
        0x87B1: "HANJA",
        0x87B2: "HANJA",
        0x87B3: "HANJA",
        0x87B4: "HANJA",
        0x87B5: "HANJA",
        0x87B6: "HANJA",
        0x87B7: "HANJA",
        0x87B8: "HANJA",
        0x87B9: "HANJA",
        0x87BA: "HANJA",
        0x87BB: "HANJA",
        0x87BC: "HANJA",
        0x87BD: "HANJA",
        0x87BE: "HANJA",
        0x87BF: "HANJA",
        0x87C0: "HANJA",
        0x87C1: "HANJA",
        0x87C2: "HANJA",
        0x87C3: "HANJA",
        0x87C4: "HANJA",
        0x87C5: "HANJA",
        0x87C6: "HANJA",
        0x87C7: "HANJA",
        0x87C8: "HANJA",
        0x87C9: "HANJA",
        0x87CA: "HANJA",
        0x87CB: "HANJA",
        0x87CC: "HANJA",
        0x87CD: "HANJA",
        0x87CE: "HANJA",
        0x87CF: "HANJA",
        0x87D0: "HANJA",
        0x87D1: "HANJA",
        0x87D2: "HANJA",
        0x87D3: "HANJA",
        0x87D4: "HANJA",
        0x87D5: "HANJA",
        0x87D6: "HANJA",
        0x87D7: "HANJA",
        0x87D8: "HANJA",
        0x87D9: "HANJA",
        0x87DA: "HANJA",
        0x87DB: "HANJA",
        0x87DC: "HANJA",
        0x87DD: "HANJA",
        0x87DE: "HANJA",
        0x87DF: "HANJA",
        0x87E0: "HANJA",
        0x87E1: "HANJA",
        0x87E2: "HANJA",
        0x87E3: "HANJA",
        0x87E4: "HANJA",
        0x87E5: "HANJA",
        0x87E6: "HANJA",
        0x87E7: "HANJA",
        0x87E8: "HANJA",
        0x87E9: "HANJA",
        0x87EA: "HANJA",
        0x87EB: "HANJA",
        0x87EC: "HANJA",
        0x87ED: "HANJA",
        0x87EE: "HANJA",
        0x87EF: "HANJA",
        0x87F0: "HANJA",
        0x87F1: "HANJA",
        0x87F2: "HANJA",
        0x87F3: "HANJA",
        0x87F4: "HANJA",
        0x87F5: "HANJA",
        0x87F6: "HANJA",
        0x87F7: "HANJA",
        0x87F8: "HANJA",
        0x87F9: "HANJA",
        0x87FA: "HANJA",
        0x87FB: "HANJA",
        0x87FC: "HANJA",
        0x87FD: "HANJA",
        0x87FE: "HANJA",
        0x87FF: "HANJA",
        0x8800: "HANJA",
        0x8801: "HANJA",
        0x8802: "HANJA",
        0x8803: "HANJA",
        0x8804: "HANJA",
        0x8805: "HANJA",
        0x8806: "HANJA",
        0x8807: "HANJA",
        0x8808: "HANJA",
        0x8809: "HANJA",
        0x880A: "HANJA",
        0x880B: "HANJA",
        0x880C: "HANJA",
        0x880D: "HANJA",
        0x880E: "HANJA",
        0x880F: "HANJA",
        0x8810: "HANJA",
        0x8811: "HANJA",
        0x8812: "HANJA",
        0x8813: "HANJA",
        0x8814: "HANJA",
        0x8815: "HANJA",
        0x8816: "HANJA",
        0x8817: "HANJA",
        0x8818: "HANJA",
        0x8819: "HANJA",
        0x881A: "HANJA",
        0x881B: "HANJA",
        0x881C: "HANJA",
        0x881D: "HANJA",
        0x881E: "HANJA",
        0x881F: "HANJA",
        0x8820: "HANJA",
        0x8821: "HANJA",
        0x8822: "HANJA",
        0x8823: "HANJA",
        0x8824: "HANJA",
        0x8825: "HANJA",
        0x8826: "HANJA",
        0x8827: "HANJA",
        0x8828: "HANJA",
        0x8829: "HANJA",
        0x882A: "HANJA",
        0x882B: "HANJA",
        0x882C: "HANJA",
        0x882D: "HANJA",
        0x882E: "HANJA",
        0x882F: "HANJA",
        0x8830: "HANJA",
        0x8831: "HANJA",
        0x8832: "HANJA",
        0x8833: "HANJA",
        0x8834: "HANJA",
        0x8835: "HANJA",
        0x8836: "HANJA",
        0x8837: "HANJA",
        0x8838: "HANJA",
        0x8839: "HANJA",
        0x883A: "HANJA",
        0x883B: "HANJA",
        0x883C: "HANJA",
        0x883D: "HANJA",
        0x883E: "HANJA",
        0x883F: "HANJA",
        0x8840: "HANJA",
        0x8841: "HANJA",
        0x8842: "HANJA",
        0x8843: "HANJA",
        0x8844: "HANJA",
        0x8845: "HANJA",
        0x8846: "HANJA",
        0x8847: "HANJA",
        0x8848: "HANJA",
        0x8849: "HANJA",
        0x884A: "HANJA",
        0x884B: "HANJA",
        0x884C: "HANJA",
        0x884D: "HANJA",
        0x884E: "HANJA",
        0x884F: "HANJA",
        0x8850: "HANJA",
        0x8851: "HANJA",
        0x8852: "HANJA",
        0x8853: "HANJA",
        0x8854: "HANJA",
        0x8855: "HANJA",
        0x8856: "HANJA",
        0x8857: "HANJA",
        0x8858: "HANJA",
        0x8859: "HANJA",
        0x885A: "HANJA",
        0x885B: "HANJA",
        0x885C: "HANJA",
        0x885D: "HANJA",
        0x885E: "HANJA",
        0x885F: "HANJA",
        0x8860: "HANJA",
        0x8861: "HANJA",
        0x8862: "HANJA",
        0x8863: "HANJA",
        0x8864: "HANJA",
        0x8865: "HANJA",
        0x8866: "HANJA",
        0x8867: "HANJA",
        0x8868: "HANJA",
        0x8869: "HANJA",
        0x886A: "HANJA",
        0x886B: "HANJA",
        0x886C: "HANJA",
        0x886D: "HANJA",
        0x886E: "HANJA",
        0x886F: "HANJA",
        0x8870: "HANJA",
        0x8871: "HANJA",
        0x8872: "HANJA",
        0x8873: "HANJA",
        0x8874: "HANJA",
        0x8875: "HANJA",
        0x8876: "HANJA",
        0x8877: "HANJA",
        0x8878: "HANJA",
        0x8879: "HANJA",
        0x887A: "HANJA",
        0x887B: "HANJA",
        0x887C: "HANJA",
        0x887D: "HANJA",
        0x887E: "HANJA",
        0x887F: "HANJA",
        0x8880: "HANJA",
        0x8881: "HANJA",
        0x8882: "HANJA",
        0x8883: "HANJA",
        0x8884: "HANJA",
        0x8885: "HANJA",
        0x8886: "HANJA",
        0x8887: "HANJA",
        0x8888: "HANJA",
        0x8889: "HANJA",
        0x888A: "HANJA",
        0x888B: "HANJA",
        0x888C: "HANJA",
        0x888D: "HANJA",
        0x888E: "HANJA",
        0x888F: "HANJA",
        0x8890: "HANJA",
        0x8891: "HANJA",
        0x8892: "HANJA",
        0x8893: "HANJA",
        0x8894: "HANJA",
        0x8895: "HANJA",
        0x8896: "HANJA",
        0x8897: "HANJA",
        0x8898: "HANJA",
        0x8899: "HANJA",
        0x889A: "HANJA",
        0x889B: "HANJA",
        0x889C: "HANJA",
        0x889D: "HANJA",
        0x889E: "HANJA",
        0x889F: "HANJA",
        0x88A0: "HANJA",
        0x88A1: "HANJA",
        0x88A2: "HANJA",
        0x88A3: "HANJA",
        0x88A4: "HANJA",
        0x88A5: "HANJA",
        0x88A6: "HANJA",
        0x88A7: "HANJA",
        0x88A8: "HANJA",
        0x88A9: "HANJA",
        0x88AA: "HANJA",
        0x88AB: "HANJA",
        0x88AC: "HANJA",
        0x88AD: "HANJA",
        0x88AE: "HANJA",
        0x88AF: "HANJA",
        0x88B0: "HANJA",
        0x88B1: "HANJA",
        0x88B2: "HANJA",
        0x88B3: "HANJA",
        0x88B4: "HANJA",
        0x88B5: "HANJA",
        0x88B6: "HANJA",
        0x88B7: "HANJA",
        0x88B8: "HANJA",
        0x88B9: "HANJA",
        0x88BA: "HANJA",
        0x88BB: "HANJA",
        0x88BC: "HANJA",
        0x88BD: "HANJA",
        0x88BE: "HANJA",
        0x88BF: "HANJA",
        0x88C0: "HANJA",
        0x88C1: "HANJA",
        0x88C2: "HANJA",
        0x88C3: "HANJA",
        0x88C4: "HANJA",
        0x88C5: "HANJA",
        0x88C6: "HANJA",
        0x88C7: "HANJA",
        0x88C8: "HANJA",
        0x88C9: "HANJA",
        0x88CA: "HANJA",
        0x88CB: "HANJA",
        0x88CC: "HANJA",
        0x88CD: "HANJA",
        0x88CE: "HANJA",
        0x88CF: "HANJA",
        0x88D0: "HANJA",
        0x88D1: "HANJA",
        0x88D2: "HANJA",
        0x88D3: "HANJA",
        0x88D4: "HANJA",
        0x88D5: "HANJA",
        0x88D6: "HANJA",
        0x88D7: "HANJA",
        0x88D8: "HANJA",
        0x88D9: "HANJA",
        0x88DA: "HANJA",
        0x88DB: "HANJA",
        0x88DC: "HANJA",
        0x88DD: "HANJA",
        0x88DE: "HANJA",
        0x88DF: "HANJA",
        0x88E0: "HANJA",
        0x88E1: "HANJA",
        0x88E2: "HANJA",
        0x88E3: "HANJA",
        0x88E4: "HANJA",
        0x88E5: "HANJA",
        0x88E6: "HANJA",
        0x88E7: "HANJA",
        0x88E8: "HANJA",
        0x88E9: "HANJA",
        0x88EA: "HANJA",
        0x88EB: "HANJA",
        0x88EC: "HANJA",
        0x88ED: "HANJA",
        0x88EE: "HANJA",
        0x88EF: "HANJA",
        0x88F0: "HANJA",
        0x88F1: "HANJA",
        0x88F2: "HANJA",
        0x88F3: "HANJA",
        0x88F4: "HANJA",
        0x88F5: "HANJA",
        0x88F6: "HANJA",
        0x88F7: "HANJA",
        0x88F8: "HANJA",
        0x88F9: "HANJA",
        0x88FA: "HANJA",
        0x88FB: "HANJA",
        0x88FC: "HANJA",
        0x88FD: "HANJA",
        0x88FE: "HANJA",
        0x88FF: "HANJA",
        0x8900: "HANJA",
        0x8901: "HANJA",
        0x8902: "HANJA",
        0x8903: "HANJA",
        0x8904: "HANJA",
        0x8905: "HANJA",
        0x8906: "HANJA",
        0x8907: "HANJA",
        0x8908: "HANJA",
        0x8909: "HANJA",
        0x890A: "HANJA",
        0x890B: "HANJA",
        0x890C: "HANJA",
        0x890D: "HANJA",
        0x890E: "HANJA",
        0x890F: "HANJA",
        0x8910: "HANJA",
        0x8911: "HANJA",
        0x8912: "HANJA",
        0x8913: "HANJA",
        0x8914: "HANJA",
        0x8915: "HANJA",
        0x8916: "HANJA",
        0x8917: "HANJA",
        0x8918: "HANJA",
        0x8919: "HANJA",
        0x891A: "HANJA",
        0x891B: "HANJA",
        0x891C: "HANJA",
        0x891D: "HANJA",
        0x891E: "HANJA",
        0x891F: "HANJA",
        0x8920: "HANJA",
        0x8921: "HANJA",
        0x8922: "HANJA",
        0x8923: "HANJA",
        0x8924: "HANJA",
        0x8925: "HANJA",
        0x8926: "HANJA",
        0x8927: "HANJA",
        0x8928: "HANJA",
        0x8929: "HANJA",
        0x892A: "HANJA",
        0x892B: "HANJA",
        0x892C: "HANJA",
        0x892D: "HANJA",
        0x892E: "HANJA",
        0x892F: "HANJA",
        0x8930: "HANJA",
        0x8931: "HANJA",
        0x8932: "HANJA",
        0x8933: "HANJA",
        0x8934: "HANJA",
        0x8935: "HANJA",
        0x8936: "HANJA",
        0x8937: "HANJA",
        0x8938: "HANJA",
        0x8939: "HANJA",
        0x893A: "HANJA",
        0x893B: "HANJA",
        0x893C: "HANJA",
        0x893D: "HANJA",
        0x893E: "HANJA",
        0x893F: "HANJA",
        0x8940: "HANJA",
        0x8941: "HANJA",
        0x8942: "HANJA",
        0x8943: "HANJA",
        0x8944: "HANJA",
        0x8945: "HANJA",
        0x8946: "HANJA",
        0x8947: "HANJA",
        0x8948: "HANJA",
        0x8949: "HANJA",
        0x894A: "HANJA",
        0x894B: "HANJA",
        0x894C: "HANJA",
        0x894D: "HANJA",
        0x894E: "HANJA",
        0x894F: "HANJA",
        0x8950: "HANJA",
        0x8951: "HANJA",
        0x8952: "HANJA",
        0x8953: "HANJA",
        0x8954: "HANJA",
        0x8955: "HANJA",
        0x8956: "HANJA",
        0x8957: "HANJA",
        0x8958: "HANJA",
        0x8959: "HANJA",
        0x895A: "HANJA",
        0x895B: "HANJA",
        0x895C: "HANJA",
        0x895D: "HANJA",
        0x895E: "HANJA",
        0x895F: "HANJA",
        0x8960: "HANJA",
        0x8961: "HANJA",
        0x8962: "HANJA",
        0x8963: "HANJA",
        0x8964: "HANJA",
        0x8965: "HANJA",
        0x8966: "HANJA",
        0x8967: "HANJA",
        0x8968: "HANJA",
        0x8969: "HANJA",
        0x896A: "HANJA",
        0x896B: "HANJA",
        0x896C: "HANJA",
        0x896D: "HANJA",
        0x896E: "HANJA",
        0x896F: "HANJA",
        0x8970: "HANJA",
        0x8971: "HANJA",
        0x8972: "HANJA",
        0x8973: "HANJA",
        0x8974: "HANJA",
        0x8975: "HANJA",
        0x8976: "HANJA",
        0x8977: "HANJA",
        0x8978: "HANJA",
        0x8979: "HANJA",
        0x897A: "HANJA",
        0x897B: "HANJA",
        0x897C: "HANJA",
        0x897D: "HANJA",
        0x897E: "HANJA",
        0x897F: "HANJA",
        0x8980: "HANJA",
        0x8981: "HANJA",
        0x8982: "HANJA",
        0x8983: "HANJA",
        0x8984: "HANJA",
        0x8985: "HANJA",
        0x8986: "HANJA",
        0x8987: "HANJA",
        0x8988: "HANJA",
        0x8989: "HANJA",
        0x898A: "HANJA",
        0x898B: "HANJA",
        0x898C: "HANJA",
        0x898D: "HANJA",
        0x898E: "HANJA",
        0x898F: "HANJA",
        0x8990: "HANJA",
        0x8991: "HANJA",
        0x8992: "HANJA",
        0x8993: "HANJA",
        0x8994: "HANJA",
        0x8995: "HANJA",
        0x8996: "HANJA",
        0x8997: "HANJA",
        0x8998: "HANJA",
        0x8999: "HANJA",
        0x899A: "HANJA",
        0x899B: "HANJA",
        0x899C: "HANJA",
        0x899D: "HANJA",
        0x899E: "HANJA",
        0x899F: "HANJA",
        0x89A0: "HANJA",
        0x89A1: "HANJA",
        0x89A2: "HANJA",
        0x89A3: "HANJA",
        0x89A4: "HANJA",
        0x89A5: "HANJA",
        0x89A6: "HANJA",
        0x89A7: "HANJA",
        0x89A8: "HANJA",
        0x89A9: "HANJA",
        0x89AA: "HANJA",
        0x89AB: "HANJA",
        0x89AC: "HANJA",
        0x89AD: "HANJA",
        0x89AE: "HANJA",
        0x89AF: "HANJA",
        0x89B0: "HANJA",
        0x89B1: "HANJA",
        0x89B2: "HANJA",
        0x89B3: "HANJA",
        0x89B4: "HANJA",
        0x89B5: "HANJA",
        0x89B6: "HANJA",
        0x89B7: "HANJA",
        0x89B8: "HANJA",
        0x89B9: "HANJA",
        0x89BA: "HANJA",
        0x89BB: "HANJA",
        0x89BC: "HANJA",
        0x89BD: "HANJA",
        0x89BE: "HANJA",
        0x89BF: "HANJA",
        0x89C0: "HANJA",
        0x89C1: "HANJA",
        0x89C2: "HANJA",
        0x89C3: "HANJA",
        0x89C4: "HANJA",
        0x89C5: "HANJA",
        0x89C6: "HANJA",
        0x89C7: "HANJA",
        0x89C8: "HANJA",
        0x89C9: "HANJA",
        0x89CA: "HANJA",
        0x89CB: "HANJA",
        0x89CC: "HANJA",
        0x89CD: "HANJA",
        0x89CE: "HANJA",
        0x89CF: "HANJA",
        0x89D0: "HANJA",
        0x89D1: "HANJA",
        0x89D2: "HANJA",
        0x89D3: "HANJA",
        0x89D4: "HANJA",
        0x89D5: "HANJA",
        0x89D6: "HANJA",
        0x89D7: "HANJA",
        0x89D8: "HANJA",
        0x89D9: "HANJA",
        0x89DA: "HANJA",
        0x89DB: "HANJA",
        0x89DC: "HANJA",
        0x89DD: "HANJA",
        0x89DE: "HANJA",
        0x89DF: "HANJA",
        0x89E0: "HANJA",
        0x89E1: "HANJA",
        0x89E2: "HANJA",
        0x89E3: "HANJA",
        0x89E4: "HANJA",
        0x89E5: "HANJA",
        0x89E6: "HANJA",
        0x89E7: "HANJA",
        0x89E8: "HANJA",
        0x89E9: "HANJA",
        0x89EA: "HANJA",
        0x89EB: "HANJA",
        0x89EC: "HANJA",
        0x89ED: "HANJA",
        0x89EE: "HANJA",
        0x89EF: "HANJA",
        0x89F0: "HANJA",
        0x89F1: "HANJA",
        0x89F2: "HANJA",
        0x89F3: "HANJA",
        0x89F4: "HANJA",
        0x89F5: "HANJA",
        0x89F6: "HANJA",
        0x89F7: "HANJA",
        0x89F8: "HANJA",
        0x89F9: "HANJA",
        0x89FA: "HANJA",
        0x89FB: "HANJA",
        0x89FC: "HANJA",
        0x89FD: "HANJA",
        0x89FE: "HANJA",
        0x89FF: "HANJA",
        0x8A00: "HANJA",
        0x8A01: "HANJA",
        0x8A02: "HANJA",
        0x8A03: "HANJA",
        0x8A04: "HANJA",
        0x8A05: "HANJA",
        0x8A06: "HANJA",
        0x8A07: "HANJA",
        0x8A08: "HANJA",
        0x8A09: "HANJA",
        0x8A0A: "HANJA",
        0x8A0B: "HANJA",
        0x8A0C: "HANJA",
        0x8A0D: "HANJA",
        0x8A0E: "HANJA",
        0x8A0F: "HANJA",
        0x8A10: "HANJA",
        0x8A11: "HANJA",
        0x8A12: "HANJA",
        0x8A13: "HANJA",
        0x8A14: "HANJA",
        0x8A15: "HANJA",
        0x8A16: "HANJA",
        0x8A17: "HANJA",
        0x8A18: "HANJA",
        0x8A19: "HANJA",
        0x8A1A: "HANJA",
        0x8A1B: "HANJA",
        0x8A1C: "HANJA",
        0x8A1D: "HANJA",
        0x8A1E: "HANJA",
        0x8A1F: "HANJA",
        0x8A20: "HANJA",
        0x8A21: "HANJA",
        0x8A22: "HANJA",
        0x8A23: "HANJA",
        0x8A24: "HANJA",
        0x8A25: "HANJA",
        0x8A26: "HANJA",
        0x8A27: "HANJA",
        0x8A28: "HANJA",
        0x8A29: "HANJA",
        0x8A2A: "HANJA",
        0x8A2B: "HANJA",
        0x8A2C: "HANJA",
        0x8A2D: "HANJA",
        0x8A2E: "HANJA",
        0x8A2F: "HANJA",
        0x8A30: "HANJA",
        0x8A31: "HANJA",
        0x8A32: "HANJA",
        0x8A33: "HANJA",
        0x8A34: "HANJA",
        0x8A35: "HANJA",
        0x8A36: "HANJA",
        0x8A37: "HANJA",
        0x8A38: "HANJA",
        0x8A39: "HANJA",
        0x8A3A: "HANJA",
        0x8A3B: "HANJA",
        0x8A3C: "HANJA",
        0x8A3D: "HANJA",
        0x8A3E: "HANJA",
        0x8A3F: "HANJA",
        0x8A40: "HANJA",
        0x8A41: "HANJA",
        0x8A42: "HANJA",
        0x8A43: "HANJA",
        0x8A44: "HANJA",
        0x8A45: "HANJA",
        0x8A46: "HANJA",
        0x8A47: "HANJA",
        0x8A48: "HANJA",
        0x8A49: "HANJA",
        0x8A4A: "HANJA",
        0x8A4B: "HANJA",
        0x8A4C: "HANJA",
        0x8A4D: "HANJA",
        0x8A4E: "HANJA",
        0x8A4F: "HANJA",
        0x8A50: "HANJA",
        0x8A51: "HANJA",
        0x8A52: "HANJA",
        0x8A53: "HANJA",
        0x8A54: "HANJA",
        0x8A55: "HANJA",
        0x8A56: "HANJA",
        0x8A57: "HANJA",
        0x8A58: "HANJA",
        0x8A59: "HANJA",
        0x8A5A: "HANJA",
        0x8A5B: "HANJA",
        0x8A5C: "HANJA",
        0x8A5D: "HANJA",
        0x8A5E: "HANJA",
        0x8A5F: "HANJA",
        0x8A60: "HANJA",
        0x8A61: "HANJA",
        0x8A62: "HANJA",
        0x8A63: "HANJA",
        0x8A64: "HANJA",
        0x8A65: "HANJA",
        0x8A66: "HANJA",
        0x8A67: "HANJA",
        0x8A68: "HANJA",
        0x8A69: "HANJA",
        0x8A6A: "HANJA",
        0x8A6B: "HANJA",
        0x8A6C: "HANJA",
        0x8A6D: "HANJA",
        0x8A6E: "HANJA",
        0x8A6F: "HANJA",
        0x8A70: "HANJA",
        0x8A71: "HANJA",
        0x8A72: "HANJA",
        0x8A73: "HANJA",
        0x8A74: "HANJA",
        0x8A75: "HANJA",
        0x8A76: "HANJA",
        0x8A77: "HANJA",
        0x8A78: "HANJA",
        0x8A79: "HANJA",
        0x8A7A: "HANJA",
        0x8A7B: "HANJA",
        0x8A7C: "HANJA",
        0x8A7D: "HANJA",
        0x8A7E: "HANJA",
        0x8A7F: "HANJA",
        0x8A80: "HANJA",
        0x8A81: "HANJA",
        0x8A82: "HANJA",
        0x8A83: "HANJA",
        0x8A84: "HANJA",
        0x8A85: "HANJA",
        0x8A86: "HANJA",
        0x8A87: "HANJA",
        0x8A88: "HANJA",
        0x8A89: "HANJA",
        0x8A8A: "HANJA",
        0x8A8B: "HANJA",
        0x8A8C: "HANJA",
        0x8A8D: "HANJA",
        0x8A8E: "HANJA",
        0x8A8F: "HANJA",
        0x8A90: "HANJA",
        0x8A91: "HANJA",
        0x8A92: "HANJA",
        0x8A93: "HANJA",
        0x8A94: "HANJA",
        0x8A95: "HANJA",
        0x8A96: "HANJA",
        0x8A97: "HANJA",
        0x8A98: "HANJA",
        0x8A99: "HANJA",
        0x8A9A: "HANJA",
        0x8A9B: "HANJA",
        0x8A9C: "HANJA",
        0x8A9D: "HANJA",
        0x8A9E: "HANJA",
        0x8A9F: "HANJA",
        0x8AA0: "HANJA",
        0x8AA1: "HANJA",
        0x8AA2: "HANJA",
        0x8AA3: "HANJA",
        0x8AA4: "HANJA",
        0x8AA5: "HANJA",
        0x8AA6: "HANJA",
        0x8AA7: "HANJA",
        0x8AA8: "HANJA",
        0x8AA9: "HANJA",
        0x8AAA: "HANJA",
        0x8AAB: "HANJA",
        0x8AAC: "HANJA",
        0x8AAD: "HANJA",
        0x8AAE: "HANJA",
        0x8AAF: "HANJA",
        0x8AB0: "HANJA",
        0x8AB1: "HANJA",
        0x8AB2: "HANJA",
        0x8AB3: "HANJA",
        0x8AB4: "HANJA",
        0x8AB5: "HANJA",
        0x8AB6: "HANJA",
        0x8AB7: "HANJA",
        0x8AB8: "HANJA",
        0x8AB9: "HANJA",
        0x8ABA: "HANJA",
        0x8ABB: "HANJA",
        0x8ABC: "HANJA",
        0x8ABD: "HANJA",
        0x8ABE: "HANJA",
        0x8ABF: "HANJA",
        0x8AC0: "HANJA",
        0x8AC1: "HANJA",
        0x8AC2: "HANJA",
        0x8AC3: "HANJA",
        0x8AC4: "HANJA",
        0x8AC5: "HANJA",
        0x8AC6: "HANJA",
        0x8AC7: "HANJA",
        0x8AC8: "HANJA",
        0x8AC9: "HANJA",
        0x8ACA: "HANJA",
        0x8ACB: "HANJA",
        0x8ACC: "HANJA",
        0x8ACD: "HANJA",
        0x8ACE: "HANJA",
        0x8ACF: "HANJA",
        0x8AD0: "HANJA",
        0x8AD1: "HANJA",
        0x8AD2: "HANJA",
        0x8AD3: "HANJA",
        0x8AD4: "HANJA",
        0x8AD5: "HANJA",
        0x8AD6: "HANJA",
        0x8AD7: "HANJA",
        0x8AD8: "HANJA",
        0x8AD9: "HANJA",
        0x8ADA: "HANJA",
        0x8ADB: "HANJA",
        0x8ADC: "HANJA",
        0x8ADD: "HANJA",
        0x8ADE: "HANJA",
        0x8ADF: "HANJA",
        0x8AE0: "HANJA",
        0x8AE1: "HANJA",
        0x8AE2: "HANJA",
        0x8AE3: "HANJA",
        0x8AE4: "HANJA",
        0x8AE5: "HANJA",
        0x8AE6: "HANJA",
        0x8AE7: "HANJA",
        0x8AE8: "HANJA",
        0x8AE9: "HANJA",
        0x8AEA: "HANJA",
        0x8AEB: "HANJA",
        0x8AEC: "HANJA",
        0x8AED: "HANJA",
        0x8AEE: "HANJA",
        0x8AEF: "HANJA",
        0x8AF0: "HANJA",
        0x8AF1: "HANJA",
        0x8AF2: "HANJA",
        0x8AF3: "HANJA",
        0x8AF4: "HANJA",
        0x8AF5: "HANJA",
        0x8AF6: "HANJA",
        0x8AF7: "HANJA",
        0x8AF8: "HANJA",
        0x8AF9: "HANJA",
        0x8AFA: "HANJA",
        0x8AFB: "HANJA",
        0x8AFC: "HANJA",
        0x8AFD: "HANJA",
        0x8AFE: "HANJA",
        0x8AFF: "HANJA",
        0x8B00: "HANJA",
        0x8B01: "HANJA",
        0x8B02: "HANJA",
        0x8B03: "HANJA",
        0x8B04: "HANJA",
        0x8B05: "HANJA",
        0x8B06: "HANJA",
        0x8B07: "HANJA",
        0x8B08: "HANJA",
        0x8B09: "HANJA",
        0x8B0A: "HANJA",
        0x8B0B: "HANJA",
        0x8B0C: "HANJA",
        0x8B0D: "HANJA",
        0x8B0E: "HANJA",
        0x8B0F: "HANJA",
        0x8B10: "HANJA",
        0x8B11: "HANJA",
        0x8B12: "HANJA",
        0x8B13: "HANJA",
        0x8B14: "HANJA",
        0x8B15: "HANJA",
        0x8B16: "HANJA",
        0x8B17: "HANJA",
        0x8B18: "HANJA",
        0x8B19: "HANJA",
        0x8B1A: "HANJA",
        0x8B1B: "HANJA",
        0x8B1C: "HANJA",
        0x8B1D: "HANJA",
        0x8B1E: "HANJA",
        0x8B1F: "HANJA",
        0x8B20: "HANJA",
        0x8B21: "HANJA",
        0x8B22: "HANJA",
        0x8B23: "HANJA",
        0x8B24: "HANJA",
        0x8B25: "HANJA",
        0x8B26: "HANJA",
        0x8B27: "HANJA",
        0x8B28: "HANJA",
        0x8B29: "HANJA",
        0x8B2A: "HANJA",
        0x8B2B: "HANJA",
        0x8B2C: "HANJA",
        0x8B2D: "HANJA",
        0x8B2E: "HANJA",
        0x8B2F: "HANJA",
        0x8B30: "HANJA",
        0x8B31: "HANJA",
        0x8B32: "HANJA",
        0x8B33: "HANJA",
        0x8B34: "HANJA",
        0x8B35: "HANJA",
        0x8B36: "HANJA",
        0x8B37: "HANJA",
        0x8B38: "HANJA",
        0x8B39: "HANJA",
        0x8B3A: "HANJA",
        0x8B3B: "HANJA",
        0x8B3C: "HANJA",
        0x8B3D: "HANJA",
        0x8B3E: "HANJA",
        0x8B3F: "HANJA",
        0x8B40: "HANJA",
        0x8B41: "HANJA",
        0x8B42: "HANJA",
        0x8B43: "HANJA",
        0x8B44: "HANJA",
        0x8B45: "HANJA",
        0x8B46: "HANJA",
        0x8B47: "HANJA",
        0x8B48: "HANJA",
        0x8B49: "HANJA",
        0x8B4A: "HANJA",
        0x8B4B: "HANJA",
        0x8B4C: "HANJA",
        0x8B4D: "HANJA",
        0x8B4E: "HANJA",
        0x8B4F: "HANJA",
        0x8B50: "HANJA",
        0x8B51: "HANJA",
        0x8B52: "HANJA",
        0x8B53: "HANJA",
        0x8B54: "HANJA",
        0x8B55: "HANJA",
        0x8B56: "HANJA",
        0x8B57: "HANJA",
        0x8B58: "HANJA",
        0x8B59: "HANJA",
        0x8B5A: "HANJA",
        0x8B5B: "HANJA",
        0x8B5C: "HANJA",
        0x8B5D: "HANJA",
        0x8B5E: "HANJA",
        0x8B5F: "HANJA",
        0x8B60: "HANJA",
        0x8B61: "HANJA",
        0x8B62: "HANJA",
        0x8B63: "HANJA",
        0x8B64: "HANJA",
        0x8B65: "HANJA",
        0x8B66: "HANJA",
        0x8B67: "HANJA",
        0x8B68: "HANJA",
        0x8B69: "HANJA",
        0x8B6A: "HANJA",
        0x8B6B: "HANJA",
        0x8B6C: "HANJA",
        0x8B6D: "HANJA",
        0x8B6E: "HANJA",
        0x8B6F: "HANJA",
        0x8B70: "HANJA",
        0x8B71: "HANJA",
        0x8B72: "HANJA",
        0x8B73: "HANJA",
        0x8B74: "HANJA",
        0x8B75: "HANJA",
        0x8B76: "HANJA",
        0x8B77: "HANJA",
        0x8B78: "HANJA",
        0x8B79: "HANJA",
        0x8B7A: "HANJA",
        0x8B7B: "HANJA",
        0x8B7C: "HANJA",
        0x8B7D: "HANJA",
        0x8B7E: "HANJA",
        0x8B7F: "HANJA",
        0x8B80: "HANJA",
        0x8B81: "HANJA",
        0x8B82: "HANJA",
        0x8B83: "HANJA",
        0x8B84: "HANJA",
        0x8B85: "HANJA",
        0x8B86: "HANJA",
        0x8B87: "HANJA",
        0x8B88: "HANJA",
        0x8B89: "HANJA",
        0x8B8A: "HANJA",
        0x8B8B: "HANJA",
        0x8B8C: "HANJA",
        0x8B8D: "HANJA",
        0x8B8E: "HANJA",
        0x8B8F: "HANJA",
        0x8B90: "HANJA",
        0x8B91: "HANJA",
        0x8B92: "HANJA",
        0x8B93: "HANJA",
        0x8B94: "HANJA",
        0x8B95: "HANJA",
        0x8B96: "HANJA",
        0x8B97: "HANJA",
        0x8B98: "HANJA",
        0x8B99: "HANJA",
        0x8B9A: "HANJA",
        0x8B9B: "HANJA",
        0x8B9C: "HANJA",
        0x8B9D: "HANJA",
        0x8B9E: "HANJA",
        0x8B9F: "HANJA",
        0x8BA0: "HANJA",
        0x8BA1: "HANJA",
        0x8BA2: "HANJA",
        0x8BA3: "HANJA",
        0x8BA4: "HANJA",
        0x8BA5: "HANJA",
        0x8BA6: "HANJA",
        0x8BA7: "HANJA",
        0x8BA8: "HANJA",
        0x8BA9: "HANJA",
        0x8BAA: "HANJA",
        0x8BAB: "HANJA",
        0x8BAC: "HANJA",
        0x8BAD: "HANJA",
        0x8BAE: "HANJA",
        0x8BAF: "HANJA",
        0x8BB0: "HANJA",
        0x8BB1: "HANJA",
        0x8BB2: "HANJA",
        0x8BB3: "HANJA",
        0x8BB4: "HANJA",
        0x8BB5: "HANJA",
        0x8BB6: "HANJA",
        0x8BB7: "HANJA",
        0x8BB8: "HANJA",
        0x8BB9: "HANJA",
        0x8BBA: "HANJA",
        0x8BBB: "HANJA",
        0x8BBC: "HANJA",
        0x8BBD: "HANJA",
        0x8BBE: "HANJA",
        0x8BBF: "HANJA",
        0x8BC0: "HANJA",
        0x8BC1: "HANJA",
        0x8BC2: "HANJA",
        0x8BC3: "HANJA",
        0x8BC4: "HANJA",
        0x8BC5: "HANJA",
        0x8BC6: "HANJA",
        0x8BC7: "HANJA",
        0x8BC8: "HANJA",
        0x8BC9: "HANJA",
        0x8BCA: "HANJA",
        0x8BCB: "HANJA",
        0x8BCC: "HANJA",
        0x8BCD: "HANJA",
        0x8BCE: "HANJA",
        0x8BCF: "HANJA",
        0x8BD0: "HANJA",
        0x8BD1: "HANJA",
        0x8BD2: "HANJA",
        0x8BD3: "HANJA",
        0x8BD4: "HANJA",
        0x8BD5: "HANJA",
        0x8BD6: "HANJA",
        0x8BD7: "HANJA",
        0x8BD8: "HANJA",
        0x8BD9: "HANJA",
        0x8BDA: "HANJA",
        0x8BDB: "HANJA",
        0x8BDC: "HANJA",
        0x8BDD: "HANJA",
        0x8BDE: "HANJA",
        0x8BDF: "HANJA",
        0x8BE0: "HANJA",
        0x8BE1: "HANJA",
        0x8BE2: "HANJA",
        0x8BE3: "HANJA",
        0x8BE4: "HANJA",
        0x8BE5: "HANJA",
        0x8BE6: "HANJA",
        0x8BE7: "HANJA",
        0x8BE8: "HANJA",
        0x8BE9: "HANJA",
        0x8BEA: "HANJA",
        0x8BEB: "HANJA",
        0x8BEC: "HANJA",
        0x8BED: "HANJA",
        0x8BEE: "HANJA",
        0x8BEF: "HANJA",
        0x8BF0: "HANJA",
        0x8BF1: "HANJA",
        0x8BF2: "HANJA",
        0x8BF3: "HANJA",
        0x8BF4: "HANJA",
        0x8BF5: "HANJA",
        0x8BF6: "HANJA",
        0x8BF7: "HANJA",
        0x8BF8: "HANJA",
        0x8BF9: "HANJA",
        0x8BFA: "HANJA",
        0x8BFB: "HANJA",
        0x8BFC: "HANJA",
        0x8BFD: "HANJA",
        0x8BFE: "HANJA",
        0x8BFF: "HANJA",
        0x8C00: "HANJA",
        0x8C01: "HANJA",
        0x8C02: "HANJA",
        0x8C03: "HANJA",
        0x8C04: "HANJA",
        0x8C05: "HANJA",
        0x8C06: "HANJA",
        0x8C07: "HANJA",
        0x8C08: "HANJA",
        0x8C09: "HANJA",
        0x8C0A: "HANJA",
        0x8C0B: "HANJA",
        0x8C0C: "HANJA",
        0x8C0D: "HANJA",
        0x8C0E: "HANJA",
        0x8C0F: "HANJA",
        0x8C10: "HANJA",
        0x8C11: "HANJA",
        0x8C12: "HANJA",
        0x8C13: "HANJA",
        0x8C14: "HANJA",
        0x8C15: "HANJA",
        0x8C16: "HANJA",
        0x8C17: "HANJA",
        0x8C18: "HANJA",
        0x8C19: "HANJA",
        0x8C1A: "HANJA",
        0x8C1B: "HANJA",
        0x8C1C: "HANJA",
        0x8C1D: "HANJA",
        0x8C1E: "HANJA",
        0x8C1F: "HANJA",
        0x8C20: "HANJA",
        0x8C21: "HANJA",
        0x8C22: "HANJA",
        0x8C23: "HANJA",
        0x8C24: "HANJA",
        0x8C25: "HANJA",
        0x8C26: "HANJA",
        0x8C27: "HANJA",
        0x8C28: "HANJA",
        0x8C29: "HANJA",
        0x8C2A: "HANJA",
        0x8C2B: "HANJA",
        0x8C2C: "HANJA",
        0x8C2D: "HANJA",
        0x8C2E: "HANJA",
        0x8C2F: "HANJA",
        0x8C30: "HANJA",
        0x8C31: "HANJA",
        0x8C32: "HANJA",
        0x8C33: "HANJA",
        0x8C34: "HANJA",
        0x8C35: "HANJA",
        0x8C36: "HANJA",
        0x8C37: "HANJA",
        0x8C38: "HANJA",
        0x8C39: "HANJA",
        0x8C3A: "HANJA",
        0x8C3B: "HANJA",
        0x8C3C: "HANJA",
        0x8C3D: "HANJA",
        0x8C3E: "HANJA",
        0x8C3F: "HANJA",
        0x8C40: "HANJA",
        0x8C41: "HANJA",
        0x8C42: "HANJA",
        0x8C43: "HANJA",
        0x8C44: "HANJA",
        0x8C45: "HANJA",
        0x8C46: "HANJA",
        0x8C47: "HANJA",
        0x8C48: "HANJA",
        0x8C49: "HANJA",
        0x8C4A: "HANJA",
        0x8C4B: "HANJA",
        0x8C4C: "HANJA",
        0x8C4D: "HANJA",
        0x8C4E: "HANJA",
        0x8C4F: "HANJA",
        0x8C50: "HANJA",
        0x8C51: "HANJA",
        0x8C52: "HANJA",
        0x8C53: "HANJA",
        0x8C54: "HANJA",
        0x8C55: "HANJA",
        0x8C56: "HANJA",
        0x8C57: "HANJA",
        0x8C58: "HANJA",
        0x8C59: "HANJA",
        0x8C5A: "HANJA",
        0x8C5B: "HANJA",
        0x8C5C: "HANJA",
        0x8C5D: "HANJA",
        0x8C5E: "HANJA",
        0x8C5F: "HANJA",
        0x8C60: "HANJA",
        0x8C61: "HANJA",
        0x8C62: "HANJA",
        0x8C63: "HANJA",
        0x8C64: "HANJA",
        0x8C65: "HANJA",
        0x8C66: "HANJA",
        0x8C67: "HANJA",
        0x8C68: "HANJA",
        0x8C69: "HANJA",
        0x8C6A: "HANJA",
        0x8C6B: "HANJA",
        0x8C6C: "HANJA",
        0x8C6D: "HANJA",
        0x8C6E: "HANJA",
        0x8C6F: "HANJA",
        0x8C70: "HANJA",
        0x8C71: "HANJA",
        0x8C72: "HANJA",
        0x8C73: "HANJA",
        0x8C74: "HANJA",
        0x8C75: "HANJA",
        0x8C76: "HANJA",
        0x8C77: "HANJA",
        0x8C78: "HANJA",
        0x8C79: "HANJA",
        0x8C7A: "HANJA",
        0x8C7B: "HANJA",
        0x8C7C: "HANJA",
        0x8C7D: "HANJA",
        0x8C7E: "HANJA",
        0x8C7F: "HANJA",
        0x8C80: "HANJA",
        0x8C81: "HANJA",
        0x8C82: "HANJA",
        0x8C83: "HANJA",
        0x8C84: "HANJA",
        0x8C85: "HANJA",
        0x8C86: "HANJA",
        0x8C87: "HANJA",
        0x8C88: "HANJA",
        0x8C89: "HANJA",
        0x8C8A: "HANJA",
        0x8C8B: "HANJA",
        0x8C8C: "HANJA",
        0x8C8D: "HANJA",
        0x8C8E: "HANJA",
        0x8C8F: "HANJA",
        0x8C90: "HANJA",
        0x8C91: "HANJA",
        0x8C92: "HANJA",
        0x8C93: "HANJA",
        0x8C94: "HANJA",
        0x8C95: "HANJA",
        0x8C96: "HANJA",
        0x8C97: "HANJA",
        0x8C98: "HANJA",
        0x8C99: "HANJA",
        0x8C9A: "HANJA",
        0x8C9B: "HANJA",
        0x8C9C: "HANJA",
        0x8C9D: "HANJA",
        0x8C9E: "HANJA",
        0x8C9F: "HANJA",
        0x8CA0: "HANJA",
        0x8CA1: "HANJA",
        0x8CA2: "HANJA",
        0x8CA3: "HANJA",
        0x8CA4: "HANJA",
        0x8CA5: "HANJA",
        0x8CA6: "HANJA",
        0x8CA7: "HANJA",
        0x8CA8: "HANJA",
        0x8CA9: "HANJA",
        0x8CAA: "HANJA",
        0x8CAB: "HANJA",
        0x8CAC: "HANJA",
        0x8CAD: "HANJA",
        0x8CAE: "HANJA",
        0x8CAF: "HANJA",
        0x8CB0: "HANJA",
        0x8CB1: "HANJA",
        0x8CB2: "HANJA",
        0x8CB3: "HANJA",
        0x8CB4: "HANJA",
        0x8CB5: "HANJA",
        0x8CB6: "HANJA",
        0x8CB7: "HANJA",
        0x8CB8: "HANJA",
        0x8CB9: "HANJA",
        0x8CBA: "HANJA",
        0x8CBB: "HANJA",
        0x8CBC: "HANJA",
        0x8CBD: "HANJA",
        0x8CBE: "HANJA",
        0x8CBF: "HANJA",
        0x8CC0: "HANJA",
        0x8CC1: "HANJA",
        0x8CC2: "HANJA",
        0x8CC3: "HANJA",
        0x8CC4: "HANJA",
        0x8CC5: "HANJA",
        0x8CC6: "HANJA",
        0x8CC7: "HANJA",
        0x8CC8: "HANJA",
        0x8CC9: "HANJA",
        0x8CCA: "HANJA",
        0x8CCB: "HANJA",
        0x8CCC: "HANJA",
        0x8CCD: "HANJA",
        0x8CCE: "HANJA",
        0x8CCF: "HANJA",
        0x8CD0: "HANJA",
        0x8CD1: "HANJA",
        0x8CD2: "HANJA",
        0x8CD3: "HANJA",
        0x8CD4: "HANJA",
        0x8CD5: "HANJA",
        0x8CD6: "HANJA",
        0x8CD7: "HANJA",
        0x8CD8: "HANJA",
        0x8CD9: "HANJA",
        0x8CDA: "HANJA",
        0x8CDB: "HANJA",
        0x8CDC: "HANJA",
        0x8CDD: "HANJA",
        0x8CDE: "HANJA",
        0x8CDF: "HANJA",
        0x8CE0: "HANJA",
        0x8CE1: "HANJA",
        0x8CE2: "HANJA",
        0x8CE3: "HANJA",
        0x8CE4: "HANJA",
        0x8CE5: "HANJA",
        0x8CE6: "HANJA",
        0x8CE7: "HANJA",
        0x8CE8: "HANJA",
        0x8CE9: "HANJA",
        0x8CEA: "HANJA",
        0x8CEB: "HANJA",
        0x8CEC: "HANJA",
        0x8CED: "HANJA",
        0x8CEE: "HANJA",
        0x8CEF: "HANJA",
        0x8CF0: "HANJA",
        0x8CF1: "HANJA",
        0x8CF2: "HANJA",
        0x8CF3: "HANJA",
        0x8CF4: "HANJA",
        0x8CF5: "HANJA",
        0x8CF6: "HANJA",
        0x8CF7: "HANJA",
        0x8CF8: "HANJA",
        0x8CF9: "HANJA",
        0x8CFA: "HANJA",
        0x8CFB: "HANJA",
        0x8CFC: "HANJA",
        0x8CFD: "HANJA",
        0x8CFE: "HANJA",
        0x8CFF: "HANJA",
        0x8D00: "HANJA",
        0x8D01: "HANJA",
        0x8D02: "HANJA",
        0x8D03: "HANJA",
        0x8D04: "HANJA",
        0x8D05: "HANJA",
        0x8D06: "HANJA",
        0x8D07: "HANJA",
        0x8D08: "HANJA",
        0x8D09: "HANJA",
        0x8D0A: "HANJA",
        0x8D0B: "HANJA",
        0x8D0C: "HANJA",
        0x8D0D: "HANJA",
        0x8D0E: "HANJA",
        0x8D0F: "HANJA",
        0x8D10: "HANJA",
        0x8D11: "HANJA",
        0x8D12: "HANJA",
        0x8D13: "HANJA",
        0x8D14: "HANJA",
        0x8D15: "HANJA",
        0x8D16: "HANJA",
        0x8D17: "HANJA",
        0x8D18: "HANJA",
        0x8D19: "HANJA",
        0x8D1A: "HANJA",
        0x8D1B: "HANJA",
        0x8D1C: "HANJA",
        0x8D1D: "HANJA",
        0x8D1E: "HANJA",
        0x8D1F: "HANJA",
        0x8D20: "HANJA",
        0x8D21: "HANJA",
        0x8D22: "HANJA",
        0x8D23: "HANJA",
        0x8D24: "HANJA",
        0x8D25: "HANJA",
        0x8D26: "HANJA",
        0x8D27: "HANJA",
        0x8D28: "HANJA",
        0x8D29: "HANJA",
        0x8D2A: "HANJA",
        0x8D2B: "HANJA",
        0x8D2C: "HANJA",
        0x8D2D: "HANJA",
        0x8D2E: "HANJA",
        0x8D2F: "HANJA",
        0x8D30: "HANJA",
        0x8D31: "HANJA",
        0x8D32: "HANJA",
        0x8D33: "HANJA",
        0x8D34: "HANJA",
        0x8D35: "HANJA",
        0x8D36: "HANJA",
        0x8D37: "HANJA",
        0x8D38: "HANJA",
        0x8D39: "HANJA",
        0x8D3A: "HANJA",
        0x8D3B: "HANJA",
        0x8D3C: "HANJA",
        0x8D3D: "HANJA",
        0x8D3E: "HANJA",
        0x8D3F: "HANJA",
        0x8D40: "HANJA",
        0x8D41: "HANJA",
        0x8D42: "HANJA",
        0x8D43: "HANJA",
        0x8D44: "HANJA",
        0x8D45: "HANJA",
        0x8D46: "HANJA",
        0x8D47: "HANJA",
        0x8D48: "HANJA",
        0x8D49: "HANJA",
        0x8D4A: "HANJA",
        0x8D4B: "HANJA",
        0x8D4C: "HANJA",
        0x8D4D: "HANJA",
        0x8D4E: "HANJA",
        0x8D4F: "HANJA",
        0x8D50: "HANJA",
        0x8D51: "HANJA",
        0x8D52: "HANJA",
        0x8D53: "HANJA",
        0x8D54: "HANJA",
        0x8D55: "HANJA",
        0x8D56: "HANJA",
        0x8D57: "HANJA",
        0x8D58: "HANJA",
        0x8D59: "HANJA",
        0x8D5A: "HANJA",
        0x8D5B: "HANJA",
        0x8D5C: "HANJA",
        0x8D5D: "HANJA",
        0x8D5E: "HANJA",
        0x8D5F: "HANJA",
        0x8D60: "HANJA",
        0x8D61: "HANJA",
        0x8D62: "HANJA",
        0x8D63: "HANJA",
        0x8D64: "HANJA",
        0x8D65: "HANJA",
        0x8D66: "HANJA",
        0x8D67: "HANJA",
        0x8D68: "HANJA",
        0x8D69: "HANJA",
        0x8D6A: "HANJA",
        0x8D6B: "HANJA",
        0x8D6C: "HANJA",
        0x8D6D: "HANJA",
        0x8D6E: "HANJA",
        0x8D6F: "HANJA",
        0x8D70: "HANJA",
        0x8D71: "HANJA",
        0x8D72: "HANJA",
        0x8D73: "HANJA",
        0x8D74: "HANJA",
        0x8D75: "HANJA",
        0x8D76: "HANJA",
        0x8D77: "HANJA",
        0x8D78: "HANJA",
        0x8D79: "HANJA",
        0x8D7A: "HANJA",
        0x8D7B: "HANJA",
        0x8D7C: "HANJA",
        0x8D7D: "HANJA",
        0x8D7E: "HANJA",
        0x8D7F: "HANJA",
        0x8D80: "HANJA",
        0x8D81: "HANJA",
        0x8D82: "HANJA",
        0x8D83: "HANJA",
        0x8D84: "HANJA",
        0x8D85: "HANJA",
        0x8D86: "HANJA",
        0x8D87: "HANJA",
        0x8D88: "HANJA",
        0x8D89: "HANJA",
        0x8D8A: "HANJA",
        0x8D8B: "HANJA",
        0x8D8C: "HANJA",
        0x8D8D: "HANJA",
        0x8D8E: "HANJA",
        0x8D8F: "HANJA",
        0x8D90: "HANJA",
        0x8D91: "HANJA",
        0x8D92: "HANJA",
        0x8D93: "HANJA",
        0x8D94: "HANJA",
        0x8D95: "HANJA",
        0x8D96: "HANJA",
        0x8D97: "HANJA",
        0x8D98: "HANJA",
        0x8D99: "HANJA",
        0x8D9A: "HANJA",
        0x8D9B: "HANJA",
        0x8D9C: "HANJA",
        0x8D9D: "HANJA",
        0x8D9E: "HANJA",
        0x8D9F: "HANJA",
        0x8DA0: "HANJA",
        0x8DA1: "HANJA",
        0x8DA2: "HANJA",
        0x8DA3: "HANJA",
        0x8DA4: "HANJA",
        0x8DA5: "HANJA",
        0x8DA6: "HANJA",
        0x8DA7: "HANJA",
        0x8DA8: "HANJA",
        0x8DA9: "HANJA",
        0x8DAA: "HANJA",
        0x8DAB: "HANJA",
        0x8DAC: "HANJA",
        0x8DAD: "HANJA",
        0x8DAE: "HANJA",
        0x8DAF: "HANJA",
        0x8DB0: "HANJA",
        0x8DB1: "HANJA",
        0x8DB2: "HANJA",
        0x8DB3: "HANJA",
        0x8DB4: "HANJA",
        0x8DB5: "HANJA",
        0x8DB6: "HANJA",
        0x8DB7: "HANJA",
        0x8DB8: "HANJA",
        0x8DB9: "HANJA",
        0x8DBA: "HANJA",
        0x8DBB: "HANJA",
        0x8DBC: "HANJA",
        0x8DBD: "HANJA",
        0x8DBE: "HANJA",
        0x8DBF: "HANJA",
        0x8DC0: "HANJA",
        0x8DC1: "HANJA",
        0x8DC2: "HANJA",
        0x8DC3: "HANJA",
        0x8DC4: "HANJA",
        0x8DC5: "HANJA",
        0x8DC6: "HANJA",
        0x8DC7: "HANJA",
        0x8DC8: "HANJA",
        0x8DC9: "HANJA",
        0x8DCA: "HANJA",
        0x8DCB: "HANJA",
        0x8DCC: "HANJA",
        0x8DCD: "HANJA",
        0x8DCE: "HANJA",
        0x8DCF: "HANJA",
        0x8DD0: "HANJA",
        0x8DD1: "HANJA",
        0x8DD2: "HANJA",
        0x8DD3: "HANJA",
        0x8DD4: "HANJA",
        0x8DD5: "HANJA",
        0x8DD6: "HANJA",
        0x8DD7: "HANJA",
        0x8DD8: "HANJA",
        0x8DD9: "HANJA",
        0x8DDA: "HANJA",
        0x8DDB: "HANJA",
        0x8DDC: "HANJA",
        0x8DDD: "HANJA",
        0x8DDE: "HANJA",
        0x8DDF: "HANJA",
        0x8DE0: "HANJA",
        0x8DE1: "HANJA",
        0x8DE2: "HANJA",
        0x8DE3: "HANJA",
        0x8DE4: "HANJA",
        0x8DE5: "HANJA",
        0x8DE6: "HANJA",
        0x8DE7: "HANJA",
        0x8DE8: "HANJA",
        0x8DE9: "HANJA",
        0x8DEA: "HANJA",
        0x8DEB: "HANJA",
        0x8DEC: "HANJA",
        0x8DED: "HANJA",
        0x8DEE: "HANJA",
        0x8DEF: "HANJA",
        0x8DF0: "HANJA",
        0x8DF1: "HANJA",
        0x8DF2: "HANJA",
        0x8DF3: "HANJA",
        0x8DF4: "HANJA",
        0x8DF5: "HANJA",
        0x8DF6: "HANJA",
        0x8DF7: "HANJA",
        0x8DF8: "HANJA",
        0x8DF9: "HANJA",
        0x8DFA: "HANJA",
        0x8DFB: "HANJA",
        0x8DFC: "HANJA",
        0x8DFD: "HANJA",
        0x8DFE: "HANJA",
        0x8DFF: "HANJA",
        0x8E00: "HANJA",
        0x8E01: "HANJA",
        0x8E02: "HANJA",
        0x8E03: "HANJA",
        0x8E04: "HANJA",
        0x8E05: "HANJA",
        0x8E06: "HANJA",
        0x8E07: "HANJA",
        0x8E08: "HANJA",
        0x8E09: "HANJA",
        0x8E0A: "HANJA",
        0x8E0B: "HANJA",
        0x8E0C: "HANJA",
        0x8E0D: "HANJA",
        0x8E0E: "HANJA",
        0x8E0F: "HANJA",
        0x8E10: "HANJA",
        0x8E11: "HANJA",
        0x8E12: "HANJA",
        0x8E13: "HANJA",
        0x8E14: "HANJA",
        0x8E15: "HANJA",
        0x8E16: "HANJA",
        0x8E17: "HANJA",
        0x8E18: "HANJA",
        0x8E19: "HANJA",
        0x8E1A: "HANJA",
        0x8E1B: "HANJA",
        0x8E1C: "HANJA",
        0x8E1D: "HANJA",
        0x8E1E: "HANJA",
        0x8E1F: "HANJA",
        0x8E20: "HANJA",
        0x8E21: "HANJA",
        0x8E22: "HANJA",
        0x8E23: "HANJA",
        0x8E24: "HANJA",
        0x8E25: "HANJA",
        0x8E26: "HANJA",
        0x8E27: "HANJA",
        0x8E28: "HANJA",
        0x8E29: "HANJA",
        0x8E2A: "HANJA",
        0x8E2B: "HANJA",
        0x8E2C: "HANJA",
        0x8E2D: "HANJA",
        0x8E2E: "HANJA",
        0x8E2F: "HANJA",
        0x8E30: "HANJA",
        0x8E31: "HANJA",
        0x8E32: "HANJA",
        0x8E33: "HANJA",
        0x8E34: "HANJA",
        0x8E35: "HANJA",
        0x8E36: "HANJA",
        0x8E37: "HANJA",
        0x8E38: "HANJA",
        0x8E39: "HANJA",
        0x8E3A: "HANJA",
        0x8E3B: "HANJA",
        0x8E3C: "HANJA",
        0x8E3D: "HANJA",
        0x8E3E: "HANJA",
        0x8E3F: "HANJA",
        0x8E40: "HANJA",
        0x8E41: "HANJA",
        0x8E42: "HANJA",
        0x8E43: "HANJA",
        0x8E44: "HANJA",
        0x8E45: "HANJA",
        0x8E46: "HANJA",
        0x8E47: "HANJA",
        0x8E48: "HANJA",
        0x8E49: "HANJA",
        0x8E4A: "HANJA",
        0x8E4B: "HANJA",
        0x8E4C: "HANJA",
        0x8E4D: "HANJA",
        0x8E4E: "HANJA",
        0x8E4F: "HANJA",
        0x8E50: "HANJA",
        0x8E51: "HANJA",
        0x8E52: "HANJA",
        0x8E53: "HANJA",
        0x8E54: "HANJA",
        0x8E55: "HANJA",
        0x8E56: "HANJA",
        0x8E57: "HANJA",
        0x8E58: "HANJA",
        0x8E59: "HANJA",
        0x8E5A: "HANJA",
        0x8E5B: "HANJA",
        0x8E5C: "HANJA",
        0x8E5D: "HANJA",
        0x8E5E: "HANJA",
        0x8E5F: "HANJA",
        0x8E60: "HANJA",
        0x8E61: "HANJA",
        0x8E62: "HANJA",
        0x8E63: "HANJA",
        0x8E64: "HANJA",
        0x8E65: "HANJA",
        0x8E66: "HANJA",
        0x8E67: "HANJA",
        0x8E68: "HANJA",
        0x8E69: "HANJA",
        0x8E6A: "HANJA",
        0x8E6B: "HANJA",
        0x8E6C: "HANJA",
        0x8E6D: "HANJA",
        0x8E6E: "HANJA",
        0x8E6F: "HANJA",
        0x8E70: "HANJA",
        0x8E71: "HANJA",
        0x8E72: "HANJA",
        0x8E73: "HANJA",
        0x8E74: "HANJA",
        0x8E75: "HANJA",
        0x8E76: "HANJA",
        0x8E77: "HANJA",
        0x8E78: "HANJA",
        0x8E79: "HANJA",
        0x8E7A: "HANJA",
        0x8E7B: "HANJA",
        0x8E7C: "HANJA",
        0x8E7D: "HANJA",
        0x8E7E: "HANJA",
        0x8E7F: "HANJA",
        0x8E80: "HANJA",
        0x8E81: "HANJA",
        0x8E82: "HANJA",
        0x8E83: "HANJA",
        0x8E84: "HANJA",
        0x8E85: "HANJA",
        0x8E86: "HANJA",
        0x8E87: "HANJA",
        0x8E88: "HANJA",
        0x8E89: "HANJA",
        0x8E8A: "HANJA",
        0x8E8B: "HANJA",
        0x8E8C: "HANJA",
        0x8E8D: "HANJA",
        0x8E8E: "HANJA",
        0x8E8F: "HANJA",
        0x8E90: "HANJA",
        0x8E91: "HANJA",
        0x8E92: "HANJA",
        0x8E93: "HANJA",
        0x8E94: "HANJA",
        0x8E95: "HANJA",
        0x8E96: "HANJA",
        0x8E97: "HANJA",
        0x8E98: "HANJA",
        0x8E99: "HANJA",
        0x8E9A: "HANJA",
        0x8E9B: "HANJA",
        0x8E9C: "HANJA",
        0x8E9D: "HANJA",
        0x8E9E: "HANJA",
        0x8E9F: "HANJA",
        0x8EA0: "HANJA",
        0x8EA1: "HANJA",
        0x8EA2: "HANJA",
        0x8EA3: "HANJA",
        0x8EA4: "HANJA",
        0x8EA5: "HANJA",
        0x8EA6: "HANJA",
        0x8EA7: "HANJA",
        0x8EA8: "HANJA",
        0x8EA9: "HANJA",
        0x8EAA: "HANJA",
        0x8EAB: "HANJA",
        0x8EAC: "HANJA",
        0x8EAD: "HANJA",
        0x8EAE: "HANJA",
        0x8EAF: "HANJA",
        0x8EB0: "HANJA",
        0x8EB1: "HANJA",
        0x8EB2: "HANJA",
        0x8EB3: "HANJA",
        0x8EB4: "HANJA",
        0x8EB5: "HANJA",
        0x8EB6: "HANJA",
        0x8EB7: "HANJA",
        0x8EB8: "HANJA",
        0x8EB9: "HANJA",
        0x8EBA: "HANJA",
        0x8EBB: "HANJA",
        0x8EBC: "HANJA",
        0x8EBD: "HANJA",
        0x8EBE: "HANJA",
        0x8EBF: "HANJA",
        0x8EC0: "HANJA",
        0x8EC1: "HANJA",
        0x8EC2: "HANJA",
        0x8EC3: "HANJA",
        0x8EC4: "HANJA",
        0x8EC5: "HANJA",
        0x8EC6: "HANJA",
        0x8EC7: "HANJA",
        0x8EC8: "HANJA",
        0x8EC9: "HANJA",
        0x8ECA: "HANJA",
        0x8ECB: "HANJA",
        0x8ECC: "HANJA",
        0x8ECD: "HANJA",
        0x8ECE: "HANJA",
        0x8ECF: "HANJA",
        0x8ED0: "HANJA",
        0x8ED1: "HANJA",
        0x8ED2: "HANJA",
        0x8ED3: "HANJA",
        0x8ED4: "HANJA",
        0x8ED5: "HANJA",
        0x8ED6: "HANJA",
        0x8ED7: "HANJA",
        0x8ED8: "HANJA",
        0x8ED9: "HANJA",
        0x8EDA: "HANJA",
        0x8EDB: "HANJA",
        0x8EDC: "HANJA",
        0x8EDD: "HANJA",
        0x8EDE: "HANJA",
        0x8EDF: "HANJA",
        0x8EE0: "HANJA",
        0x8EE1: "HANJA",
        0x8EE2: "HANJA",
        0x8EE3: "HANJA",
        0x8EE4: "HANJA",
        0x8EE5: "HANJA",
        0x8EE6: "HANJA",
        0x8EE7: "HANJA",
        0x8EE8: "HANJA",
        0x8EE9: "HANJA",
        0x8EEA: "HANJA",
        0x8EEB: "HANJA",
        0x8EEC: "HANJA",
        0x8EED: "HANJA",
        0x8EEE: "HANJA",
        0x8EEF: "HANJA",
        0x8EF0: "HANJA",
        0x8EF1: "HANJA",
        0x8EF2: "HANJA",
        0x8EF3: "HANJA",
        0x8EF4: "HANJA",
        0x8EF5: "HANJA",
        0x8EF6: "HANJA",
        0x8EF7: "HANJA",
        0x8EF8: "HANJA",
        0x8EF9: "HANJA",
        0x8EFA: "HANJA",
        0x8EFB: "HANJA",
        0x8EFC: "HANJA",
        0x8EFD: "HANJA",
        0x8EFE: "HANJA",
        0x8EFF: "HANJA",
        0x8F00: "HANJA",
        0x8F01: "HANJA",
        0x8F02: "HANJA",
        0x8F03: "HANJA",
        0x8F04: "HANJA",
        0x8F05: "HANJA",
        0x8F06: "HANJA",
        0x8F07: "HANJA",
        0x8F08: "HANJA",
        0x8F09: "HANJA",
        0x8F0A: "HANJA",
        0x8F0B: "HANJA",
        0x8F0C: "HANJA",
        0x8F0D: "HANJA",
        0x8F0E: "HANJA",
        0x8F0F: "HANJA",
        0x8F10: "HANJA",
        0x8F11: "HANJA",
        0x8F12: "HANJA",
        0x8F13: "HANJA",
        0x8F14: "HANJA",
        0x8F15: "HANJA",
        0x8F16: "HANJA",
        0x8F17: "HANJA",
        0x8F18: "HANJA",
        0x8F19: "HANJA",
        0x8F1A: "HANJA",
        0x8F1B: "HANJA",
        0x8F1C: "HANJA",
        0x8F1D: "HANJA",
        0x8F1E: "HANJA",
        0x8F1F: "HANJA",
        0x8F20: "HANJA",
        0x8F21: "HANJA",
        0x8F22: "HANJA",
        0x8F23: "HANJA",
        0x8F24: "HANJA",
        0x8F25: "HANJA",
        0x8F26: "HANJA",
        0x8F27: "HANJA",
        0x8F28: "HANJA",
        0x8F29: "HANJA",
        0x8F2A: "HANJA",
        0x8F2B: "HANJA",
        0x8F2C: "HANJA",
        0x8F2D: "HANJA",
        0x8F2E: "HANJA",
        0x8F2F: "HANJA",
        0x8F30: "HANJA",
        0x8F31: "HANJA",
        0x8F32: "HANJA",
        0x8F33: "HANJA",
        0x8F34: "HANJA",
        0x8F35: "HANJA",
        0x8F36: "HANJA",
        0x8F37: "HANJA",
        0x8F38: "HANJA",
        0x8F39: "HANJA",
        0x8F3A: "HANJA",
        0x8F3B: "HANJA",
        0x8F3C: "HANJA",
        0x8F3D: "HANJA",
        0x8F3E: "HANJA",
        0x8F3F: "HANJA",
        0x8F40: "HANJA",
        0x8F41: "HANJA",
        0x8F42: "HANJA",
        0x8F43: "HANJA",
        0x8F44: "HANJA",
        0x8F45: "HANJA",
        0x8F46: "HANJA",
        0x8F47: "HANJA",
        0x8F48: "HANJA",
        0x8F49: "HANJA",
        0x8F4A: "HANJA",
        0x8F4B: "HANJA",
        0x8F4C: "HANJA",
        0x8F4D: "HANJA",
        0x8F4E: "HANJA",
        0x8F4F: "HANJA",
        0x8F50: "HANJA",
        0x8F51: "HANJA",
        0x8F52: "HANJA",
        0x8F53: "HANJA",
        0x8F54: "HANJA",
        0x8F55: "HANJA",
        0x8F56: "HANJA",
        0x8F57: "HANJA",
        0x8F58: "HANJA",
        0x8F59: "HANJA",
        0x8F5A: "HANJA",
        0x8F5B: "HANJA",
        0x8F5C: "HANJA",
        0x8F5D: "HANJA",
        0x8F5E: "HANJA",
        0x8F5F: "HANJA",
        0x8F60: "HANJA",
        0x8F61: "HANJA",
        0x8F62: "HANJA",
        0x8F63: "HANJA",
        0x8F64: "HANJA",
        0x8F65: "HANJA",
        0x8F66: "HANJA",
        0x8F67: "HANJA",
        0x8F68: "HANJA",
        0x8F69: "HANJA",
        0x8F6A: "HANJA",
        0x8F6B: "HANJA",
        0x8F6C: "HANJA",
        0x8F6D: "HANJA",
        0x8F6E: "HANJA",
        0x8F6F: "HANJA",
        0x8F70: "HANJA",
        0x8F71: "HANJA",
        0x8F72: "HANJA",
        0x8F73: "HANJA",
        0x8F74: "HANJA",
        0x8F75: "HANJA",
        0x8F76: "HANJA",
        0x8F77: "HANJA",
        0x8F78: "HANJA",
        0x8F79: "HANJA",
        0x8F7A: "HANJA",
        0x8F7B: "HANJA",
        0x8F7C: "HANJA",
        0x8F7D: "HANJA",
        0x8F7E: "HANJA",
        0x8F7F: "HANJA",
        0x8F80: "HANJA",
        0x8F81: "HANJA",
        0x8F82: "HANJA",
        0x8F83: "HANJA",
        0x8F84: "HANJA",
        0x8F85: "HANJA",
        0x8F86: "HANJA",
        0x8F87: "HANJA",
        0x8F88: "HANJA",
        0x8F89: "HANJA",
        0x8F8A: "HANJA",
        0x8F8B: "HANJA",
        0x8F8C: "HANJA",
        0x8F8D: "HANJA",
        0x8F8E: "HANJA",
        0x8F8F: "HANJA",
        0x8F90: "HANJA",
        0x8F91: "HANJA",
        0x8F92: "HANJA",
        0x8F93: "HANJA",
        0x8F94: "HANJA",
        0x8F95: "HANJA",
        0x8F96: "HANJA",
        0x8F97: "HANJA",
        0x8F98: "HANJA",
        0x8F99: "HANJA",
        0x8F9A: "HANJA",
        0x8F9B: "HANJA",
        0x8F9C: "HANJA",
        0x8F9D: "HANJA",
        0x8F9E: "HANJA",
        0x8F9F: "HANJA",
        0x8FA0: "HANJA",
        0x8FA1: "HANJA",
        0x8FA2: "HANJA",
        0x8FA3: "HANJA",
        0x8FA4: "HANJA",
        0x8FA5: "HANJA",
        0x8FA6: "HANJA",
        0x8FA7: "HANJA",
        0x8FA8: "HANJA",
        0x8FA9: "HANJA",
        0x8FAA: "HANJA",
        0x8FAB: "HANJA",
        0x8FAC: "HANJA",
        0x8FAD: "HANJA",
        0x8FAE: "HANJA",
        0x8FAF: "HANJA",
        0x8FB0: "HANJA",
        0x8FB1: "HANJA",
        0x8FB2: "HANJA",
        0x8FB3: "HANJA",
        0x8FB4: "HANJA",
        0x8FB5: "HANJA",
        0x8FB6: "HANJA",
        0x8FB7: "HANJA",
        0x8FB8: "HANJA",
        0x8FB9: "HANJA",
        0x8FBA: "HANJA",
        0x8FBB: "HANJA",
        0x8FBC: "HANJA",
        0x8FBD: "HANJA",
        0x8FBE: "HANJA",
        0x8FBF: "HANJA",
        0x8FC0: "HANJA",
        0x8FC1: "HANJA",
        0x8FC2: "HANJA",
        0x8FC3: "HANJA",
        0x8FC4: "HANJA",
        0x8FC5: "HANJA",
        0x8FC6: "HANJA",
        0x8FC7: "HANJA",
        0x8FC8: "HANJA",
        0x8FC9: "HANJA",
        0x8FCA: "HANJA",
        0x8FCB: "HANJA",
        0x8FCC: "HANJA",
        0x8FCD: "HANJA",
        0x8FCE: "HANJA",
        0x8FCF: "HANJA",
        0x8FD0: "HANJA",
        0x8FD1: "HANJA",
        0x8FD2: "HANJA",
        0x8FD3: "HANJA",
        0x8FD4: "HANJA",
        0x8FD5: "HANJA",
        0x8FD6: "HANJA",
        0x8FD7: "HANJA",
        0x8FD8: "HANJA",
        0x8FD9: "HANJA",
        0x8FDA: "HANJA",
        0x8FDB: "HANJA",
        0x8FDC: "HANJA",
        0x8FDD: "HANJA",
        0x8FDE: "HANJA",
        0x8FDF: "HANJA",
        0x8FE0: "HANJA",
        0x8FE1: "HANJA",
        0x8FE2: "HANJA",
        0x8FE3: "HANJA",
        0x8FE4: "HANJA",
        0x8FE5: "HANJA",
        0x8FE6: "HANJA",
        0x8FE7: "HANJA",
        0x8FE8: "HANJA",
        0x8FE9: "HANJA",
        0x8FEA: "HANJA",
        0x8FEB: "HANJA",
        0x8FEC: "HANJA",
        0x8FED: "HANJA",
        0x8FEE: "HANJA",
        0x8FEF: "HANJA",
        0x8FF0: "HANJA",
        0x8FF1: "HANJA",
        0x8FF2: "HANJA",
        0x8FF3: "HANJA",
        0x8FF4: "HANJA",
        0x8FF5: "HANJA",
        0x8FF6: "HANJA",
        0x8FF7: "HANJA",
        0x8FF8: "HANJA",
        0x8FF9: "HANJA",
        0x8FFA: "HANJA",
        0x8FFB: "HANJA",
        0x8FFC: "HANJA",
        0x8FFD: "HANJA",
        0x8FFE: "HANJA",
        0x8FFF: "HANJA",
        0x9000: "HANJA",
        0x9001: "HANJA",
        0x9002: "HANJA",
        0x9003: "HANJA",
        0x9004: "HANJA",
        0x9005: "HANJA",
        0x9006: "HANJA",
        0x9007: "HANJA",
        0x9008: "HANJA",
        0x9009: "HANJA",
        0x900A: "HANJA",
        0x900B: "HANJA",
        0x900C: "HANJA",
        0x900D: "HANJA",
        0x900E: "HANJA",
        0x900F: "HANJA",
        0x9010: "HANJA",
        0x9011: "HANJA",
        0x9012: "HANJA",
        0x9013: "HANJA",
        0x9014: "HANJA",
        0x9015: "HANJA",
        0x9016: "HANJA",
        0x9017: "HANJA",
        0x9018: "HANJA",
        0x9019: "HANJA",
        0x901A: "HANJA",
        0x901B: "HANJA",
        0x901C: "HANJA",
        0x901D: "HANJA",
        0x901E: "HANJA",
        0x901F: "HANJA",
        0x9020: "HANJA",
        0x9021: "HANJA",
        0x9022: "HANJA",
        0x9023: "HANJA",
        0x9024: "HANJA",
        0x9025: "HANJA",
        0x9026: "HANJA",
        0x9027: "HANJA",
        0x9028: "HANJA",
        0x9029: "HANJA",
        0x902A: "HANJA",
        0x902B: "HANJA",
        0x902C: "HANJA",
        0x902D: "HANJA",
        0x902E: "HANJA",
        0x902F: "HANJA",
        0x9030: "HANJA",
        0x9031: "HANJA",
        0x9032: "HANJA",
        0x9033: "HANJA",
        0x9034: "HANJA",
        0x9035: "HANJA",
        0x9036: "HANJA",
        0x9037: "HANJA",
        0x9038: "HANJA",
        0x9039: "HANJA",
        0x903A: "HANJA",
        0x903B: "HANJA",
        0x903C: "HANJA",
        0x903D: "HANJA",
        0x903E: "HANJA",
        0x903F: "HANJA",
        0x9040: "HANJA",
        0x9041: "HANJA",
        0x9042: "HANJA",
        0x9043: "HANJA",
        0x9044: "HANJA",
        0x9045: "HANJA",
        0x9046: "HANJA",
        0x9047: "HANJA",
        0x9048: "HANJA",
        0x9049: "HANJA",
        0x904A: "HANJA",
        0x904B: "HANJA",
        0x904C: "HANJA",
        0x904D: "HANJA",
        0x904E: "HANJA",
        0x904F: "HANJA",
        0x9050: "HANJA",
        0x9051: "HANJA",
        0x9052: "HANJA",
        0x9053: "HANJA",
        0x9054: "HANJA",
        0x9055: "HANJA",
        0x9056: "HANJA",
        0x9057: "HANJA",
        0x9058: "HANJA",
        0x9059: "HANJA",
        0x905A: "HANJA",
        0x905B: "HANJA",
        0x905C: "HANJA",
        0x905D: "HANJA",
        0x905E: "HANJA",
        0x905F: "HANJA",
        0x9060: "HANJA",
        0x9061: "HANJA",
        0x9062: "HANJA",
        0x9063: "HANJA",
        0x9064: "HANJA",
        0x9065: "HANJA",
        0x9066: "HANJA",
        0x9067: "HANJA",
        0x9068: "HANJA",
        0x9069: "HANJA",
        0x906A: "HANJA",
        0x906B: "HANJA",
        0x906C: "HANJA",
        0x906D: "HANJA",
        0x906E: "HANJA",
        0x906F: "HANJA",
        0x9070: "HANJA",
        0x9071: "HANJA",
        0x9072: "HANJA",
        0x9073: "HANJA",
        0x9074: "HANJA",
        0x9075: "HANJA",
        0x9076: "HANJA",
        0x9077: "HANJA",
        0x9078: "HANJA",
        0x9079: "HANJA",
        0x907A: "HANJA",
        0x907B: "HANJA",
        0x907C: "HANJA",
        0x907D: "HANJA",
        0x907E: "HANJA",
        0x907F: "HANJA",
        0x9080: "HANJA",
        0x9081: "HANJA",
        0x9082: "HANJA",
        0x9083: "HANJA",
        0x9084: "HANJA",
        0x9085: "HANJA",
        0x9086: "HANJA",
        0x9087: "HANJA",
        0x9088: "HANJA",
        0x9089: "HANJA",
        0x908A: "HANJA",
        0x908B: "HANJA",
        0x908C: "HANJA",
        0x908D: "HANJA",
        0x908E: "HANJA",
        0x908F: "HANJA",
        0x9090: "HANJA",
        0x9091: "HANJA",
        0x9092: "HANJA",
        0x9093: "HANJA",
        0x9094: "HANJA",
        0x9095: "HANJA",
        0x9096: "HANJA",
        0x9097: "HANJA",
        0x9098: "HANJA",
        0x9099: "HANJA",
        0x909A: "HANJA",
        0x909B: "HANJA",
        0x909C: "HANJA",
        0x909D: "HANJA",
        0x909E: "HANJA",
        0x909F: "HANJA",
        0x90A0: "HANJA",
        0x90A1: "HANJA",
        0x90A2: "HANJA",
        0x90A3: "HANJA",
        0x90A4: "HANJA",
        0x90A5: "HANJA",
        0x90A6: "HANJA",
        0x90A7: "HANJA",
        0x90A8: "HANJA",
        0x90A9: "HANJA",
        0x90AA: "HANJA",
        0x90AB: "HANJA",
        0x90AC: "HANJA",
        0x90AD: "HANJA",
        0x90AE: "HANJA",
        0x90AF: "HANJA",
        0x90B0: "HANJA",
        0x90B1: "HANJA",
        0x90B2: "HANJA",
        0x90B3: "HANJA",
        0x90B4: "HANJA",
        0x90B5: "HANJA",
        0x90B6: "HANJA",
        0x90B7: "HANJA",
        0x90B8: "HANJA",
        0x90B9: "HANJA",
        0x90BA: "HANJA",
        0x90BB: "HANJA",
        0x90BC: "HANJA",
        0x90BD: "HANJA",
        0x90BE: "HANJA",
        0x90BF: "HANJA",
        0x90C0: "HANJA",
        0x90C1: "HANJA",
        0x90C2: "HANJA",
        0x90C3: "HANJA",
        0x90C4: "HANJA",
        0x90C5: "HANJA",
        0x90C6: "HANJA",
        0x90C7: "HANJA",
        0x90C8: "HANJA",
        0x90C9: "HANJA",
        0x90CA: "HANJA",
        0x90CB: "HANJA",
        0x90CC: "HANJA",
        0x90CD: "HANJA",
        0x90CE: "HANJA",
        0x90CF: "HANJA",
        0x90D0: "HANJA",
        0x90D1: "HANJA",
        0x90D2: "HANJA",
        0x90D3: "HANJA",
        0x90D4: "HANJA",
        0x90D5: "HANJA",
        0x90D6: "HANJA",
        0x90D7: "HANJA",
        0x90D8: "HANJA",
        0x90D9: "HANJA",
        0x90DA: "HANJA",
        0x90DB: "HANJA",
        0x90DC: "HANJA",
        0x90DD: "HANJA",
        0x90DE: "HANJA",
        0x90DF: "HANJA",
        0x90E0: "HANJA",
        0x90E1: "HANJA",
        0x90E2: "HANJA",
        0x90E3: "HANJA",
        0x90E4: "HANJA",
        0x90E5: "HANJA",
        0x90E6: "HANJA",
        0x90E7: "HANJA",
        0x90E8: "HANJA",
        0x90E9: "HANJA",
        0x90EA: "HANJA",
        0x90EB: "HANJA",
        0x90EC: "HANJA",
        0x90ED: "HANJA",
        0x90EE: "HANJA",
        0x90EF: "HANJA",
        0x90F0: "HANJA",
        0x90F1: "HANJA",
        0x90F2: "HANJA",
        0x90F3: "HANJA",
        0x90F4: "HANJA",
        0x90F5: "HANJA",
        0x90F6: "HANJA",
        0x90F7: "HANJA",
        0x90F8: "HANJA",
        0x90F9: "HANJA",
        0x90FA: "HANJA",
        0x90FB: "HANJA",
        0x90FC: "HANJA",
        0x90FD: "HANJA",
        0x90FE: "HANJA",
        0x90FF: "HANJA",
        0x9100: "HANJA",
        0x9101: "HANJA",
        0x9102: "HANJA",
        0x9103: "HANJA",
        0x9104: "HANJA",
        0x9105: "HANJA",
        0x9106: "HANJA",
        0x9107: "HANJA",
        0x9108: "HANJA",
        0x9109: "HANJA",
        0x910A: "HANJA",
        0x910B: "HANJA",
        0x910C: "HANJA",
        0x910D: "HANJA",
        0x910E: "HANJA",
        0x910F: "HANJA",
        0x9110: "HANJA",
        0x9111: "HANJA",
        0x9112: "HANJA",
        0x9113: "HANJA",
        0x9114: "HANJA",
        0x9115: "HANJA",
        0x9116: "HANJA",
        0x9117: "HANJA",
        0x9118: "HANJA",
        0x9119: "HANJA",
        0x911A: "HANJA",
        0x911B: "HANJA",
        0x911C: "HANJA",
        0x911D: "HANJA",
        0x911E: "HANJA",
        0x911F: "HANJA",
        0x9120: "HANJA",
        0x9121: "HANJA",
        0x9122: "HANJA",
        0x9123: "HANJA",
        0x9124: "HANJA",
        0x9125: "HANJA",
        0x9126: "HANJA",
        0x9127: "HANJA",
        0x9128: "HANJA",
        0x9129: "HANJA",
        0x912A: "HANJA",
        0x912B: "HANJA",
        0x912C: "HANJA",
        0x912D: "HANJA",
        0x912E: "HANJA",
        0x912F: "HANJA",
        0x9130: "HANJA",
        0x9131: "HANJA",
        0x9132: "HANJA",
        0x9133: "HANJA",
        0x9134: "HANJA",
        0x9135: "HANJA",
        0x9136: "HANJA",
        0x9137: "HANJA",
        0x9138: "HANJA",
        0x9139: "HANJA",
        0x913A: "HANJA",
        0x913B: "HANJA",
        0x913C: "HANJA",
        0x913D: "HANJA",
        0x913E: "HANJA",
        0x913F: "HANJA",
        0x9140: "HANJA",
        0x9141: "HANJA",
        0x9142: "HANJA",
        0x9143: "HANJA",
        0x9144: "HANJA",
        0x9145: "HANJA",
        0x9146: "HANJA",
        0x9147: "HANJA",
        0x9148: "HANJA",
        0x9149: "HANJA",
        0x914A: "HANJA",
        0x914B: "HANJA",
        0x914C: "HANJA",
        0x914D: "HANJA",
        0x914E: "HANJA",
        0x914F: "HANJA",
        0x9150: "HANJA",
        0x9151: "HANJA",
        0x9152: "HANJA",
        0x9153: "HANJA",
        0x9154: "HANJA",
        0x9155: "HANJA",
        0x9156: "HANJA",
        0x9157: "HANJA",
        0x9158: "HANJA",
        0x9159: "HANJA",
        0x915A: "HANJA",
        0x915B: "HANJA",
        0x915C: "HANJA",
        0x915D: "HANJA",
        0x915E: "HANJA",
        0x915F: "HANJA",
        0x9160: "HANJA",
        0x9161: "HANJA",
        0x9162: "HANJA",
        0x9163: "HANJA",
        0x9164: "HANJA",
        0x9165: "HANJA",
        0x9166: "HANJA",
        0x9167: "HANJA",
        0x9168: "HANJA",
        0x9169: "HANJA",
        0x916A: "HANJA",
        0x916B: "HANJA",
        0x916C: "HANJA",
        0x916D: "HANJA",
        0x916E: "HANJA",
        0x916F: "HANJA",
        0x9170: "HANJA",
        0x9171: "HANJA",
        0x9172: "HANJA",
        0x9173: "HANJA",
        0x9174: "HANJA",
        0x9175: "HANJA",
        0x9176: "HANJA",
        0x9177: "HANJA",
        0x9178: "HANJA",
        0x9179: "HANJA",
        0x917A: "HANJA",
        0x917B: "HANJA",
        0x917C: "HANJA",
        0x917D: "HANJA",
        0x917E: "HANJA",
        0x917F: "HANJA",
        0x9180: "HANJA",
        0x9181: "HANJA",
        0x9182: "HANJA",
        0x9183: "HANJA",
        0x9184: "HANJA",
        0x9185: "HANJA",
        0x9186: "HANJA",
        0x9187: "HANJA",
        0x9188: "HANJA",
        0x9189: "HANJA",
        0x918A: "HANJA",
        0x918B: "HANJA",
        0x918C: "HANJA",
        0x918D: "HANJA",
        0x918E: "HANJA",
        0x918F: "HANJA",
        0x9190: "HANJA",
        0x9191: "HANJA",
        0x9192: "HANJA",
        0x9193: "HANJA",
        0x9194: "HANJA",
        0x9195: "HANJA",
        0x9196: "HANJA",
        0x9197: "HANJA",
        0x9198: "HANJA",
        0x9199: "HANJA",
        0x919A: "HANJA",
        0x919B: "HANJA",
        0x919C: "HANJA",
        0x919D: "HANJA",
        0x919E: "HANJA",
        0x919F: "HANJA",
        0x91A0: "HANJA",
        0x91A1: "HANJA",
        0x91A2: "HANJA",
        0x91A3: "HANJA",
        0x91A4: "HANJA",
        0x91A5: "HANJA",
        0x91A6: "HANJA",
        0x91A7: "HANJA",
        0x91A8: "HANJA",
        0x91A9: "HANJA",
        0x91AA: "HANJA",
        0x91AB: "HANJA",
        0x91AC: "HANJA",
        0x91AD: "HANJA",
        0x91AE: "HANJA",
        0x91AF: "HANJA",
        0x91B0: "HANJA",
        0x91B1: "HANJA",
        0x91B2: "HANJA",
        0x91B3: "HANJA",
        0x91B4: "HANJA",
        0x91B5: "HANJA",
        0x91B6: "HANJA",
        0x91B7: "HANJA",
        0x91B8: "HANJA",
        0x91B9: "HANJA",
        0x91BA: "HANJA",
        0x91BB: "HANJA",
        0x91BC: "HANJA",
        0x91BD: "HANJA",
        0x91BE: "HANJA",
        0x91BF: "HANJA",
        0x91C0: "HANJA",
        0x91C1: "HANJA",
        0x91C2: "HANJA",
        0x91C3: "HANJA",
        0x91C4: "HANJA",
        0x91C5: "HANJA",
        0x91C6: "HANJA",
        0x91C7: "HANJA",
        0x91C8: "HANJA",
        0x91C9: "HANJA",
        0x91CA: "HANJA",
        0x91CB: "HANJA",
        0x91CC: "HANJA",
        0x91CD: "HANJA",
        0x91CE: "HANJA",
        0x91CF: "HANJA",
        0x91D0: "HANJA",
        0x91D1: "HANJA",
        0x91D2: "HANJA",
        0x91D3: "HANJA",
        0x91D4: "HANJA",
        0x91D5: "HANJA",
        0x91D6: "HANJA",
        0x91D7: "HANJA",
        0x91D8: "HANJA",
        0x91D9: "HANJA",
        0x91DA: "HANJA",
        0x91DB: "HANJA",
        0x91DC: "HANJA",
        0x91DD: "HANJA",
        0x91DE: "HANJA",
        0x91DF: "HANJA",
        0x91E0: "HANJA",
        0x91E1: "HANJA",
        0x91E2: "HANJA",
        0x91E3: "HANJA",
        0x91E4: "HANJA",
        0x91E5: "HANJA",
        0x91E6: "HANJA",
        0x91E7: "HANJA",
        0x91E8: "HANJA",
        0x91E9: "HANJA",
        0x91EA: "HANJA",
        0x91EB: "HANJA",
        0x91EC: "HANJA",
        0x91ED: "HANJA",
        0x91EE: "HANJA",
        0x91EF: "HANJA",
        0x91F0: "HANJA",
        0x91F1: "HANJA",
        0x91F2: "HANJA",
        0x91F3: "HANJA",
        0x91F4: "HANJA",
        0x91F5: "HANJA",
        0x91F6: "HANJA",
        0x91F7: "HANJA",
        0x91F8: "HANJA",
        0x91F9: "HANJA",
        0x91FA: "HANJA",
        0x91FB: "HANJA",
        0x91FC: "HANJA",
        0x91FD: "HANJA",
        0x91FE: "HANJA",
        0x91FF: "HANJA",
        0x9200: "HANJA",
        0x9201: "HANJA",
        0x9202: "HANJA",
        0x9203: "HANJA",
        0x9204: "HANJA",
        0x9205: "HANJA",
        0x9206: "HANJA",
        0x9207: "HANJA",
        0x9208: "HANJA",
        0x9209: "HANJA",
        0x920A: "HANJA",
        0x920B: "HANJA",
        0x920C: "HANJA",
        0x920D: "HANJA",
        0x920E: "HANJA",
        0x920F: "HANJA",
        0x9210: "HANJA",
        0x9211: "HANJA",
        0x9212: "HANJA",
        0x9213: "HANJA",
        0x9214: "HANJA",
        0x9215: "HANJA",
        0x9216: "HANJA",
        0x9217: "HANJA",
        0x9218: "HANJA",
        0x9219: "HANJA",
        0x921A: "HANJA",
        0x921B: "HANJA",
        0x921C: "HANJA",
        0x921D: "HANJA",
        0x921E: "HANJA",
        0x921F: "HANJA",
        0x9220: "HANJA",
        0x9221: "HANJA",
        0x9222: "HANJA",
        0x9223: "HANJA",
        0x9224: "HANJA",
        0x9225: "HANJA",
        0x9226: "HANJA",
        0x9227: "HANJA",
        0x9228: "HANJA",
        0x9229: "HANJA",
        0x922A: "HANJA",
        0x922B: "HANJA",
        0x922C: "HANJA",
        0x922D: "HANJA",
        0x922E: "HANJA",
        0x922F: "HANJA",
        0x9230: "HANJA",
        0x9231: "HANJA",
        0x9232: "HANJA",
        0x9233: "HANJA",
        0x9234: "HANJA",
        0x9235: "HANJA",
        0x9236: "HANJA",
        0x9237: "HANJA",
        0x9238: "HANJA",
        0x9239: "HANJA",
        0x923A: "HANJA",
        0x923B: "HANJA",
        0x923C: "HANJA",
        0x923D: "HANJA",
        0x923E: "HANJA",
        0x923F: "HANJA",
        0x9240: "HANJA",
        0x9241: "HANJA",
        0x9242: "HANJA",
        0x9243: "HANJA",
        0x9244: "HANJA",
        0x9245: "HANJA",
        0x9246: "HANJA",
        0x9247: "HANJA",
        0x9248: "HANJA",
        0x9249: "HANJA",
        0x924A: "HANJA",
        0x924B: "HANJA",
        0x924C: "HANJA",
        0x924D: "HANJA",
        0x924E: "HANJA",
        0x924F: "HANJA",
        0x9250: "HANJA",
        0x9251: "HANJA",
        0x9252: "HANJA",
        0x9253: "HANJA",
        0x9254: "HANJA",
        0x9255: "HANJA",
        0x9256: "HANJA",
        0x9257: "HANJA",
        0x9258: "HANJA",
        0x9259: "HANJA",
        0x925A: "HANJA",
        0x925B: "HANJA",
        0x925C: "HANJA",
        0x925D: "HANJA",
        0x925E: "HANJA",
        0x925F: "HANJA",
        0x9260: "HANJA",
        0x9261: "HANJA",
        0x9262: "HANJA",
        0x9263: "HANJA",
        0x9264: "HANJA",
        0x9265: "HANJA",
        0x9266: "HANJA",
        0x9267: "HANJA",
        0x9268: "HANJA",
        0x9269: "HANJA",
        0x926A: "HANJA",
        0x926B: "HANJA",
        0x926C: "HANJA",
        0x926D: "HANJA",
        0x926E: "HANJA",
        0x926F: "HANJA",
        0x9270: "HANJA",
        0x9271: "HANJA",
        0x9272: "HANJA",
        0x9273: "HANJA",
        0x9274: "HANJA",
        0x9275: "HANJA",
        0x9276: "HANJA",
        0x9277: "HANJA",
        0x9278: "HANJA",
        0x9279: "HANJA",
        0x927A: "HANJA",
        0x927B: "HANJA",
        0x927C: "HANJA",
        0x927D: "HANJA",
        0x927E: "HANJA",
        0x927F: "HANJA",
        0x9280: "HANJA",
        0x9281: "HANJA",
        0x9282: "HANJA",
        0x9283: "HANJA",
        0x9284: "HANJA",
        0x9285: "HANJA",
        0x9286: "HANJA",
        0x9287: "HANJA",
        0x9288: "HANJA",
        0x9289: "HANJA",
        0x928A: "HANJA",
        0x928B: "HANJA",
        0x928C: "HANJA",
        0x928D: "HANJA",
        0x928E: "HANJA",
        0x928F: "HANJA",
        0x9290: "HANJA",
        0x9291: "HANJA",
        0x9292: "HANJA",
        0x9293: "HANJA",
        0x9294: "HANJA",
        0x9295: "HANJA",
        0x9296: "HANJA",
        0x9297: "HANJA",
        0x9298: "HANJA",
        0x9299: "HANJA",
        0x929A: "HANJA",
        0x929B: "HANJA",
        0x929C: "HANJA",
        0x929D: "HANJA",
        0x929E: "HANJA",
        0x929F: "HANJA",
        0x92A0: "HANJA",
        0x92A1: "HANJA",
        0x92A2: "HANJA",
        0x92A3: "HANJA",
        0x92A4: "HANJA",
        0x92A5: "HANJA",
        0x92A6: "HANJA",
        0x92A7: "HANJA",
        0x92A8: "HANJA",
        0x92A9: "HANJA",
        0x92AA: "HANJA",
        0x92AB: "HANJA",
        0x92AC: "HANJA",
        0x92AD: "HANJA",
        0x92AE: "HANJA",
        0x92AF: "HANJA",
        0x92B0: "HANJA",
        0x92B1: "HANJA",
        0x92B2: "HANJA",
        0x92B3: "HANJA",
        0x92B4: "HANJA",
        0x92B5: "HANJA",
        0x92B6: "HANJA",
        0x92B7: "HANJA",
        0x92B8: "HANJA",
        0x92B9: "HANJA",
        0x92BA: "HANJA",
        0x92BB: "HANJA",
        0x92BC: "HANJA",
        0x92BD: "HANJA",
        0x92BE: "HANJA",
        0x92BF: "HANJA",
        0x92C0: "HANJA",
        0x92C1: "HANJA",
        0x92C2: "HANJA",
        0x92C3: "HANJA",
        0x92C4: "HANJA",
        0x92C5: "HANJA",
        0x92C6: "HANJA",
        0x92C7: "HANJA",
        0x92C8: "HANJA",
        0x92C9: "HANJA",
        0x92CA: "HANJA",
        0x92CB: "HANJA",
        0x92CC: "HANJA",
        0x92CD: "HANJA",
        0x92CE: "HANJA",
        0x92CF: "HANJA",
        0x92D0: "HANJA",
        0x92D1: "HANJA",
        0x92D2: "HANJA",
        0x92D3: "HANJA",
        0x92D4: "HANJA",
        0x92D5: "HANJA",
        0x92D6: "HANJA",
        0x92D7: "HANJA",
        0x92D8: "HANJA",
        0x92D9: "HANJA",
        0x92DA: "HANJA",
        0x92DB: "HANJA",
        0x92DC: "HANJA",
        0x92DD: "HANJA",
        0x92DE: "HANJA",
        0x92DF: "HANJA",
        0x92E0: "HANJA",
        0x92E1: "HANJA",
        0x92E2: "HANJA",
        0x92E3: "HANJA",
        0x92E4: "HANJA",
        0x92E5: "HANJA",
        0x92E6: "HANJA",
        0x92E7: "HANJA",
        0x92E8: "HANJA",
        0x92E9: "HANJA",
        0x92EA: "HANJA",
        0x92EB: "HANJA",
        0x92EC: "HANJA",
        0x92ED: "HANJA",
        0x92EE: "HANJA",
        0x92EF: "HANJA",
        0x92F0: "HANJA",
        0x92F1: "HANJA",
        0x92F2: "HANJA",
        0x92F3: "HANJA",
        0x92F4: "HANJA",
        0x92F5: "HANJA",
        0x92F6: "HANJA",
        0x92F7: "HANJA",
        0x92F8: "HANJA",
        0x92F9: "HANJA",
        0x92FA: "HANJA",
        0x92FB: "HANJA",
        0x92FC: "HANJA",
        0x92FD: "HANJA",
        0x92FE: "HANJA",
        0x92FF: "HANJA",
        0x9300: "HANJA",
        0x9301: "HANJA",
        0x9302: "HANJA",
        0x9303: "HANJA",
        0x9304: "HANJA",
        0x9305: "HANJA",
        0x9306: "HANJA",
        0x9307: "HANJA",
        0x9308: "HANJA",
        0x9309: "HANJA",
        0x930A: "HANJA",
        0x930B: "HANJA",
        0x930C: "HANJA",
        0x930D: "HANJA",
        0x930E: "HANJA",
        0x930F: "HANJA",
        0x9310: "HANJA",
        0x9311: "HANJA",
        0x9312: "HANJA",
        0x9313: "HANJA",
        0x9314: "HANJA",
        0x9315: "HANJA",
        0x9316: "HANJA",
        0x9317: "HANJA",
        0x9318: "HANJA",
        0x9319: "HANJA",
        0x931A: "HANJA",
        0x931B: "HANJA",
        0x931C: "HANJA",
        0x931D: "HANJA",
        0x931E: "HANJA",
        0x931F: "HANJA",
        0x9320: "HANJA",
        0x9321: "HANJA",
        0x9322: "HANJA",
        0x9323: "HANJA",
        0x9324: "HANJA",
        0x9325: "HANJA",
        0x9326: "HANJA",
        0x9327: "HANJA",
        0x9328: "HANJA",
        0x9329: "HANJA",
        0x932A: "HANJA",
        0x932B: "HANJA",
        0x932C: "HANJA",
        0x932D: "HANJA",
        0x932E: "HANJA",
        0x932F: "HANJA",
        0x9330: "HANJA",
        0x9331: "HANJA",
        0x9332: "HANJA",
        0x9333: "HANJA",
        0x9334: "HANJA",
        0x9335: "HANJA",
        0x9336: "HANJA",
        0x9337: "HANJA",
        0x9338: "HANJA",
        0x9339: "HANJA",
        0x933A: "HANJA",
        0x933B: "HANJA",
        0x933C: "HANJA",
        0x933D: "HANJA",
        0x933E: "HANJA",
        0x933F: "HANJA",
        0x9340: "HANJA",
        0x9341: "HANJA",
        0x9342: "HANJA",
        0x9343: "HANJA",
        0x9344: "HANJA",
        0x9345: "HANJA",
        0x9346: "HANJA",
        0x9347: "HANJA",
        0x9348: "HANJA",
        0x9349: "HANJA",
        0x934A: "HANJA",
        0x934B: "HANJA",
        0x934C: "HANJA",
        0x934D: "HANJA",
        0x934E: "HANJA",
        0x934F: "HANJA",
        0x9350: "HANJA",
        0x9351: "HANJA",
        0x9352: "HANJA",
        0x9353: "HANJA",
        0x9354: "HANJA",
        0x9355: "HANJA",
        0x9356: "HANJA",
        0x9357: "HANJA",
        0x9358: "HANJA",
        0x9359: "HANJA",
        0x935A: "HANJA",
        0x935B: "HANJA",
        0x935C: "HANJA",
        0x935D: "HANJA",
        0x935E: "HANJA",
        0x935F: "HANJA",
        0x9360: "HANJA",
        0x9361: "HANJA",
        0x9362: "HANJA",
        0x9363: "HANJA",
        0x9364: "HANJA",
        0x9365: "HANJA",
        0x9366: "HANJA",
        0x9367: "HANJA",
        0x9368: "HANJA",
        0x9369: "HANJA",
        0x936A: "HANJA",
        0x936B: "HANJA",
        0x936C: "HANJA",
        0x936D: "HANJA",
        0x936E: "HANJA",
        0x936F: "HANJA",
        0x9370: "HANJA",
        0x9371: "HANJA",
        0x9372: "HANJA",
        0x9373: "HANJA",
        0x9374: "HANJA",
        0x9375: "HANJA",
        0x9376: "HANJA",
        0x9377: "HANJA",
        0x9378: "HANJA",
        0x9379: "HANJA",
        0x937A: "HANJA",
        0x937B: "HANJA",
        0x937C: "HANJA",
        0x937D: "HANJA",
        0x937E: "HANJA",
        0x937F: "HANJA",
        0x9380: "HANJA",
        0x9381: "HANJA",
        0x9382: "HANJA",
        0x9383: "HANJA",
        0x9384: "HANJA",
        0x9385: "HANJA",
        0x9386: "HANJA",
        0x9387: "HANJA",
        0x9388: "HANJA",
        0x9389: "HANJA",
        0x938A: "HANJA",
        0x938B: "HANJA",
        0x938C: "HANJA",
        0x938D: "HANJA",
        0x938E: "HANJA",
        0x938F: "HANJA",
        0x9390: "HANJA",
        0x9391: "HANJA",
        0x9392: "HANJA",
        0x9393: "HANJA",
        0x9394: "HANJA",
        0x9395: "HANJA",
        0x9396: "HANJA",
        0x9397: "HANJA",
        0x9398: "HANJA",
        0x9399: "HANJA",
        0x939A: "HANJA",
        0x939B: "HANJA",
        0x939C: "HANJA",
        0x939D: "HANJA",
        0x939E: "HANJA",
        0x939F: "HANJA",
        0x93A0: "HANJA",
        0x93A1: "HANJA",
        0x93A2: "HANJA",
        0x93A3: "HANJA",
        0x93A4: "HANJA",
        0x93A5: "HANJA",
        0x93A6: "HANJA",
        0x93A7: "HANJA",
        0x93A8: "HANJA",
        0x93A9: "HANJA",
        0x93AA: "HANJA",
        0x93AB: "HANJA",
        0x93AC: "HANJA",
        0x93AD: "HANJA",
        0x93AE: "HANJA",
        0x93AF: "HANJA",
        0x93B0: "HANJA",
        0x93B1: "HANJA",
        0x93B2: "HANJA",
        0x93B3: "HANJA",
        0x93B4: "HANJA",
        0x93B5: "HANJA",
        0x93B6: "HANJA",
        0x93B7: "HANJA",
        0x93B8: "HANJA",
        0x93B9: "HANJA",
        0x93BA: "HANJA",
        0x93BB: "HANJA",
        0x93BC: "HANJA",
        0x93BD: "HANJA",
        0x93BE: "HANJA",
        0x93BF: "HANJA",
        0x93C0: "HANJA",
        0x93C1: "HANJA",
        0x93C2: "HANJA",
        0x93C3: "HANJA",
        0x93C4: "HANJA",
        0x93C5: "HANJA",
        0x93C6: "HANJA",
        0x93C7: "HANJA",
        0x93C8: "HANJA",
        0x93C9: "HANJA",
        0x93CA: "HANJA",
        0x93CB: "HANJA",
        0x93CC: "HANJA",
        0x93CD: "HANJA",
        0x93CE: "HANJA",
        0x93CF: "HANJA",
        0x93D0: "HANJA",
        0x93D1: "HANJA",
        0x93D2: "HANJA",
        0x93D3: "HANJA",
        0x93D4: "HANJA",
        0x93D5: "HANJA",
        0x93D6: "HANJA",
        0x93D7: "HANJA",
        0x93D8: "HANJA",
        0x93D9: "HANJA",
        0x93DA: "HANJA",
        0x93DB: "HANJA",
        0x93DC: "HANJA",
        0x93DD: "HANJA",
        0x93DE: "HANJA",
        0x93DF: "HANJA",
        0x93E0: "HANJA",
        0x93E1: "HANJA",
        0x93E2: "HANJA",
        0x93E3: "HANJA",
        0x93E4: "HANJA",
        0x93E5: "HANJA",
        0x93E6: "HANJA",
        0x93E7: "HANJA",
        0x93E8: "HANJA",
        0x93E9: "HANJA",
        0x93EA: "HANJA",
        0x93EB: "HANJA",
        0x93EC: "HANJA",
        0x93ED: "HANJA",
        0x93EE: "HANJA",
        0x93EF: "HANJA",
        0x93F0: "HANJA",
        0x93F1: "HANJA",
        0x93F2: "HANJA",
        0x93F3: "HANJA",
        0x93F4: "HANJA",
        0x93F5: "HANJA",
        0x93F6: "HANJA",
        0x93F7: "HANJA",
        0x93F8: "HANJA",
        0x93F9: "HANJA",
        0x93FA: "HANJA",
        0x93FB: "HANJA",
        0x93FC: "HANJA",
        0x93FD: "HANJA",
        0x93FE: "HANJA",
        0x93FF: "HANJA",
        0x9400: "HANJA",
        0x9401: "HANJA",
        0x9402: "HANJA",
        0x9403: "HANJA",
        0x9404: "HANJA",
        0x9405: "HANJA",
        0x9406: "HANJA",
        0x9407: "HANJA",
        0x9408: "HANJA",
        0x9409: "HANJA",
        0x940A: "HANJA",
        0x940B: "HANJA",
        0x940C: "HANJA",
        0x940D: "HANJA",
        0x940E: "HANJA",
        0x940F: "HANJA",
        0x9410: "HANJA",
        0x9411: "HANJA",
        0x9412: "HANJA",
        0x9413: "HANJA",
        0x9414: "HANJA",
        0x9415: "HANJA",
        0x9416: "HANJA",
        0x9417: "HANJA",
        0x9418: "HANJA",
        0x9419: "HANJA",
        0x941A: "HANJA",
        0x941B: "HANJA",
        0x941C: "HANJA",
        0x941D: "HANJA",
        0x941E: "HANJA",
        0x941F: "HANJA",
        0x9420: "HANJA",
        0x9421: "HANJA",
        0x9422: "HANJA",
        0x9423: "HANJA",
        0x9424: "HANJA",
        0x9425: "HANJA",
        0x9426: "HANJA",
        0x9427: "HANJA",
        0x9428: "HANJA",
        0x9429: "HANJA",
        0x942A: "HANJA",
        0x942B: "HANJA",
        0x942C: "HANJA",
        0x942D: "HANJA",
        0x942E: "HANJA",
        0x942F: "HANJA",
        0x9430: "HANJA",
        0x9431: "HANJA",
        0x9432: "HANJA",
        0x9433: "HANJA",
        0x9434: "HANJA",
        0x9435: "HANJA",
        0x9436: "HANJA",
        0x9437: "HANJA",
        0x9438: "HANJA",
        0x9439: "HANJA",
        0x943A: "HANJA",
        0x943B: "HANJA",
        0x943C: "HANJA",
        0x943D: "HANJA",
        0x943E: "HANJA",
        0x943F: "HANJA",
        0x9440: "HANJA",
        0x9441: "HANJA",
        0x9442: "HANJA",
        0x9443: "HANJA",
        0x9444: "HANJA",
        0x9445: "HANJA",
        0x9446: "HANJA",
        0x9447: "HANJA",
        0x9448: "HANJA",
        0x9449: "HANJA",
        0x944A: "HANJA",
        0x944B: "HANJA",
        0x944C: "HANJA",
        0x944D: "HANJA",
        0x944E: "HANJA",
        0x944F: "HANJA",
        0x9450: "HANJA",
        0x9451: "HANJA",
        0x9452: "HANJA",
        0x9453: "HANJA",
        0x9454: "HANJA",
        0x9455: "HANJA",
        0x9456: "HANJA",
        0x9457: "HANJA",
        0x9458: "HANJA",
        0x9459: "HANJA",
        0x945A: "HANJA",
        0x945B: "HANJA",
        0x945C: "HANJA",
        0x945D: "HANJA",
        0x945E: "HANJA",
        0x945F: "HANJA",
        0x9460: "HANJA",
        0x9461: "HANJA",
        0x9462: "HANJA",
        0x9463: "HANJA",
        0x9464: "HANJA",
        0x9465: "HANJA",
        0x9466: "HANJA",
        0x9467: "HANJA",
        0x9468: "HANJA",
        0x9469: "HANJA",
        0x946A: "HANJA",
        0x946B: "HANJA",
        0x946C: "HANJA",
        0x946D: "HANJA",
        0x946E: "HANJA",
        0x946F: "HANJA",
        0x9470: "HANJA",
        0x9471: "HANJA",
        0x9472: "HANJA",
        0x9473: "HANJA",
        0x9474: "HANJA",
        0x9475: "HANJA",
        0x9476: "HANJA",
        0x9477: "HANJA",
        0x9478: "HANJA",
        0x9479: "HANJA",
        0x947A: "HANJA",
        0x947B: "HANJA",
        0x947C: "HANJA",
        0x947D: "HANJA",
        0x947E: "HANJA",
        0x947F: "HANJA",
        0x9480: "HANJA",
        0x9481: "HANJA",
        0x9482: "HANJA",
        0x9483: "HANJA",
        0x9484: "HANJA",
        0x9485: "HANJA",
        0x9486: "HANJA",
        0x9487: "HANJA",
        0x9488: "HANJA",
        0x9489: "HANJA",
        0x948A: "HANJA",
        0x948B: "HANJA",
        0x948C: "HANJA",
        0x948D: "HANJA",
        0x948E: "HANJA",
        0x948F: "HANJA",
        0x9490: "HANJA",
        0x9491: "HANJA",
        0x9492: "HANJA",
        0x9493: "HANJA",
        0x9494: "HANJA",
        0x9495: "HANJA",
        0x9496: "HANJA",
        0x9497: "HANJA",
        0x9498: "HANJA",
        0x9499: "HANJA",
        0x949A: "HANJA",
        0x949B: "HANJA",
        0x949C: "HANJA",
        0x949D: "HANJA",
        0x949E: "HANJA",
        0x949F: "HANJA",
        0x94A0: "HANJA",
        0x94A1: "HANJA",
        0x94A2: "HANJA",
        0x94A3: "HANJA",
        0x94A4: "HANJA",
        0x94A5: "HANJA",
        0x94A6: "HANJA",
        0x94A7: "HANJA",
        0x94A8: "HANJA",
        0x94A9: "HANJA",
        0x94AA: "HANJA",
        0x94AB: "HANJA",
        0x94AC: "HANJA",
        0x94AD: "HANJA",
        0x94AE: "HANJA",
        0x94AF: "HANJA",
        0x94B0: "HANJA",
        0x94B1: "HANJA",
        0x94B2: "HANJA",
        0x94B3: "HANJA",
        0x94B4: "HANJA",
        0x94B5: "HANJA",
        0x94B6: "HANJA",
        0x94B7: "HANJA",
        0x94B8: "HANJA",
        0x94B9: "HANJA",
        0x94BA: "HANJA",
        0x94BB: "HANJA",
        0x94BC: "HANJA",
        0x94BD: "HANJA",
        0x94BE: "HANJA",
        0x94BF: "HANJA",
        0x94C0: "HANJA",
        0x94C1: "HANJA",
        0x94C2: "HANJA",
        0x94C3: "HANJA",
        0x94C4: "HANJA",
        0x94C5: "HANJA",
        0x94C6: "HANJA",
        0x94C7: "HANJA",
        0x94C8: "HANJA",
        0x94C9: "HANJA",
        0x94CA: "HANJA",
        0x94CB: "HANJA",
        0x94CC: "HANJA",
        0x94CD: "HANJA",
        0x94CE: "HANJA",
        0x94CF: "HANJA",
        0x94D0: "HANJA",
        0x94D1: "HANJA",
        0x94D2: "HANJA",
        0x94D3: "HANJA",
        0x94D4: "HANJA",
        0x94D5: "HANJA",
        0x94D6: "HANJA",
        0x94D7: "HANJA",
        0x94D8: "HANJA",
        0x94D9: "HANJA",
        0x94DA: "HANJA",
        0x94DB: "HANJA",
        0x94DC: "HANJA",
        0x94DD: "HANJA",
        0x94DE: "HANJA",
        0x94DF: "HANJA",
        0x94E0: "HANJA",
        0x94E1: "HANJA",
        0x94E2: "HANJA",
        0x94E3: "HANJA",
        0x94E4: "HANJA",
        0x94E5: "HANJA",
        0x94E6: "HANJA",
        0x94E7: "HANJA",
        0x94E8: "HANJA",
        0x94E9: "HANJA",
        0x94EA: "HANJA",
        0x94EB: "HANJA",
        0x94EC: "HANJA",
        0x94ED: "HANJA",
        0x94EE: "HANJA",
        0x94EF: "HANJA",
        0x94F0: "HANJA",
        0x94F1: "HANJA",
        0x94F2: "HANJA",
        0x94F3: "HANJA",
        0x94F4: "HANJA",
        0x94F5: "HANJA",
        0x94F6: "HANJA",
        0x94F7: "HANJA",
        0x94F8: "HANJA",
        0x94F9: "HANJA",
        0x94FA: "HANJA",
        0x94FB: "HANJA",
        0x94FC: "HANJA",
        0x94FD: "HANJA",
        0x94FE: "HANJA",
        0x94FF: "HANJA",
        0x9500: "HANJA",
        0x9501: "HANJA",
        0x9502: "HANJA",
        0x9503: "HANJA",
        0x9504: "HANJA",
        0x9505: "HANJA",
        0x9506: "HANJA",
        0x9507: "HANJA",
        0x9508: "HANJA",
        0x9509: "HANJA",
        0x950A: "HANJA",
        0x950B: "HANJA",
        0x950C: "HANJA",
        0x950D: "HANJA",
        0x950E: "HANJA",
        0x950F: "HANJA",
        0x9510: "HANJA",
        0x9511: "HANJA",
        0x9512: "HANJA",
        0x9513: "HANJA",
        0x9514: "HANJA",
        0x9515: "HANJA",
        0x9516: "HANJA",
        0x9517: "HANJA",
        0x9518: "HANJA",
        0x9519: "HANJA",
        0x951A: "HANJA",
        0x951B: "HANJA",
        0x951C: "HANJA",
        0x951D: "HANJA",
        0x951E: "HANJA",
        0x951F: "HANJA",
        0x9520: "HANJA",
        0x9521: "HANJA",
        0x9522: "HANJA",
        0x9523: "HANJA",
        0x9524: "HANJA",
        0x9525: "HANJA",
        0x9526: "HANJA",
        0x9527: "HANJA",
        0x9528: "HANJA",
        0x9529: "HANJA",
        0x952A: "HANJA",
        0x952B: "HANJA",
        0x952C: "HANJA",
        0x952D: "HANJA",
        0x952E: "HANJA",
        0x952F: "HANJA",
        0x9530: "HANJA",
        0x9531: "HANJA",
        0x9532: "HANJA",
        0x9533: "HANJA",
        0x9534: "HANJA",
        0x9535: "HANJA",
        0x9536: "HANJA",
        0x9537: "HANJA",
        0x9538: "HANJA",
        0x9539: "HANJA",
        0x953A: "HANJA",
        0x953B: "HANJA",
        0x953C: "HANJA",
        0x953D: "HANJA",
        0x953E: "HANJA",
        0x953F: "HANJA",
        0x9540: "HANJA",
        0x9541: "HANJA",
        0x9542: "HANJA",
        0x9543: "HANJA",
        0x9544: "HANJA",
        0x9545: "HANJA",
        0x9546: "HANJA",
        0x9547: "HANJA",
        0x9548: "HANJA",
        0x9549: "HANJA",
        0x954A: "HANJA",
        0x954B: "HANJA",
        0x954C: "HANJA",
        0x954D: "HANJA",
        0x954E: "HANJA",
        0x954F: "HANJA",
        0x9550: "HANJA",
        0x9551: "HANJA",
        0x9552: "HANJA",
        0x9553: "HANJA",
        0x9554: "HANJA",
        0x9555: "HANJA",
        0x9556: "HANJA",
        0x9557: "HANJA",
        0x9558: "HANJA",
        0x9559: "HANJA",
        0x955A: "HANJA",
        0x955B: "HANJA",
        0x955C: "HANJA",
        0x955D: "HANJA",
        0x955E: "HANJA",
        0x955F: "HANJA",
        0x9560: "HANJA",
        0x9561: "HANJA",
        0x9562: "HANJA",
        0x9563: "HANJA",
        0x9564: "HANJA",
        0x9565: "HANJA",
        0x9566: "HANJA",
        0x9567: "HANJA",
        0x9568: "HANJA",
        0x9569: "HANJA",
        0x956A: "HANJA",
        0x956B: "HANJA",
        0x956C: "HANJA",
        0x956D: "HANJA",
        0x956E: "HANJA",
        0x956F: "HANJA",
        0x9570: "HANJA",
        0x9571: "HANJA",
        0x9572: "HANJA",
        0x9573: "HANJA",
        0x9574: "HANJA",
        0x9575: "HANJA",
        0x9576: "HANJA",
        0x9577: "HANJA",
        0x9578: "HANJA",
        0x9579: "HANJA",
        0x957A: "HANJA",
        0x957B: "HANJA",
        0x957C: "HANJA",
        0x957D: "HANJA",
        0x957E: "HANJA",
        0x957F: "HANJA",
        0x9580: "HANJA",
        0x9581: "HANJA",
        0x9582: "HANJA",
        0x9583: "HANJA",
        0x9584: "HANJA",
        0x9585: "HANJA",
        0x9586: "HANJA",
        0x9587: "HANJA",
        0x9588: "HANJA",
        0x9589: "HANJA",
        0x958A: "HANJA",
        0x958B: "HANJA",
        0x958C: "HANJA",
        0x958D: "HANJA",
        0x958E: "HANJA",
        0x958F: "HANJA",
        0x9590: "HANJA",
        0x9591: "HANJA",
        0x9592: "HANJA",
        0x9593: "HANJA",
        0x9594: "HANJA",
        0x9595: "HANJA",
        0x9596: "HANJA",
        0x9597: "HANJA",
        0x9598: "HANJA",
        0x9599: "HANJA",
        0x959A: "HANJA",
        0x959B: "HANJA",
        0x959C: "HANJA",
        0x959D: "HANJA",
        0x959E: "HANJA",
        0x959F: "HANJA",
        0x95A0: "HANJA",
        0x95A1: "HANJA",
        0x95A2: "HANJA",
        0x95A3: "HANJA",
        0x95A4: "HANJA",
        0x95A5: "HANJA",
        0x95A6: "HANJA",
        0x95A7: "HANJA",
        0x95A8: "HANJA",
        0x95A9: "HANJA",
        0x95AA: "HANJA",
        0x95AB: "HANJA",
        0x95AC: "HANJA",
        0x95AD: "HANJA",
        0x95AE: "HANJA",
        0x95AF: "HANJA",
        0x95B0: "HANJA",
        0x95B1: "HANJA",
        0x95B2: "HANJA",
        0x95B3: "HANJA",
        0x95B4: "HANJA",
        0x95B5: "HANJA",
        0x95B6: "HANJA",
        0x95B7: "HANJA",
        0x95B8: "HANJA",
        0x95B9: "HANJA",
        0x95BA: "HANJA",
        0x95BB: "HANJA",
        0x95BC: "HANJA",
        0x95BD: "HANJA",
        0x95BE: "HANJA",
        0x95BF: "HANJA",
        0x95C0: "HANJA",
        0x95C1: "HANJA",
        0x95C2: "HANJA",
        0x95C3: "HANJA",
        0x95C4: "HANJA",
        0x95C5: "HANJA",
        0x95C6: "HANJA",
        0x95C7: "HANJA",
        0x95C8: "HANJA",
        0x95C9: "HANJA",
        0x95CA: "HANJA",
        0x95CB: "HANJA",
        0x95CC: "HANJA",
        0x95CD: "HANJA",
        0x95CE: "HANJA",
        0x95CF: "HANJA",
        0x95D0: "HANJA",
        0x95D1: "HANJA",
        0x95D2: "HANJA",
        0x95D3: "HANJA",
        0x95D4: "HANJA",
        0x95D5: "HANJA",
        0x95D6: "HANJA",
        0x95D7: "HANJA",
        0x95D8: "HANJA",
        0x95D9: "HANJA",
        0x95DA: "HANJA",
        0x95DB: "HANJA",
        0x95DC: "HANJA",
        0x95DD: "HANJA",
        0x95DE: "HANJA",
        0x95DF: "HANJA",
        0x95E0: "HANJA",
        0x95E1: "HANJA",
        0x95E2: "HANJA",
        0x95E3: "HANJA",
        0x95E4: "HANJA",
        0x95E5: "HANJA",
        0x95E6: "HANJA",
        0x95E7: "HANJA",
        0x95E8: "HANJA",
        0x95E9: "HANJA",
        0x95EA: "HANJA",
        0x95EB: "HANJA",
        0x95EC: "HANJA",
        0x95ED: "HANJA",
        0x95EE: "HANJA",
        0x95EF: "HANJA",
        0x95F0: "HANJA",
        0x95F1: "HANJA",
        0x95F2: "HANJA",
        0x95F3: "HANJA",
        0x95F4: "HANJA",
        0x95F5: "HANJA",
        0x95F6: "HANJA",
        0x95F7: "HANJA",
        0x95F8: "HANJA",
        0x95F9: "HANJA",
        0x95FA: "HANJA",
        0x95FB: "HANJA",
        0x95FC: "HANJA",
        0x95FD: "HANJA",
        0x95FE: "HANJA",
        0x95FF: "HANJA",
        0x9600: "HANJA",
        0x9601: "HANJA",
        0x9602: "HANJA",
        0x9603: "HANJA",
        0x9604: "HANJA",
        0x9605: "HANJA",
        0x9606: "HANJA",
        0x9607: "HANJA",
        0x9608: "HANJA",
        0x9609: "HANJA",
        0x960A: "HANJA",
        0x960B: "HANJA",
        0x960C: "HANJA",
        0x960D: "HANJA",
        0x960E: "HANJA",
        0x960F: "HANJA",
        0x9610: "HANJA",
        0x9611: "HANJA",
        0x9612: "HANJA",
        0x9613: "HANJA",
        0x9614: "HANJA",
        0x9615: "HANJA",
        0x9616: "HANJA",
        0x9617: "HANJA",
        0x9618: "HANJA",
        0x9619: "HANJA",
        0x961A: "HANJA",
        0x961B: "HANJA",
        0x961C: "HANJA",
        0x961D: "HANJA",
        0x961E: "HANJA",
        0x961F: "HANJA",
        0x9620: "HANJA",
        0x9621: "HANJA",
        0x9622: "HANJA",
        0x9623: "HANJA",
        0x9624: "HANJA",
        0x9625: "HANJA",
        0x9626: "HANJA",
        0x9627: "HANJA",
        0x9628: "HANJA",
        0x9629: "HANJA",
        0x962A: "HANJA",
        0x962B: "HANJA",
        0x962C: "HANJA",
        0x962D: "HANJA",
        0x962E: "HANJA",
        0x962F: "HANJA",
        0x9630: "HANJA",
        0x9631: "HANJA",
        0x9632: "HANJA",
        0x9633: "HANJA",
        0x9634: "HANJA",
        0x9635: "HANJA",
        0x9636: "HANJA",
        0x9637: "HANJA",
        0x9638: "HANJA",
        0x9639: "HANJA",
        0x963A: "HANJA",
        0x963B: "HANJA",
        0x963C: "HANJA",
        0x963D: "HANJA",
        0x963E: "HANJA",
        0x963F: "HANJA",
        0x9640: "HANJA",
        0x9641: "HANJA",
        0x9642: "HANJA",
        0x9643: "HANJA",
        0x9644: "HANJA",
        0x9645: "HANJA",
        0x9646: "HANJA",
        0x9647: "HANJA",
        0x9648: "HANJA",
        0x9649: "HANJA",
        0x964A: "HANJA",
        0x964B: "HANJA",
        0x964C: "HANJA",
        0x964D: "HANJA",
        0x964E: "HANJA",
        0x964F: "HANJA",
        0x9650: "HANJA",
        0x9651: "HANJA",
        0x9652: "HANJA",
        0x9653: "HANJA",
        0x9654: "HANJA",
        0x9655: "HANJA",
        0x9656: "HANJA",
        0x9657: "HANJA",
        0x9658: "HANJA",
        0x9659: "HANJA",
        0x965A: "HANJA",
        0x965B: "HANJA",
        0x965C: "HANJA",
        0x965D: "HANJA",
        0x965E: "HANJA",
        0x965F: "HANJA",
        0x9660: "HANJA",
        0x9661: "HANJA",
        0x9662: "HANJA",
        0x9663: "HANJA",
        0x9664: "HANJA",
        0x9665: "HANJA",
        0x9666: "HANJA",
        0x9667: "HANJA",
        0x9668: "HANJA",
        0x9669: "HANJA",
        0x966A: "HANJA",
        0x966B: "HANJA",
        0x966C: "HANJA",
        0x966D: "HANJA",
        0x966E: "HANJA",
        0x966F: "HANJA",
        0x9670: "HANJA",
        0x9671: "HANJA",
        0x9672: "HANJA",
        0x9673: "HANJA",
        0x9674: "HANJA",
        0x9675: "HANJA",
        0x9676: "HANJA",
        0x9677: "HANJA",
        0x9678: "HANJA",
        0x9679: "HANJA",
        0x967A: "HANJA",
        0x967B: "HANJA",
        0x967C: "HANJA",
        0x967D: "HANJA",
        0x967E: "HANJA",
        0x967F: "HANJA",
        0x9680: "HANJA",
        0x9681: "HANJA",
        0x9682: "HANJA",
        0x9683: "HANJA",
        0x9684: "HANJA",
        0x9685: "HANJA",
        0x9686: "HANJA",
        0x9687: "HANJA",
        0x9688: "HANJA",
        0x9689: "HANJA",
        0x968A: "HANJA",
        0x968B: "HANJA",
        0x968C: "HANJA",
        0x968D: "HANJA",
        0x968E: "HANJA",
        0x968F: "HANJA",
        0x9690: "HANJA",
        0x9691: "HANJA",
        0x9692: "HANJA",
        0x9693: "HANJA",
        0x9694: "HANJA",
        0x9695: "HANJA",
        0x9696: "HANJA",
        0x9697: "HANJA",
        0x9698: "HANJA",
        0x9699: "HANJA",
        0x969A: "HANJA",
        0x969B: "HANJA",
        0x969C: "HANJA",
        0x969D: "HANJA",
        0x969E: "HANJA",
        0x969F: "HANJA",
        0x96A0: "HANJA",
        0x96A1: "HANJA",
        0x96A2: "HANJA",
        0x96A3: "HANJA",
        0x96A4: "HANJA",
        0x96A5: "HANJA",
        0x96A6: "HANJA",
        0x96A7: "HANJA",
        0x96A8: "HANJA",
        0x96A9: "HANJA",
        0x96AA: "HANJA",
        0x96AB: "HANJA",
        0x96AC: "HANJA",
        0x96AD: "HANJA",
        0x96AE: "HANJA",
        0x96AF: "HANJA",
        0x96B0: "HANJA",
        0x96B1: "HANJA",
        0x96B2: "HANJA",
        0x96B3: "HANJA",
        0x96B4: "HANJA",
        0x96B5: "HANJA",
        0x96B6: "HANJA",
        0x96B7: "HANJA",
        0x96B8: "HANJA",
        0x96B9: "HANJA",
        0x96BA: "HANJA",
        0x96BB: "HANJA",
        0x96BC: "HANJA",
        0x96BD: "HANJA",
        0x96BE: "HANJA",
        0x96BF: "HANJA",
        0x96C0: "HANJA",
        0x96C1: "HANJA",
        0x96C2: "HANJA",
        0x96C3: "HANJA",
        0x96C4: "HANJA",
        0x96C5: "HANJA",
        0x96C6: "HANJA",
        0x96C7: "HANJA",
        0x96C8: "HANJA",
        0x96C9: "HANJA",
        0x96CA: "HANJA",
        0x96CB: "HANJA",
        0x96CC: "HANJA",
        0x96CD: "HANJA",
        0x96CE: "HANJA",
        0x96CF: "HANJA",
        0x96D0: "HANJA",
        0x96D1: "HANJA",
        0x96D2: "HANJA",
        0x96D3: "HANJA",
        0x96D4: "HANJA",
        0x96D5: "HANJA",
        0x96D6: "HANJA",
        0x96D7: "HANJA",
        0x96D8: "HANJA",
        0x96D9: "HANJA",
        0x96DA: "HANJA",
        0x96DB: "HANJA",
        0x96DC: "HANJA",
        0x96DD: "HANJA",
        0x96DE: "HANJA",
        0x96DF: "HANJA",
        0x96E0: "HANJA",
        0x96E1: "HANJA",
        0x96E2: "HANJA",
        0x96E3: "HANJA",
        0x96E4: "HANJA",
        0x96E5: "HANJA",
        0x96E6: "HANJA",
        0x96E7: "HANJA",
        0x96E8: "HANJA",
        0x96E9: "HANJA",
        0x96EA: "HANJA",
        0x96EB: "HANJA",
        0x96EC: "HANJA",
        0x96ED: "HANJA",
        0x96EE: "HANJA",
        0x96EF: "HANJA",
        0x96F0: "HANJA",
        0x96F1: "HANJA",
        0x96F2: "HANJA",
        0x96F3: "HANJA",
        0x96F4: "HANJA",
        0x96F5: "HANJA",
        0x96F6: "HANJA",
        0x96F7: "HANJA",
        0x96F8: "HANJA",
        0x96F9: "HANJA",
        0x96FA: "HANJA",
        0x96FB: "HANJA",
        0x96FC: "HANJA",
        0x96FD: "HANJA",
        0x96FE: "HANJA",
        0x96FF: "HANJA",
        0x9700: "HANJA",
        0x9701: "HANJA",
        0x9702: "HANJA",
        0x9703: "HANJA",
        0x9704: "HANJA",
        0x9705: "HANJA",
        0x9706: "HANJA",
        0x9707: "HANJA",
        0x9708: "HANJA",
        0x9709: "HANJA",
        0x970A: "HANJA",
        0x970B: "HANJA",
        0x970C: "HANJA",
        0x970D: "HANJA",
        0x970E: "HANJA",
        0x970F: "HANJA",
        0x9710: "HANJA",
        0x9711: "HANJA",
        0x9712: "HANJA",
        0x9713: "HANJA",
        0x9714: "HANJA",
        0x9715: "HANJA",
        0x9716: "HANJA",
        0x9717: "HANJA",
        0x9718: "HANJA",
        0x9719: "HANJA",
        0x971A: "HANJA",
        0x971B: "HANJA",
        0x971C: "HANJA",
        0x971D: "HANJA",
        0x971E: "HANJA",
        0x971F: "HANJA",
        0x9720: "HANJA",
        0x9721: "HANJA",
        0x9722: "HANJA",
        0x9723: "HANJA",
        0x9724: "HANJA",
        0x9725: "HANJA",
        0x9726: "HANJA",
        0x9727: "HANJA",
        0x9728: "HANJA",
        0x9729: "HANJA",
        0x972A: "HANJA",
        0x972B: "HANJA",
        0x972C: "HANJA",
        0x972D: "HANJA",
        0x972E: "HANJA",
        0x972F: "HANJA",
        0x9730: "HANJA",
        0x9731: "HANJA",
        0x9732: "HANJA",
        0x9733: "HANJA",
        0x9734: "HANJA",
        0x9735: "HANJA",
        0x9736: "HANJA",
        0x9737: "HANJA",
        0x9738: "HANJA",
        0x9739: "HANJA",
        0x973A: "HANJA",
        0x973B: "HANJA",
        0x973C: "HANJA",
        0x973D: "HANJA",
        0x973E: "HANJA",
        0x973F: "HANJA",
        0x9740: "HANJA",
        0x9741: "HANJA",
        0x9742: "HANJA",
        0x9743: "HANJA",
        0x9744: "HANJA",
        0x9745: "HANJA",
        0x9746: "HANJA",
        0x9747: "HANJA",
        0x9748: "HANJA",
        0x9749: "HANJA",
        0x974A: "HANJA",
        0x974B: "HANJA",
        0x974C: "HANJA",
        0x974D: "HANJA",
        0x974E: "HANJA",
        0x974F: "HANJA",
        0x9750: "HANJA",
        0x9751: "HANJA",
        0x9752: "HANJA",
        0x9753: "HANJA",
        0x9754: "HANJA",
        0x9755: "HANJA",
        0x9756: "HANJA",
        0x9757: "HANJA",
        0x9758: "HANJA",
        0x9759: "HANJA",
        0x975A: "HANJA",
        0x975B: "HANJA",
        0x975C: "HANJA",
        0x975D: "HANJA",
        0x975E: "HANJA",
        0x975F: "HANJA",
        0x9760: "HANJA",
        0x9761: "HANJA",
        0x9762: "HANJA",
        0x9763: "HANJA",
        0x9764: "HANJA",
        0x9765: "HANJA",
        0x9766: "HANJA",
        0x9767: "HANJA",
        0x9768: "HANJA",
        0x9769: "HANJA",
        0x976A: "HANJA",
        0x976B: "HANJA",
        0x976C: "HANJA",
        0x976D: "HANJA",
        0x976E: "HANJA",
        0x976F: "HANJA",
        0x9770: "HANJA",
        0x9771: "HANJA",
        0x9772: "HANJA",
        0x9773: "HANJA",
        0x9774: "HANJA",
        0x9775: "HANJA",
        0x9776: "HANJA",
        0x9777: "HANJA",
        0x9778: "HANJA",
        0x9779: "HANJA",
        0x977A: "HANJA",
        0x977B: "HANJA",
        0x977C: "HANJA",
        0x977D: "HANJA",
        0x977E: "HANJA",
        0x977F: "HANJA",
        0x9780: "HANJA",
        0x9781: "HANJA",
        0x9782: "HANJA",
        0x9783: "HANJA",
        0x9784: "HANJA",
        0x9785: "HANJA",
        0x9786: "HANJA",
        0x9787: "HANJA",
        0x9788: "HANJA",
        0x9789: "HANJA",
        0x978A: "HANJA",
        0x978B: "HANJA",
        0x978C: "HANJA",
        0x978D: "HANJA",
        0x978E: "HANJA",
        0x978F: "HANJA",
        0x9790: "HANJA",
        0x9791: "HANJA",
        0x9792: "HANJA",
        0x9793: "HANJA",
        0x9794: "HANJA",
        0x9795: "HANJA",
        0x9796: "HANJA",
        0x9797: "HANJA",
        0x9798: "HANJA",
        0x9799: "HANJA",
        0x979A: "HANJA",
        0x979B: "HANJA",
        0x979C: "HANJA",
        0x979D: "HANJA",
        0x979E: "HANJA",
        0x979F: "HANJA",
        0x97A0: "HANJA",
        0x97A1: "HANJA",
        0x97A2: "HANJA",
        0x97A3: "HANJA",
        0x97A4: "HANJA",
        0x97A5: "HANJA",
        0x97A6: "HANJA",
        0x97A7: "HANJA",
        0x97A8: "HANJA",
        0x97A9: "HANJA",
        0x97AA: "HANJA",
        0x97AB: "HANJA",
        0x97AC: "HANJA",
        0x97AD: "HANJA",
        0x97AE: "HANJA",
        0x97AF: "HANJA",
        0x97B0: "HANJA",
        0x97B1: "HANJA",
        0x97B2: "HANJA",
        0x97B3: "HANJA",
        0x97B4: "HANJA",
        0x97B5: "HANJA",
        0x97B6: "HANJA",
        0x97B7: "HANJA",
        0x97B8: "HANJA",
        0x97B9: "HANJA",
        0x97BA: "HANJA",
        0x97BB: "HANJA",
        0x97BC: "HANJA",
        0x97BD: "HANJA",
        0x97BE: "HANJA",
        0x97BF: "HANJA",
        0x97C0: "HANJA",
        0x97C1: "HANJA",
        0x97C2: "HANJA",
        0x97C3: "HANJA",
        0x97C4: "HANJA",
        0x97C5: "HANJA",
        0x97C6: "HANJA",
        0x97C7: "HANJA",
        0x97C8: "HANJA",
        0x97C9: "HANJA",
        0x97CA: "HANJA",
        0x97CB: "HANJA",
        0x97CC: "HANJA",
        0x97CD: "HANJA",
        0x97CE: "HANJA",
        0x97CF: "HANJA",
        0x97D0: "HANJA",
        0x97D1: "HANJA",
        0x97D2: "HANJA",
        0x97D3: "HANJA",
        0x97D4: "HANJA",
        0x97D5: "HANJA",
        0x97D6: "HANJA",
        0x97D7: "HANJA",
        0x97D8: "HANJA",
        0x97D9: "HANJA",
        0x97DA: "HANJA",
        0x97DB: "HANJA",
        0x97DC: "HANJA",
        0x97DD: "HANJA",
        0x97DE: "HANJA",
        0x97DF: "HANJA",
        0x97E0: "HANJA",
        0x97E1: "HANJA",
        0x97E2: "HANJA",
        0x97E3: "HANJA",
        0x97E4: "HANJA",
        0x97E5: "HANJA",
        0x97E6: "HANJA",
        0x97E7: "HANJA",
        0x97E8: "HANJA",
        0x97E9: "HANJA",
        0x97EA: "HANJA",
        0x97EB: "HANJA",
        0x97EC: "HANJA",
        0x97ED: "HANJA",
        0x97EE: "HANJA",
        0x97EF: "HANJA",
        0x97F0: "HANJA",
        0x97F1: "HANJA",
        0x97F2: "HANJA",
        0x97F3: "HANJA",
        0x97F4: "HANJA",
        0x97F5: "HANJA",
        0x97F6: "HANJA",
        0x97F7: "HANJA",
        0x97F8: "HANJA",
        0x97F9: "HANJA",
        0x97FA: "HANJA",
        0x97FB: "HANJA",
        0x97FC: "HANJA",
        0x97FD: "HANJA",
        0x97FE: "HANJA",
        0x97FF: "HANJA",
        0x9800: "HANJA",
        0x9801: "HANJA",
        0x9802: "HANJA",
        0x9803: "HANJA",
        0x9804: "HANJA",
        0x9805: "HANJA",
        0x9806: "HANJA",
        0x9807: "HANJA",
        0x9808: "HANJA",
        0x9809: "HANJA",
        0x980A: "HANJA",
        0x980B: "HANJA",
        0x980C: "HANJA",
        0x980D: "HANJA",
        0x980E: "HANJA",
        0x980F: "HANJA",
        0x9810: "HANJA",
        0x9811: "HANJA",
        0x9812: "HANJA",
        0x9813: "HANJA",
        0x9814: "HANJA",
        0x9815: "HANJA",
        0x9816: "HANJA",
        0x9817: "HANJA",
        0x9818: "HANJA",
        0x9819: "HANJA",
        0x981A: "HANJA",
        0x981B: "HANJA",
        0x981C: "HANJA",
        0x981D: "HANJA",
        0x981E: "HANJA",
        0x981F: "HANJA",
        0x9820: "HANJA",
        0x9821: "HANJA",
        0x9822: "HANJA",
        0x9823: "HANJA",
        0x9824: "HANJA",
        0x9825: "HANJA",
        0x9826: "HANJA",
        0x9827: "HANJA",
        0x9828: "HANJA",
        0x9829: "HANJA",
        0x982A: "HANJA",
        0x982B: "HANJA",
        0x982C: "HANJA",
        0x982D: "HANJA",
        0x982E: "HANJA",
        0x982F: "HANJA",
        0x9830: "HANJA",
        0x9831: "HANJA",
        0x9832: "HANJA",
        0x9833: "HANJA",
        0x9834: "HANJA",
        0x9835: "HANJA",
        0x9836: "HANJA",
        0x9837: "HANJA",
        0x9838: "HANJA",
        0x9839: "HANJA",
        0x983A: "HANJA",
        0x983B: "HANJA",
        0x983C: "HANJA",
        0x983D: "HANJA",
        0x983E: "HANJA",
        0x983F: "HANJA",
        0x9840: "HANJA",
        0x9841: "HANJA",
        0x9842: "HANJA",
        0x9843: "HANJA",
        0x9844: "HANJA",
        0x9845: "HANJA",
        0x9846: "HANJA",
        0x9847: "HANJA",
        0x9848: "HANJA",
        0x9849: "HANJA",
        0x984A: "HANJA",
        0x984B: "HANJA",
        0x984C: "HANJA",
        0x984D: "HANJA",
        0x984E: "HANJA",
        0x984F: "HANJA",
        0x9850: "HANJA",
        0x9851: "HANJA",
        0x9852: "HANJA",
        0x9853: "HANJA",
        0x9854: "HANJA",
        0x9855: "HANJA",
        0x9856: "HANJA",
        0x9857: "HANJA",
        0x9858: "HANJA",
        0x9859: "HANJA",
        0x985A: "HANJA",
        0x985B: "HANJA",
        0x985C: "HANJA",
        0x985D: "HANJA",
        0x985E: "HANJA",
        0x985F: "HANJA",
        0x9860: "HANJA",
        0x9861: "HANJA",
        0x9862: "HANJA",
        0x9863: "HANJA",
        0x9864: "HANJA",
        0x9865: "HANJA",
        0x9866: "HANJA",
        0x9867: "HANJA",
        0x9868: "HANJA",
        0x9869: "HANJA",
        0x986A: "HANJA",
        0x986B: "HANJA",
        0x986C: "HANJA",
        0x986D: "HANJA",
        0x986E: "HANJA",
        0x986F: "HANJA",
        0x9870: "HANJA",
        0x9871: "HANJA",
        0x9872: "HANJA",
        0x9873: "HANJA",
        0x9874: "HANJA",
        0x9875: "HANJA",
        0x9876: "HANJA",
        0x9877: "HANJA",
        0x9878: "HANJA",
        0x9879: "HANJA",
        0x987A: "HANJA",
        0x987B: "HANJA",
        0x987C: "HANJA",
        0x987D: "HANJA",
        0x987E: "HANJA",
        0x987F: "HANJA",
        0x9880: "HANJA",
        0x9881: "HANJA",
        0x9882: "HANJA",
        0x9883: "HANJA",
        0x9884: "HANJA",
        0x9885: "HANJA",
        0x9886: "HANJA",
        0x9887: "HANJA",
        0x9888: "HANJA",
        0x9889: "HANJA",
        0x988A: "HANJA",
        0x988B: "HANJA",
        0x988C: "HANJA",
        0x988D: "HANJA",
        0x988E: "HANJA",
        0x988F: "HANJA",
        0x9890: "HANJA",
        0x9891: "HANJA",
        0x9892: "HANJA",
        0x9893: "HANJA",
        0x9894: "HANJA",
        0x9895: "HANJA",
        0x9896: "HANJA",
        0x9897: "HANJA",
        0x9898: "HANJA",
        0x9899: "HANJA",
        0x989A: "HANJA",
        0x989B: "HANJA",
        0x989C: "HANJA",
        0x989D: "HANJA",
        0x989E: "HANJA",
        0x989F: "HANJA",
        0x98A0: "HANJA",
        0x98A1: "HANJA",
        0x98A2: "HANJA",
        0x98A3: "HANJA",
        0x98A4: "HANJA",
        0x98A5: "HANJA",
        0x98A6: "HANJA",
        0x98A7: "HANJA",
        0x98A8: "HANJA",
        0x98A9: "HANJA",
        0x98AA: "HANJA",
        0x98AB: "HANJA",
        0x98AC: "HANJA",
        0x98AD: "HANJA",
        0x98AE: "HANJA",
        0x98AF: "HANJA",
        0x98B0: "HANJA",
        0x98B1: "HANJA",
        0x98B2: "HANJA",
        0x98B3: "HANJA",
        0x98B4: "HANJA",
        0x98B5: "HANJA",
        0x98B6: "HANJA",
        0x98B7: "HANJA",
        0x98B8: "HANJA",
        0x98B9: "HANJA",
        0x98BA: "HANJA",
        0x98BB: "HANJA",
        0x98BC: "HANJA",
        0x98BD: "HANJA",
        0x98BE: "HANJA",
        0x98BF: "HANJA",
        0x98C0: "HANJA",
        0x98C1: "HANJA",
        0x98C2: "HANJA",
        0x98C3: "HANJA",
        0x98C4: "HANJA",
        0x98C5: "HANJA",
        0x98C6: "HANJA",
        0x98C7: "HANJA",
        0x98C8: "HANJA",
        0x98C9: "HANJA",
        0x98CA: "HANJA",
        0x98CB: "HANJA",
        0x98CC: "HANJA",
        0x98CD: "HANJA",
        0x98CE: "HANJA",
        0x98CF: "HANJA",
        0x98D0: "HANJA",
        0x98D1: "HANJA",
        0x98D2: "HANJA",
        0x98D3: "HANJA",
        0x98D4: "HANJA",
        0x98D5: "HANJA",
        0x98D6: "HANJA",
        0x98D7: "HANJA",
        0x98D8: "HANJA",
        0x98D9: "HANJA",
        0x98DA: "HANJA",
        0x98DB: "HANJA",
        0x98DC: "HANJA",
        0x98DD: "HANJA",
        0x98DE: "HANJA",
        0x98DF: "HANJA",
        0x98E0: "HANJA",
        0x98E1: "HANJA",
        0x98E2: "HANJA",
        0x98E3: "HANJA",
        0x98E4: "HANJA",
        0x98E5: "HANJA",
        0x98E6: "HANJA",
        0x98E7: "HANJA",
        0x98E8: "HANJA",
        0x98E9: "HANJA",
        0x98EA: "HANJA",
        0x98EB: "HANJA",
        0x98EC: "HANJA",
        0x98ED: "HANJA",
        0x98EE: "HANJA",
        0x98EF: "HANJA",
        0x98F0: "HANJA",
        0x98F1: "HANJA",
        0x98F2: "HANJA",
        0x98F3: "HANJA",
        0x98F4: "HANJA",
        0x98F5: "HANJA",
        0x98F6: "HANJA",
        0x98F7: "HANJA",
        0x98F8: "HANJA",
        0x98F9: "HANJA",
        0x98FA: "HANJA",
        0x98FB: "HANJA",
        0x98FC: "HANJA",
        0x98FD: "HANJA",
        0x98FE: "HANJA",
        0x98FF: "HANJA",
        0x9900: "HANJA",
        0x9901: "HANJA",
        0x9902: "HANJA",
        0x9903: "HANJA",
        0x9904: "HANJA",
        0x9905: "HANJA",
        0x9906: "HANJA",
        0x9907: "HANJA",
        0x9908: "HANJA",
        0x9909: "HANJA",
        0x990A: "HANJA",
        0x990B: "HANJA",
        0x990C: "HANJA",
        0x990D: "HANJA",
        0x990E: "HANJA",
        0x990F: "HANJA",
        0x9910: "HANJA",
        0x9911: "HANJA",
        0x9912: "HANJA",
        0x9913: "HANJA",
        0x9914: "HANJA",
        0x9915: "HANJA",
        0x9916: "HANJA",
        0x9917: "HANJA",
        0x9918: "HANJA",
        0x9919: "HANJA",
        0x991A: "HANJA",
        0x991B: "HANJA",
        0x991C: "HANJA",
        0x991D: "HANJA",
        0x991E: "HANJA",
        0x991F: "HANJA",
        0x9920: "HANJA",
        0x9921: "HANJA",
        0x9922: "HANJA",
        0x9923: "HANJA",
        0x9924: "HANJA",
        0x9925: "HANJA",
        0x9926: "HANJA",
        0x9927: "HANJA",
        0x9928: "HANJA",
        0x9929: "HANJA",
        0x992A: "HANJA",
        0x992B: "HANJA",
        0x992C: "HANJA",
        0x992D: "HANJA",
        0x992E: "HANJA",
        0x992F: "HANJA",
        0x9930: "HANJA",
        0x9931: "HANJA",
        0x9932: "HANJA",
        0x9933: "HANJA",
        0x9934: "HANJA",
        0x9935: "HANJA",
        0x9936: "HANJA",
        0x9937: "HANJA",
        0x9938: "HANJA",
        0x9939: "HANJA",
        0x993A: "HANJA",
        0x993B: "HANJA",
        0x993C: "HANJA",
        0x993D: "HANJA",
        0x993E: "HANJA",
        0x993F: "HANJA",
        0x9940: "HANJA",
        0x9941: "HANJA",
        0x9942: "HANJA",
        0x9943: "HANJA",
        0x9944: "HANJA",
        0x9945: "HANJA",
        0x9946: "HANJA",
        0x9947: "HANJA",
        0x9948: "HANJA",
        0x9949: "HANJA",
        0x994A: "HANJA",
        0x994B: "HANJA",
        0x994C: "HANJA",
        0x994D: "HANJA",
        0x994E: "HANJA",
        0x994F: "HANJA",
        0x9950: "HANJA",
        0x9951: "HANJA",
        0x9952: "HANJA",
        0x9953: "HANJA",
        0x9954: "HANJA",
        0x9955: "HANJA",
        0x9956: "HANJA",
        0x9957: "HANJA",
        0x9958: "HANJA",
        0x9959: "HANJA",
        0x995A: "HANJA",
        0x995B: "HANJA",
        0x995C: "HANJA",
        0x995D: "HANJA",
        0x995E: "HANJA",
        0x995F: "HANJA",
        0x9960: "HANJA",
        0x9961: "HANJA",
        0x9962: "HANJA",
        0x9963: "HANJA",
        0x9964: "HANJA",
        0x9965: "HANJA",
        0x9966: "HANJA",
        0x9967: "HANJA",
        0x9968: "HANJA",
        0x9969: "HANJA",
        0x996A: "HANJA",
        0x996B: "HANJA",
        0x996C: "HANJA",
        0x996D: "HANJA",
        0x996E: "HANJA",
        0x996F: "HANJA",
        0x9970: "HANJA",
        0x9971: "HANJA",
        0x9972: "HANJA",
        0x9973: "HANJA",
        0x9974: "HANJA",
        0x9975: "HANJA",
        0x9976: "HANJA",
        0x9977: "HANJA",
        0x9978: "HANJA",
        0x9979: "HANJA",
        0x997A: "HANJA",
        0x997B: "HANJA",
        0x997C: "HANJA",
        0x997D: "HANJA",
        0x997E: "HANJA",
        0x997F: "HANJA",
        0x9980: "HANJA",
        0x9981: "HANJA",
        0x9982: "HANJA",
        0x9983: "HANJA",
        0x9984: "HANJA",
        0x9985: "HANJA",
        0x9986: "HANJA",
        0x9987: "HANJA",
        0x9988: "HANJA",
        0x9989: "HANJA",
        0x998A: "HANJA",
        0x998B: "HANJA",
        0x998C: "HANJA",
        0x998D: "HANJA",
        0x998E: "HANJA",
        0x998F: "HANJA",
        0x9990: "HANJA",
        0x9991: "HANJA",
        0x9992: "HANJA",
        0x9993: "HANJA",
        0x9994: "HANJA",
        0x9995: "HANJA",
        0x9996: "HANJA",
        0x9997: "HANJA",
        0x9998: "HANJA",
        0x9999: "HANJA",
        0x999A: "HANJA",
        0x999B: "HANJA",
        0x999C: "HANJA",
        0x999D: "HANJA",
        0x999E: "HANJA",
        0x999F: "HANJA",
        0x99A0: "HANJA",
        0x99A1: "HANJA",
        0x99A2: "HANJA",
        0x99A3: "HANJA",
        0x99A4: "HANJA",
        0x99A5: "HANJA",
        0x99A6: "HANJA",
        0x99A7: "HANJA",
        0x99A8: "HANJA",
        0x99A9: "HANJA",
        0x99AA: "HANJA",
        0x99AB: "HANJA",
        0x99AC: "HANJA",
        0x99AD: "HANJA",
        0x99AE: "HANJA",
        0x99AF: "HANJA",
        0x99B0: "HANJA",
        0x99B1: "HANJA",
        0x99B2: "HANJA",
        0x99B3: "HANJA",
        0x99B4: "HANJA",
        0x99B5: "HANJA",
        0x99B6: "HANJA",
        0x99B7: "HANJA",
        0x99B8: "HANJA",
        0x99B9: "HANJA",
        0x99BA: "HANJA",
        0x99BB: "HANJA",
        0x99BC: "HANJA",
        0x99BD: "HANJA",
        0x99BE: "HANJA",
        0x99BF: "HANJA",
        0x99C0: "HANJA",
        0x99C1: "HANJA",
        0x99C2: "HANJA",
        0x99C3: "HANJA",
        0x99C4: "HANJA",
        0x99C5: "HANJA",
        0x99C6: "HANJA",
        0x99C7: "HANJA",
        0x99C8: "HANJA",
        0x99C9: "HANJA",
        0x99CA: "HANJA",
        0x99CB: "HANJA",
        0x99CC: "HANJA",
        0x99CD: "HANJA",
        0x99CE: "HANJA",
        0x99CF: "HANJA",
        0x99D0: "HANJA",
        0x99D1: "HANJA",
        0x99D2: "HANJA",
        0x99D3: "HANJA",
        0x99D4: "HANJA",
        0x99D5: "HANJA",
        0x99D6: "HANJA",
        0x99D7: "HANJA",
        0x99D8: "HANJA",
        0x99D9: "HANJA",
        0x99DA: "HANJA",
        0x99DB: "HANJA",
        0x99DC: "HANJA",
        0x99DD: "HANJA",
        0x99DE: "HANJA",
        0x99DF: "HANJA",
        0x99E0: "HANJA",
        0x99E1: "HANJA",
        0x99E2: "HANJA",
        0x99E3: "HANJA",
        0x99E4: "HANJA",
        0x99E5: "HANJA",
        0x99E6: "HANJA",
        0x99E7: "HANJA",
        0x99E8: "HANJA",
        0x99E9: "HANJA",
        0x99EA: "HANJA",
        0x99EB: "HANJA",
        0x99EC: "HANJA",
        0x99ED: "HANJA",
        0x99EE: "HANJA",
        0x99EF: "HANJA",
        0x99F0: "HANJA",
        0x99F1: "HANJA",
        0x99F2: "HANJA",
        0x99F3: "HANJA",
        0x99F4: "HANJA",
        0x99F5: "HANJA",
        0x99F6: "HANJA",
        0x99F7: "HANJA",
        0x99F8: "HANJA",
        0x99F9: "HANJA",
        0x99FA: "HANJA",
        0x99FB: "HANJA",
        0x99FC: "HANJA",
        0x99FD: "HANJA",
        0x99FE: "HANJA",
        0x99FF: "HANJA",
        0x9A00: "HANJA",
        0x9A01: "HANJA",
        0x9A02: "HANJA",
        0x9A03: "HANJA",
        0x9A04: "HANJA",
        0x9A05: "HANJA",
        0x9A06: "HANJA",
        0x9A07: "HANJA",
        0x9A08: "HANJA",
        0x9A09: "HANJA",
        0x9A0A: "HANJA",
        0x9A0B: "HANJA",
        0x9A0C: "HANJA",
        0x9A0D: "HANJA",
        0x9A0E: "HANJA",
        0x9A0F: "HANJA",
        0x9A10: "HANJA",
        0x9A11: "HANJA",
        0x9A12: "HANJA",
        0x9A13: "HANJA",
        0x9A14: "HANJA",
        0x9A15: "HANJA",
        0x9A16: "HANJA",
        0x9A17: "HANJA",
        0x9A18: "HANJA",
        0x9A19: "HANJA",
        0x9A1A: "HANJA",
        0x9A1B: "HANJA",
        0x9A1C: "HANJA",
        0x9A1D: "HANJA",
        0x9A1E: "HANJA",
        0x9A1F: "HANJA",
        0x9A20: "HANJA",
        0x9A21: "HANJA",
        0x9A22: "HANJA",
        0x9A23: "HANJA",
        0x9A24: "HANJA",
        0x9A25: "HANJA",
        0x9A26: "HANJA",
        0x9A27: "HANJA",
        0x9A28: "HANJA",
        0x9A29: "HANJA",
        0x9A2A: "HANJA",
        0x9A2B: "HANJA",
        0x9A2C: "HANJA",
        0x9A2D: "HANJA",
        0x9A2E: "HANJA",
        0x9A2F: "HANJA",
        0x9A30: "HANJA",
        0x9A31: "HANJA",
        0x9A32: "HANJA",
        0x9A33: "HANJA",
        0x9A34: "HANJA",
        0x9A35: "HANJA",
        0x9A36: "HANJA",
        0x9A37: "HANJA",
        0x9A38: "HANJA",
        0x9A39: "HANJA",
        0x9A3A: "HANJA",
        0x9A3B: "HANJA",
        0x9A3C: "HANJA",
        0x9A3D: "HANJA",
        0x9A3E: "HANJA",
        0x9A3F: "HANJA",
        0x9A40: "HANJA",
        0x9A41: "HANJA",
        0x9A42: "HANJA",
        0x9A43: "HANJA",
        0x9A44: "HANJA",
        0x9A45: "HANJA",
        0x9A46: "HANJA",
        0x9A47: "HANJA",
        0x9A48: "HANJA",
        0x9A49: "HANJA",
        0x9A4A: "HANJA",
        0x9A4B: "HANJA",
        0x9A4C: "HANJA",
        0x9A4D: "HANJA",
        0x9A4E: "HANJA",
        0x9A4F: "HANJA",
        0x9A50: "HANJA",
        0x9A51: "HANJA",
        0x9A52: "HANJA",
        0x9A53: "HANJA",
        0x9A54: "HANJA",
        0x9A55: "HANJA",
        0x9A56: "HANJA",
        0x9A57: "HANJA",
        0x9A58: "HANJA",
        0x9A59: "HANJA",
        0x9A5A: "HANJA",
        0x9A5B: "HANJA",
        0x9A5C: "HANJA",
        0x9A5D: "HANJA",
        0x9A5E: "HANJA",
        0x9A5F: "HANJA",
        0x9A60: "HANJA",
        0x9A61: "HANJA",
        0x9A62: "HANJA",
        0x9A63: "HANJA",
        0x9A64: "HANJA",
        0x9A65: "HANJA",
        0x9A66: "HANJA",
        0x9A67: "HANJA",
        0x9A68: "HANJA",
        0x9A69: "HANJA",
        0x9A6A: "HANJA",
        0x9A6B: "HANJA",
        0x9A6C: "HANJA",
        0x9A6D: "HANJA",
        0x9A6E: "HANJA",
        0x9A6F: "HANJA",
        0x9A70: "HANJA",
        0x9A71: "HANJA",
        0x9A72: "HANJA",
        0x9A73: "HANJA",
        0x9A74: "HANJA",
        0x9A75: "HANJA",
        0x9A76: "HANJA",
        0x9A77: "HANJA",
        0x9A78: "HANJA",
        0x9A79: "HANJA",
        0x9A7A: "HANJA",
        0x9A7B: "HANJA",
        0x9A7C: "HANJA",
        0x9A7D: "HANJA",
        0x9A7E: "HANJA",
        0x9A7F: "HANJA",
        0x9A80: "HANJA",
        0x9A81: "HANJA",
        0x9A82: "HANJA",
        0x9A83: "HANJA",
        0x9A84: "HANJA",
        0x9A85: "HANJA",
        0x9A86: "HANJA",
        0x9A87: "HANJA",
        0x9A88: "HANJA",
        0x9A89: "HANJA",
        0x9A8A: "HANJA",
        0x9A8B: "HANJA",
        0x9A8C: "HANJA",
        0x9A8D: "HANJA",
        0x9A8E: "HANJA",
        0x9A8F: "HANJA",
        0x9A90: "HANJA",
        0x9A91: "HANJA",
        0x9A92: "HANJA",
        0x9A93: "HANJA",
        0x9A94: "HANJA",
        0x9A95: "HANJA",
        0x9A96: "HANJA",
        0x9A97: "HANJA",
        0x9A98: "HANJA",
        0x9A99: "HANJA",
        0x9A9A: "HANJA",
        0x9A9B: "HANJA",
        0x9A9C: "HANJA",
        0x9A9D: "HANJA",
        0x9A9E: "HANJA",
        0x9A9F: "HANJA",
        0x9AA0: "HANJA",
        0x9AA1: "HANJA",
        0x9AA2: "HANJA",
        0x9AA3: "HANJA",
        0x9AA4: "HANJA",
        0x9AA5: "HANJA",
        0x9AA6: "HANJA",
        0x9AA7: "HANJA",
        0x9AA8: "HANJA",
        0x9AA9: "HANJA",
        0x9AAA: "HANJA",
        0x9AAB: "HANJA",
        0x9AAC: "HANJA",
        0x9AAD: "HANJA",
        0x9AAE: "HANJA",
        0x9AAF: "HANJA",
        0x9AB0: "HANJA",
        0x9AB1: "HANJA",
        0x9AB2: "HANJA",
        0x9AB3: "HANJA",
        0x9AB4: "HANJA",
        0x9AB5: "HANJA",
        0x9AB6: "HANJA",
        0x9AB7: "HANJA",
        0x9AB8: "HANJA",
        0x9AB9: "HANJA",
        0x9ABA: "HANJA",
        0x9ABB: "HANJA",
        0x9ABC: "HANJA",
        0x9ABD: "HANJA",
        0x9ABE: "HANJA",
        0x9ABF: "HANJA",
        0x9AC0: "HANJA",
        0x9AC1: "HANJA",
        0x9AC2: "HANJA",
        0x9AC3: "HANJA",
        0x9AC4: "HANJA",
        0x9AC5: "HANJA",
        0x9AC6: "HANJA",
        0x9AC7: "HANJA",
        0x9AC8: "HANJA",
        0x9AC9: "HANJA",
        0x9ACA: "HANJA",
        0x9ACB: "HANJA",
        0x9ACC: "HANJA",
        0x9ACD: "HANJA",
        0x9ACE: "HANJA",
        0x9ACF: "HANJA",
        0x9AD0: "HANJA",
        0x9AD1: "HANJA",
        0x9AD2: "HANJA",
        0x9AD3: "HANJA",
        0x9AD4: "HANJA",
        0x9AD5: "HANJA",
        0x9AD6: "HANJA",
        0x9AD7: "HANJA",
        0x9AD8: "HANJA",
        0x9AD9: "HANJA",
        0x9ADA: "HANJA",
        0x9ADB: "HANJA",
        0x9ADC: "HANJA",
        0x9ADD: "HANJA",
        0x9ADE: "HANJA",
        0x9ADF: "HANJA",
        0x9AE0: "HANJA",
        0x9AE1: "HANJA",
        0x9AE2: "HANJA",
        0x9AE3: "HANJA",
        0x9AE4: "HANJA",
        0x9AE5: "HANJA",
        0x9AE6: "HANJA",
        0x9AE7: "HANJA",
        0x9AE8: "HANJA",
        0x9AE9: "HANJA",
        0x9AEA: "HANJA",
        0x9AEB: "HANJA",
        0x9AEC: "HANJA",
        0x9AED: "HANJA",
        0x9AEE: "HANJA",
        0x9AEF: "HANJA",
        0x9AF0: "HANJA",
        0x9AF1: "HANJA",
        0x9AF2: "HANJA",
        0x9AF3: "HANJA",
        0x9AF4: "HANJA",
        0x9AF5: "HANJA",
        0x9AF6: "HANJA",
        0x9AF7: "HANJA",
        0x9AF8: "HANJA",
        0x9AF9: "HANJA",
        0x9AFA: "HANJA",
        0x9AFB: "HANJA",
        0x9AFC: "HANJA",
        0x9AFD: "HANJA",
        0x9AFE: "HANJA",
        0x9AFF: "HANJA",
        0x9B00: "HANJA",
        0x9B01: "HANJA",
        0x9B02: "HANJA",
        0x9B03: "HANJA",
        0x9B04: "HANJA",
        0x9B05: "HANJA",
        0x9B06: "HANJA",
        0x9B07: "HANJA",
        0x9B08: "HANJA",
        0x9B09: "HANJA",
        0x9B0A: "HANJA",
        0x9B0B: "HANJA",
        0x9B0C: "HANJA",
        0x9B0D: "HANJA",
        0x9B0E: "HANJA",
        0x9B0F: "HANJA",
        0x9B10: "HANJA",
        0x9B11: "HANJA",
        0x9B12: "HANJA",
        0x9B13: "HANJA",
        0x9B14: "HANJA",
        0x9B15: "HANJA",
        0x9B16: "HANJA",
        0x9B17: "HANJA",
        0x9B18: "HANJA",
        0x9B19: "HANJA",
        0x9B1A: "HANJA",
        0x9B1B: "HANJA",
        0x9B1C: "HANJA",
        0x9B1D: "HANJA",
        0x9B1E: "HANJA",
        0x9B1F: "HANJA",
        0x9B20: "HANJA",
        0x9B21: "HANJA",
        0x9B22: "HANJA",
        0x9B23: "HANJA",
        0x9B24: "HANJA",
        0x9B25: "HANJA",
        0x9B26: "HANJA",
        0x9B27: "HANJA",
        0x9B28: "HANJA",
        0x9B29: "HANJA",
        0x9B2A: "HANJA",
        0x9B2B: "HANJA",
        0x9B2C: "HANJA",
        0x9B2D: "HANJA",
        0x9B2E: "HANJA",
        0x9B2F: "HANJA",
        0x9B30: "HANJA",
        0x9B31: "HANJA",
        0x9B32: "HANJA",
        0x9B33: "HANJA",
        0x9B34: "HANJA",
        0x9B35: "HANJA",
        0x9B36: "HANJA",
        0x9B37: "HANJA",
        0x9B38: "HANJA",
        0x9B39: "HANJA",
        0x9B3A: "HANJA",
        0x9B3B: "HANJA",
        0x9B3C: "HANJA",
        0x9B3D: "HANJA",
        0x9B3E: "HANJA",
        0x9B3F: "HANJA",
        0x9B40: "HANJA",
        0x9B41: "HANJA",
        0x9B42: "HANJA",
        0x9B43: "HANJA",
        0x9B44: "HANJA",
        0x9B45: "HANJA",
        0x9B46: "HANJA",
        0x9B47: "HANJA",
        0x9B48: "HANJA",
        0x9B49: "HANJA",
        0x9B4A: "HANJA",
        0x9B4B: "HANJA",
        0x9B4C: "HANJA",
        0x9B4D: "HANJA",
        0x9B4E: "HANJA",
        0x9B4F: "HANJA",
        0x9B50: "HANJA",
        0x9B51: "HANJA",
        0x9B52: "HANJA",
        0x9B53: "HANJA",
        0x9B54: "HANJA",
        0x9B55: "HANJA",
        0x9B56: "HANJA",
        0x9B57: "HANJA",
        0x9B58: "HANJA",
        0x9B59: "HANJA",
        0x9B5A: "HANJA",
        0x9B5B: "HANJA",
        0x9B5C: "HANJA",
        0x9B5D: "HANJA",
        0x9B5E: "HANJA",
        0x9B5F: "HANJA",
        0x9B60: "HANJA",
        0x9B61: "HANJA",
        0x9B62: "HANJA",
        0x9B63: "HANJA",
        0x9B64: "HANJA",
        0x9B65: "HANJA",
        0x9B66: "HANJA",
        0x9B67: "HANJA",
        0x9B68: "HANJA",
        0x9B69: "HANJA",
        0x9B6A: "HANJA",
        0x9B6B: "HANJA",
        0x9B6C: "HANJA",
        0x9B6D: "HANJA",
        0x9B6E: "HANJA",
        0x9B6F: "HANJA",
        0x9B70: "HANJA",
        0x9B71: "HANJA",
        0x9B72: "HANJA",
        0x9B73: "HANJA",
        0x9B74: "HANJA",
        0x9B75: "HANJA",
        0x9B76: "HANJA",
        0x9B77: "HANJA",
        0x9B78: "HANJA",
        0x9B79: "HANJA",
        0x9B7A: "HANJA",
        0x9B7B: "HANJA",
        0x9B7C: "HANJA",
        0x9B7D: "HANJA",
        0x9B7E: "HANJA",
        0x9B7F: "HANJA",
        0x9B80: "HANJA",
        0x9B81: "HANJA",
        0x9B82: "HANJA",
        0x9B83: "HANJA",
        0x9B84: "HANJA",
        0x9B85: "HANJA",
        0x9B86: "HANJA",
        0x9B87: "HANJA",
        0x9B88: "HANJA",
        0x9B89: "HANJA",
        0x9B8A: "HANJA",
        0x9B8B: "HANJA",
        0x9B8C: "HANJA",
        0x9B8D: "HANJA",
        0x9B8E: "HANJA",
        0x9B8F: "HANJA",
        0x9B90: "HANJA",
        0x9B91: "HANJA",
        0x9B92: "HANJA",
        0x9B93: "HANJA",
        0x9B94: "HANJA",
        0x9B95: "HANJA",
        0x9B96: "HANJA",
        0x9B97: "HANJA",
        0x9B98: "HANJA",
        0x9B99: "HANJA",
        0x9B9A: "HANJA",
        0x9B9B: "HANJA",
        0x9B9C: "HANJA",
        0x9B9D: "HANJA",
        0x9B9E: "HANJA",
        0x9B9F: "HANJA",
        0x9BA0: "HANJA",
        0x9BA1: "HANJA",
        0x9BA2: "HANJA",
        0x9BA3: "HANJA",
        0x9BA4: "HANJA",
        0x9BA5: "HANJA",
        0x9BA6: "HANJA",
        0x9BA7: "HANJA",
        0x9BA8: "HANJA",
        0x9BA9: "HANJA",
        0x9BAA: "HANJA",
        0x9BAB: "HANJA",
        0x9BAC: "HANJA",
        0x9BAD: "HANJA",
        0x9BAE: "HANJA",
        0x9BAF: "HANJA",
        0x9BB0: "HANJA",
        0x9BB1: "HANJA",
        0x9BB2: "HANJA",
        0x9BB3: "HANJA",
        0x9BB4: "HANJA",
        0x9BB5: "HANJA",
        0x9BB6: "HANJA",
        0x9BB7: "HANJA",
        0x9BB8: "HANJA",
        0x9BB9: "HANJA",
        0x9BBA: "HANJA",
        0x9BBB: "HANJA",
        0x9BBC: "HANJA",
        0x9BBD: "HANJA",
        0x9BBE: "HANJA",
        0x9BBF: "HANJA",
        0x9BC0: "HANJA",
        0x9BC1: "HANJA",
        0x9BC2: "HANJA",
        0x9BC3: "HANJA",
        0x9BC4: "HANJA",
        0x9BC5: "HANJA",
        0x9BC6: "HANJA",
        0x9BC7: "HANJA",
        0x9BC8: "HANJA",
        0x9BC9: "HANJA",
        0x9BCA: "HANJA",
        0x9BCB: "HANJA",
        0x9BCC: "HANJA",
        0x9BCD: "HANJA",
        0x9BCE: "HANJA",
        0x9BCF: "HANJA",
        0x9BD0: "HANJA",
        0x9BD1: "HANJA",
        0x9BD2: "HANJA",
        0x9BD3: "HANJA",
        0x9BD4: "HANJA",
        0x9BD5: "HANJA",
        0x9BD6: "HANJA",
        0x9BD7: "HANJA",
        0x9BD8: "HANJA",
        0x9BD9: "HANJA",
        0x9BDA: "HANJA",
        0x9BDB: "HANJA",
        0x9BDC: "HANJA",
        0x9BDD: "HANJA",
        0x9BDE: "HANJA",
        0x9BDF: "HANJA",
        0x9BE0: "HANJA",
        0x9BE1: "HANJA",
        0x9BE2: "HANJA",
        0x9BE3: "HANJA",
        0x9BE4: "HANJA",
        0x9BE5: "HANJA",
        0x9BE6: "HANJA",
        0x9BE7: "HANJA",
        0x9BE8: "HANJA",
        0x9BE9: "HANJA",
        0x9BEA: "HANJA",
        0x9BEB: "HANJA",
        0x9BEC: "HANJA",
        0x9BED: "HANJA",
        0x9BEE: "HANJA",
        0x9BEF: "HANJA",
        0x9BF0: "HANJA",
        0x9BF1: "HANJA",
        0x9BF2: "HANJA",
        0x9BF3: "HANJA",
        0x9BF4: "HANJA",
        0x9BF5: "HANJA",
        0x9BF6: "HANJA",
        0x9BF7: "HANJA",
        0x9BF8: "HANJA",
        0x9BF9: "HANJA",
        0x9BFA: "HANJA",
        0x9BFB: "HANJA",
        0x9BFC: "HANJA",
        0x9BFD: "HANJA",
        0x9BFE: "HANJA",
        0x9BFF: "HANJA",
        0x9C00: "HANJA",
        0x9C01: "HANJA",
        0x9C02: "HANJA",
        0x9C03: "HANJA",
        0x9C04: "HANJA",
        0x9C05: "HANJA",
        0x9C06: "HANJA",
        0x9C07: "HANJA",
        0x9C08: "HANJA",
        0x9C09: "HANJA",
        0x9C0A: "HANJA",
        0x9C0B: "HANJA",
        0x9C0C: "HANJA",
        0x9C0D: "HANJA",
        0x9C0E: "HANJA",
        0x9C0F: "HANJA",
        0x9C10: "HANJA",
        0x9C11: "HANJA",
        0x9C12: "HANJA",
        0x9C13: "HANJA",
        0x9C14: "HANJA",
        0x9C15: "HANJA",
        0x9C16: "HANJA",
        0x9C17: "HANJA",
        0x9C18: "HANJA",
        0x9C19: "HANJA",
        0x9C1A: "HANJA",
        0x9C1B: "HANJA",
        0x9C1C: "HANJA",
        0x9C1D: "HANJA",
        0x9C1E: "HANJA",
        0x9C1F: "HANJA",
        0x9C20: "HANJA",
        0x9C21: "HANJA",
        0x9C22: "HANJA",
        0x9C23: "HANJA",
        0x9C24: "HANJA",
        0x9C25: "HANJA",
        0x9C26: "HANJA",
        0x9C27: "HANJA",
        0x9C28: "HANJA",
        0x9C29: "HANJA",
        0x9C2A: "HANJA",
        0x9C2B: "HANJA",
        0x9C2C: "HANJA",
        0x9C2D: "HANJA",
        0x9C2E: "HANJA",
        0x9C2F: "HANJA",
        0x9C30: "HANJA",
        0x9C31: "HANJA",
        0x9C32: "HANJA",
        0x9C33: "HANJA",
        0x9C34: "HANJA",
        0x9C35: "HANJA",
        0x9C36: "HANJA",
        0x9C37: "HANJA",
        0x9C38: "HANJA",
        0x9C39: "HANJA",
        0x9C3A: "HANJA",
        0x9C3B: "HANJA",
        0x9C3C: "HANJA",
        0x9C3D: "HANJA",
        0x9C3E: "HANJA",
        0x9C3F: "HANJA",
        0x9C40: "HANJA",
        0x9C41: "HANJA",
        0x9C42: "HANJA",
        0x9C43: "HANJA",
        0x9C44: "HANJA",
        0x9C45: "HANJA",
        0x9C46: "HANJA",
        0x9C47: "HANJA",
        0x9C48: "HANJA",
        0x9C49: "HANJA",
        0x9C4A: "HANJA",
        0x9C4B: "HANJA",
        0x9C4C: "HANJA",
        0x9C4D: "HANJA",
        0x9C4E: "HANJA",
        0x9C4F: "HANJA",
        0x9C50: "HANJA",
        0x9C51: "HANJA",
        0x9C52: "HANJA",
        0x9C53: "HANJA",
        0x9C54: "HANJA",
        0x9C55: "HANJA",
        0x9C56: "HANJA",
        0x9C57: "HANJA",
        0x9C58: "HANJA",
        0x9C59: "HANJA",
        0x9C5A: "HANJA",
        0x9C5B: "HANJA",
        0x9C5C: "HANJA",
        0x9C5D: "HANJA",
        0x9C5E: "HANJA",
        0x9C5F: "HANJA",
        0x9C60: "HANJA",
        0x9C61: "HANJA",
        0x9C62: "HANJA",
        0x9C63: "HANJA",
        0x9C64: "HANJA",
        0x9C65: "HANJA",
        0x9C66: "HANJA",
        0x9C67: "HANJA",
        0x9C68: "HANJA",
        0x9C69: "HANJA",
        0x9C6A: "HANJA",
        0x9C6B: "HANJA",
        0x9C6C: "HANJA",
        0x9C6D: "HANJA",
        0x9C6E: "HANJA",
        0x9C6F: "HANJA",
        0x9C70: "HANJA",
        0x9C71: "HANJA",
        0x9C72: "HANJA",
        0x9C73: "HANJA",
        0x9C74: "HANJA",
        0x9C75: "HANJA",
        0x9C76: "HANJA",
        0x9C77: "HANJA",
        0x9C78: "HANJA",
        0x9C79: "HANJA",
        0x9C7A: "HANJA",
        0x9C7B: "HANJA",
        0x9C7C: "HANJA",
        0x9C7D: "HANJA",
        0x9C7E: "HANJA",
        0x9C7F: "HANJA",
        0x9C80: "HANJA",
        0x9C81: "HANJA",
        0x9C82: "HANJA",
        0x9C83: "HANJA",
        0x9C84: "HANJA",
        0x9C85: "HANJA",
        0x9C86: "HANJA",
        0x9C87: "HANJA",
        0x9C88: "HANJA",
        0x9C89: "HANJA",
        0x9C8A: "HANJA",
        0x9C8B: "HANJA",
        0x9C8C: "HANJA",
        0x9C8D: "HANJA",
        0x9C8E: "HANJA",
        0x9C8F: "HANJA",
        0x9C90: "HANJA",
        0x9C91: "HANJA",
        0x9C92: "HANJA",
        0x9C93: "HANJA",
        0x9C94: "HANJA",
        0x9C95: "HANJA",
        0x9C96: "HANJA",
        0x9C97: "HANJA",
        0x9C98: "HANJA",
        0x9C99: "HANJA",
        0x9C9A: "HANJA",
        0x9C9B: "HANJA",
        0x9C9C: "HANJA",
        0x9C9D: "HANJA",
        0x9C9E: "HANJA",
        0x9C9F: "HANJA",
        0x9CA0: "HANJA",
        0x9CA1: "HANJA",
        0x9CA2: "HANJA",
        0x9CA3: "HANJA",
        0x9CA4: "HANJA",
        0x9CA5: "HANJA",
        0x9CA6: "HANJA",
        0x9CA7: "HANJA",
        0x9CA8: "HANJA",
        0x9CA9: "HANJA",
        0x9CAA: "HANJA",
        0x9CAB: "HANJA",
        0x9CAC: "HANJA",
        0x9CAD: "HANJA",
        0x9CAE: "HANJA",
        0x9CAF: "HANJA",
        0x9CB0: "HANJA",
        0x9CB1: "HANJA",
        0x9CB2: "HANJA",
        0x9CB3: "HANJA",
        0x9CB4: "HANJA",
        0x9CB5: "HANJA",
        0x9CB6: "HANJA",
        0x9CB7: "HANJA",
        0x9CB8: "HANJA",
        0x9CB9: "HANJA",
        0x9CBA: "HANJA",
        0x9CBB: "HANJA",
        0x9CBC: "HANJA",
        0x9CBD: "HANJA",
        0x9CBE: "HANJA",
        0x9CBF: "HANJA",
        0x9CC0: "HANJA",
        0x9CC1: "HANJA",
        0x9CC2: "HANJA",
        0x9CC3: "HANJA",
        0x9CC4: "HANJA",
        0x9CC5: "HANJA",
        0x9CC6: "HANJA",
        0x9CC7: "HANJA",
        0x9CC8: "HANJA",
        0x9CC9: "HANJA",
        0x9CCA: "HANJA",
        0x9CCB: "HANJA",
        0x9CCC: "HANJA",
        0x9CCD: "HANJA",
        0x9CCE: "HANJA",
        0x9CCF: "HANJA",
        0x9CD0: "HANJA",
        0x9CD1: "HANJA",
        0x9CD2: "HANJA",
        0x9CD3: "HANJA",
        0x9CD4: "HANJA",
        0x9CD5: "HANJA",
        0x9CD6: "HANJA",
        0x9CD7: "HANJA",
        0x9CD8: "HANJA",
        0x9CD9: "HANJA",
        0x9CDA: "HANJA",
        0x9CDB: "HANJA",
        0x9CDC: "HANJA",
        0x9CDD: "HANJA",
        0x9CDE: "HANJA",
        0x9CDF: "HANJA",
        0x9CE0: "HANJA",
        0x9CE1: "HANJA",
        0x9CE2: "HANJA",
        0x9CE3: "HANJA",
        0x9CE4: "HANJA",
        0x9CE5: "HANJA",
        0x9CE6: "HANJA",
        0x9CE7: "HANJA",
        0x9CE8: "HANJA",
        0x9CE9: "HANJA",
        0x9CEA: "HANJA",
        0x9CEB: "HANJA",
        0x9CEC: "HANJA",
        0x9CED: "HANJA",
        0x9CEE: "HANJA",
        0x9CEF: "HANJA",
        0x9CF0: "HANJA",
        0x9CF1: "HANJA",
        0x9CF2: "HANJA",
        0x9CF3: "HANJA",
        0x9CF4: "HANJA",
        0x9CF5: "HANJA",
        0x9CF6: "HANJA",
        0x9CF7: "HANJA",
        0x9CF8: "HANJA",
        0x9CF9: "HANJA",
        0x9CFA: "HANJA",
        0x9CFB: "HANJA",
        0x9CFC: "HANJA",
        0x9CFD: "HANJA",
        0x9CFE: "HANJA",
        0x9CFF: "HANJA",
        0x9D00: "HANJA",
        0x9D01: "HANJA",
        0x9D02: "HANJA",
        0x9D03: "HANJA",
        0x9D04: "HANJA",
        0x9D05: "HANJA",
        0x9D06: "HANJA",
        0x9D07: "HANJA",
        0x9D08: "HANJA",
        0x9D09: "HANJA",
        0x9D0A: "HANJA",
        0x9D0B: "HANJA",
        0x9D0C: "HANJA",
        0x9D0D: "HANJA",
        0x9D0E: "HANJA",
        0x9D0F: "HANJA",
        0x9D10: "HANJA",
        0x9D11: "HANJA",
        0x9D12: "HANJA",
        0x9D13: "HANJA",
        0x9D14: "HANJA",
        0x9D15: "HANJA",
        0x9D16: "HANJA",
        0x9D17: "HANJA",
        0x9D18: "HANJA",
        0x9D19: "HANJA",
        0x9D1A: "HANJA",
        0x9D1B: "HANJA",
        0x9D1C: "HANJA",
        0x9D1D: "HANJA",
        0x9D1E: "HANJA",
        0x9D1F: "HANJA",
        0x9D20: "HANJA",
        0x9D21: "HANJA",
        0x9D22: "HANJA",
        0x9D23: "HANJA",
        0x9D24: "HANJA",
        0x9D25: "HANJA",
        0x9D26: "HANJA",
        0x9D27: "HANJA",
        0x9D28: "HANJA",
        0x9D29: "HANJA",
        0x9D2A: "HANJA",
        0x9D2B: "HANJA",
        0x9D2C: "HANJA",
        0x9D2D: "HANJA",
        0x9D2E: "HANJA",
        0x9D2F: "HANJA",
        0x9D30: "HANJA",
        0x9D31: "HANJA",
        0x9D32: "HANJA",
        0x9D33: "HANJA",
        0x9D34: "HANJA",
        0x9D35: "HANJA",
        0x9D36: "HANJA",
        0x9D37: "HANJA",
        0x9D38: "HANJA",
        0x9D39: "HANJA",
        0x9D3A: "HANJA",
        0x9D3B: "HANJA",
        0x9D3C: "HANJA",
        0x9D3D: "HANJA",
        0x9D3E: "HANJA",
        0x9D3F: "HANJA",
        0x9D40: "HANJA",
        0x9D41: "HANJA",
        0x9D42: "HANJA",
        0x9D43: "HANJA",
        0x9D44: "HANJA",
        0x9D45: "HANJA",
        0x9D46: "HANJA",
        0x9D47: "HANJA",
        0x9D48: "HANJA",
        0x9D49: "HANJA",
        0x9D4A: "HANJA",
        0x9D4B: "HANJA",
        0x9D4C: "HANJA",
        0x9D4D: "HANJA",
        0x9D4E: "HANJA",
        0x9D4F: "HANJA",
        0x9D50: "HANJA",
        0x9D51: "HANJA",
        0x9D52: "HANJA",
        0x9D53: "HANJA",
        0x9D54: "HANJA",
        0x9D55: "HANJA",
        0x9D56: "HANJA",
        0x9D57: "HANJA",
        0x9D58: "HANJA",
        0x9D59: "HANJA",
        0x9D5A: "HANJA",
        0x9D5B: "HANJA",
        0x9D5C: "HANJA",
        0x9D5D: "HANJA",
        0x9D5E: "HANJA",
        0x9D5F: "HANJA",
        0x9D60: "HANJA",
        0x9D61: "HANJA",
        0x9D62: "HANJA",
        0x9D63: "HANJA",
        0x9D64: "HANJA",
        0x9D65: "HANJA",
        0x9D66: "HANJA",
        0x9D67: "HANJA",
        0x9D68: "HANJA",
        0x9D69: "HANJA",
        0x9D6A: "HANJA",
        0x9D6B: "HANJA",
        0x9D6C: "HANJA",
        0x9D6D: "HANJA",
        0x9D6E: "HANJA",
        0x9D6F: "HANJA",
        0x9D70: "HANJA",
        0x9D71: "HANJA",
        0x9D72: "HANJA",
        0x9D73: "HANJA",
        0x9D74: "HANJA",
        0x9D75: "HANJA",
        0x9D76: "HANJA",
        0x9D77: "HANJA",
        0x9D78: "HANJA",
        0x9D79: "HANJA",
        0x9D7A: "HANJA",
        0x9D7B: "HANJA",
        0x9D7C: "HANJA",
        0x9D7D: "HANJA",
        0x9D7E: "HANJA",
        0x9D7F: "HANJA",
        0x9D80: "HANJA",
        0x9D81: "HANJA",
        0x9D82: "HANJA",
        0x9D83: "HANJA",
        0x9D84: "HANJA",
        0x9D85: "HANJA",
        0x9D86: "HANJA",
        0x9D87: "HANJA",
        0x9D88: "HANJA",
        0x9D89: "HANJA",
        0x9D8A: "HANJA",
        0x9D8B: "HANJA",
        0x9D8C: "HANJA",
        0x9D8D: "HANJA",
        0x9D8E: "HANJA",
        0x9D8F: "HANJA",
        0x9D90: "HANJA",
        0x9D91: "HANJA",
        0x9D92: "HANJA",
        0x9D93: "HANJA",
        0x9D94: "HANJA",
        0x9D95: "HANJA",
        0x9D96: "HANJA",
        0x9D97: "HANJA",
        0x9D98: "HANJA",
        0x9D99: "HANJA",
        0x9D9A: "HANJA",
        0x9D9B: "HANJA",
        0x9D9C: "HANJA",
        0x9D9D: "HANJA",
        0x9D9E: "HANJA",
        0x9D9F: "HANJA",
        0x9DA0: "HANJA",
        0x9DA1: "HANJA",
        0x9DA2: "HANJA",
        0x9DA3: "HANJA",
        0x9DA4: "HANJA",
        0x9DA5: "HANJA",
        0x9DA6: "HANJA",
        0x9DA7: "HANJA",
        0x9DA8: "HANJA",
        0x9DA9: "HANJA",
        0x9DAA: "HANJA",
        0x9DAB: "HANJA",
        0x9DAC: "HANJA",
        0x9DAD: "HANJA",
        0x9DAE: "HANJA",
        0x9DAF: "HANJA",
        0x9DB0: "HANJA",
        0x9DB1: "HANJA",
        0x9DB2: "HANJA",
        0x9DB3: "HANJA",
        0x9DB4: "HANJA",
        0x9DB5: "HANJA",
        0x9DB6: "HANJA",
        0x9DB7: "HANJA",
        0x9DB8: "HANJA",
        0x9DB9: "HANJA",
        0x9DBA: "HANJA",
        0x9DBB: "HANJA",
        0x9DBC: "HANJA",
        0x9DBD: "HANJA",
        0x9DBE: "HANJA",
        0x9DBF: "HANJA",
        0x9DC0: "HANJA",
        0x9DC1: "HANJA",
        0x9DC2: "HANJA",
        0x9DC3: "HANJA",
        0x9DC4: "HANJA",
        0x9DC5: "HANJA",
        0x9DC6: "HANJA",
        0x9DC7: "HANJA",
        0x9DC8: "HANJA",
        0x9DC9: "HANJA",
        0x9DCA: "HANJA",
        0x9DCB: "HANJA",
        0x9DCC: "HANJA",
        0x9DCD: "HANJA",
        0x9DCE: "HANJA",
        0x9DCF: "HANJA",
        0x9DD0: "HANJA",
        0x9DD1: "HANJA",
        0x9DD2: "HANJA",
        0x9DD3: "HANJA",
        0x9DD4: "HANJA",
        0x9DD5: "HANJA",
        0x9DD6: "HANJA",
        0x9DD7: "HANJA",
        0x9DD8: "HANJA",
        0x9DD9: "HANJA",
        0x9DDA: "HANJA",
        0x9DDB: "HANJA",
        0x9DDC: "HANJA",
        0x9DDD: "HANJA",
        0x9DDE: "HANJA",
        0x9DDF: "HANJA",
        0x9DE0: "HANJA",
        0x9DE1: "HANJA",
        0x9DE2: "HANJA",
        0x9DE3: "HANJA",
        0x9DE4: "HANJA",
        0x9DE5: "HANJA",
        0x9DE6: "HANJA",
        0x9DE7: "HANJA",
        0x9DE8: "HANJA",
        0x9DE9: "HANJA",
        0x9DEA: "HANJA",
        0x9DEB: "HANJA",
        0x9DEC: "HANJA",
        0x9DED: "HANJA",
        0x9DEE: "HANJA",
        0x9DEF: "HANJA",
        0x9DF0: "HANJA",
        0x9DF1: "HANJA",
        0x9DF2: "HANJA",
        0x9DF3: "HANJA",
        0x9DF4: "HANJA",
        0x9DF5: "HANJA",
        0x9DF6: "HANJA",
        0x9DF7: "HANJA",
        0x9DF8: "HANJA",
        0x9DF9: "HANJA",
        0x9DFA: "HANJA",
        0x9DFB: "HANJA",
        0x9DFC: "HANJA",
        0x9DFD: "HANJA",
        0x9DFE: "HANJA",
        0x9DFF: "HANJA",
        0x9E00: "HANJA",
        0x9E01: "HANJA",
        0x9E02: "HANJA",
        0x9E03: "HANJA",
        0x9E04: "HANJA",
        0x9E05: "HANJA",
        0x9E06: "HANJA",
        0x9E07: "HANJA",
        0x9E08: "HANJA",
        0x9E09: "HANJA",
        0x9E0A: "HANJA",
        0x9E0B: "HANJA",
        0x9E0C: "HANJA",
        0x9E0D: "HANJA",
        0x9E0E: "HANJA",
        0x9E0F: "HANJA",
        0x9E10: "HANJA",
        0x9E11: "HANJA",
        0x9E12: "HANJA",
        0x9E13: "HANJA",
        0x9E14: "HANJA",
        0x9E15: "HANJA",
        0x9E16: "HANJA",
        0x9E17: "HANJA",
        0x9E18: "HANJA",
        0x9E19: "HANJA",
        0x9E1A: "HANJA",
        0x9E1B: "HANJA",
        0x9E1C: "HANJA",
        0x9E1D: "HANJA",
        0x9E1E: "HANJA",
        0x9E1F: "HANJA",
        0x9E20: "HANJA",
        0x9E21: "HANJA",
        0x9E22: "HANJA",
        0x9E23: "HANJA",
        0x9E24: "HANJA",
        0x9E25: "HANJA",
        0x9E26: "HANJA",
        0x9E27: "HANJA",
        0x9E28: "HANJA",
        0x9E29: "HANJA",
        0x9E2A: "HANJA",
        0x9E2B: "HANJA",
        0x9E2C: "HANJA",
        0x9E2D: "HANJA",
        0x9E2E: "HANJA",
        0x9E2F: "HANJA",
        0x9E30: "HANJA",
        0x9E31: "HANJA",
        0x9E32: "HANJA",
        0x9E33: "HANJA",
        0x9E34: "HANJA",
        0x9E35: "HANJA",
        0x9E36: "HANJA",
        0x9E37: "HANJA",
        0x9E38: "HANJA",
        0x9E39: "HANJA",
        0x9E3A: "HANJA",
        0x9E3B: "HANJA",
        0x9E3C: "HANJA",
        0x9E3D: "HANJA",
        0x9E3E: "HANJA",
        0x9E3F: "HANJA",
        0x9E40: "HANJA",
        0x9E41: "HANJA",
        0x9E42: "HANJA",
        0x9E43: "HANJA",
        0x9E44: "HANJA",
        0x9E45: "HANJA",
        0x9E46: "HANJA",
        0x9E47: "HANJA",
        0x9E48: "HANJA",
        0x9E49: "HANJA",
        0x9E4A: "HANJA",
        0x9E4B: "HANJA",
        0x9E4C: "HANJA",
        0x9E4D: "HANJA",
        0x9E4E: "HANJA",
        0x9E4F: "HANJA",
        0x9E50: "HANJA",
        0x9E51: "HANJA",
        0x9E52: "HANJA",
        0x9E53: "HANJA",
        0x9E54: "HANJA",
        0x9E55: "HANJA",
        0x9E56: "HANJA",
        0x9E57: "HANJA",
        0x9E58: "HANJA",
        0x9E59: "HANJA",
        0x9E5A: "HANJA",
        0x9E5B: "HANJA",
        0x9E5C: "HANJA",
        0x9E5D: "HANJA",
        0x9E5E: "HANJA",
        0x9E5F: "HANJA",
        0x9E60: "HANJA",
        0x9E61: "HANJA",
        0x9E62: "HANJA",
        0x9E63: "HANJA",
        0x9E64: "HANJA",
        0x9E65: "HANJA",
        0x9E66: "HANJA",
        0x9E67: "HANJA",
        0x9E68: "HANJA",
        0x9E69: "HANJA",
        0x9E6A: "HANJA",
        0x9E6B: "HANJA",
        0x9E6C: "HANJA",
        0x9E6D: "HANJA",
        0x9E6E: "HANJA",
        0x9E6F: "HANJA",
        0x9E70: "HANJA",
        0x9E71: "HANJA",
        0x9E72: "HANJA",
        0x9E73: "HANJA",
        0x9E74: "HANJA",
        0x9E75: "HANJA",
        0x9E76: "HANJA",
        0x9E77: "HANJA",
        0x9E78: "HANJA",
        0x9E79: "HANJA",
        0x9E7A: "HANJA",
        0x9E7B: "HANJA",
        0x9E7C: "HANJA",
        0x9E7D: "HANJA",
        0x9E7E: "HANJA",
        0x9E7F: "HANJA",
        0x9E80: "HANJA",
        0x9E81: "HANJA",
        0x9E82: "HANJA",
        0x9E83: "HANJA",
        0x9E84: "HANJA",
        0x9E85: "HANJA",
        0x9E86: "HANJA",
        0x9E87: "HANJA",
        0x9E88: "HANJA",
        0x9E89: "HANJA",
        0x9E8A: "HANJA",
        0x9E8B: "HANJA",
        0x9E8C: "HANJA",
        0x9E8D: "HANJA",
        0x9E8E: "HANJA",
        0x9E8F: "HANJA",
        0x9E90: "HANJA",
        0x9E91: "HANJA",
        0x9E92: "HANJA",
        0x9E93: "HANJA",
        0x9E94: "HANJA",
        0x9E95: "HANJA",
        0x9E96: "HANJA",
        0x9E97: "HANJA",
        0x9E98: "HANJA",
        0x9E99: "HANJA",
        0x9E9A: "HANJA",
        0x9E9B: "HANJA",
        0x9E9C: "HANJA",
        0x9E9D: "HANJA",
        0x9E9E: "HANJA",
        0x9E9F: "HANJA",
        0x9EA0: "HANJA",
        0x9EA1: "HANJA",
        0x9EA2: "HANJA",
        0x9EA3: "HANJA",
        0x9EA4: "HANJA",
        0x9EA5: "HANJA",
        0x9EA6: "HANJA",
        0x9EA7: "HANJA",
        0x9EA8: "HANJA",
        0x9EA9: "HANJA",
        0x9EAA: "HANJA",
        0x9EAB: "HANJA",
        0x9EAC: "HANJA",
        0x9EAD: "HANJA",
        0x9EAE: "HANJA",
        0x9EAF: "HANJA",
        0x9EB0: "HANJA",
        0x9EB1: "HANJA",
        0x9EB2: "HANJA",
        0x9EB3: "HANJA",
        0x9EB4: "HANJA",
        0x9EB5: "HANJA",
        0x9EB6: "HANJA",
        0x9EB7: "HANJA",
        0x9EB8: "HANJA",
        0x9EB9: "HANJA",
        0x9EBA: "HANJA",
        0x9EBB: "HANJA",
        0x9EBC: "HANJA",
        0x9EBD: "HANJA",
        0x9EBE: "HANJA",
        0x9EBF: "HANJA",
        0x9EC0: "HANJA",
        0x9EC1: "HANJA",
        0x9EC2: "HANJA",
        0x9EC3: "HANJA",
        0x9EC4: "HANJA",
        0x9EC5: "HANJA",
        0x9EC6: "HANJA",
        0x9EC7: "HANJA",
        0x9EC8: "HANJA",
        0x9EC9: "HANJA",
        0x9ECA: "HANJA",
        0x9ECB: "HANJA",
        0x9ECC: "HANJA",
        0x9ECD: "HANJA",
        0x9ECE: "HANJA",
        0x9ECF: "HANJA",
        0x9ED0: "HANJA",
        0x9ED1: "HANJA",
        0x9ED2: "HANJA",
        0x9ED3: "HANJA",
        0x9ED4: "HANJA",
        0x9ED5: "HANJA",
        0x9ED6: "HANJA",
        0x9ED7: "HANJA",
        0x9ED8: "HANJA",
        0x9ED9: "HANJA",
        0x9EDA: "HANJA",
        0x9EDB: "HANJA",
        0x9EDC: "HANJA",
        0x9EDD: "HANJA",
        0x9EDE: "HANJA",
        0x9EDF: "HANJA",
        0x9EE0: "HANJA",
        0x9EE1: "HANJA",
        0x9EE2: "HANJA",
        0x9EE3: "HANJA",
        0x9EE4: "HANJA",
        0x9EE5: "HANJA",
        0x9EE6: "HANJA",
        0x9EE7: "HANJA",
        0x9EE8: "HANJA",
        0x9EE9: "HANJA",
        0x9EEA: "HANJA",
        0x9EEB: "HANJA",
        0x9EEC: "HANJA",
        0x9EED: "HANJA",
        0x9EEE: "HANJA",
        0x9EEF: "HANJA",
        0x9EF0: "HANJA",
        0x9EF1: "HANJA",
        0x9EF2: "HANJA",
        0x9EF3: "HANJA",
        0x9EF4: "HANJA",
        0x9EF5: "HANJA",
        0x9EF6: "HANJA",
        0x9EF7: "HANJA",
        0x9EF8: "HANJA",
        0x9EF9: "HANJA",
        0x9EFA: "HANJA",
        0x9EFB: "HANJA",
        0x9EFC: "HANJA",
        0x9EFD: "HANJA",
        0x9EFE: "HANJA",
        0x9EFF: "HANJA",
        0x9F00: "HANJA",
        0x9F01: "HANJA",
        0x9F02: "HANJA",
        0x9F03: "HANJA",
        0x9F04: "HANJA",
        0x9F05: "HANJA",
        0x9F06: "HANJA",
        0x9F07: "HANJA",
        0x9F08: "HANJA",
        0x9F09: "HANJA",
        0x9F0A: "HANJA",
        0x9F0B: "HANJA",
        0x9F0C: "HANJA",
        0x9F0D: "HANJA",
        0x9F0E: "HANJA",
        0x9F0F: "HANJA",
        0x9F10: "HANJA",
        0x9F11: "HANJA",
        0x9F12: "HANJA",
        0x9F13: "HANJA",
        0x9F14: "HANJA",
        0x9F15: "HANJA",
        0x9F16: "HANJA",
        0x9F17: "HANJA",
        0x9F18: "HANJA",
        0x9F19: "HANJA",
        0x9F1A: "HANJA",
        0x9F1B: "HANJA",
        0x9F1C: "HANJA",
        0x9F1D: "HANJA",
        0x9F1E: "HANJA",
        0x9F1F: "HANJA",
        0x9F20: "HANJA",
        0x9F21: "HANJA",
        0x9F22: "HANJA",
        0x9F23: "HANJA",
        0x9F24: "HANJA",
        0x9F25: "HANJA",
        0x9F26: "HANJA",
        0x9F27: "HANJA",
        0x9F28: "HANJA",
        0x9F29: "HANJA",
        0x9F2A: "HANJA",
        0x9F2B: "HANJA",
        0x9F2C: "HANJA",
        0x9F2D: "HANJA",
        0x9F2E: "HANJA",
        0x9F2F: "HANJA",
        0x9F30: "HANJA",
        0x9F31: "HANJA",
        0x9F32: "HANJA",
        0x9F33: "HANJA",
        0x9F34: "HANJA",
        0x9F35: "HANJA",
        0x9F36: "HANJA",
        0x9F37: "HANJA",
        0x9F38: "HANJA",
        0x9F39: "HANJA",
        0x9F3A: "HANJA",
        0x9F3B: "HANJA",
        0x9F3C: "HANJA",
        0x9F3D: "HANJA",
        0x9F3E: "HANJA",
        0x9F3F: "HANJA",
        0x9F40: "HANJA",
        0x9F41: "HANJA",
        0x9F42: "HANJA",
        0x9F43: "HANJA",
        0x9F44: "HANJA",
        0x9F45: "HANJA",
        0x9F46: "HANJA",
        0x9F47: "HANJA",
        0x9F48: "HANJA",
        0x9F49: "HANJA",
        0x9F4A: "HANJA",
        0x9F4B: "HANJA",
        0x9F4C: "HANJA",
        0x9F4D: "HANJA",
        0x9F4E: "HANJA",
        0x9F4F: "HANJA",
        0x9F50: "HANJA",
        0x9F51: "HANJA",
        0x9F52: "HANJA",
        0x9F53: "HANJA",
        0x9F54: "HANJA",
        0x9F55: "HANJA",
        0x9F56: "HANJA",
        0x9F57: "HANJA",
        0x9F58: "HANJA",
        0x9F59: "HANJA",
        0x9F5A: "HANJA",
        0x9F5B: "HANJA",
        0x9F5C: "HANJA",
        0x9F5D: "HANJA",
        0x9F5E: "HANJA",
        0x9F5F: "HANJA",
        0x9F60: "HANJA",
        0x9F61: "HANJA",
        0x9F62: "HANJA",
        0x9F63: "HANJA",
        0x9F64: "HANJA",
        0x9F65: "HANJA",
        0x9F66: "HANJA",
        0x9F67: "HANJA",
        0x9F68: "HANJA",
        0x9F69: "HANJA",
        0x9F6A: "HANJA",
        0x9F6B: "HANJA",
        0x9F6C: "HANJA",
        0x9F6D: "HANJA",
        0x9F6E: "HANJA",
        0x9F6F: "HANJA",
        0x9F70: "HANJA",
        0x9F71: "HANJA",
        0x9F72: "HANJA",
        0x9F73: "HANJA",
        0x9F74: "HANJA",
        0x9F75: "HANJA",
        0x9F76: "HANJA",
        0x9F77: "HANJA",
        0x9F78: "HANJA",
        0x9F79: "HANJA",
        0x9F7A: "HANJA",
        0x9F7B: "HANJA",
        0x9F7C: "HANJA",
        0x9F7D: "HANJA",
        0x9F7E: "HANJA",
        0x9F7F: "HANJA",
        0x9F80: "HANJA",
        0x9F81: "HANJA",
        0x9F82: "HANJA",
        0x9F83: "HANJA",
        0x9F84: "HANJA",
        0x9F85: "HANJA",
        0x9F86: "HANJA",
        0x9F87: "HANJA",
        0x9F88: "HANJA",
        0x9F89: "HANJA",
        0x9F8A: "HANJA",
        0x9F8B: "HANJA",
        0x9F8C: "HANJA",
        0x9F8D: "HANJA",
        0x9F8E: "HANJA",
        0x9F8F: "HANJA",
        0x9F90: "HANJA",
        0x9F91: "HANJA",
        0x9F92: "HANJA",
        0x9F93: "HANJA",
        0x9F94: "HANJA",
        0x9F95: "HANJA",
        0x9F96: "HANJA",
        0x9F97: "HANJA",
        0x9F98: "HANJA",
        0x9F99: "HANJA",
        0x9F9A: "HANJA",
        0x9F9B: "HANJA",
        0x9F9C: "HANJA",
        0x9F9D: "HANJA",
        0x9F9E: "HANJA",
        0x9F9F: "HANJA",
        0x9FA0: "HANJA",
        0x9FA1: "HANJA",
        0x9FA2: "HANJA",
        0x9FA3: "HANJA",
        0x9FA4: "HANJA",
        0x9FA5: "HANJA",
        0xF900: "HANJA",
        0xF901: "HANJA",
        0xF902: "HANJA",
        0xF903: "HANJA",
        0xF904: "HANJA",
        0xF905: "HANJA",
        0xF906: "HANJA",
        0xF907: "HANJA",
        0xF908: "HANJA",
        0xF909: "HANJA",
        0xF90A: "HANJA",
        0xF90B: "HANJA",
        0xF90C: "HANJA",
        0xF90D: "HANJA",
        0xF90E: "HANJA",
        0xF90F: "HANJA",
        0xF910: "HANJA",
        0xF911: "HANJA",
        0xF912: "HANJA",
        0xF913: "HANJA",
        0xF914: "HANJA",
        0xF915: "HANJA",
        0xF916: "HANJA",
        0xF917: "HANJA",
        0xF918: "HANJA",
        0xF919: "HANJA",
        0xF91A: "HANJA",
        0xF91B: "HANJA",
        0xF91C: "HANJA",
        0xF91D: "HANJA",
        0xF91E: "HANJA",
        0xF91F: "HANJA",
        0xF920: "HANJA",
        0xF921: "HANJA",
        0xF922: "HANJA",
        0xF923: "HANJA",
        0xF924: "HANJA",
        0xF925: "HANJA",
        0xF926: "HANJA",
        0xF927: "HANJA",
        0xF928: "HANJA",
        0xF929: "HANJA",
        0xF92A: "HANJA",
        0xF92B: "HANJA",
        0xF92C: "HANJA",
        0xF92D: "HANJA",
        0xF92E: "HANJA",
        0xF92F: "HANJA",
        0xF930: "HANJA",
        0xF931: "HANJA",
        0xF932: "HANJA",
        0xF933: "HANJA",
        0xF934: "HANJA",
        0xF935: "HANJA",
        0xF936: "HANJA",
        0xF937: "HANJA",
        0xF938: "HANJA",
        0xF939: "HANJA",
        0xF93A: "HANJA",
        0xF93B: "HANJA",
        0xF93C: "HANJA",
        0xF93D: "HANJA",
        0xF93E: "HANJA",
        0xF93F: "HANJA",
        0xF940: "HANJA",
        0xF941: "HANJA",
        0xF942: "HANJA",
        0xF943: "HANJA",
        0xF944: "HANJA",
        0xF945: "HANJA",
        0xF946: "HANJA",
        0xF947: "HANJA",
        0xF948: "HANJA",
        0xF949: "HANJA",
        0xF94A: "HANJA",
        0xF94B: "HANJA",
        0xF94C: "HANJA",
        0xF94D: "HANJA",
        0xF94E: "HANJA",
        0xF94F: "HANJA",
        0xF950: "HANJA",
        0xF951: "HANJA",
        0xF952: "HANJA",
        0xF953: "HANJA",
        0xF954: "HANJA",
        0xF955: "HANJA",
        0xF956: "HANJA",
        0xF957: "HANJA",
        0xF958: "HANJA",
        0xF959: "HANJA",
        0xF95A: "HANJA",
        0xF95B: "HANJA",
        0xF95C: "HANJA",
        0xF95D: "HANJA",
        0xF95E: "HANJA",
        0xF95F: "HANJA",
        0xF960: "HANJA",
        0xF961: "HANJA",
        0xF962: "HANJA",
        0xF963: "HANJA",
        0xF964: "HANJA",
        0xF965: "HANJA",
        0xF966: "HANJA",
        0xF967: "HANJA",
        0xF968: "HANJA",
        0xF969: "HANJA",
        0xF96A: "HANJA",
        0xF96B: "HANJA",
        0xF96C: "HANJA",
        0xF96D: "HANJA",
        0xF96E: "HANJA",
        0xF96F: "HANJA",
        0xF970: "HANJA",
        0xF971: "HANJA",
        0xF972: "HANJA",
        0xF973: "HANJA",
        0xF974: "HANJA",
        0xF975: "HANJA",
        0xF976: "HANJA",
        0xF977: "HANJA",
        0xF978: "HANJA",
        0xF979: "HANJA",
        0xF97A: "HANJA",
        0xF97B: "HANJA",
        0xF97C: "HANJA",
        0xF97D: "HANJA",
        0xF97E: "HANJA",
        0xF97F: "HANJA",
        0xF980: "HANJA",
        0xF981: "HANJA",
        0xF982: "HANJA",
        0xF983: "HANJA",
        0xF984: "HANJA",
        0xF985: "HANJA",
        0xF986: "HANJA",
        0xF987: "HANJA",
        0xF988: "HANJA",
        0xF989: "HANJA",
        0xF98A: "HANJA",
        0xF98B: "HANJA",
        0xF98C: "HANJA",
        0xF98D: "HANJA",
        0xF98E: "HANJA",
        0xF98F: "HANJA",
        0xF990: "HANJA",
        0xF991: "HANJA",
        0xF992: "HANJA",
        0xF993: "HANJA",
        0xF994: "HANJA",
        0xF995: "HANJA",
        0xF996: "HANJA",
        0xF997: "HANJA",
        0xF998: "HANJA",
        0xF999: "HANJA",
        0xF99A: "HANJA",
        0xF99B: "HANJA",
        0xF99C: "HANJA",
        0xF99D: "HANJA",
        0xF99E: "HANJA",
        0xF99F: "HANJA",
        0xF9A0: "HANJA",
        0xF9A1: "HANJA",
        0xF9A2: "HANJA",
        0xF9A3: "HANJA",
        0xF9A4: "HANJA",
        0xF9A5: "HANJA",
        0xF9A6: "HANJA",
        0xF9A7: "HANJA",
        0xF9A8: "HANJA",
        0xF9A9: "HANJA",
        0xF9AA: "HANJA",
        0xF9AB: "HANJA",
        0xF9AC: "HANJA",
        0xF9AD: "HANJA",
        0xF9AE: "HANJA",
        0xF9AF: "HANJA",
        0xF9B0: "HANJA",
        0xF9B1: "HANJA",
        0xF9B2: "HANJA",
        0xF9B3: "HANJA",
        0xF9B4: "HANJA",
        0xF9B5: "HANJA",
        0xF9B6: "HANJA",
        0xF9B7: "HANJA",
        0xF9B8: "HANJA",
        0xF9B9: "HANJA",
        0xF9BA: "HANJA",
        0xF9BB: "HANJA",
        0xF9BC: "HANJA",
        0xF9BD: "HANJA",
        0xF9BE: "HANJA",
        0xF9BF: "HANJA",
        0xF9C0: "HANJA",
        0xF9C1: "HANJA",
        0xF9C2: "HANJA",
        0xF9C3: "HANJA",
        0xF9C4: "HANJA",
        0xF9C5: "HANJA",
        0xF9C6: "HANJA",
        0xF9C7: "HANJA",
        0xF9C8: "HANJA",
        0xF9C9: "HANJA",
        0xF9CA: "HANJA",
        0xF9CB: "HANJA",
        0xF9CC: "HANJA",
        0xF9CD: "HANJA",
        0xF9CE: "HANJA",
        0xF9CF: "HANJA",
        0xF9D0: "HANJA",
        0xF9D1: "HANJA",
        0xF9D2: "HANJA",
        0xF9D3: "HANJA",
        0xF9D4: "HANJA",
        0xF9D5: "HANJA",
        0xF9D6: "HANJA",
        0xF9D7: "HANJA",
        0xF9D8: "HANJA",
        0xF9D9: "HANJA",
        0xF9DA: "HANJA",
        0xF9DB: "HANJA",
        0xF9DC: "HANJA",
        0xF9DD: "HANJA",
        0xF9DE: "HANJA",
        0xF9DF: "HANJA",
        0xF9E0: "HANJA",
        0xF9E1: "HANJA",
        0xF9E2: "HANJA",
        0xF9E3: "HANJA",
        0xF9E4: "HANJA",
        0xF9E5: "HANJA",
        0xF9E6: "HANJA",
        0xF9E7: "HANJA",
        0xF9E8: "HANJA",
        0xF9E9: "HANJA",
        0xF9EA: "HANJA",
        0xF9EB: "HANJA",
        0xF9EC: "HANJA",
        0xF9ED: "HANJA",
        0xF9EE: "HANJA",
        0xF9EF: "HANJA",
        0xF9F0: "HANJA",
        0xF9F1: "HANJA",
        0xF9F2: "HANJA",
        0xF9F3: "HANJA",
        0xF9F4: "HANJA",
        0xF9F5: "HANJA",
        0xF9F6: "HANJA",
        0xF9F7: "HANJA",
        0xF9F8: "HANJA",
        0xF9F9: "HANJA",
        0xF9FA: "HANJA",
        0xF9FB: "HANJA",
        0xF9FC: "HANJA",
        0xF9FD: "HANJA",
        0xF9FE: "HANJA",
        0xF9FF: "HANJA",
        0xFA00: "HANJA",
        0xFA01: "HANJA",
        0xFA02: "HANJA",
        0xFA03: "HANJA",
        0xFA04: "HANJA",
        0xFA05: "HANJA",
        0xFA06: "HANJA",
        0xFA07: "HANJA",
        0xFA08: "HANJA",
        0xFA09: "HANJA",
        0xFA0A: "HANJA",
        0xFA0B: "HANJA",
        0xFA0C: "HANJA",
        0xFA0D: "HANJA",
        0xFA0E: "HANJA",
        0xFA0F: "HANJA",
        0xFA10: "HANJA",
        0xFA11: "HANJA",
        0xFA12: "HANJA",
        0xFA13: "HANJA",
        0xFA14: "HANJA",
        0xFA15: "HANJA",
        0xFA16: "HANJA",
        0xFA17: "HANJA",
        0xFA18: "HANJA",
        0xFA19: "HANJA",
        0xFA1A: "HANJA",
        0xFA1B: "HANJA",
        0xFA1C: "HANJA",
        0xFA1D: "HANJA",
        0xFA1E: "HANJA",
        0xFA1F: "HANJA",
        0xFA20: "HANJA",
        0xFA21: "HANJA",
        0xFA22: "HANJA",
        0xFA23: "HANJA",
        0xFA24: "HANJA",
        0xFA25: "HANJA",
        0xFA26: "HANJA",
        0xFA27: "HANJA",
        0xFA28: "HANJA",
        0xFA29: "HANJA",
        0xFA2A: "HANJA",
        0xFA2B: "HANJA",
        0xFA2C: "HANJA",
        0xFA2D: "HANJA",
        0xFA30: "HANJA",
        0xFA31: "HANJA",
        0xFA32: "HANJA",
        0xFA33: "HANJA",
        0xFA34: "HANJA",
        0xFA35: "HANJA",
        0xFA36: "HANJA",
        0xFA37: "HANJA",
        0xFA38: "HANJA",
        0xFA39: "HANJA",
        0xFA3A: "HANJA",
        0xFA3B: "HANJA",
        0xFA3C: "HANJA",
        0xFA3D: "HANJA",
        0xFA3E: "HANJA",
        0xFA3F: "HANJA",
        0xFA40: "HANJA",
        0xFA41: "HANJA",
        0xFA42: "HANJA",
        0xFA43: "HANJA",
        0xFA44: "HANJA",
        0xFA45: "HANJA",
        0xFA46: "HANJA",
        0xFA47: "HANJA",
        0xFA48: "HANJA",
        0xFA49: "HANJA",
        0xFA4A: "HANJA",
        0xFA4B: "HANJA",
        0xFA4C: "HANJA",
        0xFA4D: "HANJA",
        0xFA4E: "HANJA",
        0xFA4F: "HANJA",
        0xFA50: "HANJA",
        0xFA51: "HANJA",
        0xFA52: "HANJA",
        0xFA53: "HANJA",
        0xFA54: "HANJA",
        0xFA55: "HANJA",
        0xFA56: "HANJA",
        0xFA57: "HANJA",
        0xFA58: "HANJA",
        0xFA59: "HANJA",
        0xFA5A: "HANJA",
        0xFA5B: "HANJA",
        0xFA5C: "HANJA",
        0xFA5D: "HANJA",
        0xFA5E: "HANJA",
        0xFA5F: "HANJA",
        0xFA60: "HANJA",
        0xFA61: "HANJA",
        0xFA62: "HANJA",
        0xFA63: "HANJA",
        0xFA64: "HANJA",
        0xFA65: "HANJA",
        0xFA66: "HANJA",
        0xFA67: "HANJA",
        0xFA68: "HANJA",
        0xFA69: "HANJA",
        0xFA6A: "HANJA",
        0x2F00: "KANJI",
        0x2F01: "KANJI",
        0x2F02: "KANJI",
        0x2F03: "KANJI",
        0x2F04: "KANJI",
        0x2F05: "KANJI",
        0x2F06: "KANJI",
        0x2F07: "KANJI",
        0x2F08: "KANJI",
        0x2F09: "KANJI",
        0x2F0A: "KANJI",
        0x2F0B: "KANJI",
        0x2F0C: "KANJI",
        0x2F0D: "KANJI",
        0x2F0E: "KANJI",
        0x2F0F: "KANJI",
        0x2F10: "KANJI",
        0x2F11: "KANJI",
        0x2F12: "KANJI",
        0x2F13: "KANJI",
        0x2F14: "KANJI",
        0x2F15: "KANJI",
        0x2F16: "KANJI",
        0x2F17: "KANJI",
        0x2F18: "KANJI",
        0x2F19: "KANJI",
        0x2F1A: "KANJI",
        0x2F1B: "KANJI",
        0x2F1C: "KANJI",
        0x2F1D: "KANJI",
        0x2F1E: "KANJI",
        0x2F1F: "KANJI",
        0x2F20: "KANJI",
        0x2F21: "KANJI",
        0x2F22: "KANJI",
        0x2F23: "KANJI",
        0x2F24: "KANJI",
        0x2F25: "KANJI",
        0x2F26: "KANJI",
        0x2F27: "KANJI",
        0x2F28: "KANJI",
        0x2F29: "KANJI",
        0x2F2A: "KANJI",
        0x2F2B: "KANJI",
        0x2F2C: "KANJI",
        0x2F2D: "KANJI",
        0x2F2E: "KANJI",
        0x2F2F: "KANJI",
        0x2F30: "KANJI",
        0x2F31: "KANJI",
        0x2F32: "KANJI",
        0x2F33: "KANJI",
        0x2F34: "KANJI",
        0x2F35: "KANJI",
        0x2F36: "KANJI",
        0x2F37: "KANJI",
        0x2F38: "KANJI",
        0x2F39: "KANJI",
        0x2F3A: "KANJI",
        0x2F3B: "KANJI",
        0x2F3C: "KANJI",
        0x2F3D: "KANJI",
        0x2F3E: "KANJI",
        0x2F3F: "KANJI",
        0x2F40: "KANJI",
        0x2F41: "KANJI",
        0x2F42: "KANJI",
        0x2F43: "KANJI",
        0x2F44: "KANJI",
        0x2F45: "KANJI",
        0x2F46: "KANJI",
        0x2F47: "KANJI",
        0x2F48: "KANJI",
        0x2F49: "KANJI",
        0x2F4A: "KANJI",
        0x2F4B: "KANJI",
        0x2F4C: "KANJI",
        0x2F4D: "KANJI",
        0x2F4E: "KANJI",
        0x2F4F: "KANJI",
        0x2F50: "KANJI",
        0x2F51: "KANJI",
        0x2F52: "KANJI",
        0x2F53: "KANJI",
        0x2F54: "KANJI",
        0x2F55: "KANJI",
        0x2F56: "KANJI",
        0x2F57: "KANJI",
        0x2F58: "KANJI",
        0x2F59: "KANJI",
        0x2F5A: "KANJI",
        0x2F5B: "KANJI",
        0x2F5C: "KANJI",
        0x2F5D: "KANJI",
        0x2F5E: "KANJI",
        0x2F5F: "KANJI",
        0x2F60: "KANJI",
        0x2F61: "KANJI",
        0x2F62: "KANJI",
        0x2F63: "KANJI",
        0x2F64: "KANJI",
        0x2F65: "KANJI",
        0x2F66: "KANJI",
        0x2F67: "KANJI",
        0x2F68: "KANJI",
        0x2F69: "KANJI",
        0x2F6A: "KANJI",
        0x2F6B: "KANJI",
        0x2F6C: "KANJI",
        0x2F6D: "KANJI",
        0x2F6E: "KANJI",
        0x2F6F: "KANJI",
        0x2F70: "KANJI",
        0x2F71: "KANJI",
        0x2F72: "KANJI",
        0x2F73: "KANJI",
        0x2F74: "KANJI",
        0x2F75: "KANJI",
        0x2F76: "KANJI",
        0x2F77: "KANJI",
        0x2F78: "KANJI",
        0x2F79: "KANJI",
        0x2F7A: "KANJI",
        0x2F7B: "KANJI",
        0x2F7C: "KANJI",
        0x2F7D: "KANJI",
        0x2F7E: "KANJI",
        0x2F7F: "KANJI",
        0x2F80: "KANJI",
        0x2F81: "KANJI",
        0x2F82: "KANJI",
        0x2F83: "KANJI",
        0x2F84: "KANJI",
        0x2F85: "KANJI",
        0x2F86: "KANJI",
        0x2F87: "KANJI",
        0x2F88: "KANJI",
        0x2F89: "KANJI",
        0x2F8A: "KANJI",
        0x2F8B: "KANJI",
        0x2F8C: "KANJI",
        0x2F8D: "KANJI",
        0x2F8E: "KANJI",
        0x2F8F: "KANJI",
        0x2F90: "KANJI",
        0x2F91: "KANJI",
        0x2F92: "KANJI",
        0x2F93: "KANJI",
        0x2F94: "KANJI",
        0x2F95: "KANJI",
        0x2F96: "KANJI",
        0x2F97: "KANJI",
        0x2F98: "KANJI",
        0x2F99: "KANJI",
        0x2F9A: "KANJI",
        0x2F9B: "KANJI",
        0x2F9C: "KANJI",
        0x2F9D: "KANJI",
        0x2F9E: "KANJI",
        0x2F9F: "KANJI",
        0x2FA0: "KANJI",
        0x2FA1: "KANJI",
        0x2FA2: "KANJI",
        0x2FA3: "KANJI",
        0x2FA4: "KANJI",
        0x2FA5: "KANJI",
        0x2FA6: "KANJI",
        0x2FA7: "KANJI",
        0x2FA8: "KANJI",
        0x2FA9: "KANJI",
        0x2FAA: "KANJI",
        0x2FAB: "KANJI",
        0x2FAC: "KANJI",
        0x2FAD: "KANJI",
        0x2FAE: "KANJI",
        0x2FAF: "KANJI",
        0x2FB0: "KANJI",
        0x2FB1: "KANJI",
        0x2FB2: "KANJI",
        0x2FB3: "KANJI",
        0x2FB4: "KANJI",
        0x2FB5: "KANJI",
        0x2FB6: "KANJI",
        0x2FB7: "KANJI",
        0x2FB8: "KANJI",
        0x2FB9: "KANJI",
        0x2FBA: "KANJI",
        0x2FBB: "KANJI",
        0x2FBC: "KANJI",
        0x2FBD: "KANJI",
        0x2FBE: "KANJI",
        0x2FBF: "KANJI",
        0x2FC0: "KANJI",
        0x2FC1: "KANJI",
        0x2FC2: "KANJI",
        0x2FC3: "KANJI",
        0x2FC4: "KANJI",
        0x2FC5: "KANJI",
        0x2FC6: "KANJI",
        0x2FC7: "KANJI",
        0x2FC8: "KANJI",
        0x2FC9: "KANJI",
        0x2FCA: "KANJI",
        0x2FCB: "KANJI",
        0x2FCC: "KANJI",
        0x2FCD: "KANJI",
        0x2FCE: "KANJI",
        0x2FCF: "KANJI",
        0x2FD0: "KANJI",
        0x2FD1: "KANJI",
        0x2FD2: "KANJI",
        0x2FD3: "KANJI",
        0x2FD4: "KANJI",
        0x2FD5: "KANJI",
        0x4E8C: "HANJA",
        0x4E94: "HANJA",
        0x4E5D: "HANJA",
        0xFF10: "NUMERIC",
        0xFF11: "NUMERIC",
        0xFF12: "NUMERIC",
        0xFF13: "NUMERIC",
        0xFF14: "NUMERIC",
        0xFF15: "NUMERIC",
        0xFF16: "NUMERIC",
        0xFF17: "NUMERIC",
        0xFF18: "NUMERIC",
        0xFF19: "NUMERIC",
        0xFF21: "ALPHA",
        0xFF22: "ALPHA",
        0xFF23: "ALPHA",
        0xFF24: "ALPHA",
        0xFF25: "ALPHA",
        0xFF26: "ALPHA",
        0xFF27: "ALPHA",
        0xFF28: "ALPHA",
        0xFF29: "ALPHA",
        0xFF2A: "ALPHA",
        0xFF2B: "ALPHA",
        0xFF2C: "ALPHA",
        0xFF2D: "ALPHA",
        0xFF2E: "ALPHA",
        0xFF2F: "ALPHA",
        0xFF30: "ALPHA",
        0xFF31: "ALPHA",
        0xFF32: "ALPHA",
        0xFF33: "ALPHA",
        0xFF34: "ALPHA",
        0xFF35: "ALPHA",
        0xFF36: "ALPHA",
        0xFF37: "ALPHA",
        0xFF38: "ALPHA",
        0xFF39: "ALPHA",
        0xFF3A: "ALPHA",
        0xFF41: "ALPHA",
        0xFF42: "ALPHA",
        0xFF43: "ALPHA",
        0xFF44: "ALPHA",
        0xFF45: "ALPHA",
        0xFF46: "ALPHA",
        0xFF47: "ALPHA",
        0xFF48: "ALPHA",
        0xFF49: "ALPHA",
        0xFF4A: "ALPHA",
        0xFF4B: "ALPHA",
        0xFF4C: "ALPHA",
        0xFF4D: "ALPHA",
        0xFF4E: "ALPHA",
        0xFF4F: "ALPHA",
        0xFF50: "ALPHA",
        0xFF51: "ALPHA",
        0xFF52: "ALPHA",
        0xFF53: "ALPHA",
        0xFF54: "ALPHA",
        0xFF55: "ALPHA",
        0xFF56: "ALPHA",
        0xFF57: "ALPHA",
        0xFF58: "ALPHA",
        0xFF59: "ALPHA",
        0xFF5A: "ALPHA",
        0xFF01: "SYMBOL",
        0xFF02: "SYMBOL",
        0xFF03: "SYMBOL",
        0xFF04: "SYMBOL",
        0xFF05: "SYMBOL",
        0xFF06: "SYMBOL",
        0xFF07: "SYMBOL",
        0xFF08: "SYMBOL",
        0xFF09: "SYMBOL",
        0xFF0A: "SYMBOL",
        0xFF0B: "SYMBOL",
        0xFF0C: "SYMBOL",
        0xFF0D: "SYMBOL",
        0xFF0E: "SYMBOL",
        0xFF0F: "SYMBOL",
        0xFF1A: "SYMBOL",
        0xFF1B: "SYMBOL",
        0xFF1C: "SYMBOL",
        0xFF1D: "SYMBOL",
        0xFF1E: "SYMBOL",
        0xFF1F: "SYMBOL",
        0xFF3B: "SYMBOL",
        0xFF3C: "SYMBOL",
        0xFF3D: "SYMBOL",
        0xFF3E: "SYMBOL",
        0xFF3F: "SYMBOL",
        0xFF40: "SYMBOL",
        0xFF5B: "SYMBOL",
        0xFF5C: "SYMBOL",
        0xFF5D: "SYMBOL",
        0xFF5E: "SYMBOL",
        0xFF5F: "SYMBOL",
        0xFF60: "SYMBOL",
        0xFF61: "SYMBOL",
        0xFF62: "SYMBOL",
        0xFF63: "SYMBOL",
        0xFF64: "SYMBOL",
        0xFF65: "SYMBOL",
        0xFFE0: "SYMBOL",
        0xFFE1: "SYMBOL",
        0xFFE2: "SYMBOL",
        0xFFE3: "SYMBOL",
        0xFFE4: "SYMBOL",
        0xFFE5: "SYMBOL",
        0xFFE6: "SYMBOL",
        0xFFE7: "SYMBOL",
        0xFFE8: "SYMBOL",
        0xFFE9: "SYMBOL",
        0xFFEA: "SYMBOL",
        0xFFEB: "SYMBOL",
        0xFFEC: "SYMBOL",
        0xFFED: "SYMBOL",
        0xFFEE: "SYMBOL",
        0xFFEF: "SYMBOL",
        0x2070: "NUMERIC",
        0x2071: "NUMERIC",
        0x2072: "NUMERIC",
        0x2073: "NUMERIC",
        0x2074: "NUMERIC",
        0x2075: "NUMERIC",
        0x2076: "NUMERIC",
        0x2077: "NUMERIC",
        0x2078: "NUMERIC",
        0x2079: "NUMERIC",
        0x207A: "NUMERIC",
        0x207B: "NUMERIC",
        0x207C: "NUMERIC",
        0x207D: "NUMERIC",
        0x207E: "NUMERIC",
        0x207F: "NUMERIC",
        0x2080: "NUMERIC",
        0x2081: "NUMERIC",
        0x2082: "NUMERIC",
        0x2083: "NUMERIC",
        0x2084: "NUMERIC",
        0x2085: "NUMERIC",
        0x2086: "NUMERIC",
        0x2087: "NUMERIC",
        0x2088: "NUMERIC",
        0x2089: "NUMERIC",
        0x208A: "NUMERIC",
        0x208B: "NUMERIC",
        0x208C: "NUMERIC",
        0x208D: "NUMERIC",
        0x208E: "NUMERIC",
        0x208F: "NUMERIC",
        0x2090: "NUMERIC",
        0x2091: "NUMERIC",
        0x2092: "NUMERIC",
        0x2093: "NUMERIC",
        0x2094: "NUMERIC",
        0x2095: "NUMERIC",
        0x2096: "NUMERIC",
        0x2097: "NUMERIC",
        0x2098: "NUMERIC",
        0x2099: "NUMERIC",
        0x209A: "NUMERIC",
        0x209B: "NUMERIC",
        0x209C: "NUMERIC",
        0x209D: "NUMERIC",
        0x209E: "NUMERIC",
        0x209F: "NUMERIC",
        0x2150: "NUMERIC",
        0x2151: "NUMERIC",
        0x2152: "NUMERIC",
        0x2153: "NUMERIC",
        0x2154: "NUMERIC",
        0x2155: "NUMERIC",
        0x2156: "NUMERIC",
        0x2157: "NUMERIC",
        0x2158: "NUMERIC",
        0x2159: "NUMERIC",
        0x215A: "NUMERIC",
        0x215B: "NUMERIC",
        0x215C: "NUMERIC",
        0x215D: "NUMERIC",
        0x215E: "NUMERIC",
        0x215F: "NUMERIC",
        0x2160: "NUMERIC",
        0x2161: "NUMERIC",
        0x2162: "NUMERIC",
        0x2163: "NUMERIC",
        0x2164: "NUMERIC",
        0x2165: "NUMERIC",
        0x2166: "NUMERIC",
        0x2167: "NUMERIC",
        0x2168: "NUMERIC",
        0x2169: "NUMERIC",
        0x216A: "NUMERIC",
        0x216B: "NUMERIC",
        0x216C: "NUMERIC",
        0x216D: "NUMERIC",
        0x216E: "NUMERIC",
        0x216F: "NUMERIC",
        0x2170: "NUMERIC",
        0x2171: "NUMERIC",
        0x2172: "NUMERIC",
        0x2173: "NUMERIC",
        0x2174: "NUMERIC",
        0x2175: "NUMERIC",
        0x2176: "NUMERIC",
        0x2177: "NUMERIC",
        0x2178: "NUMERIC",
        0x2179: "NUMERIC",
        0x217A: "NUMERIC",
        0x217B: "NUMERIC",
        0x217C: "NUMERIC",
        0x217D: "NUMERIC",
        0x217E: "NUMERIC",
        0x217F: "NUMERIC",
        0x2180: "NUMERIC",
        0x2181: "NUMERIC",
        0x2182: "NUMERIC",
        0x2183: "NUMERIC",
        0x2184: "NUMERIC",
        0x2185: "NUMERIC",
        0x2186: "NUMERIC",
        0x2187: "NUMERIC",
        0x2188: "NUMERIC",
        0x2189: "NUMERIC",
        0x218A: "NUMERIC",
        0x218B: "NUMERIC",
        0x218C: "NUMERIC",
        0x218D: "NUMERIC",
        0x218E: "NUMERIC",
        0x218F: "NUMERIC",
        0x2000: "SYMBOL",
        0x2001: "SYMBOL",
        0x2002: "SYMBOL",
        0x2003: "SYMBOL",
        0x2004: "SYMBOL",
        0x2005: "SYMBOL",
        0x2006: "SYMBOL",
        0x2007: "SYMBOL",
        0x2008: "SYMBOL",
        0x2009: "SYMBOL",
        0x200A: "SYMBOL",
        0x200B: "SYMBOL",
        0x200C: "SYMBOL",
        0x200D: "SYMBOL",
        0x200E: "SYMBOL",
        0x200F: "SYMBOL",
        0x2010: "SYMBOL",
        0x2011: "SYMBOL",
        0x2012: "SYMBOL",
        0x2013: "SYMBOL",
        0x2014: "SYMBOL",
        0x2015: "SYMBOL",
        0x2016: "SYMBOL",
        0x2017: "SYMBOL",
        0x2018: "SYMBOL",
        0x2019: "SYMBOL",
        0x201A: "SYMBOL",
        0x201B: "SYMBOL",
        0x201C: "SYMBOL",
        0x201D: "SYMBOL",
        0x201E: "SYMBOL",
        0x201F: "SYMBOL",
        0x2020: "SYMBOL",
        0x2021: "SYMBOL",
        0x2022: "SYMBOL",
        0x2023: "SYMBOL",
        0x2024: "SYMBOL",
        0x2025: "SYMBOL",
        0x2026: "SYMBOL",
        0x2027: "SYMBOL",
        0x2028: "SYMBOL",
        0x2029: "SYMBOL",
        0x202A: "SYMBOL",
        0x202B: "SYMBOL",
        0x202C: "SYMBOL",
        0x202D: "SYMBOL",
        0x202E: "SYMBOL",
        0x202F: "SYMBOL",
        0x2030: "SYMBOL",
        0x2031: "SYMBOL",
        0x2032: "SYMBOL",
        0x2033: "SYMBOL",
        0x2034: "SYMBOL",
        0x2035: "SYMBOL",
        0x2036: "SYMBOL",
        0x2037: "SYMBOL",
        0x2038: "SYMBOL",
        0x2039: "SYMBOL",
        0x203A: "SYMBOL",
        0x203B: "SYMBOL",
        0x203C: "SYMBOL",
        0x203D: "SYMBOL",
        0x203E: "SYMBOL",
        0x203F: "SYMBOL",
        0x2040: "SYMBOL",
        0x2041: "SYMBOL",
        0x2042: "SYMBOL",
        0x2043: "SYMBOL",
        0x2044: "SYMBOL",
        0x2045: "SYMBOL",
        0x2046: "SYMBOL",
        0x2047: "SYMBOL",
        0x2048: "SYMBOL",
        0x2049: "SYMBOL",
        0x204A: "SYMBOL",
        0x204B: "SYMBOL",
        0x204C: "SYMBOL",
        0x204D: "SYMBOL",
        0x204E: "SYMBOL",
        0x204F: "SYMBOL",
        0x2050: "SYMBOL",
        0x2051: "SYMBOL",
        0x2052: "SYMBOL",
        0x2053: "SYMBOL",
        0x2054: "SYMBOL",
        0x2055: "SYMBOL",
        0x2056: "SYMBOL",
        0x2057: "SYMBOL",
        0x2058: "SYMBOL",
        0x2059: "SYMBOL",
        0x205A: "SYMBOL",
        0x205B: "SYMBOL",
        0x205C: "SYMBOL",
        0x205D: "SYMBOL",
        0x205E: "SYMBOL",
        0x205F: "SYMBOL",
        0x2060: "SYMBOL",
        0x2061: "SYMBOL",
        0x2062: "SYMBOL",
        0x2063: "SYMBOL",
        0x2064: "SYMBOL",
        0x2065: "SYMBOL",
        0x2066: "SYMBOL",
        0x2067: "SYMBOL",
        0x2068: "SYMBOL",
        0x2069: "SYMBOL",
        0x206A: "SYMBOL",
        0x206B: "SYMBOL",
        0x206C: "SYMBOL",
        0x206D: "SYMBOL",
        0x206E: "SYMBOL",
        0x206F: "SYMBOL",
        0x20A0: "SYMBOL",
        0x20A1: "SYMBOL",
        0x20A2: "SYMBOL",
        0x20A3: "SYMBOL",
        0x20A4: "SYMBOL",
        0x20A5: "SYMBOL",
        0x20A6: "SYMBOL",
        0x20A7: "SYMBOL",
        0x20A8: "SYMBOL",
        0x20A9: "SYMBOL",
        0x20AA: "SYMBOL",
        0x20AB: "SYMBOL",
        0x20AC: "SYMBOL",
        0x20AD: "SYMBOL",
        0x20AE: "SYMBOL",
        0x20AF: "SYMBOL",
        0x20B0: "SYMBOL",
        0x20B1: "SYMBOL",
        0x20B2: "SYMBOL",
        0x20B3: "SYMBOL",
        0x20B4: "SYMBOL",
        0x20B5: "SYMBOL",
        0x20B6: "SYMBOL",
        0x20B7: "SYMBOL",
        0x20B8: "SYMBOL",
        0x20B9: "SYMBOL",
        0x20BA: "SYMBOL",
        0x20BB: "SYMBOL",
        0x20BC: "SYMBOL",
        0x20BD: "SYMBOL",
        0x20BE: "SYMBOL",
        0x20BF: "SYMBOL",
        0x20C0: "SYMBOL",
        0x20C1: "SYMBOL",
        0x20C2: "SYMBOL",
        0x20C3: "SYMBOL",
        0x20C4: "SYMBOL",
        0x20C5: "SYMBOL",
        0x20C6: "SYMBOL",
        0x20C7: "SYMBOL",
        0x20C8: "SYMBOL",
        0x20C9: "SYMBOL",
        0x20CA: "SYMBOL",
        0x20CB: "SYMBOL",
        0x20CC: "SYMBOL",
        0x20CD: "SYMBOL",
        0x20CE: "SYMBOL",
        0x20CF: "SYMBOL",
        0x20D0: "SYMBOL",
        0x20D1: "SYMBOL",
        0x20D2: "SYMBOL",
        0x20D3: "SYMBOL",
        0x20D4: "SYMBOL",
        0x20D5: "SYMBOL",
        0x20D6: "SYMBOL",
        0x20D7: "SYMBOL",
        0x20D8: "SYMBOL",
        0x20D9: "SYMBOL",
        0x20DA: "SYMBOL",
        0x20DB: "SYMBOL",
        0x20DC: "SYMBOL",
        0x20DD: "SYMBOL",
        0x20DE: "SYMBOL",
        0x20DF: "SYMBOL",
        0x20E0: "SYMBOL",
        0x20E1: "SYMBOL",
        0x20E2: "SYMBOL",
        0x20E3: "SYMBOL",
        0x20E4: "SYMBOL",
        0x20E5: "SYMBOL",
        0x20E6: "SYMBOL",
        0x20E7: "SYMBOL",
        0x20E8: "SYMBOL",
        0x20E9: "SYMBOL",
        0x20EA: "SYMBOL",
        0x20EB: "SYMBOL",
        0x20EC: "SYMBOL",
        0x20ED: "SYMBOL",
        0x20EE: "SYMBOL",
        0x20EF: "SYMBOL",
        0x20F0: "SYMBOL",
        0x20F1: "SYMBOL",
        0x20F2: "SYMBOL",
        0x20F3: "SYMBOL",
        0x20F4: "SYMBOL",
        0x20F5: "SYMBOL",
        0x20F6: "SYMBOL",
        0x20F7: "SYMBOL",
        0x20F8: "SYMBOL",
        0x20F9: "SYMBOL",
        0x20FA: "SYMBOL",
        0x20FB: "SYMBOL",
        0x20FC: "SYMBOL",
        0x20FD: "SYMBOL",
        0x20FE: "SYMBOL",
        0x20FF: "SYMBOL",
        0x2100: "SYMBOL",
        0x2101: "SYMBOL",
        0x2102: "SYMBOL",
        0x2103: "SYMBOL",
        0x2104: "SYMBOL",
        0x2105: "SYMBOL",
        0x2106: "SYMBOL",
        0x2107: "SYMBOL",
        0x2108: "SYMBOL",
        0x2109: "SYMBOL",
        0x210A: "SYMBOL",
        0x210B: "SYMBOL",
        0x210C: "SYMBOL",
        0x210D: "SYMBOL",
        0x210E: "SYMBOL",
        0x210F: "SYMBOL",
        0x2110: "SYMBOL",
        0x2111: "SYMBOL",
        0x2112: "SYMBOL",
        0x2113: "SYMBOL",
        0x2114: "SYMBOL",
        0x2115: "SYMBOL",
        0x2116: "SYMBOL",
        0x2117: "SYMBOL",
        0x2118: "SYMBOL",
        0x2119: "SYMBOL",
        0x211A: "SYMBOL",
        0x211B: "SYMBOL",
        0x211C: "SYMBOL",
        0x211D: "SYMBOL",
        0x211E: "SYMBOL",
        0x211F: "SYMBOL",
        0x2120: "SYMBOL",
        0x2121: "SYMBOL",
        0x2122: "SYMBOL",
        0x2123: "SYMBOL",
        0x2124: "SYMBOL",
        0x2125: "SYMBOL",
        0x2126: "SYMBOL",
        0x2127: "SYMBOL",
        0x2128: "SYMBOL",
        0x2129: "SYMBOL",
        0x212A: "SYMBOL",
        0x212B: "SYMBOL",
        0x212C: "SYMBOL",
        0x212D: "SYMBOL",
        0x212E: "SYMBOL",
        0x212F: "SYMBOL",
        0x2130: "SYMBOL",
        0x2131: "SYMBOL",
        0x2132: "SYMBOL",
        0x2133: "SYMBOL",
        0x2134: "SYMBOL",
        0x2135: "SYMBOL",
        0x2136: "SYMBOL",
        0x2137: "SYMBOL",
        0x2138: "SYMBOL",
        0x2139: "SYMBOL",
        0x213A: "SYMBOL",
        0x213B: "SYMBOL",
        0x213C: "SYMBOL",
        0x213D: "SYMBOL",
        0x213E: "SYMBOL",
        0x213F: "SYMBOL",
        0x2140: "SYMBOL",
        0x2141: "SYMBOL",
        0x2142: "SYMBOL",
        0x2143: "SYMBOL",
        0x2144: "SYMBOL",
        0x2145: "SYMBOL",
        0x2146: "SYMBOL",
        0x2147: "SYMBOL",
        0x2148: "SYMBOL",
        0x2149: "SYMBOL",
        0x214A: "SYMBOL",
        0x214B: "SYMBOL",
        0x214C: "SYMBOL",
        0x214D: "SYMBOL",
        0x214E: "SYMBOL",
        0x214F: "SYMBOL",
        0x2190: "SYMBOL",
        0x2191: "SYMBOL",
        0x2192: "SYMBOL",
        0x2193: "SYMBOL",
        0x2194: "SYMBOL",
        0x2195: "SYMBOL",
        0x2196: "SYMBOL",
        0x2197: "SYMBOL",
        0x2198: "SYMBOL",
        0x2199: "SYMBOL",
        0x219A: "SYMBOL",
        0x219B: "SYMBOL",
        0x219C: "SYMBOL",
        0x219D: "SYMBOL",
        0x219E: "SYMBOL",
        0x219F: "SYMBOL",
        0x21A0: "SYMBOL",
        0x21A1: "SYMBOL",
        0x21A2: "SYMBOL",
        0x21A3: "SYMBOL",
        0x21A4: "SYMBOL",
        0x21A5: "SYMBOL",
        0x21A6: "SYMBOL",
        0x21A7: "SYMBOL",
        0x21A8: "SYMBOL",
        0x21A9: "SYMBOL",
        0x21AA: "SYMBOL",
        0x21AB: "SYMBOL",
        0x21AC: "SYMBOL",
        0x21AD: "SYMBOL",
        0x21AE: "SYMBOL",
        0x21AF: "SYMBOL",
        0x21B0: "SYMBOL",
        0x21B1: "SYMBOL",
        0x21B2: "SYMBOL",
        0x21B3: "SYMBOL",
        0x21B4: "SYMBOL",
        0x21B5: "SYMBOL",
        0x21B6: "SYMBOL",
        0x21B7: "SYMBOL",
        0x21B8: "SYMBOL",
        0x21B9: "SYMBOL",
        0x21BA: "SYMBOL",
        0x21BB: "SYMBOL",
        0x21BC: "SYMBOL",
        0x21BD: "SYMBOL",
        0x21BE: "SYMBOL",
        0x21BF: "SYMBOL",
        0x21C0: "SYMBOL",
        0x21C1: "SYMBOL",
        0x21C2: "SYMBOL",
        0x21C3: "SYMBOL",
        0x21C4: "SYMBOL",
        0x21C5: "SYMBOL",
        0x21C6: "SYMBOL",
        0x21C7: "SYMBOL",
        0x21C8: "SYMBOL",
        0x21C9: "SYMBOL",
        0x21CA: "SYMBOL",
        0x21CB: "SYMBOL",
        0x21CC: "SYMBOL",
        0x21CD: "SYMBOL",
        0x21CE: "SYMBOL",
        0x21CF: "SYMBOL",
        0x21D0: "SYMBOL",
        0x21D1: "SYMBOL",
        0x21D2: "SYMBOL",
        0x21D3: "SYMBOL",
        0x21D4: "SYMBOL",
        0x21D5: "SYMBOL",
        0x21D6: "SYMBOL",
        0x21D7: "SYMBOL",
        0x21D8: "SYMBOL",
        0x21D9: "SYMBOL",
        0x21DA: "SYMBOL",
        0x21DB: "SYMBOL",
        0x21DC: "SYMBOL",
        0x21DD: "SYMBOL",
        0x21DE: "SYMBOL",
        0x21DF: "SYMBOL",
        0x21E0: "SYMBOL",
        0x21E1: "SYMBOL",
        0x21E2: "SYMBOL",
        0x21E3: "SYMBOL",
        0x21E4: "SYMBOL",
        0x21E5: "SYMBOL",
        0x21E6: "SYMBOL",
        0x21E7: "SYMBOL",
        0x21E8: "SYMBOL",
        0x21E9: "SYMBOL",
        0x21EA: "SYMBOL",
        0x21EB: "SYMBOL",
        0x21EC: "SYMBOL",
        0x21ED: "SYMBOL",
        0x21EE: "SYMBOL",
        0x21EF: "SYMBOL",
        0x21F0: "SYMBOL",
        0x21F1: "SYMBOL",
        0x21F2: "SYMBOL",
        0x21F3: "SYMBOL",
        0x21F4: "SYMBOL",
        0x21F5: "SYMBOL",
        0x21F6: "SYMBOL",
        0x21F7: "SYMBOL",
        0x21F8: "SYMBOL",
        0x21F9: "SYMBOL",
        0x21FA: "SYMBOL",
        0x21FB: "SYMBOL",
        0x21FC: "SYMBOL",
        0x21FD: "SYMBOL",
        0x21FE: "SYMBOL",
        0x21FF: "SYMBOL",
        0x2200: "SYMBOL",
        0x2201: "SYMBOL",
        0x2202: "SYMBOL",
        0x2203: "SYMBOL",
        0x2204: "SYMBOL",
        0x2205: "SYMBOL",
        0x2206: "SYMBOL",
        0x2207: "SYMBOL",
        0x2208: "SYMBOL",
        0x2209: "SYMBOL",
        0x220A: "SYMBOL",
        0x220B: "SYMBOL",
        0x220C: "SYMBOL",
        0x220D: "SYMBOL",
        0x220E: "SYMBOL",
        0x220F: "SYMBOL",
        0x2210: "SYMBOL",
        0x2211: "SYMBOL",
        0x2212: "SYMBOL",
        0x2213: "SYMBOL",
        0x2214: "SYMBOL",
        0x2215: "SYMBOL",
        0x2216: "SYMBOL",
        0x2217: "SYMBOL",
        0x2218: "SYMBOL",
        0x2219: "SYMBOL",
        0x221A: "SYMBOL",
        0x221B: "SYMBOL",
        0x221C: "SYMBOL",
        0x221D: "SYMBOL",
        0x221E: "SYMBOL",
        0x221F: "SYMBOL",
        0x2220: "SYMBOL",
        0x2221: "SYMBOL",
        0x2222: "SYMBOL",
        0x2223: "SYMBOL",
        0x2224: "SYMBOL",
        0x2225: "SYMBOL",
        0x2226: "SYMBOL",
        0x2227: "SYMBOL",
        0x2228: "SYMBOL",
        0x2229: "SYMBOL",
        0x222A: "SYMBOL",
        0x222B: "SYMBOL",
        0x222C: "SYMBOL",
        0x222D: "SYMBOL",
        0x222E: "SYMBOL",
        0x222F: "SYMBOL",
        0x2230: "SYMBOL",
        0x2231: "SYMBOL",
        0x2232: "SYMBOL",
        0x2233: "SYMBOL",
        0x2234: "SYMBOL",
        0x2235: "SYMBOL",
        0x2236: "SYMBOL",
        0x2237: "SYMBOL",
        0x2238: "SYMBOL",
        0x2239: "SYMBOL",
        0x223A: "SYMBOL",
        0x223B: "SYMBOL",
        0x223C: "SYMBOL",
        0x223D: "SYMBOL",
        0x223E: "SYMBOL",
        0x223F: "SYMBOL",
        0x2240: "SYMBOL",
        0x2241: "SYMBOL",
        0x2242: "SYMBOL",
        0x2243: "SYMBOL",
        0x2244: "SYMBOL",
        0x2245: "SYMBOL",
        0x2246: "SYMBOL",
        0x2247: "SYMBOL",
        0x2248: "SYMBOL",
        0x2249: "SYMBOL",
        0x224A: "SYMBOL",
        0x224B: "SYMBOL",
        0x224C: "SYMBOL",
        0x224D: "SYMBOL",
        0x224E: "SYMBOL",
        0x224F: "SYMBOL",
        0x2250: "SYMBOL",
        0x2251: "SYMBOL",
        0x2252: "SYMBOL",
        0x2253: "SYMBOL",
        0x2254: "SYMBOL",
        0x2255: "SYMBOL",
        0x2256: "SYMBOL",
        0x2257: "SYMBOL",
        0x2258: "SYMBOL",
        0x2259: "SYMBOL",
        0x225A: "SYMBOL",
        0x225B: "SYMBOL",
        0x225C: "SYMBOL",
        0x225D: "SYMBOL",
        0x225E: "SYMBOL",
        0x225F: "SYMBOL",
        0x2260: "SYMBOL",
        0x2261: "SYMBOL",
        0x2262: "SYMBOL",
        0x2263: "SYMBOL",
        0x2264: "SYMBOL",
        0x2265: "SYMBOL",
        0x2266: "SYMBOL",
        0x2267: "SYMBOL",
        0x2268: "SYMBOL",
        0x2269: "SYMBOL",
        0x226A: "SYMBOL",
        0x226B: "SYMBOL",
        0x226C: "SYMBOL",
        0x226D: "SYMBOL",
        0x226E: "SYMBOL",
        0x226F: "SYMBOL",
        0x2270: "SYMBOL",
        0x2271: "SYMBOL",
        0x2272: "SYMBOL",
        0x2273: "SYMBOL",
        0x2274: "SYMBOL",
        0x2275: "SYMBOL",
        0x2276: "SYMBOL",
        0x2277: "SYMBOL",
        0x2278: "SYMBOL",
        0x2279: "SYMBOL",
        0x227A: "SYMBOL",
        0x227B: "SYMBOL",
        0x227C: "SYMBOL",
        0x227D: "SYMBOL",
        0x227E: "SYMBOL",
        0x227F: "SYMBOL",
        0x2280: "SYMBOL",
        0x2281: "SYMBOL",
        0x2282: "SYMBOL",
        0x2283: "SYMBOL",
        0x2284: "SYMBOL",
        0x2285: "SYMBOL",
        0x2286: "SYMBOL",
        0x2287: "SYMBOL",
        0x2288: "SYMBOL",
        0x2289: "SYMBOL",
        0x228A: "SYMBOL",
        0x228B: "SYMBOL",
        0x228C: "SYMBOL",
        0x228D: "SYMBOL",
        0x228E: "SYMBOL",
        0x228F: "SYMBOL",
        0x2290: "SYMBOL",
        0x2291: "SYMBOL",
        0x2292: "SYMBOL",
        0x2293: "SYMBOL",
        0x2294: "SYMBOL",
        0x2295: "SYMBOL",
        0x2296: "SYMBOL",
        0x2297: "SYMBOL",
        0x2298: "SYMBOL",
        0x2299: "SYMBOL",
        0x229A: "SYMBOL",
        0x229B: "SYMBOL",
        0x229C: "SYMBOL",
        0x229D: "SYMBOL",
        0x229E: "SYMBOL",
        0x229F: "SYMBOL",
        0x22A0: "SYMBOL",
        0x22A1: "SYMBOL",
        0x22A2: "SYMBOL",
        0x22A3: "SYMBOL",
        0x22A4: "SYMBOL",
        0x22A5: "SYMBOL",
        0x22A6: "SYMBOL",
        0x22A7: "SYMBOL",
        0x22A8: "SYMBOL",
        0x22A9: "SYMBOL",
        0x22AA: "SYMBOL",
        0x22AB: "SYMBOL",
        0x22AC: "SYMBOL",
        0x22AD: "SYMBOL",
        0x22AE: "SYMBOL",
        0x22AF: "SYMBOL",
        0x22B0: "SYMBOL",
        0x22B1: "SYMBOL",
        0x22B2: "SYMBOL",
        0x22B3: "SYMBOL",
        0x22B4: "SYMBOL",
        0x22B5: "SYMBOL",
        0x22B6: "SYMBOL",
        0x22B7: "SYMBOL",
        0x22B8: "SYMBOL",
        0x22B9: "SYMBOL",
        0x22BA: "SYMBOL",
        0x22BB: "SYMBOL",
        0x22BC: "SYMBOL",
        0x22BD: "SYMBOL",
        0x22BE: "SYMBOL",
        0x22BF: "SYMBOL",
        0x22C0: "SYMBOL",
        0x22C1: "SYMBOL",
        0x22C2: "SYMBOL",
        0x22C3: "SYMBOL",
        0x22C4: "SYMBOL",
        0x22C5: "SYMBOL",
        0x22C6: "SYMBOL",
        0x22C7: "SYMBOL",
        0x22C8: "SYMBOL",
        0x22C9: "SYMBOL",
        0x22CA: "SYMBOL",
        0x22CB: "SYMBOL",
        0x22CC: "SYMBOL",
        0x22CD: "SYMBOL",
        0x22CE: "SYMBOL",
        0x22CF: "SYMBOL",
        0x22D0: "SYMBOL",
        0x22D1: "SYMBOL",
        0x22D2: "SYMBOL",
        0x22D3: "SYMBOL",
        0x22D4: "SYMBOL",
        0x22D5: "SYMBOL",
        0x22D6: "SYMBOL",
        0x22D7: "SYMBOL",
        0x22D8: "SYMBOL",
        0x22D9: "SYMBOL",
        0x22DA: "SYMBOL",
        0x22DB: "SYMBOL",
        0x22DC: "SYMBOL",
        0x22DD: "SYMBOL",
        0x22DE: "SYMBOL",
        0x22DF: "SYMBOL",
        0x22E0: "SYMBOL",
        0x22E1: "SYMBOL",
        0x22E2: "SYMBOL",
        0x22E3: "SYMBOL",
        0x22E4: "SYMBOL",
        0x22E5: "SYMBOL",
        0x22E6: "SYMBOL",
        0x22E7: "SYMBOL",
        0x22E8: "SYMBOL",
        0x22E9: "SYMBOL",
        0x22EA: "SYMBOL",
        0x22EB: "SYMBOL",
        0x22EC: "SYMBOL",
        0x22ED: "SYMBOL",
        0x22EE: "SYMBOL",
        0x22EF: "SYMBOL",
        0x22F0: "SYMBOL",
        0x22F1: "SYMBOL",
        0x22F2: "SYMBOL",
        0x22F3: "SYMBOL",
        0x22F4: "SYMBOL",
        0x22F5: "SYMBOL",
        0x22F6: "SYMBOL",
        0x22F7: "SYMBOL",
        0x22F8: "SYMBOL",
        0x22F9: "SYMBOL",
        0x22FA: "SYMBOL",
        0x22FB: "SYMBOL",
        0x22FC: "SYMBOL",
        0x22FD: "SYMBOL",
        0x22FE: "SYMBOL",
        0x22FF: "SYMBOL",
        0x2300: "SYMBOL",
        0x2301: "SYMBOL",
        0x2302: "SYMBOL",
        0x2303: "SYMBOL",
        0x2304: "SYMBOL",
        0x2305: "SYMBOL",
        0x2306: "SYMBOL",
        0x2307: "SYMBOL",
        0x2308: "SYMBOL",
        0x2309: "SYMBOL",
        0x230A: "SYMBOL",
        0x230B: "SYMBOL",
        0x230C: "SYMBOL",
        0x230D: "SYMBOL",
        0x230E: "SYMBOL",
        0x230F: "SYMBOL",
        0x2310: "SYMBOL",
        0x2311: "SYMBOL",
        0x2312: "SYMBOL",
        0x2313: "SYMBOL",
        0x2314: "SYMBOL",
        0x2315: "SYMBOL",
        0x2316: "SYMBOL",
        0x2317: "SYMBOL",
        0x2318: "SYMBOL",
        0x2319: "SYMBOL",
        0x231A: "SYMBOL",
        0x231B: "SYMBOL",
        0x231C: "SYMBOL",
        0x231D: "SYMBOL",
        0x231E: "SYMBOL",
        0x231F: "SYMBOL",
        0x2320: "SYMBOL",
        0x2321: "SYMBOL",
        0x2322: "SYMBOL",
        0x2323: "SYMBOL",
        0x2324: "SYMBOL",
        0x2325: "SYMBOL",
        0x2326: "SYMBOL",
        0x2327: "SYMBOL",
        0x2328: "SYMBOL",
        0x2329: "SYMBOL",
        0x232A: "SYMBOL",
        0x232B: "SYMBOL",
        0x232C: "SYMBOL",
        0x232D: "SYMBOL",
        0x232E: "SYMBOL",
        0x232F: "SYMBOL",
        0x2330: "SYMBOL",
        0x2331: "SYMBOL",
        0x2332: "SYMBOL",
        0x2333: "SYMBOL",
        0x2334: "SYMBOL",
        0x2335: "SYMBOL",
        0x2336: "SYMBOL",
        0x2337: "SYMBOL",
        0x2338: "SYMBOL",
        0x2339: "SYMBOL",
        0x233A: "SYMBOL",
        0x233B: "SYMBOL",
        0x233C: "SYMBOL",
        0x233D: "SYMBOL",
        0x233E: "SYMBOL",
        0x233F: "SYMBOL",
        0x2340: "SYMBOL",
        0x2341: "SYMBOL",
        0x2342: "SYMBOL",
        0x2343: "SYMBOL",
        0x2344: "SYMBOL",
        0x2345: "SYMBOL",
        0x2346: "SYMBOL",
        0x2347: "SYMBOL",
        0x2348: "SYMBOL",
        0x2349: "SYMBOL",
        0x234A: "SYMBOL",
        0x234B: "SYMBOL",
        0x234C: "SYMBOL",
        0x234D: "SYMBOL",
        0x234E: "SYMBOL",
        0x234F: "SYMBOL",
        0x2350: "SYMBOL",
        0x2351: "SYMBOL",
        0x2352: "SYMBOL",
        0x2353: "SYMBOL",
        0x2354: "SYMBOL",
        0x2355: "SYMBOL",
        0x2356: "SYMBOL",
        0x2357: "SYMBOL",
        0x2358: "SYMBOL",
        0x2359: "SYMBOL",
        0x235A: "SYMBOL",
        0x235B: "SYMBOL",
        0x235C: "SYMBOL",
        0x235D: "SYMBOL",
        0x235E: "SYMBOL",
        0x235F: "SYMBOL",
        0x2360: "SYMBOL",
        0x2361: "SYMBOL",
        0x2362: "SYMBOL",
        0x2363: "SYMBOL",
        0x2364: "SYMBOL",
        0x2365: "SYMBOL",
        0x2366: "SYMBOL",
        0x2367: "SYMBOL",
        0x2368: "SYMBOL",
        0x2369: "SYMBOL",
        0x236A: "SYMBOL",
        0x236B: "SYMBOL",
        0x236C: "SYMBOL",
        0x236D: "SYMBOL",
        0x236E: "SYMBOL",
        0x236F: "SYMBOL",
        0x2370: "SYMBOL",
        0x2371: "SYMBOL",
        0x2372: "SYMBOL",
        0x2373: "SYMBOL",
        0x2374: "SYMBOL",
        0x2375: "SYMBOL",
        0x2376: "SYMBOL",
        0x2377: "SYMBOL",
        0x2378: "SYMBOL",
        0x2379: "SYMBOL",
        0x237A: "SYMBOL",
        0x237B: "SYMBOL",
        0x237C: "SYMBOL",
        0x237D: "SYMBOL",
        0x237E: "SYMBOL",
        0x237F: "SYMBOL",
        0x2380: "SYMBOL",
        0x2381: "SYMBOL",
        0x2382: "SYMBOL",
        0x2383: "SYMBOL",
        0x2384: "SYMBOL",
        0x2385: "SYMBOL",
        0x2386: "SYMBOL",
        0x2387: "SYMBOL",
        0x2388: "SYMBOL",
        0x2389: "SYMBOL",
        0x238A: "SYMBOL",
        0x238B: "SYMBOL",
        0x238C: "SYMBOL",
        0x238D: "SYMBOL",
        0x238E: "SYMBOL",
        0x238F: "SYMBOL",
        0x2390: "SYMBOL",
        0x2391: "SYMBOL",
        0x2392: "SYMBOL",
        0x2393: "SYMBOL",
        0x2394: "SYMBOL",
        0x2395: "SYMBOL",
        0x2396: "SYMBOL",
        0x2397: "SYMBOL",
        0x2398: "SYMBOL",
        0x2399: "SYMBOL",
        0x239A: "SYMBOL",
        0x239B: "SYMBOL",
        0x239C: "SYMBOL",
        0x239D: "SYMBOL",
        0x239E: "SYMBOL",
        0x239F: "SYMBOL",
        0x23A0: "SYMBOL",
        0x23A1: "SYMBOL",
        0x23A2: "SYMBOL",
        0x23A3: "SYMBOL",
        0x23A4: "SYMBOL",
        0x23A5: "SYMBOL",
        0x23A6: "SYMBOL",
        0x23A7: "SYMBOL",
        0x23A8: "SYMBOL",
        0x23A9: "SYMBOL",
        0x23AA: "SYMBOL",
        0x23AB: "SYMBOL",
        0x23AC: "SYMBOL",
        0x23AD: "SYMBOL",
        0x23AE: "SYMBOL",
        0x23AF: "SYMBOL",
        0x23B0: "SYMBOL",
        0x23B1: "SYMBOL",
        0x23B2: "SYMBOL",
        0x23B3: "SYMBOL",
        0x23B4: "SYMBOL",
        0x23B5: "SYMBOL",
        0x23B6: "SYMBOL",
        0x23B7: "SYMBOL",
        0x23B8: "SYMBOL",
        0x23B9: "SYMBOL",
        0x23BA: "SYMBOL",
        0x23BB: "SYMBOL",
        0x23BC: "SYMBOL",
        0x23BD: "SYMBOL",
        0x23BE: "SYMBOL",
        0x23BF: "SYMBOL",
        0x23C0: "SYMBOL",
        0x23C1: "SYMBOL",
        0x23C2: "SYMBOL",
        0x23C3: "SYMBOL",
        0x23C4: "SYMBOL",
        0x23C5: "SYMBOL",
        0x23C6: "SYMBOL",
        0x23C7: "SYMBOL",
        0x23C8: "SYMBOL",
        0x23C9: "SYMBOL",
        0x23CA: "SYMBOL",
        0x23CB: "SYMBOL",
        0x23CC: "SYMBOL",
        0x23CD: "SYMBOL",
        0x23CE: "SYMBOL",
        0x23CF: "SYMBOL",
        0x23D0: "SYMBOL",
        0x23D1: "SYMBOL",
        0x23D2: "SYMBOL",
        0x23D3: "SYMBOL",
        0x23D4: "SYMBOL",
        0x23D5: "SYMBOL",
        0x23D6: "SYMBOL",
        0x23D7: "SYMBOL",
        0x23D8: "SYMBOL",
        0x23D9: "SYMBOL",
        0x23DA: "SYMBOL",
        0x23DB: "SYMBOL",
        0x23DC: "SYMBOL",
        0x23DD: "SYMBOL",
        0x23DE: "SYMBOL",
        0x23DF: "SYMBOL",
        0x23E0: "SYMBOL",
        0x23E1: "SYMBOL",
        0x23E2: "SYMBOL",
        0x23E3: "SYMBOL",
        0x23E4: "SYMBOL",
        0x23E5: "SYMBOL",
        0x23E6: "SYMBOL",
        0x23E7: "SYMBOL",
        0x23E8: "SYMBOL",
        0x23E9: "SYMBOL",
        0x23EA: "SYMBOL",
        0x23EB: "SYMBOL",
        0x23EC: "SYMBOL",
        0x23ED: "SYMBOL",
        0x23EE: "SYMBOL",
        0x23EF: "SYMBOL",
        0x23F0: "SYMBOL",
        0x23F1: "SYMBOL",
        0x23F2: "SYMBOL",
        0x23F3: "SYMBOL",
        0x23F4: "SYMBOL",
        0x23F5: "SYMBOL",
        0x23F6: "SYMBOL",
        0x23F7: "SYMBOL",
        0x23F8: "SYMBOL",
        0x23F9: "SYMBOL",
        0x23FA: "SYMBOL",
        0x23FB: "SYMBOL",
        0x23FC: "SYMBOL",
        0x23FD: "SYMBOL",
        0x23FE: "SYMBOL",
        0x23FF: "SYMBOL",
        0x2460: "SYMBOL",
        0x2461: "SYMBOL",
        0x2462: "SYMBOL",
        0x2463: "SYMBOL",
        0x2464: "SYMBOL",
        0x2465: "SYMBOL",
        0x2466: "SYMBOL",
        0x2467: "SYMBOL",
        0x2468: "SYMBOL",
        0x2469: "SYMBOL",
        0x246A: "SYMBOL",
        0x246B: "SYMBOL",
        0x246C: "SYMBOL",
        0x246D: "SYMBOL",
        0x246E: "SYMBOL",
        0x246F: "SYMBOL",
        0x2470: "SYMBOL",
        0x2471: "SYMBOL",
        0x2472: "SYMBOL",
        0x2473: "SYMBOL",
        0x2474: "SYMBOL",
        0x2475: "SYMBOL",
        0x2476: "SYMBOL",
        0x2477: "SYMBOL",
        0x2478: "SYMBOL",
        0x2479: "SYMBOL",
        0x247A: "SYMBOL",
        0x247B: "SYMBOL",
        0x247C: "SYMBOL",
        0x247D: "SYMBOL",
        0x247E: "SYMBOL",
        0x247F: "SYMBOL",
        0x2480: "SYMBOL",
        0x2481: "SYMBOL",
        0x2482: "SYMBOL",
        0x2483: "SYMBOL",
        0x2484: "SYMBOL",
        0x2485: "SYMBOL",
        0x2486: "SYMBOL",
        0x2487: "SYMBOL",
        0x2488: "SYMBOL",
        0x2489: "SYMBOL",
        0x248A: "SYMBOL",
        0x248B: "SYMBOL",
        0x248C: "SYMBOL",
        0x248D: "SYMBOL",
        0x248E: "SYMBOL",
        0x248F: "SYMBOL",
        0x2490: "SYMBOL",
        0x2491: "SYMBOL",
        0x2492: "SYMBOL",
        0x2493: "SYMBOL",
        0x2494: "SYMBOL",
        0x2495: "SYMBOL",
        0x2496: "SYMBOL",
        0x2497: "SYMBOL",
        0x2498: "SYMBOL",
        0x2499: "SYMBOL",
        0x249A: "SYMBOL",
        0x249B: "SYMBOL",
        0x249C: "SYMBOL",
        0x249D: "SYMBOL",
        0x249E: "SYMBOL",
        0x249F: "SYMBOL",
        0x24A0: "SYMBOL",
        0x24A1: "SYMBOL",
        0x24A2: "SYMBOL",
        0x24A3: "SYMBOL",
        0x24A4: "SYMBOL",
        0x24A5: "SYMBOL",
        0x24A6: "SYMBOL",
        0x24A7: "SYMBOL",
        0x24A8: "SYMBOL",
        0x24A9: "SYMBOL",
        0x24AA: "SYMBOL",
        0x24AB: "SYMBOL",
        0x24AC: "SYMBOL",
        0x24AD: "SYMBOL",
        0x24AE: "SYMBOL",
        0x24AF: "SYMBOL",
        0x24B0: "SYMBOL",
        0x24B1: "SYMBOL",
        0x24B2: "SYMBOL",
        0x24B3: "SYMBOL",
        0x24B4: "SYMBOL",
        0x24B5: "SYMBOL",
        0x24B6: "SYMBOL",
        0x24B7: "SYMBOL",
        0x24B8: "SYMBOL",
        0x24B9: "SYMBOL",
        0x24BA: "SYMBOL",
        0x24BB: "SYMBOL",
        0x24BC: "SYMBOL",
        0x24BD: "SYMBOL",
        0x24BE: "SYMBOL",
        0x24BF: "SYMBOL",
        0x24C0: "SYMBOL",
        0x24C1: "SYMBOL",
        0x24C2: "SYMBOL",
        0x24C3: "SYMBOL",
        0x24C4: "SYMBOL",
        0x24C5: "SYMBOL",
        0x24C6: "SYMBOL",
        0x24C7: "SYMBOL",
        0x24C8: "SYMBOL",
        0x24C9: "SYMBOL",
        0x24CA: "SYMBOL",
        0x24CB: "SYMBOL",
        0x24CC: "SYMBOL",
        0x24CD: "SYMBOL",
        0x24CE: "SYMBOL",
        0x24CF: "SYMBOL",
        0x24D0: "SYMBOL",
        0x24D1: "SYMBOL",
        0x24D2: "SYMBOL",
        0x24D3: "SYMBOL",
        0x24D4: "SYMBOL",
        0x24D5: "SYMBOL",
        0x24D6: "SYMBOL",
        0x24D7: "SYMBOL",
        0x24D8: "SYMBOL",
        0x24D9: "SYMBOL",
        0x24DA: "SYMBOL",
        0x24DB: "SYMBOL",
        0x24DC: "SYMBOL",
        0x24DD: "SYMBOL",
        0x24DE: "SYMBOL",
        0x24DF: "SYMBOL",
        0x24E0: "SYMBOL",
        0x24E1: "SYMBOL",
        0x24E2: "SYMBOL",
        0x24E3: "SYMBOL",
        0x24E4: "SYMBOL",
        0x24E5: "SYMBOL",
        0x24E6: "SYMBOL",
        0x24E7: "SYMBOL",
        0x24E8: "SYMBOL",
        0x24E9: "SYMBOL",
        0x24EA: "SYMBOL",
        0x24EB: "SYMBOL",
        0x24EC: "SYMBOL",
        0x24ED: "SYMBOL",
        0x24EE: "SYMBOL",
        0x24EF: "SYMBOL",
        0x24F0: "SYMBOL",
        0x24F1: "SYMBOL",
        0x24F2: "SYMBOL",
        0x24F3: "SYMBOL",
        0x24F4: "SYMBOL",
        0x24F5: "SYMBOL",
        0x24F6: "SYMBOL",
        0x24F7: "SYMBOL",
        0x24F8: "SYMBOL",
        0x24F9: "SYMBOL",
        0x24FA: "SYMBOL",
        0x24FB: "SYMBOL",
        0x24FC: "SYMBOL",
        0x24FD: "SYMBOL",
        0x24FE: "SYMBOL",
        0x24FF: "SYMBOL",
        0x2501: "SYMBOL",
        0x2502: "SYMBOL",
        0x2503: "SYMBOL",
        0x2504: "SYMBOL",
        0x2505: "SYMBOL",
        0x2506: "SYMBOL",
        0x2507: "SYMBOL",
        0x2508: "SYMBOL",
        0x2509: "SYMBOL",
        0x250A: "SYMBOL",
        0x250B: "SYMBOL",
        0x250C: "SYMBOL",
        0x250D: "SYMBOL",
        0x250E: "SYMBOL",
        0x250F: "SYMBOL",
        0x2510: "SYMBOL",
        0x2511: "SYMBOL",
        0x2512: "SYMBOL",
        0x2513: "SYMBOL",
        0x2514: "SYMBOL",
        0x2515: "SYMBOL",
        0x2516: "SYMBOL",
        0x2517: "SYMBOL",
        0x2518: "SYMBOL",
        0x2519: "SYMBOL",
        0x251A: "SYMBOL",
        0x251B: "SYMBOL",
        0x251C: "SYMBOL",
        0x251D: "SYMBOL",
        0x251E: "SYMBOL",
        0x251F: "SYMBOL",
        0x2520: "SYMBOL",
        0x2521: "SYMBOL",
        0x2522: "SYMBOL",
        0x2523: "SYMBOL",
        0x2524: "SYMBOL",
        0x2525: "SYMBOL",
        0x2526: "SYMBOL",
        0x2527: "SYMBOL",
        0x2528: "SYMBOL",
        0x2529: "SYMBOL",
        0x252A: "SYMBOL",
        0x252B: "SYMBOL",
        0x252C: "SYMBOL",
        0x252D: "SYMBOL",
        0x252E: "SYMBOL",
        0x252F: "SYMBOL",
        0x2530: "SYMBOL",
        0x2531: "SYMBOL",
        0x2532: "SYMBOL",
        0x2533: "SYMBOL",
        0x2534: "SYMBOL",
        0x2535: "SYMBOL",
        0x2536: "SYMBOL",
        0x2537: "SYMBOL",
        0x2538: "SYMBOL",
        0x2539: "SYMBOL",
        0x253A: "SYMBOL",
        0x253B: "SYMBOL",
        0x253C: "SYMBOL",
        0x253D: "SYMBOL",
        0x253E: "SYMBOL",
        0x253F: "SYMBOL",
        0x2540: "SYMBOL",
        0x2541: "SYMBOL",
        0x2542: "SYMBOL",
        0x2543: "SYMBOL",
        0x2544: "SYMBOL",
        0x2545: "SYMBOL",
        0x2546: "SYMBOL",
        0x2547: "SYMBOL",
        0x2548: "SYMBOL",
        0x2549: "SYMBOL",
        0x254A: "SYMBOL",
        0x254B: "SYMBOL",
        0x254C: "SYMBOL",
        0x254D: "SYMBOL",
        0x254E: "SYMBOL",
        0x254F: "SYMBOL",
        0x2550: "SYMBOL",
        0x2551: "SYMBOL",
        0x2552: "SYMBOL",
        0x2553: "SYMBOL",
        0x2554: "SYMBOL",
        0x2555: "SYMBOL",
        0x2556: "SYMBOL",
        0x2557: "SYMBOL",
        0x2558: "SYMBOL",
        0x2559: "SYMBOL",
        0x255A: "SYMBOL",
        0x255B: "SYMBOL",
        0x255C: "SYMBOL",
        0x255D: "SYMBOL",
        0x255E: "SYMBOL",
        0x255F: "SYMBOL",
        0x2560: "SYMBOL",
        0x2561: "SYMBOL",
        0x2562: "SYMBOL",
        0x2563: "SYMBOL",
        0x2564: "SYMBOL",
        0x2565: "SYMBOL",
        0x2566: "SYMBOL",
        0x2567: "SYMBOL",
        0x2568: "SYMBOL",
        0x2569: "SYMBOL",
        0x256A: "SYMBOL",
        0x256B: "SYMBOL",
        0x256C: "SYMBOL",
        0x256D: "SYMBOL",
        0x256E: "SYMBOL",
        0x256F: "SYMBOL",
        0x2570: "SYMBOL",
        0x2571: "SYMBOL",
        0x2572: "SYMBOL",
        0x2573: "SYMBOL",
        0x2574: "SYMBOL",
        0x2575: "SYMBOL",
        0x2576: "SYMBOL",
        0x2577: "SYMBOL",
        0x2578: "SYMBOL",
        0x2579: "SYMBOL",
        0x257A: "SYMBOL",
        0x257B: "SYMBOL",
        0x257C: "SYMBOL",
        0x257D: "SYMBOL",
        0x257E: "SYMBOL",
        0x257F: "SYMBOL",
        0x2580: "SYMBOL",
        0x2581: "SYMBOL",
        0x2582: "SYMBOL",
        0x2583: "SYMBOL",
        0x2584: "SYMBOL",
        0x2585: "SYMBOL",
        0x2586: "SYMBOL",
        0x2587: "SYMBOL",
        0x2588: "SYMBOL",
        0x2589: "SYMBOL",
        0x258A: "SYMBOL",
        0x258B: "SYMBOL",
        0x258C: "SYMBOL",
        0x258D: "SYMBOL",
        0x258E: "SYMBOL",
        0x258F: "SYMBOL",
        0x2590: "SYMBOL",
        0x2591: "SYMBOL",
        0x2592: "SYMBOL",
        0x2593: "SYMBOL",
        0x2594: "SYMBOL",
        0x2595: "SYMBOL",
        0x2596: "SYMBOL",
        0x2597: "SYMBOL",
        0x2598: "SYMBOL",
        0x2599: "SYMBOL",
        0x259A: "SYMBOL",
        0x259B: "SYMBOL",
        0x259C: "SYMBOL",
        0x259D: "SYMBOL",
        0x259E: "SYMBOL",
        0x259F: "SYMBOL",
        0x25A0: "SYMBOL",
        0x25A1: "SYMBOL",
        0x25A2: "SYMBOL",
        0x25A3: "SYMBOL",
        0x25A4: "SYMBOL",
        0x25A5: "SYMBOL",
        0x25A6: "SYMBOL",
        0x25A7: "SYMBOL",
        0x25A8: "SYMBOL",
        0x25A9: "SYMBOL",
        0x25AA: "SYMBOL",
        0x25AB: "SYMBOL",
        0x25AC: "SYMBOL",
        0x25AD: "SYMBOL",
        0x25AE: "SYMBOL",
        0x25AF: "SYMBOL",
        0x25B0: "SYMBOL",
        0x25B1: "SYMBOL",
        0x25B2: "SYMBOL",
        0x25B3: "SYMBOL",
        0x25B4: "SYMBOL",
        0x25B5: "SYMBOL",
        0x25B6: "SYMBOL",
        0x25B7: "SYMBOL",
        0x25B8: "SYMBOL",
        0x25B9: "SYMBOL",
        0x25BA: "SYMBOL",
        0x25BB: "SYMBOL",
        0x25BC: "SYMBOL",
        0x25BD: "SYMBOL",
        0x25BE: "SYMBOL",
        0x25BF: "SYMBOL",
        0x25C0: "SYMBOL",
        0x25C1: "SYMBOL",
        0x25C2: "SYMBOL",
        0x25C3: "SYMBOL",
        0x25C4: "SYMBOL",
        0x25C5: "SYMBOL",
        0x25C6: "SYMBOL",
        0x25C7: "SYMBOL",
        0x25C8: "SYMBOL",
        0x25C9: "SYMBOL",
        0x25CA: "SYMBOL",
        0x25CB: "SYMBOL",
        0x25CC: "SYMBOL",
        0x25CD: "SYMBOL",
        0x25CE: "SYMBOL",
        0x25CF: "SYMBOL",
        0x25D0: "SYMBOL",
        0x25D1: "SYMBOL",
        0x25D2: "SYMBOL",
        0x25D3: "SYMBOL",
        0x25D4: "SYMBOL",
        0x25D5: "SYMBOL",
        0x25D6: "SYMBOL",
        0x25D7: "SYMBOL",
        0x25D8: "SYMBOL",
        0x25D9: "SYMBOL",
        0x25DA: "SYMBOL",
        0x25DB: "SYMBOL",
        0x25DC: "SYMBOL",
        0x25DD: "SYMBOL",
        0x25DE: "SYMBOL",
        0x25DF: "SYMBOL",
        0x25E0: "SYMBOL",
        0x25E1: "SYMBOL",
        0x25E2: "SYMBOL",
        0x25E3: "SYMBOL",
        0x25E4: "SYMBOL",
        0x25E5: "SYMBOL",
        0x25E6: "SYMBOL",
        0x25E7: "SYMBOL",
        0x25E8: "SYMBOL",
        0x25E9: "SYMBOL",
        0x25EA: "SYMBOL",
        0x25EB: "SYMBOL",
        0x25EC: "SYMBOL",
        0x25ED: "SYMBOL",
        0x25EE: "SYMBOL",
        0x25EF: "SYMBOL",
        0x25F0: "SYMBOL",
        0x25F1: "SYMBOL",
        0x25F2: "SYMBOL",
        0x25F3: "SYMBOL",
        0x25F4: "SYMBOL",
        0x25F5: "SYMBOL",
        0x25F6: "SYMBOL",
        0x25F7: "SYMBOL",
        0x25F8: "SYMBOL",
        0x25F9: "SYMBOL",
        0x25FA: "SYMBOL",
        0x25FB: "SYMBOL",
        0x25FC: "SYMBOL",
        0x25FD: "SYMBOL",
        0x25FE: "SYMBOL",
        0x25FF: "SYMBOL",
        0x2600: "SYMBOL",
        0x2601: "SYMBOL",
        0x2602: "SYMBOL",
        0x2603: "SYMBOL",
        0x2604: "SYMBOL",
        0x2605: "SYMBOL",
        0x2606: "SYMBOL",
        0x2607: "SYMBOL",
        0x2608: "SYMBOL",
        0x2609: "SYMBOL",
        0x260A: "SYMBOL",
        0x260B: "SYMBOL",
        0x260C: "SYMBOL",
        0x260D: "SYMBOL",
        0x260E: "SYMBOL",
        0x260F: "SYMBOL",
        0x2610: "SYMBOL",
        0x2611: "SYMBOL",
        0x2612: "SYMBOL",
        0x2613: "SYMBOL",
        0x2614: "SYMBOL",
        0x2615: "SYMBOL",
        0x2616: "SYMBOL",
        0x2617: "SYMBOL",
        0x2618: "SYMBOL",
        0x2619: "SYMBOL",
        0x261A: "SYMBOL",
        0x261B: "SYMBOL",
        0x261C: "SYMBOL",
        0x261D: "SYMBOL",
        0x261E: "SYMBOL",
        0x261F: "SYMBOL",
        0x2620: "SYMBOL",
        0x2621: "SYMBOL",
        0x2622: "SYMBOL",
        0x2623: "SYMBOL",
        0x2624: "SYMBOL",
        0x2625: "SYMBOL",
        0x2626: "SYMBOL",
        0x2627: "SYMBOL",
        0x2628: "SYMBOL",
        0x2629: "SYMBOL",
        0x262A: "SYMBOL",
        0x262B: "SYMBOL",
        0x262C: "SYMBOL",
        0x262D: "SYMBOL",
        0x262E: "SYMBOL",
        0x262F: "SYMBOL",
        0x2630: "SYMBOL",
        0x2631: "SYMBOL",
        0x2632: "SYMBOL",
        0x2633: "SYMBOL",
        0x2634: "SYMBOL",
        0x2635: "SYMBOL",
        0x2636: "SYMBOL",
        0x2637: "SYMBOL",
        0x2638: "SYMBOL",
        0x2639: "SYMBOL",
        0x263A: "SYMBOL",
        0x263B: "SYMBOL",
        0x263C: "SYMBOL",
        0x263D: "SYMBOL",
        0x263E: "SYMBOL",
        0x263F: "SYMBOL",
        0x2640: "SYMBOL",
        0x2641: "SYMBOL",
        0x2642: "SYMBOL",
        0x2643: "SYMBOL",
        0x2644: "SYMBOL",
        0x2645: "SYMBOL",
        0x2646: "SYMBOL",
        0x2647: "SYMBOL",
        0x2648: "SYMBOL",
        0x2649: "SYMBOL",
        0x264A: "SYMBOL",
        0x264B: "SYMBOL",
        0x264C: "SYMBOL",
        0x264D: "SYMBOL",
        0x264E: "SYMBOL",
        0x264F: "SYMBOL",
        0x2650: "SYMBOL",
        0x2651: "SYMBOL",
        0x2652: "SYMBOL",
        0x2653: "SYMBOL",
        0x2654: "SYMBOL",
        0x2655: "SYMBOL",
        0x2656: "SYMBOL",
        0x2657: "SYMBOL",
        0x2658: "SYMBOL",
        0x2659: "SYMBOL",
        0x265A: "SYMBOL",
        0x265B: "SYMBOL",
        0x265C: "SYMBOL",
        0x265D: "SYMBOL",
        0x265E: "SYMBOL",
        0x265F: "SYMBOL",
        0x2660: "SYMBOL",
        0x2661: "SYMBOL",
        0x2662: "SYMBOL",
        0x2663: "SYMBOL",
        0x2664: "SYMBOL",
        0x2665: "SYMBOL",
        0x2666: "SYMBOL",
        0x2667: "SYMBOL",
        0x2668: "SYMBOL",
        0x2669: "SYMBOL",
        0x266A: "SYMBOL",
        0x266B: "SYMBOL",
        0x266C: "SYMBOL",
        0x266D: "SYMBOL",
        0x266E: "SYMBOL",
        0x266F: "SYMBOL",
        0x2670: "SYMBOL",
        0x2671: "SYMBOL",
        0x2672: "SYMBOL",
        0x2673: "SYMBOL",
        0x2674: "SYMBOL",
        0x2675: "SYMBOL",
        0x2676: "SYMBOL",
        0x2677: "SYMBOL",
        0x2678: "SYMBOL",
        0x2679: "SYMBOL",
        0x267A: "SYMBOL",
        0x267B: "SYMBOL",
        0x267C: "SYMBOL",
        0x267D: "SYMBOL",
        0x267E: "SYMBOL",
        0x267F: "SYMBOL",
        0x2680: "SYMBOL",
        0x2681: "SYMBOL",
        0x2682: "SYMBOL",
        0x2683: "SYMBOL",
        0x2684: "SYMBOL",
        0x2685: "SYMBOL",
        0x2686: "SYMBOL",
        0x2687: "SYMBOL",
        0x2688: "SYMBOL",
        0x2689: "SYMBOL",
        0x268A: "SYMBOL",
        0x268B: "SYMBOL",
        0x268C: "SYMBOL",
        0x268D: "SYMBOL",
        0x268E: "SYMBOL",
        0x268F: "SYMBOL",
        0x2690: "SYMBOL",
        0x2691: "SYMBOL",
        0x2692: "SYMBOL",
        0x2693: "SYMBOL",
        0x2694: "SYMBOL",
        0x2695: "SYMBOL",
        0x2696: "SYMBOL",
        0x2697: "SYMBOL",
        0x2698: "SYMBOL",
        0x2699: "SYMBOL",
        0x269A: "SYMBOL",
        0x269B: "SYMBOL",
        0x269C: "SYMBOL",
        0x269D: "SYMBOL",
        0x269E: "SYMBOL",
        0x269F: "SYMBOL",
        0x26A0: "SYMBOL",
        0x26A1: "SYMBOL",
        0x26A2: "SYMBOL",
        0x26A3: "SYMBOL",
        0x26A4: "SYMBOL",
        0x26A5: "SYMBOL",
        0x26A6: "SYMBOL",
        0x26A7: "SYMBOL",
        0x26A8: "SYMBOL",
        0x26A9: "SYMBOL",
        0x26AA: "SYMBOL",
        0x26AB: "SYMBOL",
        0x26AC: "SYMBOL",
        0x26AD: "SYMBOL",
        0x26AE: "SYMBOL",
        0x26AF: "SYMBOL",
        0x26B0: "SYMBOL",
        0x26B1: "SYMBOL",
        0x26B2: "SYMBOL",
        0x26B3: "SYMBOL",
        0x26B4: "SYMBOL",
        0x26B5: "SYMBOL",
        0x26B6: "SYMBOL",
        0x26B7: "SYMBOL",
        0x26B8: "SYMBOL",
        0x26B9: "SYMBOL",
        0x26BA: "SYMBOL",
        0x26BB: "SYMBOL",
        0x26BC: "SYMBOL",
        0x26BD: "SYMBOL",
        0x26BE: "SYMBOL",
        0x26BF: "SYMBOL",
        0x26C0: "SYMBOL",
        0x26C1: "SYMBOL",
        0x26C2: "SYMBOL",
        0x26C3: "SYMBOL",
        0x26C4: "SYMBOL",
        0x26C5: "SYMBOL",
        0x26C6: "SYMBOL",
        0x26C7: "SYMBOL",
        0x26C8: "SYMBOL",
        0x26C9: "SYMBOL",
        0x26CA: "SYMBOL",
        0x26CB: "SYMBOL",
        0x26CC: "SYMBOL",
        0x26CD: "SYMBOL",
        0x26CE: "SYMBOL",
        0x26CF: "SYMBOL",
        0x26D0: "SYMBOL",
        0x26D1: "SYMBOL",
        0x26D2: "SYMBOL",
        0x26D3: "SYMBOL",
        0x26D4: "SYMBOL",
        0x26D5: "SYMBOL",
        0x26D6: "SYMBOL",
        0x26D7: "SYMBOL",
        0x26D8: "SYMBOL",
        0x26D9: "SYMBOL",
        0x26DA: "SYMBOL",
        0x26DB: "SYMBOL",
        0x26DC: "SYMBOL",
        0x26DD: "SYMBOL",
        0x26DE: "SYMBOL",
        0x26DF: "SYMBOL",
        0x26E0: "SYMBOL",
        0x26E1: "SYMBOL",
        0x26E2: "SYMBOL",
        0x26E3: "SYMBOL",
        0x26E4: "SYMBOL",
        0x26E5: "SYMBOL",
        0x26E6: "SYMBOL",
        0x26E7: "SYMBOL",
        0x26E8: "SYMBOL",
        0x26E9: "SYMBOL",
        0x26EA: "SYMBOL",
        0x26EB: "SYMBOL",
        0x26EC: "SYMBOL",
        0x26ED: "SYMBOL",
        0x26EE: "SYMBOL",
        0x26EF: "SYMBOL",
        0x26F0: "SYMBOL",
        0x26F1: "SYMBOL",
        0x26F2: "SYMBOL",
        0x26F3: "SYMBOL",
        0x26F4: "SYMBOL",
        0x26F5: "SYMBOL",
        0x26F6: "SYMBOL",
        0x26F7: "SYMBOL",
        0x26F8: "SYMBOL",
        0x26F9: "SYMBOL",
        0x26FA: "SYMBOL",
        0x26FB: "SYMBOL",
        0x26FC: "SYMBOL",
        0x26FD: "SYMBOL",
        0x26FE: "SYMBOL",
        0x2700: "SYMBOL",
        0x2701: "SYMBOL",
        0x2702: "SYMBOL",
        0x2703: "SYMBOL",
        0x2704: "SYMBOL",
        0x2705: "SYMBOL",
        0x2706: "SYMBOL",
        0x2707: "SYMBOL",
        0x2708: "SYMBOL",
        0x2709: "SYMBOL",
        0x270A: "SYMBOL",
        0x270B: "SYMBOL",
        0x270C: "SYMBOL",
        0x270D: "SYMBOL",
        0x270E: "SYMBOL",
        0x270F: "SYMBOL",
        0x2710: "SYMBOL",
        0x2711: "SYMBOL",
        0x2712: "SYMBOL",
        0x2713: "SYMBOL",
        0x2714: "SYMBOL",
        0x2715: "SYMBOL",
        0x2716: "SYMBOL",
        0x2717: "SYMBOL",
        0x2718: "SYMBOL",
        0x2719: "SYMBOL",
        0x271A: "SYMBOL",
        0x271B: "SYMBOL",
        0x271C: "SYMBOL",
        0x271D: "SYMBOL",
        0x271E: "SYMBOL",
        0x271F: "SYMBOL",
        0x2720: "SYMBOL",
        0x2721: "SYMBOL",
        0x2722: "SYMBOL",
        0x2723: "SYMBOL",
        0x2724: "SYMBOL",
        0x2725: "SYMBOL",
        0x2726: "SYMBOL",
        0x2727: "SYMBOL",
        0x2728: "SYMBOL",
        0x2729: "SYMBOL",
        0x272A: "SYMBOL",
        0x272B: "SYMBOL",
        0x272C: "SYMBOL",
        0x272D: "SYMBOL",
        0x272E: "SYMBOL",
        0x272F: "SYMBOL",
        0x2730: "SYMBOL",
        0x2731: "SYMBOL",
        0x2732: "SYMBOL",
        0x2733: "SYMBOL",
        0x2734: "SYMBOL",
        0x2735: "SYMBOL",
        0x2736: "SYMBOL",
        0x2737: "SYMBOL",
        0x2738: "SYMBOL",
        0x2739: "SYMBOL",
        0x273A: "SYMBOL",
        0x273B: "SYMBOL",
        0x273C: "SYMBOL",
        0x273D: "SYMBOL",
        0x273E: "SYMBOL",
        0x273F: "SYMBOL",
        0x2740: "SYMBOL",
        0x2741: "SYMBOL",
        0x2742: "SYMBOL",
        0x2743: "SYMBOL",
        0x2744: "SYMBOL",
        0x2745: "SYMBOL",
        0x2746: "SYMBOL",
        0x2747: "SYMBOL",
        0x2748: "SYMBOL",
        0x2749: "SYMBOL",
        0x274A: "SYMBOL",
        0x274B: "SYMBOL",
        0x274C: "SYMBOL",
        0x274D: "SYMBOL",
        0x274E: "SYMBOL",
        0x274F: "SYMBOL",
        0x2750: "SYMBOL",
        0x2751: "SYMBOL",
        0x2752: "SYMBOL",
        0x2753: "SYMBOL",
        0x2754: "SYMBOL",
        0x2755: "SYMBOL",
        0x2756: "SYMBOL",
        0x2757: "SYMBOL",
        0x2758: "SYMBOL",
        0x2759: "SYMBOL",
        0x275A: "SYMBOL",
        0x275B: "SYMBOL",
        0x275C: "SYMBOL",
        0x275D: "SYMBOL",
        0x275E: "SYMBOL",
        0x275F: "SYMBOL",
        0x2760: "SYMBOL",
        0x2761: "SYMBOL",
        0x2762: "SYMBOL",
        0x2763: "SYMBOL",
        0x2764: "SYMBOL",
        0x2765: "SYMBOL",
        0x2766: "SYMBOL",
        0x2767: "SYMBOL",
        0x2768: "SYMBOL",
        0x2769: "SYMBOL",
        0x276A: "SYMBOL",
        0x276B: "SYMBOL",
        0x276C: "SYMBOL",
        0x276D: "SYMBOL",
        0x276E: "SYMBOL",
        0x276F: "SYMBOL",
        0x2770: "SYMBOL",
        0x2771: "SYMBOL",
        0x2772: "SYMBOL",
        0x2773: "SYMBOL",
        0x2774: "SYMBOL",
        0x2775: "SYMBOL",
        0x2776: "SYMBOL",
        0x2777: "SYMBOL",
        0x2778: "SYMBOL",
        0x2779: "SYMBOL",
        0x277A: "SYMBOL",
        0x277B: "SYMBOL",
        0x277C: "SYMBOL",
        0x277D: "SYMBOL",
        0x277E: "SYMBOL",
        0x277F: "SYMBOL",
        0x2780: "SYMBOL",
        0x2781: "SYMBOL",
        0x2782: "SYMBOL",
        0x2783: "SYMBOL",
        0x2784: "SYMBOL",
        0x2785: "SYMBOL",
        0x2786: "SYMBOL",
        0x2787: "SYMBOL",
        0x2788: "SYMBOL",
        0x2789: "SYMBOL",
        0x278A: "SYMBOL",
        0x278B: "SYMBOL",
        0x278C: "SYMBOL",
        0x278D: "SYMBOL",
        0x278E: "SYMBOL",
        0x278F: "SYMBOL",
        0x2790: "SYMBOL",
        0x2791: "SYMBOL",
        0x2792: "SYMBOL",
        0x2793: "SYMBOL",
        0x2794: "SYMBOL",
        0x2795: "SYMBOL",
        0x2796: "SYMBOL",
        0x2797: "SYMBOL",
        0x2798: "SYMBOL",
        0x2799: "SYMBOL",
        0x279A: "SYMBOL",
        0x279B: "SYMBOL",
        0x279C: "SYMBOL",
        0x279D: "SYMBOL",
        0x279E: "SYMBOL",
        0x279F: "SYMBOL",
        0x27A0: "SYMBOL",
        0x27A1: "SYMBOL",
        0x27A2: "SYMBOL",
        0x27A3: "SYMBOL",
        0x27A4: "SYMBOL",
        0x27A5: "SYMBOL",
        0x27A6: "SYMBOL",
        0x27A7: "SYMBOL",
        0x27A8: "SYMBOL",
        0x27A9: "SYMBOL",
        0x27AA: "SYMBOL",
        0x27AB: "SYMBOL",
        0x27AC: "SYMBOL",
        0x27AD: "SYMBOL",
        0x27AE: "SYMBOL",
        0x27AF: "SYMBOL",
        0x27B0: "SYMBOL",
        0x27B1: "SYMBOL",
        0x27B2: "SYMBOL",
        0x27B3: "SYMBOL",
        0x27B4: "SYMBOL",
        0x27B5: "SYMBOL",
        0x27B6: "SYMBOL",
        0x27B7: "SYMBOL",
        0x27B8: "SYMBOL",
        0x27B9: "SYMBOL",
        0x27BA: "SYMBOL",
        0x27BB: "SYMBOL",
        0x27BC: "SYMBOL",
        0x27BD: "SYMBOL",
        0x27BE: "SYMBOL",
        0x27BF: "SYMBOL",
        0x27F0: "SYMBOL",
        0x27F1: "SYMBOL",
        0x27F2: "SYMBOL",
        0x27F3: "SYMBOL",
        0x27F4: "SYMBOL",
        0x27F5: "SYMBOL",
        0x27F6: "SYMBOL",
        0x27F7: "SYMBOL",
        0x27F8: "SYMBOL",
        0x27F9: "SYMBOL",
        0x27FA: "SYMBOL",
        0x27FB: "SYMBOL",
        0x27FC: "SYMBOL",
        0x27FD: "SYMBOL",
        0x27FE: "SYMBOL",
        0x27FF: "SYMBOL",
        0x27C0: "SYMBOL",
        0x27C1: "SYMBOL",
        0x27C2: "SYMBOL",
        0x27C3: "SYMBOL",
        0x27C4: "SYMBOL",
        0x27C5: "SYMBOL",
        0x27C6: "SYMBOL",
        0x27C7: "SYMBOL",
        0x27C8: "SYMBOL",
        0x27C9: "SYMBOL",
        0x27CA: "SYMBOL",
        0x27CB: "SYMBOL",
        0x27CC: "SYMBOL",
        0x27CD: "SYMBOL",
        0x27CE: "SYMBOL",
        0x27CF: "SYMBOL",
        0x27D0: "SYMBOL",
        0x27D1: "SYMBOL",
        0x27D2: "SYMBOL",
        0x27D3: "SYMBOL",
        0x27D4: "SYMBOL",
        0x27D5: "SYMBOL",
        0x27D6: "SYMBOL",
        0x27D7: "SYMBOL",
        0x27D8: "SYMBOL",
        0x27D9: "SYMBOL",
        0x27DA: "SYMBOL",
        0x27DB: "SYMBOL",
        0x27DC: "SYMBOL",
        0x27DD: "SYMBOL",
        0x27DE: "SYMBOL",
        0x27DF: "SYMBOL",
        0x27E0: "SYMBOL",
        0x27E1: "SYMBOL",
        0x27E2: "SYMBOL",
        0x27E3: "SYMBOL",
        0x27E4: "SYMBOL",
        0x27E5: "SYMBOL",
        0x27E6: "SYMBOL",
        0x27E7: "SYMBOL",
        0x27E8: "SYMBOL",
        0x27E9: "SYMBOL",
        0x27EA: "SYMBOL",
        0x27EB: "SYMBOL",
        0x27EC: "SYMBOL",
        0x27ED: "SYMBOL",
        0x27EE: "SYMBOL",
        0x27EF: "SYMBOL",
        0x2800: "SYMBOL",
        0x2801: "SYMBOL",
        0x2802: "SYMBOL",
        0x2803: "SYMBOL",
        0x2804: "SYMBOL",
        0x2805: "SYMBOL",
        0x2806: "SYMBOL",
        0x2807: "SYMBOL",
        0x2808: "SYMBOL",
        0x2809: "SYMBOL",
        0x280A: "SYMBOL",
        0x280B: "SYMBOL",
        0x280C: "SYMBOL",
        0x280D: "SYMBOL",
        0x280E: "SYMBOL",
        0x280F: "SYMBOL",
        0x2810: "SYMBOL",
        0x2811: "SYMBOL",
        0x2812: "SYMBOL",
        0x2813: "SYMBOL",
        0x2814: "SYMBOL",
        0x2815: "SYMBOL",
        0x2816: "SYMBOL",
        0x2817: "SYMBOL",
        0x2818: "SYMBOL",
        0x2819: "SYMBOL",
        0x281A: "SYMBOL",
        0x281B: "SYMBOL",
        0x281C: "SYMBOL",
        0x281D: "SYMBOL",
        0x281E: "SYMBOL",
        0x281F: "SYMBOL",
        0x2820: "SYMBOL",
        0x2821: "SYMBOL",
        0x2822: "SYMBOL",
        0x2823: "SYMBOL",
        0x2824: "SYMBOL",
        0x2825: "SYMBOL",
        0x2826: "SYMBOL",
        0x2827: "SYMBOL",
        0x2828: "SYMBOL",
        0x2829: "SYMBOL",
        0x282A: "SYMBOL",
        0x282B: "SYMBOL",
        0x282C: "SYMBOL",
        0x282D: "SYMBOL",
        0x282E: "SYMBOL",
        0x282F: "SYMBOL",
        0x2830: "SYMBOL",
        0x2831: "SYMBOL",
        0x2832: "SYMBOL",
        0x2833: "SYMBOL",
        0x2834: "SYMBOL",
        0x2835: "SYMBOL",
        0x2836: "SYMBOL",
        0x2837: "SYMBOL",
        0x2838: "SYMBOL",
        0x2839: "SYMBOL",
        0x283A: "SYMBOL",
        0x283B: "SYMBOL",
        0x283C: "SYMBOL",
        0x283D: "SYMBOL",
        0x283E: "SYMBOL",
        0x283F: "SYMBOL",
        0x2840: "SYMBOL",
        0x2841: "SYMBOL",
        0x2842: "SYMBOL",
        0x2843: "SYMBOL",
        0x2844: "SYMBOL",
        0x2845: "SYMBOL",
        0x2846: "SYMBOL",
        0x2847: "SYMBOL",
        0x2848: "SYMBOL",
        0x2849: "SYMBOL",
        0x284A: "SYMBOL",
        0x284B: "SYMBOL",
        0x284C: "SYMBOL",
        0x284D: "SYMBOL",
        0x284E: "SYMBOL",
        0x284F: "SYMBOL",
        0x2850: "SYMBOL",
        0x2851: "SYMBOL",
        0x2852: "SYMBOL",
        0x2853: "SYMBOL",
        0x2854: "SYMBOL",
        0x2855: "SYMBOL",
        0x2856: "SYMBOL",
        0x2857: "SYMBOL",
        0x2858: "SYMBOL",
        0x2859: "SYMBOL",
        0x285A: "SYMBOL",
        0x285B: "SYMBOL",
        0x285C: "SYMBOL",
        0x285D: "SYMBOL",
        0x285E: "SYMBOL",
        0x285F: "SYMBOL",
        0x2860: "SYMBOL",
        0x2861: "SYMBOL",
        0x2862: "SYMBOL",
        0x2863: "SYMBOL",
        0x2864: "SYMBOL",
        0x2865: "SYMBOL",
        0x2866: "SYMBOL",
        0x2867: "SYMBOL",
        0x2868: "SYMBOL",
        0x2869: "SYMBOL",
        0x286A: "SYMBOL",
        0x286B: "SYMBOL",
        0x286C: "SYMBOL",
        0x286D: "SYMBOL",
        0x286E: "SYMBOL",
        0x286F: "SYMBOL",
        0x2870: "SYMBOL",
        0x2871: "SYMBOL",
        0x2872: "SYMBOL",
        0x2873: "SYMBOL",
        0x2874: "SYMBOL",
        0x2875: "SYMBOL",
        0x2876: "SYMBOL",
        0x2877: "SYMBOL",
        0x2878: "SYMBOL",
        0x2879: "SYMBOL",
        0x287A: "SYMBOL",
        0x287B: "SYMBOL",
        0x287C: "SYMBOL",
        0x287D: "SYMBOL",
        0x287E: "SYMBOL",
        0x287F: "SYMBOL",
        0x2880: "SYMBOL",
        0x2881: "SYMBOL",
        0x2882: "SYMBOL",
        0x2883: "SYMBOL",
        0x2884: "SYMBOL",
        0x2885: "SYMBOL",
        0x2886: "SYMBOL",
        0x2887: "SYMBOL",
        0x2888: "SYMBOL",
        0x2889: "SYMBOL",
        0x288A: "SYMBOL",
        0x288B: "SYMBOL",
        0x288C: "SYMBOL",
        0x288D: "SYMBOL",
        0x288E: "SYMBOL",
        0x288F: "SYMBOL",
        0x2890: "SYMBOL",
        0x2891: "SYMBOL",
        0x2892: "SYMBOL",
        0x2893: "SYMBOL",
        0x2894: "SYMBOL",
        0x2895: "SYMBOL",
        0x2896: "SYMBOL",
        0x2897: "SYMBOL",
        0x2898: "SYMBOL",
        0x2899: "SYMBOL",
        0x289A: "SYMBOL",
        0x289B: "SYMBOL",
        0x289C: "SYMBOL",
        0x289D: "SYMBOL",
        0x289E: "SYMBOL",
        0x289F: "SYMBOL",
        0x28A0: "SYMBOL",
        0x28A1: "SYMBOL",
        0x28A2: "SYMBOL",
        0x28A3: "SYMBOL",
        0x28A4: "SYMBOL",
        0x28A5: "SYMBOL",
        0x28A6: "SYMBOL",
        0x28A7: "SYMBOL",
        0x28A8: "SYMBOL",
        0x28A9: "SYMBOL",
        0x28AA: "SYMBOL",
        0x28AB: "SYMBOL",
        0x28AC: "SYMBOL",
        0x28AD: "SYMBOL",
        0x28AE: "SYMBOL",
        0x28AF: "SYMBOL",
        0x28B0: "SYMBOL",
        0x28B1: "SYMBOL",
        0x28B2: "SYMBOL",
        0x28B3: "SYMBOL",
        0x28B4: "SYMBOL",
        0x28B5: "SYMBOL",
        0x28B6: "SYMBOL",
        0x28B7: "SYMBOL",
        0x28B8: "SYMBOL",
        0x28B9: "SYMBOL",
        0x28BA: "SYMBOL",
        0x28BB: "SYMBOL",
        0x28BC: "SYMBOL",
        0x28BD: "SYMBOL",
        0x28BE: "SYMBOL",
        0x28BF: "SYMBOL",
        0x28C0: "SYMBOL",
        0x28C1: "SYMBOL",
        0x28C2: "SYMBOL",
        0x28C3: "SYMBOL",
        0x28C4: "SYMBOL",
        0x28C5: "SYMBOL",
        0x28C6: "SYMBOL",
        0x28C7: "SYMBOL",
        0x28C8: "SYMBOL",
        0x28C9: "SYMBOL",
        0x28CA: "SYMBOL",
        0x28CB: "SYMBOL",
        0x28CC: "SYMBOL",
        0x28CD: "SYMBOL",
        0x28CE: "SYMBOL",
        0x28CF: "SYMBOL",
        0x28D0: "SYMBOL",
        0x28D1: "SYMBOL",
        0x28D2: "SYMBOL",
        0x28D3: "SYMBOL",
        0x28D4: "SYMBOL",
        0x28D5: "SYMBOL",
        0x28D6: "SYMBOL",
        0x28D7: "SYMBOL",
        0x28D8: "SYMBOL",
        0x28D9: "SYMBOL",
        0x28DA: "SYMBOL",
        0x28DB: "SYMBOL",
        0x28DC: "SYMBOL",
        0x28DD: "SYMBOL",
        0x28DE: "SYMBOL",
        0x28DF: "SYMBOL",
        0x28E0: "SYMBOL",
        0x28E1: "SYMBOL",
        0x28E2: "SYMBOL",
        0x28E3: "SYMBOL",
        0x28E4: "SYMBOL",
        0x28E5: "SYMBOL",
        0x28E6: "SYMBOL",
        0x28E7: "SYMBOL",
        0x28E8: "SYMBOL",
        0x28E9: "SYMBOL",
        0x28EA: "SYMBOL",
        0x28EB: "SYMBOL",
        0x28EC: "SYMBOL",
        0x28ED: "SYMBOL",
        0x28EE: "SYMBOL",
        0x28EF: "SYMBOL",
        0x28F0: "SYMBOL",
        0x28F1: "SYMBOL",
        0x28F2: "SYMBOL",
        0x28F3: "SYMBOL",
        0x28F4: "SYMBOL",
        0x28F5: "SYMBOL",
        0x28F6: "SYMBOL",
        0x28F7: "SYMBOL",
        0x28F8: "SYMBOL",
        0x28F9: "SYMBOL",
        0x28FA: "SYMBOL",
        0x28FB: "SYMBOL",
        0x28FC: "SYMBOL",
        0x28FD: "SYMBOL",
        0x28FE: "SYMBOL",
        0x28FF: "SYMBOL",
        0x2900: "SYMBOL",
        0x2901: "SYMBOL",
        0x2902: "SYMBOL",
        0x2903: "SYMBOL",
        0x2904: "SYMBOL",
        0x2905: "SYMBOL",
        0x2906: "SYMBOL",
        0x2907: "SYMBOL",
        0x2908: "SYMBOL",
        0x2909: "SYMBOL",
        0x290A: "SYMBOL",
        0x290B: "SYMBOL",
        0x290C: "SYMBOL",
        0x290D: "SYMBOL",
        0x290E: "SYMBOL",
        0x290F: "SYMBOL",
        0x2910: "SYMBOL",
        0x2911: "SYMBOL",
        0x2912: "SYMBOL",
        0x2913: "SYMBOL",
        0x2914: "SYMBOL",
        0x2915: "SYMBOL",
        0x2916: "SYMBOL",
        0x2917: "SYMBOL",
        0x2918: "SYMBOL",
        0x2919: "SYMBOL",
        0x291A: "SYMBOL",
        0x291B: "SYMBOL",
        0x291C: "SYMBOL",
        0x291D: "SYMBOL",
        0x291E: "SYMBOL",
        0x291F: "SYMBOL",
        0x2920: "SYMBOL",
        0x2921: "SYMBOL",
        0x2922: "SYMBOL",
        0x2923: "SYMBOL",
        0x2924: "SYMBOL",
        0x2925: "SYMBOL",
        0x2926: "SYMBOL",
        0x2927: "SYMBOL",
        0x2928: "SYMBOL",
        0x2929: "SYMBOL",
        0x292A: "SYMBOL",
        0x292B: "SYMBOL",
        0x292C: "SYMBOL",
        0x292D: "SYMBOL",
        0x292E: "SYMBOL",
        0x292F: "SYMBOL",
        0x2930: "SYMBOL",
        0x2931: "SYMBOL",
        0x2932: "SYMBOL",
        0x2933: "SYMBOL",
        0x2934: "SYMBOL",
        0x2935: "SYMBOL",
        0x2936: "SYMBOL",
        0x2937: "SYMBOL",
        0x2938: "SYMBOL",
        0x2939: "SYMBOL",
        0x293A: "SYMBOL",
        0x293B: "SYMBOL",
        0x293C: "SYMBOL",
        0x293D: "SYMBOL",
        0x293E: "SYMBOL",
        0x293F: "SYMBOL",
        0x2940: "SYMBOL",
        0x2941: "SYMBOL",
        0x2942: "SYMBOL",
        0x2943: "SYMBOL",
        0x2944: "SYMBOL",
        0x2945: "SYMBOL",
        0x2946: "SYMBOL",
        0x2947: "SYMBOL",
        0x2948: "SYMBOL",
        0x2949: "SYMBOL",
        0x294A: "SYMBOL",
        0x294B: "SYMBOL",
        0x294C: "SYMBOL",
        0x294D: "SYMBOL",
        0x294E: "SYMBOL",
        0x294F: "SYMBOL",
        0x2950: "SYMBOL",
        0x2951: "SYMBOL",
        0x2952: "SYMBOL",
        0x2953: "SYMBOL",
        0x2954: "SYMBOL",
        0x2955: "SYMBOL",
        0x2956: "SYMBOL",
        0x2957: "SYMBOL",
        0x2958: "SYMBOL",
        0x2959: "SYMBOL",
        0x295A: "SYMBOL",
        0x295B: "SYMBOL",
        0x295C: "SYMBOL",
        0x295D: "SYMBOL",
        0x295E: "SYMBOL",
        0x295F: "SYMBOL",
        0x2960: "SYMBOL",
        0x2961: "SYMBOL",
        0x2962: "SYMBOL",
        0x2963: "SYMBOL",
        0x2964: "SYMBOL",
        0x2965: "SYMBOL",
        0x2966: "SYMBOL",
        0x2967: "SYMBOL",
        0x2968: "SYMBOL",
        0x2969: "SYMBOL",
        0x296A: "SYMBOL",
        0x296B: "SYMBOL",
        0x296C: "SYMBOL",
        0x296D: "SYMBOL",
        0x296E: "SYMBOL",
        0x296F: "SYMBOL",
        0x2970: "SYMBOL",
        0x2971: "SYMBOL",
        0x2972: "SYMBOL",
        0x2973: "SYMBOL",
        0x2974: "SYMBOL",
        0x2975: "SYMBOL",
        0x2976: "SYMBOL",
        0x2977: "SYMBOL",
        0x2978: "SYMBOL",
        0x2979: "SYMBOL",
        0x297A: "SYMBOL",
        0x297B: "SYMBOL",
        0x297C: "SYMBOL",
        0x297D: "SYMBOL",
        0x297E: "SYMBOL",
        0x297F: "SYMBOL",
        0x2B00: "SYMBOL",
        0x2B01: "SYMBOL",
        0x2B02: "SYMBOL",
        0x2B03: "SYMBOL",
        0x2B04: "SYMBOL",
        0x2B05: "SYMBOL",
        0x2B06: "SYMBOL",
        0x2B07: "SYMBOL",
        0x2B08: "SYMBOL",
        0x2B09: "SYMBOL",
        0x2B0A: "SYMBOL",
        0x2B0B: "SYMBOL",
        0x2B0C: "SYMBOL",
        0x2B0D: "SYMBOL",
        0x2B0E: "SYMBOL",
        0x2B0F: "SYMBOL",
        0x2B10: "SYMBOL",
        0x2B11: "SYMBOL",
        0x2B12: "SYMBOL",
        0x2B13: "SYMBOL",
        0x2B14: "SYMBOL",
        0x2B15: "SYMBOL",
        0x2B16: "SYMBOL",
        0x2B17: "SYMBOL",
        0x2B18: "SYMBOL",
        0x2B19: "SYMBOL",
        0x2B1A: "SYMBOL",
        0x2B1B: "SYMBOL",
        0x2B1C: "SYMBOL",
        0x2B1D: "SYMBOL",
        0x2B1E: "SYMBOL",
        0x2B1F: "SYMBOL",
        0x2B20: "SYMBOL",
        0x2B21: "SYMBOL",
        0x2B22: "SYMBOL",
        0x2B23: "SYMBOL",
        0x2B24: "SYMBOL",
        0x2B25: "SYMBOL",
        0x2B26: "SYMBOL",
        0x2B27: "SYMBOL",
        0x2B28: "SYMBOL",
        0x2B29: "SYMBOL",
        0x2B2A: "SYMBOL",
        0x2B2B: "SYMBOL",
        0x2B2C: "SYMBOL",
        0x2B2D: "SYMBOL",
        0x2B2E: "SYMBOL",
        0x2B2F: "SYMBOL",
        0x2B30: "SYMBOL",
        0x2B31: "SYMBOL",
        0x2B32: "SYMBOL",
        0x2B33: "SYMBOL",
        0x2B34: "SYMBOL",
        0x2B35: "SYMBOL",
        0x2B36: "SYMBOL",
        0x2B37: "SYMBOL",
        0x2B38: "SYMBOL",
        0x2B39: "SYMBOL",
        0x2B3A: "SYMBOL",
        0x2B3B: "SYMBOL",
        0x2B3C: "SYMBOL",
        0x2B3D: "SYMBOL",
        0x2B3E: "SYMBOL",
        0x2B3F: "SYMBOL",
        0x2B40: "SYMBOL",
        0x2B41: "SYMBOL",
        0x2B42: "SYMBOL",
        0x2B43: "SYMBOL",
        0x2B44: "SYMBOL",
        0x2B45: "SYMBOL",
        0x2B46: "SYMBOL",
        0x2B47: "SYMBOL",
        0x2B48: "SYMBOL",
        0x2B49: "SYMBOL",
        0x2B4A: "SYMBOL",
        0x2B4B: "SYMBOL",
        0x2B4C: "SYMBOL",
        0x2B4D: "SYMBOL",
        0x2B4E: "SYMBOL",
        0x2B4F: "SYMBOL",
        0x2B50: "SYMBOL",
        0x2B51: "SYMBOL",
        0x2B52: "SYMBOL",
        0x2B53: "SYMBOL",
        0x2B54: "SYMBOL",
        0x2B55: "SYMBOL",
        0x2B56: "SYMBOL",
        0x2B57: "SYMBOL",
        0x2B58: "SYMBOL",
        0x2B59: "SYMBOL",
        0x2B5A: "SYMBOL",
        0x2B5B: "SYMBOL",
        0x2B5C: "SYMBOL",
        0x2B5D: "SYMBOL",
        0x2B5E: "SYMBOL",
        0x2B5F: "SYMBOL",
        0x2B60: "SYMBOL",
        0x2B61: "SYMBOL",
        0x2B62: "SYMBOL",
        0x2B63: "SYMBOL",
        0x2B64: "SYMBOL",
        0x2B65: "SYMBOL",
        0x2B66: "SYMBOL",
        0x2B67: "SYMBOL",
        0x2B68: "SYMBOL",
        0x2B69: "SYMBOL",
        0x2B6A: "SYMBOL",
        0x2B6B: "SYMBOL",
        0x2B6C: "SYMBOL",
        0x2B6D: "SYMBOL",
        0x2B6E: "SYMBOL",
        0x2B6F: "SYMBOL",
        0x2B70: "SYMBOL",
        0x2B71: "SYMBOL",
        0x2B72: "SYMBOL",
        0x2B73: "SYMBOL",
        0x2B74: "SYMBOL",
        0x2B75: "SYMBOL",
        0x2B76: "SYMBOL",
        0x2B77: "SYMBOL",
        0x2B78: "SYMBOL",
        0x2B79: "SYMBOL",
        0x2B7A: "SYMBOL",
        0x2B7B: "SYMBOL",
        0x2B7C: "SYMBOL",
        0x2B7D: "SYMBOL",
        0x2B7E: "SYMBOL",
        0x2B7F: "SYMBOL",
        0x2B80: "SYMBOL",
        0x2B81: "SYMBOL",
        0x2B82: "SYMBOL",
        0x2B83: "SYMBOL",
        0x2B84: "SYMBOL",
        0x2B85: "SYMBOL",
        0x2B86: "SYMBOL",
        0x2B87: "SYMBOL",
        0x2B88: "SYMBOL",
        0x2B89: "SYMBOL",
        0x2B8A: "SYMBOL",
        0x2B8B: "SYMBOL",
        0x2B8C: "SYMBOL",
        0x2B8D: "SYMBOL",
        0x2B8E: "SYMBOL",
        0x2B8F: "SYMBOL",
        0x2B90: "SYMBOL",
        0x2B91: "SYMBOL",
        0x2B92: "SYMBOL",
        0x2B93: "SYMBOL",
        0x2B94: "SYMBOL",
        0x2B95: "SYMBOL",
        0x2B96: "SYMBOL",
        0x2B97: "SYMBOL",
        0x2B98: "SYMBOL",
        0x2B99: "SYMBOL",
        0x2B9A: "SYMBOL",
        0x2B9B: "SYMBOL",
        0x2B9C: "SYMBOL",
        0x2B9D: "SYMBOL",
        0x2B9E: "SYMBOL",
        0x2B9F: "SYMBOL",
        0x2BA0: "SYMBOL",
        0x2BA1: "SYMBOL",
        0x2BA2: "SYMBOL",
        0x2BA3: "SYMBOL",
        0x2BA4: "SYMBOL",
        0x2BA5: "SYMBOL",
        0x2BA6: "SYMBOL",
        0x2BA7: "SYMBOL",
        0x2BA8: "SYMBOL",
        0x2BA9: "SYMBOL",
        0x2BAA: "SYMBOL",
        0x2BAB: "SYMBOL",
        0x2BAC: "SYMBOL",
        0x2BAD: "SYMBOL",
        0x2BAE: "SYMBOL",
        0x2BAF: "SYMBOL",
        0x2BB0: "SYMBOL",
        0x2BB1: "SYMBOL",
        0x2BB2: "SYMBOL",
        0x2BB3: "SYMBOL",
        0x2BB4: "SYMBOL",
        0x2BB5: "SYMBOL",
        0x2BB6: "SYMBOL",
        0x2BB7: "SYMBOL",
        0x2BB8: "SYMBOL",
        0x2BB9: "SYMBOL",
        0x2BBA: "SYMBOL",
        0x2BBB: "SYMBOL",
        0x2BBC: "SYMBOL",
        0x2BBD: "SYMBOL",
        0x2BBE: "SYMBOL",
        0x2BBF: "SYMBOL",
        0x2BC0: "SYMBOL",
        0x2BC1: "SYMBOL",
        0x2BC2: "SYMBOL",
        0x2BC3: "SYMBOL",
        0x2BC4: "SYMBOL",
        0x2BC5: "SYMBOL",
        0x2BC6: "SYMBOL",
        0x2BC7: "SYMBOL",
        0x2BC8: "SYMBOL",
        0x2BC9: "SYMBOL",
        0x2BCA: "SYMBOL",
        0x2BCB: "SYMBOL",
        0x2BCC: "SYMBOL",
        0x2BCD: "SYMBOL",
        0x2BCE: "SYMBOL",
        0x2BCF: "SYMBOL",
        0x2BD0: "SYMBOL",
        0x2BD1: "SYMBOL",
        0x2BD2: "SYMBOL",
        0x2BD3: "SYMBOL",
        0x2BD4: "SYMBOL",
        0x2BD5: "SYMBOL",
        0x2BD6: "SYMBOL",
        0x2BD7: "SYMBOL",
        0x2BD8: "SYMBOL",
        0x2BD9: "SYMBOL",
        0x2BDA: "SYMBOL",
        0x2BDB: "SYMBOL",
        0x2BDC: "SYMBOL",
        0x2BDD: "SYMBOL",
        0x2BDE: "SYMBOL",
        0x2BDF: "SYMBOL",
        0x2BE0: "SYMBOL",
        0x2BE1: "SYMBOL",
        0x2BE2: "SYMBOL",
        0x2BE3: "SYMBOL",
        0x2BE4: "SYMBOL",
        0x2BE5: "SYMBOL",
        0x2BE6: "SYMBOL",
        0x2BE7: "SYMBOL",
        0x2BE8: "SYMBOL",
        0x2BE9: "SYMBOL",
        0x2BEA: "SYMBOL",
        0x2BEB: "SYMBOL",
        0x2BEC: "SYMBOL",
        0x2BED: "SYMBOL",
        0x2BEE: "SYMBOL",
        0x2BEF: "SYMBOL",
        0x2BF0: "SYMBOL",
        0x2BF1: "SYMBOL",
        0x2BF2: "SYMBOL",
        0x2BF3: "SYMBOL",
        0x2BF4: "SYMBOL",
        0x2BF5: "SYMBOL",
        0x2BF6: "SYMBOL",
        0x2BF7: "SYMBOL",
        0x2BF8: "SYMBOL",
        0x2BF9: "SYMBOL",
        0x2BFA: "SYMBOL",
        0x2BFB: "SYMBOL",
        0x2BFC: "SYMBOL",
        0x2BFD: "SYMBOL",
        0x2BFE: "SYMBOL",
        0x2BFF: "SYMBOL",
        0x2A00: "SYMBOL",
        0x2A01: "SYMBOL",
        0x2A02: "SYMBOL",
        0x2A03: "SYMBOL",
        0x2A04: "SYMBOL",
        0x2A05: "SYMBOL",
        0x2A06: "SYMBOL",
        0x2A07: "SYMBOL",
        0x2A08: "SYMBOL",
        0x2A09: "SYMBOL",
        0x2A0A: "SYMBOL",
        0x2A0B: "SYMBOL",
        0x2A0C: "SYMBOL",
        0x2A0D: "SYMBOL",
        0x2A0E: "SYMBOL",
        0x2A0F: "SYMBOL",
        0x2A10: "SYMBOL",
        0x2A11: "SYMBOL",
        0x2A12: "SYMBOL",
        0x2A13: "SYMBOL",
        0x2A14: "SYMBOL",
        0x2A15: "SYMBOL",
        0x2A16: "SYMBOL",
        0x2A17: "SYMBOL",
        0x2A18: "SYMBOL",
        0x2A19: "SYMBOL",
        0x2A1A: "SYMBOL",
        0x2A1B: "SYMBOL",
        0x2A1C: "SYMBOL",
        0x2A1D: "SYMBOL",
        0x2A1E: "SYMBOL",
        0x2A1F: "SYMBOL",
        0x2A20: "SYMBOL",
        0x2A21: "SYMBOL",
        0x2A22: "SYMBOL",
        0x2A23: "SYMBOL",
        0x2A24: "SYMBOL",
        0x2A25: "SYMBOL",
        0x2A26: "SYMBOL",
        0x2A27: "SYMBOL",
        0x2A28: "SYMBOL",
        0x2A29: "SYMBOL",
        0x2A2A: "SYMBOL",
        0x2A2B: "SYMBOL",
        0x2A2C: "SYMBOL",
        0x2A2D: "SYMBOL",
        0x2A2E: "SYMBOL",
        0x2A2F: "SYMBOL",
        0x2A30: "SYMBOL",
        0x2A31: "SYMBOL",
        0x2A32: "SYMBOL",
        0x2A33: "SYMBOL",
        0x2A34: "SYMBOL",
        0x2A35: "SYMBOL",
        0x2A36: "SYMBOL",
        0x2A37: "SYMBOL",
        0x2A38: "SYMBOL",
        0x2A39: "SYMBOL",
        0x2A3A: "SYMBOL",
        0x2A3B: "SYMBOL",
        0x2A3C: "SYMBOL",
        0x2A3D: "SYMBOL",
        0x2A3E: "SYMBOL",
        0x2A3F: "SYMBOL",
        0x2A40: "SYMBOL",
        0x2A41: "SYMBOL",
        0x2A42: "SYMBOL",
        0x2A43: "SYMBOL",
        0x2A44: "SYMBOL",
        0x2A45: "SYMBOL",
        0x2A46: "SYMBOL",
        0x2A47: "SYMBOL",
        0x2A48: "SYMBOL",
        0x2A49: "SYMBOL",
        0x2A4A: "SYMBOL",
        0x2A4B: "SYMBOL",
        0x2A4C: "SYMBOL",
        0x2A4D: "SYMBOL",
        0x2A4E: "SYMBOL",
        0x2A4F: "SYMBOL",
        0x2A50: "SYMBOL",
        0x2A51: "SYMBOL",
        0x2A52: "SYMBOL",
        0x2A53: "SYMBOL",
        0x2A54: "SYMBOL",
        0x2A55: "SYMBOL",
        0x2A56: "SYMBOL",
        0x2A57: "SYMBOL",
        0x2A58: "SYMBOL",
        0x2A59: "SYMBOL",
        0x2A5A: "SYMBOL",
        0x2A5B: "SYMBOL",
        0x2A5C: "SYMBOL",
        0x2A5D: "SYMBOL",
        0x2A5E: "SYMBOL",
        0x2A5F: "SYMBOL",
        0x2A60: "SYMBOL",
        0x2A61: "SYMBOL",
        0x2A62: "SYMBOL",
        0x2A63: "SYMBOL",
        0x2A64: "SYMBOL",
        0x2A65: "SYMBOL",
        0x2A66: "SYMBOL",
        0x2A67: "SYMBOL",
        0x2A68: "SYMBOL",
        0x2A69: "SYMBOL",
        0x2A6A: "SYMBOL",
        0x2A6B: "SYMBOL",
        0x2A6C: "SYMBOL",
        0x2A6D: "SYMBOL",
        0x2A6E: "SYMBOL",
        0x2A6F: "SYMBOL",
        0x2A70: "SYMBOL",
        0x2A71: "SYMBOL",
        0x2A72: "SYMBOL",
        0x2A73: "SYMBOL",
        0x2A74: "SYMBOL",
        0x2A75: "SYMBOL",
        0x2A76: "SYMBOL",
        0x2A77: "SYMBOL",
        0x2A78: "SYMBOL",
        0x2A79: "SYMBOL",
        0x2A7A: "SYMBOL",
        0x2A7B: "SYMBOL",
        0x2A7C: "SYMBOL",
        0x2A7D: "SYMBOL",
        0x2A7E: "SYMBOL",
        0x2A7F: "SYMBOL",
        0x2A80: "SYMBOL",
        0x2A81: "SYMBOL",
        0x2A82: "SYMBOL",
        0x2A83: "SYMBOL",
        0x2A84: "SYMBOL",
        0x2A85: "SYMBOL",
        0x2A86: "SYMBOL",
        0x2A87: "SYMBOL",
        0x2A88: "SYMBOL",
        0x2A89: "SYMBOL",
        0x2A8A: "SYMBOL",
        0x2A8B: "SYMBOL",
        0x2A8C: "SYMBOL",
        0x2A8D: "SYMBOL",
        0x2A8E: "SYMBOL",
        0x2A8F: "SYMBOL",
        0x2A90: "SYMBOL",
        0x2A91: "SYMBOL",
        0x2A92: "SYMBOL",
        0x2A93: "SYMBOL",
        0x2A94: "SYMBOL",
        0x2A95: "SYMBOL",
        0x2A96: "SYMBOL",
        0x2A97: "SYMBOL",
        0x2A98: "SYMBOL",
        0x2A99: "SYMBOL",
        0x2A9A: "SYMBOL",
        0x2A9B: "SYMBOL",
        0x2A9C: "SYMBOL",
        0x2A9D: "SYMBOL",
        0x2A9E: "SYMBOL",
        0x2A9F: "SYMBOL",
        0x2AA0: "SYMBOL",
        0x2AA1: "SYMBOL",
        0x2AA2: "SYMBOL",
        0x2AA3: "SYMBOL",
        0x2AA4: "SYMBOL",
        0x2AA5: "SYMBOL",
        0x2AA6: "SYMBOL",
        0x2AA7: "SYMBOL",
        0x2AA8: "SYMBOL",
        0x2AA9: "SYMBOL",
        0x2AAA: "SYMBOL",
        0x2AAB: "SYMBOL",
        0x2AAC: "SYMBOL",
        0x2AAD: "SYMBOL",
        0x2AAE: "SYMBOL",
        0x2AAF: "SYMBOL",
        0x2AB0: "SYMBOL",
        0x2AB1: "SYMBOL",
        0x2AB2: "SYMBOL",
        0x2AB3: "SYMBOL",
        0x2AB4: "SYMBOL",
        0x2AB5: "SYMBOL",
        0x2AB6: "SYMBOL",
        0x2AB7: "SYMBOL",
        0x2AB8: "SYMBOL",
        0x2AB9: "SYMBOL",
        0x2ABA: "SYMBOL",
        0x2ABB: "SYMBOL",
        0x2ABC: "SYMBOL",
        0x2ABD: "SYMBOL",
        0x2ABE: "SYMBOL",
        0x2ABF: "SYMBOL",
        0x2AC0: "SYMBOL",
        0x2AC1: "SYMBOL",
        0x2AC2: "SYMBOL",
        0x2AC3: "SYMBOL",
        0x2AC4: "SYMBOL",
        0x2AC5: "SYMBOL",
        0x2AC6: "SYMBOL",
        0x2AC7: "SYMBOL",
        0x2AC8: "SYMBOL",
        0x2AC9: "SYMBOL",
        0x2ACA: "SYMBOL",
        0x2ACB: "SYMBOL",
        0x2ACC: "SYMBOL",
        0x2ACD: "SYMBOL",
        0x2ACE: "SYMBOL",
        0x2ACF: "SYMBOL",
        0x2AD0: "SYMBOL",
        0x2AD1: "SYMBOL",
        0x2AD2: "SYMBOL",
        0x2AD3: "SYMBOL",
        0x2AD4: "SYMBOL",
        0x2AD5: "SYMBOL",
        0x2AD6: "SYMBOL",
        0x2AD7: "SYMBOL",
        0x2AD8: "SYMBOL",
        0x2AD9: "SYMBOL",
        0x2ADA: "SYMBOL",
        0x2ADB: "SYMBOL",
        0x2ADC: "SYMBOL",
        0x2ADD: "SYMBOL",
        0x2ADE: "SYMBOL",
        0x2ADF: "SYMBOL",
        0x2AE0: "SYMBOL",
        0x2AE1: "SYMBOL",
        0x2AE2: "SYMBOL",
        0x2AE3: "SYMBOL",
        0x2AE4: "SYMBOL",
        0x2AE5: "SYMBOL",
        0x2AE6: "SYMBOL",
        0x2AE7: "SYMBOL",
        0x2AE8: "SYMBOL",
        0x2AE9: "SYMBOL",
        0x2AEA: "SYMBOL",
        0x2AEB: "SYMBOL",
        0x2AEC: "SYMBOL",
        0x2AED: "SYMBOL",
        0x2AEE: "SYMBOL",
        0x2AEF: "SYMBOL",
        0x2AF0: "SYMBOL",
        0x2AF1: "SYMBOL",
        0x2AF2: "SYMBOL",
        0x2AF3: "SYMBOL",
        0x2AF4: "SYMBOL",
        0x2AF5: "SYMBOL",
        0x2AF6: "SYMBOL",
        0x2AF7: "SYMBOL",
        0x2AF8: "SYMBOL",
        0x2AF9: "SYMBOL",
        0x2AFA: "SYMBOL",
        0x2AFB: "SYMBOL",
        0x2AFC: "SYMBOL",
        0x2AFD: "SYMBOL",
        0x2AFE: "SYMBOL",
        0x2AFF: "SYMBOL",
        0x3300: "SYMBOL",
        0x3301: "SYMBOL",
        0x3302: "SYMBOL",
        0x3303: "SYMBOL",
        0x3304: "SYMBOL",
        0x3305: "SYMBOL",
        0x3306: "SYMBOL",
        0x3307: "SYMBOL",
        0x3308: "SYMBOL",
        0x3309: "SYMBOL",
        0x330A: "SYMBOL",
        0x330B: "SYMBOL",
        0x330C: "SYMBOL",
        0x330D: "SYMBOL",
        0x330E: "SYMBOL",
        0x330F: "SYMBOL",
        0x3310: "SYMBOL",
        0x3311: "SYMBOL",
        0x3312: "SYMBOL",
        0x3313: "SYMBOL",
        0x3314: "SYMBOL",
        0x3315: "SYMBOL",
        0x3316: "SYMBOL",
        0x3317: "SYMBOL",
        0x3318: "SYMBOL",
        0x3319: "SYMBOL",
        0x331A: "SYMBOL",
        0x331B: "SYMBOL",
        0x331C: "SYMBOL",
        0x331D: "SYMBOL",
        0x331E: "SYMBOL",
        0x331F: "SYMBOL",
        0x3320: "SYMBOL",
        0x3321: "SYMBOL",
        0x3322: "SYMBOL",
        0x3323: "SYMBOL",
        0x3324: "SYMBOL",
        0x3325: "SYMBOL",
        0x3326: "SYMBOL",
        0x3327: "SYMBOL",
        0x3328: "SYMBOL",
        0x3329: "SYMBOL",
        0x332A: "SYMBOL",
        0x332B: "SYMBOL",
        0x332C: "SYMBOL",
        0x332D: "SYMBOL",
        0x332E: "SYMBOL",
        0x332F: "SYMBOL",
        0x3330: "SYMBOL",
        0x3331: "SYMBOL",
        0x3332: "SYMBOL",
        0x3333: "SYMBOL",
        0x3334: "SYMBOL",
        0x3335: "SYMBOL",
        0x3336: "SYMBOL",
        0x3337: "SYMBOL",
        0x3338: "SYMBOL",
        0x3339: "SYMBOL",
        0x333A: "SYMBOL",
        0x333B: "SYMBOL",
        0x333C: "SYMBOL",
        0x333D: "SYMBOL",
        0x333E: "SYMBOL",
        0x333F: "SYMBOL",
        0x3340: "SYMBOL",
        0x3341: "SYMBOL",
        0x3342: "SYMBOL",
        0x3343: "SYMBOL",
        0x3344: "SYMBOL",
        0x3345: "SYMBOL",
        0x3346: "SYMBOL",
        0x3347: "SYMBOL",
        0x3348: "SYMBOL",
        0x3349: "SYMBOL",
        0x334A: "SYMBOL",
        0x334B: "SYMBOL",
        0x334C: "SYMBOL",
        0x334D: "SYMBOL",
        0x334E: "SYMBOL",
        0x334F: "SYMBOL",
        0x3350: "SYMBOL",
        0x3351: "SYMBOL",
        0x3352: "SYMBOL",
        0x3353: "SYMBOL",
        0x3354: "SYMBOL",
        0x3355: "SYMBOL",
        0x3356: "SYMBOL",
        0x3357: "SYMBOL",
        0x3358: "SYMBOL",
        0x3359: "SYMBOL",
        0x335A: "SYMBOL",
        0x335B: "SYMBOL",
        0x335C: "SYMBOL",
        0x335D: "SYMBOL",
        0x335E: "SYMBOL",
        0x335F: "SYMBOL",
        0x3360: "SYMBOL",
        0x3361: "SYMBOL",
        0x3362: "SYMBOL",
        0x3363: "SYMBOL",
        0x3364: "SYMBOL",
        0x3365: "SYMBOL",
        0x3366: "SYMBOL",
        0x3367: "SYMBOL",
        0x3368: "SYMBOL",
        0x3369: "SYMBOL",
        0x336A: "SYMBOL",
        0x336B: "SYMBOL",
        0x336C: "SYMBOL",
        0x336D: "SYMBOL",
        0x336E: "SYMBOL",
        0x336F: "SYMBOL",
        0x3370: "SYMBOL",
        0x3371: "SYMBOL",
        0x3372: "SYMBOL",
        0x3373: "SYMBOL",
        0x3374: "SYMBOL",
        0x3375: "SYMBOL",
        0x3376: "SYMBOL",
        0x3377: "SYMBOL",
        0x3378: "SYMBOL",
        0x3379: "SYMBOL",
        0x337A: "SYMBOL",
        0x337B: "SYMBOL",
        0x337C: "SYMBOL",
        0x337D: "SYMBOL",
        0x337E: "SYMBOL",
        0x337F: "SYMBOL",
        0x3380: "SYMBOL",
        0x3381: "SYMBOL",
        0x3382: "SYMBOL",
        0x3383: "SYMBOL",
        0x3384: "SYMBOL",
        0x3385: "SYMBOL",
        0x3386: "SYMBOL",
        0x3387: "SYMBOL",
        0x3388: "SYMBOL",
        0x3389: "SYMBOL",
        0x338A: "SYMBOL",
        0x338B: "SYMBOL",
        0x338C: "SYMBOL",
        0x338D: "SYMBOL",
        0x338E: "SYMBOL",
        0x338F: "SYMBOL",
        0x3390: "SYMBOL",
        0x3391: "SYMBOL",
        0x3392: "SYMBOL",
        0x3393: "SYMBOL",
        0x3394: "SYMBOL",
        0x3395: "SYMBOL",
        0x3396: "SYMBOL",
        0x3397: "SYMBOL",
        0x3398: "SYMBOL",
        0x3399: "SYMBOL",
        0x339A: "SYMBOL",
        0x339B: "SYMBOL",
        0x339C: "SYMBOL",
        0x339D: "SYMBOL",
        0x339E: "SYMBOL",
        0x339F: "SYMBOL",
        0x33A0: "SYMBOL",
        0x33A1: "SYMBOL",
        0x33A2: "SYMBOL",
        0x33A3: "SYMBOL",
        0x33A4: "SYMBOL",
        0x33A5: "SYMBOL",
        0x33A6: "SYMBOL",
        0x33A7: "SYMBOL",
        0x33A8: "SYMBOL",
        0x33A9: "SYMBOL",
        0x33AA: "SYMBOL",
        0x33AB: "SYMBOL",
        0x33AC: "SYMBOL",
        0x33AD: "SYMBOL",
        0x33AE: "SYMBOL",
        0x33AF: "SYMBOL",
        0x33B0: "SYMBOL",
        0x33B1: "SYMBOL",
        0x33B2: "SYMBOL",
        0x33B3: "SYMBOL",
        0x33B4: "SYMBOL",
        0x33B5: "SYMBOL",
        0x33B6: "SYMBOL",
        0x33B7: "SYMBOL",
        0x33B8: "SYMBOL",
        0x33B9: "SYMBOL",
        0x33BA: "SYMBOL",
        0x33BB: "SYMBOL",
        0x33BC: "SYMBOL",
        0x33BD: "SYMBOL",
        0x33BE: "SYMBOL",
        0x33BF: "SYMBOL",
        0x33C0: "SYMBOL",
        0x33C1: "SYMBOL",
        0x33C2: "SYMBOL",
        0x33C3: "SYMBOL",
        0x33C4: "SYMBOL",
        0x33C5: "SYMBOL",
        0x33C6: "SYMBOL",
        0x33C7: "SYMBOL",
        0x33C8: "SYMBOL",
        0x33C9: "SYMBOL",
        0x33CA: "SYMBOL",
        0x33CB: "SYMBOL",
        0x33CC: "SYMBOL",
        0x33CD: "SYMBOL",
        0x33CE: "SYMBOL",
        0x33CF: "SYMBOL",
        0x33D0: "SYMBOL",
        0x33D1: "SYMBOL",
        0x33D2: "SYMBOL",
        0x33D3: "SYMBOL",
        0x33D4: "SYMBOL",
        0x33D5: "SYMBOL",
        0x33D6: "SYMBOL",
        0x33D7: "SYMBOL",
        0x33D8: "SYMBOL",
        0x33D9: "SYMBOL",
        0x33DA: "SYMBOL",
        0x33DB: "SYMBOL",
        0x33DC: "SYMBOL",
        0x33DD: "SYMBOL",
        0x33DE: "SYMBOL",
        0x33DF: "SYMBOL",
        0x33E0: "SYMBOL",
        0x33E1: "SYMBOL",
        0x33E2: "SYMBOL",
        0x33E3: "SYMBOL",
        0x33E4: "SYMBOL",
        0x33E5: "SYMBOL",
        0x33E6: "SYMBOL",
        0x33E7: "SYMBOL",
        0x33E8: "SYMBOL",
        0x33E9: "SYMBOL",
        0x33EA: "SYMBOL",
        0x33EB: "SYMBOL",
        0x33EC: "SYMBOL",
        0x33ED: "SYMBOL",
        0x33EE: "SYMBOL",
        0x33EF: "SYMBOL",
        0x33F0: "SYMBOL",
        0x33F1: "SYMBOL",
        0x33F2: "SYMBOL",
        0x33F3: "SYMBOL",
        0x33F4: "SYMBOL",
        0x33F5: "SYMBOL",
        0x33F6: "SYMBOL",
        0x33F7: "SYMBOL",
        0x33F8: "SYMBOL",
        0x33F9: "SYMBOL",
        0x33FA: "SYMBOL",
        0x33FB: "SYMBOL",
        0x33FC: "SYMBOL",
        0x33FD: "SYMBOL",
        0x33FE: "SYMBOL",
        0x33FF: "SYMBOL",
        0x3200: "SYMBOL",
        0x3201: "SYMBOL",
        0x3202: "SYMBOL",
        0x3203: "SYMBOL",
        0x3204: "SYMBOL",
        0x3205: "SYMBOL",
        0x3206: "SYMBOL",
        0x3207: "SYMBOL",
        0x3208: "SYMBOL",
        0x3209: "SYMBOL",
        0x320A: "SYMBOL",
        0x320B: "SYMBOL",
        0x320C: "SYMBOL",
        0x320D: "SYMBOL",
        0x320E: "SYMBOL",
        0x320F: "SYMBOL",
        0x3210: "SYMBOL",
        0x3211: "SYMBOL",
        0x3212: "SYMBOL",
        0x3213: "SYMBOL",
        0x3214: "SYMBOL",
        0x3215: "SYMBOL",
        0x3216: "SYMBOL",
        0x3217: "SYMBOL",
        0x3218: "SYMBOL",
        0x3219: "SYMBOL",
        0x321A: "SYMBOL",
        0x321B: "SYMBOL",
        0x321C: "SYMBOL",
        0x321D: "SYMBOL",
        0x321E: "SYMBOL",
        0x321F: "SYMBOL",
        0x3220: "SYMBOL",
        0x3221: "SYMBOL",
        0x3222: "SYMBOL",
        0x3223: "SYMBOL",
        0x3224: "SYMBOL",
        0x3225: "SYMBOL",
        0x3226: "SYMBOL",
        0x3227: "SYMBOL",
        0x3228: "SYMBOL",
        0x3229: "SYMBOL",
        0x322A: "SYMBOL",
        0x322B: "SYMBOL",
        0x322C: "SYMBOL",
        0x322D: "SYMBOL",
        0x322E: "SYMBOL",
        0x322F: "SYMBOL",
        0x3230: "SYMBOL",
        0x3231: "SYMBOL",
        0x3232: "SYMBOL",
        0x3233: "SYMBOL",
        0x3234: "SYMBOL",
        0x3235: "SYMBOL",
        0x3236: "SYMBOL",
        0x3237: "SYMBOL",
        0x3238: "SYMBOL",
        0x3239: "SYMBOL",
        0x323A: "SYMBOL",
        0x323B: "SYMBOL",
        0x323C: "SYMBOL",
        0x323D: "SYMBOL",
        0x323E: "SYMBOL",
        0x323F: "SYMBOL",
        0x3240: "SYMBOL",
        0x3241: "SYMBOL",
        0x3242: "SYMBOL",
        0x3243: "SYMBOL",
        0x3244: "SYMBOL",
        0x3245: "SYMBOL",
        0x3246: "SYMBOL",
        0x3247: "SYMBOL",
        0x3248: "SYMBOL",
        0x3249: "SYMBOL",
        0x324A: "SYMBOL",
        0x324B: "SYMBOL",
        0x324C: "SYMBOL",
        0x324D: "SYMBOL",
        0x324E: "SYMBOL",
        0x324F: "SYMBOL",
        0x3250: "SYMBOL",
        0x3251: "SYMBOL",
        0x3252: "SYMBOL",
        0x3253: "SYMBOL",
        0x3254: "SYMBOL",
        0x3255: "SYMBOL",
        0x3256: "SYMBOL",
        0x3257: "SYMBOL",
        0x3258: "SYMBOL",
        0x3259: "SYMBOL",
        0x325A: "SYMBOL",
        0x325B: "SYMBOL",
        0x325C: "SYMBOL",
        0x325D: "SYMBOL",
        0x325E: "SYMBOL",
        0x325F: "SYMBOL",
        0x3260: "SYMBOL",
        0x3261: "SYMBOL",
        0x3262: "SYMBOL",
        0x3263: "SYMBOL",
        0x3264: "SYMBOL",
        0x3265: "SYMBOL",
        0x3266: "SYMBOL",
        0x3267: "SYMBOL",
        0x3268: "SYMBOL",
        0x3269: "SYMBOL",
        0x326A: "SYMBOL",
        0x326B: "SYMBOL",
        0x326C: "SYMBOL",
        0x326D: "SYMBOL",
        0x326E: "SYMBOL",
        0x326F: "SYMBOL",
        0x3270: "SYMBOL",
        0x3271: "SYMBOL",
        0x3272: "SYMBOL",
        0x3273: "SYMBOL",
        0x3274: "SYMBOL",
        0x3275: "SYMBOL",
        0x3276: "SYMBOL",
        0x3277: "SYMBOL",
        0x3278: "SYMBOL",
        0x3279: "SYMBOL",
        0x327A: "SYMBOL",
        0x327B: "SYMBOL",
        0x327C: "SYMBOL",
        0x327D: "SYMBOL",
        0x327E: "SYMBOL",
        0x327F: "SYMBOL",
        0x3280: "SYMBOL",
        0x3281: "SYMBOL",
        0x3282: "SYMBOL",
        0x3283: "SYMBOL",
        0x3284: "SYMBOL",
        0x3285: "SYMBOL",
        0x3286: "SYMBOL",
        0x3287: "SYMBOL",
        0x3288: "SYMBOL",
        0x3289: "SYMBOL",
        0x328A: "SYMBOL",
        0x328B: "SYMBOL",
        0x328C: "SYMBOL",
        0x328D: "SYMBOL",
        0x328E: "SYMBOL",
        0x328F: "SYMBOL",
        0x3290: "SYMBOL",
        0x3291: "SYMBOL",
        0x3292: "SYMBOL",
        0x3293: "SYMBOL",
        0x3294: "SYMBOL",
        0x3295: "SYMBOL",
        0x3296: "SYMBOL",
        0x3297: "SYMBOL",
        0x3298: "SYMBOL",
        0x3299: "SYMBOL",
        0x329A: "SYMBOL",
        0x329B: "SYMBOL",
        0x329C: "SYMBOL",
        0x329D: "SYMBOL",
        0x329E: "SYMBOL",
        0x329F: "SYMBOL",
        0x32A0: "SYMBOL",
        0x32A1: "SYMBOL",
        0x32A2: "SYMBOL",
        0x32A3: "SYMBOL",
        0x32A4: "SYMBOL",
        0x32A5: "SYMBOL",
        0x32A6: "SYMBOL",
        0x32A7: "SYMBOL",
        0x32A8: "SYMBOL",
        0x32A9: "SYMBOL",
        0x32AA: "SYMBOL",
        0x32AB: "SYMBOL",
        0x32AC: "SYMBOL",
        0x32AD: "SYMBOL",
        0x32AE: "SYMBOL",
        0x32AF: "SYMBOL",
        0x32B0: "SYMBOL",
        0x32B1: "SYMBOL",
        0x32B2: "SYMBOL",
        0x32B3: "SYMBOL",
        0x32B4: "SYMBOL",
        0x32B5: "SYMBOL",
        0x32B6: "SYMBOL",
        0x32B7: "SYMBOL",
        0x32B8: "SYMBOL",
        0x32B9: "SYMBOL",
        0x32BA: "SYMBOL",
        0x32BB: "SYMBOL",
        0x32BC: "SYMBOL",
        0x32BD: "SYMBOL",
        0x32BE: "SYMBOL",
        0x32BF: "SYMBOL",
        0x32C0: "SYMBOL",
        0x32C1: "SYMBOL",
        0x32C2: "SYMBOL",
        0x32C3: "SYMBOL",
        0x32C4: "SYMBOL",
        0x32C5: "SYMBOL",
        0x32C6: "SYMBOL",
        0x32C7: "SYMBOL",
        0x32C8: "SYMBOL",
        0x32C9: "SYMBOL",
        0x32CA: "SYMBOL",
        0x32CB: "SYMBOL",
        0x32CC: "SYMBOL",
        0x32CD: "SYMBOL",
        0x32CE: "SYMBOL",
        0x32CF: "SYMBOL",
        0x32D0: "SYMBOL",
        0x32D1: "SYMBOL",
        0x32D2: "SYMBOL",
        0x32D3: "SYMBOL",
        0x32D4: "SYMBOL",
        0x32D5: "SYMBOL",
        0x32D6: "SYMBOL",
        0x32D7: "SYMBOL",
        0x32D8: "SYMBOL",
        0x32D9: "SYMBOL",
        0x32DA: "SYMBOL",
        0x32DB: "SYMBOL",
        0x32DC: "SYMBOL",
        0x32DD: "SYMBOL",
        0x32DE: "SYMBOL",
        0x32DF: "SYMBOL",
        0x32E0: "SYMBOL",
        0x32E1: "SYMBOL",
        0x32E2: "SYMBOL",
        0x32E3: "SYMBOL",
        0x32E4: "SYMBOL",
        0x32E5: "SYMBOL",
        0x32E6: "SYMBOL",
        0x32E7: "SYMBOL",
        0x32E8: "SYMBOL",
        0x32E9: "SYMBOL",
        0x32EA: "SYMBOL",
        0x32EB: "SYMBOL",
        0x32EC: "SYMBOL",
        0x32ED: "SYMBOL",
        0x32EE: "SYMBOL",
        0x32EF: "SYMBOL",
        0x32F0: "SYMBOL",
        0x32F1: "SYMBOL",
        0x32F2: "SYMBOL",
        0x32F3: "SYMBOL",
        0x32F4: "SYMBOL",
        0x32F5: "SYMBOL",
        0x32F6: "SYMBOL",
        0x32F7: "SYMBOL",
        0x32F8: "SYMBOL",
        0x32F9: "SYMBOL",
        0x32FA: "SYMBOL",
        0x32FB: "SYMBOL",
        0x32FC: "SYMBOL",
        0x32FD: "SYMBOL",
        0x32FE: "SYMBOL",
        0x3000: "SYMBOL",
        0x3001: "SYMBOL",
        0x3002: "SYMBOL",
        0x3003: "SYMBOL",
        0x3004: "SYMBOL",
        0x3005: "SYMBOL",
        0x3006: "SYMBOL",
        0x3007: "SYMBOL",
        0x3008: "SYMBOL",
        0x3009: "SYMBOL",
        0x300A: "SYMBOL",
        0x300B: "SYMBOL",
        0x300C: "SYMBOL",
        0x300D: "SYMBOL",
        0x300E: "SYMBOL",
        0x300F: "SYMBOL",
        0x3010: "SYMBOL",
        0x3011: "SYMBOL",
        0x3012: "SYMBOL",
        0x3013: "SYMBOL",
        0x3014: "SYMBOL",
        0x3015: "SYMBOL",
        0x3016: "SYMBOL",
        0x3017: "SYMBOL",
        0x3018: "SYMBOL",
        0x3019: "SYMBOL",
        0x301A: "SYMBOL",
        0x301B: "SYMBOL",
        0x301C: "SYMBOL",
        0x301D: "SYMBOL",
        0x301E: "SYMBOL",
        0x301F: "SYMBOL",
        0x3020: "SYMBOL",
        0x3021: "SYMBOL",
        0x3022: "SYMBOL",
        0x3023: "SYMBOL",
        0x3024: "SYMBOL",
        0x3025: "SYMBOL",
        0x3026: "SYMBOL",
        0x3027: "SYMBOL",
        0x3028: "SYMBOL",
        0x3029: "SYMBOL",
        0x302A: "SYMBOL",
        0x302B: "SYMBOL",
        0x302C: "SYMBOL",
        0x302D: "SYMBOL",
        0x302E: "SYMBOL",
        0x302F: "SYMBOL",
        0x3030: "SYMBOL",
        0x3031: "SYMBOL",
        0x3032: "SYMBOL",
        0x3033: "SYMBOL",
        0x3034: "SYMBOL",
        0x3035: "SYMBOL",
        0x3036: "SYMBOL",
        0x3037: "SYMBOL",
        0x3038: "SYMBOL",
        0x3039: "SYMBOL",
        0x303A: "SYMBOL",
        0x303B: "SYMBOL",
        0x303C: "SYMBOL",
        0x303D: "SYMBOL",
        0x303E: "SYMBOL",
        0x303F: "SYMBOL",
        0xFE30: "SYMBOL",
        0xFE31: "SYMBOL",
        0xFE32: "SYMBOL",
        0xFE33: "SYMBOL",
        0xFE34: "SYMBOL",
        0xFE35: "SYMBOL",
        0xFE36: "SYMBOL",
        0xFE37: "SYMBOL",
        0xFE38: "SYMBOL",
        0xFE39: "SYMBOL",
        0xFE3A: "SYMBOL",
        0xFE3B: "SYMBOL",
        0xFE3C: "SYMBOL",
        0xFE3D: "SYMBOL",
        0xFE3E: "SYMBOL",
        0xFE3F: "SYMBOL",
        0xFE40: "SYMBOL",
        0xFE41: "SYMBOL",
        0xFE42: "SYMBOL",
        0xFE43: "SYMBOL",
        0xFE44: "SYMBOL",
        0xFE45: "SYMBOL",
        0xFE46: "SYMBOL",
        0xFE47: "SYMBOL",
        0xFE48: "SYMBOL",
        0xFE49: "SYMBOL",
        0xFE4A: "SYMBOL",
        0xFE4B: "SYMBOL",
        0xFE4C: "SYMBOL",
        0xFE4D: "SYMBOL",
        0xFE4E: "SYMBOL",
        0xFE4F: "SYMBOL",
        0xFE50: "SYMBOL",
        0xFE51: "SYMBOL",
        0xFE52: "SYMBOL",
        0xFE53: "SYMBOL",
        0xFE54: "SYMBOL",
        0xFE55: "SYMBOL",
        0xFE56: "SYMBOL",
        0xFE57: "SYMBOL",
        0xFE58: "SYMBOL",
        0xFE59: "SYMBOL",
        0xFE5A: "SYMBOL",
        0xFE5B: "SYMBOL",
        0xFE5C: "SYMBOL",
        0xFE5D: "SYMBOL",
        0xFE5E: "SYMBOL",
        0xFE5F: "SYMBOL",
        0xFE60: "SYMBOL",
        0xFE61: "SYMBOL",
        0xFE62: "SYMBOL",
        0xFE63: "SYMBOL",
        0xFE64: "SYMBOL",
        0xFE65: "SYMBOL",
        0xFE66: "SYMBOL",
        0xFE67: "SYMBOL",
        0xFE68: "SYMBOL",
        0xFE69: "SYMBOL",
        0xFE6A: "SYMBOL",
        0xFE6B: "SYMBOL",
    },
)


def character_category_map(ch):
    if len(get_emoji(ch)) != 0:
        return "EMOJI"
    return categories[ord(ch)]


"""
CATEGORY_NAME: Name of category. you have to define DEFAULT class.
INVOKE: 1/0:   always invoke unknown word processing, evan when the word can be found in the lexicon
GROUP:  1/0:   make a new word by grouping the same chracter category
LENGTH: n:     1 to n length new words are added
"""
invokeMap = {
    "DEFAULT": 0,
    "SPACE": 0,
    "HANJA": 0,
    "KANJI": 0,
    "SYMBOL": 1,
    "NUMERIC": 1,
    "ALPHA": 1,
    "HANGUL": 0,
    "HIRAGANA": 1,
    "KATAKANA": 1,
    "HANJANUMERIC": 1,
    "GREEK": 1,
    "CYRILLIC": 1,
}
groupMap = {
    "DEFAULT": 1,
    "SPACE": 1,
    "HANJA": 0,
    "KANJI": 0,
    "SYMBOL": 1,
    "NUMERIC": 1,
    "ALPHA": 1,
    "HANGUL": 1,
    "HIRAGANA": 1,
    "KATAKANA": 1,
    "HANJANUMERIC": 1,
    "GREEK": 1,
    "CYRILLIC": 1,
}


class CharacterDefinition(object):
    invokeMap = []
    groupMap = []

    @staticmethod
    def getCharacterClass(ch):
        return character_category_map(ch)

    @staticmethod
    def isInvoke(ch):
        return invokeMap[character_category_map(ch)]

    @staticmethod
    def isGroup(ch):
        return groupMap[character_category_map(ch)]

    @staticmethod
    def isHangul(ch):
        if 0xAC00 < ord(ch) < 0xD7A3:  # 한글 음절 범위 (가 ~ 힣)
            return True
        return False

    @staticmethod
    def hasCoda(ch):
        if ((ord(ch) - 0xAC00) % 0x001C) == 0:
            return False  # 종성 없음 (ex. 키)
        else:
            return True  # 종성 있음 (ex. 킥)
